package Absyn "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 absyn.rml
  module:      Absyn
  description: Abstract syntax
 
  RCS: $Id: Absyn.rml 1801 2005-06-03 06:46:30Z x02lucpo $
 
  This file defines the abstract syntax for Modelica in RML.  It mainly
  contains datatypes for constructing the abstract syntax tree
  (AST), functions for building and altering RML datatypes and a few functions 
  for printing the AST.
   
  absyn.rml's constructors are primarily used by the walker 
  (Compiler/absyn_builder/walker.g) which takes an ANTLR internal syntax tree and
  converts it into an RML abstract syntax tree.
 
  When the AST has been built, it is normaly used by explode.rml in order to
  build the scode (See explode.rml). It is also possile to send the AST do 
  the dumper (dump.rml) in order to print it.
  
  For details regarding the abstract syntax tree, check out the grammar in 
  the Modelica language specification.
   
 "

public 
type Ident = String " An identifier, for example a variable name ";

public 
type ClassList = list<Class>;

type PathOption = Option<Path>;

uniontype Program " - Programs, the top level construct 
   A program is simply a list of class definitions declared at top
    level in the source file, combined with a within statement that
    indicates the hieractical position of the program. 
   "
  record PROGRAM
    ClassList classlist "List of classes ";
    Within within_ "Within statement ";
  end PROGRAM;
  record BEGIN_DEFINITION
    Path path "For split definitions";
    Restriction restriction "Class restriction ";
    Boolean partial_ "Partial ";
    Boolean encapsulated_ "Encapsulated ";
  end BEGIN_DEFINITION;
  record END_DEFINITION
    Ident ident "For split definitions ";
  end END_DEFINITION;
  record COMP_DEFINITION
    ElementSpec elementspec "For split definitions";
    PathOption pathoption "insert into. 
						       Default, NONE ";
  end COMP_DEFINITION;
  record IMPORT_DEFINITION
    ElementSpec elementspec "For split definitions";
    PathOption pathoption "insert into. 
						       Default, NONE ";
  end IMPORT_DEFINITION;
end Program;

public 
uniontype Within " Within statements "
  record WITHIN
    Path path;
  end WITHIN;
  record TOP
  end TOP;
end Within;

public 
uniontype Class " - Classes 
   A class definition consists of a name, a flag to indicate if this 
   class is declared as `partial', the declared class restriction, 
   and the body of the declaration. "
  record CLASS
    Ident name "Name ";
    Boolean partial_ "Partial ";
    Boolean final_ "Final ";
    Boolean encapsulated_ "Encapsulated ";
    Restriction restricion "Restricion ";
    ClassDef body "Body ";
    String string "File name class is defined in";
  end CLASS;
end Class;

public 
type ClassPartList = list<ClassPart>;

type StringOption = Option<String>;

type ArrayDimOption = Option<ArrayDim>;

type ElementArgList = list<ElementArg>;

type CommentOption = Option<Comment>;

type EnumLiteralList = list<EnumLiteral>;

type PathList = list<Path>;

uniontype ClassDef " The `ClassDef' type contains the definition part of a class 
   declaration.  The definition is either explicit, with a list of 
   parts (`public', `protected', `equationc' and `algorithm'), or it 
   is a definition derived from another class or an enumeration type. 
   For a derived type, the  type contains the name of the derived class and an optional 
   array dimension and a list of modifications. An enumeration type contains a list of "
  record PARTS
    ClassPartList classpartlist;
    StringOption stringoption "string comment ";
  end PARTS;
  record DERIVED
    Path path;
    ArrayDimOption  arrayDimOption "";
    ElementAttributes elementattributes;
    ElementArgList elementarglist;
    CommentOption comment "comment ";
  end DERIVED;
  record ENUMERATION
    EnumLiteralList enumliterallist;
    CommentOption comment "comment";
  end ENUMERATION;
  record OVERLOAD
    PathList pathlist "function names ";
    CommentOption commentoption;
  end OVERLOAD;
end ClassDef;

public 
uniontype EnumLiteral " EnumLiteral, which is a name in an enumeration and an optional
   Comment."
  record ENUMLITERAL
    Ident literal "Literal ";
    CommentOption comment "comment ";
  end ENUMLITERAL;
end EnumLiteral;

public 
type ElementItemList = list<ElementItem>;

type EquationItemList = list<EquationItem>;

type AlgorithmItemList = list<AlgorithmItem>;

type AnnotationOption = Option<Annotation>;

uniontype ClassPart " A class definition contains several parts.  There are public and 
   protected component declarations, type definitions and `extends' 
   clauses, collectively called elements.  There are also equation 
   sections and algorithm sections. The EXTERNAL part is used only by functions 
   which can be declared as external C or FORTRAN functions. "
  record PUBLIC
    ElementItemList elementitemlist;
  end PUBLIC;
  record PROTECTED
    ElementItemList elementitemlist;
  end PROTECTED;
  record EQUATIONS
    EquationItemList equationitemlist;
  end EQUATIONS;
  record INITIALEQUATIONS
    EquationItemList equationitemlist;
  end INITIALEQUATIONS;
  record ALGORITHMS
    AlgorithmItemList algorithmitemlist;
  end ALGORITHMS;
  record INITIALALGORITHMS
    AlgorithmItemList algorithmitemlist;
  end INITIALALGORITHMS;
  record EXTERNAL
    ExternalDecl externaldecl;
    AnnotationOption annotationoption;
  end EXTERNAL;
end ClassPart;

public 
uniontype ElementItem " An element item is either an element or an annotation "
  record ELEMENTITEM
    Element element;
  end ELEMENTITEM;
  record ANNOTATIONITEM
    Annotation annotation_;
  end ANNOTATIONITEM;
end ElementItem;

public 
type ConstrainClassOption = Option<ConstrainClass>;

uniontype Element " - Elements 
  The basic element type in Modelica "
  record ELEMENT
    Boolean final_ "final ";
    Boolean replaceable_ "replaceable ";
    InnerOuter inner_outer "inner/outer ";
    Ident ident "Element name ";
    ElementSpec elementspec "Actual element specification";
    String string "Source code file ";
    Integer integer "Line number ";
    ConstrainClassOption constrainclassoption "only valid for classdef and component";
  end ELEMENT;
end Element;

public 
type ConstrainClass = ElementSpec "Constraining type, must be extendes ";

public 
type ComponentItemList = list<ComponentItem>;

uniontype ElementSpec " An element is something that occurs in a public or protected
    section in a class definition.  There is one constructor in the
    `ElementSpec' type for each possible element type.  There are
    class definitions (`CLASSDEF'), `extends' clauses (`EXTENDS')
    and component declarations (`COMPONENTS').
    
    As an example, if the element `extends TwoPin;' appears
    in the source, it is represented in the AST as
    `EXTENDS(IDENT(\"TwoPin\"),{})'.
   "
  record CLASSDEF
    Boolean replaceable_ "replaceable ";
    Class class_;
  end CLASSDEF;
  record EXTENDS
    Path path;
    ElementArgList elementarglist;
  end EXTENDS;
  record IMPORT
    Import import_;
    CommentOption commentoption;
  end IMPORT;
  record COMPONENTS
    ElementAttributes elementattributes "1.1 contains Araydim also.";
    Path path "type name ";
    ComponentItemList componentitemlist " One of the keyword inner and outer CAN be given to reference an inner or
      outer component. Thus there are three disjoint possibilities. *";
  end COMPONENTS;
end ElementSpec;

public 
uniontype InnerOuter "1.1 contains Araydim also."
  record INNER
  end INNER;
  record OUTER
  end OUTER;
  record UNSPECIFIED
  end UNSPECIFIED;
end InnerOuter;

public 
uniontype Import "Import statements, different kinds  "
  record NAMED_IMPORT
    Ident ident;
    Path path;
  end NAMED_IMPORT;
  record QUAL_IMPORT
    Path path;
  end QUAL_IMPORT;
  record UNQUAL_IMPORT
    Path path;
  end UNQUAL_IMPORT;
end Import;

public 
uniontype ComponentItem "Collection of component and an optional comment "
  record COMPONENTITEM
    Component component;
    CommentOption commentoption;
  end COMPONENTITEM;
end ComponentItem;

public 
type ModificationOption = Option<Modification>;

uniontype Component "Some kind of Modelica entity (object or variable) "
  record COMPONENT
    Ident ident "component name ";
    ArrayDim arraydim "Array dimensions, if any ";
    ModificationOption modificationoption "Optional modification ";
  end COMPONENT;
end Component;

public 
uniontype EquationItem " Several component declarations can be grouped together in one 
   `ElementSpec' by writing them on the same line in the source. 
   This type contains the information specific to one component. "
  record EQUATIONITEM
    Equation equation_;
    CommentOption commentoption;
  end EQUATIONITEM;
  record EQUATIONITEMANN
    Annotation annotation_;
  end EQUATIONITEMANN;
end EquationItem;

public 
uniontype AlgorithmItem " Info specific for an algorithm item. "
  record ALGORITHMITEM
    Algorithm algorithm_;
    CommentOption commentoption;
  end ALGORITHMITEM;
  record ALGORITHMITEMANN
    Annotation annotation_;
  end ALGORITHMITEMANN;
end AlgorithmItem;

public 
type TupleExpEquationItemList = tuple<Exp,EquationItemList>;

type TupleExpEquationItemListList = list<TupleExpEquationItemList>;

uniontype Equation "Information on one (kind) of equation, different constructors for different
     kinds of equations "
  record EQ_IF
    Exp exp1 "Conditional expression ";
    EquationItemList equationitemlist2 "true branch ";
    TupleExpEquationItemListList tupleexpequationitemlistlist3 "elseif branches ";
    EquationItemList equationitemlist4 "else branch ";
  end EQ_IF;
  record EQ_EQUALS
    Exp exp1;
    Exp exp2 "Standard 2-side eqn";
  end EQ_EQUALS;
  record EQ_CONNECT
    ComponentRef componentref1;
    ComponentRef componentref2 "Connect stmt ";
  end EQ_CONNECT;
  record EQ_FOR
    Ident ident;
    Exp exp;
    EquationItemList for_loops "For-loops ";
  end EQ_FOR;
  record EQ_WHEN_E
    Exp condition "Condition ";
    EquationItemList equationitemlist "Loop body ";
    TupleExpEquationItemListList tupleexpequationitemlistlist "else when ";
  end EQ_WHEN_E;
  record EQ_NORETCALL
    Ident ident;
    FunctionArgs functionargs "fcalls without return value ";
  end EQ_NORETCALL;
end Equation;

public 
type TupleExpAlgorithmItemList = tuple<Exp,AlgorithmItemList>;

type TupleExpAlgorithmItemListList = list<TupleExpAlgorithmItemList>;

uniontype Algorithm " The `Algorithm' type describes one algorithm statement in an 
   algorithm section.  It does not describe a whole algorithm.  The 
   reason this type is named like this is that the name of the 
   grammar rule for algorithm statements is `algorithm'. "
  record ALG_ASSIGN
    ComponentRef componentref;
    Exp exp;
  end ALG_ASSIGN;
  record ALG_TUPLE_ASSIGN
    Exp tuple_ "tuple";
    Exp value "value";
  end ALG_TUPLE_ASSIGN;
  record ALG_IF
    Exp exp1;
    AlgorithmItemList algorithmitemlist2 "true branch ";
    TupleExpAlgorithmItemListList elseif3 "elseif ";
    AlgorithmItemList algorithmitemlist4 "else branch ";
  end ALG_IF;
  record ALG_FOR
    Ident ident;
    Exp exp;
    AlgorithmItemList algorithmitemlist;
  end ALG_FOR;
  record ALG_WHILE
    Exp exp;
    AlgorithmItemList algorithmitemlist;
  end ALG_WHILE;
  record ALG_WHEN_A
    Exp exp;
    AlgorithmItemList algorithmitemlist;
    TupleExpAlgorithmItemListList elsewhen_ "elsewhen ";
  end ALG_WHEN_A;
  record ALG_NORETCALL
    ComponentRef componentref;
    FunctionArgs functionargs "general fcalls without return value ";
  end ALG_NORETCALL;
end Algorithm;

public 
type ExpOption = Option<Exp>;

uniontype Modification " - Modifications 
   Modifications are described by the `Modification' type.  There 
   are two forms of modifications: redeclarations and component 
   modifications. "
  record CLASSMOD
    ElementArgList elementarglist;
    ExpOption expoption;
  end CLASSMOD;
end Modification;

public 
uniontype ElementArg "Wrapper for things that modify elements, modifications and redeclarations "
  record MODIFICATION
    Boolean boolean;
    Each each_;
    ComponentRef componentref;
    ModificationOption modificationoption;
    StringOption stringoption;
  end MODIFICATION;
  record REDECLARATION
    Boolean boolean;
    Each each_;
    ElementSpec elementspec;
    ConstrainClassOption constrainclassoption;
  end REDECLARATION;
end ElementArg;

public 
uniontype Each " - Each attribute 
   The each keyword can be present in both MODIFICATION's and REDECLARATION's. "
  record EACH
  end EACH;
  record NON_EACH
  end NON_EACH;
end Each;

public 
uniontype ElementAttributes " - Component attributes "
  record ATTR
    Boolean flow_ "flow ";
    Variability variability "parameter, constant etc. ";
    Direction direction;
    ArrayDim arraydim "1.1";
  end ATTR;
end ElementAttributes;

public 
uniontype Variability "Dete "
  record VAR
  end VAR;
  record DISCRETE
  end DISCRETE;
  record PARAM
  end PARAM;
  record CONST
  end CONST;
end Variability;

public 
uniontype Direction
  record INPUT
  end INPUT;
  record OUTPUT
  end OUTPUT;
  record BIDIR " Component attributes are "
  end BIDIR;
end Direction;

public 
type ArrayDim = list<Subscript> " properties of components which are applied by type prefixes. 
 As an example, declaring a component as `input Real x;' will 
 give the attributes `ATTR({},false,VAR,INPUT)'. 
 - Array dimensions ";

public 
type TupleExpExp = tuple<Exp,Exp>;

type TupleExpExpList = list<TupleExpExp>;

type ExpList = list<Exp>;

type ExpListList = list<ExpList>;

uniontype Exp " Components in Modelica can be scalar or arrays with one or more 
   dimensions. This datatype is used to indicate the dimensionality 
   of a component or a type definition. 
   - Expressions "
  record INTEGER
    Integer integer;
  end INTEGER;
  record REAL
    Real real;
  end REAL;
  record CREF
    ComponentRef componentref;
  end CREF;
  record STRING
    String string;
  end STRING;
  record BOOL
    Boolean boolean;
  end BOOL;
  record BINARY
    Exp exp1;
    Operator operator2;
    Exp exp3 "Binary operations, e.g. a*b ";
  end BINARY;
  record UNARY
    Operator operator;
    Exp exp "Unary operations, e.g. -(x) ";
  end UNARY;
  record LBINARY
    Exp exp1;
    Operator operator2;
    Exp exp3 "Logical binary operations: and, or ";
  end LBINARY;
  record LUNARY
    Operator operator;
    Exp exp "Logical unary operations: not ";
  end LUNARY;
  record RELATION
    Exp exp1;
    Operator operator2;
    Exp exp3 "Relations, e.g. a >= 0 ";
  end RELATION;
  record IFEXP
    Exp exp1;
    Exp exp2;
    Exp exp3;
    TupleExpExpList tupleexpexplist4 "If expressions ";
  end IFEXP;
  record CALL
    ComponentRef componentref;
    FunctionArgs functionargs "Function calls ";
  end CALL;
  record ARRAY
    ExpList explist "Array construction using '{','}' or 'array' ";
  end ARRAY;
  record MATRIX
    ExpListList explistlist "Matrix construction using '{', '}' ";
  end MATRIX;
  record RANGE
    Exp exp1;
    ExpOption expoption2;
    Exp exp3 "Range expressions, e.g. 1:10 or 1:0.5:10 ";
  end RANGE;
  record TUPLE
    ExpList explist "Tuples used in function calls returning several values ";
  end TUPLE;
  record END "array access operator for last element, e.g. a{end}:=1; "
  end END;
  record CODE
    Code code "Modelica AST Code constructors ";
  end CODE;
end Exp;

public 
uniontype Code " The `Exp' datatype is the container of a Modelica expression. "
  record C_TYPENAME
    Path path;
  end C_TYPENAME;
  record C_VARIABLENAME
    ComponentRef componentref;
  end C_VARIABLENAME;
  record C_EQUATIONSECTION
    Boolean boolean;
    EquationItemList equationitemlist;
  end C_EQUATIONSECTION;
  record C_ALGORITHMSECTION
    Boolean boolean;
    AlgorithmItemList algorithmitemlist;
  end C_ALGORITHMSECTION;
  record C_ELEMENT
    Element element;
  end C_ELEMENT;
  record C_EXPRESSION
    Exp exp;
  end C_EXPRESSION;
  record C_MODIFICATION
    Modification modification " The 'Code' datatype is used for Meta-programming. It orgiginates from the Code quotation. ";
  end C_MODIFICATION;
end Code;

public 
type NamedArgList = list<NamedArg>;

uniontype FunctionArgs
  record FUNCTIONARGS
    ExpList explist;
    NamedArgList namedarglist;
  end FUNCTIONARGS;
  record FOR_ITER_FARG
    Exp exp1;
    Ident ident2;
    Exp exp3 " The `FunctionArgs' datatype consists of a list of positional arguments ";
  end FOR_ITER_FARG;
end FunctionArgs;

public 
uniontype NamedArg " followed by a list of named arguments (Modelica v2.0) "
  record NAMEDARG
    Ident ident;
    Exp exp " The `NamedArg' datatype consist of an Identifier for the argument and an expression ";
  end NAMEDARG;
end NamedArg;

public 
uniontype Operator " giving the value of the argument "
  record ADD
  end ADD;
  record SUB
  end SUB;
  record MUL
  end MUL;
  record DIV
  end DIV;
  record POW
  end POW;
  record UPLUS
  end UPLUS;
  record UMINUS
  end UMINUS;
  record AND
  end AND;
  record OR
  end OR;
  record NOT
  end NOT;
  record LESS
  end LESS;
  record LESSEQ
  end LESSEQ;
  record GREATER
  end GREATER;
  record GREATEREQ
  end GREATEREQ;
  record EQUAL
  end EQUAL;
  record NEQUAL
  end NEQUAL;
end Operator;

public 
uniontype Subscript " - Subscripts "
  record NOSUB
  end NOSUB;
  record SUBSCRIPT
    Exp exp " The `Subscript' datatype is used both in array declarations and ";
  end SUBSCRIPT;
end Subscript;

public 
type SubscriptList = list<Subscript>;

uniontype ComponentRef " component references.  This might seem strange, but it is 
   inherited from the grammar.  The `NOSUB' constructor means that 
   the dimension size is undefined when used in a declaration, and 
   when it is used in a component reference it means a slice of the 
   whole dimension. 
   - Component references and paths "
  record CREF_QUAL
    Ident ident;
    SubscriptList subscriptlist;
    ComponentRef componentref;
  end CREF_QUAL;
  record CREF_IDENT
    Ident ident;
    SubscriptList subscriptlist;
  end CREF_IDENT;
end ComponentRef;

public 
uniontype Path
  record QUALIFIED
    Ident ident;
    Path path;
  end QUALIFIED;
  record IDENT
    Ident ident " A component reference is the fully or partially qualified name of ";
  end IDENT;
end Path;

public 
uniontype Restriction " a component.  It is represented as a list of 
   identifier--subscript pairs.  The type `Path', on the other hand, 
   is used to store references to class names, or names inside 
   class definitions. 
   - Restrictions "
  record R_CLASS
  end R_CLASS;
  record R_MODEL
  end R_MODEL;
  record R_RECORD
  end R_RECORD;
  record R_BLOCK
  end R_BLOCK;
  record R_CONNECTOR
  end R_CONNECTOR;
  record R_TYPE
  end R_TYPE;
  record R_PACKAGE
  end R_PACKAGE;
  record R_FUNCTION
  end R_FUNCTION;
  record R_ENUMERATION
  end R_ENUMERATION;
  record R_PREDEFINED_INT
  end R_PREDEFINED_INT;
  record R_PREDEFINED_REAL
  end R_PREDEFINED_REAL;
  record R_PREDEFINED_STRING
  end R_PREDEFINED_STRING;
  record R_PREDEFINED_BOOL
  end R_PREDEFINED_BOOL;
  record R_PREDEFINED_ENUM " These constructors each correspond to a different kind of class "
  end R_PREDEFINED_ENUM;
end Restriction;

public 
uniontype Annotation " declaration in Modelica, except the last four, which are used 
   for the predefined types.  The parser assigns each class 
   declaration one of the restrictions, and the actual class 
   definition is checked for conformance during translation.  The 
   predefined types are created in the `Builtin' module and are 
   assigned special restrictions. 
   Annotation "
  record ANNOTATION
    ElementArgList elementarglist " An Annotation is a class_modification. ";
  end ANNOTATION;
end Annotation;

public 
uniontype Comment " Comment "
  record COMMENT
    AnnotationOption annotationoption;
    StringOption stringoption;
  end COMMENT;
end Comment;

public 
type IdentOption = Option<Ident>;

type ComponentRefOption = Option<ComponentRef>;

uniontype ExternalDecl "ExternalDecl 
  Declaration of an external function call "
  record EXTERNALDECL
    IdentOption identoption;
    StringOption stringoption "The name of the external function ";
    ComponentRefOption componentrefoption "Lanugage of the external function ";
    ExpList explist "ouput parameter as return value only positional arguments, i.e. expression list";
  end EXTERNALDECL;
end ExternalDecl;

protected import OpenModelica.Compiler.Debug " From here down, only absyn helper functions should be present. 
 Thus,no actual absyn datatype definitions. ";

protected import OpenModelica.Compiler.Dump;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Print;

public function element_spec_name " relation: element_spec_name
 
  The `ElementSpec' type contans the name of the element, and this
  relation extracts this name.
 "
  input ElementSpec in_elementspec;
  output Ident out_ident;
algorithm 
  out_ident:=
  matchcontinue (in_elementspec)
    local Ident n;
    case CLASSDEF(_,CLASS(n,_,_,_,_,_,_)) then n; 
    case COMPONENTS(_,_,{COMPONENTITEM(COMPONENT(n,_,_),_)}) then n; 
    case EXTENDS(_,_)
      equation 
        Print.print_buf("# element_spec_name EXTENDS\n"); then fail();
  end matchcontinue;
end element_spec_name;

public function path_string " relation: path_string
 
  This function simply converts a `Path' to a `string'.
 "
  input Path path;
  output String s;
algorithm 
  s := path_string2(path, ".");
end path_string;

public function opt_path_string " relation: opt_path_string
 
  Returns a path converted to string or an empty string if nothing exist
 "
  input PathOption in_pathoption;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_pathoption)
    local
      Ident str;
      Path p;
    case (NONE) then ""; 
    case (SOME(p))
      equation 
        str = path_string(p); then str;
  end matchcontinue;
end opt_path_string;

public function path_string2 " relation: 
 
  Helper function to path_string
 "
  input Path in_path;
  input String in_string;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_path,in_string)
    local
      Ident s,ns,s1,ss,str;
      Path n;
    case (IDENT(s),_) then s; 
    case (QUALIFIED(s,n),str)
      equation 
        ns = path_string2(n, str);
        s1 = string_append(s, str);
        ss = string_append(s1, ns); then ss;
  end matchcontinue;
end path_string2;

public function path_last_ident " relation: path_last_ident
 
  Returns the last ident (After last dot) in a paht
 "
  input Path in_path;
  output Ident out_ident;
algorithm 
  out_ident:=
  matchcontinue (in_path)
    local
      Ident res,n;
      Path p;
    case (QUALIFIED(_,p))
      equation 
        res = path_last_ident(p); then res;
    case (IDENT(n)) then n; 
  end matchcontinue;
end path_last_ident;

public function path_first_ident " relation: path_first_ident
 
  Returns the last ident (After last dot) in a paht
 "
  input Path in_path;
  output Ident out_ident;
algorithm 
  out_ident:=
  matchcontinue (in_path)
    local
      Ident n;
      Path p;
    case (QUALIFIED(n,p)) then n; 
    case (IDENT(n)) then n; 
  end matchcontinue;
end path_first_ident;

public function get_cref_from_exp " relation: get_cref_from_exp
 
  Returns a flattened list of the component references 
  in an expression
 "
  input Exp in_exp;
  output ComponentRefList out_componentreflist;
protected 
  type ComponentRefList = list<ComponentRef>;
algorithm 
  out_componentreflist:=
  matchcontinue (in_exp)
    local
      type ComponentRefListList = list<ComponentRefList>;
      type ComponentRefListListList = list<ComponentRefListList>;
      ComponentRef cr;
      ComponentRefList l1,l2,res,res1,l3;
      Exp e1,e2,e3;
      Operator op;
      TupleExpExpList e4;
      FunctionArgs farg;
      ComponentRefListList res2;
      ExpList expl;
      ExpListList expll;
    case (INTEGER(_)) then {}; 
    case (REAL(_)) then {}; 
    case (STRING(_)) then {}; 
    case (BOOL(_)) then {}; 
    case (CREF(cr)) then {cr}; 
    case (BINARY(e1,op,e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res = list_append(l1, l2); then res;
    case (UNARY(op,e1))
      equation 
        res = get_cref_from_exp(e1); then res;
    case (LBINARY(e1,op,e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res = list_append(l1, l2); then res;
    case (LUNARY(op,e1))
      equation 
        res = get_cref_from_exp(e1); then res;
    case (RELATION(e1,op,e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res = list_append(l1, l2); then res;
    case (IFEXP(e1,e2,e3,e4))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res1 = list_append(l1, l2);
        l3 = get_cref_from_exp(e3);
        res = list_append(res1, l3) "TODO elseif's e4 "; then res;
    case (CALL(_,farg))
      equation 
        res = get_cref_from_farg(farg) "Util.list_map(expl,get_cref_from_exp) => res"; then res;
    case (ARRAY(expl))
      local ComponentRefListList res1;
      equation 
        res1 = Util.list_map(expl, get_cref_from_exp);
        res = Util.list_flatten(res1); then res;
    case (MATRIX(expll))
      local ComponentRefListListList res1;
      equation 
        res1 = Util.list_list_map(expll, get_cref_from_exp);
        res2 = Util.list_flatten(res1);
        res = Util.list_flatten(res2); then res;
    case (RANGE(e1,SOME(e3),e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res1 = list_append(l1, l2);
        l3 = get_cref_from_exp(e3);
        res = list_append(res1, l3); then res;
    case (RANGE(e1,NONE,e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res = list_append(l1, l2); then res;
    case (TUPLE(expl))
      equation 
        Print.print_buf("Not implemented yet\n"); then {};
  end matchcontinue;
end get_cref_from_exp;

protected function get_cref_from_farg " relation: get_cref_from_farg
  
  Returns the flattened list of all component references 
  present in a list of function arguments.
 "
  input FunctionArgs in_functionargs;
  output ComponentRefList out_componentreflist;
protected 
  type ComponentRefList = list<ComponentRef>;
algorithm 
  out_componentreflist:=
  matchcontinue (in_functionargs)
    local
      type ComponentRefListList = list<ComponentRefList>;
      ComponentRefListList l1,l2;
      ComponentRefList fl1,fl2,res;
      ExpList expl;
      NamedArgList nargl;
    case (FUNCTIONARGS(expl,nargl))
      equation 
        l1 = Util.list_map(expl, get_cref_from_exp);
        fl1 = Util.list_flatten(l1);
        l2 = Util.list_map(nargl, get_cref_from_narg);
        fl2 = Util.list_flatten(l2);
        res = list_append(fl1, fl2); then res;
  end matchcontinue;
end get_cref_from_farg;

protected function get_cref_from_narg " relation: get_cref_from_narg
  
  Returns the flattened list of all component references 
  present in a list of named function arguments.
 "
  input NamedArg in_namedarg;
  output ComponentRefList out_componentreflist;
protected 
  type ComponentRefList = list<ComponentRef>;
algorithm 
  out_componentreflist:=
  matchcontinue (in_namedarg)
    local
      ComponentRefList res;
      Exp exp;
    case (NAMEDARG(_,exp))
      equation 
        res = get_cref_from_exp(exp); then res;
  end matchcontinue;
end get_cref_from_narg;

public function join_paths " relation: join_paths
 
  This function joins two paths
 "
  input Path in_path1;
  input Path in_path2;
  output Path out_path;
algorithm 
  out_path:=
  matchcontinue (in_path1,in_path2)
    local
      Ident str;
      Path p2,p_1,p;
    case (IDENT(str),p2) then QUALIFIED(str,p2); 
    case (QUALIFIED(str,p),p2)
      equation 
        p_1 = join_paths(p, p2); then QUALIFIED(str,p_1);
  end matchcontinue;
end join_paths;

public function strip_last " relation: strip_last
 
  Returns the path given as argument to the function
  minus the last ident.
 "
  input Path in_path;
  output Path out_path;
algorithm 
  out_path:=
  matchcontinue (in_path)
    local
      Ident str;
      Path p_1,p;
    case (IDENT(_)) then fail(); 
    case (QUALIFIED(str,IDENT(_))) then IDENT(str); 
    case (QUALIFIED(str,p))
      equation 
        p_1 = strip_last(p); then QUALIFIED(str,p_1);
  end matchcontinue;
end strip_last;

public function cref_to_path " relation: cref_to_path
 
  This function converts a `ComponentRef' to a `Path', if possible.
  If the component reference contains subscripts, it will silently
  fail.
 "
  input ComponentRef in_componentref;
  output Path out_path;
algorithm 
  out_path:=
  matchcontinue (in_componentref)
    local
      Ident i;
      Path p;
      ComponentRef c;
    case CREF_IDENT(i,{}) then IDENT(i); 
    case CREF_QUAL(i,{},c)
      equation 
        p = cref_to_path(c); then QUALIFIED(i,p);
  end matchcontinue;
end cref_to_path;

public function path_to_cref " relation: path_to_cref
 
  This function converts a Path to a ComponentRef.
  
 "
  input Path in_path;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_path)
    local
      Ident i;
      ComponentRef c;
      Path p;
    case IDENT(i) then CREF_IDENT(i,{}); 
    case QUALIFIED(i,p)
      equation 
        c = path_to_cref(p); then CREF_QUAL(i,{},c);
  end matchcontinue;
end path_to_cref;

public function join_crefs
  input ComponentRef in_componentref1;
  input ComponentRef in_componentref2;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_componentref1,in_componentref2)
    local
      Ident id;
      SubscriptList sub;
      ComponentRef cr2,cr_1,cr;
    case (CREF_IDENT(id,sub),cr2) then CREF_QUAL(id,sub,cr2); 
    case (CREF_QUAL(id,sub,cr),cr2)
      equation 
        cr_1 = join_crefs(cr, cr2); then CREF_QUAL(id,sub,cr_1);
  end matchcontinue;
end join_crefs;

public function cref_get_first " relation: cref_get_first
 
  Returns first ident from a ComponentRef
 "
  input ComponentRef in_componentref;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_componentref)
    local Ident i;
    case (CREF_IDENT(i,_)) then CREF_IDENT(i,{}); 
    case (CREF_QUAL(i,_,_)) then CREF_IDENT(i,{}); 
  end matchcontinue;
end cref_get_first;

public function restr_string " relation: restr_string
  
  Maps a class restriction to the corresponding string for printing
 "
  input Restriction in_restriction;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_restriction)
    case R_CLASS() then "CLASS"; 
    case R_MODEL() then "MODEL"; 
    case R_RECORD() then "RECORD"; 
    case R_BLOCK() then "BLOCK"; 
    case R_CONNECTOR() then "CONNECTOR"; 
    case R_TYPE() then "TYPE"; 
    case R_PACKAGE() then "PACKAGE"; 
    case R_FUNCTION() then "FUNCTION"; 
    case R_PREDEFINED_INT() then "PREDEFINED_INT"; 
    case R_PREDEFINED_REAL() then "PREDEFINED_REAL"; 
    case R_PREDEFINED_STRING() then "PREDEFINED_STRING"; 
    case R_PREDEFINED_BOOL() then "PREDEFINED_BOOL"; 
  end matchcontinue;
end restr_string;

public function print_restr " relation: print_restr
 
  This is a utility function for printing a `Restriction'.  The code
  is excluded for brevity.
 "
  input Restriction restr;
protected 
  Ident str;
algorithm 
  str := restr_string(restr);
  Print.print_buf(str);
end print_restr;

public function last_classname " relation: last_classname
 
  Returns the path (=namse) of the last class in a program
 "
  input Program in_program;
  output Path out_path;
algorithm 
  out_path:=
  matchcontinue (in_program)
    local
      Ident id;
      ClassList lst;
    case (PROGRAM(lst,_))
      equation 
        CLASS(id,_,_,_,_,_,_) = Util.list_last(lst); then IDENT(id);
  end matchcontinue;
end last_classname;

public function class_filename " relation class_filename
  author: PA
 
  Retrieves the filename where the class is stored.
 "
  input Class in_class;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_class)
    local Ident filename;
    case (CLASS(_,_,_,_,_,_,filename)) then filename; 
  end matchcontinue;
end class_filename;

public function set_class_filename " relation class_filename
  author: PA
 
  Sets the filename where the class is stored.
 "
  input Class in_class;
  input String in_string;
  output Class out_class;
algorithm 
  out_class:=
  matchcontinue (in_class,in_string)
    local
      Ident n,filename;
      Boolean p,f,e;
      Restriction r;
      ClassDef body;
    case (CLASS(n,p,f,e,r,body,_),filename) then CLASS(n,p,f,e,r,body,filename); 
  end matchcontinue;
end set_class_filename;

public function print_absyn_exp " relation: print_absyn_exp 
 
  Prints an Exp
 "
  input Exp in_exp;
algorithm 
  _:=
  matchcontinue (in_exp)
    local ExpList es;
    case (ARRAY(es))
      equation 
        Debug.print("This is an array: ");
        Debug.print("[");
        Debug.print("]"); then ();
    case (TUPLE(es))
      equation 
        Debug.print("(");
        Debug.print(")"); then ();
    case (_)
      equation 
        Debug.print("\nNo tuple."); then ();
  end matchcontinue;
end print_absyn_exp;

public function cref_equal " relation: cref_equal
 
  Checks if the name of a ComponentRef is equal to the name of 
  another ComponentRef
  
 "
  input ComponentRef in_componentref1;
  input ComponentRef in_componentref2;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_componentref1,in_componentref2)
    local
      Ident id,id2;
      ComponentRef cr1,cr2;
    case (CREF_IDENT(id,_),CREF_IDENT(id2,_))
      equation 
        equality(id = id2); then true;
    case (CREF_QUAL(id,_,cr1),CREF_QUAL(id2,_,cr2))
      equation 
        equality(id = id2);
        true = cref_equal(cr1, cr2); then true;
    case (_,_) then false; 
  end matchcontinue;
end cref_equal;

public function is_package_restriction " relation is_package_restriction
  checks if the provided parameter is a package or not
 "
  input Restriction in_restriction;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_restriction)
    case (R_PACKAGE()) then true; 
    case (_) then false; 
  end matchcontinue;
end is_package_restriction;
end Absyn;

