package Algorithm "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Algorithm.rml
  module:      Algorithm
  description: Algorithm datatypes
 
  RCS: $Id: Algorithm.rml 1733 2005-04-27 06:47:45Z haklu $
 
  This file contains data types and functions for managing
  algorithm sections. The algorithms in the AST are analyzed by the `Inst'
  module (Inst.rml) which uses this module to represent the algorithms. No
  processing of any kind, except for building the datastructure is 
  done in this module.
  
  It is used primarily by Inst.rml which both provides its input data
  and uses its \"output\" data.
  
 "

public import OpenModelica.Compiler.Exp;

public import OpenModelica.Compiler.Types;

public import OpenModelica.Compiler.SCode;

public 
type Ident = String;

public 
type StatementList = list<Statement>;

uniontype Algorithm " The `Algorithm' type corresponds to a whole algorithm secion. 
   It is simple a list of algorithm statements. "
  record ALGORITHM
    StatementList statementlist;
  end ALGORITHM;
end Algorithm;

public 
type Exp_ExpList = list<Exp.Exp>;

uniontype Statement " There are four kinds of statements.  Assignments (`a := b;'),
    if statements (`if A then B; elseif C; else D;'), for loops
    (`for i in 1:10 loop ...; end for;') and when statements
    (`when E do S; end when;'). "
  record ASSIGN
    Exp.Type type_;
    Exp.ComponentRef componentref;
    Exp.Exp exp;
  end ASSIGN;
  record TUPLE_ASSIGN
    Exp.Type type_;
    Exp_ExpList exp_explist;
    Exp.Exp exp;
  end TUPLE_ASSIGN;
  record ASSIGN_ARR
    Exp.Type type_;
    Exp.ComponentRef componentref;
    Exp.Exp exp;
  end ASSIGN_ARR;
  record IF
    Exp.Exp exp;
    StatementList statementlist;
    Else else_;
  end IF;
  record FOR
    Exp.Type type_;
    Boolean boolean;
    Ident ident;
    Exp.Exp exp;
    StatementList statementlist;
  end FOR;
  record WHILE
    Exp.Exp exp;
    StatementList statementlist;
  end WHILE;
  record WHEN
    Exp.Exp exp;
    StatementList statementlist;
  end WHEN;
  record ASSERT
    Exp.Exp exp1;
    Exp.Exp exp2;
  end ASSERT;
end Statement;

public 
uniontype Else " An if statements can one or more `elseif' branches and an
    optional `else' branch. "
  record NOELSE
  end NOELSE;
  record ELSEIF
    Exp.Exp exp;
    StatementList statementlist;
    Else else_;
  end ELSEIF;
  record ELSE
    StatementList statementlist;
  end ELSE;
end Else;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Dump;

protected import OpenModelica.Compiler.Debug;

protected import OpenModelica.Compiler.Absyn;

public function make_assignment
  input Exp.Exp in_exp1;
  input Types.Properties in_properties2;
  input Exp.Exp in_exp3;
  input Types.Properties in_properties4;
  input SCode.Accessibility in_accessibility5;
  output Statement out_statement;
algorithm 
  out_statement:=
  matchcontinue (in_exp1,in_properties2,in_exp3,in_properties4,in_accessibility5)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.Exp lhs,rhs,e,rhs_1;
      Types.Properties lprop,rprop,lhprop,rhprop;
      Exp.Type t,crt;
      Exp.ComponentRef c;
      TupleTypes_TTypeAbsyn_PathOption lt,rt;
    case (lhs,lprop,rhs,rprop,_)
      equation 
        true = Types.is_prop_any_const(lprop);
        Print.print_error_buf("# Trying to assign to constant component\n");
        Print.print_error_buf("    ");
        Exp.print_exp(lhs);
        Print.print_error_buf(" := ");
        Exp.print_exp(rhs);
        Print.print_buf("\n"); then fail();
    case (e,_,_,_,SCode.RO())
      equation 
        Print.print_error_buf("# Can't assign to (readonly) ");
        Exp.print_exp(e);
        Print.print_error_buf("\n"); then fail();
    case (Exp.CREF(c,crt),lhprop,rhs,rhprop,_) "LS: Replaced \"as Types.PROP(_,false)\" from lhprop\", by the first predicate "
      equation 
        false = Types.is_prop_any_const(lhprop);
        rhs_1 = Types.match_prop(rhs, rhprop, lhprop);
        false = Types.is_prop_array(lhprop);
        t = get_prop_exp_type(lhprop); then ASSIGN(t,c,rhs_1);
    case (Exp.CREF(c,crt),lhprop,rhs,rhprop,_)
      equation 
        false = Types.is_prop_any_const(lhprop);
        rhs_1 = Types.match_prop(rhs, rhprop, lhprop);
        true = Types.is_prop_array(lhprop);
        t = get_prop_exp_type(lhprop); then ASSIGN_ARR(t,c,rhs_1);
    case (lhs,lprop,rhs,rprop,_)
      equation 
        lt = Types.get_prop_type(lprop);
        rt = Types.get_prop_type(rprop);
        false = Types.equivtypes(lt, rt);
        Print.print_error_buf("# Type mismatch in assignment\n");
        Print.print_error_buf("    ");
        Exp.print_exp(lhs);
        Print.print_error_buf(" := ");
        Exp.print_exp(rhs);
        Print.print_buf("\n");
        Print.print_error_buf("    ");
        Types.print_type(lt);
        Print.print_error_buf(" := ");
        Types.print_type(rt);
        Print.print_buf("\n"); then fail();
    case (lhs,lprop,rhs,rprop,_)
      equation 
        Print.print_buf("- Algorithm.make_assignment failed\n");
        Print.print_buf("    ");
        Exp.print_exp(lhs);
        Print.print_buf(" := ");
        Exp.print_exp(rhs);
        Print.print_buf("\n"); then fail();
  end matchcontinue;
end make_assignment;

public function make_tuple_assignment
  input Exp_ExpList in_exp_explist;
  input Types_PropertiesList in_types_propertieslist;
  input Exp.Exp in_exp;
  input Types.Properties in_properties;
  output Statement out_statement;
protected 
  type Types_PropertiesList = list<Types.Properties>;
algorithm 
  out_statement:=
  matchcontinue (in_exp_explist,in_types_propertieslist,in_exp,in_properties)
    local
      type BooleanList = list<Boolean>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type TupleTypes_TTypeAbsyn_PathOptionList = list<TupleTypes_TTypeAbsyn_PathOption>;
      type Types_ConstList = list<Types.Const>;
      BooleanList bvals;
      Exp_ExpList lhs,expl;
      Types_PropertiesList lprop,lhprops;
      Exp.Exp rhs,rhs_1;
      Types.Properties rprop;
      TupleTypes_TTypeAbsyn_PathOptionList lhrtypes,tpl;
      Types_ConstList clist;
    case (lhs,lprop,rhs,rprop)
      equation 
        bvals = Util.list_map(lprop, Types.is_prop_any_const);
        true = Util.bool_or_list(bvals);
        Print.print_error_buf(
          "# Trying to assign to constant component in tuple assignment\n");
        Print.print_error_buf("    ");
        Dump.print_list(lhs, Exp.print_exp, ", ");
        Print.print_error_buf(" := ");
        Exp.print_exp(rhs);
        Print.print_error_buf("\n"); then fail();
    case (expl,lhprops,rhs,Types.PROP_TUPLE((Types.T_TUPLE(tpl),_),Types.TUPLE_CONST(clist)))
      equation 
        bvals = Util.list_map(lhprops, Types.is_prop_any_const);
        false = Util.bool_or_list(bvals);
        lhrtypes = Util.list_map(lhprops, Types.get_prop_type);
        (rhs_1,_) = Types.match_type_list(rhs, tpl, lhrtypes); then TUPLE_ASSIGN(Exp.OTHER(),expl,rhs);
    case (lhs,lprop,rhs,rprop) then fail(); 
  end matchcontinue;
end make_tuple_assignment;

protected function get_prop_exp_type " relation: get_prop_exp_type
 
  Returns the expression type for a given Properties by calling
  get_type_exp_type. Used by make_assignment.
 "
  input Types.Properties p;
  output Exp.Type t;
protected 
  type Absyn_PathOption = Option<Absyn.Path>;
  type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
  TupleTypes_TTypeAbsyn_PathOption ty;
algorithm 
  ty := Types.get_prop_type(p);
  t := get_type_exp_type(ty);
end get_prop_exp_type;

protected function get_type_exp_type " relation: get_type_exp_type
 
  Returns the expression type for a given Type module type. Used only by
  get_prop_exp_type. 
 "
  input Types.Type in_type;
  output Exp.Type out_type;
algorithm 
  out_type:=
  matchcontinue (in_type)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.Type t_1;
      TupleTypes_TTypeAbsyn_PathOption t;
    case ((Types.T_INTEGER(_),_)) then Exp.INT(); 
    case ((Types.T_REAL(_),_)) then Exp.REAL(); 
    case ((Types.T_STRING(_),_)) then Exp.STRING(); 
    case ((Types.T_BOOL(_),_)) then Exp.BOOL(); 
    case ((Types.T_ARRAY(_,t),_))
      equation 
        t_1 = get_type_exp_type(t); then t_1;
    case ((_,_)) then Exp.OTHER();  "was fail but records must be handled somehow "
  end matchcontinue;
end get_type_exp_type;

public function make_if
  input Exp.Exp in_exp1;
  input Types.Properties in_properties2;
  input StatementList in_statementlist3;
  input TupleExp_ExpTypes_PropertiesStatementListList in_tupleexp_exptypes_propertiesstatementlistlist4;
  input StatementList in_statementlist5;
  output Statement out_statement;
protected 
  type TupleExp_ExpTypes_PropertiesStatementList = tuple<Exp.Exp,Types.Properties,StatementList>;
  type TupleExp_ExpTypes_PropertiesStatementListList = list<TupleExp_ExpTypes_PropertiesStatementList>;
algorithm 
  out_statement:=
  matchcontinue (in_exp1,in_properties2,in_statementlist3,in_tupleexp_exptypes_propertiesstatementlistlist4,in_statementlist5)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Else else_;
      Exp.Exp e;
      StatementList tb,fb;
      TupleExp_ExpTypes_PropertiesStatementListList eib;
      TupleTypes_TTypeAbsyn_PathOption t;
    case (e,Types.PROP((Types.T_BOOL(_),_),_),tb,eib,fb)
      equation 
        else_ = make_else(eib, fb); then IF(e,tb,else_);
    case (e,Types.PROP(t,_),_,_,_)
      equation 
        Print.print_error_buf("# Type error in if conditional (");
        Exp.print_exp(e);
        Print.print_buf(")\n");
        Print.print_error_buf("    Expected Boolean, got ");
        Types.print_type(t);
        Print.print_error_buf("\n"); then fail();
  end matchcontinue;
end make_if;

protected function make_else
  input TupleExp_ExpTypes_PropertiesStatementListList in_tupleexp_exptypes_propertiesstatementlistlist;
  input StatementList in_statementlist;
  output Else out_else;
protected 
  type TupleExp_ExpTypes_PropertiesStatementList = tuple<Exp.Exp,Types.Properties,StatementList>;
  type TupleExp_ExpTypes_PropertiesStatementListList = list<TupleExp_ExpTypes_PropertiesStatementList>;
algorithm 
  out_else:=
  matchcontinue (in_tupleexp_exptypes_propertiesstatementlistlist,in_statementlist)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      StatementList fb,b;
      Else else_;
      Exp.Exp e;
      TupleExp_ExpTypes_PropertiesStatementListList xs;
      TupleTypes_TTypeAbsyn_PathOption t;
    case ({},{}) then NOELSE(); 
    case ({},fb) then ELSE(fb); 
    case ((e,Types.PROP((Types.T_BOOL(_),_),_),b) :: xs,fb)
      equation 
        else_ = make_else(xs, fb); then ELSEIF(e,b,else_);
    case ((e,Types.PROP(t,_),_) :: _,_)
      equation 
        Print.print_error_buf("# Type error in elseif conditional (");
        Exp.print_exp(e);
        Print.print_error_buf(")\n");
        Print.print_error_buf("    Expected Boolean, got ");
        Types.print_type(t);
        Print.print_error_buf("\n"); then fail();
  end matchcontinue;
end make_else;

public function make_for
  input Ident in_ident;
  input Exp.Exp in_exp;
  input Types.Properties in_properties;
  input StatementList in_statementlist;
  output Statement out_statement;
algorithm 
  out_statement:=
  matchcontinue (in_ident,in_exp,in_properties,in_statementlist)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Boolean array;
      Exp.Type et;
      Ident i;
      Exp.Exp e;
      TupleTypes_TTypeAbsyn_PathOption t;
      StatementList stmts;
    case (i,e,Types.PROP((Types.T_ARRAY(_,t),_),_),stmts)
      equation 
        array = Types.is_array(t);
        et = Types.elab_type(t); then FOR(et,array,i,e,stmts);
    case (_,_,Types.PROP(t,_),_)
      equation 
        Print.print_error_buf("# Type error in for conditional.\n");
        Print.print_error_buf("    Expected array, got ");
        Types.print_type(t);
        Print.print_error_buf("\n"); then fail();
  end matchcontinue;
end make_for;

public function make_while
  input Exp.Exp in_exp;
  input Types.Properties in_properties;
  input StatementList in_statementlist;
  output Statement out_statement;
algorithm 
  out_statement:=
  matchcontinue (in_exp,in_properties,in_statementlist)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.Exp e;
      StatementList stmts;
      TupleTypes_TTypeAbsyn_PathOption t;
    case (e,Types.PROP((Types.T_BOOL(_),_),_),stmts) then WHILE(e,stmts); 
    case (_,Types.PROP(t,_),_)
      equation 
        Print.print_error_buf("# Type error in while conditional.\n");
        Print.print_error_buf("    Expected Boolean, got ");
        Types.print_type(t);
        Print.print_error_buf("\n"); then fail();
  end matchcontinue;
end make_while;

public function make_when_a
  input Exp.Exp in_exp;
  input Types.Properties in_properties;
  input StatementList in_statementlist;
  output Statement out_statement;
algorithm 
  out_statement:=
  matchcontinue (in_exp,in_properties,in_statementlist)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.Exp e;
      StatementList stmts;
      TupleTypes_TTypeAbsyn_PathOption t;
    case (e,Types.PROP((Types.T_BOOL(_),_),_),stmts) then WHEN(e,stmts); 
    case (e,Types.PROP((Types.T_ARRAY(_,(Types.T_BOOL(_),_)),_),_),stmts) then WHEN(e,stmts); 
    case (_,Types.PROP(t,_),_)
      equation 
        Print.print_error_buf("# Type error in when conditional.\n");
        Print.print_error_buf("    Expected Boolean scalar or vector, got ");
        Types.print_type(t);
        Print.print_error_buf("\n"); then fail();
  end matchcontinue;
end make_when_a;

public function make_assert
  input Exp.Exp in_exp1;
  input Exp.Exp in_exp2;
  input Types.Properties in_properties3;
  input Types.Properties in_properties4;
  output Statement out_statement;
algorithm 
  out_statement:=
  matchcontinue (in_exp1,in_exp2,in_properties3,in_properties4)
    local Exp.Exp cond,msg;
    case (cond,msg,Types.PROP((Types.T_BOOL(_),_),_),Types.PROP((Types.T_STRING(_),_),_)) then ASSERT(cond,msg);  "RML does not handle the pattern below T_BOOL(_), hence we need to
	     implement this differently.   rule	not let T_BOOL(_) = condt &  
	 Print.print_buf \"# Type error in assert condition.\n\" &
	 Print.print_buf \" Expected Boolean, got \" &
	 Types.print_type condt & Print.print_buf \"\n\"
	 --------------------------------------------
	 make_assert(_,_,Types.PROP(condt,_),_) => fail 
	 "
  end matchcontinue;
end make_assert;

public function get_all_exps " relation: get_all_exps
  
  This function goes through the Algorithm structure and finds all the
  expressions and returns them in a list
 "
  input Algorithm in_algorithm;
  output Exp_ExpList out_exp_explist;
algorithm 
  out_exp_explist:=
  matchcontinue (in_algorithm)
    local
      Exp_ExpList exps;
      StatementList stmts;
    case ALGORITHM(stmts)
      equation 
        exps = get_all_exps_stmts(stmts); then exps;
  end matchcontinue;
end get_all_exps;

protected function get_all_exps_stmts
  input StatementList stmts;
  output Exp_ExpList exps;
protected 
  type Exp_ExpListList = list<Exp_ExpList>;
  Exp_ExpListList expslist;
algorithm 
  expslist := Util.list_map(stmts, get_all_exps_stmt);
  exps := Util.list_flatten(expslist);
end get_all_exps_stmts;

protected function get_all_exps_stmt
  input Statement in_statement;
  output Exp_ExpList out_exp_explist;
algorithm 
  out_exp_explist:=
  matchcontinue (in_statement)
    local
      Exp.Exp crexp,exp,e1,e2;
      Exp.Type expty;
      Exp.ComponentRef cr;
      Exp_ExpList exps,explist,exps1,elseexps;
      StatementList stmts;
      Else else_;
      Boolean flag;
      Ident id;
    case ASSIGN(expty,cr,exp)
      equation 
        crexp = cref_to_exp(cr); then {crexp,exp};
    case TUPLE_ASSIGN(expty,explist,exp)
      equation 
        exps = list_append(explist, {exp}); then exps;
    case ASSIGN_ARR(expty,cr,exp)
      equation 
        crexp = cref_to_exp(cr); then {crexp,exp};
    case IF(exp,stmts,else_)
      equation 
        exps1 = get_all_exps_stmts(stmts);
        elseexps = get_all_exps_else(else_);
        exps = list_append(exps1, elseexps); then exp :: exps;
    case FOR(expty,flag,id,exp,stmts)
      equation 
        exps = get_all_exps_stmts(stmts); then exp :: exps;
    case WHILE(exp,stmts)
      equation 
        exps = get_all_exps_stmts(stmts); then exp :: exps;
    case WHEN(exp,stmts)
      equation 
        exps = get_all_exps_stmts(stmts); then exp :: exps;
    case ASSERT(e1,e2) then {e1,e2}; 
    case _
      equation 
        Debug.fprintln("failtrace", "-- get_all_exps_stmt failed"); then fail();
  end matchcontinue;
end get_all_exps_stmt;

protected function get_all_exps_else
  input Else in_else;
  output Exp_ExpList out_exp_explist;
algorithm 
  out_exp_explist:=
  matchcontinue (in_else)
    local
      Exp_ExpList exps1,elseexps,exps;
      Exp.Exp exp;
      StatementList stmts;
      Else else_;
    case NOELSE() then {}; 
    case ELSEIF(exp,stmts,else_)
      equation 
        exps1 = get_all_exps_stmts(stmts);
        elseexps = get_all_exps_else(else_);
        exps = list_append(exps1, elseexps); then exp :: exps;
    case ELSE(stmts)
      equation 
        exps = get_all_exps_stmts(stmts); then exps;
  end matchcontinue;
end get_all_exps_else;

protected function cref_to_exp
  input Exp.ComponentRef in_componentref;
  output Exp.Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_componentref)
    local Exp.ComponentRef cref;
    case cref then Exp.CREF(cref,Exp.OTHER()); 
  end matchcontinue;
end cref_to_exp;
end Algorithm;

