package ClassLoader "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 ClassLoader.rml
  module:      ClassLoader
  description: Loading of classes from $MODELICAPATH.
 
  RCS: $Id: ClassLoader.rml 1733 2005-04-27 06:47:45Z haklu $
 
  This module loads classes from $MODELICAPATH. It exports only one
  relation: the load_class function. It is currently (2004-09-27) 
  only used by Ceval.rml when using the \"loadclass\" function in 
  the interactive environment.
  
 "

public import OpenModelica.Compiler.Absyn;

protected import OpenModelica.Compiler.System;

protected import OpenModelica.Compiler.Lookup;

protected import OpenModelica.Compiler.Interactive;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Parser;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Env;

protected import OpenModelica.Compiler.Dump;

public function load_class " relation: load_class
  This function takes a 'Path' and the $MODELICAPATH as a string
  and tries to load the class from the path.
  If the classname is qualified, the complete package is loaded. 
  E.g. load_class(Modelica.SIunits.Voltage) -> whole Modelica package loaded.
 "
  input Absyn.Path in_path;
  input String in_string;
  output Absyn.Program out_program;
algorithm 
  out_program:=
  matchcontinue (in_path,in_string)
    local
      type StringList = list<String>;
      String gd,classname,mp,pack;
      StringList mps;
      Absyn.Program p;
      Absyn.Path rest,path;
    case (Absyn.IDENT(classname),mp) "Simple names: Just load the file if it can be found in $MODELICAPATH "
      equation 
        gd = System.group_delimiter();
        mps = System.strtok(mp, gd);
        p = load_class_from_mps(classname, mps); then p;
    case (Absyn.QUALIFIED(pack,rest),mp) "Qualified names: First check if it is defined in a file pack.mo "
      equation 
        gd = System.group_delimiter();
        mps = System.strtok(mp, gd);
        p = load_class_from_mps(pack, mps); then p;
    case ((path = Absyn.QUALIFIED(pack,rest)),mp) "Qualified names: Else, load the complete package and then check that  the package contains the file "
      equation 
        gd = System.group_delimiter();
        mps = System.strtok(mp, gd);
        p = load_complete_package_from_mps(pack, mps, Absyn.TOP(), Absyn.PROGRAM({},Absyn.TOP()));
        _ = Interactive.get_pathed_class_in_program(path, p); then p;
    case (_,_)
      equation 
        print("load_class failed\n"); then fail();
  end matchcontinue;
end load_class;

protected function exist_regular_file " relation: exist_regular_file
 
  Checks if a file exists
 "
  input String filename;
algorithm 
  0 := System.regular_file_exist(filename);
end exist_regular_file;

protected function exist_directory_file " relation: exist_directory_file
 
  Checks if a directory exist
 "
  input String filename;
algorithm 
  0 := System.directory_exist(filename);
end exist_directory_file;

protected function load_class_from_mps " relation: load_class_from_mps
 
  Loads a class or classes from a set of paths in modelicapath
 "
  input Absyn.Ident in_ident;
  input StringList in_stringlist;
  output Absyn.Program out_program;
protected 
  type StringList = list<String>;
algorithm 
  out_program:=
  matchcontinue (in_ident,in_stringlist)
    local
      Absyn.Program p;
      String class_,mp;
      StringList mps;
    case (class_,mp :: mps)
      equation 
        p = load_class_from_mp(class_, mp); then p;
    case (class_,_ :: mps)
      equation 
        p = load_class_from_mps(class_, mps); then p;
  end matchcontinue;
end load_class_from_mps;

protected function load_class_from_mp " relation: load_class_from_mp
  
  This function loads a modelica class \"className\" from the file path 
  \"<mp>/className.mo\" or it loads complete package from 
  \"<mp>/className/package.mo\"
 "
  input Absyn.Ident in_ident;
  input String in_string;
  output Absyn.Program out_program;
algorithm 
  out_program:=
  matchcontinue (in_ident,in_string)
    local
      String pd,classfile,classfile_1,class_,mp,dirfile,packfile;
      Absyn.Program p;
    case (class_,mp)
      equation 
        pd = System.path_delimiter();
        classfile = string_append(class_, ".mo");
        classfile_1 = Util.string_append_list({mp,pd,classfile});
        exist_regular_file(classfile_1);
        print("parsing ");
        print(classfile_1);
        print("\n");
        p = Parser.parse(classfile_1); then p;
    case (class_,mp)
      equation 
        pd = System.path_delimiter();
        dirfile = Util.string_append_list({mp,pd,class_});
        packfile = Util.string_append_list({dirfile,pd,"package.mo"});
        exist_directory_file(dirfile);
        exist_regular_file(packfile);
        Print.print_buf(
          "Class is package stored in a directory, loading whole package(incl. subdir)\n");
        p = load_complete_package_from_mp(class_, mp, Absyn.TOP(), Absyn.PROGRAM({},Absyn.TOP())); then p;
    case (_,_)
      equation 
        print("load_class_from_mp failed\n"); then fail();
  end matchcontinue;
end load_class_from_mp;

protected function load_complete_package_from_mps " relation: load_complete_package_from_mps
 
  Loads a whole package from the ModelicaPaths defined in MODELICAPATH
 "
  input Absyn.Ident in_ident;
  input StringList in_stringlist;
  input Absyn.Within in_within;
  input Absyn.Program in_program;
  output Absyn.Program out_program;
protected 
  type StringList = list<String>;
algorithm 
  out_program:=
  matchcontinue (in_ident,in_stringlist,in_within,in_program)
    local
      Absyn.Program p,oldp;
      String pack,mp;
      Absyn.Within within_;
      StringList mps;
    case (pack,mp :: _,within_,oldp)
      equation 
        p = load_complete_package_from_mp(pack, mp, within_, oldp); then p;
    case (pack,_ :: mps,within_,oldp)
      equation 
        p = load_complete_package_from_mps(pack, mps, within_, oldp); then p;
  end matchcontinue;
end load_complete_package_from_mps;

protected function load_complete_package_from_mp " relation: load_complete_package_from_mp
  
  Loads a whole package from the ModelicaPaths defined in MODELICAPATH
 "
  input Absyn.Ident in_ident;
  input String in_string;
  input Absyn.Within in_within;
  input Absyn.Program in_program;
  output Absyn.Program out_program;
algorithm 
  out_program:=
  matchcontinue (in_ident,in_string,in_within,in_program)
    local
      type Absyn_ClassList = list<Absyn.Class>;
      type StringList = list<String>;
      String pd,mp_1,packagefile,subdirstr,pack,mp;
      Absyn_ClassList p1,oldc;
      Absyn.Within w1,within_;
      Absyn.Program p1_1,p2,p;
      StringList subdirs;
      Absyn.Path wpath_1,wpath;
    case (pack,mp,(within_ = Absyn.TOP()),Absyn.PROGRAM(oldc,_))
      equation 
        pd = System.path_delimiter();
        mp_1 = Util.string_append_list({mp,pd,pack});
        packagefile = Util.string_append_list({mp_1,pd,"package.mo"});
        exist_regular_file(packagefile);
        print("parsing ");
        print(packagefile);
        print("\n");
        Absyn.PROGRAM(p1,w1) = Parser.parse(packagefile);
        Print.print_buf("loading ");
        Print.print_buf(packagefile);
        Print.print_buf("\n");
        p1_1 = Interactive.update_program(Absyn.PROGRAM(p1,w1), Absyn.PROGRAM(oldc,Absyn.TOP()));
        subdirs = System.sub_directories(mp_1);
        subdirstr = Util.string_delimit_list(subdirs, ", ");
        print("subdirs =");
        print(subdirstr);
        print("\n");
        p2 = load_complete_subdirs(subdirs, pack, mp_1, within_, p1_1);
        p = load_complete_subfiles(pack, mp_1, within_, p2); then p;
    case (pack,mp,(within_ = Absyn.WITHIN(wpath)),Absyn.PROGRAM(oldc,_))
      equation 
        pd = System.path_delimiter();
        mp_1 = Util.string_append_list({mp,pd,pack});
        packagefile = Util.string_append_list({mp_1,pd,"package.mo"});
        exist_regular_file(packagefile);
        print("parsing ");
        print(packagefile);
        print("\n");
        Absyn.PROGRAM(p1,w1) = Parser.parse(packagefile);
        Print.print_buf("loading ");
        Print.print_buf(packagefile);
        Print.print_buf("\n");
        p1_1 = Interactive.update_program(Absyn.PROGRAM(p1,Absyn.WITHIN(wpath)), 
          Absyn.PROGRAM(oldc,Absyn.TOP()));
        subdirs = System.sub_directories(mp_1);
        subdirstr = Util.string_delimit_list(subdirs, ", ");
        print("subdirs =");
        print(subdirstr);
        print("\n");
        p2 = load_complete_subdirs(subdirs, pack, mp_1, within_, p1_1);
        wpath_1 = Absyn.join_paths(wpath, Absyn.IDENT(pack));
        p = load_complete_subfiles(pack, mp_1, within_, p2); then p;
    case (_,_,_,_) then fail(); 
  end matchcontinue;
end load_complete_package_from_mp;

protected function load_complete_subdirs
  input StringList in_stringlist;
  input Absyn.Ident in_ident;
  input String in_string;
  input Absyn.Within in_within;
  input Absyn.Program in_program;
  output Absyn.Program out_program;
protected 
  type StringList = list<String>;
algorithm 
  out_program:=
  matchcontinue (in_stringlist,in_ident,in_string,in_within,in_program)
    local
      type Absyn_ClassList = list<Absyn.Class>;
      Absyn.Within w,w2,within_;
      Absyn_ClassList oldcls;
      Absyn.Path pack_1,pack2;
      Absyn.Program p,p_1,oldp;
      String pack,pack1,mp;
      StringList packs;
    case ({},_,_,w,Absyn.PROGRAM(oldcls,w2)) then Absyn.PROGRAM(oldcls,w2); 
    case (pack :: packs,pack1,mp,(within_ = Absyn.WITHIN(pack2)),oldp)
      equation 
        pack_1 = Absyn.join_paths(pack2, Absyn.IDENT(pack1));
        p = load_complete_package_from_mp(pack, mp, Absyn.WITHIN(pack_1), oldp);
        p_1 = load_complete_subdirs(packs, pack1, mp, within_, p); then p_1;
    case (pack :: packs,pack1,mp,(within_ = Absyn.TOP()),oldp)
      equation 
        pack_1 = Absyn.join_paths(Absyn.IDENT(pack1), Absyn.IDENT(pack));
        p = load_complete_package_from_mp(pack, mp, Absyn.WITHIN(Absyn.IDENT(pack1)), oldp);
        p_1 = load_complete_subdirs(packs, pack1, mp, within_, p); then p_1;
    case (pack :: packs,pack1,mp,within_,p)
      equation 
        p_1 = load_complete_subdirs(packs, pack1, mp, within_, p); then p_1;
    case (_,_,_,_,_)
      equation 
        print("load_complete_subdirs failed\n"); then fail();
  end matchcontinue;
end load_complete_subdirs;

protected function load_complete_subfiles
  input Absyn.Ident in_ident;
  input String in_string;
  input Absyn.Within in_within;
  input Absyn.Program in_program;
  output Absyn.Program out_program;
algorithm 
  out_program:=
  matchcontinue (in_ident,in_string,in_within,in_program)
    local
      type StringList = list<String>;
      StringList mofiles;
      Absyn.Path within_1,within_;
      Absyn.Program p,oldp;
      String pack,mp;
    case (pack,mp,Absyn.WITHIN(within_),oldp)
      equation 
        mofiles = System.mo_files(mp) "Here *.mo files in same directory as package.mo should be loaded as sub-packages";
        within_1 = Absyn.join_paths(within_, Absyn.IDENT(pack));
        p = load_subpackage_files(mofiles, mp, Absyn.WITHIN(within_1), oldp); then p;
    case (pack,mp,Absyn.TOP(),oldp)
      equation 
        mofiles = System.mo_files(mp) "Here *.mo files in same directory as package.mo should be loaded as sub-packages";
        p = load_subpackage_files(mofiles, mp, Absyn.WITHIN(Absyn.IDENT(pack)), oldp); then p;
    case (_,_,_,_)
      equation 
        print("load_complete_subfiles failed\n"); then fail();
  end matchcontinue;
end load_complete_subfiles;

protected function load_subpackage_files
  input StringList in_stringlist;
  input String in_string;
  input Absyn.Within in_within;
  input Absyn.Program in_program;
  output Absyn.Program out_program;
protected 
  type StringList = list<String>;
algorithm 
  out_program:=
  matchcontinue (in_stringlist,in_string,in_within,in_program)
    local
      type Absyn_ClassList = list<Absyn.Class>;
      String mp,pd,f_1,f;
      Absyn.Within within_,w;
      Absyn_ClassList cls,oldc;
      Absyn.Program p_1,p_2;
      StringList fs;
    case ({},mp,within_,Absyn.PROGRAM(cls,w)) then Absyn.PROGRAM(cls,w); 
    case (f :: fs,mp,within_,Absyn.PROGRAM(oldc,_))
      equation 
        pd = System.path_delimiter();
        f_1 = Util.string_append_list({mp,pd,f});
        print("parsing ");
        print(f_1);
        print("\n");
        Absyn.PROGRAM(cls,_) = Parser.parse(f_1);
        Print.print_buf("loading ");
        Print.print_buf(f_1);
        Print.print_buf("\n");
        p_1 = Interactive.update_program(Absyn.PROGRAM(cls,within_), Absyn.PROGRAM(oldc,Absyn.TOP()));
        p_2 = load_subpackage_files(fs, mp, within_, p_1); then p_2;
    case (_,_,_,_)
      equation 
        print("load_subpackage_files failed\n"); then fail();
  end matchcontinue;
end load_subpackage_files;
end ClassLoader;

