package Connect "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Connect.rml
  module:      Connect
  description: Connection set management
 
  RCS: $Id: Connect.rml 1733 2005-04-27 06:47:45Z haklu $
 
  Connections generate connection sets (datatype SET is described below)
  which are constructed during instantiation.  When a connection 
  set is generated, it is used to create a number of equations. 
  The kind of equations created depends on the type of the set. 
  
  Connect.rml is called from Inst.rml and is responsible for 
  creation of all connect-equations later passed to the DAE module 
  in DAE.rml.
  
 "

public import OpenModelica.Compiler.Exp;

public import OpenModelica.Compiler.Static;

public import OpenModelica.Compiler.DAE;

public import OpenModelica.Compiler.Env;

public 
uniontype Face
  record INNER
  end INNER;
  record OUTER " This type indicates whether a connector is an inner or an outer
    connector.  "
  end OUTER;
end Face;

public 
type Exp_ComponentRefList = list<Exp.ComponentRef>;

type TupleExp_ComponentRefFace = tuple<Exp.ComponentRef,Face>;

type TupleExp_ComponentRefFaceList = list<TupleExp_ComponentRefFace>;

uniontype Set " This type indicates whether a connector is an inner or an outer
    connector.  "
  record EQU
    Exp_ComponentRefList exp_componentreflist;
  end EQU;
  record FLOW
    TupleExp_ComponentRefFaceList tupleexp_componentreffacelist;
  end FLOW;
end Set;

public 
type Sets = list<Set>;

public function add_equ "
  - Adding
 
  The two functions `add_eq' and `add_flow' addes a variable to a
  connection set.  The first function is used to add a non-flow
  variable, and the second is used to add a flow variable.  When
  two component are to be added to a collection of connection sets,
  the connections sets containg the components have to be located.
  If no such set exists, a new set containing only the new component
  is created.
 
  If the connection sets containing the two components are not the
  same, they are merged.
 "
  input Sets ss;
  input Exp.ComponentRef r1;
  input Exp.ComponentRef r2;
  output Sets ss_1;
protected 
  Set s1,s2;
  Sets ss_1;
algorithm 
  s1 := find_equ_set(ss, r1);
  s2 := find_equ_set(ss, r2);
  ss_1 := merge(ss, s1, s2);
end add_equ;

public function add_flow
  input Sets ss;
  input Exp.ComponentRef r1;
  input Face d1;
  input Exp.ComponentRef r2;
  input Face d2;
  output Sets ss_1;
protected 
  Set s1,s2;
  Sets ss_1;
algorithm 
  s1 := find_flow_set(ss, r1, d1);
  s2 := find_flow_set(ss, r2, d2);
  ss_1 := merge(ss, s1, s2);
end add_flow;

public function add_array_equ "For connecting two arrays, an equal equation for each index should be generated."
  input Sets in_sets1;
  input Exp.ComponentRef in_componentref2;
  input Exp.ComponentRef in_componentref3;
  input Integer in_integer4;
  output Sets out_sets;
algorithm 
  out_sets:=
  matchcontinue (in_sets1,in_componentref2,in_componentref3,in_integer4)
    local
      Sets s,ss_1,ss_2,ss;
      Exp.ComponentRef r1_1,r2_1,r1,r2;
      Integer i_1,i;
      Set s1,s2;
    case (s,_,_,0) then s; 
    case (ss,r1,r2,i)
      equation 
        r1_1 = Exp.subscript_cref(r1, {Exp.INDEX(Exp.ICONST(i))});
        r2_1 = Exp.subscript_cref(r2, {Exp.INDEX(Exp.ICONST(i))});
        i_1 = i - 1;
        s1 = find_equ_set(ss, r1_1);
        s2 = find_equ_set(ss, r2_1);
        ss_1 = merge(ss, s1, s2);
        ss_2 = add_array_equ(ss_1, r1, r2, i_1); then ss_2;
  end matchcontinue;
end add_array_equ;

public function equations "
  - Equation generation
 
   relation: equations
 
  From a number of connection sets, this function generates a list
  of equations.
 "
  input Sets in_sets;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_sets)
    local
      DAE_ElementList dae1,dae2,dae;
      Exp_ComponentRefList cs;
      Sets ss;
    case {} then {}; 
    case EQU(cs) :: ss
      equation 
        dae1 = equ_equations(cs);
        dae2 = equations(ss);
        dae = list_append(dae1, dae2); then dae;
    case FLOW(cs) :: ss
      local TupleExp_ComponentRefFaceList cs;
      equation 
        dae1 = flow_equations(cs);
        dae2 = equations(ss);
        dae = list_append(dae1, dae2); then dae;
  end matchcontinue;
end equations;

protected function equ_equations " relation: equ_equations
  
  A non-flow connection set contains a number of components.
  Generating the equation from this set means equating all the
  components.  For n components, this will give n-1 equations.
 
  For example, if the set contains the components `x', `y.a' and
  `z.b', the equations generated will me `x = y.a' and `y.a = z.b'.
 "
  input Exp_ComponentRefList in_exp_componentreflist;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_exp_componentreflist)
    local
      DAE_ElementList eq;
      Exp.ComponentRef x,y;
      Exp_ComponentRefList cs;
    case {_} then {}; 
    case x :: y :: cs
      equation 
        eq = equ_equations(y :: cs); then DAE.EQUATION(Exp.CREF(x,Exp.OTHER()),Exp.CREF(y,Exp.OTHER())) :: eq;
  end matchcontinue;
end equ_equations;

protected function flow_equations " relation: flow_equations
  
  Generating equations from a flow connection set is a little
  trickier that from a non-flow set.  Only one equation is
  generated, but it has to consider whether the comoponents were
  inner or outer connectors.
 
  This function uses `flow_sum' to create the sum of all components
  (some of which will be negated), and the returns the equation
  where this sum is equal to 0.0.
 "
  input TupleExp_ComponentRefFaceList cs;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
  Exp.Exp sum;
algorithm 
  sum := flow_sum(cs);
  out_dae_elementlist := {DAE.EQUATION(sum,Exp.RCONST(0.0))};
end flow_equations;

protected function flow_sum " relation: flow_sum
  
  This function creates an exression expressing the sum of all
  components in the given list.  Before adding the component to the
  sum, it is passed to `sign_flow' which will negate all outer
  connectors.
 "
  input TupleExp_ComponentRefFaceList in_tupleexp_componentreffacelist;
  output Exp.Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_tupleexp_componentreffacelist)
    local
      Exp.Exp exp,exp1,exp2;
      Exp.ComponentRef c;
      Face f;
      TupleExp_ComponentRefFaceList cs;
    case {(c,f)}
      equation 
        exp = sign_flow(c, f); then exp;
    case ((c,f) :: cs)
      equation 
        exp1 = sign_flow(c, f);
        exp2 = flow_sum(cs); then Exp.BINARY(exp1,Exp.ADD(Exp.REAL()),exp2);
  end matchcontinue;
end flow_sum;

protected function sign_flow " relation: sign_flow
 
  This function takes a name of a component and a `Face', returns an
  expression.  If the face is `INNER' the expression simply contains
  the component reference, but if it is `OUTER', the expression is
  negated.
 "
  input Exp.ComponentRef in_componentref;
  input Face in_face;
  output Exp.Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_componentref,in_face)
    local Exp.ComponentRef c;
    case (c,INNER()) then Exp.CREF(c,Exp.OTHER()); 
    case (c,OUTER()) then Exp.UNARY(Exp.UMINUS(Exp.REAL()),Exp.CREF(c,Exp.OTHER())); 
  end matchcontinue;
end sign_flow;

protected function find_equ_set "
  - Lookup
  
  These functions are used to find and create connection sets.
 
   relation: find_equ_set
 
  This function finds a non-flow connection set that contains the
  component named by the second argument.  If no such set is found,
  a new set is created.
 "
  input Sets in_sets;
  input Exp.ComponentRef in_componentref;
  output Set out_set;
algorithm 
  out_set:=
  matchcontinue (in_sets,in_componentref)
    local
      Set s;
      Exp.ComponentRef c;
      Sets ss;
    case ({},c)
      equation 
        s = new_equ_set(c); then s;
    case (s :: _,c)
      equation 
        find_in_set(s, c); then s;
    case (_ :: ss,c)
      equation 
        s = find_equ_set(ss, c); then s;
  end matchcontinue;
end find_equ_set;

protected function find_flow_set " relation: find_flow_set
 
  This function finds a flow connection set that contains the
  component named by the second argument.  If no such set is found,
  a new set is created.
 "
  input Sets in_sets;
  input Exp.ComponentRef in_componentref;
  input Face in_face;
  output Set out_set;
algorithm 
  out_set:=
  matchcontinue (in_sets,in_componentref,in_face)
    local
      Set s;
      Exp.ComponentRef c;
      Face d;
      Sets ss;
    case ({},c,d)
      equation 
        s = new_flow_set(c, d); then s;
    case (s :: _,c,d)
      equation 
        find_in_set(s, c); then s;
    case (_ :: ss,c,d)
      equation 
        s = find_flow_set(ss, c, d); then s;
  end matchcontinue;
end find_flow_set;

protected function find_in_set " relation: find_in_set
  
  This function checks if a componet already appears in a given
  connection set.
 "
  input Set in_set;
  input Exp.ComponentRef in_componentref;
algorithm 
  _:=
  matchcontinue (in_set,in_componentref)
    local
      Exp_ComponentRefList cs;
      Exp.ComponentRef c;
    case (EQU(cs),c)
      equation 
        find_in_set_equ(cs, c); then ();
    case (FLOW(cs),c)
      local TupleExp_ComponentRefFaceList cs;
      equation 
        find_in_set_flow(cs, c); then ();
  end matchcontinue;
end find_in_set;

protected function find_in_set_equ " relation: find_in_set_equ
  
  This is a version of `find_in_set' which is specialized on
  non-flow connection sets
 "
  input Exp_ComponentRefList in_exp_componentreflist;
  input Exp.ComponentRef in_componentref;
algorithm 
  _:=
  matchcontinue (in_exp_componentreflist,in_componentref)
    local
      Exp.ComponentRef c1,c2;
      Exp_ComponentRefList cs;
    case (c1 :: _,c2)
      equation 
        Static.eq_cref(c1, c2); then ();
    case (_ :: cs,c2)
      equation 
        find_in_set_equ(cs, c2); then ();
  end matchcontinue;
end find_in_set_equ;

protected function find_in_set_flow
  input TupleExp_ComponentRefFaceList in_tupleexp_componentreffacelist;
  input Exp.ComponentRef in_componentref;
algorithm 
  _:=
  matchcontinue (in_tupleexp_componentreffacelist,in_componentref)
    local
      Exp.ComponentRef c1,c2;
      TupleExp_ComponentRefFaceList cs;
    case ((c1,_) :: _,c2)
      equation 
        Static.eq_cref(c1, c2); then ();
    case (_ :: cs,c2)
      equation 
        find_in_set_flow(cs, c2); then ();
  end matchcontinue;
end find_in_set_flow;

protected function new_equ_set " relation: new_equ_set
 
  This function creates a new non-flow connection set containing
  only the given component.
 "
  input Exp.ComponentRef in_componentref;
  output Set out_set;
algorithm 
  out_set:=
  matchcontinue (in_componentref)
    local Exp.ComponentRef c;
    case c then EQU({c}); 
  end matchcontinue;
end new_equ_set;

protected function new_flow_set " relation: new_flow_set
 
  This function creates a new-flow connection set containing only
  the given component.
 "
  input Exp.ComponentRef in_componentref;
  input Face in_face;
  output Set out_set;
algorithm 
  out_set:=
  matchcontinue (in_componentref,in_face)
    local
      Exp.ComponentRef c;
      Face d;
    case (c,d) then FLOW({(c,d)}); 
  end matchcontinue;
end new_flow_set;

protected function merge "
  - Merging
  
  The result of merging two connection sets is the intersection of
  the two sets.
 "
  input Sets in_sets1;
  input Set in_set2;
  input Set in_set3;
  output Sets out_sets;
algorithm 
  out_sets:=
  matchcontinue (in_sets1,in_set2,in_set3)
    local
      Sets ss,ss_1;
      Set s1,s2;
      Exp_ComponentRefList cs,cs1,cs2;
    case (ss,s1,s2)
      equation 
        equality(s1 = s2); then ss;
    case (ss,(s1 = EQU(cs1)),(s2 = EQU(cs2)))
      equation 
        cs = list_append(cs1, cs2);
        ss_1 = remove_set2(ss, s1, s2); then EQU(cs) :: ss_1;
    case (ss,(s1 = FLOW(cs1)),(s2 = FLOW(cs2)))
      local TupleExp_ComponentRefFaceList cs,cs1,cs2;
      equation 
        cs = list_append(cs1, cs2);
        ss_1 = remove_set2(ss, s1, s2); then FLOW(cs) :: ss_1;
  end matchcontinue;
end merge;

protected function remove_set2 " relation: remove_set2
  
  This function removes the two sets given in the second and third
  argument from the collection of sets given in the first argument.
 "
  input Sets in_sets1;
  input Set in_set2;
  input Set in_set3;
  output Sets out_sets;
algorithm 
  out_sets:=
  matchcontinue (in_sets1,in_set2,in_set3)
    local
      Sets ss_1,ss;
      Set s,s1,s2;
    case ({},_,_) then {}; 
    case (s :: ss,s1,s2)
      equation 
        equality(s = s1);
        ss_1 = remove_set(ss, s2); then ss_1;
    case (s :: ss,s1,s2)
      equation 
        equality(s = s2);
        ss_1 = remove_set(ss, s1); then ss_1;
    case (s :: ss,s1,s2)
      equation 
        ss_1 = remove_set2(ss, s1, s2); then s :: ss_1;
  end matchcontinue;
end remove_set2;

protected function remove_set " relation: remove_set
 
  This function removes one set from a list of sets.
 "
  input Sets in_sets;
  input Set in_set;
  output Sets out_sets;
algorithm 
  out_sets:=
  matchcontinue (in_sets,in_set)
    local
      Set s,s1;
      Sets ss,ss_1;
    case ({},_) then {}; 
    case (s :: ss,s1)
      equation 
        equality(s = s1); then ss;
    case (s :: ss,s1)
      equation 
        ss_1 = remove_set(ss, s1); then s :: ss_1;
  end matchcontinue;
end remove_set;

public function unconnected_flow_equations "Unconnected flow variables. 
   relation: unconnected_flow_equations 
 
  This function will generate set-to-zero equations for INNER flow variables.
  It can not generate for outer flow varaibles, since we do not yet know if 
  these are connected or not. This is only known in the preceding recursive 
  call. However, the top call must generate for both inner and outer 
  connectors, hence the last argument, true for top call"
  input Sets in_sets;
  input DAE_ElementList in_dae_elementlist;
  input Env.Env in_env;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_sets,in_dae_elementlist,in_env,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      Exp_ComponentRefList v1,v2,vars,vars2,unconnectedvars;
      DAE_ElementList dae_1,dae;
      Sets csets;
      Env_FrameList env;
    case (csets,dae,env,true)
      equation 
        v1 = Env.local_outside_connector_flowvars(env) "if outermost call look at both inner and outer unconnected connectors ";
        v2 = Env.local_inside_connector_flowvars(env);
        vars = list_append(v1, v2);
        vars2 = get_all_flow_variables(csets);
        unconnectedvars = remove_variables(vars, vars2);
        dae_1 = generate_zeroflow_equations(unconnectedvars); then dae_1;
    case (csets,dae,env,true) "If not outermost call, only look at unconnected inner connectors "
      equation 
        vars = Env.local_inside_connector_flowvars(env);
        vars2 = get_all_flow_variables(csets);
        unconnectedvars = remove_variables(vars, vars2);
        dae_1 = generate_zeroflow_equations(unconnectedvars); then dae_1;
    case (csets,dae,env,_) then {}; 
  end matchcontinue;
end unconnected_flow_equations;

protected function remove_variables
  input Exp_ComponentRefList in_exp_componentreflist1;
  input Exp_ComponentRefList in_exp_componentreflist2;
  output Exp_ComponentRefList out_exp_componentreflist;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_exp_componentreflist1,in_exp_componentreflist2)
    local
      Exp_ComponentRefList vars,vars_1,res,removelist;
      Exp.ComponentRef r1;
    case (vars,{}) then vars; 
    case (vars,r1 :: removelist)
      equation 
        vars_1 = remove_variable(r1, vars);
        res = remove_variables(vars_1, removelist); then res;
  end matchcontinue;
end remove_variables;

protected function remove_variable
  input Exp.ComponentRef in_componentref;
  input Exp_ComponentRefList in_exp_componentreflist;
  output Exp_ComponentRefList out_exp_componentreflist;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_componentref,in_exp_componentreflist)
    local
      Exp.ComponentRef cr,cr2;
      Exp_ComponentRefList xs,res;
    case (cr,{}) then {}; 
    case (cr,cr2 :: xs)
      equation 
        true = Exp.cref_equal(cr, cr2); then xs;
    case (cr,cr2 :: xs)
      equation 
        res = remove_variable(cr, xs); then cr2 :: res;
  end matchcontinue;
end remove_variable;

protected function generate_zeroflow_equations
  input Exp_ComponentRefList in_exp_componentreflist;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_exp_componentreflist)
    local
      DAE_ElementList res;
      Exp.ComponentRef cr;
      Exp_ComponentRefList xs;
    case ({}) then {}; 
    case (cr :: xs)
      equation 
        res = generate_zeroflow_equations(xs); then DAE.EQUATION(Exp.CREF(cr,Exp.REAL()),Exp.RCONST(0.0)) :: res;
  end matchcontinue;
end generate_zeroflow_equations;

protected function get_all_flow_variables
  input Sets in_sets;
  output Exp_ComponentRefList out_exp_componentreflist;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_sets)
    local
      Exp_ComponentRefList res1,res2,res;
      TupleExp_ComponentRefFaceList varlst;
      Sets xs;
    case {} then {}; 
    case (FLOW(varlst) :: xs)
      equation 
        res1 = Util.list_map(varlst, Util.tuple2_1);
        res2 = get_all_flow_variables(xs);
        res = list_append(res1, res2); then res;
    case (EQU(res1) :: xs)
      equation 
        res = get_all_flow_variables(xs); then res;
  end matchcontinue;
end get_all_flow_variables;

protected function get_inner_flow_variables
  input Sets in_sets;
  output Exp_ComponentRefList out_exp_componentreflist;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_sets)
    local
      Exp_ComponentRefList res1,res2,res;
      TupleExp_ComponentRefFaceList vars;
      Sets xs;
    case ({}) then {}; 
    case (FLOW(vars) :: xs)
      equation 
        res1 = get_inner_flow_variables_2(vars);
        res2 = get_inner_flow_variables(xs);
        res = list_append(res1, res2); then res;
    case (EQU(_) :: xs)
      equation 
        res = get_inner_flow_variables(xs); then res;
    case (_) then fail(); 
  end matchcontinue;
end get_inner_flow_variables;

protected function get_inner_flow_variables_2
  input TupleExp_ComponentRefFaceList in_tupleexp_componentreffacelist;
  output Exp_ComponentRefList out_exp_componentreflist;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_tupleexp_componentreffacelist)
    local
      Exp_ComponentRefList res;
      Exp.ComponentRef cr;
      TupleExp_ComponentRefFaceList xs;
    case ({}) then {}; 
    case ((cr,INNER()) :: xs)
      equation 
        res = get_inner_flow_variables_2(xs); then cr :: res;
    case (_ :: xs)
      equation 
        res = get_inner_flow_variables_2(xs); then res;
    case (_) then fail(); 
  end matchcontinue;
end get_inner_flow_variables_2;

protected import OpenModelica.Compiler.Dump "
  - Printing
 
  These are a few functions used for printing a description of the
  connection sets.  The implementation is excluded from the report
  for brevity.
 
!ignorecode";

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Debug;

public function print_sets " relation: print_sets
 
  Prints a description of a number of connection sets to the
  standard output.
 "
  input Sets in_sets;
algorithm 
  _:=
  matchcontinue (in_sets)
    local
      Set x;
      Sets xs;
    case {} then (); 
    case x :: xs
      equation 
        print_set(x);
        print_sets(xs); then ();
  end matchcontinue;
end print_sets;

protected function print_set
  input Set in_set;
algorithm 
  _:=
  matchcontinue (in_set)
    local Exp_ComponentRefList cs;
    case EQU(cs)
      equation 
        Print.print_buf(" non-flow set: { ");
        Exp.print_list(cs, Exp.print_component_ref, ", ");
        Print.print_buf(" }\n"); then ();
    case FLOW(cs)
      local TupleExp_ComponentRefFaceList cs;
      equation 
        Print.print_buf(" flow set: { ");
        Exp.print_list(cs, print_flow_ref, ", ");
        Print.print_buf(" }\n"); then ();
  end matchcontinue;
end print_set;

protected function print_flow_ref
  input TupleExp_ComponentRefFace in_tupleexp_componentrefface;
algorithm 
  _:=
  matchcontinue (in_tupleexp_componentrefface)
    local Exp.ComponentRef c;
    case ((c,INNER()))
      equation 
        Exp.print_component_ref(c);
        Print.print_buf(" INNER"); then ();
    case ((c,OUTER()))
      equation 
        Exp.print_component_ref(c);
        Print.print_buf(" OUTER"); then ();
  end matchcontinue;
end print_flow_ref;
end Connect;

