package DAE "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 DAE.rml
  module:      DAE
  description: DAE management and output
 
  RCS: $Id: DAE.rml 1821 2005-06-21 12:12:32Z petar $
  
  This module defines data structures for DAE equations and 
  declarations of variables and functions. It also exports some help
  relations for other modules. The DAE data structure is the result of
  flattening, containing only flat modelica, i.e. equations, algorithms,
  variables and functions. 
 
  
 
   - Module header "

public import OpenModelica.Compiler.Absyn;

public import OpenModelica.Compiler.Exp;

public import OpenModelica.Compiler.Algorithm;

public import OpenModelica.Compiler.Types;

public import OpenModelica.Compiler.Values;

public import OpenModelica.Compiler.ClassInf;

public 
type Ident = String;

public 
type InstDims = list<Exp.Subscript>;

public 
type StartValue = Option<Exp.Exp>;

public 
uniontype VarKind
  record VARIABLE
  end VARIABLE;
  record DISCRETE
  end DISCRETE;
  record PARAM
  end PARAM;
  record CONST
  end CONST;
end VarKind;

public 
type StringList = list<String>;

uniontype Type
  record REAL
  end REAL;
  record INT
  end INT;
  record BOOL
  end BOOL;
  record STRING
  end STRING;
  record ENUM
  end ENUM;
  record ENUMERATION
    StringList stringlist;
  end ENUMERATION;
end Type;

public 
uniontype Flow
  record FLOW
  end FLOW;
  record NON_FLOW
  end NON_FLOW;
  record NON_CONNECTOR
  end NON_CONNECTOR;
end Flow;

public 
uniontype VarDirection
  record INPUT
  end INPUT;
  record OUTPUT
  end OUTPUT;
  record BIDIR
  end BIDIR;
end VarDirection;

public 
type Exp_ExpOption = Option<Exp.Exp>;

type Absyn_PathList = list<Absyn.Path>;

type VariableAttributesOption = Option<VariableAttributes>;

type Absyn_CommentOption = Option<Absyn.Comment>;

type ElementList = list<Element>;

type ElementOption = Option<Element>;

uniontype Element "LS: Removed VARVAL completely, and added Exp.Exp as an option to VAR "
  record VAR
    Exp.ComponentRef componentref;
    VarKind varkind "varible name ";
    VarDirection vardirection "variable, constant, parameter, etc. ";
    Type type_ "input, output or bidir ";
    Exp_ExpOption exp_expoption "one of the builtin types ";
    InstDims instdims "Binding expression e.g. for parameters";
    StartValue startvalue "dimension of original component ";
    Flow flow_ "value of start attribute ";
    Absyn_PathList absyn_pathlist "Flow of connector variable. Needed for 
						unconnected flow variables ";
    VariableAttributesOption variableattributesoption;
    Absyn_CommentOption absyn_commentoption;
  end VAR;
  record DEFINE
    Exp.ComponentRef componentref;
    Exp.Exp exp;
  end DEFINE;
  record INITIALDEFINE
    Exp.ComponentRef componentref;
    Exp.Exp exp;
  end INITIALDEFINE;
  record EQUATION
    Exp.Exp exp1;
    Exp.Exp exp2;
  end EQUATION;
  record WHEN_EQUATION
    Exp.Exp condition "Condition ";
    ElementList equations "Equations ";
    ElementOption elementoption "Elsewhen should be of type WHEN_EQUATION";
  end WHEN_EQUATION;
  record IF_EQUATION
    Exp.Exp condition1 "Condition ";
    ElementList elementlist2 "Equations of true branch";
    ElementList elementlist3 "Equations of false branch";
  end IF_EQUATION;
  record INITIALEQUATION
    Exp.Exp exp1;
    Exp.Exp exp2;
  end INITIALEQUATION;
  record ALGORITHM
    Algorithm.Algorithm algorithm_;
  end ALGORITHM;
  record INITIALALGORITHM
    Algorithm.Algorithm algorithm_;
  end INITIALALGORITHM;
  record COMP
    Ident ident;
    DAElist daelist;
  end COMP;
  record FUNCTION
    Absyn.Path path;
    DAElist daelist;
    Types.Type type_;
  end FUNCTION;
  record EXTFUNCTION
    Absyn.Path path;
    DAElist daelist;
    Types.Type type_;
    ExternalDecl externaldecl;
  end EXTFUNCTION;
  record ASSERT
    Exp.Exp exp;
  end ASSERT;
  record REINIT
    Exp.ComponentRef componentref;
    Exp.Exp exp;
  end REINIT;
end Element;

public 
type StringOption = Option<String>;

type RealOption = Option<Real>;

type TupleRealOptionRealOption = tuple<RealOption,RealOption>;

type BooleanOption = Option<Boolean>;

type StateSelectOption = Option<StateSelect>;

type IntegerOption = Option<Integer>;

type TupleIntegerOptionIntegerOption = tuple<IntegerOption,IntegerOption>;

type TupleExp_ExpOptionExp_ExpOption = tuple<Exp_ExpOption,Exp_ExpOption>;

uniontype VariableAttributes
  record VAR_ATTR_REAL
    StringOption stringoption1;
    StringOption quantity2 "quantity ";
    StringOption unit3 "unit ";
    TupleRealOptionRealOption displayunit4 "displayUnit ";
    RealOption realoption5 "min , max ";
    BooleanOption booleanoption6 "Initial value ";
    RealOption realoption7 "fixed - true: default for parameter/constant, false - default for other variables ";
    StateSelectOption nominal8 "nominal ";
  end VAR_ATTR_REAL;
  record VAR_ATTR_INT
    StringOption stringoption;
    TupleIntegerOptionIntegerOption quantity "quantity ";
    IntegerOption integeroption "min , max ";
    BooleanOption booleanoption "Initial value  fixed - true: default for parameter/constant, false - default for other variables ";
  end VAR_ATTR_INT;
  record VAR_ATTR_BOOL
    StringOption stringoption;
    BooleanOption quantity "quantity ";
    BooleanOption booleanoption "Initial value  fixed - true: default for parameter/constant, false - default for other variables ";
  end VAR_ATTR_BOOL;
  record VAR_ATTR_STRING
    StringOption stringoption1;
    StringOption stringoption2 "quantity  Initial value ";
  end VAR_ATTR_STRING;
  record VAR_ATTR_ENUMERATION
    StringOption stringoption;
    TupleExp_ExpOptionExp_ExpOption quantity "quantity ";
    Exp_ExpOption exp_expoption "min , max ";
    BooleanOption booleanoption "start fixed - true: default for parameter/constant, false - default for other variables ";
  end VAR_ATTR_ENUMERATION;
end VariableAttributes;

public 
uniontype StateSelect
  record NEVER
  end NEVER;
  record AVOID
  end AVOID;
  record DEFAULT
  end DEFAULT;
  record PREFER
  end PREFER;
  record ALWAYS
  end ALWAYS;
end StateSelect;

public 
uniontype ExtArg
  record EXTARG
    Exp.ComponentRef componentref;
    Types.Attributes attributes;
    Types.Type type_;
  end EXTARG;
  record EXTARGEXP
    Exp.Exp exp;
    Types.Type type_;
  end EXTARGEXP;
  record EXTARGSIZE
    Exp.ComponentRef componentref;
    Types.Attributes attributes;
    Types.Type type_;
    Exp.Exp exp;
  end EXTARGSIZE;
  record NOEXTARG
  end NOEXTARG;
end ExtArg;

public 
type ExtArgList = list<ExtArg>;

uniontype ExternalDecl
  record EXTERNALDECL
    Ident ident;
    ExtArgList extarglist "external function name ";
    ExtArg parameters "parameters ";
    String string "return type  language ";
  end EXTERNALDECL;
end ExternalDecl;

public 
uniontype DAElist "
 LS: Could have extended DAElist to have two lists, one with 
 functions and one with elements, but that would mean a lot of
 change in places where two daes are built and appended, in 
 Inst.rml 
 FUNCTION could also consist of SCode.Class instead.
 "
  record DAE
    ElementList elementlist;
  end DAE;
end DAElist;

protected import OpenModelica.Compiler.RTOpts " - Relations
 
  The only functions in this module are for printing, and are not
  included in the report.
 
!ignorecode";

protected import OpenModelica.Compiler.Graphviz;

protected import OpenModelica.Compiler.Dump;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Ceval;

protected import OpenModelica.Compiler.ModUtil;

protected import OpenModelica.Compiler.Debug;

public function dump " relation: dump
 
  This function prints the DAE in the standard output format.
 "
  input DAElist in_daelist;
algorithm 
  _:=
  matchcontinue (in_daelist)
    local ElementList daelist;
    case DAE(daelist)
      equation 
        dump_list(daelist, dump_function);
        dump_list(daelist, dump_comp_element); then ();
  end matchcontinue;
end dump;

public function dump2
  input DAElist in_daelist;
algorithm 
  _:=
  matchcontinue (in_daelist)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Ident comment_str,ident,str,extdeclstr;
      Exp.ComponentRef cr;
      Exp.Exp e,e1,e2;
      InstDims dims;
      Exp_ExpOption start;
      VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      ElementList xs;
      DAElist lst,dae;
      Absyn.Path path;
      TupleTypes_TTypeAbsyn_PathOption tp;
      ExternalDecl extdecl;
    case DAE(VAR(cr,_,_,_,SOME(e),dims,start,_,_,dae_var_attr,comment) :: xs)
      equation 
        Print.print_buf("VAR(");
        Exp.print_component_ref(cr);
        Print.print_buf("=");
        Exp.print_exp(e);
        Print.print_buf(",dims=");
        Dump.print_list(dims, Exp.print_subscript, ", ");
        comment_str = Dump.unparse_comment_option(comment) "	dump_start_value start & ";
        print("  comment:");
        print(comment_str);
        print(",\n ");
        dump_variable_attributes(dae_var_attr);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case DAE(VAR(cr,_,_,_,NONE,_,start,_,_,dae_var_attr,comment) :: xs)
      equation 
        Print.print_buf("VAR(");
        Exp.print_component_ref(cr);
        comment_str = Dump.unparse_comment_option(comment) "	dump_start_value start & ";
        print("  comment:");
        print(comment_str);
        print(",\n ");
        dump_variable_attributes(dae_var_attr);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case DAE(DEFINE(cr,_) :: xs)
      equation 
        Print.print_buf("DEFINE(");
        Exp.print_component_ref(cr);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case DAE(INITIALDEFINE(cr,_) :: xs)
      equation 
        Print.print_buf("INITIALDEFINE(");
        Exp.print_component_ref(cr);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case DAE(EQUATION(e1,e2) :: xs)
      equation 
        Print.print_buf("EQUATION(");
        Exp.print_exp(e1);
        Print.print_buf(" = ");
        Exp.print_exp(e2);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case DAE(INITIALEQUATION(e1,e2) :: xs)
      equation 
        Print.print_buf("INITIALEQUATION(");
        Exp.print_exp(e1);
        Print.print_buf(" = ");
        Exp.print_exp(e2);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case (DAE(ALGORITHM(_) :: xs))
      equation 
        Print.print_buf("ALGORITHM(...)");
        dump2(DAE(xs)); then ();
    case (DAE(INITIALALGORITHM(_) :: xs))
      equation 
        Print.print_buf("INITIALALGORITHM(...)");
        dump2(DAE(xs)); then ();
    case (DAE(COMP(ident,lst) :: xs))
      equation 
        Print.print_buf("COMP(");
        Print.print_buf(ident);
        dump2(lst);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case (DAE(FUNCTION(_,_,_) :: xs))
      equation 
        Print.print_buf("FUNCTION(...)\n");
        dump2(DAE(xs)); then ();
    case (DAE(EXTFUNCTION(path,dae,tp,extdecl) :: xs))
      equation 
        Print.print_buf("EXTFUNCTION(\n");
        str = Absyn.path_string(path);
        Print.print_buf(str);
        Print.print_buf(", ");
        dump2(dae);
        Print.print_buf(", ");
        Types.print_type(tp);
        Print.print_buf(", ");
        extdeclstr = dump_ext_decl_str(extdecl);
        Print.print_buf(extdeclstr);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case (DAE(ASSERT(e) :: xs))
      equation 
        Print.print_buf("ASSERT(\n");
        Exp.print_exp(e);
        Print.print_buf(")\n");
        dump2(DAE(xs)); then ();
    case (DAE({})) then (); 
    case (_)
      equation 
        Print.print_buf("dump2 failed\n"); then ();
  end matchcontinue;
end dump2;

protected function dump_start_value
  input StartValue in_startvalue;
algorithm 
  _:=
  matchcontinue (in_startvalue)
    local Exp.Exp e;
    case (SOME(e))
      equation 
        Print.print_buf("(start=");
        Exp.print_exp(e);
        Print.print_buf(")"); then ();
    case (_) then (); 
  end matchcontinue;
end dump_start_value;

protected function dump_start_value_str
  input StartValue in_startvalue;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_startvalue)
    local
      Ident s,res;
      Exp.Exp e;
    case (SOME(e))
      equation 
        s = Exp.print_exp_str(e);
        res = Util.string_append_list({"(start=",s,")"}); then res;
    case (_) then ""; 
  end matchcontinue;
end dump_start_value_str;

public function dump_ext_decl_str
  input ExternalDecl in_externaldecl;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_externaldecl)
    local
      Ident extargsstr,rettystr,str,id,lang;
      ExtArgList extargs;
      ExtArg retty;
    case EXTERNALDECL(id,extargs,retty,lang)
      equation 
        extargsstr = Dump.get_string_list(extargs, dump_ext_arg_str, ",");
        rettystr = dump_ext_arg_str(retty);
        str = Util.string_append_list(
          {"EXTERNALDECL(",id,", (",extargsstr,"), ",rettystr,", \"",
          lang,"\")"}); then str;
  end matchcontinue;
end dump_ext_decl_str;

public function dump_ext_arg_str
  input ExtArg in_extarg;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_extarg)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Ident crstr,dirstr,tystr,str,dimstr;
      Exp.ComponentRef cr;
      Boolean fl;
      SCode.Accessibility acc;
      SCode.Variability var;
      Absyn.Direction dir;
      TupleTypes_TTypeAbsyn_PathOption ty;
      Exp.Exp exp,dim;
      Types.Attributes attr;
    case NOEXTARG() then "void"; 
    case EXTARG(cr,Types.ATTR(fl,acc,var,dir),ty)
      equation 
        crstr = Exp.print_component_ref_str(cr);
        dirstr = Dump.direction_symbol(dir);
        tystr = Types.get_type_name(ty);
        str = Util.string_append_list({dirstr," ",tystr," ",crstr}); then str;
    case EXTARGEXP(exp,ty)
      equation 
        crstr = Exp.print_exp_str(exp);
        tystr = Types.get_type_name(ty);
        str = Util.string_append_list({"(",tystr,") ",crstr}); then str;
    case EXTARGSIZE(cr,attr,ty,dim)
      equation 
        crstr = Exp.print_component_ref_str(cr);
        dimstr = Exp.print_exp_str(dim);
        str = Util.string_append_list({"size(",crstr,",",dimstr,")"}); then str;
  end matchcontinue;
end dump_ext_arg_str;

public function dump_str " relation: dump_str
  
  This function prints the DAE to a string.
"
  input DAElist in_daelist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_daelist)
    local
      type IdentList = list<Ident>;
      IdentList flist,clist,slist;
      Ident str;
      ElementList daelist;
    case DAE(daelist)
      equation 
        flist = Util.list_map(daelist, dump_function_str);
        clist = Util.list_map(daelist, dump_comp_element_str);
        slist = list_append(flist, clist);
        str = Util.string_append_list(slist); then str;
  end matchcontinue;
end dump_str;

protected function dump_list
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
  end FuncType1;
algorithm 
  _:=
  matchcontinue (in_vtype_alist,in_functype1)
    local
      Type_a first;
      VType_aList rest;
      FuncType1 r;
    case ({},_) then (); 
    case (first :: rest,r)
      equation 
        r(first);
        dump_list(rest, r); then ();
  end matchcontinue;
end dump_list;

protected function dump_list_str
  input ElementList daelist;
  input FuncType1 func;
  output String str;
protected 
  function FuncType1
    input Element in_element;
    output String out_string;
  end FuncType1;
  type IdentList = list<Ident>;
  IdentList slist;
algorithm 
  slist := Util.list_map(daelist, func);
  str := Util.string_append_list(slist);
end dump_list_str;

protected function dump_comp_element
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      Ident n;
      ElementList l;
    case COMP(n,DAE(l))
      equation 
        false = RTOpts.modelica_output();
        Print.print_buf("fclass ");
        Print.print_buf(n);
        Print.print_buf("\n");
        dump_elements(l);
        Print.print_buf("end ");
        Print.print_buf(n);
        Print.print_buf(";\n"); then ();
    case COMP(n,DAE(l))
      equation 
        true = RTOpts.modelica_output();
        Print.print_buf("class ");
        Print.print_buf(n);
        Print.print_buf("\n");
        dump_elements(l);
        Print.print_buf("end ");
        Print.print_buf(n);
        Print.print_buf(";\n"); then ();
    case _ then (); 
  end matchcontinue;
end dump_comp_element;

protected function dump_comp_element_str
  input Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_element)
    local
      Ident s1,s2,s3,s4,s5,s6,str,n;
      ElementList l;
    case COMP(n,DAE(l))
      equation 
        false = RTOpts.modelica_output();
        s1 = string_append("fclass ", n);
        s2 = string_append(s1, "\n");
        s3 = dump_elements_str(l);
        s4 = string_append(s2, s3);
        s5 = string_append(s4, "end ");
        s6 = string_append(s5, n);
        str = string_append(s6, ";\n"); then str;
    case COMP(n,DAE(l))
      equation 
        true = RTOpts.modelica_output();
        s1 = string_append("class ", n);
        s2 = string_append(s1, "\n");
        s3 = dump_elements_str(l);
        s4 = string_append(s2, s3);
        s5 = string_append(s4, "end ");
        s6 = string_append(s5, n);
        str = string_append(s6, ";\n"); then str;
    case _ then ""; 
  end matchcontinue;
end dump_comp_element_str;

public function dump_elements
  input ElementList l;
algorithm 
  dump_vars(l);
  Print.print_buf("initial equation\n");
  dump_list(l, dump_initialequation);
  Print.print_buf("equation\n");
  dump_list(l, dump_equation);
  dump_list(l, dump_initialalgorithm);
  dump_list(l, dump_algorithm);
  dump_list(l, dump_comp_element);
end dump_elements;

protected function dump_elements_str
  input ElementList l;
  output String str;
protected 
  Ident s1,s2,s3,s4,s5,initeqstr,initalgstr,eqstr,algstr;
  Boolean noiniteq,noinitalg,noeq,noalg;
algorithm 
  s1 := dump_vars_str(l);
  s2 := dump_initialequations_str(l);
  s3 := dump_equations_str(l);
  s4 := dump_initialalgorithms_str(l);
  s5 := dump_algorithms_str(l);
  noiniteq := Util.string_equal(s2, "");
  noinitalg := Util.string_equal(s4, "");
  noeq := Util.string_equal(s3, "");
  noalg := Util.string_equal(s5, "");
  initeqstr := Dump.select_string(noiniteq, "", "initial equation\n");
  initalgstr := Dump.select_string(noinitalg, "", "initial algorithm\n");
  eqstr := Dump.select_string(noeq, "", "equation\n");
  algstr := Dump.select_string(noalg, "", "algorithm\n");
  str := Util.string_append_list({s1,initeqstr,s2,initalgstr,s4,eqstr,s3,algstr,s5});
end dump_elements_str;

protected function dump_algorithms_str
  input ElementList in_elementlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementlist)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Ident s1,s2,s3,str;
      Algorithm_StatementList stmts;
      ElementList xs;
    case (ALGORITHM(Algorithm.ALGORITHM(stmts)) :: xs)
      equation 
        s1 = Dump.get_string_list(stmts, pp_statement_str, "");
        s2 = string_append("algorithm\n", s1);
        s3 = dump_algorithms_str(xs);
        str = string_append(s1, s3); then str;
    case (_ :: xs)
      equation 
        str = dump_algorithms_str(xs); then str;
    case ({}) then ""; 
  end matchcontinue;
end dump_algorithms_str;

protected function dump_initialalgorithms_str
  input ElementList in_elementlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementlist)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Ident s1,s2,s3,str;
      Algorithm_StatementList stmts;
      ElementList xs;
    case (INITIALALGORITHM(Algorithm.ALGORITHM(stmts)) :: xs)
      equation 
        s1 = Dump.get_string_list(stmts, pp_statement_str, "");
        s2 = string_append("algorithm\n", s1);
        s3 = dump_initialalgorithms_str(xs);
        str = string_append(s1, s3); then str;
    case (_ :: xs)
      equation 
        str = dump_initialalgorithms_str(xs); then str;
    case ({}) then ""; 
  end matchcontinue;
end dump_initialalgorithms_str;

protected function dump_equations_str
  input ElementList in_elementlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementlist)
    local
      Ident s1,s2,s3,s4,s4_1,s5,s6,str,s;
      Exp.Exp e1,e2,e;
      ElementList xs,xs1,xs2;
      Exp.ComponentRef c,cr;
    case (EQUATION(e1,e2) :: xs)
      equation 
        s1 = Exp.print_exp_str(e1);
        s2 = string_append("  ", s1);
        s3 = string_append(s2, " = ");
        s4 = Exp.print_exp_str(e2);
        s4_1 = string_append(s3, s4);
        s5 = string_append(s4_1, ";\n");
        s6 = dump_equations_str(xs);
        str = string_append(s5, s6); then str;
    case (DEFINE(c,e) :: xs)
      equation 
        s1 = Exp.print_component_ref_str(c);
        s2 = string_append("  ", s1);
        s3 = string_append(s2, " = ");
        s4 = Exp.print_exp_str(e);
        s4_1 = string_append(s3, s4);
        s5 = string_append(s4_1, ";\n");
        s6 = dump_equations_str(xs);
        str = string_append(s5, s6); then str;
    case (ASSERT(e) :: xs)
      equation 
        s = Exp.print_exp_str(e);
        s2 = dump_equations_str(xs);
        str = Util.string_append_list({s,";\n",s2}); then str;
    case (IF_EQUATION(c,xs1,xs2) :: xs)
      local Exp.Exp c;
      equation 
        s = Exp.print_exp_str(c);
        s1 = dump_equations_str(xs1);
        s2 = dump_equations_str(xs2);
        s3 = dump_equations_str(xs);
        str = Util.string_append_list({"  if ",s," then\n",s1,"  else\n",s2,"  end if;\n",s3}); then str;
    case (WHEN_EQUATION(c,xs1,SOME(xs2)) :: xs)
      local
        Exp.Exp c;
        Element xs2;
      equation 
        s = Exp.print_exp_str(c);
        s1 = dump_equations_str(xs1);
        s2 = dump_equations_str(xs2 :: xs);
        str = Util.string_append_list({"when ",s," then\n",s1,"  else",s2}); then str;
    case (WHEN_EQUATION(c,xs1,NONE) :: xs)
      local Exp.Exp c;
      equation 
        s = Exp.print_exp_str(c);
        s1 = dump_equations_str(xs1);
        s3 = dump_equations_str(xs);
        str = Util.string_append_list({"  when ",s," then\n",s1,"  end when;\n",s3}); then str;
    case (REINIT(cr,e) :: xs)
      equation 
        s = Exp.print_component_ref_str(cr);
        s1 = Exp.print_exp_str(e);
        s2 = dump_equations_str(xs);
        str = Util.string_append_list({"  reinit(",s,",",s1,");\n",s2}); then str;
    case (_ :: xs)
      equation 
        str = dump_equations_str(xs); then str;
    case ({}) then ""; 
  end matchcontinue;
end dump_equations_str;

protected function dump_initialequations_str
  input ElementList in_elementlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementlist)
    local
      Ident s1,s2,s3,s4,s4_1,s5,s6,str;
      Exp.Exp e1,e2,e;
      ElementList xs;
      Exp.ComponentRef c;
    case (INITIALEQUATION(e1,e2) :: xs)
      equation 
        s1 = Exp.print_exp_str(e1);
        s2 = string_append("  ", s1);
        s3 = string_append(s2, " = ");
        s4 = Exp.print_exp_str(e2);
        s4_1 = string_append(s3, s4);
        s5 = string_append(s4_1, ";\n");
        s6 = dump_equations_str(xs);
        str = string_append(s5, s6); then str;
    case (INITIALDEFINE(c,e) :: xs)
      equation 
        s1 = Exp.print_component_ref_str(c);
        s2 = string_append("  ", s1);
        s3 = string_append(s2, " := ");
        s4 = Exp.print_exp_str(e);
        s4_1 = string_append(s3, s4);
        s5 = string_append(s4_1, ";\n");
        s6 = dump_initialequations_str(xs);
        str = string_append(s5, s6); then str;
    case (_ :: xs)
      equation 
        str = dump_initialequations_str(xs); then str;
    case ({}) then ""; 
  end matchcontinue;
end dump_initialequations_str;

protected function dump_vars "replace this with dump_list when everything works after the VARVAL removal 
  project "
  input ElementList in_elementlist;
algorithm 
  _:=
  matchcontinue (in_elementlist)
    local
      Element first;
      ElementList rest;
    case {} then (); 
    case first :: rest
      equation 
        dump_var(first);
        dump_vars(rest); then ();
  end matchcontinue;
end dump_vars;

protected function dump_vars_str
  input ElementList in_elementlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementlist)
    local
      Ident s1,s2,str;
      Element first;
      ElementList rest;
    case {} then ""; 
    case first :: rest
      equation 
        s1 = dump_var_str(first);
        s2 = dump_vars_str(rest);
        str = string_append(s1, s2); then str;
  end matchcontinue;
end dump_vars_str;

protected function dump_kind
  input VarKind in_varkind;
algorithm 
  _:=
  matchcontinue (in_varkind)
    case CONST()
      equation 
        Print.print_buf(" constant  "); then ();
    case PARAM()
      equation 
        Print.print_buf(" parameter "); then ();
    case DISCRETE()
      equation 
        Print.print_buf(" discrete  "); then ();
    case VARIABLE()
      equation 
        Print.print_buf("           "); then ();
  end matchcontinue;
end dump_kind;

protected function dump_kind_str
  input VarKind in_varkind;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_varkind)
    case CONST() then "constant "; 
    case PARAM() then "parameter "; 
    case DISCRETE() then "discrete "; 
    case VARIABLE() then ""; 
  end matchcontinue;
end dump_kind_str;

protected function dump_direction
  input VarDirection in_vardirection;
algorithm 
  _:=
  matchcontinue (in_vardirection)
    case INPUT()
      equation 
        Print.print_buf(" input  "); then ();
    case OUTPUT()
      equation 
        Print.print_buf(" output "); then ();
    case BIDIR()
      equation 
        Print.print_buf("        "); then ();
  end matchcontinue;
end dump_direction;

public function dump_direction_str
  input VarDirection in_vardirection;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_vardirection)
    case INPUT() then "input "; 
    case OUTPUT() then "output "; 
    case BIDIR() then ""; 
  end matchcontinue;
end dump_direction_str;

protected function dump_stateSelect_str
  input StateSelect in_stateselect;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_stateselect)
    case (NEVER()) then "StateSelect.never"; 
    case (AVOID()) then "StateSelect.avoid"; 
    case (PREFER()) then "StateSelect.prefer"; 
    case (ALWAYS()) then "StateSelect.always"; 
    case (DEFAULT()) then "StateSelect.default"; 
  end matchcontinue;
end dump_stateSelect_str;

public function dump_variable_attributes
  input VariableAttributesOption attr;
protected 
  Ident res;
algorithm 
  res := dump_variable_attributes_str(attr);
  Print.print_buf(res);
end dump_variable_attributes;

protected function string_to_string
  input String str;
  output String str_1;
protected 
  Ident str_1;
algorithm 
  str_1 := Util.string_append_list({"\"",str,"\""});
end string_to_string;

public function dump_variable_attributes_str
  input VariableAttributesOption in_variableattributesoption;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_variableattributesoption)
    local
      type IdentOption = Option<Ident>;
      Ident quantity,unit_str,displayUnit_str,stateSel_str,min_str,max_str,nominal_str,Initial_str,fixed_str,res_1,res1,res;
      Boolean is_empty;
      IdentOption quant,unit,displayUnit;
      RealOption min,max,Initial,nominal;
      BooleanOption fixed;
      StateSelectOption stateSel;
    case (SOME(VAR_ATTR_REAL(quant,unit,displayUnit,(min,max),Initial,fixed,nominal,stateSel)))
      equation 
        quantity = Dump.get_option_with_concat_str(quant, string_to_string, "quantity = ");
        unit_str = Dump.get_option_with_concat_str(unit, string_to_string, "unit = ");
        displayUnit_str = Dump.get_option_with_concat_str(displayUnit, string_to_string, "displayUnit = ");
        stateSel_str = Dump.get_option_with_concat_str(stateSel, dump_stateSelect_str, "StateSelect = ");
        min_str = Dump.get_option_with_concat_str(min, real_string, "min = ");
        max_str = Dump.get_option_with_concat_str(max, real_string, "max = ");
        nominal_str = Dump.get_option_with_concat_str(nominal, real_string, "nominal = ");
        Initial_str = Dump.get_option_with_concat_str(Initial, real_string, "start = ");
        fixed_str = Dump.get_option_with_concat_str(fixed, Dump.print_bool_str, "fixed = ");
        res_1 = Util.string_delimit_list_no_empty(
          {quantity,unit_str,displayUnit_str,min_str,max_str,
          Initial_str,fixed_str,nominal_str,stateSel_str}, ", ");
        res1 = Util.string_append_list({"(",res_1,")"});
        is_empty = Util.is_empty_string(res_1);
        res = Util.if_(is_empty, "", res1); then res;
    case (SOME(VAR_ATTR_INT(quant,(min,max),Initial,fixed)))
      local IntegerOption min,max,Initial;
      equation 
        quantity = Dump.get_option_with_concat_str(quant, string_to_string, "quantity = ");
        min_str = Dump.get_option_with_concat_str(min, int_string, "min = ");
        max_str = Dump.get_option_with_concat_str(max, int_string, "max = ");
        Initial_str = Dump.get_option_with_concat_str(Initial, int_string, "start = ");
        fixed_str = Dump.get_option_with_concat_str(fixed, Dump.print_bool_str, "fixed = ");
        res_1 = Util.string_delimit_list_no_empty({quantity,min_str,max_str,Initial_str,fixed_str}, ", ");
        res1 = Util.string_append_list({"(",res_1,")"});
        is_empty = Util.is_empty_string(res_1);
        res = Util.if_(is_empty, "", res1); then res;
    case (SOME(VAR_ATTR_BOOL(quant,Initial,fixed)))
      local BooleanOption Initial;
      equation 
        quantity = Dump.get_option_with_concat_str(quant, string_to_string, "quantity = ");
        Initial_str = Dump.get_option_with_concat_str(Initial, Dump.print_bool_str, "start = ");
        fixed_str = Dump.get_option_with_concat_str(fixed, Dump.print_bool_str, "fixed = ");
        res_1 = Util.string_delimit_list_no_empty({quantity,Initial_str,fixed_str}, ", ");
        res1 = Util.string_append_list({"(",res_1,")"});
        is_empty = Util.is_empty_string(res_1);
        res = Util.if_(is_empty, "", res1); then res;
    case (SOME(VAR_ATTR_STRING(quant,Initial)))
      local IdentOption Initial;
      equation 
        quantity = Dump.get_option_with_concat_str(quant, string_to_string, "quantity = ");
        Initial_str = Dump.get_option_with_concat_str(Initial, string_to_string, "start = ");
        res_1 = Util.string_delimit_list_no_empty({quantity,Initial_str}, ", ");
        res1 = Util.string_append_list({"(",res_1,")"});
        is_empty = Util.is_empty_string(res_1);
        res = Util.if_(is_empty, "", res1); then res;
    case (SOME(VAR_ATTR_ENUMERATION(quant,(min,max),Initial,fixed)))
      local Exp_ExpOption min,max,Initial;
      equation 
        quantity = Dump.get_option_with_concat_str(quant, string_to_string, "quantity = ");
        min_str = Dump.get_option_with_concat_str(min, Exp.print_exp_str, "min = ");
        max_str = Dump.get_option_with_concat_str(max, Exp.print_exp_str, "max = ");
        Initial_str = Dump.get_option_with_concat_str(Initial, Exp.print_exp_str, "start = ");
        fixed_str = Dump.get_option_with_concat_str(fixed, Dump.print_bool_str, "fixed = ");
        res_1 = Util.string_delimit_list_no_empty({quantity,min_str,max_str,Initial_str,fixed_str}, ", ");
        res1 = Util.string_append_list({"(",res_1,")"});
        is_empty = Util.is_empty_string(res_1);
        res = Util.if_(is_empty, "", res1); then res;
    case (NONE) then ""; 
    case (_) then "unknown VariableAttributes"; 
  end matchcontinue;
end dump_variable_attributes_str;

public function dump_type
  input Type in_type;
algorithm 
  _:=
  matchcontinue (in_type)
    local
      type IdentList = list<Ident>;
      IdentList l;
    case INT()
      equation 
        Print.print_buf("Integer "); then ();
    case REAL()
      equation 
        Print.print_buf("Real    "); then ();
    case BOOL()
      equation 
        Print.print_buf("Boolean "); then ();
    case STRING()
      equation 
        Print.print_buf("String  "); then ();
    case ENUM()
      equation 
        Print.print_buf("Enum "); then ();
    case ENUMERATION(l)
      equation 
        Print.print_buf("Enumeration(");
        Dump.print_list(l, print, ",");
        Print.print_buf(") "); then ();
  end matchcontinue;
end dump_type;

public function dump_type_str
  input Type in_type;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_type)
    local
      type IdentList = list<Ident>;
      Ident s1,s2,str;
      IdentList l;
    case INT() then "Integer "; 
    case REAL() then "Real "; 
    case BOOL() then "Boolean "; 
    case STRING() then "String "; 
    case ENUM() then "Enum "; 
    case ENUMERATION(l)
      equation 
        s1 = Util.string_delimit_list(l, ", ");
        s2 = string_append("enumeration(", s1);
        str = string_append(s2, ") "); then str;
  end matchcontinue;
end dump_type_str;

protected function dump_var
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      Exp.ComponentRef id;
      VarKind kind;
      VarDirection dir;
      Type typ;
      Exp_ExpOption start;
      Flow flow_;
      Absyn_PathList classlst,class_;
      VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      Exp.Exp e;
    case VAR(id,kind,dir,typ,NONE,_,start,flow_,classlst,dae_var_attr,comment)
      equation 
        dump_kind(kind);
        dump_direction(dir);
        dump_type(typ);
        Exp.print_component_ref(id);
        dump_comment_option(comment) "	dump_start_value start & ";
        dump_variable_attributes(dae_var_attr);
        Print.print_buf(";\n"); then ();
    case VAR(id,kind,dir,typ,SOME(e),_,start,flow_,class_,dae_var_attr,comment)
      equation 
        dump_kind(kind);
        dump_direction(dir);
        dump_type(typ);
        Exp.print_component_ref(id);
        dump_variable_attributes(dae_var_attr) "	dump_start_value start & ";
        Print.print_buf(" = ") "	dump_start_value start & ";
        Exp.print_exp(e);
        Print.print_buf(";\n"); then ();
    case (_) then (); 
  end matchcontinue;
end dump_var;

protected function dump_var_str
  input Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_element)
    local
      Ident s1,s2,s3,s4,comment_str,s5,str,s6;
      Exp.ComponentRef id;
      VarKind kind;
      VarDirection dir;
      Type typ;
      Exp_ExpOption start;
      Flow flow_;
      Absyn_PathList classlst;
      VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      Exp.Exp e;
    case VAR(id,kind,dir,typ,NONE,_,start,flow_,classlst,dae_var_attr,comment)
      equation 
        s1 = dump_kind_str(kind);
        s2 = dump_direction_str(dir);
        s3 = dump_type_str(typ);
        s4 = Exp.print_component_ref_str(id);
        comment_str = dump_comment_option_str(comment) "	dump_start_value_str start => s5 & ";
        s5 = dump_variable_attributes_str(dae_var_attr);
        str = Util.string_append_list({s1,s2,s3,s4,s5,comment_str,";\n"}) "	Util.list_map(classlst,Absyn.path_string) => classstrlst & 
	Util.string_delimit_list(classstrlst, \", \") => classstr & "; then str;
    case VAR(id,kind,dir,typ,SOME(e),_,start,flow_,classlst,dae_var_attr,comment)
      equation 
        s1 = dump_kind_str(kind);
        s2 = dump_direction_str(dir);
        s3 = dump_type_str(typ);
        s4 = Exp.print_component_ref_str(id);
        s5 = Exp.print_exp_str(e);
        comment_str = dump_comment_option_str(comment) "	dump_start_value_str start => s6 & ";
        s6 = dump_variable_attributes_str(dae_var_attr);
        str = Util.string_append_list({s1,s2,s3,s4," = ",s5,s6,comment_str,";\n"}) "	Util.list_map(classlst,Absyn.path_string) => classstrlst & 
	Util.string_delimit_list(classstrlst, \", \") => classstr & "; then str;
    case (_) then ""; 
  end matchcontinue;
end dump_var_str;

protected function dump_comment_option_str
  input Absyn_CommentOption in_absyn_commentoption;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_absyn_commentoption)
    local
      type Absyn_AnnotationOption = Option<Absyn.Annotation>;
      Ident str,cmt;
      Absyn_AnnotationOption annopt;
    case (NONE) then ""; 
    case (SOME(Absyn.COMMENT(annopt,SOME(cmt))))
      equation 
        str = Util.string_append_list({" \"",cmt,"\""}); then str;
    case (SOME(Absyn.COMMENT(annopt,NONE))) then ""; 
  end matchcontinue;
end dump_comment_option_str;

protected function dump_comment_option
  input Absyn_CommentOption comment;
protected 
  Ident str;
algorithm 
  str := dump_comment_option_str(comment);
  Print.print_buf(str);
end dump_comment_option;

protected function dump_equation
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      Exp.Exp e1,e2,e;
      Exp.ComponentRef c;
    case (EQUATION(e1,e2))
      equation 
        Print.print_buf("  ");
        Exp.print_exp(e1);
        Print.print_buf(" = ");
        Exp.print_exp(e2);
        Print.print_buf(";\n"); then ();
    case (DEFINE(c,e))
      equation 
        Print.print_buf("  ");
        Exp.print_component_ref(c);
        Print.print_buf(" ::= ");
        Exp.print_exp(e);
        Print.print_buf(";\n"); then ();
    case (ASSERT(e))
      equation 
        Exp.print_exp(e);
        Print.print_buf(";\n"); then ();
    case _ then (); 
  end matchcontinue;
end dump_equation;

protected function dump_initialequation
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      Exp.Exp e1,e2,e;
      Exp.ComponentRef c;
    case (INITIALEQUATION(e1,e2))
      equation 
        Print.print_buf("  ");
        Exp.print_exp(e1);
        Print.print_buf(" = ");
        Exp.print_exp(e2);
        Print.print_buf(";\n"); then ();
    case (INITIALDEFINE(c,e))
      equation 
        Print.print_buf("  ");
        Exp.print_component_ref(c);
        Print.print_buf(" ::= ");
        Exp.print_exp(e);
        Print.print_buf(";\n"); then ();
    case _ then (); 
  end matchcontinue;
end dump_initialequation;

protected function dump_equation_str
  input Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_element)
    local
      Ident s1,s2,s3,s4,s5,str,s;
      Exp.Exp e1,e2,e;
      Exp.ComponentRef c;
    case (EQUATION(e1,e2))
      equation 
        s1 = Exp.print_exp_str(e1);
        s2 = string_append("  ", s1);
        s3 = string_append(s2, " = ");
        s4 = Exp.print_exp_str(e2);
        s5 = string_append(s3, s4);
        str = string_append(s5, ";\n"); then str;
    case (DEFINE(c,e))
      equation 
        s1 = Exp.print_component_ref_str(c);
        s2 = string_append("  ", s1);
        s3 = string_append(" ::= ", s2);
        s4 = Exp.print_exp_str(e);
        s5 = string_append(s3, s4);
        str = string_append(s5, ";\n"); then str;
    case (ASSERT(e))
      equation 
        s = Exp.print_exp_str(e);
        str = string_append(s, ";\n"); then str;
    case _ then ""; 
  end matchcontinue;
end dump_equation_str;

public function dump_algorithm
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Algorithm_StatementList stmts;
    case ALGORITHM(Algorithm.ALGORITHM(stmts))
      equation 
        Print.print_buf("algorithm\n");
        Dump.print_list(stmts, pp_statement, ""); then ();
    case _ then (); 
  end matchcontinue;
end dump_algorithm;

protected function dump_initialalgorithm
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Algorithm_StatementList stmts;
    case INITIALALGORITHM(Algorithm.ALGORITHM(stmts))
      equation 
        Print.print_buf("initial algorithm\n");
        Dump.print_list(stmts, pp_statement, ""); then ();
    case _ then (); 
  end matchcontinue;
end dump_initialalgorithm;

public function dump_algorithm_str
  input Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_element)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Ident s1,str;
      Algorithm_StatementList stmts;
    case ALGORITHM(Algorithm.ALGORITHM(stmts))
      equation 
        s1 = Dump.get_string_list(stmts, pp_statement_str, "");
        str = string_append("algorithm\n", s1); then str;
    case _ then ""; 
  end matchcontinue;
end dump_algorithm_str;

protected function dump_initialalgorithm_str
  input Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_element)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Ident s1,str;
      Algorithm_StatementList stmts;
    case INITIALALGORITHM(Algorithm.ALGORITHM(stmts))
      equation 
        s1 = Dump.get_string_list(stmts, pp_statement_str, "");
        str = string_append("initial algorithm\n", s1); then str;
    case _ then ""; 
  end matchcontinue;
end dump_initialalgorithm_str;

protected function dump_function "LS "
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Ident fstr;
      Absyn.Path fpath;
      ElementList dae;
      TupleTypes_TTypeAbsyn_PathOption t;
    case FUNCTION(fpath,DAE(dae),t)
      equation 
        Print.print_buf("function ");
        fstr = Absyn.path_string(fpath);
        Print.print_buf(fstr);
        Print.print_buf("\n");
        dump_elements(dae);
        Print.print_buf("end ");
        Print.print_buf(fstr);
        Print.print_buf(";\n\n"); then ();
    case _ then (); 
  end matchcontinue;
end dump_function;

protected function dump_function_str "LS "
  input Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_element)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Ident fstr,daestr,str;
      Absyn.Path fpath;
      ElementList dae;
      TupleTypes_TTypeAbsyn_PathOption t;
    case FUNCTION(fpath,DAE(dae),t)
      equation 
        fstr = Absyn.path_string(fpath);
        daestr = dump_elements_str(dae);
        str = Util.string_append_list({"function ",fstr,"\n",daestr,"end ",fstr,";\n\n"}); then str;
    case _ then ""; 
  end matchcontinue;
end dump_function_str;

protected function pp_statement " relation: pp_statement
 
  Prettyprint an algorithm statement
 "
  input Algorithm.Statement alg;
algorithm 
  pp_stmt(alg, 2);
end pp_statement;

protected function pp_statement_str
  input Algorithm.Statement alg;
  output String str;
algorithm 
  str := pp_stmt_str(alg, 2);
end pp_statement_str;

protected function pp_stmt
  input Algorithm.Statement in_statement;
  input Integer in_integer;
algorithm 
  _:=
  matchcontinue (in_statement,in_integer)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Exp.ComponentRef c;
      Exp.Exp e;
      Integer i,i_1;
      Algorithm_StatementList then_,stmts;
      Algorithm.Else else_;
      Ident id;
    case (Algorithm.ASSIGN(_,c,e),i)
      equation 
        indent(i);
        Exp.print_component_ref(c);
        Print.print_buf(" := ");
        Exp.print_exp(e);
        Print.print_buf(";\n"); then ();
    case (Algorithm.ASSIGN_ARR(_,c,e),i)
      equation 
        indent(i);
        Exp.print_component_ref(c);
        Print.print_buf(" := ");
        Exp.print_exp(e);
        Print.print_buf(";\n"); then ();
    case (Algorithm.IF(e,then_,else_),i)
      equation 
        indent(i);
        Print.print_buf("if ");
        Exp.print_exp(e);
        Print.print_buf(" then\n");
        i_1 = i + 2;
        pp_stmt_list(then_, i_1);
        pp_else(else_, i);
        indent(i);
        Print.print_buf("end if;\n"); then ();
    case (Algorithm.FOR(_,_,id,e,stmts),i)
      equation 
        indent(i);
        Print.print_buf("for ");
        Print.print_buf(id);
        Print.print_buf(" in ");
        Exp.print_exp(e);
        Print.print_buf(" loop\n");
        i_1 = i + 2;
        pp_stmt_list(stmts, i_1);
        indent(i);
        Print.print_buf("end for;\n"); then ();
    case (Algorithm.WHILE(e,stmts),i)
      equation 
        indent(i);
        Print.print_buf("while ");
        Exp.print_exp(e);
        Print.print_buf(" loop\n");
        i_1 = i + 2;
        pp_stmt_list(stmts, i_1);
        indent(i);
        Print.print_buf("end while;\n"); then ();
    case (Algorithm.WHEN(e,stmts),i)
      equation 
        indent(i);
        Print.print_buf("when ");
        Exp.print_exp(e);
        Print.print_buf(" do\n");
        i_1 = i + 2;
        pp_stmt_list(stmts, i_1);
        indent(i);
        Print.print_buf("end when;\n"); then ();
    case (_,i)
      equation 
        indent(i);
        Print.print_buf("**ALGORITHM**;\n"); then ();
  end matchcontinue;
end pp_stmt;

protected function pp_stmt_str
  input Algorithm.Statement in_statement;
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_statement,in_integer)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Ident s1,s2,s3,s4,s5,s6,str,s7,s8,s9,s10,s11,id;
      Exp.ComponentRef c;
      Exp.Exp e;
      Integer i,i_1;
      Algorithm_StatementList then_,stmts;
      Algorithm.Else else_;
    case (Algorithm.ASSIGN(_,c,e),i)
      equation 
        s1 = indent_str(i);
        s2 = Exp.print_component_ref_str(c);
        s3 = string_append(s1, s2);
        s4 = string_append(s3, " := ");
        s5 = Exp.print_exp_str(e);
        s6 = string_append(s4, s5);
        str = string_append(s6, ";\n"); then str;
    case (Algorithm.ASSIGN_ARR(_,c,e),i)
      equation 
        s1 = indent_str(i);
        s2 = Exp.print_component_ref_str(c);
        s3 = string_append(s1, s2);
        s4 = string_append(s3, " := ");
        s5 = Exp.print_exp_str(e);
        s6 = string_append(s4, s5);
        str = string_append(s6, ";\n"); then str;
    case (Algorithm.IF(e,then_,else_),i)
      equation 
        s1 = indent_str(i);
        s2 = string_append(s1, "if ");
        s3 = Exp.print_exp_str(e);
        s4 = string_append(s2, s3);
        s5 = string_append(s4, " then\n");
        i_1 = i + 2;
        s6 = pp_stmt_list_str(then_, i_1);
        s7 = string_append(s5, s6);
        s8 = pp_else_str(else_, i);
        s9 = string_append(s7, s8);
        s10 = indent_str(i);
        s11 = string_append(s9, s10);
        str = string_append(s11, "end if;\n"); then str;
    case (Algorithm.FOR(_,_,id,e,stmts),i)
      equation 
        s1 = indent_str(i);
        s2 = string_append(s1, "for ");
        s3 = string_append(s2, id);
        s4 = string_append(s3, " in ");
        s5 = Exp.print_exp_str(e);
        s6 = string_append(s4, s5);
        s7 = string_append(s6, " loop\n");
        i_1 = i + 2;
        s8 = pp_stmt_list_str(stmts, i_1);
        s9 = string_append(s7, s8);
        s10 = indent_str(i);
        s11 = string_append(s9, s10);
        str = string_append(s11, "end for;\n"); then str;
    case (Algorithm.WHILE(e,stmts),i)
      equation 
        s1 = indent_str(i);
        s2 = string_append(s1, "while ");
        s3 = Exp.print_exp_str(e);
        s4 = string_append(s2, s3);
        s5 = string_append(s4, " loop\n");
        i_1 = i + 2;
        s6 = pp_stmt_list_str(stmts, i_1);
        s7 = string_append(s5, s6);
        s8 = indent_str(i);
        s9 = string_append(s7, s8);
        str = string_append(s9, "end while;\n"); then str;
    case (Algorithm.WHEN(e,stmts),i)
      equation 
        s1 = indent_str(i);
        s2 = string_append(s1, "when ");
        s3 = Exp.print_exp_str(e);
        s4 = string_append(s2, s3);
        s5 = string_append(s4, " do\n");
        i_1 = i + 2;
        s6 = pp_stmt_list_str(stmts, i_1);
        s7 = string_append(s5, s6);
        s8 = indent_str(i);
        s9 = string_append(s7, s8);
        str = string_append(s9, "end when;\n"); then str;
    case (_,i)
      equation 
        s1 = indent_str(i);
        str = string_append(s1, "**ALGORITHM**;\n"); then str;
  end matchcontinue;
end pp_stmt_str;

protected function pp_stmt_list
  input Algorithm_StatementList in_algorithm_statementlist;
  input Integer in_integer;
protected 
  type Algorithm_StatementList = list<Algorithm.Statement>;
algorithm 
  _:=
  matchcontinue (in_algorithm_statementlist,in_integer)
    local
      Algorithm.Statement stmt;
      Algorithm_StatementList stmts;
      Integer i;
    case ({},_) then (); 
    case (stmt :: stmts,i)
      equation 
        pp_stmt(stmt, i);
        pp_stmt_list(stmts, i); then ();
  end matchcontinue;
end pp_stmt_list;

protected function pp_stmt_list_str
  input Algorithm_StatementList in_algorithm_statementlist;
  input Integer in_integer;
  output String out_string;
protected 
  type Algorithm_StatementList = list<Algorithm.Statement>;
algorithm 
  out_string:=
  matchcontinue (in_algorithm_statementlist,in_integer)
    local
      Ident s1,s2,str;
      Algorithm.Statement stmt;
      Algorithm_StatementList stmts;
      Integer i;
    case ({},_) then ""; 
    case (stmt :: stmts,i)
      equation 
        s1 = pp_stmt_str(stmt, i);
        s2 = pp_stmt_list_str(stmts, i);
        str = string_append(s1, s2); then str;
  end matchcontinue;
end pp_stmt_list_str;

protected function pp_else
  input Algorithm.Else in_else;
  input Integer in_integer;
algorithm 
  _:=
  matchcontinue (in_else,in_integer)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Integer i_1,i;
      Exp.Exp e;
      Algorithm_StatementList then_,stmts;
      Algorithm.Else else_;
    case (Algorithm.NOELSE(),_) then (); 
    case (Algorithm.ELSEIF(e,then_,else_),i)
      equation 
        indent(i);
        Print.print_buf("elseif ");
        Exp.print_exp(e);
        Print.print_buf(" then\n");
        i_1 = i + 2;
        pp_stmt_list(then_, i_1);
        pp_else(else_, i); then ();
    case (Algorithm.ELSE(stmts),i)
      equation 
        indent(i);
        Print.print_buf("else\n");
        i_1 = i + 2;
        pp_stmt_list(stmts, i_1); then ();
  end matchcontinue;
end pp_else;

protected function pp_else_str
  input Algorithm.Else in_else;
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_else,in_integer)
    local
      type Algorithm_StatementList = list<Algorithm.Statement>;
      Ident s1,s2,s3,s4,s5,s6,s7,s8,str;
      Integer i_1,i;
      Exp.Exp e;
      Algorithm_StatementList then_,stmts;
      Algorithm.Else else_;
    case (Algorithm.NOELSE(),_) then ""; 
    case (Algorithm.ELSEIF(e,then_,else_),i)
      equation 
        s1 = indent_str(i);
        s2 = string_append(s1, "elseif ");
        s3 = Exp.print_exp_str(e);
        s4 = string_append(s2, s3);
        s5 = string_append(s4, " then\n");
        i_1 = i + 2;
        s6 = pp_stmt_list_str(then_, i_1);
        s7 = string_append(s5, s6);
        s8 = pp_else_str(else_, i);
        str = string_append(s7, s8); then str;
    case (Algorithm.ELSE(stmts),i)
      equation 
        s1 = indent_str(i);
        s2 = string_append(s1, "else\n");
        i_1 = i + 2;
        s3 = pp_stmt_list_str(stmts, i_1);
        str = string_append(s2, s3); then str;
  end matchcontinue;
end pp_else_str;

protected function indent
  input Integer in_integer;
algorithm 
  _:=
  matchcontinue (in_integer)
    local Integer i_1,i;
    case 0 then (); 
    case i
      equation 
        Print.print_buf(" ");
        i_1 = i - 1;
        indent(i_1); then ();
  end matchcontinue;
end indent;

protected function indent_str
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer)
    local
      Integer i_1,i;
      Ident s1,str;
    case 0 then ""; 
    case i
      equation 
        i_1 = i - 1;
        s1 = indent_str(i_1);
        str = string_append(" ", s1); then str;
  end matchcontinue;
end indent_str;

public function get_matching_elements "LS "
  input ElementList elist;
  input FuncType1 cond;
  output ElementList elist;
protected 
  function FuncType1
    input Element in_element;
  end FuncType1;
algorithm 
  elist := get_matching(elist, cond);
end get_matching_elements;

public function get_matching "LS "
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
  output VType_aList out_vtype_alist;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
  end FuncType1;
algorithm 
  out_vtype_alist:=
  matchcontinue (in_vtype_alist,in_functype1)
    local
      VType_aList vl_1,vl;
      Type_a v;
      FuncType1 cond;
    case ({},_) then {}; 
    case (v :: vl,cond)
      equation 
        cond(v);
        vl_1 = get_matching(vl, cond); then v :: vl_1;
    case (v :: vl,cond)
      equation 
        failure(cond(v));
        vl_1 = get_matching(vl, cond); then vl_1;
  end matchcontinue;
end get_matching;

public function is_parameter "LS "
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    case VAR(_,PARAM(),_,_,_,_,_,_,_,_,_) then (); 
  end matchcontinue;
end is_parameter;

public function is_comp "LS "
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    case COMP(_,_) then (); 
  end matchcontinue;
end is_comp;

public function get_output_vars "LS "
  input ElementList vl;
  output ElementList vl_1;
protected 
  ElementList vl_1;
algorithm 
  vl_1 := get_matching_elements(vl, is_output_var);
end get_output_vars;

public function get_bidir_vars "LS "
  input ElementList vl;
  output ElementList vl_1;
protected 
  ElementList vl_1;
algorithm 
  vl_1 := get_matching_elements(vl, is_bidir_var);
end get_bidir_vars;

public function get_input_vars "HJ "
  input ElementList vl;
  output ElementList vl_1;
protected 
  ElementList vl_1;
algorithm 
  vl_1 := get_matching_elements(vl, is_input_var);
end get_input_vars;

public function generate_dae_type "Generate a Types.Type from a DAE.Type
  Is needed when investigating the DAE and want to e.g. evaluate expressions.
 "
  input Type in_type;
  output Types.Type out_type;
algorithm 
  out_type:=
  matchcontinue (in_type)
    case (REAL()) then ((Types.T_REAL({}),NONE)); 
    case (INT()) then ((Types.T_INTEGER({}),NONE)); 
    case (BOOL()) then ((Types.T_BOOL({}),NONE)); 
    case (STRING()) then ((Types.T_STRING({}),NONE)); 
  end matchcontinue;
end generate_dae_type;

public function set_component_type "  relation: set_component_type
  This function takes a dae element list and a type name and inserts the type name 
  into each Var  (variable) of the dae. This type name is the origin of the variable.
 "
  input ElementList in_elementlist;
  input Absyn.Path in_path;
  output ElementList out_elementlist;
algorithm 
  out_elementlist:=
  matchcontinue (in_elementlist,in_path)
    local
      ElementList xs_1,xs;
      Exp.ComponentRef cr;
      VarKind kind;
      VarDirection dir;
      Type tp;
      Exp_ExpOption bind,start;
      InstDims dim;
      Flow flow_;
      Absyn_PathList lst;
      VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      Absyn.Path newtype;
      Element x;
    case ({},_) then {}; 
    case (VAR(cr,kind,dir,tp,bind,dim,start,flow_,lst,dae_var_attr,comment) :: xs,newtype)
      equation 
        xs_1 = set_component_type(xs, newtype); then VAR(cr,kind,dir,tp,bind,dim,start,flow_,newtype :: lst,
          dae_var_attr,comment) :: xs_1;
    case (x :: xs,newtype)
      equation 
        xs_1 = set_component_type(xs, newtype); then x :: xs_1;
  end matchcontinue;
end set_component_type;

public function is_output_var "LS "
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      Exp.ComponentRef n;
      Type ty;
    case VAR(n,VARIABLE(),OUTPUT(),ty,_,_,_,_,_,_,_) then (); 
  end matchcontinue;
end is_output_var;

public function is_bidir_var "LS "
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      Exp.ComponentRef n;
      Type ty;
    case VAR(n,VARIABLE(),BIDIR(),ty,_,_,_,_,_,_,_) then (); 
  end matchcontinue;
end is_bidir_var;

public function is_input_var "HJ "
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      Exp.ComponentRef n;
      Type ty;
    case VAR(n,VARIABLE(),INPUT(),ty,_,_,_,_,_,_,_) then (); 
  end matchcontinue;
end is_input_var;

protected function is_not_var "LS "
  input Element e;
algorithm 
  failure(is_var(e));
end is_not_var;

public function is_var "LS "
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    case VAR(_,_,_,_,_,_,_,_,_,_,_) then (); 
  end matchcontinue;
end is_var;

public function is_algorithm
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    case ALGORITHM(_) then (); 
  end matchcontinue;
end is_algorithm;

public function is_function
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    case FUNCTION(_,_,_) then (); 
    case EXTFUNCTION(_,_,_,_) then (); 
  end matchcontinue;
end is_function;

public function dump_debug "

 Dump the data structures in a 
 paranthesised way

"
  input DAElist in_daelist;
algorithm 
  _:=
  matchcontinue (in_daelist)
    local ElementList elist;
    case DAE(elist)
      equation 
        Print.print_buf("DAE(");
        dump_debug_elist(elist);
        Print.print_buf(")"); then ();
  end matchcontinue;
end dump_debug;

protected function dump_debug_elist
  input ElementList in_elementlist;
algorithm 
  _:=
  matchcontinue (in_elementlist)
    local
      Element first;
      ElementList rest;
    case {} then (); 
    case first :: rest
      equation 
        dump_debug_element(first);
        Print.print_buf("\n");
        dump_debug_elist(rest); then ();
  end matchcontinue;
end dump_debug_elist;

public function dump_debug_element
  input Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Ident comment_str,tmp_str,n,fstr;
      Exp.ComponentRef cr;
      VarKind vk;
      VarDirection vd;
      Type ty;
      VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      Exp.Exp e,exp,e1,e2;
      DAElist l;
      Absyn.Path fpath;
      TupleTypes_TTypeAbsyn_PathOption t;
    case VAR(cr,vk,vd,ty,NONE,_,_,_,_,dae_var_attr,comment)
      equation 
        Print.print_buf("VAR(");
        Exp.print_component_ref(cr);
        Print.print_buf(", ");
        dump_kind(vk);
        comment_str = Dump.unparse_comment_option(comment);
        Print.print_buf("  comment:");
        Print.print_buf(comment_str);
        tmp_str = dump_variable_attributes_str(dae_var_attr);
        Print.print_buf(tmp_str);
        Print.print_buf(")"); then ();
    case VAR(cr,vk,vd,ty,SOME(e),_,_,_,_,dae_var_attr,comment)
      equation 
        Print.print_buf("VAR(");
        Exp.print_component_ref(cr);
        Print.print_buf(", ");
        dump_kind(vk);
        Print.print_buf(", ");
        comment_str = Dump.unparse_comment_option(comment);
        Print.print_buf("  comment:");
        Print.print_buf(comment_str);
        tmp_str = dump_variable_attributes_str(dae_var_attr);
        Exp.print_exp(e);
        Print.print_buf(tmp_str);
        Print.print_buf(")"); then ();
    case DEFINE(cr,exp)
      equation 
        Print.print_buf("DEFINE(");
        Exp.print_component_ref(cr);
        Print.print_buf(", ");
        Exp.print_exp(exp);
        Print.print_buf(")"); then ();
    case INITIALDEFINE(cr,exp)
      equation 
        Print.print_buf("INITIALDEFINE(");
        Exp.print_component_ref(cr);
        Print.print_buf(", ");
        Exp.print_exp(exp);
        Print.print_buf(")"); then ();
    case EQUATION(e1,e2)
      equation 
        Print.print_buf("EQUATION(");
        Exp.print_exp(e1);
        Print.print_buf(",");
        Exp.print_exp(e2);
        Print.print_buf(")"); then ();
    case INITIALEQUATION(e1,e2)
      equation 
        Print.print_buf("INITIALEQUATION(");
        Exp.print_exp(e1);
        Print.print_buf(",");
        Exp.print_exp(e2);
        Print.print_buf(")"); then ();
    case ALGORITHM(_)
      equation 
        Print.print_buf("ALGORITHM()"); then ();
    case INITIALALGORITHM(_)
      equation 
        Print.print_buf("INITIALALGORITHM()"); then ();
    case COMP(n,l)
      equation 
        Print.print_buf("COMP(");
        Print.print_buf(n);
        Print.print_buf(",");
        dump_debug(l);
        Print.print_buf(")"); then ();
    case FUNCTION(fpath,l,t)
      equation 
        Print.print_buf("FUNCTION(");
        fstr = Absyn.path_string(fpath);
        Print.print_buf(fstr);
        Print.print_buf(",");
        Types.print_type(t);
        Print.print_buf(",");
        dump_debug(l);
        Print.print_buf(")"); then ();
    case _
      equation 
        Print.print_buf("UNKNOWN "); then ();
  end matchcontinue;
end dump_debug_element;

public function find_element
  input ElementList in_elementlist;
  input FuncType1 in_functype1;
  output ElementOption out_elementoption;
protected 
  function FuncType1
    input Element in_element;
  end FuncType1;
algorithm 
  out_elementoption:=
  matchcontinue (in_elementlist,in_functype1)
    local
      Element e;
      ElementList rest;
      FuncType1 f;
      ElementOption e_1;
    case ({},_) then NONE; 
    case (e :: rest,f)
      equation 
        f(e); then SOME(e);
    case (e :: rest,f)
      equation 
        failure(f(e));
        e_1 = find_element(rest, f); then e_1;
  end matchcontinue;
end find_element;

public function dump_graphviz "
 Graphviz functions to visualize 
 the dae
"
  input DAElist dae;
protected 
  Graphviz.Node r;
algorithm 
  r := build_graphviz(dae);
  Graphviz.dump(r);
end dump_graphviz;

protected function build_graphviz
  input DAElist in_daelist;
  output Graphviz.Node out_node;
algorithm 
  out_node:=
  matchcontinue (in_daelist)
    local
      type Graphviz_NodeList = list<Graphviz.Node>;
      ElementList vars,nonvars,els;
      Graphviz_NodeList nonvarnodes,varnodes,nodelist;
    case DAE(els)
      equation 
        vars = get_matching_elements(els, is_var);
        nonvars = get_matching_elements(els, is_not_var);
        nonvarnodes = build_gr_list(nonvars);
        varnodes = build_gr_vars(vars);
        nodelist = list_append(nonvarnodes, varnodes); then Graphviz.NODE("DAE",{},nodelist);
  end matchcontinue;
end build_graphviz;

protected function build_gr_list
  input ElementList in_elementlist;
  output Graphviz_NodeList out_graphviz_nodelist;
protected 
  type Graphviz_NodeList = list<Graphviz.Node>;
algorithm 
  out_graphviz_nodelist:=
  matchcontinue (in_elementlist)
    local
      Graphviz.Node node;
      Graphviz_NodeList nodelist;
      Element el;
      ElementList rest;
    case {} then {}; 
    case el :: rest
      equation 
        node = build_gr_element(el);
        nodelist = build_gr_list(rest); then node :: nodelist;
  end matchcontinue;
end build_gr_list;

protected function build_gr_vars
  input ElementList in_elementlist;
  output Graphviz_NodeList out_graphviz_nodelist;
protected 
  type Graphviz_NodeList = list<Graphviz.Node>;
algorithm 
  out_graphviz_nodelist:=
  matchcontinue (in_elementlist)
    local
      type IdentList = list<Ident>;
      IdentList strlist;
      ElementList vars;
    case {} then {}; 
    case vars
      equation 
        (strlist,_) = build_gr_strlist(vars, build_gr_var_str, 10); then {Graphviz.LNODE("VARS",strlist,{Graphviz.box},{})};
  end matchcontinue;
end build_gr_vars;

public function build_gr_strlist
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
  input Integer in_integer;
  output StringList out_stringlist;
  output VType_aList out_vtype_alist;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
    output String out_string;
  end FuncType1;
algorithm 
  (out_stringlist,out_vtype_alist):=
  matchcontinue (in_vtype_alist,in_functype1,in_integer)
    local
      type IdentList = list<Ident>;
      VType_aList ignored,rest;
      FuncType1 printer;
      Integer count,count_1;
      IdentList strlist;
      Ident str;
      Type_a var;
    case ({},_,_) then ({},{}); 
    case (ignored,printer,count)
      equation 
        (count <= 0) = true; then ({"..."},ignored);
    case (var :: rest,printer,count)
      equation 
        (count > 0) = true;
        count_1 = count - 1;
        (strlist,ignored) = build_gr_strlist(rest, printer, count_1);
        str = printer(var); then (str :: strlist,ignored);
  end matchcontinue;
end build_gr_strlist;

protected function build_gr_var_str
  input Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_element)
    local
      Ident str,expstr,str_1,str_2;
      Exp.ComponentRef cr;
      Exp.Exp exp;
    case VAR(cr,_,_,_,NONE,_,_,_,_,_,_)
      equation 
        str = Exp.print_component_ref_str(cr); then str;
    case VAR(cr,_,_,_,SOME(exp),_,_,_,_,_,_)
      equation 
        str = Exp.print_component_ref_str(cr);
        expstr = print_exp_str_special(exp);
        str_1 = string_append(str, " = ");
        str_2 = string_append(str_1, expstr); then str_2;
  end matchcontinue;
end build_gr_var_str;

protected function print_exp_str_special
  input Exp.Exp in_exp;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_exp)
    local
      Ident s_1,s_2,s,str;
      Exp.Exp exp;
    case Exp.SCONST(s)
      equation 
        s_1 = string_append("\\\"", s);
        s_2 = string_append(s_1, "\\\""); then s_2;
    case exp
      equation 
        str = Exp.print_exp_str(exp); then str;
  end matchcontinue;
end print_exp_str_special;

protected function build_gr_element
  input Element in_element;
  output Graphviz.Node out_node;
algorithm 
  out_node:=
  matchcontinue (in_element)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Ident crstr,vkstr,expstr,expstr_1,e1str,e2str,n,fstr;
      Exp.ComponentRef cr;
      VarKind vk;
      VarDirection vd;
      Type ty;
      Exp.Exp exp,e1,e2;
      Graphviz.Node node;
      DAElist dae;
      Absyn.Path fpath;
    case VAR(cr,vk,vd,ty,NONE,_,_,_,_,_,_)
      equation 
        crstr = Exp.print_component_ref_str(cr);
        vkstr = dump_kind_str(vk); then Graphviz.LNODE("VAR",{crstr,vkstr},{},{});
    case VAR(cr,vk,vd,ty,SOME(exp),_,_,_,_,_,_)
      equation 
        crstr = Exp.print_component_ref_str(cr);
        vkstr = dump_kind_str(vk);
        expstr = print_exp_str_special(exp);
        expstr_1 = string_append("= ", expstr); then Graphviz.LNODE("VAR",{crstr,vkstr,expstr_1},{},{});
    case DEFINE(cr,exp)
      equation 
        crstr = Exp.print_component_ref_str(cr);
        expstr = print_exp_str_special(exp);
        expstr_1 = string_append("= ", expstr); then Graphviz.LNODE("DEFINE",{crstr,expstr_1},{},{});
    case EQUATION(e1,e2)
      equation 
        e1str = print_exp_str_special(e1);
        e2str = print_exp_str_special(e2); then Graphviz.LNODE("EQUATION",{e1str,"=",e2str},{},{});
    case ALGORITHM(_) then Graphviz.NODE("ALGORITHM",{},{}); 
    case INITIALDEFINE(cr,exp)
      equation 
        crstr = Exp.print_component_ref_str(cr);
        expstr = print_exp_str_special(exp);
        expstr_1 = string_append("= ", expstr); then Graphviz.LNODE("INITIALDEFINE",{crstr,expstr_1},{},{});
    case INITIALEQUATION(e1,e2)
      equation 
        e1str = print_exp_str_special(e1);
        e2str = print_exp_str_special(e2); then Graphviz.LNODE("INITIALEQUATION",{e1str,"=",e2str},{},{});
    case INITIALALGORITHM(_) then Graphviz.NODE("INITIALALGORITHM",{},{}); 
    case COMP(n,dae)
      equation 
        node = build_graphviz(dae); then Graphviz.LNODE("COMP",{n},{},{node});
    case FUNCTION(fpath,dae,ty)
      local TupleTypes_TTypeAbsyn_PathOption ty;
      equation 
        node = build_graphviz(dae);
        fstr = Absyn.path_string(fpath); then Graphviz.LNODE("FUNCTION",{fstr},{},{node});
  end matchcontinue;
end build_gr_element;

public function get_variable_bindings_str " relation: get_variable_bindings_str
  This function takes a `DAE.Element' list and returns a comma separated string of variable bindings.
  E.g. model A Real x=1; Real y=2; end A; => \"1,2\"
 "
  input ElementList elts;
  output String str;
protected 
  ElementList varlst;
algorithm 
  varlst := get_variable_list(elts);
  str := get_bindings_str(varlst);
end get_variable_bindings_str;

protected function get_variable_list
  input ElementList in_elementlist;
  output ElementList out_elementlist;
algorithm 
  out_elementlist:=
  matchcontinue (in_elementlist)
    local
      ElementList res,lst;
      Exp.ComponentRef a;
      VarKind b;
      VarDirection c;
      Type d;
      Exp_ExpOption e,g;
      InstDims f;
      Flow h;
      Absyn_PathList i;
      VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
    case VAR(a,b,c,d,e,f,g,h,i,dae_var_attr,comment) :: lst
      equation 
        res = get_variable_list(lst); then VAR(a,b,c,d,e,f,g,h,i,dae_var_attr,comment) :: res;
    case _ :: lst
      equation 
        res = get_variable_list(lst); then res;
    case {} then {}; 
  end matchcontinue;
end get_variable_list;

protected function get_bindings_str
  input ElementList in_elementlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementlist)
    local
      Ident expstr,s3,s4,str,s1,s2;
      Element v;
      Exp.ComponentRef cr;
      Exp.Exp e;
      ElementList lst;
    case ((v = VAR(cr,_,_,_,SOME(e),_,_,_,_,_,_)) :: (lst = _ :: _))
      equation 
        expstr = Exp.print_exp_str(e);
        s3 = string_append(expstr, ",");
        s4 = get_bindings_str(lst);
        str = string_append(s3, s4); then str;
    case ((v = VAR(cr,_,_,_,NONE,_,_,_,_,_,_)) :: (lst = _ :: _))
      equation 
        s1 = "-,";
        s2 = get_bindings_str(lst);
        str = string_append(s1, s2); then str;
    case ({(v = VAR(cr,_,_,_,SOME(e),_,_,_,_,_,_))})
      equation 
        str = Exp.print_exp_str(e); then str;
    case ({(v = VAR(cr,_,_,_,NONE,_,_,_,_,_,_))}) then ""; 
  end matchcontinue;
end get_bindings_str;

public function to_flow
  input Boolean in_boolean;
  input ClassInf.State in_state;
  output Flow out_flow;
algorithm 
  out_flow:=
  matchcontinue (in_boolean,in_state)
    case (true,_) then FLOW(); 
    case (_,ClassInf.CONNECTOR(_)) then NON_FLOW(); 
    case (_,_) then NON_CONNECTOR(); 
  end matchcontinue;
end to_flow;

public function get_flow_variables
  input ElementList in_elementlist;
  output Exp_ComponentRefList out_exp_componentreflist;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_elementlist)
    local
      Exp_ComponentRefList res,res1,res1_1,res2;
      Exp.ComponentRef cr;
      ElementList xs,lst;
      Ident id;
    case ({}) then {}; 
    case (VAR(cr,_,_,_,_,_,_,FLOW(),_,_,_) :: xs)
      equation 
        res = get_flow_variables(xs); then cr :: res;
    case (COMP(id,DAE(lst)) :: xs)
      equation 
        res1 = get_flow_variables(lst);
        res1_1 = get_flow_variables_2(res1, id);
        res2 = get_flow_variables(xs);
        res = list_append(res1_1, res2); then res;
    case (_ :: xs)
      equation 
        res = get_flow_variables(xs); then res;
  end matchcontinue;
end get_flow_variables;

protected function get_flow_variables_2
  input Exp_ComponentRefList in_exp_componentreflist;
  input Ident in_ident;
  output Exp_ComponentRefList out_exp_componentreflist;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_exp_componentreflist,in_ident)
    local
      Ident id;
      Exp_ComponentRefList res,xs;
      Exp.ComponentRef cr_1,cr;
    case ({},id) then {}; 
    case (cr :: xs,id)
      equation 
        res = get_flow_variables_2(xs, id);
        cr_1 = Exp.join_crefs(Exp.CREF_IDENT(id,{}), cr); then cr_1 :: res;
  end matchcontinue;
end get_flow_variables_2;

public function dae_to_record_value " relation: dae_to_record_value
  Transforms a list of elements into a record value.
  TODO: This does not work for records inside records. 
  For a general approach we need to build an environment from the DAE and then
  instead investigate the variables and lookup their values from the created environment.
 "
  input Absyn.Path in_path;
  input ElementList in_elementlist;
  input Boolean in_boolean;
  output Values.Value out_value;
algorithm 
  out_value:=
  matchcontinue (in_path,in_elementlist,in_boolean)
    local
      type Values_ValueList = list<Values.Value>;
      type IdentList = list<Ident>;
      Absyn.Path cname;
      Values.Value value,res;
      Values_ValueList vals;
      IdentList names;
      Ident cr_str;
      Exp.ComponentRef cr;
      Exp.Exp rhs;
      ElementList rest;
      Boolean impl;
    case (cname,{},_) then Values.RECORD(cname,{},{}); 
    case (cname,EQUATION(Exp.CREF(cr,_),rhs) :: rest,impl)
      equation 
        (value,_) = Ceval.ceval({}, rhs, impl, NONE, NONE);
        Values.RECORD(cname,vals,names) = dae_to_record_value(cname, rest, impl);
        cr_str = Exp.print_component_ref_str(cr); then Values.RECORD(cname,value :: vals,cr_str :: names);
    case (cname,_ :: rest,impl)
      equation 
        res = dae_to_record_value(cname, rest, impl); then res;
    case (_,_,_)
      equation 
        Debug.fprint("failtrace", "-dae_to_record_value failed\n"); then fail();
  end matchcontinue;
end dae_to_record_value;

public function to_modelica_form "Transforms all variables from a.b.c to a_b_c, etc"
  input DAElist in_daelist;
  output DAElist out_daelist;
algorithm 
  out_daelist:=
  matchcontinue (in_daelist)
    local ElementList elts_1,elts;
    case (DAE(elts))
      equation 
        elts_1 = to_modelica_form_elts(elts); then DAE(elts_1);
  end matchcontinue;
end to_modelica_form;

protected function to_modelica_form_elts
  input ElementList in_elementlist;
  output ElementList out_elementlist;
algorithm 
  out_elementlist:=
  matchcontinue (in_elementlist)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Ident str,str_1,id;
      ElementList elts_1,elts,welts_1,welts,telts_1,eelts_1,telts,eelts;
      Exp_ExpOption d_1,d,f;
      Exp.ComponentRef cr,cr_1;
      VarKind a;
      VarDirection b;
      Type c;
      InstDims e;
      Flow g;
      Absyn_PathList h;
      VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      Exp.Exp e_1,e1_1,e2_1,e1,e2;
      Element elt_1,elt;
      DAElist dae_1,dae;
      Absyn.Path p;
      TupleTypes_TTypeAbsyn_PathOption t;
    case ({}) then {}; 
    case (VAR(cr,a,b,c,d,e,f,g,h,dae_var_attr,comment) :: elts)
      equation 
        str = Exp.print_component_ref_str(cr);
        str_1 = Util.string_replace_char(str, #".", #"_");
        elts_1 = to_modelica_form_elts(elts);
        d_1 = to_modelica_form_exp_opt(d); then VAR(Exp.CREF_IDENT(str_1,{}),a,b,c,d_1,e,f,g,h,dae_var_attr,
          comment) :: elts_1;
    case (DEFINE(cr,e) :: elts)
      local Exp.Exp e;
      equation 
        e_1 = to_modelica_form_exp(e);
        cr_1 = to_modelica_form_cref(cr);
        elts_1 = to_modelica_form_elts(elts); then DEFINE(cr_1,e_1) :: elts_1;
    case (INITIALDEFINE(cr,e) :: elts)
      local Exp.Exp e;
      equation 
        e_1 = to_modelica_form_exp(e);
        cr_1 = to_modelica_form_cref(cr);
        elts_1 = to_modelica_form_elts(elts); then INITIALDEFINE(cr_1,e_1) :: elts_1;
    case (EQUATION(e1,e2) :: elts)
      equation 
        e1_1 = to_modelica_form_exp(e1);
        e2_1 = to_modelica_form_exp(e2);
        elts_1 = to_modelica_form_elts(elts); then EQUATION(e1_1,e2_1) :: elts_1;
    case (WHEN_EQUATION(e1,welts,SOME(elt)) :: elts)
      equation 
        e1_1 = to_modelica_form_exp(e1);
        welts_1 = to_modelica_form_elts(welts);
        {elt_1} = to_modelica_form_elts({elt});
        elts_1 = to_modelica_form_elts(elts); then WHEN_EQUATION(e1_1,welts_1,SOME(elt_1)) :: elts_1;
    case (WHEN_EQUATION(e1,welts,NONE) :: elts)
      equation 
        e1_1 = to_modelica_form_exp(e1);
        welts_1 = to_modelica_form_elts(welts);
        elts_1 = to_modelica_form_elts(elts); then WHEN_EQUATION(e1_1,welts_1,NONE) :: elts_1;
    case (IF_EQUATION(e1,telts,eelts) :: elts)
      equation 
        e1_1 = to_modelica_form_exp(e1);
        telts_1 = to_modelica_form_elts(telts);
        eelts_1 = to_modelica_form_elts(eelts);
        elts_1 = to_modelica_form_elts(elts); then IF_EQUATION(e1_1,telts_1,eelts_1) :: elts_1;
    case (INITIALEQUATION(e1,e2) :: elts)
      equation 
        e1_1 = to_modelica_form_exp(e1);
        e2_1 = to_modelica_form_exp(e2);
        elts_1 = to_modelica_form_elts(elts); then INITIALEQUATION(e1_1,e2_1) :: elts_1;
    case (ALGORITHM(a) :: elts)
      local Algorithm.Algorithm a;
      equation 
        print("to_modelica_form_elts(ALGORITHM) not impl. yet\n");
        elts_1 = to_modelica_form_elts(elts); then ALGORITHM(a) :: elts_1;
    case (INITIALALGORITHM(a) :: elts)
      local Algorithm.Algorithm a;
      equation 
        print("to_modelica_form_elts(INITIALALGORITHM) not impl. yet\n");
        elts_1 = to_modelica_form_elts(elts); then INITIALALGORITHM(a) :: elts_1;
    case (COMP(id,dae) :: elts)
      equation 
        dae_1 = to_modelica_form(dae);
        elts_1 = to_modelica_form_elts(elts); then COMP(id,dae_1) :: elts_1;
    case (FUNCTION(p,dae,t) :: elts)
      equation 
        dae_1 = to_modelica_form(dae);
        elts_1 = to_modelica_form_elts(elts); then FUNCTION(p,dae_1,t) :: elts_1;
    case (EXTFUNCTION(p,dae,t,d) :: elts)
      local ExternalDecl d;
      equation 
        elts_1 = to_modelica_form_elts(elts);
        dae_1 = to_modelica_form(dae); then EXTFUNCTION(p,dae,t,d) :: elts_1;
    case (ASSERT(e) :: elts)
      local Exp.Exp e;
      equation 
        elts_1 = to_modelica_form_elts(elts);
        e_1 = to_modelica_form_exp(e); then ASSERT(e_1) :: elts_1;
  end matchcontinue;
end to_modelica_form_elts;

protected function to_modelica_form_exp_opt
  input Exp_ExpOption in_exp_expoption;
  output Exp_ExpOption out_exp_expoption;
algorithm 
  out_exp_expoption:=
  matchcontinue (in_exp_expoption)
    local Exp.Exp e_1,e;
    case (SOME(e))
      equation 
        e_1 = to_modelica_form_exp(e); then SOME(e_1);
    case (NONE) then NONE; 
  end matchcontinue;
end to_modelica_form_exp_opt;

protected function to_modelica_form_cref
  input Exp.ComponentRef cr;
  output Exp.ComponentRef out_componentref;
protected 
  Ident str,str_1;
algorithm 
  str := Exp.print_component_ref_str(cr);
  str_1 := Util.string_replace_char(str, #".", #"_");
  out_componentref := Exp.CREF_IDENT(str_1,{});
end to_modelica_form_cref;

protected function to_modelica_form_exp
  input Exp.Exp in_exp;
  output Exp.Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp)
    local
      type Exp_ExpList = list<Exp.Exp>;
      Exp.ComponentRef cr_1,cr;
      Exp.Type t;
      Exp.Exp e1_1,e2_1,e1,e2,e_1,e,e3_1,e3;
      Exp.Operator op;
      Exp_ExpList expl_1,expl;
      Absyn.Path f;
      Boolean b;
      Integer i;
      Exp_ExpOption eopt_1,eopt;
    case (Exp.CREF(cr,t))
      equation 
        cr_1 = to_modelica_form_cref(cr); then Exp.CREF(cr_1,t);
    case (Exp.BINARY(e1,op,e2))
      equation 
        e1_1 = to_modelica_form_exp(e1);
        e2_1 = to_modelica_form_exp(e2); then Exp.BINARY(e1_1,op,e2_1);
    case (Exp.LBINARY(e1,op,e2))
      equation 
        e1_1 = to_modelica_form_exp(e1);
        e2_1 = to_modelica_form_exp(e2); then Exp.LBINARY(e1_1,op,e2_1);
    case (Exp.UNARY(op,e))
      equation 
        e_1 = to_modelica_form_exp(e); then Exp.UNARY(op,e_1);
    case (Exp.LUNARY(op,e))
      equation 
        e_1 = to_modelica_form_exp(e); then Exp.LUNARY(op,e_1);
    case (Exp.RELATION(e1,op,e2))
      equation 
        e1_1 = to_modelica_form_exp(e1);
        e2_1 = to_modelica_form_exp(e2); then Exp.RELATION(e1_1,op,e2_1);
    case (Exp.IFEXP(e1,e2,e3))
      equation 
        e1_1 = to_modelica_form_exp(e1);
        e2_1 = to_modelica_form_exp(e2);
        e3_1 = to_modelica_form_exp(e3); then Exp.IFEXP(e1_1,e2_1,e3_1);
    case (Exp.CALL(f,expl,t,b))
      local Boolean t;
      equation 
        expl_1 = Util.list_map(expl, to_modelica_form_exp); then Exp.CALL(f,expl_1,t,b);
    case (Exp.ARRAY(t,b,expl))
      equation 
        expl_1 = Util.list_map(expl, to_modelica_form_exp); then Exp.ARRAY(t,b,expl_1);
    case (Exp.TUPLE(expl))
      equation 
        expl_1 = Util.list_map(expl, to_modelica_form_exp); then Exp.TUPLE(expl_1);
    case (Exp.CAST(t,e))
      equation 
        e_1 = to_modelica_form_exp(e); then Exp.CAST(t,e_1);
    case (Exp.ASUB(e,i))
      equation 
        e_1 = to_modelica_form_exp(e); then Exp.ASUB(e_1,i);
    case (Exp.SIZE(e,eopt))
      equation 
        e_1 = to_modelica_form_exp(e);
        eopt_1 = to_modelica_form_exp_opt(eopt); then Exp.SIZE(e_1,eopt_1);
    case (e) then e; 
  end matchcontinue;
end to_modelica_form_exp;

public function get_named_function " relation: get_named_function
  return the FUNCTION with the given name. Returns empty list if not found
  TODO: Only top level functions are checked. Add recursing into the DAE
  and path name checking.
  TODO: External functions?
 "
  input Absyn.Path in_path;
  input ElementList in_elementlist;
  output ElementList out_elementlist;
algorithm 
  out_elementlist:=
  matchcontinue (in_path,in_elementlist)
    local
      Absyn.Path path,elpath;
      Element el;
      ElementList rest,res;
    case (_,{}) then {}; 
    case (path,(el = FUNCTION(elpath,_,_)) :: rest)
      equation 
        true = ModUtil.path_equal(path, elpath); then {el};
    case (path,el :: rest)
      equation 
        res = get_named_function(path, rest); then res;
    case (_,_)
      equation 
        Debug.fprintln("failtrace", "-- get_named_function failed"); then fail();
  end matchcontinue;
end get_named_function;

public function get_all_exps " relation: get_all_exps
  
  This function goes through the DAE structure and finds all the
  expressions and returns them in a list
 "
  input ElementList elements;
  output Exp_ExpList exps;
protected 
  type Exp_ExpList = list<Exp.Exp>;
  type Exp_ExpListList = list<Exp_ExpList>;
  Exp_ExpListList expslist;
algorithm 
  expslist := Util.list_map(elements, get_all_exps_element);
  exps := Util.list_flatten(expslist);
end get_all_exps;

protected function cref_to_exp
  input Exp.ComponentRef in_componentref;
  output Exp.Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_componentref)
    local Exp.ComponentRef cref;
    case cref then Exp.CREF(cref,Exp.OTHER()); 
  end matchcontinue;
end cref_to_exp;

protected function get_all_exps_element " relation: get_all_exps_element
  
  Helper to get_all_exps. Implements get_all_exps for different kinds of
  elements 
 "
  input Element in_element;
  output Exp_ExpList out_exp_explist;
protected 
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  out_exp_explist:=
  matchcontinue (in_element)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Exp_ExpListList = list<Exp_ExpList>;
      Exp_ExpList e1,e2,e3,exps,explist1,explist2,exps1,exps2,exps3;
      Exp.Exp crefexp,exp,cond;
      Exp.ComponentRef cref;
      VarKind vk;
      VarDirection vd;
      Type ty;
      Exp_ExpOption bndexp,startvalexp;
      InstDims instdims;
      Flow flow_;
      Absyn_PathList pathlist;
      VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      ElementList ellist,elements,eqs,eqstrueb,eqsfalseb;
      ElementOption elsewhenopt;
      Algorithm.Algorithm alg;
      Ident id,fname,lang;
      Absyn.Path path;
      Exp_ExpListList argexps,expslist;
      ExtArgList args;
      ExtArg retarg;
    case VAR(cref,vk,vd,ty,bndexp,instdims,startvalexp,flow_,pathlist,dae_var_attr,comment) "VAR "
      equation 
        e1 = Util.option_to_list(bndexp);
        e2 = Util.option_to_list(startvalexp);
        e3 = Util.list_map(instdims, get_all_exps_subscript);
        e3 = Util.list_flatten(e3);
        crefexp = cref_to_exp(cref);
        exps = Util.list_flatten({e1,e2,e3,{crefexp}}); then exps;
    case DEFINE(cref,exp)
      equation 
        crefexp = cref_to_exp(cref); then {crefexp,exp};
    case INITIALDEFINE(cref,exp)
      equation 
        crefexp = cref_to_exp(cref); then {crefexp,exp};
    case EQUATION(e1,e2)
      local Exp.Exp e1,e2; then {e1,e2};
    case WHEN_EQUATION(cond,eqs,elsewhenopt)
      equation 
        ellist = Util.option_to_list(elsewhenopt);
        elements = list_append(eqs, ellist);
        exps = get_all_exps(elements); then cond :: exps;
    case IF_EQUATION(cond,eqstrueb,eqsfalseb)
      equation 
        explist1 = get_all_exps(eqstrueb);
        explist2 = get_all_exps(eqsfalseb);
        exps = Util.list_flatten({{cond},explist1,explist2}); then exps;
    case INITIALEQUATION(e1,e2)
      local Exp.Exp e1,e2; then {e1,e2};
    case ALGORITHM(alg)
      equation 
        exps = Algorithm.get_all_exps(alg); then exps;
    case INITIALALGORITHM(alg)
      equation 
        exps = Algorithm.get_all_exps(alg); then exps;
    case COMP(id,DAE(elements))
      equation 
        exps = get_all_exps(elements); then exps;
    case FUNCTION(path,DAE(elements),ty)
      local TupleTypes_TTypeAbsyn_PathOption ty;
      equation 
        exps1 = get_all_exps(elements);
        exps2 = Types.get_all_exps(ty);
        exps = list_append(exps1, exps2); then exps;
    case EXTFUNCTION(path,DAE(elements),ty,EXTERNALDECL(fname,args,retarg,lang))
      local TupleTypes_TTypeAbsyn_PathOption ty;
      equation 
        exps1 = get_all_exps(elements);
        exps2 = Types.get_all_exps(ty);
        exps3 = get_all_exps_extarg(retarg);
        argexps = Util.list_map(args, get_all_exps_extarg);
        expslist = list_append({exps1,exps2,exps3}, argexps);
        exps = Util.list_flatten(expslist); then exps;
    case ASSERT(exp) then {exp}; 
    case _
      equation 
        Debug.fprintln("failtrace", "-- get_all_exps_element failed"); then fail();
  end matchcontinue;
end get_all_exps_element;

protected function get_all_exps_subscript " relation: get_all_exps_subscript
  
  Get all exps from a Subscript 
 "
  input Exp.Subscript in_subscript;
  output Exp_ExpList out_exp_explist;
protected 
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  out_exp_explist:=
  matchcontinue (in_subscript)
    local Exp.Exp e;
    case Exp.WHOLEDIM() then {}; 
    case Exp.SLICE(e) then {e}; 
    case Exp.INDEX(e) then {e}; 
    case _
      equation 
        Debug.fprintln("failtrace", "-- get_all_exps_subscript failed"); then fail();
  end matchcontinue;
end get_all_exps_subscript;

protected function get_all_exps_extarg " relation: get_all_exps_extarg
  
  Get all exps from an ExtArg 
 "
  input ExtArg in_extarg;
  output Exp_ExpList out_exp_explist;
protected 
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  out_exp_explist:=
  matchcontinue (in_extarg)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.Exp exp1,crefexp,exp;
      Exp_ExpList explist,exps,tyexps;
      Exp.ComponentRef cref;
      Types.Attributes attr;
      TupleTypes_TTypeAbsyn_PathOption ty;
    case EXTARG(cref,attr,ty)
      equation 
        exp1 = cref_to_exp(cref);
        explist = Types.get_all_exps(ty);
        exps = list_append({exp1}, explist); then exps;
    case EXTARGEXP(exp1,ty)
      equation 
        explist = Types.get_all_exps(ty);
        exps = list_append({exp1}, explist); then exps;
    case EXTARGSIZE(cref,attr,ty,exp)
      equation 
        crefexp = cref_to_exp(cref);
        tyexps = Types.get_all_exps(ty);
        exps = Util.list_flatten({{crefexp},tyexps,{exp}}); then exps;
    case NOEXTARG() then {}; 
    case _
      equation 
        Debug.fprintln("failtrace", "-- get_all_exps_extarg failed"); then fail();
  end matchcontinue;
end get_all_exps_extarg;
end DAE;

