package Debug "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Debug.rml
  module:      Debug
  description: debug printing
 
  RCS: $Id: Debug.rml 1733 2005-04-27 06:47:45Z haklu $
 
  Printing routines for debug output of strings. Also flag controlled
  printing. When flag controlled printing functions are called, printing is 
  done only if the given flag is among the flags given in the runtime 
  arguments, to +d-flag, i.e. if +d=inst,lookup is given in the command line,
  only calls containing these flags will actually print something, e.g.:
  fprint(\"inst\", \"Starting instantiation...\"). See runtime/rtopts.c for
  implementation of flag checking.
 
 "

protected import OpenModelica.Compiler.RTOpts;

protected import OpenModelica.Compiler.Print;

public function print "PR 
  This realtion is used for debug printing. 
 "
  input String s;
algorithm 
  fprint("olddebug", s);
end print;

public function fprint "LS: Flag controlled debugging "
  input String in_string1;
  input String in_string2;
algorithm 
  _:=
  matchcontinue (in_string1,in_string2)
    local String flag,str;
    case (flag,str)
      equation 
        true = RTOpts.debug_flag(flag);
        Print.print_error_buf(str); then ();
    case (_,_) then (); 
  end matchcontinue;
end fprint;

public function fprintln
  input String in_string1;
  input String in_string2;
algorithm 
  _:=
  matchcontinue (in_string1,in_string2)
    local String flag,str;
    case (flag,str)
      equation 
        true = RTOpts.debug_flag(flag);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then ();
    case (_,_) then (); 
  end matchcontinue;
end fprintln;

public function fprintl
  input String in_string;
  input StringList in_stringlist;
protected 
  type StringList = list<String>;
algorithm 
  _:=
  matchcontinue (in_string,in_stringlist)
    local
      String str,flag;
      StringList strlist;
    case (flag,strlist)
      equation 
        true = RTOpts.debug_flag(flag);
        str = make_string(strlist);
        Print.print_error_buf(str); then ();
    case (_,_) then (); 
  end matchcontinue;
end fprintl;

protected function make_string
  input StringList in_stringlist;
  output String out_string;
protected 
  type StringList = list<String>;
algorithm 
  out_string:=
  matchcontinue (in_stringlist)
    local
      String res,res_1,first;
      StringList rest;
    case {} then ""; 
    case first :: rest
      equation 
        res = make_string(rest);
        res_1 = string_append(first, res); then res_1;
  end matchcontinue;
end make_string;

public function fcall "Call the given function (2nd arg) if the flag given in 1st arg is set "
  input String in_string;
  input FuncType1 in_functype1;
  input Type_a in_type_a;
protected 
  function FuncType1
    input Type_a in_type_a;
  protected 
    replaceable type Type_a;
  end FuncType1;
  replaceable type Type_a;
algorithm 
  _:=
  matchcontinue (in_string,in_functype1,in_type_a)
    local
      String flag;
      FuncType1 func;
      Type_a str;
    case (flag,func,str)
      equation 
        true = RTOpts.debug_flag(flag);
        func(str); then ();
    case (_,_,_) then (); 
  end matchcontinue;
end fcall;

public function fcall0 "Call the given function (2nd arg) if the flag given in 1st arg is set "
  input String in_string;
  input FuncType1 in_functype1;
protected 
  function FuncType1
  end FuncType1;
algorithm 
  _:=
  matchcontinue (in_string,in_functype1)
    local
      String flag;
      FuncType1 func;
    case (flag,func)
      equation 
        true = RTOpts.debug_flag(flag);
        func(); then ();
    case (_,_) then (); 
  end matchcontinue;
end fcall0;

public function fcallret "Call the given function (2nd arg) if the flag given in 1st arg is set, and 
  return its return value "
  input String in_string;
  input FuncType1 in_functype1;
  input Type_a in_type_a;
  input Type_b in_type_b;
  output Type_b out_type_b;
protected 
  function FuncType1
    input Type_a in_type_a;
    output Type_b out_type_b;
  protected 
    replaceable type Type_a;
    replaceable type Type_b;
  end FuncType1;
  replaceable type Type_a;
  replaceable type Type_b;
algorithm 
  out_type_b:=
  matchcontinue (in_string,in_functype1,in_type_a,in_type_b)
    local
      Type_b res,def;
      String flag;
      FuncType1 func;
      Type_a arg;
    case (flag,func,arg,def)
      equation 
        true = RTOpts.debug_flag(flag);
        res = func(arg); then res;
    case (_,_,_,def) then def; 
  end matchcontinue;
end fcallret;

public function bcall
  input Boolean in_boolean;
  input FuncType1 in_functype1;
  input Type_a in_type_a;
protected 
  function FuncType1
    input Type_a in_type_a;
  protected 
    replaceable type Type_a;
  end FuncType1;
  replaceable type Type_a;
algorithm 
  _:=
  matchcontinue (in_boolean,in_functype1,in_type_a)
    local
      FuncType1 func;
      Type_a str;
    case (true,func,str)
      equation 
        func(str); then ();
    case (false,_,_) then (); 
  end matchcontinue;
end bcall;

public function bcall_2
  input Boolean in_boolean;
  input FuncType1 in_functype1;
  input Type_a in_type_a;
  input Type_b in_type_b;
protected 
  function FuncType1
    input Type_a in_type_a;
    input Type_b in_type_b;
  protected 
    replaceable type Type_a;
    replaceable type Type_b;
  end FuncType1;
  replaceable type Type_a;
  replaceable type Type_b;
algorithm 
  _:=
  matchcontinue (in_boolean,in_functype1,in_type_a,in_type_b)
    local
      FuncType1 func;
      Type_a a;
      Type_b b;
    case (true,func,a,b)
      equation 
        func(a, b); then ();
    case (false,_,_,_) then (); 
  end matchcontinue;
end bcall_2;

public function notfcall "Call the given function (2nd arg) if the flag given in 1st arg is NOT set "
  input String in_string;
  input FuncType1 in_functype1;
  input Type_a in_type_a;
protected 
  function FuncType1
    input Type_a in_type_a;
  protected 
    replaceable type Type_a;
  end FuncType1;
  replaceable type Type_a;
algorithm 
  _:=
  matchcontinue (in_string,in_functype1,in_type_a)
    local
      String flag;
      FuncType1 func;
      Type_a str;
    case (flag,func,str)
      equation 
        false = RTOpts.debug_flag(flag);
        func(str); then ();
    case (_,_,_) then (); 
  end matchcontinue;
end notfcall;

public function fprint_list
  input String in_string1;
  input VType_aList in_vtype_alist2;
  input FuncType1 in_functype13;
  input String in_string4;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
  end FuncType1;
algorithm 
  _:=
  matchcontinue (in_string1,in_vtype_alist2,in_functype13,in_string4)
    local
      String flag,sep;
      VType_aList lst;
      FuncType1 func;
    case (flag,lst,func,sep)
      equation 
        true = RTOpts.debug_flag(flag);
        print_list(lst, func, sep); then ();
    case (_,_,_,_) then (); 
  end matchcontinue;
end fprint_list;

protected function print_list
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
  input String in_string;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
  end FuncType1;
algorithm 
  _:=
  matchcontinue (in_vtype_alist,in_functype1,in_string)
    local
      Type_a h;
      FuncType1 r;
      VType_aList t;
      String sep;
    case ({},_,_) then (); 
    case ({h},r,_)
      equation 
        r(h); then ();
    case (h :: t,r,sep)
      equation 
        r(h);
        Print.print_error_buf(sep);
        print_list(t, r, sep); then ();
  end matchcontinue;
end print_list;
end Debug;

