package Dump "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Dump.rml
  module:      Dump
  description: debug printing
 
  RCS: $Id: Dump.rml 1798 2005-05-31 06:32:26Z petar $
 
  Printing routines for debugging of the AST.  These functions do
  nothing but print the data structures to the standard output.
 
  The main entrypoint for this module is the function \"dump\" which
  takes an entire program as an argument, and prints it all in
  Modelica source form. The other interface functions can be used
  to print smaller portions of a program.
 "

public import OpenModelica.Compiler.Absyn;

public import OpenModelica.Compiler.Interactive;

public 
type Ident = String;

protected import OpenModelica.Compiler.ClassInf "!ignorecode";

protected import OpenModelica.Compiler.RTOpts;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Debug;

public function dump "Relations "
  input Absyn.Program in_program;
algorithm 
  _:=
  matchcontinue (in_program)
    local
      type Absyn_ClassList = list<Absyn.Class>;
      Absyn.Class c;
      Absyn_ClassList cs;
      Absyn.Within w;
    case Absyn.PROGRAM({},_) then (); 
    case Absyn.PROGRAM(c :: cs,w)
      equation 
        print_class(c);
        Print.print_buf("\n");
        dump(Absyn.PROGRAM(cs,w)); then ();
  end matchcontinue;
end dump;

public function unparse_str
  input Absyn.Program in_program;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_program)
    local
      type Absyn_ClassList = list<Absyn.Class>;
      Ident s1,s2,s3,str;
      Absyn_ClassList cs;
      Absyn.Within w;
    case Absyn.PROGRAM({},_) then ""; 
    case Absyn.PROGRAM(cs,w)
      equation 
        s1 = unparse_within(0, w);
        s2 = unparse_class_list(0, cs);
        s3 = string_append(s1, s2);
        str = string_append(s3, "\n"); then str;
  end matchcontinue;
end unparse_str;

public function unparse_class_list
  input Integer in_integer;
  input Absyn_ClassList in_absyn_classlist;
  output String out_string;
protected 
  type Absyn_ClassList = list<Absyn.Class>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_absyn_classlist)
    local
      Ident s1,s2,res;
      Integer i;
      Absyn.Class c;
      Absyn_ClassList cs;
    case (_,{}) then ""; 
    case (i,c :: cs)
      equation 
        s1 = unparse_class_str(i, c, "", "", "");
        s2 = unparse_class_list(i, cs);
        res = Util.string_append_list({s1,";\n",s2}); then res;
  end matchcontinue;
end unparse_class_list;

public function unparse_within
  input Integer in_integer;
  input Absyn.Within in_within;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_within)
    local
      Ident s1,s2,str;
      Integer i;
      Absyn.Path p;
    case (_,Absyn.TOP()) then ""; 
    case (i,Absyn.WITHIN(p))
      equation 
        s1 = indent_str(i);
        s2 = Absyn.path_string(p);
        str = Util.string_append_list({s1,"within ",s2,";\n"}); then str;
  end matchcontinue;
end unparse_within;

protected function unparse_class_str
  input Integer in_integer1;
  input Absyn.Class in_class2;
  input String in_string3;
  input String in_string4;
  input String in_string5;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer1,in_class2,in_string3,in_string4,in_string5)
    local
      type Absyn_ClassPartList = list<Absyn.ClassPart>;
      type IdentOption = Option<Ident>;
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_SubscriptListOption = Option<Absyn_SubscriptList>;
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type Absyn_EnumLiteralList = list<Absyn.EnumLiteral>;
      Ident is,s1,s2,s2_1,s3,s4,s5,str,n,fi,re,io,s6,s7,s8,s9;
      Integer i_1,i,indent;
      Boolean p,f,e;
      Absyn.Restriction r;
      Absyn_ClassPartList parts;
      IdentOption optcmt;
      Absyn_SubscriptListOption dim;
      Absyn.ElementAttributes attr;
      Absyn_ElementArgList m;
      Absyn_CommentOption cmt;
      Absyn_EnumLiteralList l;
    case (i,Absyn.CLASS(n,p,f,e,r,Absyn.PARTS(parts,optcmt),_),fi,re,io)
      equation 
        is = indent_str(i);
        s1 = select_string(p, "partial ", "");
        s2 = select_string(f, "final ", "");
        s2_1 = select_string(e, "encapsulated ", "");
        s3 = unparse_restriction_str(r);
        i_1 = i + 1;
        s4 = unparse_class_part_str_lst(i_1, parts, true);
        s5 = unparse_string_comment_option(optcmt);
        str = Util.string_append_list({is,s2_1,s1,s2,re,io,s3," ",n,s5,"\n",s4,is,"end ",n}); then str;
    case (indent,Absyn.CLASS(n,p,f,e,r,Absyn.DERIVED(i,dim,attr,m,optcmt),_),fi,re,io)
      local
        Absyn.Path i;
        Absyn_CommentOption optcmt;
      equation 
        is = indent_str(indent);
        s1 = select_string(p, "partial ", "");
        s2 = select_string(f, "final ", "");
        s2_1 = select_string(e, "encapsulated ", "");
        s3 = unparse_restriction_str(r);
        s4 = unparse_elementattr_str(attr);
        s5 = string_append(s1, s2);
        s6 = Absyn.path_string(i);
        s7 = get_option_str(dim, print_arraydim_str);
        s8 = unparse_mod1_str(m);
        s9 = unparse_comment_option(optcmt);
        str = Util.string_append_list({is,s2_1,s1,s2,re,io,s3," ",n,"= ",s4,s5,s6,s7,s8,s9}); then str;
    case (i,Absyn.CLASS(n,p,f,e,r,Absyn.ENUMERATION(l,cmt),_),fi,re,io)
      equation 
        is = indent_str(i);
        s1 = select_string(p, "partial ", "");
        s2 = select_string(f, "final ", "");
        s2_1 = select_string(e, "encapsulated ", "");
        s3 = unparse_restriction_str(r);
        s4 = unparse_enumliterals(l);
        s5 = unparse_comment_option(cmt);
        str = Util.string_append_list({is,s2_1,s1,s2,re,io,s3," ",n,"= enumeration(",s4,")",s5}); then str;
  end matchcontinue;
end unparse_class_str;

public function unparse_comment_option
  input Absyn_CommentOption in_absyn_commentoption;
  output String out_string;
protected 
  type Absyn_CommentOption = Option<Absyn.Comment>;
algorithm 
  out_string:=
  matchcontinue (in_absyn_commentoption)
    local
      type Absyn_AnnotationOption = Option<Absyn.Annotation>;
      Ident s1,str,cmt;
      Absyn_AnnotationOption annopt;
    case (NONE) then ""; 
    case (SOME(Absyn.COMMENT(annopt,SOME(cmt))))
      equation 
        s1 = unparse_annotation_option(0, annopt);
        str = Util.string_append_list({" \"",cmt,"\"",s1}); then str;
    case (SOME(Absyn.COMMENT(annopt,NONE)))
      equation 
        str = unparse_annotation_option(0, annopt); then str;
  end matchcontinue;
end unparse_comment_option;

protected function unparse_enumliterals
  input Absyn_EnumLiteralList in_absyn_enumliterallist;
  output String out_string;
protected 
  type Absyn_EnumLiteralList = list<Absyn.EnumLiteral>;
algorithm 
  out_string:=
  matchcontinue (in_absyn_enumliterallist)
    local
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Ident s1,s2,res,str,str2;
      Absyn_CommentOption optcmt,optcmt2;
      Absyn.EnumLiteral a;
      Absyn_EnumLiteralList b;
    case ({}) then ""; 
    case (Absyn.ENUMLITERAL(str,optcmt) :: a :: b)
      equation 
        s1 = unparse_comment_option(optcmt);
        s2 = unparse_enumliterals(a :: b);
        res = Util.string_append_list({str,s1,", ",s2}); then res;
    case ({Absyn.ENUMLITERAL(str,optcmt),Absyn.ENUMLITERAL(str2,optcmt2)})
      equation 
        s1 = unparse_comment_option(optcmt);
        s2 = unparse_comment_option(optcmt2);
        res = Util.string_append_list({str,s1,", ",str2,s2}); then res;
  end matchcontinue;
end unparse_enumliterals;

protected function unparse_restriction_str
  input Absyn.Restriction in_restriction;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_restriction)
    case Absyn.R_CLASS() then "class"; 
    case Absyn.R_MODEL() then "model"; 
    case Absyn.R_RECORD() then "record"; 
    case Absyn.R_BLOCK() then "block"; 
    case Absyn.R_CONNECTOR() then "connector"; 
    case Absyn.R_TYPE() then "type"; 
    case Absyn.R_PACKAGE() then "package"; 
    case Absyn.R_FUNCTION() then "function"; 
    case Absyn.R_PREDEFINED_INT() then "Integer"; 
    case Absyn.R_PREDEFINED_REAL() then "Real"; 
    case Absyn.R_PREDEFINED_STRING() then "String"; 
    case Absyn.R_PREDEFINED_BOOL() then "Boolean"; 
  end matchcontinue;
end unparse_restriction_str;

public function dump_istmt
  input Interactive.InteractiveStmts in_interactivestmts;
algorithm 
  _:=
  matchcontinue (in_interactivestmts)
    local
      type Interactive_InteractiveStmtList = list<Interactive.InteractiveStmt>;
      Absyn.AlgorithmItem alg;
      Absyn.Exp expr;
      Interactive_InteractiveStmtList l;
      Boolean sc;
    case (Interactive.ISTMTS({Interactive.IALG(alg)},_))
      equation 
        Print.print_buf("IALG(");
        print_algorithmitem(alg);
        Print.print_buf(")\n"); then ();
    case (Interactive.ISTMTS({Interactive.IEXP(expr)},_))
      equation 
        Print.print_buf("IEXP(");
        print_exp(expr);
        Print.print_buf(")\n"); then ();
    case (Interactive.ISTMTS(Interactive.IALG(alg) :: l,sc))
      equation 
        Print.print_buf("IALG(");
        print_algorithmitem(alg);
        Print.print_buf(",");
        dump_istmt(Interactive.ISTMTS(l,sc)); then ();
    case (Interactive.ISTMTS(Interactive.IEXP(expr) :: l,sc))
      equation 
        Print.print_buf("IEXP(");
        print_exp(expr);
        Print.print_buf(",");
        dump_istmt(Interactive.ISTMTS(l,sc)); then ();
    case (_) then (); 
  end matchcontinue;
end dump_istmt;

protected function print_class
  input Absyn.Class in_class;
algorithm 
  _:=
  matchcontinue (in_class)
    local
      type Absyn_ClassPartList = list<Absyn.ClassPart>;
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      type Absyn_EnumLiteralList = list<Absyn.EnumLiteral>;
      Ident n;
      Boolean p,f,e;
      Absyn.Restriction r;
      Absyn_ClassPartList parts;
      Absyn.Path i;
      Absyn_ElementArgList m;
      Absyn_EnumLiteralList l;
    case (Absyn.CLASS(n,p,f,e,r,Absyn.PARTS(parts,_),_))
      equation 
        print_class_restriction(r);
        Print.print_buf("(");
        Print.print_buf(n);
        Print.print_buf(",");
        print_select(p, "PARTIAL,", "");
        print_list_debug("print_class1", parts, print_class_part, ",");
        Print.print_buf(")"); then ();
    case (Absyn.CLASS(n,p,f,e,r,Absyn.DERIVED(i,_,_,m,_),_))
      equation 
        print_class_restriction(r);
        Print.print_buf("(");
        Print.print_buf(n);
        Print.print_buf(",");
        print_select(p, "PARTIAL,", "");
        Print.print_buf(") = ");
        print_path(i);
        print_class_modification(m); then ();
    case (Absyn.CLASS(n,p,f,e,r,Absyn.ENUMERATION(l,_),_))
      equation 
        print_class_restriction(r);
        Print.print_buf("(");
        Print.print_buf(n);
        Print.print_buf(",");
        print_select(p, "PARTIAL,", "");
        Print.print_buf("ENUMERATION(...)\n"); then ();
  end matchcontinue;
end print_class;

protected function print_class_restriction ""
  input Absyn.Restriction re;
algorithm 
  Absyn.print_restr(re);
end print_class_restriction;

protected function print_class_modification ""
  input Absyn_ElementArgList in_absyn_elementarglist;
protected 
  type Absyn_ElementArgList = list<Absyn.ElementArg>;
algorithm 
  _:=
  matchcontinue (in_absyn_elementarglist)
    local Absyn_ElementArgList l;
    case ({}) then (); 
    case (l)
      equation 
        Print.print_buf("(");
        print_list_debug("print_class_modification", l, print_element_arg, ",");
        Print.print_buf(")"); then ();
  end matchcontinue;
end print_class_modification;

protected function unparse_class_modification_str
  input Absyn.Modification in_modification;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_modification)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      Ident s1,s2,str;
      Absyn_ElementArgList l;
      Absyn.Exp e;
    case (Absyn.CLASSMOD({},_)) then ""; 
    case (Absyn.CLASSMOD(l,NONE))
      equation 
        s1 = get_string_list(l, unparse_element_arg_str, ",");
        s2 = string_append("(", s1);
        str = string_append(s2, ")"); then str;
    case (Absyn.CLASSMOD(_,SOME(e)))
      equation 
        s1 = print_exp_str(e);
        str = Util.string_append_list({"=",s1}); then str;
  end matchcontinue;
end unparse_class_modification_str;

protected function print_element_arg
  input Absyn.ElementArg in_elementarg;
algorithm 
  _:=
  matchcontinue (in_elementarg)
    local
      type Absyn_ModificationOption = Option<Absyn.Modification>;
      type IdentOption = Option<Ident>;
      Boolean f;
      Absyn.Each each_;
      Absyn.ComponentRef r;
      Absyn_ModificationOption optm;
      IdentOption optcmt;
      Absyn.ElementSpec spec;
    case (Absyn.MODIFICATION(f,each_,r,optm,optcmt))
      equation 
        Print.print_buf("ARG(");
        print_select(f, "FINAL,", ",");
        print_component_ref(r);
        print_opt_modification(optm);
        Print.print_buf(")"); then ();
    case (Absyn.REDECLARATION(f,each_,spec,_))
      equation 
        Print.print_buf("REDECLARATION(");
        print_select(f, "FINAL,", ",");
        print_elementspec(spec);
        Print.print_buf(",_)"); then ();
  end matchcontinue;
end print_element_arg;

protected function unparse_element_arg_str
  input Absyn.ElementArg in_elementarg;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementarg)
    local
      type Absyn_ModificationOption = Option<Absyn.Modification>;
      type IdentOption = Option<Ident>;
      Ident s1,s2,s3,s4,s5,str;
      Boolean f,repl;
      Absyn.Each each_;
      Absyn.ComponentRef r;
      Absyn_ModificationOption optm;
      IdentOption optstr;
      Absyn.ElementSpec spec,constropt;
      Absyn.Class cl;
    case (Absyn.MODIFICATION(f,each_,r,optm,optstr))
      equation 
        s1 = unparse_each_str(each_);
        s2 = select_string(f, "final ", "");
        s3 = print_component_ref_str(r);
        s4 = unparse_opt_modification_str(optm);
        s5 = unparse_string_comment_option(optstr);
        str = Util.string_append_list({s1,s2,s3,s4,s5}); then str;
    case (Absyn.REDECLARATION(f,each_,spec,SOME(constropt)))
      equation 
        s1 = unparse_each_str(each_);
        s2 = select_string(f, "final ", "");
        s3 = unparse_elementspec_str(0, spec, s2, "", "");
        s4 = unparse_elementspec_str(0, constropt, "", "", "");
        str = Util.string_append_list({s1,s2,"redeclare ",s3," ",s4}); then str;
    case (Absyn.REDECLARATION(f,each_,(spec = Absyn.CLASSDEF(repl,cl)),SOME(constropt)))
      equation 
        s1 = unparse_each_str(each_) "Since unparse_elementspec_str does not print \"replaceable\" we need to check for it here";
        s2 = select_string(f, "final ", "");
        s3 = select_string(repl, "replaceable ", "");
        s4 = unparse_elementspec_str(0, spec, "", s3, "");
        s5 = unparse_elementspec_str(0, constropt, "", "", "");
        str = Util.string_append_list({s1,s2,"redeclare ",s4," ",s5}); then str;
    case (Absyn.REDECLARATION(f,each_,(spec = Absyn.CLASSDEF(repl,cl)),NONE))
      equation 
        s1 = unparse_each_str(each_) "Since unparse_elementspec_str does not print \"replaceable\" we need to check for it here";
        s2 = select_string(f, "final ", "");
        s3 = select_string(repl, "replaceable ", "");
        s4 = unparse_elementspec_str(0, spec, "", s3, "");
        str = Util.string_append_list({s1,s2,"redeclare ",s4}); then str;
    case (Absyn.REDECLARATION(f,each_,spec,NONE))
      equation 
        s1 = unparse_each_str(each_);
        s2 = select_string(f, "final ", "");
        s3 = unparse_elementspec_str(0, spec, "", "", "");
        str = Util.string_append_list({s1,s2,"redeclare ",s3}); then str;
  end matchcontinue;
end unparse_element_arg_str;

protected function unparse_each_str ""
  input Absyn.Each in_each;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_each)
    case (Absyn.EACH()) then "each "; 
    case (Absyn.NON_EACH()) then ""; 
  end matchcontinue;
end unparse_each_str;

protected function print_class_part
  input Absyn.ClassPart in_classpart;
algorithm 
  _:=
  matchcontinue (in_classpart)
    local
      type Absyn_ElementItemList = list<Absyn.ElementItem>;
      type Absyn_EquationItemList = list<Absyn.EquationItem>;
      type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
      Absyn_ElementItemList el;
      Absyn_EquationItemList eqs;
      Absyn.ExternalDecl edecl;
    case (Absyn.PUBLIC(el))
      equation 
        Print.print_buf("PUBLIC(");
        print_elementitems(el);
        Print.print_buf(")"); then ();
    case (Absyn.PROTECTED(el))
      equation 
        Print.print_buf("PROTECTED(");
        print_elementitems(el);
        Print.print_buf(")"); then ();
    case (Absyn.EQUATIONS(eqs))
      equation 
        Print.print_buf("EQUATIONS(");
        print_list_debug("print_class_part", eqs, print_equationitem, ",");
        Print.print_buf(")"); then ();
    case (Absyn.INITIALEQUATIONS(eqs))
      equation 
        Print.print_buf("INITIALEQUATIONS(");
        print_list_debug("print_class_part", eqs, print_equationitem, ",");
        Print.print_buf(")"); then ();
    case (Absyn.ALGORITHMS(eqs))
      local Absyn_AlgorithmItemList eqs;
      equation 
        Print.print_buf("ALGORITHMS(");
        print_list_debug("print_class_part", eqs, print_algorithmitem, ",");
        Print.print_buf(")"); then ();
    case (Absyn.INITIALALGORITHMS(eqs))
      local Absyn_AlgorithmItemList eqs;
      equation 
        Print.print_buf("INITIALALGORITHMS(");
        print_list_debug("print_class_part", eqs, print_algorithmitem, ",");
        Print.print_buf(")"); then ();
    case (Absyn.EXTERNAL(edecl,_))
      equation 
        Print.print_buf("EXTERNAL(");
        print_external_decl(edecl);
        Print.print_buf(")"); then ();
  end matchcontinue;
end print_class_part;

protected function print_external_decl
  input Absyn.ExternalDecl in_externaldecl;
algorithm 
  _:=
  matchcontinue (in_externaldecl)
    local
      type IdentOption = Option<Ident>;
      type Absyn_ComponentRefOption = Option<Absyn.ComponentRef>;
      type Absyn_ExpList = list<Absyn.Exp>;
      Ident idstr,crefstr,expstr,str,lang;
      IdentOption id;
      Absyn_ComponentRefOption cref;
      Absyn_ExpList exps;
    case Absyn.EXTERNALDECL(id,NONE,cref,exps)
      equation 
        idstr = get_option_str(id, identity);
        crefstr = get_option_str(cref, print_component_ref_str);
        expstr = print_list_str(exps, print_exp_str, ",");
        str = Util.string_append_list({idstr,", ",crefstr,", (",expstr,")"});
        Print.print_buf(str); then ();
    case Absyn.EXTERNALDECL(id,SOME(lang),cref,exps)
      equation 
        idstr = get_option_str(id, identity);
        crefstr = get_option_str(cref, print_component_ref_str);
        expstr = print_list_str(exps, print_exp_str, ",");
        str = Util.string_append_list({idstr,", \"",lang,"\", ",crefstr,", (",expstr,")"});
        Print.print_buf(str); then ();
  end matchcontinue;
end print_external_decl;

protected function unparse_class_part_str_lst
  input Integer in_integer;
  input Absyn_ClassPartList in_absyn_classpartlist;
  input Boolean in_boolean;
  output String out_string;
protected 
  type Absyn_ClassPartList = list<Absyn.ClassPart>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_absyn_classpartlist,in_boolean)
    local
      Ident s1,s2,res;
      Integer i;
      Absyn.ClassPart x;
      Absyn_ClassPartList xs;
      Boolean skippublic;
    case (_,{},false) then ""; 
    case (i,x :: xs,skippublic)
      equation 
        s1 = unparse_class_part_str(i, x, skippublic);
        s2 = unparse_class_part_str_lst(i, xs, false);
        res = string_append(s1, s2); then res;
  end matchcontinue;
end unparse_class_part_str_lst;

protected function unparse_class_part_str
  input Integer in_integer;
  input Absyn.ClassPart in_classpart;
  input Boolean in_boolean;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_classpart,in_boolean)
    local
      type Absyn_ElementItemList = list<Absyn.ElementItem>;
      type Absyn_EquationItemList = list<Absyn.EquationItem>;
      type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
      type IdentOption = Option<Ident>;
      type Absyn_ExpList = list<Absyn.Exp>;
      type Absyn_AnnotationOption = Option<Absyn.Annotation>;
      Integer i,i_1;
      Ident s1,is,str,langstr,outputstr,expstr,annstr,ident,res;
      Absyn_ElementItemList el;
      Absyn_EquationItemList eqs;
      IdentOption lang;
      Absyn.ComponentRef output_;
      Absyn_ExpList expl;
      Absyn_AnnotationOption ann;
    case (i,Absyn.PUBLIC({}),_) then ""; 
    case (i,Absyn.PROTECTED({}),_) then ""; 
    case (i,Absyn.EQUATIONS({}),_) then ""; 
    case (i,Absyn.INITIALEQUATIONS({}),_) then ""; 
    case (i,Absyn.ALGORITHMS({}),_) then ""; 
    case (i,Absyn.INITIALALGORITHMS({}),_) then ""; 
    case (i,Absyn.PUBLIC(el),true)
      equation 
        s1 = unparse_elementitem_str_lst(i, el);
        i_1 = i - 1;
        is = indent_str(i_1);
        str = Util.string_append_list({is,s1}); then str;
    case (i,Absyn.PUBLIC(el),false)
      equation 
        s1 = unparse_elementitem_str_lst(i, el);
        i_1 = i - 1;
        is = indent_str(i_1);
        str = Util.string_append_list({is,"public \n",s1}); then str;
    case (i,Absyn.PROTECTED(el),_)
      equation 
        s1 = unparse_elementitem_str_lst(i, el);
        i_1 = i - 1;
        is = indent_str(i_1);
        str = Util.string_append_list({is,"protected \n",s1}); then str;
    case (i,Absyn.EQUATIONS(eqs),_)
      equation 
        s1 = unparse_equationitem_str_lst(i, eqs, ";\n");
        i_1 = i - 1;
        is = indent_str(i_1);
        str = Util.string_append_list({"\n",is,"equation \n",s1}); then str;
    case (i,Absyn.INITIALEQUATIONS(eqs),_)
      equation 
        s1 = unparse_equationitem_str_lst(i, eqs, ";\n");
        i_1 = i - 1;
        is = indent_str(i_1);
        str = Util.string_append_list({"\n",is,"initial equation \n",s1}); then str;
    case (i,Absyn.ALGORITHMS(eqs),_)
      local Absyn_AlgorithmItemList eqs;
      equation 
        s1 = unparse_algorithm_str_lst(i, eqs, "\n");
        i_1 = i - 1;
        is = indent_str(i_1);
        str = Util.string_append_list({is,"algorithm \n",s1}); then str;
    case (i,Absyn.INITIALALGORITHMS(eqs),_)
      local Absyn_AlgorithmItemList eqs;
      equation 
        s1 = unparse_algorithm_str_lst(i, eqs, "\n");
        i_1 = i - 1;
        is = indent_str(i_1);
        str = Util.string_append_list({is,"initial algorithm \n",s1}); then str;
    case (i,Absyn.EXTERNAL(Absyn.EXTERNALDECL(SOME(ident),lang,SOME(output_),expl),ann),_)
      equation 
        langstr = get_extlang_str(lang);
        outputstr = print_component_ref_str(output_);
        expstr = print_list_str(expl, print_exp_str, ",");
        s1 = string_append(langstr, " ");
        is = indent_str(i);
        annstr = unparse_annotation_option(i, ann);
        str = Util.string_append_list(
          {"\n",is,"external ",langstr," ",outputstr,"=",ident,"(",
          expstr,") ",annstr,";\n"}); then str;
    case (i,Absyn.EXTERNAL(Absyn.EXTERNALDECL(SOME(ident),lang,NONE,expl),ann),_)
      equation 
        langstr = get_extlang_str(lang);
        expstr = print_list_str(expl, print_exp_str, ",");
        s1 = string_append(langstr, " ");
        is = indent_str(i);
        annstr = unparse_annotation_option(i, ann);
        str = Util.string_append_list(
          {"\n",is,"external ",langstr," ",ident,"(",expstr,") ",
          annstr,";\n"}); then str;
    case (i,Absyn.EXTERNAL(Absyn.EXTERNALDECL(NONE,lang,NONE,_),ann),_)
      equation 
        is = indent_str(i);
        langstr = get_extlang_str(lang);
        annstr = unparse_annotation_option(i, ann);
        res = Util.string_append_list({"\n",is,"external ",langstr," ",annstr,";\n"}); then res;
  end matchcontinue;
end unparse_class_part_str;

protected function get_extlang_str
  input StringOption in_stringoption;
  output String out_string;
protected 
  type StringOption = Option<String>;
algorithm 
  out_string:=
  matchcontinue (in_stringoption)
    local Ident res,str;
    case (NONE) then ""; 
    case (SOME(str))
      equation 
        res = Util.string_append_list({"\"",str,"\""}); then res;
  end matchcontinue;
end get_extlang_str;

protected function print_elementitems
  input Absyn_ElementItemList in_absyn_elementitemlist;
protected 
  type Absyn_ElementItemList = list<Absyn.ElementItem>;
algorithm 
  _:=
  matchcontinue (in_absyn_elementitemlist)
    local
      Absyn.Element e;
      Absyn.Annotation a;
      Absyn_ElementItemList els;
    case {} then (); 
    case {Absyn.ELEMENTITEM(e)}
      equation 
        print_element(e); then ();
    case ({Absyn.ANNOTATIONITEM(a)})
      equation 
        print_annotation(a); then ();
    case Absyn.ELEMENTITEM(e) :: els
      equation 
        print_element(e);
        print_elementitems(els); then ();
    case Absyn.ANNOTATIONITEM(a) :: els
      equation 
        print_annotation(a);
        print_elementitems(els); then ();
    case _
      equation 
        Print.print_buf("Error print_elementitems\n"); then ();
  end matchcontinue;
end print_elementitems;

protected function print_annotation
  input Absyn.Annotation in_annotation;
algorithm 
  _:=
  matchcontinue (in_annotation)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      Absyn_ElementArgList mod;
    case (Absyn.ANNOTATION(mod))
      equation 
        Print.print_buf("ANNOTATION(");
        print_modification(Absyn.CLASSMOD(mod,NONE));
        Print.print_buf(")"); then ();
  end matchcontinue;
end print_annotation;

protected function unparse_elementitem_str_lst
  input Integer in_integer;
  input Absyn_ElementItemList in_absyn_elementitemlist;
  output String out_string;
protected 
  type Absyn_ElementItemList = list<Absyn.ElementItem>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_absyn_elementitemlist)
    local
      Ident s1,s2,res;
      Integer i;
      Absyn.ElementItem x;
      Absyn_ElementItemList xs;
    case (_,{}) then ""; 
    case (i,x :: xs)
      equation 
        s1 = unparse_elementitem_str(i, x);
        s2 = unparse_elementitem_str_lst(i, xs);
        res = Util.string_append_list({s1,"\n",s2}); then res;
  end matchcontinue;
end unparse_elementitem_str_lst;

protected function unparse_elementitem_str
  input Integer in_integer;
  input Absyn.ElementItem in_elementitem;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_elementitem)
    local
      Ident str,s1;
      Integer i;
      Absyn.Element e;
      Absyn.Annotation a;
    case (i,Absyn.ELEMENTITEM(e))
      equation 
        str = unparse_element_str(i, e); then str;
    case (i,Absyn.ANNOTATIONITEM(a))
      equation 
        s1 = unparse_annotation_option(i, SOME(a));
        str = string_append(s1, ";"); then str;
  end matchcontinue;
end unparse_elementitem_str;

public function unparse_annotation_option
  input Integer in_integer;
  input Absyn_AnnotationOption in_absyn_annotationoption;
  output String out_string;
protected 
  type Absyn_AnnotationOption = Option<Absyn.Annotation>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_absyn_annotationoption)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      Ident s1,s2,str,is;
      Absyn_ElementArgList mod;
      Integer i;
    case (0,SOME(Absyn.ANNOTATION(mod)))
      equation 
        s1 = unparse_class_modification_str(Absyn.CLASSMOD(mod,NONE));
        s2 = string_append(" annotation", s1);
        str = string_append(s2, ""); then str;
    case (i,SOME(Absyn.ANNOTATION(mod)))
      equation 
        s1 = unparse_class_modification_str(Absyn.CLASSMOD(mod,NONE));
        is = indent_str(i);
        str = Util.string_append_list({is,"annotation",s1}); then str;
    case (_,NONE) then ""; 
  end matchcontinue;
end unparse_annotation_option;

protected function print_element
  input Absyn.Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local
      type Absyn_ElementSpecOption = Option<Absyn.ElementSpec>;
      Ident s1,filename;
      Boolean final_,repl;
      Absyn.InnerOuter inout;
      Absyn.ElementSpec spec;
      Integer lineno;
      Absyn_ElementSpecOption constr;
    case (Absyn.ELEMENT(final_,repl,inout,_,spec,filename,lineno,constr))
      equation 
        Print.print_buf("ELEMENT(");
        print_select(final_, "FINAL,", "");
        print_innerouter(inout);
        Print.print_buf(", ");
        print_elementspec(spec);
        Print.print_buf(", ");
        Print.print_buf(filename);
        Print.print_buf(", ");
        s1 = int_string(lineno);
        Print.print_buf(s1);
        Print.print_buf(")"); then ();
  end matchcontinue;
end print_element;

protected function unparse_element_str
  input Integer in_integer;
  input Absyn.Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_element)
    local
      type Absyn_ElementSpecOption = Option<Absyn.ElementSpec>;
      Ident s1,s2,s3,s4,s5,is,str,filename;
      Integer i,lineno;
      Boolean final_,repl;
      Absyn.InnerOuter inout;
      Absyn.ElementSpec spec;
      Absyn_ElementSpecOption constr;
    case (i,Absyn.ELEMENT(final_,repl,inout,_,spec,filename,lineno,constr))
      equation 
        s1 = select_string(final_, "final ", "");
        s2 = select_string(repl, "replaceable ", "");
        s3 = unparse_innerouter_str(inout);
        s4 = unparse_elementspec_str(i, spec, s1, s2, s3);
        s5 = int_string(lineno);
        is = indent_str(i);
        str = Util.string_append_list({s4,";"}) "	Util.string_append_list({s4, \" \", filename,\" \", s5,\";\"}) => str "; then str;
  end matchcontinue;
end unparse_element_str;

protected function print_innerouter
  input Absyn.InnerOuter in_innerouter;
algorithm 
  _:=
  matchcontinue (in_innerouter)
    case (Absyn.INNER())
      equation 
        Print.print_buf("INNER"); then ();
    case (Absyn.OUTER())
      equation 
        Print.print_buf("OUTER"); then ();
    case (Absyn.UNSPECIFIED())
      equation 
        Print.print_buf(" "); then ();
  end matchcontinue;
end print_innerouter;

protected function unparse_innerouter_str
  input Absyn.InnerOuter in_innerouter;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_innerouter)
    case (Absyn.INNER()) then "inner "; 
    case (Absyn.OUTER()) then "outer "; 
    case (Absyn.UNSPECIFIED()) then ""; 
  end matchcontinue;
end unparse_innerouter_str;

public function print_elementspec
  input Absyn.ElementSpec in_elementspec;
algorithm 
  _:=
  matchcontinue (in_elementspec)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      type Absyn_ComponentItemList = list<Absyn.ComponentItem>;
      Boolean repl;
      Absyn.Class cl;
      Absyn.Path p,t;
      Absyn_ElementArgList l;
      Absyn.ElementAttributes attr;
      Absyn_ComponentItemList cs;
      Absyn.Import i;
    case (Absyn.CLASSDEF(repl,cl))
      equation 
        Print.print_buf("CLASSDEF(");
        print_select(repl, "REPLACEABLE,", "");
        print_class(cl);
        Print.print_buf(")"); then ();
    case (Absyn.EXTENDS(p,l))
      equation 
        Print.print_buf("EXTENDS(");
        print_path(p);
        Print.print_buf(",");
        print_list_debug("print_elementspec", l, print_element_arg, ",");
        Print.print_buf(")"); then ();
    case (Absyn.COMPONENTS(attr,t,cs))
      equation 
        Print.print_buf("COMPONENTS(");
        print_elementattr(attr);
        Print.print_buf(",");
        print_path(t);
        Print.print_buf(",");
        print_list_debug("print_elementspec", cs, print_componentitem, ",");
        Print.print_buf(")"); then ();
    case (Absyn.IMPORT(i,_))
      equation 
        Print.print_buf("IMPORT(");
        print_import(i);
        Print.print_buf(")"); then ();
    case (_)
      equation 
        Print.print_buf(" ##ERROR## "); then ();
  end matchcontinue;
end print_elementspec;

protected function unparse_elementspec_str
  input Integer in_integer1;
  input Absyn.ElementSpec in_elementspec2;
  input String in_string3;
  input String in_string4;
  input String in_string5;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer1,in_elementspec2,in_string3,in_string4,in_string5)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      type Absyn_ComponentItemList = list<Absyn.ComponentItem>;
      Ident str,f,r,io,s1,s2,is,s3,ad;
      Integer i,indent;
      Boolean repl;
      Absyn.Class cl;
      Absyn.Path p,t;
      Absyn_ElementArgList l;
      Absyn.ElementAttributes attr;
      Absyn_ComponentItemList cs;
    case (i,Absyn.CLASSDEF(repl,cl),f,r,io)
      equation 
        str = unparse_class_str(i, cl, f, r, io); then str;
    case (i,Absyn.EXTENDS(p,{}),f,r,io)
      equation 
        s1 = Absyn.path_string(p);
        s2 = string_append("extends ", s1);
        is = indent_str(i);
        str = Util.string_append_list({is,f,r,io,s2}); then str;
    case (i,Absyn.EXTENDS(p,l),f,r,io)
      equation 
        s1 = Absyn.path_string(p);
        s2 = string_append("extends ", s1);
        s3 = get_string_list(l, unparse_element_arg_str, ",");
        is = indent_str(i);
        str = Util.string_append_list({is,f,r,io,s2,"(",s3,")"}); then str;
    case (i,Absyn.COMPONENTS(attr,t,cs),f,r,io)
      equation 
        s1 = Absyn.path_string(t);
        s2 = unparse_elementattr_str(attr);
        ad = unparse_arraydim_in_attr(attr);
        s3 = get_string_list(cs, unparse_componentitem_str, ",");
        is = indent_str(i);
        str = Util.string_append_list({is,f,r,io,s2,s1,ad," ",s3}); then str;
    case (indent,Absyn.IMPORT(i,_),f,r,io)
      local Absyn.Import i;
      equation 
        s1 = unparse_import_str(i);
        s2 = string_append("import ", s1);
        is = indent_str(indent);
        str = Util.string_append_list({is,f,r,io,s2}); then str;
    case (_,_,_,_,_)
      equation 
        Print.print_buf(" ##ERROR## "); then "";
  end matchcontinue;
end unparse_elementspec_str;

public function print_import
  input Absyn.Import in_import;
algorithm 
  _:=
  matchcontinue (in_import)
    local
      Ident i;
      Absyn.Path p;
    case (Absyn.NAMED_IMPORT(i,p))
      equation 
        Print.print_buf(i);
        Print.print_buf(" = ");
        print_path(p); then ();
    case (Absyn.QUAL_IMPORT(p))
      equation 
        print_path(p); then ();
    case (Absyn.UNQUAL_IMPORT(p))
      equation 
        print_path(p);
        Print.print_buf(".*"); then ();
  end matchcontinue;
end print_import;

protected function unparse_import_str
  input Absyn.Import in_import;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_import)
    local
      Ident s1,s2,str,i;
      Absyn.Path p;
    case (Absyn.NAMED_IMPORT(i,p))
      equation 
        s1 = string_append(i, " = ");
        s2 = Absyn.path_string(p);
        str = string_append(s1, s2); then str;
    case (Absyn.QUAL_IMPORT(p))
      equation 
        str = Absyn.path_string(p); then str;
    case (Absyn.UNQUAL_IMPORT(p))
      equation 
        s1 = Absyn.path_string(p);
        str = string_append(s1, ".*"); then str;
  end matchcontinue;
end unparse_import_str;

protected function print_elementattr
  input Absyn.ElementAttributes in_elementattributes;
algorithm 
  _:=
  matchcontinue (in_elementattributes)
    local
      type IdentList = list<Ident>;
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      Ident fs,vs,ds;
      IdentList sl;
      Boolean fl;
      Absyn.Variability var;
      Absyn.Direction dir;
      Absyn_SubscriptList adim;
    case (Absyn.ATTR(fl,var,dir,adim))
      equation 
        Print.print_buf("ATTR(");
        fs = select_string(fl, "FLOW,", "");
        vs = variability_symbol(var);
        ds = direction_symbol(dir);
        sl = {fs,vs,ds};
        print_list_debug("print_elementattr", sl, Print.print_buf, ",");
        Print.print_buf(")"); then ();
    case (_)
      equation 
        Print.print_buf(" ##ERROR## print_elementattr"); then ();
  end matchcontinue;
end print_elementattr;

protected function unparse_elementattr_str
  input Absyn.ElementAttributes in_elementattributes;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementattributes)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      Ident fs,vs,ds,str;
      Boolean fl;
      Absyn.Variability var;
      Absyn.Direction dir;
      Absyn_SubscriptList adim;
    case (Absyn.ATTR(fl,var,dir,adim))
      equation 
        fs = select_string(fl, "flow ", "");
        vs = unparse_variability_symbol_str(var);
        ds = unparse_direction_symbol_str(dir);
        str = Util.string_append_list({fs,vs,ds}); then str;
    case (_)
      equation 
        Print.print_buf(" ##ERROR## unparse_elementattr_str"); then "";
  end matchcontinue;
end unparse_elementattr_str;

protected function unparse_arraydim_in_attr
  input Absyn.ElementAttributes in_elementattributes;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_elementattributes)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      Ident str;
      Absyn_SubscriptList adim;
    case (Absyn.ATTR(_,_,_,adim))
      equation 
        str = print_arraydim_str(adim); then str;
    case (_) then ""; 
  end matchcontinue;
end unparse_arraydim_in_attr;

public function variability_symbol
  input Absyn.Variability in_variability;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_variability)
    case (Absyn.VAR()) then ""; 
    case (Absyn.DISCRETE()) then "DISCRETE"; 
    case (Absyn.PARAM()) then "PARAM"; 
    case (Absyn.CONST()) then "CONST"; 
  end matchcontinue;
end variability_symbol;

public function direction_symbol
  input Absyn.Direction in_direction;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_direction)
    case (Absyn.BIDIR()) then ""; 
    case (Absyn.INPUT()) then "INPUT"; 
    case (Absyn.OUTPUT()) then "OUTPUT"; 
  end matchcontinue;
end direction_symbol;

protected function unparse_variability_symbol_str
  input Absyn.Variability in_variability;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_variability)
    case (Absyn.VAR()) then ""; 
    case (Absyn.DISCRETE()) then "discrete "; 
    case (Absyn.PARAM()) then "parameter "; 
    case (Absyn.CONST()) then "constant "; 
  end matchcontinue;
end unparse_variability_symbol_str;

protected function unparse_direction_symbol_str
  input Absyn.Direction in_direction;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_direction)
    case (Absyn.BIDIR()) then ""; 
    case (Absyn.INPUT()) then "input "; 
    case (Absyn.OUTPUT()) then "output "; 
  end matchcontinue;
end unparse_direction_symbol_str;

public function print_component ""
  input Absyn.Component in_component;
algorithm 
  _:=
  matchcontinue (in_component)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_ModificationOption = Option<Absyn.Modification>;
      Ident n;
      Absyn_SubscriptList a;
      Absyn_ModificationOption m;
    case (Absyn.COMPONENT(n,a,m))
      equation 
        Print.print_buf(n);
        print_arraydim(a);
        print_option(m, print_modification); then ();
  end matchcontinue;
end print_component;

protected function print_componentitem
  input Absyn.ComponentItem in_componentitem;
algorithm 
  _:=
  matchcontinue (in_componentitem)
    local
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Absyn.Component c;
      Absyn_CommentOption optcmt;
    case (Absyn.COMPONENTITEM(c,optcmt))
      equation 
        Print.print_buf("COMPONENTITEM(");
        print_component(c);
        Print.print_buf(",<comment>)"); then ();
  end matchcontinue;
end print_componentitem;

protected function unparse_component_str
  input Absyn.Component in_component;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_component)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_ModificationOption = Option<Absyn.Modification>;
      Ident s1,s2,s3,str,n;
      Absyn_SubscriptList a;
      Absyn_ModificationOption m;
    case (Absyn.COMPONENT(n,a,m))
      equation 
        s1 = print_arraydim_str(a);
        s2 = string_append(n, s1);
        s3 = get_option_str(m, unparse_modification_str);
        str = string_append(s2, s3); then str;
  end matchcontinue;
end unparse_component_str;

protected function unparse_componentitem_str
  input Absyn.ComponentItem in_componentitem;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_componentitem)
    local
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Ident s1,s2,str;
      Absyn.Component c;
      Absyn_CommentOption cmtopt;
    case (Absyn.COMPONENTITEM(c,cmtopt))
      equation 
        s1 = unparse_component_str(c);
        s2 = unparse_comment_option(cmtopt);
        str = string_append(s1, s2); then str;
  end matchcontinue;
end unparse_componentitem_str;

public function print_arraydim ""
  input Absyn.ArrayDim s;
algorithm 
  print_subscripts(s);
end print_arraydim;

public function print_arraydim_str
  input Absyn.ArrayDim s;
  output String str;
algorithm 
  str := print_subscripts_str(s);
end print_arraydim_str;

protected function print_subscript
  input Absyn.Subscript in_subscript;
algorithm 
  _:=
  matchcontinue (in_subscript)
    local Absyn.Exp e1;
    case (Absyn.NOSUB())
      equation 
        Print.print_buf(":"); then ();
    case (Absyn.SUBSCRIPT(e1))
      equation 
        print_exp(e1); then ();
  end matchcontinue;
end print_subscript;

protected function print_subscript_str
  input Absyn.Subscript in_subscript;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_subscript)
    local
      Ident s;
      Absyn.Exp e1;
    case (Absyn.NOSUB()) then ":"; 
    case (Absyn.SUBSCRIPT(e1))
      equation 
        s = print_exp_str(e1); then s;
  end matchcontinue;
end print_subscript_str;

protected function print_opt_modification "Modifications "
  input Absyn_ModificationOption in_absyn_modificationoption;
protected 
  type Absyn_ModificationOption = Option<Absyn.Modification>;
algorithm 
  _:=
  matchcontinue (in_absyn_modificationoption)
    local Absyn.Modification m;
    case (SOME(m))
      equation 
        print_modification(m); then ();
    case (NONE) then (); 
  end matchcontinue;
end print_opt_modification;

protected function print_modification
  input Absyn.Modification in_modification;
algorithm 
  _:=
  matchcontinue (in_modification)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      type Absyn_ExpOption = Option<Absyn.Exp>;
      Absyn_ElementArgList l;
      Absyn_ExpOption e;
    case (Absyn.CLASSMOD(l,e))
      equation 
        print_mod1(l);
        print_mod2(e); then ();
    case (_)
      equation 
        Print.print_buf("( ** MODIFICATION ** )"); then ();
  end matchcontinue;
end print_modification;

protected function print_mod1
  input Absyn_ElementArgList in_absyn_elementarglist;
protected 
  type Absyn_ElementArgList = list<Absyn.ElementArg>;
algorithm 
  _:=
  matchcontinue (in_absyn_elementarglist)
    local Absyn_ElementArgList l;
    case {} then (); 
    case l
      equation 
        Print.print_buf("(");
        print_list_debug("print_mod1", l, print_element_arg, ",");
        Print.print_buf(")"); then ();
  end matchcontinue;
end print_mod1;

protected function print_mod2
  input Absyn_ExpOption in_absyn_expoption;
protected 
  type Absyn_ExpOption = Option<Absyn.Exp>;
algorithm 
  _:=
  matchcontinue (in_absyn_expoption)
    local Absyn.Exp e;
    case NONE then (); 
    case SOME(e)
      equation 
        Print.print_buf(" = ");
        print_exp(e); then ();
  end matchcontinue;
end print_mod2;

protected function unparse_opt_modification_str
  input Absyn_ModificationOption in_absyn_modificationoption;
  output String out_string;
protected 
  type Absyn_ModificationOption = Option<Absyn.Modification>;
algorithm 
  out_string:=
  matchcontinue (in_absyn_modificationoption)
    local
      Ident str;
      Absyn.Modification opt;
    case (SOME(opt))
      equation 
        str = unparse_modification_str(opt); then str;
    case (NONE) then ""; 
  end matchcontinue;
end unparse_opt_modification_str;

public function unparse_modification_str
  input Absyn.Modification in_modification;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_modification)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      type Absyn_ExpOption = Option<Absyn.Exp>;
      Ident s1,s2,str;
      Absyn_ElementArgList l;
      Absyn_ExpOption e;
    case (Absyn.CLASSMOD({},NONE)) then "()"; 
    case (Absyn.CLASSMOD(l,e))
      equation 
        s1 = unparse_mod1_str(l);
        s2 = unparse_mod2_str(e);
        str = string_append(s1, s2); then str;
    case (_)
      equation 
        Print.print_buf(" Failure MODIFICATION \n"); then "";
  end matchcontinue;
end unparse_modification_str;

protected function unparse_mod1_str
  input Absyn_ElementArgList in_absyn_elementarglist;
  output String out_string;
protected 
  type Absyn_ElementArgList = list<Absyn.ElementArg>;
algorithm 
  out_string:=
  matchcontinue (in_absyn_elementarglist)
    local
      Ident s1,s2,str;
      Absyn_ElementArgList l;
    case {} then ""; 
    case l
      equation 
        s1 = get_string_list(l, unparse_element_arg_str, ",");
        s2 = string_append("(", s1);
        str = string_append(s2, ")"); then str;
  end matchcontinue;
end unparse_mod1_str;

protected function unparse_mod2_str
  input Absyn_ExpOption in_absyn_expoption;
  output String out_string;
protected 
  type Absyn_ExpOption = Option<Absyn.Exp>;
algorithm 
  out_string:=
  matchcontinue (in_absyn_expoption)
    local
      Ident s1,str;
      Absyn.Exp e;
    case NONE then ""; 
    case SOME(e)
      equation 
        s1 = print_exp_str(e);
        str = string_append("=", s1); then str;
  end matchcontinue;
end unparse_mod2_str;

public function print_equation "Equations "
  input Absyn.Equation in_equation;
algorithm 
  _:=
  matchcontinue (in_equation)
    local
      type Absyn_EquationItemList = list<Absyn.EquationItem>;
      type TupleAbsyn_ExpAbsyn_EquationItemList = tuple<Absyn.Exp,Absyn_EquationItemList>;
      type TupleAbsyn_ExpAbsyn_EquationItemListList = list<TupleAbsyn_ExpAbsyn_EquationItemList>;
      Absyn.Exp e,e1,e2;
      Absyn_EquationItemList tb,fb,el;
      TupleAbsyn_ExpAbsyn_EquationItemListList eb;
      Ident i;
    case (Absyn.EQ_IF(e,tb,eb,fb))
      equation 
        Print.print_buf("IF (");
        print_exp(e);
        Print.print_buf(") THEN ");
        print_list_debug("print_equation", tb, print_equationitem, ";");
        print_list_debug("print_equation", eb, print_eq_elseif, " ");
        Print.print_buf(" ELSE ");
        print_list_debug("print_equation", fb, print_equationitem, ";"); then ();
    case (Absyn.EQ_EQUALS(e1,e2))
      equation 
        Print.print_buf("EQ_EQUALS(");
        print_exp(e1);
        Print.print_buf(",");
        print_exp(e2);
        Print.print_buf(")"); then ();
    case (Absyn.EQ_CONNECT(e1,e2))
      local Absyn.ComponentRef e1,e2;
      equation 
        Print.print_buf("EQ_CONNECT(");
        print_component_ref(e1);
        Print.print_buf(",");
        print_component_ref(e2);
        Print.print_buf(")"); then ();
    case Absyn.EQ_FOR(i,e,el)
      equation 
        Print.print_buf("FOR ");
        Print.print_buf(i);
        Print.print_buf(" in ");
        print_exp(e);
        Print.print_buf(" {");
        print_list_debug("print_equation", el, print_equationitem, ";");
        Print.print_buf("}"); then ();
    case (_)
      equation 
        Print.print_buf(" ** UNKNOWN EQUATION ** "); then ();
  end matchcontinue;
end print_equation;

protected function print_equationitem
  input Absyn.EquationItem in_equationitem;
algorithm 
  _:=
  matchcontinue (in_equationitem)
    local Absyn.Equation eq;
    case Absyn.EQUATIONITEM(eq,_)
      equation 
        Print.print_buf("EQUATIONITEM(");
        print_equation(eq);
        Print.print_buf(", <comment>)\n"); then ();
    case Absyn.EQUATIONITEMANN(_)
      equation 
        Print.print_buf("EQUATIONITEMANN(<annotation>)\n"); then ();
  end matchcontinue;
end print_equationitem;

protected function unparse_equation_str
  input Integer in_integer;
  input Absyn.Equation in_equation;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_equation)
    local
      type Absyn_EquationItemList = list<Absyn.EquationItem>;
      type TupleAbsyn_ExpAbsyn_EquationItemList = tuple<Absyn.Exp,Absyn_EquationItemList>;
      type TupleAbsyn_ExpAbsyn_EquationItemListList = list<TupleAbsyn_ExpAbsyn_EquationItemList>;
      Ident s1,s2,is,str,s3,s4,id;
      Integer i_1,i,indent;
      Absyn.Exp e,e1,e2,exp;
      Absyn_EquationItemList tb,fb,el,eql;
      TupleAbsyn_ExpAbsyn_EquationItemListList eb,eqlelse;
      Absyn.FunctionArgs fargs;
    case (i,Absyn.EQ_IF(e,tb,{},{}))
      equation 
        s1 = print_exp_str(e);
        i_1 = i + 1;
        s2 = unparse_equationitem_str_lst(i_1, tb, ";\n");
        is = indent_str(i);
        str = Util.string_append_list({"if ",s1," then\n",is,s2,is,"end if"}); then str;
    case (i,Absyn.EQ_IF(e,tb,eb,fb))
      equation 
        s1 = print_exp_str(e);
        i_1 = i + 1;
        s2 = unparse_equationitem_str_lst(i_1, tb, ";\n");
        s3 = unparse_eq_elseif_str_lst(i_1, eb, "\n");
        s4 = unparse_equationitem_str_lst(i_1, fb, ";\n");
        is = indent_str(i);
        str = Util.string_append_list(
          {is,"if ",s1," then\n",s2,s3,"\n",is,"else\n",s4,"\n",is,
          "end if"}); then str;
    case (i,Absyn.EQ_EQUALS(e1,e2))
      equation 
        s1 = print_exp_str(e1);
        s2 = print_exp_str(e2);
        is = indent_str(i);
        str = Util.string_append_list({is,s1,"=",s2}); then str;
    case (i,Absyn.EQ_CONNECT(e1,e2))
      local Absyn.ComponentRef e1,e2;
      equation 
        s1 = print_component_ref_str(e1);
        s2 = print_component_ref_str(e2);
        is = indent_str(i);
        str = Util.string_append_list({is,"connect(",s1,",",s2,")"}); then str;
    case (indent,Absyn.EQ_FOR(i,e,el))
      local Ident i;
      equation 
        s1 = print_exp_str(e);
        s2 = unparse_equationitem_str_lst(indent, el, ";\n");
        is = indent_str(indent);
        str = Util.string_append_list({is,"for ",i," in ",s1," loop\n",s2,"\n",is,"end for"}); then str;
    case (i,Absyn.EQ_NORETCALL(id,fargs))
      equation 
        s2 = print_function_args_str(fargs);
        str = Util.string_append_list({id,"(",s2,")"}); then str;
    case (i,Absyn.EQ_WHEN_E(exp,eql,eqlelse))
      equation 
        s1 = print_exp_str(exp);
        i_1 = i + 1;
        s2 = unparse_equationitem_str_lst(i_1, eql, ";\n");
        is = indent_str(i);
        s4 = unparse_eq_elsewhen_str_lst(i_1, eqlelse);
        str = Util.string_append_list({is,"when ",s1," then\n",is,s2,is,s4,"\n",is,"end when"}); then str;
    case (_,_)
      equation 
        Print.print_buf(" ** Failure! UNKNOWN EQUATION ** "); then "";
  end matchcontinue;
end unparse_equation_str;

protected function unparse_equationitem_str_lst
  input Integer in_integer;
  input Absyn_EquationItemList in_absyn_equationitemlist;
  input String in_string;
  output String out_string;
protected 
  type Absyn_EquationItemList = list<Absyn.EquationItem>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_absyn_equationitemlist,in_string)
    local
      Ident s1,s2,res,sep;
      Integer i;
      Absyn.EquationItem x;
      Absyn_EquationItemList xs;
    case (_,{},_) then ""; 
    case (i,x :: xs,sep)
      equation 
        s1 = unparse_equationitem_str(i, x);
        s2 = unparse_equationitem_str_lst(i, xs, sep);
        res = Util.string_append_list({s1,sep,s2}); then res;
  end matchcontinue;
end unparse_equationitem_str_lst;

protected function unparse_equationitem_str
  input Integer in_integer;
  input Absyn.EquationItem in_equationitem;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_equationitem)
    local
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Ident s1,s2,str;
      Integer i;
      Absyn.Equation eq;
      Absyn_CommentOption optcmt;
      Absyn.Annotation ann;
    case (i,Absyn.EQUATIONITEM(eq,optcmt))
      equation 
        s1 = unparse_equation_str(i, eq);
        s2 = unparse_comment_option(optcmt);
        str = string_append(s1, s2); then str;
    case (i,Absyn.EQUATIONITEMANN(ann))
      equation 
        str = unparse_annotation_option(i, SOME(ann)); then str;
  end matchcontinue;
end unparse_equationitem_str;

protected function print_eq_elseif ""
  input TupleAbsyn_ExpAbsyn_EquationItemList in_tupleabsyn_expabsyn_equationitemlist;
protected 
  type Absyn_EquationItemList = list<Absyn.EquationItem>;
  type TupleAbsyn_ExpAbsyn_EquationItemList = tuple<Absyn.Exp,Absyn_EquationItemList>;
algorithm 
  _:=
  matchcontinue (in_tupleabsyn_expabsyn_equationitemlist)
    local
      Absyn.Exp e;
      Absyn_EquationItemList el;
    case ((e,el))
      equation 
        Print.print_buf(" ELSEIF ");
        print_exp(e);
        Print.print_buf(" THEN ");
        print_list_debug("print_eq_elseif", el, print_equationitem, ";"); then ();
  end matchcontinue;
end print_eq_elseif;

protected function unparse_eq_elseif_str_lst
  input Integer in_integer;
  input TupleAbsyn_ExpAbsyn_EquationItemListList in_tupleabsyn_expabsyn_equationitemlistlist;
  input String in_string;
  output String out_string;
protected 
  type Absyn_EquationItemList = list<Absyn.EquationItem>;
  type TupleAbsyn_ExpAbsyn_EquationItemList = tuple<Absyn.Exp,Absyn_EquationItemList>;
  type TupleAbsyn_ExpAbsyn_EquationItemListList = list<TupleAbsyn_ExpAbsyn_EquationItemList>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_tupleabsyn_expabsyn_equationitemlistlist,in_string)
    local
      Ident s1,res,sep,s2;
      Integer i;
      TupleAbsyn_ExpAbsyn_EquationItemList x1,x,x2;
      TupleAbsyn_ExpAbsyn_EquationItemListList xs;
    case (_,{},_) then ""; 
    case (i,{x1},sep)
      equation 
        s1 = unparse_eq_elseif_str(i, x1);
        res = Util.string_append_list({s1,sep}); then res;
    case (i,x :: (xs = _ :: _),sep)
      equation 
        s2 = unparse_eq_elseif_str_lst(i, xs, sep);
        s1 = unparse_eq_elseif_str(i, x);
        res = Util.string_append_list({s1,sep,s2}); then res;
    case (i,{x1,x2},sep)
      equation 
        s1 = unparse_eq_elseif_str(i, x1);
        s2 = unparse_eq_elseif_str(i, x2);
        res = Util.string_append_list({s1,sep,s2}); then res;
  end matchcontinue;
end unparse_eq_elseif_str_lst;

protected function unparse_eq_elseif_str
  input Integer in_integer;
  input TupleAbsyn_ExpAbsyn_EquationItemList in_tupleabsyn_expabsyn_equationitemlist;
  output String out_string;
protected 
  type Absyn_EquationItemList = list<Absyn.EquationItem>;
  type TupleAbsyn_ExpAbsyn_EquationItemList = tuple<Absyn.Exp,Absyn_EquationItemList>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_tupleabsyn_expabsyn_equationitemlist)
    local
      Ident s1,s2,is,res;
      Integer i_1,i;
      Absyn.Exp e;
      Absyn_EquationItemList el;
    case (i,(e,el))
      equation 
        s1 = print_exp_str(e);
        s2 = unparse_equationitem_str_lst(i, el, ";\n");
        i_1 = i - 1;
        is = indent_str(i_1);
        res = Util.string_append_list({"\n",is,"elseif ",s1," then\n",s2}); then res;
  end matchcontinue;
end unparse_eq_elseif_str;

protected function print_algorithmitem "Algorithm clauses "
  input Absyn.AlgorithmItem in_algorithmitem;
algorithm 
  _:=
  matchcontinue (in_algorithmitem)
    local
      Absyn.Algorithm alg;
      Absyn.Annotation ann;
    case (Absyn.ALGORITHMITEM(alg,_))
      equation 
        Print.print_buf("ALGORITHMITEM(");
        print_algorithm(alg);
        Print.print_buf(")\n"); then ();
    case (Absyn.ALGORITHMITEMANN(ann))
      equation 
        Print.print_buf("ALGORITHMITEMANN(<annotation>)\n"); then ();
  end matchcontinue;
end print_algorithmitem;

public function print_algorithm
  input Absyn.Algorithm in_algorithm;
algorithm 
  _:=
  matchcontinue (in_algorithm)
    local
      type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
      type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
      type TupleAbsyn_ExpAbsyn_AlgorithmItemListList = list<TupleAbsyn_ExpAbsyn_AlgorithmItemList>;
      Absyn.ComponentRef cr;
      Absyn.Exp exp,e1,e2,e;
      Absyn_AlgorithmItemList tb,fb,el,al;
      TupleAbsyn_ExpAbsyn_AlgorithmItemListList eb;
      Ident i;
    case (Absyn.ALG_ASSIGN(cr,exp))
      equation 
        Print.print_buf("ALG_ASSIGN(");
        print_component_ref(cr);
        Print.print_buf(" := ");
        print_exp(exp);
        Print.print_buf(")"); then ();
    case (Absyn.ALG_TUPLE_ASSIGN(e1,e2))
      equation 
        Print.print_buf("ALG_TUPLE_ASSIGN(");
        print_exp(e1);
        Print.print_buf(" := ");
        print_exp(e2);
        Print.print_buf(")"); then ();
    case (Absyn.ALG_IF(e,tb,eb,fb))
      equation 
        Print.print_buf("IF (");
        print_exp(e);
        Print.print_buf(") THEN ");
        print_list_debug("print_algorithm", tb, print_algorithmitem, ";");
        print_list_debug("print_algorithm", eb, print_alg_elseif, " ");
        Print.print_buf(" ELSE ");
        print_list_debug("print_algorithm", fb, print_algorithmitem, ";"); then ();
    case Absyn.ALG_FOR(i,e,el)
      equation 
        Print.print_buf("FOR ");
        Print.print_buf(i);
        Print.print_buf(" in ");
        print_exp(e);
        Print.print_buf(" {");
        print_list_debug("print_algorithm", el, print_algorithmitem, ";");
        Print.print_buf("}"); then ();
    case Absyn.ALG_WHILE(e,al)
      equation 
        Print.print_buf("WHILE ");
        print_exp(e);
        Print.print_buf(" {");
        print_list_debug("print_algorithm", al, print_algorithmitem, ";");
        Print.print_buf("}"); then ();
    case Absyn.ALG_WHEN_A(e,al,el) " rule	Print.print_buf \"WHEN_E \" & print_exp(e) &
	Print.print_buf \" {\" & print_list_debug(\"print_algorithm\",al, print_algorithmitem, \";\") & Print.print_buf \"}\"
	 ----------------------------------------------------------
	print_algorithm Absyn.ALG_WHEN_E(e,al)
"
      local TupleAbsyn_ExpAbsyn_AlgorithmItemListList el;
      equation 
        Print.print_buf("WHEN_A ");
        print_exp(e);
        Print.print_buf(" {");
        print_list_debug("print_algorithm", al, print_algorithmitem, ";");
        Print.print_buf("}"); then ();
    case (_)
      equation 
        Print.print_buf(" ** UNKNOWN ALGORITHM CLAUSE ** "); then ();
  end matchcontinue;
end print_algorithm;

protected function unparse_algorithm_str_lst
  input Integer in_integer;
  input Absyn_AlgorithmItemList in_absyn_algorithmitemlist;
  input String in_string;
  output String out_string;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_absyn_algorithmitemlist,in_string)
    local
      Ident s1,res,sep,s2;
      Integer i;
      Absyn.AlgorithmItem x,x1,x2;
      Absyn_AlgorithmItemList xs;
    case (_,{},_) then ""; 
    case (i,{x},sep)
      equation 
        s1 = unparse_algorithm_str(i, x);
        res = string_append(s1, sep); then res;
    case (i,{x1,x2},sep)
      equation 
        s1 = unparse_algorithm_str(i, x1);
        s2 = unparse_algorithm_str(i, x2);
        res = Util.string_append_list({s1,sep,s2}); then res;
    case (i,x :: xs,sep)
      equation 
        s1 = unparse_algorithm_str(i, x);
        s2 = unparse_algorithm_str_lst(i, xs, sep);
        res = Util.string_append_list({s1,sep,s2}); then res;
  end matchcontinue;
end unparse_algorithm_str_lst;

public function unparse_algorithm_str
  input Integer in_integer;
  input Absyn.AlgorithmItem in_algorithmitem;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_algorithmitem)
    local
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
      type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
      type TupleAbsyn_ExpAbsyn_AlgorithmItemListList = list<TupleAbsyn_ExpAbsyn_AlgorithmItemList>;
      Ident s1,s2,s3,is,str,s4,s5,str_1;
      Integer i,i_1,ident_1,ident;
      Absyn.ComponentRef cr;
      Absyn.Exp exp,e;
      Absyn_CommentOption optcmt;
      Absyn_AlgorithmItemList tb,fb,el,al;
      TupleAbsyn_ExpAbsyn_AlgorithmItemListList eb,al2;
      Absyn.FunctionArgs fargs;
      Absyn.Annotation ann;
    case (i,Absyn.ALGORITHMITEM(Absyn.ALG_ASSIGN(cr,exp),optcmt))
      equation 
        s1 = print_component_ref_str(cr);
        s2 = print_exp_str(exp);
        s3 = unparse_comment_option(optcmt);
        is = indent_str(i);
        str = Util.string_append_list({is,s1,":=",s2,s3,";"}); then str;
    case (i,Absyn.ALGORITHMITEM(Absyn.ALG_IF(e,tb,eb,fb),optcmt))
      equation 
        s1 = print_exp_str(e);
        i_1 = i + 1;
        s2 = unparse_algorithm_str_lst(i, tb, "\n");
        s3 = unparse_alg_elseif_str_lst(i_1, eb, "\n");
        s4 = unparse_algorithm_str_lst(i, fb, "\n");
        s5 = unparse_comment_option(optcmt);
        is = indent_str(i);
        str = Util.string_append_list(
          {is,"if ",s1," then \n",is,s2,s3,"\n",is,"else ",s4,"\n",is,
          "end if",s5,";"}); then str;
    case (ident,Absyn.ALGORITHMITEM(Absyn.ALG_FOR(i,e,el),optcmt))
      local Ident i;
      equation 
        s1 = print_exp_str(e);
        ident_1 = ident + 1;
        s2 = unparse_algorithm_str_lst(ident_1, el, "\n");
        s3 = unparse_comment_option(optcmt);
        is = indent_str(ident);
        str = Util.string_append_list(
          {is,"for ",i," in ",s1," loop\n",is,s2,"\n",is,"end for",s3,
          ";"}); then str;
    case (i,Absyn.ALGORITHMITEM(Absyn.ALG_WHILE(e,al),optcmt))
      equation 
        s1 = print_exp_str(e);
        i_1 = i + 1;
        s2 = unparse_algorithm_str_lst(i_1, al, "\n");
        s3 = unparse_comment_option(optcmt);
        is = indent_str(i);
        str = Util.string_append_list(
          {is,"while (",s1,") loop\n",is,s2,"\n",is,"end while",s3,";"}); then str;
    case (i,Absyn.ALGORITHMITEM(Absyn.ALG_WHEN_A(e,al,al2),optcmt))
      equation 
        s1 = print_exp_str(e);
        i_1 = i + 1;
        s2 = unparse_algorithm_str_lst(i_1, al, "\n");
        s3 = unparse_comment_option(optcmt);
        is = indent_str(i);
        s4 = unparse_alg_elsewhen_str_lst(i_1, al2);
        str = Util.string_append_list(
          {is,"when ",s1," then\n",is,s2,is,s4,"\n",is,"end when",s3,
          ";"}); then str;
    case (i,Absyn.ALGORITHMITEM(Absyn.ALG_NORETCALL(cr,fargs),optcmt))
      equation 
        s1 = print_component_ref_str(cr);
        s2 = print_function_args_str(fargs);
        s3 = unparse_comment_option(optcmt);
        is = indent_str(i);
        str = Util.string_append_list({is,s1,"(",s2,")",s3,";"}); then str;
    case (i,Absyn.ALGORITHMITEMANN(ann))
      equation 
        str = unparse_annotation_option(i, SOME(ann));
        str_1 = string_append(str, ";"); then str_1;
    case (_,_)
      equation 
        Print.print_buf(" *Failure! UNKNOWN ALGORITHM CLAUSE ** "); then "";
  end matchcontinue;
end unparse_algorithm_str;

protected function unparse_alg_elsewhen_str_lst
  input Integer in_integer;
  input TupleAbsyn_ExpAbsyn_AlgorithmItemListList in_tupleabsyn_expabsyn_algorithmitemlistlist;
  output String out_string;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemListList = list<TupleAbsyn_ExpAbsyn_AlgorithmItemList>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_tupleabsyn_expabsyn_algorithmitemlistlist)
    local
      Ident res,s1,s2;
      Integer i;
      TupleAbsyn_ExpAbsyn_AlgorithmItemList x,x1,x2;
      TupleAbsyn_ExpAbsyn_AlgorithmItemListList xs;
    case (_,{}) then ""; 
    case (i,{x})
      equation 
        res = unparse_alg_elsewhen_str(i, x); then res;
    case (i,{x1,x2})
      equation 
        s1 = unparse_alg_elsewhen_str(i, x1);
        s2 = unparse_alg_elsewhen_str(i, x2);
        res = Util.string_append_list({s1,"\n",s2}); then res;
    case (i,x :: (xs = _ :: _))
      equation 
        s1 = unparse_alg_elsewhen_str(i, x);
        s2 = unparse_alg_elsewhen_str_lst(i, xs);
        res = Util.string_append_list({s1,"\n",s2}); then res;
  end matchcontinue;
end unparse_alg_elsewhen_str_lst;

protected function unparse_alg_elsewhen_str
  input Integer in_integer;
  input TupleAbsyn_ExpAbsyn_AlgorithmItemList in_tupleabsyn_expabsyn_algorithmitemlist;
  output String out_string;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_tupleabsyn_expabsyn_algorithmitemlist)
    local
      Ident is,s1,s2,res;
      Integer i;
      Absyn.Exp exp;
      Absyn_AlgorithmItemList algl;
    case (i,(exp,algl))
      equation 
        is = indent_str(i);
        s1 = unparse_algorithm_str_lst(i, algl, "\n");
        s2 = print_exp_str(exp);
        res = Util.string_append_list({"elsewhen ",s2," then\n",s1}); then res;
  end matchcontinue;
end unparse_alg_elsewhen_str;

protected function unparse_eq_elsewhen_str_lst
  input Integer in_integer;
  input TupleAbsyn_ExpAbsyn_EquationItemListList in_tupleabsyn_expabsyn_equationitemlistlist;
  output String out_string;
protected 
  type Absyn_EquationItemList = list<Absyn.EquationItem>;
  type TupleAbsyn_ExpAbsyn_EquationItemList = tuple<Absyn.Exp,Absyn_EquationItemList>;
  type TupleAbsyn_ExpAbsyn_EquationItemListList = list<TupleAbsyn_ExpAbsyn_EquationItemList>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_tupleabsyn_expabsyn_equationitemlistlist)
    local
      Ident res,s1,s2;
      Integer i;
      TupleAbsyn_ExpAbsyn_EquationItemList x,x1,x2;
      TupleAbsyn_ExpAbsyn_EquationItemListList xs;
    case (_,{}) then ""; 
    case (i,{x})
      equation 
        res = unparse_eq_elsewhen_str(i, x); then res;
    case (i,{x1,x2})
      equation 
        s1 = unparse_eq_elsewhen_str(i, x1);
        s2 = unparse_eq_elsewhen_str(i, x2);
        res = Util.string_append_list({s1,"\n",s2}); then res;
    case (i,x :: xs)
      equation 
        s1 = unparse_eq_elsewhen_str(i, x);
        s2 = unparse_eq_elsewhen_str_lst(i, xs);
        res = Util.string_append_list({s1,"\n",s2}); then res;
  end matchcontinue;
end unparse_eq_elsewhen_str_lst;

protected function unparse_eq_elsewhen_str
  input Integer in_integer;
  input TupleAbsyn_ExpAbsyn_EquationItemList in_tupleabsyn_expabsyn_equationitemlist;
  output String out_string;
protected 
  type Absyn_EquationItemList = list<Absyn.EquationItem>;
  type TupleAbsyn_ExpAbsyn_EquationItemList = tuple<Absyn.Exp,Absyn_EquationItemList>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_tupleabsyn_expabsyn_equationitemlist)
    local
      Ident is,s1,s2,res;
      Integer i;
      Absyn.Exp exp;
      Absyn_EquationItemList eql;
    case (i,(exp,eql))
      equation 
        is = indent_str(i);
        s1 = unparse_equationitem_str_lst(i, eql, ";\n");
        s2 = print_exp_str(exp);
        res = Util.string_append_list({"elsewhen ",s2," then\n",s1}); then res;
  end matchcontinue;
end unparse_eq_elsewhen_str;

protected function print_alg_elseif
  input TupleAbsyn_ExpAbsyn_AlgorithmItemList in_tupleabsyn_expabsyn_algorithmitemlist;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
algorithm 
  _:=
  matchcontinue (in_tupleabsyn_expabsyn_algorithmitemlist)
    local
      Absyn.Exp e;
      Absyn_AlgorithmItemList el;
    case ((e,el))
      equation 
        Print.print_buf(" ELSEIF ");
        print_exp(e);
        Print.print_buf(" THEN ");
        print_list_debug("print_alg_elseif", el, print_algorithmitem, ";"); then ();
  end matchcontinue;
end print_alg_elseif;

protected function unparse_alg_elseif_str_lst
  input Integer in_integer;
  input TupleAbsyn_ExpAbsyn_AlgorithmItemListList in_tupleabsyn_expabsyn_algorithmitemlistlist;
  input String in_string;
  output String out_string;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemListList = list<TupleAbsyn_ExpAbsyn_AlgorithmItemList>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_tupleabsyn_expabsyn_algorithmitemlistlist,in_string)
    local
      Ident s2,s1,res,sep;
      Integer i;
      TupleAbsyn_ExpAbsyn_AlgorithmItemList x,x1,x2;
      TupleAbsyn_ExpAbsyn_AlgorithmItemListList xs;
    case (_,{},_) then ""; 
    case (i,x :: (xs = _ :: _),sep)
      equation 
        s2 = unparse_alg_elseif_str_lst(i, xs, sep);
        s1 = unparse_alg_elseif_str(i, x);
        res = Util.string_append_list({s1,sep,s2}); then res;
    case (i,{x1,x2},sep)
      equation 
        s1 = unparse_alg_elseif_str(i, x1);
        s2 = unparse_alg_elseif_str(i, x2);
        res = Util.string_append_list({s1,sep,s2}); then res;
  end matchcontinue;
end unparse_alg_elseif_str_lst;

protected function unparse_alg_elseif_str
  input Integer in_integer;
  input TupleAbsyn_ExpAbsyn_AlgorithmItemList in_tupleabsyn_expabsyn_algorithmitemlist;
  output String out_string;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
algorithm 
  out_string:=
  matchcontinue (in_integer,in_tupleabsyn_expabsyn_algorithmitemlist)
    local
      Ident s1,s2,is,str;
      Integer i_1,i;
      Absyn.Exp e;
      Absyn_AlgorithmItemList el;
    case (i,(e,el))
      equation 
        s1 = print_exp_str(e);
        s2 = unparse_algorithm_str_lst(i, el, "\n");
        i_1 = i - 1;
        is = indent_str(i_1);
        str = Util.string_append_list({is,"elseif ",s1," then\n",s2}); then str;
  end matchcontinue;
end unparse_alg_elseif_str;

public function print_component_ref "Component references and paths 
   function: print_component_ref
 
  Print a `ComponentRef'.
 "
  input Absyn.ComponentRef in_componentref;
algorithm 
  _:=
  matchcontinue (in_componentref)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      Ident s;
      Absyn_SubscriptList subs;
      Absyn.ComponentRef cr;
    case Absyn.CREF_IDENT(s,subs)
      equation 
        Print.print_buf(s);
        print_subscripts(subs); then ();
    case Absyn.CREF_QUAL(s,subs,cr)
      equation 
        Print.print_buf(s);
        print_subscripts(subs);
        Print.print_buf(".");
        print_component_ref(cr); then ();
  end matchcontinue;
end print_component_ref;

public function print_subscripts
  input Absyn_SubscriptList in_absyn_subscriptlist;
protected 
  type Absyn_SubscriptList = list<Absyn.Subscript>;
algorithm 
  _:=
  matchcontinue (in_absyn_subscriptlist)
    local Absyn_SubscriptList l;
    case {} then (); 
    case l
      equation 
        Print.print_buf("[");
        print_list_debug("print_subscripts", l, print_subscript, ",");
        Print.print_buf("]"); then ();
  end matchcontinue;
end print_subscripts;

public function print_component_ref_str " function: print_component_ref_str
 
  Print a `ComponentRef' and return as a string.
 "
  input Absyn.ComponentRef in_componentref;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_componentref)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      Ident subsstr,s_1,s,crs,s_2,s_3;
      Absyn_SubscriptList subs;
      Absyn.ComponentRef cr;
    case Absyn.CREF_IDENT(s,subs)
      equation 
        subsstr = print_subscripts_str(subs);
        s_1 = string_append(s, subsstr); then s_1;
    case Absyn.CREF_QUAL(s,subs,cr)
      equation 
        crs = print_component_ref_str(cr);
        subsstr = print_subscripts_str(subs);
        s_1 = string_append(s, subsstr);
        s_2 = string_append(s_1, ".");
        s_3 = string_append(s_2, crs); then s_3;
  end matchcontinue;
end print_component_ref_str;

public function print_subscripts_str
  input Absyn_SubscriptList in_absyn_subscriptlist;
  output String out_string;
protected 
  type Absyn_SubscriptList = list<Absyn.Subscript>;
algorithm 
  out_string:=
  matchcontinue (in_absyn_subscriptlist)
    local
      Ident s,s_1,s_2;
      Absyn_SubscriptList l;
    case {} then ""; 
    case l
      equation 
        s = print_list_str(l, print_subscript_str, ",");
        s_1 = string_append("[", s);
        s_2 = string_append(s_1, "]"); then s_2;
  end matchcontinue;
end print_subscripts_str;

public function print_path " function: print_path
 
  Print a `Path'.
 "
  input Absyn.Path p;
protected 
  Ident s;
algorithm 
  s := Absyn.path_string(p);
  Print.print_buf(s);
end print_path;

protected function print_path_str " function: print_path
 
  Print a `Path'.
 "
  input Absyn.Path p;
  output String s;
algorithm 
  s := Absyn.path_string(p);
end print_path_str;

public function print_exp "- Expressions 
   function: print_exp
 
  This function prints a complete expression.
 "
  input Absyn.Exp in_exp;
algorithm 
  _:=
  matchcontinue (in_exp)
    local
      type TupleAbsyn_ExpAbsyn_Exp = tuple<Absyn.Exp,Absyn.Exp>;
      type TupleAbsyn_ExpAbsyn_ExpList = list<TupleAbsyn_ExpAbsyn_Exp>;
      type Absyn_ExpList = list<Absyn.Exp>;
      type Absyn_ExpListList = list<Absyn_ExpList>;
      Ident s,sym;
      Integer x;
      Absyn.ComponentRef c,fcn;
      Absyn.Exp e1,e2,e,t,f,start,stop,step;
      Absyn.Operator op;
      TupleAbsyn_ExpAbsyn_ExpList lst;
      Absyn.FunctionArgs args;
      Absyn_ExpList es;
    case (Absyn.INTEGER(x))
      equation 
        s = int_string(x);
        Print.print_buf(s); then ();
    case (Absyn.REAL(x))
      local Real x;
      equation 
        s = real_string(x);
        Print.print_buf(s); then ();
    case (Absyn.CREF(c))
      equation 
        print_component_ref(c); then ();
    case (Absyn.STRING(s))
      equation 
        Print.print_buf("\"");
        Print.print_buf(s);
        Print.print_buf("\""); then ();
    case (Absyn.BOOL(false))
      equation 
        Print.print_buf("FALSE"); then ();
    case (Absyn.BOOL(true))
      equation 
        Print.print_buf("TRUE"); then ();
    case (Absyn.BINARY(e1,op,e2))
      equation 
        sym = op_symbol(op);
        print_exp(e1);
        Print.print_buf(sym);
        print_exp(e2); then ();
    case (Absyn.UNARY(op,e))
      equation 
        sym = op_symbol(op);
        Print.print_buf(sym);
        print_exp(e); then ();
    case (Absyn.LBINARY(e1,op,e2))
      equation 
        sym = op_symbol(op);
        print_exp(e1);
        Print.print_buf(sym);
        print_exp(e2); then ();
    case (Absyn.LUNARY(op,e))
      equation 
        sym = op_symbol(op);
        Print.print_buf(sym);
        print_exp(e); then ();
    case (Absyn.RELATION(e1,op,e2))
      equation 
        sym = op_symbol(op);
        print_exp(e1);
        Print.print_buf(sym);
        print_exp(e2); then ();
    case (Absyn.IFEXP(c,t,f,lst))
      local Absyn.Exp c;
      equation 
        Print.print_buf("if ");
        print_exp(c);
        Print.print_buf(" then ");
        print_exp(t);
        Print.print_buf(" else ");
        print_exp(f); then ();
    case (Absyn.CALL(fcn,args))
      equation 
        print_component_ref(fcn);
        Print.print_buf("(");
        print_function_args(args);
        Print.print_buf(")"); then ();
    case Absyn.ARRAY(es)
      equation 
        Print.print_buf("ARRAY[");
        print_list_debug("print_exp", es, print_exp, ",");
        Print.print_buf("]"); then ();
    case Absyn.TUPLE(es) "PR. "
      equation 
        Print.print_buf("TUPLE");
        Print.print_buf("(");
        print_list_debug("print_exp", es, print_exp, ",");
        Print.print_buf(")"); then ();
    case Absyn.MATRIX(es)
      local Absyn_ExpListList es;
      equation 
        Print.print_buf("MATRIX[");
        print_list_debug("print_exp", es, print_row, ";");
        Print.print_buf("]"); then ();
    case Absyn.RANGE(start,NONE,stop)
      equation 
        Print.print_buf("RANGE(");
        print_exp(start);
        Print.print_buf(":");
        print_exp(stop);
        Print.print_buf(")"); then ();
    case Absyn.RANGE(start,SOME(step),stop)
      equation 
        Print.print_buf("RANGE(");
        print_exp(start);
        Print.print_buf(":");
        print_exp(step);
        Print.print_buf(":");
        print_exp(stop);
        Print.print_buf(")"); then ();
    case Absyn.END()
      equation 
        Print.print_buf("end"); then ();
    case (_)
      equation 
        Print.print_buf("#UNKNOWN EXPRESSION#"); then ();
  end matchcontinue;
end print_exp;

protected function print_function_args
  input Absyn.FunctionArgs in_functionargs;
algorithm 
  _:=
  matchcontinue (in_functionargs)
    local
      type Absyn_ExpList = list<Absyn.Exp>;
      type Absyn_NamedArgList = list<Absyn.NamedArg>;
      Absyn_ExpList expargs;
      Absyn_NamedArgList nargs;
      Absyn.Exp exp,iterexp;
      Ident id;
    case Absyn.FUNCTIONARGS(expargs,nargs)
      equation 
        Print.print_buf("FUNCTIONARGS(");
        print_list_debug("print_exp", expargs, print_exp, ", ");
        Print.print_buf(", ");
        print_list_debug("print_namedarg", nargs, print_named_arg, ", ");
        Print.print_buf(")"); then ();
    case Absyn.FOR_ITER_FARG(exp,id,iterexp)
      equation 
        Print.print_buf("FOR_ITER_FARG(");
        print_exp(exp);
        Print.print_buf(", ");
        Print.print_buf(id);
        Print.print_buf(", ");
        print_exp(iterexp);
        Print.print_buf(")"); then ();
  end matchcontinue;
end print_function_args;

protected function print_function_args_str
  input Absyn.FunctionArgs in_functionargs;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_functionargs)
    local
      type Absyn_ExpList = list<Absyn.Exp>;
      type Absyn_NamedArgList = list<Absyn.NamedArg>;
      Ident s1,s2,s3,str,estr,istr,id;
      Absyn_ExpList expargs;
      Absyn_NamedArgList nargs;
      Absyn.Exp exp,iterexp;
    case Absyn.FUNCTIONARGS((expargs = _ :: _),(nargs = _ :: _))
      equation 
        s1 = print_list_str(expargs, print_exp_str, ", ") "Both positional and named arguments ";
        s2 = string_append(s1, ", ");
        s3 = print_list_str(nargs, print_named_arg_str, ", ");
        str = string_append(s2, s3); then str;
    case Absyn.FUNCTIONARGS({},nargs)
      equation 
        str = print_list_str(nargs, print_named_arg_str, ", ") "Only named arguments "; then str;
    case Absyn.FUNCTIONARGS(expargs,{})
      equation 
        str = print_list_str(expargs, print_exp_str, ", ") "Only positional arguments "; then str;
    case Absyn.FOR_ITER_FARG(exp,id,iterexp)
      equation 
        estr = print_exp_str(exp);
        istr = print_exp_str(iterexp);
        str = Util.string_append_list({estr," for ",id," in ",istr}); then str;
  end matchcontinue;
end print_function_args_str;

public function print_named_arg
  input Absyn.NamedArg in_namedarg;
algorithm 
  _:=
  matchcontinue (in_namedarg)
    local
      Ident ident;
      Absyn.Exp e;
    case Absyn.NAMEDARG(ident,e)
      equation 
        Print.print_buf(ident);
        Print.print_buf("=");
        print_exp(e); then ();
  end matchcontinue;
end print_named_arg;

protected function print_named_arg_str
  input Absyn.NamedArg in_namedarg;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_namedarg)
    local
      Ident s1,s2,str,ident;
      Absyn.Exp e;
    case Absyn.NAMEDARG(ident,e)
      equation 
        s1 = string_append(ident, "=");
        s2 = print_exp_str(e);
        str = string_append(s1, s2); then str;
  end matchcontinue;
end print_named_arg_str;

protected function print_row ""
  input Absyn_ExpList es;
protected 
  type Absyn_ExpList = list<Absyn.Exp>;
algorithm 
  print_list_debug("print_row", es, print_exp, ",");
end print_row;

protected function exp_priority " function: exp_priority

 Returns a priority number for an expression.
 This function is used to output parenthesis when needed. eg 3*(1+2) should output 3*(1+2) 
 and not 3*1+2
 "
  input Absyn.Exp in_exp;
  output Integer out_integer;
algorithm 
  out_integer:=
  matchcontinue (in_exp)
    case (Absyn.INTEGER(_)) then 0; 
    case (Absyn.REAL(_)) then 0; 
    case (Absyn.STRING(_)) then 0; 
    case (Absyn.BOOL(_)) then 0; 
    case (Absyn.CREF(_)) then 0; 
    case (Absyn.END()) then 0; 
    case (Absyn.CALL(_,_)) then 0; 
    case (Absyn.ARRAY(_)) then 0; 
    case (Absyn.MATRIX(_)) then 0; 
    case (Absyn.BINARY(_,Absyn.POW(),_)) then 1; 
    case (Absyn.BINARY(_,Absyn.DIV(),_)) then 2; 
    case (Absyn.BINARY(_,Absyn.MUL(),_)) then 3; 
    case (Absyn.UNARY(Absyn.UPLUS(),_)) then 4; 
    case (Absyn.UNARY(Absyn.UMINUS(),_)) then 4; 
    case (Absyn.BINARY(_,Absyn.ADD(),_)) then 5; 
    case (Absyn.BINARY(_,Absyn.SUB(),_)) then 5; 
    case (Absyn.RELATION(_,Absyn.LESS(),_)) then 6; 
    case (Absyn.RELATION(_,Absyn.LESSEQ(),_)) then 6; 
    case (Absyn.RELATION(_,Absyn.GREATER(),_)) then 6; 
    case (Absyn.RELATION(_,Absyn.GREATEREQ(),_)) then 6; 
    case (Absyn.RELATION(_,Absyn.EQUAL(),_)) then 6; 
    case (Absyn.RELATION(_,Absyn.NEQUAL(),_)) then 6; 
    case (Absyn.LUNARY(Absyn.NOT(),_)) then 7; 
    case (Absyn.LBINARY(_,Absyn.AND(),_)) then 8; 
    case (Absyn.LBINARY(_,Absyn.OR(),_)) then 9; 
    case (Absyn.RANGE(_,_,_)) then 10; 
    case (Absyn.IFEXP(_,_,_,_)) then 11; 
    case (Absyn.TUPLE(_)) then 12; 
    case (_) then 13; 
  end matchcontinue;
end exp_priority;

protected function parenthesize "expressuion prio
  parent expr prio"
  input String in_string1;
  input Integer in_integer2;
  input Integer in_integer3;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_string1,in_integer2,in_integer3)
    local
      Ident str_1,str;
      Integer pparent,pexpr;
    case (str,pparent,pexpr) "expressuion prio parent expr prio"
      equation 
        (pparent > pexpr) = true;
        str_1 = Util.string_append_list({"(",str,")"}); then str_1;
    case (str,_,_) then str; 
  end matchcontinue;
end parenthesize;

public function print_exp_str " function: print_exp
 
  This function prints a complete expression.
 "
  input Absyn.Exp in_exp;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_exp)
    local
      type TupleAbsyn_ExpAbsyn_Exp = tuple<Absyn.Exp,Absyn.Exp>;
      type TupleAbsyn_ExpAbsyn_ExpList = list<TupleAbsyn_ExpAbsyn_Exp>;
      type Absyn_ExpList = list<Absyn.Exp>;
      type Absyn_ExpListList = list<Absyn_ExpList>;
      Ident s,s_1,s_2,sym,s1,s2,s1_1,s2_1,cs,ts,fs,cs_1,ts_1,fs_1,el,str,argsstr,s3,s3_1,res,res_1;
      Integer x,p,p1,p2,pc,pt,pf,pstart,pstop,pstep;
      Absyn.ComponentRef c,fcn;
      Boolean b;
      Absyn.Exp e,e1,e2,t,f,start,stop,step;
      Absyn.Operator op;
      TupleAbsyn_ExpAbsyn_ExpList elseif_;
      Absyn.FunctionArgs args;
      Absyn_ExpList es;
    case (Absyn.INTEGER(x))
      equation 
        s = int_string(x); then s;
    case (Absyn.REAL(x))
      local Real x;
      equation 
        s = real_string(x); then s;
    case (Absyn.CREF(c))
      equation 
        s = print_component_ref_str(c); then s;
    case (Absyn.STRING(s))
      equation 
        s_1 = string_append("\"", s);
        s_2 = string_append(s_1, "\""); then s_2;
    case (Absyn.BOOL(b))
      equation 
        s = print_bool_str(b); then s;
    case ((e = Absyn.BINARY(e1,op,e2)))
      equation 
        sym = op_symbol(op);
        s1 = print_exp_str(e1);
        s2 = print_exp_str(e2);
        p = exp_priority(e);
        p1 = exp_priority(e1);
        p2 = exp_priority(e2);
        s1_1 = parenthesize(s1, p1, p);
        s2_1 = parenthesize(s2, p2, p);
        s = string_append(s1_1, sym);
        s_1 = string_append(s, s2_1); then s_1;
    case ((e = Absyn.UNARY(op,e1)))
      equation 
        sym = op_symbol(op);
        s = print_exp_str(e1);
        p = exp_priority(e);
        p1 = exp_priority(e1);
        s_1 = parenthesize(s, p1, p);
        s_2 = string_append(sym, s_1); then s_2;
    case ((e = Absyn.LBINARY(e1,op,e2)))
      equation 
        sym = op_symbol(op);
        s1 = print_exp_str(e1);
        s2 = print_exp_str(e2);
        p = exp_priority(e);
        p1 = exp_priority(e1);
        p2 = exp_priority(e2);
        s1_1 = parenthesize(s1, p1, p);
        s2_1 = parenthesize(s2, p1, p);
        s = string_append(s1_1, sym);
        s_1 = string_append(s, s2_1); then s_1;
    case ((e = Absyn.LUNARY(op,e1)))
      equation 
        sym = op_symbol(op);
        s = print_exp_str(e1);
        p = exp_priority(e);
        p1 = exp_priority(e1);
        s_1 = parenthesize(s, p1, p);
        s_2 = string_append(sym, s_1); then s_2;
    case ((e = Absyn.RELATION(e1,op,e2)))
      equation 
        sym = op_symbol(op);
        s1 = print_exp_str(e1);
        s2 = print_exp_str(e2);
        p = exp_priority(e);
        p1 = exp_priority(e1);
        p2 = exp_priority(e2);
        s1_1 = parenthesize(s1, p1, p);
        s2_1 = parenthesize(s2, p1, p);
        s = string_append(s1_1, sym);
        s_1 = string_append(s, s2_1); then s_1;
    case ((e = Absyn.IFEXP(c,t,f,elseif_)))
      local Absyn.Exp c;
      equation 
        cs = print_exp_str(c);
        ts = print_exp_str(t);
        fs = print_exp_str(f);
        p = exp_priority(e);
        pc = exp_priority(c);
        pt = exp_priority(t);
        pf = exp_priority(f);
        cs_1 = parenthesize(cs, pc, p);
        ts_1 = parenthesize(ts, pt, p);
        fs_1 = parenthesize(fs, pf, p);
        el = print_elseif_str(elseif_);
        str = Util.string_append_list({"if ",cs_1," then ",ts_1,el," else ",fs_1}); then str;
    case (Absyn.CALL(fcn,args))
      equation 
        fs = print_component_ref_str(fcn);
        argsstr = print_function_args_str(args);
        s = string_append(fs, "(");
        s_1 = string_append(s, argsstr);
        s_2 = string_append(s_1, ")"); then s_2;
    case Absyn.ARRAY(es)
      equation 
        s = print_list_str(es, print_exp_str, ",") "Does not need parentheses ";
        s_1 = string_append("{", s);
        s_2 = string_append(s_1, "}"); then s_2;
    case Absyn.TUPLE(es)
      equation 
        s = print_list_str(es, print_exp_str, ",") "Does not need parentheses ";
        s_1 = string_append("(", s);
        s_2 = string_append(s_1, ")"); then s_2;
    case Absyn.MATRIX(es)
      local Absyn_ExpListList es;
      equation 
        s = print_list_str(es, print_row_str, ";") "Does not need parentheses ";
        s_1 = string_append("[", s);
        s_2 = string_append(s_1, "]"); then s_2;
    case ((e = Absyn.RANGE(start,NONE,stop)))
      equation 
        s1 = print_exp_str(start);
        s3 = print_exp_str(stop);
        p = exp_priority(e);
        pstart = exp_priority(start);
        pstop = exp_priority(stop);
        s1_1 = parenthesize(s1, pstart, p);
        s3_1 = parenthesize(s3, pstop, p);
        s = Util.string_append_list({s1_1,":",s3_1}); then s;
    case ((e = Absyn.RANGE(start,SOME(step),stop)))
      equation 
        s1 = print_exp_str(start);
        s2 = print_exp_str(step);
        s3 = print_exp_str(stop);
        p = exp_priority(e);
        pstart = exp_priority(start);
        pstop = exp_priority(stop);
        pstep = exp_priority(step);
        s1_1 = parenthesize(s1, pstart, p);
        s3_1 = parenthesize(s3, pstop, p);
        s2_1 = parenthesize(s2, pstep, p);
        s = Util.string_append_list({s1_1,":",s2_1,":",s3_1}); then s;
    case (Absyn.CODE(c))
      local Absyn.Code c;
      equation 
        res = print_code_str(c);
        res_1 = Util.string_append_list({"Code(",res,")"}); then res_1;
    case Absyn.END() then "end"; 
    case (_) then "#UNKNOWN EXPRESSION#"; 
  end matchcontinue;
end print_exp_str;

public function print_code_str
  input Absyn.Code in_code;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_code)
    local
      type Absyn_EquationItemList = list<Absyn.EquationItem>;
      type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
      Ident s,s1,s2,res;
      Absyn.Path p;
      Absyn.ComponentRef cr;
      Boolean b;
      Absyn_EquationItemList eqitems;
      Absyn_AlgorithmItemList algitems;
      Absyn.Element elt;
      Absyn.Exp exp;
      Absyn.Modification m;
    case (Absyn.C_TYPENAME(p))
      equation 
        s = print_path_str(p); then s;
    case (Absyn.C_VARIABLENAME(cr))
      equation 
        s = print_component_ref_str(cr); then s;
    case (Absyn.C_EQUATIONSECTION(b,eqitems))
      equation 
        s1 = select_string(b, "initial ", "");
        s2 = unparse_equationitem_str_lst(1, eqitems, ";\n");
        res = Util.string_append_list({s1,"equation ",s2}); then res;
    case (Absyn.C_ALGORITHMSECTION(b,algitems))
      equation 
        s1 = select_string(b, "initial ", "");
        s2 = unparse_algorithm_str_lst(1, algitems, ";\n");
        res = Util.string_append_list({s1,"algorithm ",s2}); then res;
    case (Absyn.C_ELEMENT(elt))
      equation 
        res = unparse_element_str(1, elt); then res;
    case (Absyn.C_EXPRESSION(exp))
      equation 
        res = print_exp_str(exp); then res;
    case (Absyn.C_MODIFICATION(m))
      equation 
        res = unparse_modification_str(m); then res;
  end matchcontinue;
end print_code_str;

protected function print_elseif_str
  input TupleAbsyn_ExpAbsyn_ExpList in_tupleabsyn_expabsyn_explist;
  output String out_string;
protected 
  type TupleAbsyn_ExpAbsyn_Exp = tuple<Absyn.Exp,Absyn.Exp>;
  type TupleAbsyn_ExpAbsyn_ExpList = list<TupleAbsyn_ExpAbsyn_Exp>;
algorithm 
  out_string:=
  matchcontinue (in_tupleabsyn_expabsyn_explist)
    local
      Ident s1,s2,s3,str;
      Absyn.Exp ec,ee;
      TupleAbsyn_ExpAbsyn_ExpList rest;
    case ({}) then ""; 
    case ((ec,ee) :: rest)
      equation 
        s1 = print_exp_str(ec);
        s2 = print_exp_str(ee);
        s3 = print_elseif_str(rest);
        str = Util.string_append_list({" elseif ",s1," then ",s2,s3}); then str;
  end matchcontinue;
end print_elseif_str;

protected function print_row_str
  input Absyn_ExpList es;
  output String s;
protected 
  type Absyn_ExpList = list<Absyn.Exp>;
algorithm 
  s := print_list_str(es, print_exp_str, ",");
end print_row_str;

protected function print_list_str " relation: print_list_str
 
  Same as print_list, except it returns a string
  instead of printing
 "
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
  input String in_string;
  output String out_string;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
    output String out_string;
  end FuncType1;
algorithm 
  out_string:=
  matchcontinue (in_vtype_alist,in_functype1,in_string)
    local
      Ident s,srest,s_1,s_2,sep;
      Type_a h;
      FuncType1 r;
      VType_aList t;
    case ({},_,_) then ""; 
    case ({h},r,_)
      equation 
        s = r(h); then s;
    case (h :: t,r,sep)
      equation 
        s = r(h);
        srest = print_list_str(t, r, sep);
        s_1 = string_append(s, sep);
        s_2 = string_append(s_1, srest); then s_2;
  end matchcontinue;
end print_list_str;

public function op_symbol " function: op_symbol
 
  Make a string describing different operators.
 "
  input Absyn.Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (Absyn.ADD()) then " + "; 
    case (Absyn.SUB()) then " - "; 
    case (Absyn.MUL()) then "*"; 
    case (Absyn.DIV()) then "/"; 
    case (Absyn.POW()) then "^"; 
    case (Absyn.UMINUS()) then "-"; 
    case (Absyn.UPLUS()) then "+"; 
    case (Absyn.AND()) then " and "; 
    case (Absyn.OR()) then " or "; 
    case (Absyn.NOT()) then "not "; 
    case (Absyn.LESS()) then " < "; 
    case (Absyn.LESSEQ()) then " <= "; 
    case (Absyn.GREATER()) then " > "; 
    case (Absyn.GREATEREQ()) then " >= "; 
    case (Absyn.EQUAL()) then " == "; 
    case (Absyn.NEQUAL()) then " <> "; 
  end matchcontinue;
end op_symbol;

public function select_string "- Utility functions
 
  These are utility functions used in some of the other
  relations. "
  input Boolean in_boolean1;
  input String in_string2;
  input String in_string3;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_boolean1,in_string2,in_string3)
    local Ident a,b;
    case (true,a,b) then a; 
    case (false,a,b) then b; 
  end matchcontinue;
end select_string;

public function print_select
  input Boolean f;
  input String yes;
  input String no;
protected 
  Ident res;
algorithm 
  res := select_string(f, yes, no);
  Print.print_buf(res);
end print_select;

public function print_option
  input VType_aOption in_vtype_aoption;
  input FuncType1 in_functype1;
protected 
  replaceable type Type_a;
  type VType_aOption = Option<Type_a>;
  function FuncType1
    input Type_a in_type_a;
  end FuncType1;
algorithm 
  _:=
  matchcontinue (in_vtype_aoption,in_functype1)
    local
      Type_a x;
      FuncType1 r;
    case (NONE,_) then (); 
    case (SOME(x),r)
      equation 
        r(x); then ();
  end matchcontinue;
end print_option;

public function print_list_debug
  input String in_string1;
  input VType_aList in_vtype_alist2;
  input FuncType1 in_functype13;
  input String in_string4;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
  end FuncType1;
algorithm 
  _:=
  matchcontinue (in_string1,in_vtype_alist2,in_functype13,in_string4)
    local
      Ident caller,s1,sep;
      Type_a h;
      FuncType1 r;
      VType_aList rest;
    case (_,{},_,_)
      equation 
        Debug.fprintln("dumptr", "print_list_debug-1"); then ();
    case (caller,{h},r,_)
      equation 
        Debug.fprintl("dumptr", {"print_list_debug-2 from ",caller,"\n"});
        r(h);
        Debug.fprintln("dumptr", "//print_list_debug-2"); then ();
    case (caller,h :: rest,r,sep)
      equation 
        s1 = string_append("print_list_debug-3 from ", caller);
        Debug.fprintl("dumptr", {s1,"\n"});
        r(h);
        Print.print_buf(sep);
        Debug.fprintln("dumptr", "//print_list_debug-3");
        print_list_debug(s1, rest, r, sep); then ();
  end matchcontinue;
end print_list_debug;

public function print_list
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
  input String in_string;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
  end FuncType1;
algorithm 
  _:=
  matchcontinue (in_vtype_alist,in_functype1,in_string)
    local
      Type_a h;
      FuncType1 r;
      VType_aList t;
      Ident sep;
    case ({},_,_) then (); 
    case ({h},r,_)
      equation 
        r(h); then ();
    case (h :: t,r,sep)
      equation 
        r(h);
        Print.print_buf(sep);
        print_list(t, r, sep); then ();
  end matchcontinue;
end print_list;

public function get_string_list
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
  input String in_string;
  output String out_string;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
    output String out_string;
  end FuncType1;
algorithm 
  out_string:=
  matchcontinue (in_vtype_alist,in_functype1,in_string)
    local
      Ident s,s_1,srest,s_2,sep;
      Type_a h;
      FuncType1 r;
      VType_aList t;
    case ({},_,_) then ""; 
    case ({h},r,_)
      equation 
        s = r(h); then s;
    case (h :: t,r,sep)
      equation 
        s = r(h);
        s_1 = string_append(s, sep);
        srest = get_string_list(t, r, sep);
        s_2 = string_append(s_1, srest); then s_2;
  end matchcontinue;
end get_string_list;

public function print_bool
  input Boolean b;
algorithm 
  print_select(b, "true", "false");
end print_bool;

public function get_option_str
  input VType_aOption in_vtype_aoption;
  input FuncType1 in_functype1;
  output String out_string;
protected 
  replaceable type Type_a;
  type VType_aOption = Option<Type_a>;
  function FuncType1
    input Type_a in_type_a;
    output String out_string;
  end FuncType1;
algorithm 
  out_string:=
  matchcontinue (in_vtype_aoption,in_functype1)
    local
      Ident str;
      Type_a a;
      FuncType1 r;
    case (SOME(a),r)
      equation 
        str = r(a); then str;
    case (NONE,_) then ""; 
  end matchcontinue;
end get_option_str;

public function get_option_with_concat_str
  input VType_aOption in_vtype_aoption;
  input FuncType1 in_functype1;
  input String in_string;
  output String out_string;
protected 
  replaceable type Type_a;
  type VType_aOption = Option<Type_a>;
  function FuncType1
    input Type_a in_type_a;
    output String out_string;
  end FuncType1;
algorithm 
  out_string:=
  matchcontinue (in_vtype_aoption,in_functype1,in_string)
    local
      Ident str,str_1,default_str;
      Type_a a;
      FuncType1 r;
    case (SOME(a),r,default_str)
      equation 
        str = r(a);
        str_1 = string_append(default_str, str); then str_1;
    case (NONE,_,default_str) then ""; 
  end matchcontinue;
end get_option_with_concat_str;

protected function unparse_string_comment_option
  input StringOption in_stringoption;
  output String out_string;
protected 
  type StringOption = Option<String>;
algorithm 
  out_string:=
  matchcontinue (in_stringoption)
    local Ident str,s;
    case (NONE) then ""; 
    case (SOME(s))
      equation 
        str = Util.string_append_list({" \"",s,"\""}); then str;
  end matchcontinue;
end unparse_string_comment_option;

protected function identity
  input Type_a in_type_a;
  output Type_a out_type_a;
protected 
  replaceable type Type_a;
algorithm 
  out_type_a:=
  matchcontinue (in_type_a)
    local Type_a x;
    case (x) then x; 
  end matchcontinue;
end identity;

public function print_bool_str
  input Boolean b;
  output String s;
algorithm 
  s := select_string(b, "true", "false");
end print_bool_str;

protected function stringlist_append
  input StringList in_stringlist;
  output String out_string;
protected 
  type StringList = list<String>;
algorithm 
  out_string:=
  matchcontinue (in_stringlist)
    local
      type IdentList = list<Ident>;
      Ident s2,str,s1;
      IdentList xs;
    case ({}) then ""; 
    case (s1 :: xs)
      equation 
        s2 = stringlist_append(xs);
        str = string_append(s1, s2); then str;
  end matchcontinue;
end stringlist_append;

public function indent_str
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer)
    local
      Integer i_1,i;
      Ident s1,res;
    case (0) then ""; 
    case (i)
      equation 
        i_1 = i - 1;
        s1 = indent_str(i_1);
        res = string_append(s1, "  ") "Indent using two whitespaces "; then res;
  end matchcontinue;
end indent_str;
end Dump;

