package Env "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Env.rml
  module:      Env
  description: Environmane management
 
  RCS: $Id: Env.rml 1835 2005-06-27 13:42:20Z petar $
 
  An environment is a stack of frames, where each frame contains a
  number of class and variable bindings.
  Each frame consist of:
  - a frame name (corresponding to the class partially instantiated in that frame)
  - a binary tree containing a list of classes
  - a binary tree containing a list of functions (functions are overloaded so serveral
						     function names can exist)
  - a list of unnamed items consisting of import statements
 
  As an example lets consider the following Modelica code:
  package A
    package B
     import Modelica.SIunits.*;
     constant Voltage V=3.3;
     function foo
     end foo;
     model M1
       Real x,y;  
     end M1;
     model M2
     end M2;
   end B;
  end A;
  
  When instantiating M1 we will first create the environment for its surrounding scope
  by a recursive instantiation on A.B giving the environment:
   { 
   FRAME(\"A\", {Class:B},{},{},false) ,  
   FRAME(\"B\", {Class:M1, Class:M2, Variable:V}, {Type:foo},
	   {import Modelica.SIunits.*},false)
   }
 
  Then, the class M1 is instantiated in a new scope/Frame giving the environment:
   { 
   FRAME(\"A\", {Class:B},{},{},false) ,  
   FRAME(\"B\", {Class:M1, Class:M2, Variable:V}, {Type:foo}, 
	   {Import Modelica.SIunits.*},false),
   FRAME(\"M1, {Variable:x, Variable:y},{},{},false)
   }

  NOTE: The instance hierachy (components and variables) and the class hierachy 
  (packages and classes) are combined into the same data structure, enabling a 
  uniform lookup mechanism
 "

public import OpenModelica.Compiler.Absyn;

public import OpenModelica.Compiler.Values;

public import OpenModelica.Compiler.SCode;

public import OpenModelica.Compiler.Types;

public import OpenModelica.Compiler.ClassInf;

public import OpenModelica.Compiler.Exp;

public 
type Ident = String;

public 
type IdentOption = Option<Ident>;

type ItemList = list<Item>;

uniontype Frame
  record FRAME
    IdentOption identoption1;
    BinTree bintree2 "Class name ";
    BinTree bintree3 "List of classes and variables which must be uniquely named";
    ItemList itemlist4 "List of types, which DOES NOT be uniquely named, eg. size have several types";
    Boolean boolean5 "list of unnamed items (imports)* encapsulated ";
  end FRAME;
end Frame;

public 
type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;

type TupleSCode_ElementTypes_ModOption = Option<TupleSCode_ElementTypes_Mod>;

type Types_TypeList = list<Types.Type>;

uniontype Item "bool=true means that FRAME is created due to encapsulated class  "
  record VAR
    Types.Var var;
    TupleSCode_ElementTypes_ModOption tuplescode_elementtypes_modoption;
    Boolean boolean "if it typed or not ";
  end VAR;
  record CLASS
    SCode.Class class_;
    Env env;
  end CLASS;
  record TYPE
    Types_TypeList types_typelist "list since several types with the same name can exist in the same scope (overloading) ";
  end TYPE;
  record IMPORT
    Absyn.Import import_;
  end IMPORT;
end Item;

public 
type Env = list<Frame>;

public 
type TreeValueOption = Option<TreeValue>;

type BinTreeOption = Option<BinTree>;

uniontype BinTree " The binary tree data structure
  ==============================
  The binary tree data structure used for the environment is generic and can 
  be used in any RML application.
  The Tree data structure BinTree is defined as:*"
  record TREENODE
    TreeValueOption treevalueoption;
    BinTreeOption value "Value ";
    BinTreeOption bintreeoption "left subtree  right subtree ";
  end TREENODE;
end BinTree;

public 
uniontype TreeValue "Each node in the binary tree can have a value associated with it."
  record TREEVALUE
    Key key "Key ";
    Value value "Value ";
  end TREEVALUE;
end TreeValue;

public 
type Key = Ident "Key ";

public 
type Value = Item;

protected import OpenModelica.Compiler.Dump;

protected import OpenModelica.Compiler.Graphviz;

protected import OpenModelica.Compiler.DAE;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.System;

public constant Env empty_env={} " - Values ";

public function new_frame " - Relations 
   relation: new_frame
  This function creates a new frame, which includes setting up the hashtable for the frame 
"
  input Boolean enc;
  output Frame out_frame;
protected 
  BinTree ht,httypes;
algorithm 
  ht := tree_new();
  httypes := tree_new();
  out_frame := FRAME(NONE,ht,httypes,{},enc);
end new_frame;

public function open_scope " relation: open_scope
 
  Opening a new scope in the environment mans adding a new frame on
  top of the stack of frames.
 "
  input Env in_env;
  input Boolean in_boolean;
  input IdentOption in_identoption;
  output Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_boolean,in_identoption)
    local
      Frame frame;
      Env env_1,env;
      Boolean encflag;
      Ident id;
    case (env,encflag,SOME(id)) "encapsulated?"
      equation 
        frame = new_frame(encflag);
        env_1 = name_scope(frame :: env, id); then env_1;
    case (env,encflag,NONE)
      equation 
        frame = new_frame(encflag); then frame :: env;
  end matchcontinue;
end open_scope;

protected function name_scope " relation: name_scope
 
  This function names the current scope, giving it an identifier.
  Scopes needs to be named for several reasons. First, it is needed for debugging purposes, since
  it is easier to follow the environment if we know what the current class being instantiated is.
 
  Secondly, it is needed when expanding type names in the context of flattening of the inheritance
  hierarchy. The reason for this is that types of inherited components needs to be expanded such that ** the types can be looked up from the environment of the base class.
 "
  input Env in_env;
  input Ident in_ident;
  output Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_ident)
    local
      type ValueList = list<Value>;
      BinTree ht,httypes;
      ValueList imps;
      Boolean encflag;
      Env res;
      Ident id;
    case (FRAME(_,ht,httypes,imps,encflag) :: res,id) then FRAME(SOME(id),ht,httypes,imps,encflag) :: res; 
  end matchcontinue;
end name_scope;

public function extend_frame_c " relation: extend_frame_c
 
  This function adds a class definition to the environment.
 "
  input Env in_env;
  input SCode.Class in_class;
  output Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_class)
    local
      type KeyOption = Option<Key>;
      type ValueList = list<Value>;
      BinTree ht_1,ht,httypes;
      Env env,fs;
      IdentOption id;
      ValueList imps;
      Boolean encflag;
      SCode.Class c;
      Ident n;
    case ((env = FRAME(id,ht,httypes,imps,encflag) :: fs),(c = SCode.CLASS(n,_,_,_,_))) "
  rule	Print.print_buf \"extend_frame_c:\n\" &
	print_env env &
	Print.print_buf \" + \" &
	Print.print_buf n &
	Print.print_buf \" = \" &
	print_env ((FRAME((n,CLASS(c,env))::items,encflag)::fs)) &
	Print.print_buf \"\n\" 
	----------------
	extend_frame_c(env as (FRAME(id,items,encflag)::fs),c as SCode.CLASS(n,_,_,_))
	  => ((FRAME(id,(n,CLASS(c,env))::items,encflag)::fs))
"
      equation 
        (ht_1) = tree_add(ht, n, CLASS(c,env), System.hash); then FRAME(id,ht_1,httypes,imps,encflag) :: fs;
    case (_,_)
      equation 
        print("extend_frame_c FAILED\n"); then fail();
  end matchcontinue;
end extend_frame_c;

public function extend_frame_classes
  input Env in_env;
  input SCode.Program in_program;
  output Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_program)
    local
      type SCode_ClassList = list<SCode.Class>;
      Env env,env_1,env_2;
      SCode.Class c;
      SCode_ClassList cs;
    case (env,{}) then env; 
    case (env,c :: cs)
      equation 
        env_1 = extend_frame_c(env, c);
        env_2 = extend_frame_classes(env_1, cs); then env_2;
  end matchcontinue;
end extend_frame_classes;

public function extend_frame_v " relation: extend_frame_v
 
  This function adds a component to the environment.
 "
  input Env in_env;
  input Types.Var in_var;
  input TupleSCode_ElementTypes_ModOption in_tuplescode_elementtypes_modoption;
  input Boolean in_boolean;
  output Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_var,in_tuplescode_elementtypes_modoption,in_boolean)
    local
      type KeyOption = Option<Key>;
      type ValueList = list<Value>;
      BinTree ht_1,ht,httypes;
      IdentOption id;
      ValueList imps;
      Boolean encflag,i;
      Env fs;
      Types.Var v;
      Ident n;
      TupleSCode_ElementTypes_ModOption c;
    case (FRAME(id,ht,httypes,imps,encflag) :: fs,(v = Types.VAR(n,_,_,_,_)),c,i) "
  rule	Print.print_buf \"extend_frame_v:\n\" &
   print_env env &
   Print.print_buf \" + \" &
   Print.print_buf n &
	Print.print_buf \" = \" &
   print_env (FRAME(id,(n,VAR(v))::items,encflag)::fs) &
   Print.print_buf \"\n\" 
   ----------------
   extend_frame_v(env as (FRAME(id,items,encflag)::fs),v as Types.VAR(n,_,_,_,_))
   => (FRAME(id,(n,VAR(v))::items,encflag)::fs)
 "
      equation 
        (ht_1) = tree_add(ht, n, VAR(v,c,i), System.hash); then FRAME(id,ht_1,httypes,imps,encflag) :: fs;
  end matchcontinue;
end extend_frame_v;

public function update_frame_v " relation: update_frame_v
 
  This function updates a component already added to the environment, but that prior to the
  update did not have any binding. (I.e this function is called in the second stage of 
   instantiation with declare before use.)
"
  input Env in_env;
  input Types.Var in_var;
  input Boolean in_boolean;
  output Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_var,in_boolean)
    local
      type KeyOption = Option<Key>;
      type ValueList = list<Value>;
      Boolean i,encflag;
      TupleSCode_ElementTypes_ModOption c;
      BinTree ht_1,ht,httypes;
      IdentOption sid;
      ValueList imps;
      Env fs,frames;
      Types.Var v;
      Ident n,id;
    case ({},_,i) then {}; 
    case (FRAME(sid,ht,httypes,imps,encflag) :: fs,(v = Types.VAR(n,_,_,_,_)),i)
      equation 
        VAR(_,c,_) = tree_get(ht, n, System.hash);
        (ht_1) = tree_add(ht, n, VAR(v,c,i), System.hash); then FRAME(sid,ht_1,httypes,imps,encflag) :: fs;
    case (FRAME(sid,ht,httypes,imps,encflag) :: fs,(v = Types.VAR(n,_,_,_,_)),i)
      equation 
        frames = update_frame_v(fs, v, i) "Also check frames above, e.g. when variable is in base class "; then FRAME(sid,ht,httypes,imps,encflag) :: frames;
    case (FRAME(sid,ht,httypes,imps,encflag) :: fs,Types.VAR(n,_,_,_,_),_)
      equation 
        Print.print_buf("- update_frame_v, variable ");
        Print.print_buf(n);
        Print.print_buf(" not found\n rest of env:");
        print_env(fs);
        Print.print_buf("\n"); then FRAME(sid,ht,httypes,imps,encflag) :: fs;
    case (_,(v = Types.VAR(id,_,_,_,_)),_)
      equation 
        Print.print_buf("- update_frame_v failed\n");
        Print.print_buf("  - variable: ");
        Types.print_var(v);
        Print.print_buf("\n"); then fail();
  end matchcontinue;
end update_frame_v;

public function extend_frame_t " relation: extend_frame_t
 
  This function adds a type to the environment.  Types in the
  environment are used for looking up constants etc. inside class
  definitions, such as packages.  For each type in the environment,
  there is a class definition with the same name in the
  environment.
 "
  input Env in_env;
  input Ident in_ident;
  input Types.Type in_type;
  output Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_ident,in_type)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type TupleTypes_TTypeAbsyn_PathOptionList = list<TupleTypes_TTypeAbsyn_PathOption>;
      type KeyOption = Option<Key>;
      type ValueList = list<Value>;
      TupleTypes_TTypeAbsyn_PathOptionList tps;
      BinTree httypes_1,ht,httypes;
      IdentOption sid;
      ValueList imps;
      Boolean encflag;
      Env fs;
      Ident n;
      TupleTypes_TTypeAbsyn_PathOption t;
    case (FRAME(sid,ht,httypes,imps,encflag) :: fs,n,t) "
  rule	Print.print_buf \"extend_frame_t:\n\" &
	print_env env &
	Print.print_buf \" + \" &
	Print.print_buf n &
	Print.print_buf \" = \" &
	print_env (FRAME(sid,(n,TYPE(t))::items,encflag)::fs) &
	Print.print_buf \"\n\" 
	----------------
	extend_frame_t(env as (FRAME(sid,items,encflag)::fs), n, t)
	  => (FRAME(sid,(n,TYPE(t))::items,encflag)::fs)
"
      equation 
        TYPE(tps) = tree_get(httypes, n, System.hash) "Other types with that name allready exist, add this type as well ";
        (httypes_1) = tree_add(httypes, n, TYPE(t :: tps), System.hash); then FRAME(sid,ht,httypes_1,imps,encflag) :: fs;
    case (FRAME(sid,ht,httypes,imps,encflag) :: fs,n,t)
      equation 
        failure(TYPE(_) = tree_get(httypes, n, System.hash)) "No other types exists ";
        (httypes_1) = tree_add(httypes, n, TYPE({t}), System.hash); then FRAME(sid,ht,httypes_1,imps,encflag) :: fs;
  end matchcontinue;
end extend_frame_t;

public function extend_frame_i
  input Env in_env;
  input Absyn.Import in_import;
  output Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_import)
    local
      type KeyOption = Option<Key>;
      type ValueList = list<Value>;
      IdentOption sid;
      BinTree ht,httypes;
      ValueList imps;
      Boolean encflag;
      Env fs;
      Absyn.Import imp;
    case (FRAME(sid,ht,httypes,imps,encflag) :: fs,imp) then FRAME(sid,ht,httypes,IMPORT(imp) :: imps,encflag) :: fs; 
  end matchcontinue;
end extend_frame_i;

public function top_frame
  input Env in_env;
  output Frame out_frame;
algorithm 
  out_frame:=
  matchcontinue (in_env)
    local
      Frame fr,elt;
      Env lst;
    case ({fr}) then fr; 
    case (elt :: (lst = _ :: _))
      equation 
        fr = top_frame(lst); then fr;
  end matchcontinue;
end top_frame;

public function get_env_path " relation: get_env_path
 
  This function returns all partially instantiated parents as an Absyn.Path option
  I.e. it collects all identifiers of each frame until it reaches the topmost
  unnamed frame. If the environment is only the topmost frame, NONE is returned.
"
  input Env in_env;
  output Absyn_PathOption out_absyn_pathoption;
protected 
  type Absyn_PathOption = Option<Absyn.Path>;
algorithm 
  out_absyn_pathoption:=
  matchcontinue (in_env)
    local
      Ident id;
      Absyn.Path path,path_1;
      Env rest;
    case ({FRAME(SOME(id),_,_,_,_),FRAME(NONE,_,_,_,_)}) then SOME(Absyn.IDENT(id)); 
    case (FRAME(SOME(id),_,_,_,_) :: rest)
      equation 
        SOME(path) = get_env_path(rest);
        path_1 = Absyn.join_paths(path, Absyn.IDENT(id)); then SOME(path_1);
    case (_) then NONE; 
  end matchcontinue;
end get_env_path;

public function print_env_path_str
  input Env in_env;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_env)
    local
      Absyn.Path path;
      Ident pathstr;
      Env env;
    case (env)
      equation 
        SOME(path) = get_env_path(env);
        pathstr = Absyn.path_string(path); then pathstr;
    case (env) then "<global scope>"; 
  end matchcontinue;
end print_env_path_str;

public function print_env_path
  input Env in_env;
algorithm 
  _:=
  matchcontinue (in_env)
    local
      Absyn.Path path;
      Ident pathstr;
      Env env;
    case (env)
      equation 
        SOME(path) = get_env_path(env);
        pathstr = Absyn.path_string(path);
        Print.print_buf(pathstr); then ();
    case (env)
      equation 
        Print.print_buf("TOPENV"); then ();
  end matchcontinue;
end print_env_path;

public function print_env_str
  input Env env;
  output String res;
protected 
  Ident s1;
algorithm 
  s1 := Print.get_string();
  Print.clear_buf();
  print_env(env);
  res := Print.get_string();
  Print.print_buf(s1);
end print_env_str;

public function print_env
  input Env in_env;
algorithm 
  _:=
  matchcontinue (in_env)
    local
      Frame fr;
      Env frs;
    case {}
      equation 
        Print.print_buf("Empty env\n"); then ();
    case fr :: frs
      equation 
        print_frame(fr);
        print_env(frs); then ();
  end matchcontinue;
end print_env;

public function print_frame
  input Frame in_frame;
algorithm 
  _:=
  matchcontinue (in_frame)
    local
      type ValueList = list<Value>;
      Ident sid;
      BinTree ht,httypes;
      ValueList imps;
      Boolean encflag;
    case FRAME(SOME(sid),ht,httypes,imps,encflag)
      equation 
        Print.print_buf("FRAME: ");
        Print.print_buf(sid);
        Print.print_buf(" (enc=");
        Dump.print_bool(encflag);
        Print.print_buf(") \nclasses and vars:\n=============\n");
        print_bintree(ht);
        Print.print_buf("   Types:\n======\n");
        print_bintree(httypes);
        Print.print_buf("   Imports:\n=======\n");
        print_imports(imps);
        Print.print_buf("\n\n\n"); then ();
    case FRAME(NONE,ht,httypes,imps,encflag)
      equation 
        Print.print_buf("FRAME: ");
        Print.print_buf("unnamed");
        Print.print_buf(" (enc=");
        Dump.print_bool(encflag);
        Print.print_buf(")\nclasses and vars:\n=============\n ");
        print_bintree(ht);
        Print.print_buf("\n   Types:\n======\n");
        print_bintree(httypes);
        Print.print_buf("\n   Imports:\n========\n");
        print_imports(imps);
        Print.print_buf("\n"); then ();
  end matchcontinue;
end print_frame;

protected function print_frame_contents
  input TupleIdentItemList in_tupleidentitemlist;
protected 
  type TupleIdentItem = tuple<Ident,Item>;
  type TupleIdentItemList = list<TupleIdentItem>;
algorithm 
  _:=
  matchcontinue (in_tupleidentitemlist)
    local
      type TupleKeyValue = tuple<Ident,Value>;
      type TupleKeyValueList = list<TupleKeyValue>;
      TupleKeyValue e;
      Ident n;
      Value i;
      TupleKeyValueList rst;
    case {} then (); 
    case {(e = (n,i))}
      equation 
        print_frame_element(e); then ();
    case ((e = (n,i)) :: rst)
      equation 
        print_frame_element(e);
        Print.print_buf("\n");
        print_frame_contents(rst); then ();
  end matchcontinue;
end print_frame_contents;

protected function print_imports
  input ItemList in_itemlist;
algorithm 
  _:=
  matchcontinue (in_itemlist)
    local
      type ValueList = list<Value>;
      Value e;
      ValueList rst;
    case {} then (); 
    case {e}
      equation 
        print_frame_element(("",e)); then ();
    case (e :: rst)
      equation 
        print_frame_element(("",e));
        Print.print_buf(",");
        print_imports(rst); then ();
  end matchcontinue;
end print_imports;

protected function print_frame_element
  input TupleIdentItem in_tupleidentitem;
protected 
  type TupleIdentItem = tuple<Ident,Item>;
algorithm 
  _:=
  matchcontinue (in_tupleidentitem)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type TupleTypes_TTypeAbsyn_PathOptionList = list<TupleTypes_TTypeAbsyn_PathOption>;
      Ident s,n,lenstr;
      Types.Var tv;
      SCode.Variability var;
      TupleTypes_TTypeAbsyn_PathOption tp;
      SCode.Element elt;
      Boolean i;
      Types.Binding bnd;
      Integer len;
      TupleTypes_TTypeAbsyn_PathOptionList lst;
      Absyn.Import imp;
    case ((n,VAR((tv = Types.VAR(_,Types.ATTR(_,_,var,_),_,tp,_)),SOME((elt,_)),i)))
      equation 
        Print.print_buf("v:");
        Print.print_buf(n);
        s = SCode.variability_string(var);
        Print.print_buf(s);
        Print.print_buf(" (");
        SCode.print_element(elt);
        Print.print_buf(") [");
        Types.print_type(tp);
        Print.print_buf("]");
        Print.print_buf("{");
        Types.print_var(tv);
        Print.print_buf("}");
        Print.print_buf("\n"); then ();
    case ((n,VAR(Types.VAR(_,_,_,_,bnd),NONE,i)))
      equation 
        Print.print_buf("v:");
        Print.print_buf(n);
        Print.print_buf("\n"); then ();
    case ((n,CLASS(_,_)))
      equation 
        Print.print_buf("c:");
        Print.print_buf(n);
        Print.print_buf("\n"); then ();
    case ((n,TYPE(lst)))
      equation 
        Print.print_buf("t:");
        Print.print_buf(n);
        len = list_length(lst);
        lenstr = int_string(len);
        Print.print_buf("(");
        Print.print_buf(lenstr);
        Print.print_buf(")\n"); then ();
    case ((n,IMPORT(imp)))
      equation 
        Print.print_buf("imp:");
        Dump.print_import(imp);
        Print.print_buf("\n"); then ();
  end matchcontinue;
end print_frame_element;

public function print_env_graphviz
  input TupleEnvString in_tupleenvstring;
protected 
  type TupleEnvString = tuple<Env,String>;
algorithm 
  _:=
  matchcontinue (in_tupleenvstring)
    local
      Graphviz.Node r;
      Env env;
      Ident str;
    case ((env,str))
      equation 
        r = build_env_graphviz((env,str));
        Graphviz.dump(r); then ();
  end matchcontinue;
end print_env_graphviz;

protected function build_env_graphviz
  input TupleEnvString in_tupleenvstring;
  output Graphviz.Node out_node;
protected 
  type TupleEnvString = tuple<Env,String>;
algorithm 
  out_node:=
  matchcontinue (in_tupleenvstring)
    local
      type Graphviz_NodeList = list<Graphviz.Node>;
      Ident str_1,str;
      Graphviz_NodeList nodelist;
      Env env;
    case ((env,str))
      equation 
        str_1 = string_append("ROOT: ", str);
        nodelist = build_env_graphviz_2(env); then Graphviz.NODE(str_1,{},nodelist);
  end matchcontinue;
end build_env_graphviz;

protected function build_env_graphviz_2
  input Env in_env;
  output Graphviz_NodeList out_graphviz_nodelist;
protected 
  type Graphviz_NodeList = list<Graphviz.Node>;
algorithm 
  out_graphviz_nodelist:=
  matchcontinue (in_env)
    local
      Graphviz_NodeList nodelist;
      Graphviz.Node node;
      Frame frame;
      Env rest;
    case {} then {}; 
    case frame :: rest
      equation 
        nodelist = build_env_graphviz_2(rest);
        node = build_frame_graphviz(frame); then node :: nodelist;
  end matchcontinue;
end build_env_graphviz_2;

protected function is_var_item
  input TupleType_aItem in_tupletype_aitem;
protected 
  replaceable type Type_a;
  type TupleType_aItem = tuple<Type_a,Item>;
algorithm 
  _:=
  matchcontinue (in_tupletype_aitem)
    case ((_,VAR(_,_,_))) then (); 
  end matchcontinue;
end is_var_item;

protected function is_class_item
  input TupleType_aItem in_tupletype_aitem;
protected 
  replaceable type Type_a;
  type TupleType_aItem = tuple<Type_a,Item>;
algorithm 
  _:=
  matchcontinue (in_tupletype_aitem)
    case ((_,CLASS(_,_))) then (); 
  end matchcontinue;
end is_class_item;

protected function is_type_item
  input TupleType_aItem in_tupletype_aitem;
protected 
  replaceable type Type_a;
  type TupleType_aItem = tuple<Type_a,Item>;
algorithm 
  _:=
  matchcontinue (in_tupletype_aitem)
    case ((_,TYPE(_))) then (); 
  end matchcontinue;
end is_type_item;

protected function build_frame_graphviz
  input Frame in_frame;
  output Graphviz.Node out_node;
algorithm 
  out_node:=
  matchcontinue (in_frame)
    local
      type KeyOption = Option<Key>;
      type ValueList = list<Value>;
      IdentOption sid;
      BinTree ht,httypes;
      ValueList imps;
      Boolean encflag;
    case FRAME(sid,ht,httypes,imps,encflag) then Graphviz.NODE("FRAME",{},{}); 
  end matchcontinue;
end build_frame_graphviz;

protected function build_item_listnode
  input TupleIdentItemList items;
  input FuncType1 cond;
  input String name;
  output Graphviz.Node out_node;
protected 
  type TupleIdentItem = tuple<Ident,Item>;
  type TupleIdentItemList = list<TupleIdentItem>;
  function FuncType1
    input TupleIdentItem in_tupleidentitem;
  end FuncType1;
  type TupleKeyValue = tuple<Ident,Value>;
  type TupleKeyValueList = list<TupleKeyValue>;
  TupleKeyValueList selitems;
  Graphviz.Node node;
algorithm 
  selitems := DAE.get_matching(items, cond);
  node := build_item_listnode_2(selitems, 1);
  out_node := Graphviz.NODE(name,{},{node});
end build_item_listnode;

protected function build_item_listnode_2
  input TupleIdentItemList in_tupleidentitemlist;
  input Integer in_integer;
  output Graphviz.Node out_node;
protected 
  type TupleIdentItem = tuple<Ident,Item>;
  type TupleIdentItemList = list<TupleIdentItem>;
algorithm 
  out_node:=
  matchcontinue (in_tupleidentitemlist,in_integer)
    local
      type IdentList = list<Ident>;
      type TupleKeyValue = tuple<Ident,Value>;
      type TupleKeyValueList = list<TupleKeyValue>;
      IdentList strlist;
      Ident cstr;
      TupleKeyValueList items,ignored;
      Integer count,count_1;
      Graphviz.Node restnode;
    case (items,count)
      equation 
        (strlist,{}) = DAE.build_gr_strlist(items, build_item_str, 10);
        cstr = int_string(count); then Graphviz.LNODE(cstr,strlist,{Graphviz.box},{});
    case (items,count)
      equation 
        (strlist,ignored) = DAE.build_gr_strlist(items, build_item_str, 10);
        cstr = int_string(count);
        count_1 = count + 1;
        restnode = build_item_listnode_2(ignored, count_1); then Graphviz.LNODE(cstr,strlist,{Graphviz.box},{restnode});
  end matchcontinue;
end build_item_listnode_2;

protected function build_item_str
  input TupleIdentItem in_tupleidentitem;
  output String out_string;
protected 
  type TupleIdentItem = tuple<Ident,Item>;
algorithm 
  out_string:=
  matchcontinue (in_tupleidentitem)
    local
      Ident s,id;
      SCode.Class cls;
      Env env;
    case ((id,VAR(_,_,_)))
      equation 
        s = string_append("VAR: ", id); then s;
    case ((id,CLASS(cls,env)))
      equation 
        s = string_append("CLASS: ", id) "build_env_graphviz env => r &"; then s;
    case ((id,TYPE(_)))
      equation 
        s = string_append("TYPE: ", id); then s;
  end matchcontinue;
end build_item_str;

public function myhash "
relation build_item_graphviz : (Ident, Item) => Graphviz.Node =

  axiom	build_item_graphviz (id, VAR(_)) => Graphviz.LNODE(\"VAR\",{id},{},{})

  axiom	build_item_graphviz (id, CLASS(cls, env)) => Graphviz.LNODE(\"CLASS\",{id},{},{})

  axiom	build_item_graphviz (id, TYPE(_)) => Graphviz.LNODE(\"TYPE\",{id},{},{})

end

   BinTree implementation "
  input Key str;
  output Integer res;
algorithm 
  res := System.hash(str);
end myhash;

protected function tree_new
  output BinTree out_bintree;
algorithm 
  out_bintree:=
  matchcontinue ()
    case () then TREENODE(NONE,NONE,NONE); 
  end matchcontinue;
end tree_new;

public function tree_add
  input BinTree in_bintree;
  input Key in_key;
  input Value in_value;
  input FuncType1 in_functype1;
  output BinTree out_bintree;
protected 
  function FuncType1
    input Key in_key;
    output Integer out_integer;
  end FuncType1;
algorithm 
  out_bintree:=
  matchcontinue (in_bintree,in_key,in_value,in_functype1)
    local
      function FuncType1
        input String in_string;
        output Integer out_integer;
      end FuncType1;
      Ident key,rkey;
      Value value,rval;
      BinTreeOption left,right;
      FuncType1 hashfunc;
      Integer hval,rhval;
      BinTree t_1,t,right_1,left_1;
    case (TREENODE(NONE,NONE,NONE),key,value,_) then TREENODE(SOME(TREEVALUE(key,value)),NONE,NONE); 
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),left,right),key,value,hashfunc)
      equation 
        equality(rkey = key) "Replace this node "; then TREENODE(SOME(TREEVALUE(rkey,value)),left,right);
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),left,(right = SOME(t))),key,value,hashfunc)
      equation 
        hval = hashfunc(key) "Insert to right subtree";
        rhval = hashfunc(rkey);
        (hval > rhval) = true;
        t_1 = tree_add(t, key, value, hashfunc); then TREENODE(SOME(TREEVALUE(rkey,rval)),left,SOME(t_1));
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),left,(right = NONE)),key,value,hashfunc)
      equation 
        hval = hashfunc(key) "Insert to right node";
        rhval = hashfunc(rkey);
        (hval > rhval) = true;
        right_1 = tree_add(TREENODE(NONE,NONE,NONE), key, value, hashfunc); then TREENODE(SOME(TREEVALUE(rkey,rval)),left,SOME(right_1));
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),(left = SOME(t)),right),key,value,hashfunc)
      equation 
        hval = hashfunc(key) "Insert to left subtree";
        rhval = hashfunc(rkey);
        (hval <= rhval) = true;
        t_1 = tree_add(t, key, value, hashfunc); then TREENODE(SOME(TREEVALUE(rkey,rval)),SOME(t_1),right);
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),(left = NONE),right),key,value,hashfunc)
      equation 
        hval = hashfunc(key) "Insert to left node";
        rhval = hashfunc(rkey);
        (hval <= rhval) = true;
        left_1 = tree_add(TREENODE(NONE,NONE,NONE), key, value, hashfunc); then TREENODE(SOME(TREEVALUE(rkey,rval)),SOME(left_1),right);
    case (_,_,_,_)
      equation 
        print("tree_add failed\n"); then fail();
  end matchcontinue;
end tree_add;

public function local_outside_connector_flowvars
  input Env in_env;
  output Exp_ComponentRefList out_exp_componentreflist;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_env)
    local
      type KeyOption = Option<Key>;
      Exp_ComponentRefList res;
      IdentOption sid;
      BinTree ht;
    case (FRAME(sid,ht,_,_,_) :: _)
      equation 
        res = local_outside_connector_flowvars2(SOME(ht)); then res;
  end matchcontinue;
end local_outside_connector_flowvars;

protected function local_outside_connector_flowvars2
  input BinTreeOption in_bintreeoption;
  output Exp_ComponentRefList out_exp_componentreflist;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_bintreeoption)
    local
      type Types_VarList = list<Types.Var>;
      Exp_ComponentRefList lst1,lst2,lst3,res;
      Ident id;
      Types_VarList vars;
      BinTreeOption l,r;
    case (NONE) then {}; 
    case (SOME(TREENODE(SOME(TREEVALUE(_,VAR(Types.VAR(id,_,_,(Types.T_COMPLEX(ClassInf.CONNECTOR(_),vars),_),_),_,_))),l,r)))
      equation 
        lst1 = local_outside_connector_flowvars2(l);
        lst2 = local_outside_connector_flowvars2(r);
        lst3 = Types.flow_variables(vars, Exp.CREF_IDENT(id,{}));
        res = Util.list_flatten({lst1,lst2,lst3}); then res;
    case (SOME(TREENODE(SOME(_),l,r)))
      equation 
        lst1 = local_outside_connector_flowvars2(l);
        lst2 = local_outside_connector_flowvars2(r);
        res = list_append(lst1, lst2); then res;
  end matchcontinue;
end local_outside_connector_flowvars2;

public function local_inside_connector_flowvars
  input Env in_env;
  output Exp_ComponentRefList out_exp_componentreflist;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_env)
    local
      type KeyOption = Option<Key>;
      Exp_ComponentRefList res;
      IdentOption sid;
      BinTree ht;
    case (FRAME(sid,ht,_,_,_) :: _)
      equation 
        res = local_inside_connector_flowvars2(SOME(ht)); then res;
  end matchcontinue;
end local_inside_connector_flowvars;

protected function local_inside_connector_flowvars2
  input BinTreeOption in_bintreeoption;
  output Exp_ComponentRefList out_exp_componentreflist;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_bintreeoption)
    local
      type Types_VarList = list<Types.Var>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp_ComponentRefList lst1,lst2,res,lst3;
      Ident id;
      BinTreeOption l,r;
      Types_VarList vars;
      TupleTypes_TTypeAbsyn_PathOption t;
    case (NONE) then {}; 
    case (SOME(TREENODE(SOME(TREEVALUE(_,VAR(Types.VAR(id,_,_,(Types.T_COMPLEX(ClassInf.CONNECTOR(_),_),_),_),_,_))),l,r)))
      equation 
        lst1 = local_inside_connector_flowvars2(l) "If CONNECTOR then  outside and not inside, skip.. ";
        lst2 = local_inside_connector_flowvars2(r);
        res = list_append(lst1, lst2); then res;
    case (SOME(TREENODE(SOME(TREEVALUE(_,VAR(Types.VAR(id,_,_,(Types.T_COMPLEX(_,vars),_),_),_,_))),l,r)))
      equation 
        lst1 = local_inside_connector_flowvars3(vars, id) "... else retrieve connectors as subcomponents";
        lst2 = local_inside_connector_flowvars2(l);
        lst3 = local_inside_connector_flowvars2(r);
        res = Util.list_flatten({lst1,lst2,lst3}); then res;
    case (SOME(TREENODE(SOME(TREEVALUE(_,VAR(Types.VAR(id,_,_,t,_),_,_))),l,r)))
      equation 
        lst1 = local_inside_connector_flowvars2(l) "if not complex, skip ";
        lst2 = local_inside_connector_flowvars2(r);
        res = list_append(lst1, lst2); then res;
  end matchcontinue;
end local_inside_connector_flowvars2;

protected function local_inside_connector_flowvars3
  input Types_VarList in_types_varlist;
  input Ident in_ident;
  output Exp_ComponentRefList out_exp_componentreflist;
protected 
  type Types_VarList = list<Types.Var>;
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_exp_componentreflist:=
  matchcontinue (in_types_varlist,in_ident)
    local
      Exp_ComponentRefList lst1,lst2,res;
      Ident id,oid;
      Types_VarList vars,xs;
    case ({},_) then {}; 
    case (Types.VAR(id,_,_,(Types.T_COMPLEX(ClassInf.CONNECTOR(_),vars),_),_) :: xs,oid)
      equation 
        lst1 = local_inside_connector_flowvars3(xs, oid);
        lst2 = Types.flow_variables(vars, Exp.CREF_QUAL(oid,{},Exp.CREF_IDENT(id,{})));
        res = list_append(lst1, lst2); then res;
    case (_ :: xs,oid)
      equation 
        res = local_inside_connector_flowvars3(xs, oid); then res;
  end matchcontinue;
end local_inside_connector_flowvars3;

public function tree_get
  input BinTree in_bintree;
  input Key in_key;
  input FuncType1 in_functype1;
  output Value out_value;
protected 
  function FuncType1
    input Key in_key;
    output Integer out_integer;
  end FuncType1;
algorithm 
  out_value:=
  matchcontinue (in_bintree,in_key,in_functype1)
    local
      function FuncType1
        input String in_string;
        output Integer out_integer;
      end FuncType1;
      Ident rkey,key;
      Value rval,res;
      BinTreeOption left,right;
      FuncType1 hashfunc;
      Integer hval,rhval;
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),left,right),key,hashfunc)
      equation 
        equality(rkey = key); then rval;
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),left,SOME(right)),key,hashfunc)
      local BinTree right;
      equation 
        hval = hashfunc(key) "Search to the right";
        rhval = hashfunc(rkey);
        (hval > rhval) = true;
        res = tree_get(right, key, hashfunc); then res;
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),SOME(left),right),key,hashfunc)
      local BinTree left;
      equation 
        hval = hashfunc(key) "Search to the left";
        rhval = hashfunc(rkey);
        (hval <= rhval) = true;
        res = tree_get(left, key, hashfunc); then res;
  end matchcontinue;
end tree_get;

public function print_bintree
  input BinTree in_bintree;
algorithm 
  _:=
  matchcontinue (in_bintree)
    local
      Ident rkey;
      Value rval;
      BinTreeOption l,r;
    case (TREENODE(SOME(TREEVALUE(rkey,rval)),l,r))
      equation 
        print_frame_element((rkey,rval));
        Dump.print_option(l, print_bintree);
        Dump.print_option(r, print_bintree); then ();
    case (TREENODE(NONE,l,r))
      equation 
        Dump.print_option(l, print_bintree);
        Dump.print_option(r, print_bintree); then ();
  end matchcontinue;
end print_bintree;
end Env;

