package Exp "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Exp.rml
  module:      Exp
  description: Expressions
 
  This file contains the module `Exp', which contains data types for
  describing expressions, after they have been examined by the
  static analyzer in the module `StaticExp'.  There are of course
  great similarities with the expression types in the `Absyn'
  module, but there are also several important differences.
 
  No overloading of operators occur, and subscripts have been
  checked to see if they are slices.  All expressions are also type
  consistent, and all implicit type conversions in the AST are made
  explicit here.
 
  Some expression simplification and solving is also done here. This is used
  for symbolic transformations before simulation, in order to rearrange
  equations into a form needed by simulation tools. simplify, solve,
  exp_contains, exp_equal are part of this code.
 
  This module also contains functions for printing expressions, to io or to
  strings. Also graphviz output is supported.
 "

public import OpenModelica.Compiler.Absyn;

public import OpenModelica.Compiler.Graphviz;

public 
type Ident = String " - Identifiers ";

public 
uniontype Type " Define `Ident' as an alias for `string' and use it for all 
   identifiers in Modelica. 
   - Basic types "
  record INT
  end INT;
  record REAL
  end REAL;
  record BOOL
  end BOOL;
  record STRING
  end STRING;
  record ENUM
  end ENUM;
  record OTHER " These types are not used as expression types (see the `Types'
    module for expression types).  They are used to parameterize
    operators which may work on several simple types.   - Expressions "
  end OTHER;
end Type;

public 
type ExpList = list<Exp>;

type TupleExpBoolean = tuple<Exp,Boolean>;

type TupleExpBooleanList = list<TupleExpBoolean>;

type TupleExpBooleanListList = list<TupleExpBooleanList>;

type ExpOption = Option<Exp>;

uniontype Exp " These types are not used as expression types (see the `Types'
    module for expression types).  They are used to parameterize
    operators which may work on several simple types. 
   - Expressions "
  record ICONST
    Integer integer "Integer constants ";
  end ICONST;
  record RCONST
    Real real "Real constants ";
  end RCONST;
  record SCONST
    String string "String constants ";
  end SCONST;
  record BCONST
    Boolean boolean "Bool constants ";
  end BCONST;
  record CREF
    ComponentRef componentref;
    Type type_ "component references, e.g. a.b{2}.c{1} ";
  end CREF;
  record BINARY
    Exp exp1;
    Operator operator2;
    Exp exp3 "Binary operations, e.g. a+4 ";
  end BINARY;
  record UNARY
    Operator operator;
    Exp exp "Unary operations, -(4*x) ";
  end UNARY;
  record LBINARY
    Exp exp1;
    Operator operator2;
    Exp exp3 "Logical binary operations: and, or ";
  end LBINARY;
  record LUNARY
    Operator operator;
    Exp exp "Logical unary operations: not ";
  end LUNARY;
  record RELATION
    Exp exp1;
    Operator operator2;
    Exp exp3 "Relation, e.g. a <= 0 ";
  end RELATION;
  record IFEXP
    Exp exp1;
    Exp exp2;
    Exp exp3 "If expressions ";
  end IFEXP;
  record CALL
    Absyn.Path path;
    ExpList explist;
    Boolean tuple_ "tuple ";
    Boolean boolean "builtin  Function call ";
  end CALL;
  record ARRAY
    Type type_;
    Boolean boolean "scalar for codegen";
    ExpList explist "Array constructor, e.g. {1,3,4} ";
  end ARRAY;
  record MATRIX
    Type type_;
    Integer integer;
    TupleExpBooleanListList tupleexpbooleanlistlist "scalar Matrix constructor. e.g. {1,0;0,1} ";
  end MATRIX;
  record RANGE
    Type type1;
    Exp exp2;
    ExpOption expoption3;
    Exp exp4 "Range constructor, e.g. 1:0.5:10 ";
  end RANGE;
  record TUPLE
    ExpList explist "PR. Tuples, used in func calls returning several 
								  arguments ";
  end TUPLE;
  record CAST
    Type type_;
    Exp exp "Cast operator ";
  end CAST;
  record ASUB
    Exp exp;
    Integer integer "Array subscripts ";
  end ASUB;
  record SIZE
    Exp exp;
    ExpOption expoption "The ssize operator ";
  end SIZE;
  record CODE
    Absyn.Code code;
    Type type_ "Modelica AST constructor ";
  end CODE;
  record REDUCTION
    Absyn.Path path;
    Exp expr "expr";
    Ident ident;
    Exp exp "range Reduction expression ";
  end REDUCTION;
  record END "array index to last element, e.g. a{end}:=1;   The `Exp' datatype closely corresponds to the `Absyn.Exp'
    datatype, but is used for statically analyzed expressions.  It
    includes explicit type promotions and typed (non-overloaded)
    operators. It also contains expression indexing with the `ASUB'
    constructor.  Indexing arbitrary array expressions is currently
    not supported in Modelica, but it is needed here.   LS: Added size for external function interface *"
  end END;
end Exp;

public 
uniontype Operator "array index to last element, e.g. a{end}:=1; 
   The `Exp' datatype closely corresponds to the `Absyn.Exp'
    datatype, but is used for statically analyzed expressions.  It
    includes explicit type promotions and typed (non-overloaded)
    operators. It also contains expression indexing with the `ASUB'
    constructor.  Indexing arbitrary array expressions is currently
    not supported in Modelica, but it is needed here. 
   LS: Added size for external function interface *"
  record ADD
    Type type_;
  end ADD;
  record SUB
    Type type_;
  end SUB;
  record MUL
    Type type_;
  end MUL;
  record DIV
    Type type_;
  end DIV;
  record POW
    Type type_;
  end POW;
  record UMINUS
    Type type_;
  end UMINUS;
  record UPLUS
    Type type_;
  end UPLUS;
  record UMINUS_ARR
    Type type_;
  end UMINUS_ARR;
  record UPLUS_ARR
    Type type_;
  end UPLUS_ARR;
  record ADD_ARR
    Type type_;
  end ADD_ARR;
  record SUB_ARR
    Type type_;
  end SUB_ARR;
  record MUL_SCALAR_ARRAY
    Type type_ "a * { b, c }            ";
  end MUL_SCALAR_ARRAY;
  record MUL_ARRAY_SCALAR
    Type type_ "{a, b} * c              ";
  end MUL_ARRAY_SCALAR;
  record MUL_SCALAR_PRODUCT
    Type type_ "{a, b} * {c, d}         ";
  end MUL_SCALAR_PRODUCT;
  record MUL_MATRIX_PRODUCT
    Type type_ "{{..},..} * {{..},{..}} ";
  end MUL_MATRIX_PRODUCT;
  record DIV_ARRAY_SCALAR
    Type type_ "{a, b} / c              ";
  end DIV_ARRAY_SCALAR;
  record POW_ARR
    Type type_;
  end POW_ARR;
  record AND
  end AND;
  record OR
  end OR;
  record NOT
  end NOT;
  record LESS
    Type type_;
  end LESS;
  record LESSEQ
    Type type_;
  end LESSEQ;
  record GREATER
    Type type_;
  end GREATER;
  record GREATEREQ
    Type type_;
  end GREATEREQ;
  record EQUAL
    Type type_;
  end EQUAL;
  record NEQUAL
    Type type_;
  end NEQUAL;
  record USERDEFINED
    Absyn.Path path "The FQ name of the overloaded operator function   Operators which are overloaded in the abstract syntax are here
    made type-specific.  The integer addition operator (`ADD(INT)')
    and the real addition operator (`ADD(REAL)') are two distinct
    operators.   - Componebnt references ";
  end USERDEFINED;
end Operator;

public 
type SubscriptList = list<Subscript>;

uniontype ComponentRef "The FQ name of the overloaded operator function 
   Operators which are overloaded in the abstract syntax are here
    made type-specific.  The integer addition operator (`ADD(INT)')
    and the real addition operator (`ADD(REAL)') are two distinct
    operators. 
   - Componebnt references "
  record CREF_QUAL
    Ident ident;
    SubscriptList subscriptlist;
    ComponentRef componentref;
  end CREF_QUAL;
  record CREF_IDENT
    Ident ident;
    SubscriptList subscriptlist;
  end CREF_IDENT;
end ComponentRef;

public 
uniontype Subscript
  record WHOLEDIM "a{:,1} "
  end WHOLEDIM;
  record SLICE
    Exp exp "a{1:3,1}, a{1:2:10,2} ";
  end SLICE;
  record INDEX
    Exp exp " The `Subscript' and `ComponentRef' datatypes are simple ";
  end INDEX;
end Subscript;

protected import OpenModelica.Compiler.RTOpts;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.ModUtil;

protected import OpenModelica.Compiler.Derive;

protected import OpenModelica.Compiler.System;

protected import OpenModelica.Compiler.Dump;

protected constant Exp rconstone=RCONST(1.0);

public function cref_to_path " relation: cref_to_path
 
  This function converts a `ComponentRef' to a `Path', if possible.
  If the component reference contains subscripts, it will silently
  fail.
 "
  input ComponentRef in_componentref;
  output Absyn.Path out_path;
algorithm 
  out_path:=
  matchcontinue (in_componentref)
    local
      Ident i;
      Absyn.Path p;
      ComponentRef c;
    case CREF_IDENT(i,{}) then Absyn.IDENT(i); 
    case CREF_QUAL(i,{},c)
      equation 
        p = cref_to_path(c); then Absyn.QUALIFIED(i,p);
  end matchcontinue;
end cref_to_path;

public function path_to_cref " relation: path_to_cref
 
  This function converts a `Absyn.Path' to a `ComponentRef'.
 "
  input Absyn.Path in_path;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_path)
    local
      Ident i;
      ComponentRef c;
      Absyn.Path p;
    case Absyn.IDENT(i) then CREF_IDENT(i,{}); 
    case Absyn.QUALIFIED(i,p)
      equation 
        c = path_to_cref(p); then CREF_QUAL(i,{},c);
  end matchcontinue;
end path_to_cref;

public function cref_str " relation: path_string
 
  This function simply converts a `Path' to a `string'.
 "
  input ComponentRef in_componentref;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_componentref)
    local
      Ident s,ns,s1,ss;
      ComponentRef n;
    case (CREF_IDENT(s,_)) then s; 
    case (CREF_QUAL(s,_,n))
      equation 
        ns = cref_str(n);
        s1 = string_append(s, ".");
        ss = string_append(s1, ns); then ss;
  end matchcontinue;
end cref_str;

public function cref_modelica_str "As above but uses '_' instead of '.' "
  input ComponentRef in_componentref;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_componentref)
    local
      Ident s,ns,s1,ss;
      ComponentRef n;
    case (CREF_IDENT(s,_)) then s; 
    case (CREF_QUAL(s,_,n))
      equation 
        ns = cref_modelica_str(n);
        s1 = string_append(s, "_");
        ss = string_append(s1, ns); then ss;
  end matchcontinue;
end cref_modelica_str;

public function is_range
  input Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    case RANGE(_,_,_,_) then true; 
    case _ then false; 
  end matchcontinue;
end is_range;

public function is_zero " relation: is_zero
 
  Returns true f an expression is constant and has the value zero, 
  otherwise false
 "
  input Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    local
      Integer ival;
      Real rzero,rval;
      Boolean res;
      Type t;
      Exp e;
    case (ICONST(ival))
      equation 
        (ival == 0) = true; then true;
    case (RCONST(rval))
      equation 
        rzero = int_real(0) "Due to bug in rml, go trough a cast from int ";
        (rzero ==. rval) = true; then true;
    case (CAST(t,e))
      equation 
        res = is_zero(e) "Casting to zero is still zero "; then res;
    case (_) then false; 
  end matchcontinue;
end is_zero;

public function is_const " relation: is_const
 
  Returns true f an expression is constant 
  otherwise false
 "
  input Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    local
      Boolean res;
      Operator op;
      Exp e;
      Integer ival;
      Real rval;
      Type t;
    case (UNARY(op,e))
      equation 
        res = is_const(e); then res;
    case (ICONST(ival)) then true; 
    case (RCONST(rval)) then true; 
    case (CAST(t,e))
      equation 
        res = is_const(e) "Casting to zero is still zero "; then res;
    case (_) then false; 
  end matchcontinue;
end is_const;

public function is_relation
  input Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    local
      Boolean b1,b2,res;
      Exp e1,e2;
    case (RELATION(_,_,_)) then true; 
    case (LUNARY(_,RELATION(_,_,_))) then true; 
    case (LBINARY(e1,_,e2))
      equation 
        b1 = is_relation(e1);
        b2 = is_relation(e2);
        res = bool_or(b1, b2); then res;
    case (_) then false; 
  end matchcontinue;
end is_relation;

public function get_relations
  input Exp in_exp;
  output ExpList out_explist;
algorithm 
  out_explist:=
  matchcontinue (in_exp)
    local
      Exp e,e1,e2,cond,tb,fb;
      ExpList rellst1,rellst2,rellst,rellst3,rellst4,xs;
      Type t;
      Boolean sc;
    case ((e = RELATION(_,_,_))) then {e}; 
    case (LBINARY(e1,_,e2))
      equation 
        rellst1 = get_relations(e1);
        rellst2 = get_relations(e2);
        rellst = list_append(rellst1, rellst2); then rellst;
    case (LUNARY(_,e))
      equation 
        rellst = get_relations(e); then rellst;
    case (BINARY(e1,_,e2))
      equation 
        rellst1 = get_relations(e1);
        rellst2 = get_relations(e2);
        rellst = list_append(rellst1, rellst2); then rellst;
    case (IFEXP(cond,tb,fb))
      equation 
        rellst1 = get_relations(cond);
        rellst2 = get_relations(tb);
        rellst3 = get_relations(fb);
        rellst4 = list_append(rellst1, rellst2);
        rellst = list_append(rellst3, rellst4); then rellst;
    case (ARRAY(_,_,{e}))
      equation 
        rellst = get_relations(e); then rellst;
    case (ARRAY(t,sc,e :: xs))
      equation 
        rellst1 = get_relations(ARRAY(t,sc,xs));
        rellst2 = get_relations(e);
        rellst = list_append(rellst1, rellst2); then rellst;
    case (UNARY(_,e))
      equation 
        rellst = get_relations(e); then rellst;
    case (_) then {}; 
  end matchcontinue;
end get_relations;

public function join_crefs " LS:
  Replace this with a call to print_component_ref_str
  and comparison of the resulting strings, after the _str functions are 
  working
 "
  input ComponentRef in_componentref1;
  input ComponentRef in_componentref2;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_componentref1,in_componentref2)
    local
      Ident id;
      SubscriptList sub;
      ComponentRef cr2,cr_1,cr;
    case (CREF_IDENT(id,sub),cr2) then CREF_QUAL(id,sub,cr2); 
    case (CREF_QUAL(id,sub,cr),cr2)
      equation 
        cr_1 = join_crefs(cr, cr2); then CREF_QUAL(id,sub,cr_1);
  end matchcontinue;
end join_crefs;

public function cref_equal
  input ComponentRef in_componentref1;
  input ComponentRef in_componentref2;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_componentref1,in_componentref2)
    local
      Ident n1,n2,s1,s2;
      SubscriptList idx1,idx2;
      ComponentRef cr1,cr2;
    case (CREF_IDENT(n1,idx1),CREF_IDENT(n2,idx2)) " IGNORES subscript list (indexes) for now. Too complex *  Another option is to call print_component_ref_str for both refs and   compare the resulting strings "
      equation 
        equality(n1 = n2);
        true = subscript_equal(idx1, idx2); then true;
    case (CREF_QUAL(n1,idx1,cr1),CREF_QUAL(n2,idx2,cr2))
      equation 
        equality(n1 = n2);
        true = cref_equal(cr1, cr2);
        true = subscript_equal(idx1, idx2); then true;
    case (cr1,cr2)
      equation 
        s1 = print_component_ref_str(cr1) "There is a bug here somewhere or in RML.
	 * Therefore as a last resort, print the strings and compare.
	 *";
        s2 = print_component_ref_str(cr2);
        equality(s1 = s2); then true;
    case (_,_) then false; 
  end matchcontinue;
end cref_equal;

protected function subscript_equal
  input SubscriptList in_subscriptlist1;
  input SubscriptList in_subscriptlist2;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_subscriptlist1,in_subscriptlist2)
    local
      Boolean res;
      Integer i1,i2;
      SubscriptList xs1,xs2;
    case (INDEX(ICONST(i1)) :: xs1,INDEX(ICONST(i2)) :: xs2) "P.A: Just compare constant integer values for now. Might need to lookup variables
	   in environment for evaluating arbitrary expressions in future impl. "
      equation 
        (i1 == i2) = true;
        res = subscript_equal(xs1, xs2); then res;
    case ({},{}) then true; 
    case (_,_) then false; 
  end matchcontinue;
end subscript_equal;

public function prepend_string_cref
  input String in_string;
  input ComponentRef in_componentref;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_string,in_componentref)
    local
      Ident i_1,p,i;
      SubscriptList s;
      ComponentRef c;
    case (p,CREF_QUAL(i,s,c))
      equation 
        i_1 = string_append(p, i); then CREF_QUAL(i_1,s,c);
    case (p,CREF_IDENT(i,s))
      equation 
        i_1 = string_append(p, i); then CREF_IDENT(i_1,s);
  end matchcontinue;
end prepend_string_cref;

public function extend_cref " relation: extend_cref
 
  The `extend_cref' function extends a `ComponentRef' by appending
  an identifier and a (possibly empty) list of subscripts.  Adding
  the identifier `a' to the component reference `x.y{10}' would
  produce the component reference `x.y{10}.a', for instance.
 "
  input ComponentRef in_componentref;
  input Ident in_ident;
  input SubscriptList in_subscriptlist;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_componentref,in_ident,in_subscriptlist)
    local
      Ident i1,i;
      SubscriptList s1,s;
      ComponentRef c_1,c;
    case (CREF_IDENT(i1,s1),i,s) then CREF_QUAL(i1,s1,CREF_IDENT(i,s)); 
    case (CREF_QUAL(i1,s1,c),i,s)
      equation 
        c_1 = extend_cref(c, i, s); then CREF_QUAL(i1,s1,c_1);
  end matchcontinue;
end extend_cref;

public function subscript_cref " relation: subscript_cref
 
  The 'subscript_cref' function adds a subscript to the 'ComponentRef'
  For instance 'a.b' with subscript 10 becomes 'a.b{10} and 'c.d{1,2} with subscript 
  3,4 becomes 'c.d{1,2,3,4}' 
 "
  input ComponentRef in_componentref;
  input SubscriptList in_subscriptlist;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_componentref,in_subscriptlist)
    local
      SubscriptList newsub_1,sub,newsub;
      Ident id;
      ComponentRef cref_1,cref;
    case (CREF_IDENT(id,sub),newsub)
      equation 
        newsub_1 = list_append(sub, newsub); then CREF_IDENT(id,newsub_1);
    case (CREF_QUAL(id,sub,cref),newsub)
      equation 
        cref_1 = subscript_cref(cref, newsub); then CREF_QUAL(id,sub,cref_1);
  end matchcontinue;
end subscript_cref;

public function int_subscripts " - Utility functions
 
  These are utility functions used in some of the other
  relations. 
   relation: int_subscripts
 
  This function describes the function between a list of integers
  and a list of `Exp.Subscript' where each integer is converted to
  an integer indexing expression.
 "
  input IntegerList in_integerlist;
  output SubscriptList out_subscriptlist;
protected 
  type IntegerList = list<Integer>;
algorithm 
  out_subscriptlist:=
  matchcontinue (in_integerlist)
    local
      SubscriptList xs_1;
      Integer x;
      IntegerList xs;
    case {} then {}; 
    case x :: xs
      equation 
        xs_1 = int_subscripts(xs); then INDEX(ICONST(x)) :: xs_1;
  end matchcontinue;
end int_subscripts;

public function simplify " relation: simplify
 
  This function does some very basic simplification on expressions.
  It is not intended to be used to simplify expressions provided by
  the model, but to simplify unnecessarily complex expressions
  constructed during instantiation. 
  PA. Added rules for binary, unary operations
  and multiplication with zero, addition with zero, etc. Useful when deriving 
  equations and then want to simplify
 "
  input Exp in_exp;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp)
    local
      type BooleanList = list<Boolean>;
      Exp e,e1,e2,e3,e_1,e1_1,e2_1,factor,e1_2,e1_3,e2_2,e2_3,exp,c,f,t_1,f_1,exp_1;
      Type tp,tp2,t;
      ExpList expl,expl_1,exps;
      Real v,rv;
      Integer i_1,i,n;
      Boolean b;
      BooleanList bls;
      SubscriptList s,s_1;
      ComponentRef c_1;
      Operator op;
    case (BINARY(e1,ADD(tp),UNARY(UMINUS(tp2),e2))) "a+(-b) and (-b)+a "
      equation 
        e = simplify(BINARY(e1,SUB(tp),e2)); then e;
    case (BINARY(UNARY(UMINUS(tp2),e2),ADD(tp),e1))
      equation 
        e = simplify(BINARY(e1,SUB(tp),e2)); then e;
    case (BINARY(e1,DIV(tp),BINARY(e2,DIV(tp2),e3)))
      equation 
        e = simplify(BINARY(BINARY(e1,MUL(tp),e3),DIV(tp2),e2)) "a/b/c => (a*c)/b) "; then e;
    case (BINARY(BINARY(e1,DIV(tp2),e2),DIV(tp),e3))
      equation 
        e = simplify(BINARY(e1,DIV(tp2),BINARY(e2,MUL(tp),e3))) "(a/b)/c => a/(b*c)) "; then e;
    case (BINARY(e1,DIV(tp),e2))
      equation 
        (expl = _ :: _ :: _) = terms(e1) "(a1+a2+...+an) / b  => a1/b+a2/b+...+an/b ";
        expl_1 = Util.list_map_1(expl, divide, e2);
        e = make_sum(expl_1);
        e_1 = simplify(e); then e_1;
    case (BINARY(e1,DIV(tp),e2))
      equation 
        e1_1 = simplify(e1) "e1 / e2 where e1 and e2 can be factorized with common factor
	 e.g  a*b/c*a => b/c ";
        e2_1 = simplify(e2);
        factor = gcd(e1_1, e2_1);
        e1_2 = remove_factor(factor, e1_1);
        e1_3 = simplify(e1_2);
        e2_2 = remove_factor(factor, e2_1);
        e2_3 = simplify(e2_2); then BINARY(e1_3,DIV(tp),e2_3);
    case (CAST(REAL(),RCONST(v))) then RCONST(v); 
    case (CAST(REAL(),ICONST(v)))
      local Integer v;
      equation 
        rv = int_real(v); then RCONST(rv);
    case ASUB(e,i) "Array and Matrix stuff "
      equation 
        ARRAY(t,b,exps) = simplify(e);
        i_1 = i - 1;
        exp = list_nth(exps, i_1); then exp;
    case ASUB(e,i)
      local
        TupleExpBooleanListList exps;
        TupleExpBooleanList expl;
      equation 
        MATRIX(t,n,exps) = simplify(e);
        i_1 = i - 1;
        (expl) = list_nth(exps, i_1);
        (expl_1,bls) = Util.split_tuple2_list(expl);
        b = Util.bool_and_list(bls); then ARRAY(t,b,expl_1);
    case ASUB(e,i)
      local Exp t;
      equation 
        IFEXP(c,t,f) = simplify(e);
        t_1 = simplify(ASUB(t,i));
        f_1 = simplify(ASUB(f,i)); then IFEXP(c,t_1,f_1);
    case ASUB(e,i)
      local Ident n;
      equation 
        CREF(CREF_IDENT(n,s),t) = simplify(e);
        s_1 = subscripts_append(s, i); then CREF(CREF_IDENT(n,s_1),t);
    case ASUB(e,i)
      local
        Ident n;
        ComponentRef c;
      equation 
        CREF(CREF_QUAL(n,s,c),t) = simplify(e);
        CREF(c_1,t) = simplify(ASUB(CREF(c,t),i)); then CREF(CREF_QUAL(n,s,c_1),t);
    case ASUB(e,i)
      equation 
        e = simplify_asub(e, i) "For arbitrary vector operations, e.g (a+b-c){1} => a{1}+b{1}-c{1} "; then e;
    case ((exp = UNARY(op,e1))) "Operations "
      equation 
        e1_1 = simplify(e1);
        exp_1 = UNARY(op,e1_1);
        e = simplify_unary(exp_1, op, e1_1); then e;
    case (exp)
      equation 
        exp_1 = simplify_add_and_mul(exp); then exp_1;
    case ((exp = BINARY(e1,op,e2)))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        exp_1 = BINARY(e1_1,op,e2_1);
        e = simplify_binary(exp_1, op, e1_1, e2_1); then e;
    case ((exp = RELATION(e1,op,e2)))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        exp_1 = RELATION(e1_1,op,e2_1);
        e = simplify_binary(exp_1, op, e1_1, e2_1); then e;
    case ((exp = LUNARY(op,e1)))
      equation 
        e1_1 = simplify(e1);
        exp_1 = LUNARY(op,e1_1);
        e = simplify_unary(exp_1, op, e1_1); then e;
    case ((exp = LBINARY(e1,op,e2)))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        exp_1 = LBINARY(e1_1,op,e2_1);
        e = simplify_binary(exp_1, op, e1_1, e2_1); then e;
    case e then e; 
  end matchcontinue;
end simplify;

protected function simplify_add_and_mul
  input Exp in_exp;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp)
    local
      Exp e1_1,e2_1,exp_1,exp,e1,e2;
      Type tp;
    case ((exp = BINARY(e1,ADD(tp),e2)))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        exp_1 = simplify_binary_add_coeff(BINARY(e1_1,ADD(tp),e2_1)); then exp_1;
    case ((exp = BINARY(e1,MUL(tp),e2)))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        exp_1 = simplify_binary_mul_coeff(BINARY(e1_1,MUL(tp),e2_1)); then exp_1;
    case (_) then fail(); 
  end matchcontinue;
end simplify_add_and_mul;

protected function simplify_binary_add_coeff
  input Exp in_exp;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp)
    local
      Exp e1_2,e2_2,e1,e2,exp,e3;
      Real coeff1,coeff2,res_coeff;
      Integer res_int;
      Type tp1,tp2;
    case (BINARY(e1,ADD(REAL()),e2)) "  rule  dump_exp_str(exp,0) => e1_str & print \"simplify_binary_add_coeff: \n\" & print e1_str &  print \"\n\" &          int_eq(1,2) => true          ----------------------------  	simplify_binary_add_coeff (exp) => fail "
      equation 
        (e1_2,coeff1) = simplify_binary_add_coeff2(e1);
        (e2_2,coeff2) = simplify_binary_add_coeff2(e2);
        true = exp_equal(e1_2, e2_2);
        res_coeff = coeff1 +. coeff2; then BINARY(RCONST(res_coeff),MUL(REAL()),e1_2);
    case (BINARY(e1,ADD(INT()),e2))
      equation 
        (e1_2,coeff1) = simplify_binary_add_coeff2(e1);
        (e2_2,coeff2) = simplify_binary_add_coeff2(e2);
        true = exp_equal(e1_2, e2_2);
        res_coeff = coeff1 +. coeff2;
        res_int = real_int(res_coeff); then BINARY(ICONST(res_int),MUL(INT()),e1_2);
    case (BINARY(BINARY(e1,ADD(tp1),e2),ADD(tp2),e3))
      equation 
        exp = simplify_two_add_exp(e2, e3, e1, tp1); then exp;
    case (BINARY(BINARY(e1,ADD(tp1),e2),ADD(tp2),e3))
      equation 
        exp = simplify_two_add_exp(e1, e3, e2, tp1); then exp;
    case (BINARY(e3,ADD(tp2),BINARY(e1,ADD(tp1),e2)))
      equation 
        exp = simplify_two_add_exp(e2, e3, e1, tp1); then exp;
    case (BINARY(e3,ADD(tp2),BINARY(e1,ADD(tp1),e2)))
      equation 
        exp = simplify_two_add_exp(e1, e3, e2, tp1); then exp;
    case (_) then fail(); 
  end matchcontinue;
end simplify_binary_add_coeff;

protected function simplify_two_add_exp "
  this function simplify the three exp like this:
  e3 + <coeff> * e1
  used for: x*x+x+x => x * x + x * 2
 "
  input Exp in_exp1;
  input Exp in_exp2;
  input Exp in_exp3;
  input Type in_type4;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_exp2,in_exp3,in_type4)
    local
      Exp e1_1,e2_1,e3_1,e4,e1,e2,e3,e1_2,e2_2;
      Real coeff1,coeff2,res_coeff;
      Integer res_int;
    case (e1,e2,e3,INT())
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        e3_1 = simplify(e3);
        true = is_const(e1_1);
        true = is_const(e2_1);
        e4 = simplify_binary_const(ADD(INT()), e1, e2); then BINARY(e3_1,ADD(INT()),e4);
    case (e1,e2,e3,REAL())
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        e3_1 = simplify(e3);
        true = is_const(e1_1);
        true = is_const(e2_1);
        e4 = simplify_binary_const(ADD(REAL()), e1, e2); then BINARY(e3_1,ADD(REAL()),e4);
    case (e1,e2,e3,INT())
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        e3_1 = simplify(e3);
        true = exp_equal(e1_1, e2_1);
        (e1_2,coeff1) = simplify_binary_add_coeff2(e1_1);
        (e2_2,coeff2) = simplify_binary_add_coeff2(e2_1);
        res_coeff = coeff1 +. coeff2;
        res_int = real_int(res_coeff); then BINARY(e3_1,ADD(INT()),BINARY(e1_2,MUL(INT()),ICONST(res_int)));
    case (e1,e2,e3,REAL())
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        e3_1 = simplify(e3);
        true = exp_equal(e1_1, e2_1);
        (e1_2,coeff1) = simplify_binary_add_coeff2(e1_1);
        (e2_2,coeff2) = simplify_binary_add_coeff2(e2_1);
        res_coeff = coeff1 +. coeff2; then BINARY(e3_1,ADD(REAL()),BINARY(e1_2,MUL(REAL()),RCONST(res_coeff)));
    case (_,_,_,_) then fail(); 
  end matchcontinue;
end simplify_two_add_exp;

protected function simplify_binary_add_coeff2 "
  this relatin checks for x+x+x+x and returns (x,4.0)
 "
  input Exp in_exp;
  output Exp out_exp;
  output Real out_real;
algorithm 
  (out_exp,out_real):=
  matchcontinue (in_exp)
    local
      Exp exp_1,exp,e1,e1_1,e2_1,e1_2,e2_2,e2;
      ComponentRef cr;
      Real coeff,coeff_1,coeff1,coeff2,res_coeff;
      Type tp;
    case (exp) "  rule  dump_exp_str(exp,0) => e1_str & print \"coeff2: \n\" & print e1_str &  print \"\n\" &          int_eq(1,2) => true          ----------------------------  	simplify_binary_add_coeff2 (exp) => fail "
      equation 
        (exp_1 = CREF(cr,_)) = simplify(exp); then (exp_1,1.0);
    case (exp)
      equation 
        BINARY(RCONST(coeff),MUL(_),e1) = simplify(exp); then (e1,coeff);
    case (exp)
      equation 
        BINARY(e1,MUL(_),RCONST(coeff)) = simplify(exp); then (e1,coeff);
    case (exp)
      local Integer coeff;
      equation 
        BINARY(e1,MUL(_),ICONST(coeff)) = simplify(exp);
        coeff_1 = int_real(coeff); then (e1,coeff_1);
    case (exp)
      local Integer coeff;
      equation 
        BINARY(ICONST(coeff),MUL(_),e1) = simplify(exp);
        coeff_1 = int_real(coeff); then (e1,coeff_1);
    case (BINARY(e1,ADD(tp),e2))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        (e1_2,coeff1) = simplify_binary_add_coeff2(e1_1);
        (e2_2,coeff2) = simplify_binary_add_coeff2(e2_1);
        true = exp_equal(e1_2, e2_2);
        res_coeff = coeff1 +. coeff2; then (e1_2,res_coeff);
    case (BINARY(e1,ADD(tp),e2))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        true = exp_equal(e1_1, e2_1); then (e1_1,2.0);
    case (_) then fail(); 
  end matchcontinue;
end simplify_binary_add_coeff2;

protected function simplify_binary_mul_coeff " 
 this function simplifies the expression x*x*x*x to x^4
"
  input Exp in_exp;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp)
    local
      Exp e1_2,e2_2,e1,e2,exp,e3;
      Real coeff1,coeff2,res_coeff;
      Integer res_int;
      Type tp1,tp2;
    case (BINARY(e1,MUL(REAL()),e2)) "  rule  print_exp_str(exp) => e1_str & print \"simplify_binary_mul_coeff: \" & print e1_str &  print \"\n\" &          int_eq(1,2) => true          ----------------------------  	simplify_binary_mul_coeff (exp) => fail "
      equation 
        (e1_2,coeff1) = simplify_binary_mul_coeff2(e1);
        (e2_2,coeff2) = simplify_binary_mul_coeff2(e2);
        true = exp_equal(e1_2, e2_2);
        res_coeff = coeff1 +. coeff2; then BINARY(e1_2,POW(REAL()),RCONST(res_coeff));
    case (BINARY(e1,MUL(INT()),e2))
      equation 
        (e1_2,coeff1) = simplify_binary_mul_coeff2(e1);
        (e2_2,coeff2) = simplify_binary_mul_coeff2(e2);
        true = exp_equal(e1_2, e2_2);
        res_coeff = coeff1 +. coeff2;
        res_int = real_int(res_coeff); then BINARY(e1_2,POW(INT()),ICONST(res_int));
    case (BINARY(BINARY(e1,MUL(tp1),e2),MUL(tp2),e3))
      equation 
        exp = simplify_two_mul_exp(e2, e3, e1, tp1); then exp;
    case (BINARY(BINARY(e1,MUL(tp1),e2),MUL(tp2),e3))
      equation 
        exp = simplify_two_mul_exp(e1, e3, e2, tp1); then exp;
    case (BINARY(e3,MUL(tp2),BINARY(e1,MUL(tp1),e2)))
      equation 
        exp = simplify_two_mul_exp(e2, e3, e1, tp1); then exp;
    case (BINARY(e3,MUL(tp2),BINARY(e1,MUL(tp1),e2)))
      equation 
        exp = simplify_two_mul_exp(e1, e3, e2, tp1); then exp;
    case (_) then fail(); 
  end matchcontinue;
end simplify_binary_mul_coeff;

protected function simplify_two_mul_exp "
  this function simplify the three exp like this:
  e3 + <coeff> * e1
  used for: x*x*x+x => x ^ 2
 "
  input Exp in_exp1;
  input Exp in_exp2;
  input Exp in_exp3;
  input Type in_type4;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_exp2,in_exp3,in_type4)
    local
      Exp e1_1,e2_1,e3_1,e4,e1,e2,e3,e1_2,e2_2;
      Real coeff1,coeff2,res_coeff;
      Integer res_int;
    case (e1,e2,e3,INT())
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        e3_1 = simplify(e3);
        true = is_const(e1_1);
        true = is_const(e2_1);
        e4 = simplify_binary_const(MUL(INT()), e1, e2); then BINARY(e3_1,MUL(INT()),e4);
    case (e1,e2,e3,REAL())
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        e3_1 = simplify(e3);
        true = is_const(e1_1);
        true = is_const(e2_1);
        e4 = simplify_binary_const(MUL(REAL()), e1, e2); then BINARY(e3_1,MUL(REAL()),e4);
    case (e1,e2,e3,INT())
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        e3_1 = simplify(e3);
        (e1_2,coeff1) = simplify_binary_mul_coeff2(e1_1);
        (e2_2,coeff2) = simplify_binary_mul_coeff2(e2_1);
        true = exp_equal(e1_2, e2_2);
        res_coeff = coeff1 +. coeff2;
        res_int = real_int(res_coeff); then BINARY(e3_1,MUL(INT()),BINARY(e1_2,POW(INT()),ICONST(res_int)));
    case (e1,e2,e3,REAL())
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        e3_1 = simplify(e3);
        (e1_2,coeff1) = simplify_binary_mul_coeff2(e1_1);
        (e2_2,coeff2) = simplify_binary_mul_coeff2(e2_1);
        true = exp_equal(e1_2, e2_2);
        res_coeff = coeff1 +. coeff2; then BINARY(e3_1,MUL(REAL()),BINARY(e1_2,POW(REAL()),RCONST(res_coeff)));
    case (_,_,_,_) then fail(); 
  end matchcontinue;
end simplify_two_mul_exp;

protected function simplify_binary_mul_coeff2 "
 this function takes an expression X*X*X*X*X* and return (X,5.0)
 to be used for X^5
"
  input Exp in_exp;
  output Exp out_exp;
  output Real out_real;
algorithm 
  (out_exp,out_real):=
  matchcontinue (in_exp)
    local
      Exp exp_1,exp,e1,e1_1,e2_1,e1_2,e2_2,e2;
      ComponentRef cr;
      Real coeff,coeff_1,coeff1,coeff2,res_coeff;
      Type tp;
    case (exp)
      equation 
        (exp_1 = CREF(cr,_)) = simplify(exp); then (exp_1,1.0);
    case (exp)
      equation 
        BINARY(e1,POW(_),RCONST(coeff)) = simplify(exp); then (e1,coeff);
    case (exp)
      equation 
        BINARY(RCONST(coeff),POW(_),e1) = simplify(exp); then (e1,coeff);
    case (exp)
      local Integer coeff;
      equation 
        BINARY(e1,POW(_),ICONST(coeff)) = simplify(exp);
        coeff_1 = int_real(coeff); then (e1,coeff_1);
    case (exp)
      local Integer coeff;
      equation 
        BINARY(ICONST(coeff),POW(_),e1) = simplify(exp);
        coeff_1 = int_real(coeff); then (e1,coeff_1);
    case (BINARY(e1,MUL(tp),e2))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        (e1_2,coeff1) = simplify_binary_mul_coeff2(e1_1);
        (e2_2,coeff2) = simplify_binary_mul_coeff2(e2_1);
        true = exp_equal(e1_2, e2_2);
        res_coeff = coeff1 +. coeff2; then (e1_2,res_coeff);
    case (BINARY(e1,MUL(tp),e2))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        true = exp_equal(e1_1, e2_1); then (e1_1,2.0);
    case (_) then fail(); 
  end matchcontinue;
end simplify_binary_mul_coeff2;

protected function simplify_asub " relation: simplify_asub
  This function simplifies array subscripts on vector operations
 "
  input Exp in_exp;
  input Integer in_integer;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp,in_integer)
    local
      Exp e_1,e,e1_1,e2_1,e1,e2;
      Type t;
      Integer indx;
      ComponentRef cr;
    case (UNARY(UMINUS_ARR(t),e),indx)
      equation 
        e_1 = simplify_asub(e, indx); then UNARY(UMINUS_ARR(t),e_1);
    case (UNARY(UPLUS_ARR(t),e),indx)
      equation 
        e_1 = simplify_asub(e, indx); then UNARY(UPLUS_ARR(t),e_1);
    case (BINARY(e1,SUB_ARR(t),e2),indx)
      equation 
        e1_1 = simplify_asub(e1, indx);
        e2_1 = simplify_asub(e2, indx); then BINARY(e1_1,SUB_ARR(t),e2_1);
    case (BINARY(e1,MUL_SCALAR_ARRAY(t),e2),indx)
      equation 
        e2_1 = simplify_asub(e2, indx);
        e1_1 = simplify(e1); then BINARY(e1_1,MUL_SCALAR_ARRAY(t),e2_1);
    case (BINARY(e1,MUL_ARRAY_SCALAR(t),e2),indx)
      equation 
        e1_1 = simplify_asub(e1, indx);
        e2_1 = simplify(e2); then BINARY(e1_1,MUL_ARRAY_SCALAR(t),e2_1);
    case (BINARY(e1,DIV_ARRAY_SCALAR(t),e2),indx)
      equation 
        e1_1 = simplify_asub(e1, indx);
        e2_1 = simplify(e2); then BINARY(e1_1,DIV_ARRAY_SCALAR(t),e2_1);
    case ((e = CREF(cr,t)),indx)
      equation 
        e_1 = simplify(ASUB(e,indx)); then e_1;
  end matchcontinue;
end simplify_asub;

protected function divide " relation: divide
  author: PA
 
  divides two expressions.
 "
  input Exp in_exp1;
  input Exp in_exp2;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_exp2)
    local Exp e1,e2;
    case (e1,e2) then BINARY(e1,DIV(REAL()),e2); 
  end matchcontinue;
end divide;

protected function remove_factor "expr "
  input Exp in_exp1;
  input Exp in_exp2;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_exp2)
    local
      type IdentList = list<Ident>;
      Exp e1,factor,expr,exp;
      ExpList rest,e2s,e1s,factors_1;
      Ident fs,es,factorsstr;
      IdentList elst;
    case (factor,expr) "factor  expr  updated expr "
      equation 
        e1 :: rest = factors(factor);
        e2s = factors(expr);
        {} = Util.list_setdifference_p(e2s, e1 :: rest, exp_equal) "factor = expr "; then e1;
    case (factor,expr)
      equation 
        e1s = factors(factor);
        e2s = factors(expr);
        factors_1 = Util.list_setdifference_p(e2s, e1s, exp_equal);
        exp = make_product(factors_1); then exp;
    case (factor,expr)
      equation 
        fs = print_exp_str(factor);
        es = print_exp_str(expr);
        print("remove_factor failed, factor:");
        print(fs);
        print(" expr:");
        print(es);
        print("\n");
        e2s = factors(expr);
        elst = Util.list_map(e2s, print_exp_str);
        factorsstr = Util.string_delimit_list(elst, ", ");
        print(" factors:");
        print(factorsstr);
        print("\n"); then fail();
  end matchcontinue;
end remove_factor;

protected function gcd " relation: gcd
  Return the greatest common divisor expression from two expressions.
  If no common divisor besides a numerical expression can be found, the function fails.
 "
  input Exp e1;
  input Exp e2;
  output Exp product;
protected 
  ExpList e1s,e2s,factor;
algorithm 
  e1s := factors(e1);
  e2s := factors(e2);
  (factor = _ :: _) := Util.list_intersection_p(e1s, e2s, exp_equal);
  product := make_product(factor);
end gcd;

protected function no_factors " relation no_factors
  Helper function to factors.
  If a factor list is empty, the expression has no subfactors.
  But the complete expression is then a factor for larger expressions,
  returned by this function.
 "
  input ExpList in_explist;
  input Exp in_exp;
  output ExpList out_explist;
algorithm 
  out_explist:=
  matchcontinue (in_explist,in_exp)
    local
      Exp e;
      ExpList lst;
    case ({},e) then {e}; 
    case (lst,_) then lst; 
  end matchcontinue;
end no_factors;

protected function negate " relation: negate
  author: PA
 
  Negates an expression.
 "
  input Exp e;
  output Exp out_exp;
protected 
  Type t;
algorithm 
  t := typeof(e);
  out_exp := UNARY(UMINUS(t),e);
end negate;

public function terms "  relation: terms
  author: PA
 
  Returns the terms of the expression if any as a list of expressiosn
 "
  input Exp in_exp;
  output ExpList out_explist;
algorithm 
  out_explist:=
  matchcontinue (in_exp)
    local
      ExpList f1,f2,res,f2_1;
      Exp e1,e2,e;
      ComponentRef cr;
    case (BINARY(e1,ADD(_),e2))
      equation 
        f1 = terms(e1);
        f2 = terms(e2);
        res = list_append(f1, f2); then res;
    case (BINARY(e1,SUB(_),e2))
      equation 
        f1 = terms(e1);
        f2 = terms(e2);
        f2_1 = Util.list_map(f2, negate);
        res = list_append(f1, f2_1); then res;
    case ((e = BINARY(_,MUL(_),_))) then {e}; 
    case ((e = BINARY(_,DIV(_),_))) then {e}; 
    case ((e = CREF(cr,_))) then {e}; 
    case ((e = ICONST(_))) then {e}; 
    case ((e = RCONST(_))) then {e}; 
    case ((e = SCONST(_))) then {e}; 
    case ((e = UNARY(_,_))) then {e}; 
    case ((e = IFEXP(_,_,_))) then {e}; 
    case ((e = CALL(_,_,_,_))) then {e}; 
    case ((e = ARRAY(_,_,_))) then {e}; 
    case ((e = MATRIX(_,_,_))) then {e}; 
    case ((e = RANGE(_,_,_,_))) then {e}; 
    case ((e = CAST(_,_))) then {e}; 
    case ((e = ASUB(_,_))) then {e}; 
    case ((e = SIZE(_,_))) then {e}; 
    case ((e = REDUCTION(_,_,_,_))) then {e}; 
    case (_) then {}; 
  end matchcontinue;
end terms;

public function factors " relation: factors
  Returns the factors of the expression if any as a list of expressions
 "
  input Exp in_exp;
  output ExpList out_explist;
algorithm 
  out_explist:=
  matchcontinue (in_exp)
    local
      ExpList f1,f2,f1_1,f2_1,res,f2_2;
      Exp e1,e2,e;
      ComponentRef cr;
    case (BINARY(e1,MUL(_),e2))
      equation 
        f1 = factors(e1) "Both subexpression has factors";
        f2 = factors(e2);
        f1_1 = no_factors(f1, e1);
        f2_1 = no_factors(f2, e2);
        res = list_append(f1_1, f2_1); then res;
    case (BINARY(e1,DIV(REAL()),e2))
      equation 
        f1 = factors(e1);
        f2 = factors(e2);
        f1_1 = no_factors(f1, e1);
        f2_1 = no_factors(f2, e2);
        f2_2 = inverse_factors(f2_1);
        res = list_append(f1_1, f2_2); then res;
    case ((e = CREF(cr,_))) then {e}; 
    case ((e = ICONST(_))) then {e}; 
    case ((e = RCONST(_))) then {e}; 
    case ((e = SCONST(_))) then {e}; 
    case ((e = UNARY(_,_))) then {e}; 
    case ((e = IFEXP(_,_,_))) then {e}; 
    case ((e = CALL(_,_,_,_))) then {e}; 
    case ((e = ARRAY(_,_,_))) then {e}; 
    case ((e = MATRIX(_,_,_))) then {e}; 
    case ((e = RANGE(_,_,_,_))) then {e}; 
    case ((e = CAST(_,_))) then {e}; 
    case ((e = ASUB(_,_))) then {e}; 
    case ((e = SIZE(_,_))) then {e}; 
    case ((e = REDUCTION(_,_,_,_))) then {e}; 
    case (_) then {}; 
  end matchcontinue;
end factors;

protected function inverse_factors " relation inverse_factors
  Takes a list of expressions and returns each expression in the list inversed.
  For example inverse_factors {a,3+b} => {1/a, 1/3+b}
 "
  input ExpList in_explist;
  output ExpList out_explist;
algorithm 
  out_explist:=
  matchcontinue (in_explist)
    local
      ExpList es_1,es;
      Exp e;
    case ({}) then {}; 
    case (e :: es)
      equation 
        REAL() = typeof(e);
        es_1 = inverse_factors(es); then BINARY(RCONST(1.0),DIV(REAL()),e) :: es_1;
    case (e :: es)
      equation 
        INT() = typeof(e);
        es_1 = inverse_factors(es); then BINARY(ICONST(1),DIV(INT()),e) :: es_1;
  end matchcontinue;
end inverse_factors;

protected function make_product " relation: make_product
  Takes a list of expressions an makes a product expression multiplying all 
  elements in the list.
 "
  input ExpList in_explist;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_explist)
    local
      type IdentList = list<Ident>;
      Exp e1,e2;
      Type tp;
      ExpList rest,lst;
      IdentList explst;
      Ident str;
    case ({e1}) then e1; 
    case ({e1,e2})
      equation 
        tp = typeof(e1) "Take type info from e1, ok since type checking already performed. "; then BINARY(e1,MUL(tp),e2);
    case (e1 :: rest)
      equation 
        e2 = make_product(rest);
        tp = typeof(e2); then BINARY(e1,MUL(tp),e2);
    case (lst)
      equation 
        print("-make_product failed, exp lst:");
        explst = Util.list_map(lst, print_exp_str);
        str = Util.string_delimit_list(explst, ", ");
        print(str);
        print("\n"); then fail();
  end matchcontinue;
end make_product;

public function make_sum " relation: make_sum
  Takes a list of expressions an makes a sum expression adding all 
  elements in the list.
 "
  input ExpList in_explist;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_explist)
    local
      type IdentList = list<Ident>;
      Exp e1,e2;
      Type tp;
      ExpList rest,lst;
      IdentList explst;
      Ident str;
    case ({e1}) then e1; 
    case ({e1,e2})
      equation 
        tp = typeof(e1) "Take type info from e1, ok since type checking already performed. "; then BINARY(e1,ADD(tp),e2);
    case (e1 :: rest)
      equation 
        e2 = make_sum(rest);
        tp = typeof(e2); then BINARY(e1,ADD(tp),e2);
    case (lst)
      equation 
        print("-make_sum failed, exp lst:");
        explst = Util.list_map(lst, print_exp_str);
        str = Util.string_delimit_list(explst, ", ");
        print(str);
        print("\n"); then fail();
  end matchcontinue;
end make_sum;

public function type_builtin " relation: type_builtin
 
  Returns true if type is one of the builtin types.
 "
  input Type in_type_;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_type)
    case (INT()) then true; 
    case (REAL()) then true; 
    case (STRING()) then true; 
    case (BOOL()) then true; 
    case (_) then false; 
  end matchcontinue;
end type_builtin;

public function typeof " relation typeof
  Retrieves the Type of the Expression
 "
  input Exp in_exp;
  output Type out_type_;
algorithm 
  out_type:=
  matchcontinue (in_exp)
    local
      Type tp;
      Operator op;
      Exp e1,e2,e3,e;
    case (ICONST(_)) then INT(); 
    case (RCONST(_)) then REAL(); 
    case (SCONST(_)) then STRING(); 
    case (BCONST(_)) then BOOL(); 
    case (CREF(_,tp)) then tp; 
    case (BINARY(_,op,_))
      equation 
        tp = typeof_op(op); then tp;
    case (UNARY(op,_))
      equation 
        tp = typeof_op(op); then tp;
    case (LBINARY(_,op,_))
      equation 
        tp = typeof_op(op); then tp;
    case (LUNARY(op,_))
      equation 
        tp = typeof_op(op); then tp;
    case (RELATION(_,op,_))
      equation 
        tp = typeof_op(op); then tp;
    case (IFEXP(e1,e2,e3))
      equation 
        tp = typeof(e2); then tp;
    case (CALL(_,_,_,_)) then OTHER();  "Not always true "
    case (ARRAY(tp,_,_)) then tp; 
    case (MATRIX(tp,_,_)) then tp; 
    case (RANGE(tp,_,_,_)) then tp; 
    case (CAST(tp,_)) then tp; 
    case (ASUB(e,_))
      equation 
        tp = typeof(e); then tp;
    case (CODE(_,tp)) then tp; 
    case (REDUCTION(_,e,_,_))
      equation 
        tp = typeof(e); then tp;
    case (END()) then OTHER();  "Can be any type. "
  end matchcontinue;
end typeof;

protected function typeof_op
  input Operator in_operator;
  output Type out_type_;
algorithm 
  out_type:=
  matchcontinue (in_operator)
    local Type t;
    case (ADD(t)) then t; 
    case (SUB(t)) then t; 
    case (MUL(t)) then t; 
    case (DIV(t)) then t; 
    case (POW(t)) then t; 
    case (UMINUS(t)) then t; 
    case (UPLUS(t)) then t; 
    case (UMINUS_ARR(t)) then t; 
    case (UPLUS_ARR(t)) then t; 
    case (ADD_ARR(t)) then t; 
    case (SUB_ARR(t)) then t; 
    case (MUL_SCALAR_ARRAY(t)) then t; 
    case (MUL_SCALAR_PRODUCT(t)) then t; 
    case (MUL_MATRIX_PRODUCT(t)) then t; 
    case (DIV_ARRAY_SCALAR(t)) then t; 
    case (POW_ARR(t)) then t; 
    case (AND()) then BOOL(); 
    case (OR()) then BOOL(); 
    case (NOT()) then BOOL(); 
    case (LESS(t)) then t; 
    case (LESSEQ(t)) then t; 
    case (GREATER(t)) then t; 
    case (GREATEREQ(t)) then t; 
    case (EQUAL(t)) then t; 
    case (NEQUAL(t)) then t; 
    case (USERDEFINED(t))
      local Absyn.Path t; then OTHER();
  end matchcontinue;
end typeof_op;

protected function is_const_one
  input Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    local
      Real rval;
      Exp e;
    case e
      equation 
        rval = int_real(1);
        equality(e = RCONST(rval)); then true;
    case ICONST(1) then true; 
    case (_) then false; 
  end matchcontinue;
end is_const_one;

protected function is_const_minus_one
  input Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    local Real rval,v;
    case RCONST(v)
      equation 
        rval = int_real(-1);
        (v ==. rval) = true; then true;
    case ICONST(-1) then true; 
    case (_) then false; 
  end matchcontinue;
end is_const_minus_one;

protected function is_const_zero
  input Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    local
      Real rval;
      Exp e;
    case e
      equation 
        rval = int_real(0);
        equality(e = RCONST(rval)); then true;
    case ICONST(0) then true; 
    case (_) then false; 
  end matchcontinue;
end is_const_zero;

protected function simplify_binary_const
  input Operator in_operator1;
  input Exp in_exp2;
  input Exp in_exp3;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_operator1,in_exp2,in_exp3)
    local
      Integer e3,e1,e2;
      Real e2_1,e1_1;
      Operator op;
    case (ADD(_),ICONST(e1),ICONST(e2)) "  rule  print_exp_str(e1) => e1_str &          print_exp_str(e2) => e2_str &          binop_symbol_1 oper => op_str &          print \"simplify_binary_const: \" & print e1_str & print \" \" & print op_str & print \" \" & print e2_str & print \"\n\" &          int_eq(1,2) => true          ----------------------------  	simplify_binary_const (oper, e1, e2) => fail "
      equation 
        e3 = e1 + e2; then ICONST(e3);
    case (ADD(_),RCONST(e1),RCONST(e2))
      local Real e3,e1,e2;
      equation 
        e3 = e1 +. e2; then RCONST(e3);
    case (ADD(_),RCONST(e1),ICONST(e2))
      local Real e3,e1;
      equation 
        e2_1 = int_real(e2);
        e3 = e1 +. e2_1; then RCONST(e3);
    case (ADD(_),ICONST(e1),RCONST(e2))
      local Real e3,e2;
      equation 
        e1_1 = int_real(e1);
        e3 = e1_1 +. e2; then RCONST(e3);
    case (SUB(_),ICONST(e1),ICONST(e2))
      equation 
        e3 = e1 - e2; then ICONST(e3);
    case (SUB(_),RCONST(e1),RCONST(e2))
      local Real e3,e1,e2;
      equation 
        e3 = e1 -. e2; then RCONST(e3);
    case (SUB(_),RCONST(e1),ICONST(e2))
      local Real e3,e1;
      equation 
        e2_1 = int_real(e2);
        e3 = e1 -. e2_1; then RCONST(e3);
    case (SUB(_),ICONST(e1),RCONST(e2))
      local Real e3,e2;
      equation 
        e1_1 = int_real(e1);
        e3 = e1_1 -. e2; then RCONST(e3);
    case (MUL(_),ICONST(e1),ICONST(e2))
      equation 
        e3 = e1*e2; then ICONST(e3);
    case (MUL(_),RCONST(e1),RCONST(e2))
      local Real e3,e1,e2;
      equation 
        e3 = e1*.e2; then RCONST(e3);
    case (MUL(_),RCONST(e1),ICONST(e2))
      local Real e3,e1;
      equation 
        e2_1 = int_real(e2);
        e3 = e1*.e2_1; then RCONST(e3);
    case (MUL(_),ICONST(e1),RCONST(e2))
      local Real e3,e2;
      equation 
        e1_1 = int_real(e1);
        e3 = e1_1*.e2; then RCONST(e3);
    case (DIV(_),ICONST(e1),ICONST(e2))
      equation 
        e3 = e1/e2; then ICONST(e3);
    case (DIV(_),RCONST(e1),RCONST(e2))
      local Real e3,e1,e2;
      equation 
        e3 = e1/.e2; then RCONST(e3);
    case (DIV(_),RCONST(e1),ICONST(e2))
      local Real e3,e1;
      equation 
        e2_1 = int_real(e2);
        e3 = e1/.e2_1; then RCONST(e3);
    case (DIV(_),ICONST(e1),RCONST(e2))
      local Real e3,e2;
      equation 
        e1_1 = int_real(e1);
        e3 = e1_1/.e2; then RCONST(e3);
    case (op,e1,e2)
      local Exp e1,e2; then fail();
  end matchcontinue;
end simplify_binary_const;

protected function simplify_binary
  input Exp in_exp1;
  input Operator in_operator2;
  input Exp in_exp3;
  input Exp in_exp4;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_operator2,in_exp3,in_exp4)
    local
      Exp e1_1,e2_1,e3,e,e1,e2,e_1,res;
      Operator oper;
      Type ty,ty2,ty1,tp2,tp;
    case (e,oper,e1,e2) "  rule  dump_exp_str(e1,0) => e1_str &          dump_exp_str(e2,0) => e2_str &          binop_symbol_1 oper => op_str &          print \"simplify_binary: \" & print e1_str & print \" \" & print op_str & print \" \" & print e2_str & print \"\n\" &          int_eq(1,2) => true          ----------------------------  	simplify_binary (e, oper, e1, e2) => fail "
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2);
        true = is_const(e1_1);
        true = is_const(e2_1);
        e3 = simplify_binary_const(oper, e1_1, e2_1); then e3;
    case (_,ADD(ty),e1,e2)
      equation 
        true = is_const_zero(e1);
        e2_1 = simplify(e2); then e2_1;
    case (_,ADD(ty),e1,e2)
      equation 
        true = is_const_zero(e2);
        e1_1 = simplify(e1); then e1_1;
    case (_,SUB(ty),e1,e2)
      equation 
        true = is_const_zero(e1);
        e = UNARY(UMINUS(ty),e2);
        e_1 = simplify(e); then e_1;
    case (_,SUB(ty),e1,e2)
      equation 
        true = is_const_zero(e2);
        e1_1 = simplify(e1); then e1_1;
    case (_,SUB(ty),e1,e2)
      equation 
        true = is_const_zero(e2);
        e1_1 = simplify(e1); then e1_1;
    case (_,SUB(ty),e1,UNARY(UMINUS(ty2),e2))
      equation 
        e = simplify(BINARY(e1,ADD(ty),e2)) "a-(-b) = a+b "; then e;
    case (_,MUL(ty),e1,e2)
      equation 
        true = is_const_zero(e1); then e1;
    case (_,MUL(ty),e1,e2)
      equation 
        true = is_const_zero(e2); then e2;
    case (_,MUL(ty),e1,e2)
      equation 
        true = is_const_one(e1);
        e2_1 = simplify(e2); then e2_1;
    case (_,MUL(ty),e1,e2)
      equation 
        true = is_const_one(e2);
        e1_1 = simplify(e1); then e1_1;
    case (_,MUL(ty),e1,e2)
      equation 
        true = is_const_minus_one(e1);
        e = simplify(UNARY(UMINUS(ty),e2)); then e;
    case (_,MUL(ty),e1,e2)
      equation 
        true = is_const_minus_one(e2);
        e1_1 = simplify(e1); then UNARY(UMINUS(ty),e1_1);
    case (_,MUL(ty),UNARY(UMINUS(ty1),e1),UNARY(UMINUS(ty2),e2))
      equation 
        e = simplify(BINARY(e1,MUL(ty),e2)); then e;
    case (_,MUL(ty),e1,UNARY(UMINUS(ty2),e2))
      equation 
        e1_1 = simplify(UNARY(UMINUS(ty),e1)) "e1 * -e2 => -e1 * e2 ";
        e2_1 = simplify(e2); then BINARY(e1_1,MUL(ty),e2_1);
    case (_,DIV(ty),e1,e2)
      equation 
        true = is_const_zero(e1); then RCONST(0.0);
    case (_,DIV(ty),e1,e2)
      equation 
        true = is_const_zero(e2);
        Print.print_error_buf("# Error, division by zero.\n"); then fail();
    case (_,DIV(ty),e1,e2)
      equation 
        true = is_const_one(e2);
        e1_1 = simplify(e1); then e1_1;
    case (_,DIV(ty),e1,e2)
      equation 
        true = is_const_minus_one(e2);
        e1_1 = simplify(e1); then UNARY(UMINUS(ty),e1_1);
    case (_,DIV(ty),UNARY(UMINUS(ty1),e1),UNARY(UMINUS(ty2),e2))
      equation 
        e1_1 = simplify(e1);
        e2_1 = simplify(e2); then BINARY(e1_1,DIV(ty),e2_1);
    case (_,DIV(ty),e1,UNARY(UMINUS(ty2),e2))
      equation 
        e1_1 = simplify(UNARY(UMINUS(ty),e1)) "e1 / -e2  => -e1 / e2 ";
        e2_1 = simplify(e2); then BINARY(e1_1,DIV(ty),e2_1);
    case (_,DIV(tp2),BINARY(e2,MUL(tp),e3),e1)
      equation 
        true = is_const(e3) "(c1*x)/c2 ";
        true = is_const(e1);
        e = simplify(BINARY(BINARY(e1,DIV(tp2),e3),MUL(tp),e2)); then e;
    case (_,DIV(tp2),BINARY(e2,MUL(tp),e3),e1)
      equation 
        true = is_const(e3) "(x*c1)/c2 ";
        true = is_const(e2);
        print("is const2 \n");
        e = simplify(BINARY(BINARY(e2,DIV(tp2),e3),MUL(tp),e1)); then e;
    case (_,POW(_),CREF(e1,tp),e)
      local ComponentRef e1;
      equation 
        e_1 = simplify(e) "e1^e2, where e2 is one ";
        true = is_const_one(e_1); then CREF(e1,tp);
    case (_,POW(_),CREF(e1,tp),e)
      local ComponentRef e1;
      equation 
        e_1 = simplify(e) "e1^e2, where e2 is zero ";
        true = is_zero(e_1);
        res = create_const_one(tp); then res;
    case (e,_,_,_) then e; 
  end matchcontinue;
end simplify_binary;

protected function create_const_one " relation: create_const_one
  Creates a constant value one, given a type INT or REAL
 "
  input Type in_type_;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_type)
    local Real realv;
    case (REAL())
      equation 
        realv = int_real(1); then RCONST(realv);
    case (INT()) then ICONST(1); 
  end matchcontinue;
end create_const_one;

protected function simplify_unary
  input Exp in_exp1;
  input Operator in_operator2;
  input Exp in_exp3;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_operator2,in_exp3)
    local
      Type ty,ty1,ty2;
      Exp e1,e1_1,e_1,e2,e;
    case (_,UPLUS(ty),e1) then e1; 
    case (_,UMINUS(ty),e1)
      equation 
        e1_1 = simplify(e1);
        true = is_const_zero(e1_1); then e1_1;
    case (_,UMINUS(ty),BINARY(e1,SUB(ty1),e2))
      equation 
        e_1 = simplify(BINARY(e2,SUB(ty1),e1)) "-(a-b) => b-a "; then e_1;
    case (_,UMINUS(ty),BINARY(e1,ADD(ty1),e2))
      equation 
        e_1 = simplify(BINARY(UNARY(UMINUS(ty),e1),ADD(ty1),UNARY(UMINUS(ty),e2))) "-(a+b) => b-a "; then e_1;
    case (_,UMINUS(ty),BINARY(e1,DIV(ty1),e2))
      equation 
        e_1 = simplify(BINARY(UNARY(UMINUS(ty),e1),DIV(ty1),e2)) "-(a/b) => -a/b "; then e_1;
    case (_,UMINUS(ty),BINARY(e1,MUL(ty1),e2))
      equation 
        e_1 = simplify(BINARY(UNARY(UMINUS(ty),e1),MUL(ty1),e2)) "-(a*b) => -a*b "; then e_1;
    case (_,UMINUS(ty1),UNARY(UMINUS(ty2),e1)) "Should we check ty1 == ty2? "
      equation 
        e1_1 = simplify(e1); then e1_1;
    case (e,_,_) then e; 
  end matchcontinue;
end simplify_unary;

public function contain_functioncall " relation: contain_functioncall
  Returns true if expression or subexpression is a functioncall.
  otherwise false 
"
  input Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    local
      type BooleanList = list<Boolean>;
      Exp e1,e2,e,e3;
      Boolean res;
      BooleanList blst;
      ExpList elst;
      TupleExpBooleanList flatexplst;
      TupleExpBooleanListList explst;
      ExpOption optexp;
    case (CALL(_,_,_,_)) then true; 
    case (BINARY(e1,_,e2)) "Binary "
      equation 
        true = contain_functioncall(e1); then true;
    case (BINARY(e1,_,e2))
      equation 
        true = contain_functioncall(e2); then true;
    case (UNARY(_,e)) "Unary "
      equation 
        res = contain_functioncall(e); then res;
    case (LBINARY(e1,_,e2)) "LBinary "
      equation 
        true = contain_functioncall(e1); then true;
    case (LBINARY(e1,_,e2))
      equation 
        true = contain_functioncall(e2); then true;
    case (LUNARY(_,e)) "LUnary "
      equation 
        res = contain_functioncall(e); then res;
    case (RELATION(e1,_,e2)) "Relation "
      equation 
        true = contain_functioncall(e1); then true;
    case (RELATION(e1,_,e2))
      equation 
        true = contain_functioncall(e2); then true;
    case (IFEXP(e1,e2,e3)) "If exp"
      equation 
        true = contain_functioncall(e1); then true;
    case (IFEXP(e1,e2,e3))
      equation 
        true = contain_functioncall(e2); then true;
    case (IFEXP(e1,e2,e3))
      equation 
        true = contain_functioncall(e3); then true;
    case (ARRAY(_,_,elst)) "Array "
      equation 
        blst = Util.list_map(elst, contain_functioncall);
        res = Util.bool_or_list(blst); then res;
    case (MATRIX(_,_,explst)) "Matrix "
      equation 
        flatexplst = Util.list_flatten(explst);
        elst = Util.list_map(flatexplst, Util.tuple2_1);
        blst = Util.list_map(elst, contain_functioncall);
        res = Util.bool_or_list(blst); then res;
    case (RANGE(_,e1,optexp,e2)) "Range "
      equation 
        true = contain_functioncall(e1); then true;
    case (RANGE(_,e1,optexp,e2))
      equation 
        true = contain_functioncall(e2); then true;
    case (RANGE(_,e1,SOME(e),e2))
      equation 
        true = contain_functioncall(e); then true;
    case (TUPLE(_)) then true; 
    case (CAST(_,e))
      equation 
        res = contain_functioncall(e); then res;
    case (SIZE(e1,e2)) "Size "
      local ExpOption e2;
      equation 
        true = contain_functioncall(e1); then true;
    case (SIZE(e1,SOME(e2)))
      equation 
        true = contain_functioncall(e2); then true;
    case (_) then false; 
  end matchcontinue;
end contain_functioncall;

public function unelab_exp
  input Exp in_exp;
  output Absyn.Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp)
    local
      Integer i;
      Real r;
      Ident s;
      Boolean b;
      Absyn.ComponentRef cr_1;
      ComponentRef cr;
      Type t;
    case (ICONST(i)) then Absyn.INTEGER(i); 
    case (RCONST(r)) then Absyn.REAL(r); 
    case (SCONST(s)) then Absyn.STRING(s); 
    case (BCONST(b)) then Absyn.BOOL(b); 
    case (CREF(cr,t))
      equation 
        cr_1 = unelab_cref(cr); then Absyn.CREF(cr_1);
  end matchcontinue;
end unelab_exp;

public function unelab_cref
  input ComponentRef in_componentref;
  output Absyn.ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_componentref)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      Absyn_SubscriptList subs_1;
      Ident id;
      SubscriptList subs;
      Absyn.ComponentRef cr_1;
      ComponentRef cr;
    case (CREF_IDENT(id,subs))
      equation 
        subs_1 = unelab_subscripts(subs); then Absyn.CREF_IDENT(id,subs_1);
    case (CREF_QUAL(id,subs,cr))
      equation 
        cr_1 = unelab_cref(cr);
        subs_1 = unelab_subscripts(subs); then Absyn.CREF_QUAL(id,subs_1,cr_1);
  end matchcontinue;
end unelab_cref;

protected function unelab_subscripts
  input SubscriptList in_subscriptlist;
  output Absyn_SubscriptList out_absyn_subscriptlist;
protected 
  type Absyn_SubscriptList = list<Absyn.Subscript>;
algorithm 
  out_absyn_subscriptlist:=
  matchcontinue (in_subscriptlist)
    local
      Absyn_SubscriptList xs_1;
      SubscriptList xs;
      Absyn.Exp e_1;
      Exp e;
    case ({}) then {}; 
    case (WHOLEDIM() :: xs)
      equation 
        xs_1 = unelab_subscripts(xs); then Absyn.NOSUB() :: xs_1;
    case (SLICE(e) :: xs)
      equation 
        xs_1 = unelab_subscripts(xs);
        e_1 = unelab_exp(e); then Absyn.SUBSCRIPT(e_1) :: xs_1;
    case (INDEX(e) :: xs)
      equation 
        xs_1 = unelab_subscripts(xs);
        e_1 = unelab_exp(e); then Absyn.SUBSCRIPT(e_1) :: xs_1;
  end matchcontinue;
end unelab_subscripts;

public function to_exp_cref "Only support for indexed subscripts of integers"
  input Absyn.ComponentRef in_componentref;
  output ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_componentref)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      SubscriptList subs_1;
      Ident id;
      Absyn_SubscriptList subs;
      ComponentRef cr_1;
      Absyn.ComponentRef cr;
    case (Absyn.CREF_IDENT(id,subs))
      equation 
        subs_1 = to_exp_cref_subs(subs); then CREF_IDENT(id,subs_1);
    case (Absyn.CREF_QUAL(id,subs,cr))
      equation 
        cr_1 = to_exp_cref(cr);
        subs_1 = to_exp_cref_subs(subs); then CREF_QUAL(id,subs_1,cr_1);
  end matchcontinue;
end to_exp_cref;

protected function to_exp_cref_subs
  input Absyn_SubscriptList in_absyn_subscriptlist;
  output SubscriptList out_subscriptlist;
protected 
  type Absyn_SubscriptList = list<Absyn.Subscript>;
algorithm 
  out_subscriptlist:=
  matchcontinue (in_absyn_subscriptlist)
    local
      SubscriptList xs_1;
      Integer i;
      Absyn_SubscriptList xs;
    case ({}) then {}; 
    case (Absyn.SUBSCRIPT(Absyn.INTEGER(i)) :: xs)
      equation 
        xs_1 = to_exp_cref_subs(xs); then INDEX(ICONST(i)) :: xs_1;
    case (_ :: xs)
      equation 
        Print.print_error_buf("#Error elaborating TypeName.");
        xs_1 = to_exp_cref_subs(xs); then xs_1;
  end matchcontinue;
end to_exp_cref_subs;

protected function subscripts_append " relation: subscripts_append
 
  This function takes a subscript list and adds a new subscript.
  But there are a few special cases.  When the last existing
  subscript is a slice, it is replaced by the slice indexed by the
  new subscript.
 "
  input SubscriptList in_subscriptlist;
  input Integer in_integer;
  output SubscriptList out_subscriptlist;
algorithm 
  out_subscriptlist:=
  matchcontinue (in_subscriptlist,in_integer)
    local
      Integer i;
      Exp e_1,e;
      Subscript s;
      SubscriptList ss_1,ss;
    case ({},i) then {INDEX(ICONST(i))}; 
    case ({WHOLEDIM()},i) then {INDEX(ICONST(i))}; 
    case ({SLICE(e)},i)
      equation 
        e_1 = simplify(ASUB(e,i)); then {INDEX(e_1)};
    case ({(s = INDEX(_))},i) then {s,INDEX(ICONST(i))}; 
    case (s :: ss,i)
      equation 
        ss_1 = subscripts_append(ss, i); then s :: ss_1;
  end matchcontinue;
end subscripts_append;

public function type_string "
  - Printing expressions
 
  This module provides some functions to print data to the standard
  output.  This is used for error messages, and for debugging the
  semantic description.
 
  !ignorecode"
  input Type in_type_;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_type)
    case INT() then "INT"; 
    case REAL() then "REAL"; 
    case BOOL() then "BOOL"; 
    case STRING() then "STRING"; 
  end matchcontinue;
end type_string;

public function print_component_ref " relation: print_component_ref
 
  Print a `ComponentRef'.
 "
  input ComponentRef in_componentref;
algorithm 
  _:=
  matchcontinue (in_componentref)
    local
      Ident s;
      SubscriptList subs;
      ComponentRef cr;
    case CREF_IDENT(s,subs)
      equation 
        print_component_ref2(s, subs); then ();
    case CREF_QUAL(s,subs,cr) "Does not handle names with underscores "
      equation 
        true = RTOpts.modelica_output();
        print_component_ref2(s, subs);
        Print.print_buf("__");
        print_component_ref(cr); then ();
    case CREF_QUAL(s,subs,cr)
      equation 
        false = RTOpts.modelica_output();
        print_component_ref2(s, subs);
        Print.print_buf(".");
        print_component_ref(cr); then ();
  end matchcontinue;
end print_component_ref;

protected function print_component_ref2
  input String in_string;
  input SubscriptList in_subscriptlist;
algorithm 
  _:=
  matchcontinue (in_string,in_subscriptlist)
    local
      Ident s;
      SubscriptList l;
    case (s,{})
      equation 
        Print.print_buf(s); then ();
    case (s,l)
      equation 
        true = RTOpts.modelica_output();
        Print.print_buf(s);
        Print.print_buf("_L");
        print_list(l, print_subscript, ",");
        Print.print_buf("_R"); then ();
    case (s,l)
      equation 
        false = RTOpts.modelica_output();
        Print.print_buf(s);
        Print.print_buf("[");
        print_list(l, print_subscript, ",");
        Print.print_buf("]"); then ();
  end matchcontinue;
end print_component_ref2;

public function print_subscript " relation: print_subscript
 
  Print a `Subscript'.
 "
  input Subscript in_subscript;
algorithm 
  _:=
  matchcontinue (in_subscript)
    local Exp e1;
    case (WHOLEDIM())
      equation 
        Print.print_buf(":"); then ();
    case (INDEX(e1))
      equation 
        print_exp(e1); then ();
    case (SLICE(e1))
      equation 
        print_exp(e1); then ();
  end matchcontinue;
end print_subscript;

public function print_exp " relation: print_exp
 
  This function prints a complete expression.
 "
  input Exp e;
algorithm 
  print_exp2(e, 0);
end print_exp;

protected function print_exp2
  input Exp in_exp;
  input Integer in_integer;
algorithm 
  _:=
  matchcontinue (in_exp,in_integer)
    local
      Ident s,sym,fs,rstr,str;
      Integer x,pri2_1,pri2,pri3,pri1,i;
      Real r;
      ComponentRef c;
      Exp e1,e2,e21,e22,e,t,f,start,stop,step,cr,dim,exp,iterexp;
      Operator op;
      Type ty,ty2;
      Absyn.Path fcn;
      ExpList args,es;
    case (ICONST(x),_)
      equation 
        s = int_string(x);
        Print.print_buf(s); then ();
    case (RCONST(x),_)
      local Real x;
      equation 
        s = real_string(x);
        Print.print_buf(s); then ();
    case (SCONST(s),_)
      equation 
        Print.print_buf("\"");
        Print.print_buf(s);
        Print.print_buf("\""); then ();
    case (BCONST(false),_)
      equation 
        Print.print_buf("false"); then ();
    case (BCONST(true),_)
      equation 
        Print.print_buf("true"); then ();
    case (CREF(c,_),_)
      equation 
        print_component_ref(c); then ();
    case (BINARY(e1,(op = SUB(ty)),(e2 = BINARY(e21,SUB(ty2),e22))),pri1)
      equation 
        sym = binop_symbol(op);
        pri2_1 = binop_priority(op);
        pri2 = pri2_1 + 1;
        pri3 = print_leftpar(pri1, pri2) "binary minus have higher priority than itself ";
        print_exp2(e1, pri3);
        Print.print_buf(sym);
        print_exp2(e2, pri2);
        print_rightpar(pri1, pri2); then ();
    case (BINARY(e1,op,e2),pri1)
      equation 
        sym = binop_symbol(op);
        pri2 = binop_priority(op);
        pri3 = print_leftpar(pri1, pri2);
        print_exp2(e1, pri3);
        Print.print_buf(sym);
        print_exp2(e2, pri2);
        print_rightpar(pri1, pri2); then ();
    case (UNARY(op,e),pri1)
      equation 
        sym = unaryop_symbol(op);
        pri2 = unaryop_priority(op);
        pri3 = print_leftpar(pri1, pri2);
        Print.print_buf(sym);
        print_exp2(e, pri3);
        print_rightpar(pri1, pri2); then ();
    case (LBINARY(e1,op,e2),pri1)
      equation 
        sym = lbinop_symbol(op);
        pri2 = lbinop_priority(op);
        pri3 = print_leftpar(pri1, pri2);
        print_exp2(e1, pri3);
        Print.print_buf(sym);
        print_exp2(e2, pri2);
        print_rightpar(pri1, pri2); then ();
    case (LUNARY(op,e),pri1)
      equation 
        sym = lunaryop_symbol(op);
        pri2 = lunaryop_priority(op);
        pri3 = print_leftpar(pri1, pri2);
        Print.print_buf(sym);
        print_exp2(e, pri3);
        print_rightpar(pri1, pri2); then ();
    case (RELATION(e1,op,e2),pri1)
      equation 
        sym = relop_symbol(op);
        pri2 = relop_priority(op);
        pri3 = print_leftpar(pri1, pri2);
        print_exp2(e1, pri3);
        Print.print_buf(sym);
        print_exp2(e2, pri2);
        print_rightpar(pri1, pri2); then ();
    case (IFEXP(c,t,f),_)
      local Exp c;
      equation 
        Print.print_buf("if ");
        print_exp2(c, 0);
        Print.print_buf(" then ");
        print_exp2(t, 0);
        Print.print_buf(" else ");
        print_exp2(f, 0); then ();
    case (CALL(fcn,args,_,_),_)
      equation 
        fs = Absyn.path_string(fcn);
        Print.print_buf(fs);
        Print.print_buf("(");
        print_list(args, print_exp, ",");
        Print.print_buf(")"); then ();
    case (ARRAY(_,_,es),_)
      equation 
        Print.print_buf("{");
        print_list(es, print_exp, ",");
        Print.print_buf("}"); then ();
    case (TUPLE(es),_) "PR."
      equation 
        Print.print_buf("(");
        print_list(es, print_exp, ",");
        Print.print_buf(")"); then ();
    case (MATRIX(_,_,es),_)
      local TupleExpBooleanListList es;
      equation 
        Print.print_buf("<matrix>[");
        print_list(es, print_row, ";");
        Print.print_buf("]"); then ();
    case (RANGE(_,start,NONE,stop),pri1)
      equation 
        pri2 = 41;
        pri3 = print_leftpar(pri1, pri2);
        print_exp2(start, pri3);
        Print.print_buf(":");
        print_exp2(stop, pri3);
        print_rightpar(pri1, pri2); then ();
    case (RANGE(_,start,SOME(step),stop),pri1)
      equation 
        pri2 = 41;
        pri3 = print_leftpar(pri1, pri2);
        print_exp2(start, pri3);
        Print.print_buf(":");
        print_exp2(step, pri3);
        Print.print_buf(":");
        print_exp2(stop, pri3);
        print_rightpar(pri1, pri2); then ();
    case (CAST(REAL(),ICONST(i)),_)
      equation 
        false = RTOpts.modelica_output();
        r = int_real(i);
        rstr = real_string(r);
        Print.print_buf(rstr); then ();
    case (CAST(REAL(),e),_)
      equation 
        false = RTOpts.modelica_output();
        Print.print_buf("Real(");
        print_exp(e);
        Print.print_buf(")"); then ();
    case (CAST(REAL(),e),_)
      equation 
        true = RTOpts.modelica_output();
        print_exp(e); then ();
    case (ASUB(e,i),pri1)
      equation 
        pri2 = 51;
        pri3 = print_leftpar(pri1, pri2);
        print_exp2(e, pri3);
        print_rightpar(pri1, pri2);
        Print.print_buf("<asub>[");
        s = int_string(i);
        Print.print_buf(s);
        Print.print_buf("]"); then ();
    case ((e = SIZE(cr,SOME(dim))),_)
      equation 
        str = print_exp_str(e);
        Print.print_buf(str); then ();
    case ((e = SIZE(cr,NONE)),_)
      equation 
        str = print_exp_str(e);
        Print.print_buf(str); then ();
    case ((e = REDUCTION(fcn,exp,i,iterexp)),_)
      local Ident i;
      equation 
        str = print_exp_str(e);
        Print.print_buf(str); then ();
    case (_,_)
      equation 
        Print.print_buf("#UNKNOWN EXPRESSION# ----eee "); then ();
  end matchcontinue;
end print_exp2;

protected function print_leftpar
  input Integer in_integer1;
  input Integer in_integer2;
  output Integer out_integer;
algorithm 
  out_integer:=
  matchcontinue (in_integer1,in_integer2)
    local Integer x,y,pri1,pri2;
    case (x,y)
      equation 
        (x > y) = true;
        Print.print_buf("("); then 0;
    case (pri1,pri2) then pri2; 
  end matchcontinue;
end print_leftpar;

protected function print_rightpar
  input Integer in_integer1;
  input Integer in_integer2;
algorithm 
  _:=
  matchcontinue (in_integer1,in_integer2)
    local Integer x,y;
    case (x,y)
      equation 
        (x > y) = true;
        Print.print_buf(")"); then ();
    case (_,_) then (); 
  end matchcontinue;
end print_rightpar;

public function binop_priority "LS: Changed precedence for unary +-
   which must be higher than binary operators but lower than power
   according to e.g. matlab 
  LS: Changed precedence for binary - , should be higher than + and also
   itself, but this is specially handled in print_exp2 and print_exp2_str 
  priorities:
 
    and, or		10
    not		11
    <, >, =, != etc.	21
    bin +		32
    bin -		33
    			35
    /			36
    unary +, unary -	37
    ^			38
    :			41
    {}			51
 
 "
  input Operator in_operator;
  output Integer out_integer;
algorithm 
  out_integer:=
  matchcontinue (in_operator)
    case (ADD(_)) then 32; 
    case (SUB(_)) then 33; 
    case (ADD_ARR(_)) then 32; 
    case (SUB_ARR(_)) then 33; 
    case (MUL(_)) then 35; 
    case (MUL_SCALAR_ARRAY(_)) then 35; 
    case (MUL_ARRAY_SCALAR(_)) then 35; 
    case (MUL_SCALAR_PRODUCT(_)) then 35; 
    case (MUL_MATRIX_PRODUCT(_)) then 35; 
    case (DIV(_)) then 36; 
    case (DIV_ARRAY_SCALAR(_)) then 36; 
    case (POW(_)) then 38; 
  end matchcontinue;
end binop_priority;

public function unaryop_priority
  input Operator in_operator;
  output Integer out_integer;
algorithm 
  out_integer:=
  matchcontinue (in_operator)
    case (UMINUS(_)) then 37; 
    case (UPLUS(_)) then 37; 
  end matchcontinue;
end unaryop_priority;

public function lbinop_priority
  input Operator in_operator;
  output Integer out_integer;
algorithm 
  out_integer:=
  matchcontinue (in_operator)
    case (AND()) then 10; 
    case (OR()) then 10; 
  end matchcontinue;
end lbinop_priority;

public function lunaryop_priority
  input Operator in_operator;
  output Integer out_integer;
algorithm 
  out_integer:=
  matchcontinue (in_operator)
    case (NOT()) then 11; 
  end matchcontinue;
end lunaryop_priority;

public function relop_priority
  input Operator in_operator;
  output Integer out_integer;
algorithm 
  out_integer:=
  matchcontinue (in_operator)
    case (LESS(_)) then 21; 
    case (LESSEQ(_)) then 21; 
    case (GREATER(_)) then 21; 
    case (GREATEREQ(_)) then 21; 
    case (EQUAL(_)) then 21; 
    case (NEQUAL(_)) then 21; 
  end matchcontinue;
end relop_priority;

public function make_real_add "
   relation: make_real_add
  Construct an add node of the two expressions of type REAL
 "
  input Exp in_exp1;
  input Exp in_exp2;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_exp2)
    local Exp e1,e2;
    case (e1,e2) then BINARY(e1,ADD(REAL()),e2); 
  end matchcontinue;
end make_real_add;

public function make_real_array " relation: make_real_array
  Construct an array node of an Exp list of type REAL
 "
  input ExpList in_explist;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_explist)
    local ExpList expl;
    case (expl) then ARRAY(REAL(),false,expl); 
  end matchcontinue;
end make_real_array;

public function binop_symbol
  input Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    local
      Ident s;
      Operator op;
    case op
      equation 
        false = RTOpts.typeinfo();
        s = binop_symbol_1(op); then s;
    case op
      equation 
        true = RTOpts.typeinfo();
        s = binop_symbol_2(op); then s;
  end matchcontinue;
end binop_symbol;

public function binop_symbol_1 ""
  input Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (ADD(_)) then " + "; 
    case (SUB(_)) then " - "; 
    case (MUL(_)) then " * "; 
    case (DIV(_)) then " / "; 
    case (POW(_)) then " ^ "; 
    case (ADD_ARR(_)) then " + "; 
    case (SUB_ARR(_)) then " - "; 
    case (MUL_SCALAR_ARRAY(_)) then " * "; 
    case (MUL_ARRAY_SCALAR(_)) then " * "; 
    case (MUL_SCALAR_PRODUCT(_)) then " * "; 
    case (MUL_MATRIX_PRODUCT(_)) then " * "; 
    case (DIV_ARRAY_SCALAR(_)) then " / ";  "
  rule	int_string dim => s1 &
	string_append(\" <int{\",s1) => s2 &
	string_append(s2,\"} *> \") => s3
	-------------------------------
	binop_symbol(INT_ARRAY_SCALAR_MUL(dim)) => s3

  rule	int_string dim => s1 &
	string_append(\" <real{\",s1) => s2 &
	string_append(s2,\"} *> \") => s3
	-------------------------------
	binop_symbol(REAL_ARRAY_SCALAR_MUL(dim)) => s3

  axiom	binop_symbol(_) => \" <unknown binop> \"
"
  end matchcontinue;
end binop_symbol_1;

protected function binop_symbol_2
  input Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    local
      Ident ts,s,s_1;
      Type t;
    case (ADD(t))
      equation 
        ts = type_string(t);
        s = string_append(" +<", ts);
        s_1 = string_append(s, "> "); then s_1;
    case (SUB(t)) then " - "; 
    case (MUL(t)) then " * "; 
    case (DIV(t))
      equation 
        ts = type_string(t);
        s = string_append(" /<", ts);
        s_1 = string_append(s, "> "); then s_1;
    case (POW(t)) then " ^ "; 
    case (ADD_ARR(_)) then " + "; 
    case (SUB_ARR(_)) then " - "; 
    case (MUL_SCALAR_ARRAY(_)) then " * "; 
    case (MUL_ARRAY_SCALAR(_)) then " * "; 
    case (MUL_SCALAR_PRODUCT(_)) then " * "; 
    case (MUL_MATRIX_PRODUCT(_)) then " * "; 
    case (DIV_ARRAY_SCALAR(_)) then " / ";  "
  rule	int_string dim => s1 &
	string_append(\" <int{\",s1) => s2 &
	string_append(s2,\"} *> \") => s3
	-------------------------------
	binop_symbol(INT_ARRAY_SCALAR_MUL(dim)) => s3

  rule	int_string dim => s1 &
	string_append(\" <real{\",s1) => s2 &
	string_append(s2,\"} *> \") => s3
	-------------------------------
	binop_symbol(REAL_ARRAY_SCALAR_MUL(dim)) => s3

  axiom	binop_symbol(_) => \" <unknown binop> \"
"
  end matchcontinue;
end binop_symbol_2;

public function unaryop_symbol
  input Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (UMINUS(_)) then "-"; 
    case (UPLUS(_)) then "+"; 
  end matchcontinue;
end unaryop_symbol;

public function lbinop_symbol
  input Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (AND()) then " AND "; 
    case (OR()) then " OR "; 
  end matchcontinue;
end lbinop_symbol;

public function lunaryop_symbol
  input Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (NOT()) then " NOT "; 
  end matchcontinue;
end lunaryop_symbol;

public function relop_symbol
  input Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (LESS(_)) then " < "; 
    case (LESSEQ(_)) then " <= "; 
    case (GREATER(_)) then " > "; 
    case (GREATEREQ(_)) then " >= "; 
    case (EQUAL(_)) then " = "; 
    case (NEQUAL(_)) then " <> "; 
  end matchcontinue;
end relop_symbol;

public function print_list
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
  input String in_string;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
  end FuncType1;
algorithm 
  _:=
  matchcontinue (in_vtype_alist,in_functype1,in_string)
    local
      Type_a h;
      FuncType1 r;
      VType_aList t;
      Ident sep;
    case ({},_,_) then (); 
    case ({h},r,_)
      equation 
        r(h); then ();
    case (h :: t,r,sep)
      equation 
        r(h);
        Print.print_buf(sep);
        print_list(t, r, sep); then ();
  end matchcontinue;
end print_list;

protected function print_row ""
  input TupleExpBooleanList es;
protected 
  ExpList es_1;
algorithm 
  es_1 := Util.list_map(es, Util.tuple2_1);
  print_list(es_1, print_exp, ",");
end print_row;

public function print_component_ref_str "!includecode
  
   LS: print functions that return a string instead of printing *
   Had to duplicate the huge print_exp2 and modify              *
   An alternative would be to implement \"sprint\" somehow        *
   which would need internal state, with reset and              *
   get_string methods                                           *
                                                                
   Once these are tested and ok, the print_exp* above can be    *
   replaced by a call to these _str functions and printing      *
   the result.                                                  *
  
   relation: print_component_ref
 
  Print a `ComponentRef'.
 "
  input ComponentRef in_componentref;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_componentref)
    local
      Ident s,str,strrest,str_1,str_2;
      SubscriptList subs;
      ComponentRef cr;
    case (CREF_IDENT(s,{})) then s;  "optimize "
    case CREF_IDENT(s,subs)
      equation 
        str = print_component_ref2_str(s, subs); then str;
    case CREF_QUAL(s,subs,cr) "Does not handle names with underscores "
      equation 
        true = RTOpts.modelica_output();
        str = print_component_ref2_str(s, subs);
        strrest = print_component_ref_str(cr);
        str_1 = string_append(str, "__");
        str_2 = string_append(str_1, strrest); then str_2;
    case CREF_QUAL(s,subs,cr)
      equation 
        false = RTOpts.modelica_output();
        str = print_component_ref2_str(s, subs);
        strrest = print_component_ref_str(cr);
        str_1 = string_append(str, ".");
        str_2 = string_append(str_1, strrest); then str_2;
  end matchcontinue;
end print_component_ref_str;

protected function print_component_ref2_str
  input Ident in_ident;
  input SubscriptList in_subscriptlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_ident,in_subscriptlist)
    local
      Ident s,str,str_1,str_2,str_3;
      SubscriptList l;
    case (s,{}) then s; 
    case (s,l)
      equation 
        true = RTOpts.modelica_output();
        str = print_list_str(l, print_subscript_str, ",");
        str_1 = string_append(s, "_L");
        str_2 = string_append(str_1, str);
        str_3 = string_append(str_2, "_R"); then str_3;
    case (s,l)
      equation 
        false = RTOpts.modelica_output();
        str = print_list_str(l, print_subscript_str, ",");
        str_1 = string_append(s, "[");
        str_2 = string_append(str_1, str);
        str_3 = string_append(str_2, "]"); then str_3;
  end matchcontinue;
end print_component_ref2_str;

public function print_list_str " relation: print_list_str
 
  Same as print_list, except it returns a string
  instead of printing
 "
  input VType_aList in_vtype_alist;
  input FuncType1 in_functype1;
  input String in_string;
  output String out_string;
protected 
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  function FuncType1
    input Type_a in_type_a;
    output String out_string;
  end FuncType1;
algorithm 
  out_string:=
  matchcontinue (in_vtype_alist,in_functype1,in_string)
    local
      Ident s,srest,s_1,s_2,sep;
      Type_a h;
      FuncType1 r;
      VType_aList t;
    case ({},_,_) then ""; 
    case ({h},r,_)
      equation 
        s = r(h); then s;
    case (h :: t,r,sep)
      equation 
        s = r(h);
        srest = print_list_str(t, r, sep);
        s_1 = string_append(s, sep);
        s_2 = string_append(s_1, srest); then s_2;
  end matchcontinue;
end print_list_str;

public function print_subscript_str " relation: print_subscript_str
 
  Print a `Subscript'.
 "
  input Subscript in_subscript;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_subscript)
    local
      Ident s;
      Exp e1;
    case (WHOLEDIM()) then ":"; 
    case (INDEX(e1))
      equation 
        s = print_exp_str(e1); then s;
    case (SLICE(e1))
      equation 
        s = print_exp_str(e1); then s;
  end matchcontinue;
end print_subscript_str;

public function print_exp_str " relation: print_exp_str
 
  This function prints a complete expression.
 "
  input Exp e;
  output String s;
algorithm 
  s := print_exp2_str(e, 0);
end print_exp_str;

protected function print_exp2_str
  input Exp in_exp;
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_exp,in_integer)
    local
      Ident s,s_1,s_2,sym,s1,s2,s3,s4,s_3,ifstr,thenstr,elsestr,s_4,fs,argstr,s5,s_5,res,res2,crstr,dimstr,str,expstr,iterstr,id;
      Integer x,pri2_1,pri2,pri3,pri1,ival,i;
      Real rval;
      ComponentRef c;
      Exp e1,e2,e21,e22,e,t,f,start,stop,step,cr,dim,exp,iterexp;
      Operator op;
      Type ty,ty2;
      Absyn.Path fcn;
      ExpList args,es;
    case (END(),_) then "end"; 
    case (ICONST(x),_)
      equation 
        s = int_string(x); then s;
    case (RCONST(x),_)
      local Real x;
      equation 
        s = real_string(x); then s;
    case (SCONST(s),_)
      equation 
        s_1 = string_append("\"", s);
        s_2 = string_append(s_1, "\""); then s_2;
    case (BCONST(false),_) then "false"; 
    case (BCONST(true),_) then "true"; 
    case (CREF(c,_),_)
      equation 
        s = print_component_ref_str(c); then s;
    case (BINARY(e1,(op = SUB(ty)),(e2 = BINARY(e21,SUB(ty2),e22))),pri1)
      equation 
        sym = binop_symbol(op);
        pri2_1 = binop_priority(op);
        pri2 = pri2_1 + 1;
        (s1,pri3) = print_leftpar_str(pri1, pri2) "binary minus have higher priority than itself ";
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, sym);
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (BINARY(e1,op,e2),pri1)
      equation 
        sym = binop_symbol(op);
        pri2 = binop_priority(op);
        (s1,pri3) = print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, sym);
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (UNARY(op,e),pri1)
      equation 
        sym = unaryop_symbol(op);
        pri2 = unaryop_priority(op);
        (s1,pri3) = print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e, pri3);
        s3 = print_rightpar_str(pri1, pri2);
        s = string_append(s1, sym);
        s_1 = string_append(s, s2);
        s_2 = string_append(s_1, s3); then s_2;
    case (LBINARY(e1,op,e2),pri1)
      equation 
        sym = lbinop_symbol(op);
        pri2 = lbinop_priority(op);
        (s1,pri3) = print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, sym);
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (LUNARY(op,e),pri1)
      equation 
        sym = lunaryop_symbol(op);
        pri2 = lunaryop_priority(op);
        (s1,pri3) = print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e, pri3);
        s3 = print_rightpar_str(pri1, pri2);
        s = string_append(s1, sym);
        s_1 = string_append(s, s2);
        s_2 = string_append(s_1, s3); then s_2;
    case (RELATION(e1,op,e2),pri1)
      equation 
        sym = relop_symbol(op);
        pri2 = relop_priority(op);
        (s1,pri3) = print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, sym);
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (IFEXP(c,t,f),_)
      local Exp c;
      equation 
        ifstr = print_exp2_str(c, 0);
        thenstr = print_exp2_str(t, 0);
        elsestr = print_exp2_str(f, 0);
        s = string_append("if ", ifstr);
        s_1 = string_append(s, " then ");
        s_2 = string_append(s_1, thenstr);
        s_3 = string_append(s_2, " else ");
        s_4 = string_append(s_3, elsestr); then s_4;
    case (CALL(fcn,args,_,_),_)
      equation 
        fs = Absyn.path_string(fcn);
        argstr = print_list_str(args, print_exp_str, ",");
        s = string_append(fs, "(");
        s_1 = string_append(s, argstr);
        s_2 = string_append(s_1, ")"); then s_2;
    case (ARRAY(_,_,es),_)
      equation 
        s = print_list_str(es, print_exp_str, ",");
        s_1 = string_append("{", s);
        s_2 = string_append(s_1, "}"); then s_2;
    case (TUPLE(es),_)
      equation 
        s = print_list_str(es, print_exp_str, ",");
        s_1 = string_append("(", s);
        s_2 = string_append(s_1, ")"); then s_2;
    case (MATRIX(_,_,es),_)
      local TupleExpBooleanListList es;
      equation 
        s = print_list_str(es, print_row_str, "},{");
        s_1 = string_append("{{", s);
        s_2 = string_append(s_1, "}}"); then s_2;
    case (RANGE(_,start,NONE,stop),pri1)
      equation 
        pri2 = 41;
        (s1,pri3) = print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(start, pri3);
        s3 = print_exp2_str(stop, pri3);
        s4 = print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, ":");
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (RANGE(_,start,SOME(step),stop),pri1)
      equation 
        pri2 = 41;
        (s1,pri3) = print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(start, pri3);
        s3 = print_exp2_str(step, pri3);
        s4 = print_exp2_str(stop, pri3);
        s5 = print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, ":");
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, ":");
        s_4 = string_append(s_3, s4);
        s_5 = string_append(s_4, s5); then s_5;
    case (CAST(REAL(),ICONST(ival)),_)
      equation 
        false = RTOpts.modelica_output();
        rval = int_real(ival);
        res = real_string(rval); then res;
    case (CAST(REAL(),UNARY(UMINUS(_),ICONST(ival))),_)
      equation 
        false = RTOpts.modelica_output();
        rval = int_real(ival);
        res = real_string(rval);
        res2 = string_append("-", res); then res2;
    case (CAST(REAL(),e),_)
      equation 
        false = RTOpts.modelica_output();
        s = print_exp_str(e);
        s_1 = string_append("Real(", s);
        s_2 = string_append(s_1, ")"); then s_2;
    case (CAST(REAL(),e),_)
      equation 
        true = RTOpts.modelica_output();
        s = print_exp_str(e); then s;
    case (ASUB(e,i),pri1)
      equation 
        pri2 = 51;
        (s1,pri3) = print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e, pri3);
        s3 = print_rightpar_str(pri1, pri2);
        s4 = int_string(i);
        s = string_append(s1, s2);
        s_1 = string_append(s, s3);
        s_2 = string_append(s_1, "[");
        s_3 = string_append(s_2, s4);
        s_4 = string_append(s_3, "]"); then s_4;
    case (SIZE(cr,SOME(dim)),_)
      equation 
        crstr = print_exp_str(cr);
        dimstr = print_exp_str(dim);
        str = Util.string_append_list({"size(",crstr,",",dimstr,")"}); then str;
    case (SIZE(cr,NONE),_)
      equation 
        crstr = print_exp_str(cr);
        str = Util.string_append_list({"size(",crstr,")"}); then str;
    case (REDUCTION(fcn,exp,id,iterexp),_)
      equation 
        fs = Absyn.path_string(fcn);
        expstr = print_exp_str(exp);
        iterstr = print_exp_str(iterexp);
        str = Util.string_append_list({"<reduction>",fs,"(",expstr," for ",id," in ",iterstr,")"}); then str;
    case (_,_) then "#UNKNOWN EXPRESSION# ----eee "; 
  end matchcontinue;
end print_exp2_str;

public function print_row_str
  input TupleExpBooleanList es;
  output String s;
protected 
  ExpList es_1;
algorithm 
  es_1 := Util.list_map(es, Util.tuple2_1);
  s := print_list_str(es_1, print_exp_str, ",");
end print_row_str;

public function print_leftpar_str
  input Integer in_integer1;
  input Integer in_integer2;
  output String out_string;
  output Integer out_integer;
algorithm 
  (out_string,out_integer):=
  matchcontinue (in_integer1,in_integer2)
    local Integer x,y,pri1,pri2;
    case (x,y)
      equation 
        (x > y) = true; then ("(",0);
    case (pri1,pri2) then ("",pri2); 
  end matchcontinue;
end print_leftpar_str;

public function print_rightpar_str
  input Integer in_integer1;
  input Integer in_integer2;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_integer1,in_integer2)
    local Integer x,y;
    case (x,y)
      equation 
        (x > y) = true; then ")";
    case (_,_) then ""; 
  end matchcontinue;
end print_rightpar_str;

public function exp_equal
  input Exp in_exp1;
  input Exp in_exp2;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp1,in_exp2)
    local
      type BooleanList = list<Boolean>;
      Integer c1,c2,i1,i2;
      Ident id1,id2;
      Boolean b1,c1_1,c2_1,b2,res,b3;
      Exp e11,e12,e21,e22,e1,e2,e13,e23,r1,r2;
      Operator op1,op2;
      BooleanList bs;
      Absyn.Path path1,path2;
      ExpList expl1,expl2;
      Type tp1,tp2;
    case (ICONST(c1),ICONST(c2)) then c1 == c2; 
    case (RCONST(c1),RCONST(c2))
      local Real c1,c2; then c1 ==. c2;
    case (SCONST(c1),SCONST(c2))
      local Ident c1,c2;
      equation 
        equality(c1 = c2); then true;
    case (BCONST(c1),BCONST(c2))
      local Boolean c1,c2;
      equation 
        b1 = bool_and(c1, c2);
        c1_1 = bool_not(c1);
        c2_1 = bool_not(c2);
        b2 = bool_and(c1_1, c2_1);
        res = bool_or(b1, b2); then res;
    case (CREF(c1,_),CREF(c2,_))
      local ComponentRef c1,c2;
      equation 
        res = cref_equal(c1, c2); then res;
    case (BINARY(e11,op1,e12),BINARY(e21,op2,e22))
      equation 
        b1 = operator_equal(op1, op2);
        b2 = exp_equal(e11, e21);
        b3 = exp_equal(e12, e22);
        res = Util.bool_and_list({b1,b2,b3}); then res;
    case (LBINARY(e11,op1,e12),LBINARY(e21,op2,e22))
      equation 
        b1 = operator_equal(op1, op2);
        b2 = exp_equal(e11, e21);
        b3 = exp_equal(e12, e22);
        res = Util.bool_and_list({b1,b2,b3}); then res;
    case (UNARY(op1,e1),UNARY(op2,e2))
      equation 
        b1 = operator_equal(op1, op2);
        b2 = exp_equal(e1, e2);
        res = bool_and(b1, b2); then res;
    case (LUNARY(op1,e1),LUNARY(op2,e2))
      equation 
        b1 = operator_equal(op1, op2);
        b2 = exp_equal(e1, e2);
        res = bool_and(b1, b2); then res;
    case (RELATION(e11,op1,e12),RELATION(e21,op2,e22))
      equation 
        b1 = operator_equal(op1, op2);
        b2 = exp_equal(e11, e21);
        b3 = exp_equal(e12, e22);
        res = Util.bool_and_list({b1,b2,b3}); then res;
    case (IFEXP(e11,e12,e13),IFEXP(e21,e22,e23))
      equation 
        b1 = exp_equal(e13, e23);
        b2 = exp_equal(e11, e21);
        b3 = exp_equal(e12, e22);
        res = Util.bool_and_list({b1,b2,b3}); then res;
    case (CALL(path1,expl1,_,_),CALL(path2,expl2,_,_))
      equation 
        b1 = ModUtil.path_equal(path1, path2);
        bs = Util.list_thread_map(expl1, expl2, exp_equal);
        res = Util.bool_and_list(b1 :: bs); then res;
    case (ARRAY(tp1,_,expl1),ARRAY(tp2,_,expl2))
      equation 
        equality(tp1 = tp2);
        bs = Util.list_thread_map(expl1, expl2, exp_equal);
        res = Util.bool_and_list(bs); then res;
    case (MATRIX(_,_,_),MATRIX(_,_,_))
      equation 
        print("exp_equal for MATRIX not impl. yet.\n"); then false;
    case (RANGE(tp1,e11,NONE,e13),RANGE(tp2,e21,NONE,e23))
      equation 
        b1 = exp_equal(e13, e23);
        b2 = exp_equal(e11, e21);
        res = Util.bool_and_list({b1,b2}); then res;
    case (RANGE(tp1,e11,SOME(e12),e13),RANGE(tp2,e21,SOME(e22),e23))
      equation 
        b1 = exp_equal(e13, e23);
        b2 = exp_equal(e11, e21);
        b3 = exp_equal(e12, e22);
        res = Util.bool_and_list({b1,b2,b3}); then res;
    case (TUPLE(expl1),TUPLE(expl2))
      equation 
        bs = Util.list_thread_map(expl1, expl2, exp_equal);
        res = Util.bool_and_list(bs); then res;
    case (CAST(tp1,e1),CAST(tp2,e2))
      equation 
        equality(tp1 = tp2);
        res = exp_equal(e1, e2); then res;
    case (ASUB(e1,i1),ASUB(e2,i2))
      equation 
        b1 = (i1 == i2);
        b2 = exp_equal(e1, e2);
        res = bool_and(b1, b2); then res;
    case (SIZE(e1,NONE),SIZE(e2,NONE))
      equation 
        res = exp_equal(e1, e2); then res;
    case (SIZE(e1,SOME(e11)),SIZE(e2,SOME(e22)))
      equation 
        b1 = exp_equal(e1, e2);
        b2 = exp_equal(e11, e22);
        res = bool_and(b1, b2); then res;
    case (CODE(_,_),CODE(_,_))
      equation 
        print("exp_equal on CODE not impl.\n"); then false;
    case (REDUCTION(path1,e1,id1,r1),REDUCTION(path2,e2,id2,r2))
      equation 
        equality(id1 = id2);
        b1 = ModUtil.path_equal(path1, path2);
        b2 = exp_equal(e1, e2);
        b3 = exp_equal(r1, r2);
        res = Util.bool_and_list({b1,b2,b3}); then res;
    case (END(),END()) then true; 
    case (_,_) then false; 
  end matchcontinue;
end exp_equal;

protected function operator_equal
  input Operator in_operator1;
  input Operator in_operator2;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_operator1,in_operator2)
    local
      Boolean res;
      Absyn.Path p1,p2;
    case (ADD(_),ADD(_)) then true; 
    case (SUB(_),SUB(_)) then true; 
    case (MUL(_),MUL(_)) then true; 
    case (DIV(_),DIV(_)) then true; 
    case (POW(_),POW(_)) then true; 
    case (UMINUS(_),UMINUS(_)) then true; 
    case (UMINUS_ARR(_),UMINUS_ARR(_)) then true; 
    case (UPLUS_ARR(_),UPLUS_ARR(_)) then true; 
    case (ADD_ARR(_),ADD_ARR(_)) then true; 
    case (SUB_ARR(_),SUB_ARR(_)) then true; 
    case (MUL_SCALAR_ARRAY(_),MUL_SCALAR_ARRAY(_)) then true; 
    case (MUL_ARRAY_SCALAR(_),MUL_ARRAY_SCALAR(_)) then true; 
    case (MUL_SCALAR_PRODUCT(_),MUL_SCALAR_PRODUCT(_)) then true; 
    case (MUL_MATRIX_PRODUCT(_),MUL_MATRIX_PRODUCT(_)) then true; 
    case (DIV_ARRAY_SCALAR(_),DIV_ARRAY_SCALAR(_)) then true; 
    case (POW_ARR(_),POW_ARR(_)) then true; 
    case (AND(),AND()) then true; 
    case (OR(),OR()) then true; 
    case (NOT(),NOT()) then true; 
    case (LESS(_),LESS(_)) then true; 
    case (LESSEQ(_),LESSEQ(_)) then true; 
    case (GREATER(_),GREATER(_)) then true; 
    case (GREATEREQ(_),GREATEREQ(_)) then true; 
    case (EQUAL(_),EQUAL(_)) then true; 
    case (NEQUAL(_),NEQUAL(_)) then true; 
    case (USERDEFINED(p1),USERDEFINED(p2))
      equation 
        res = ModUtil.path_equal(p1, p2); then res;
    case (_,_) then false; 
  end matchcontinue;
end operator_equal;

public function replace_exp_list "target list"
  input Exp in_exp1;
  input ExpList in_explist2;
  input ExpList in_explist3;
  output Exp out_exp;
  output Integer out_integer;
algorithm 
  (out_exp,out_integer):=
  matchcontinue (in_exp1,in_explist2,in_explist3)
    local
      Exp e,e_1,e_2,s1,t1;
      Integer c1,c2,c;
      ExpList sr,tr;
    case (e,{},{}) then (e,0); 
    case (e,s1 :: sr,t1 :: tr)
      equation 
        (e_1,c1) = replace_exp(e, s1, t1);
        (e_2,c2) = replace_exp_list(e_1, sr, tr);
        c = c1 + c2; then (e_2,c);
  end matchcontinue;
end replace_exp_list;

public function replace_exp "target expr"
  input Exp in_exp1;
  input Exp in_exp2;
  input Exp in_exp3;
  output Exp out_exp;
  output Integer out_integer;
algorithm 
  (out_exp,out_integer):=
  matchcontinue (in_exp1,in_exp2,in_exp3)
    local
      type IntegerList = list<Integer>;
      Exp expr,source,target,e1_1,e2_1,e1,e2,e3_1,e3,e_1,r_1,e,r,s;
      Integer c1,c2,c,c3,cnt_1,b,i;
      Operator op;
      ExpList expl_1,expl;
      IntegerList cnt;
      Absyn.Path path,p;
      Boolean t;
      Type tp,a;
      Ident id;
    case (expr,source,target) "expr source expr target expr"
      equation 
        true = exp_equal(expr, source); then (target,1);
    case (BINARY(e1,op,e2),source,target)
      equation 
        (e1_1,c1) = replace_exp(e1, source, target);
        (e2_1,c2) = replace_exp(e2, source, target);
        c = c1 + c2; then (BINARY(e1_1,op,e2_1),c);
    case (LBINARY(e1,op,e2),source,target)
      equation 
        (e1_1,c1) = replace_exp(e1, source, target);
        (e2_1,c2) = replace_exp(e2, source, target);
        c = c1 + c2; then (LBINARY(e1_1,op,e2_1),c);
    case (UNARY(op,e1),source,target)
      equation 
        (e1_1,c) = replace_exp(e1, source, target); then (UNARY(op,e1_1),c);
    case (LUNARY(op,e1),source,target)
      equation 
        (e1_1,c) = replace_exp(e1, source, target); then (LUNARY(op,e1_1),c);
    case (RELATION(e1,op,e2),source,target)
      equation 
        (e1_1,c1) = replace_exp(e1, source, target);
        (e2_1,c2) = replace_exp(e2, source, target);
        c = c1 + c2; then (RELATION(e1_1,op,e2_1),c);
    case (IFEXP(e1,e2,e3),source,target)
      equation 
        (e1_1,c1) = replace_exp(e1, source, target);
        (e2_1,c2) = replace_exp(e2, source, target);
        (e3_1,c3) = replace_exp(e3, source, target);
        c = Util.list_reduce({c1,c2,c3}, int_add); then (IFEXP(e1_1,e2_1,e3_1),c);
    case (CALL(path,expl,t,c),source,target)
      local Boolean c;
      equation 
        expl_1 = Util.list_map_2_2(expl, replace_exp, source, target);
        (expl_1,cnt) = Util.split_tuple2_list(expl_1);
        cnt_1 = Util.list_reduce(cnt, int_add); then (CALL(path,expl_1,t,c),cnt_1);
    case (ARRAY(tp,c,expl),source,target)
      local Boolean c;
      equation 
        expl_1 = Util.list_map_2_2(expl, replace_exp, source, target);
        (expl_1,cnt) = Util.split_tuple2_list(expl_1);
        cnt_1 = Util.list_reduce(cnt, int_add); then (ARRAY(tp,c,expl_1),cnt_1);
    case (MATRIX(a,b,c),source,target)
      local TupleExpBooleanListList c;
      equation 
        print("replace_exp for matrix not impl."); then (MATRIX(a,b,c),0);
    case (RANGE(tp,e1,NONE,e2),source,target)
      equation 
        (e1_1,c1) = replace_exp(e1, source, target);
        (e2_1,c2) = replace_exp(e2, source, target);
        c = c1 + c2; then (RANGE(tp,e1_1,NONE,e2_1),c);
    case (RANGE(tp,e1,SOME(e3),e2),source,target)
      equation 
        (e1_1,c1) = replace_exp(e1, source, target);
        (e2_1,c2) = replace_exp(e2, source, target);
        (e3_1,c3) = replace_exp(e3, source, target);
        c = Util.list_reduce({c1,c2,c3}, int_add); then (RANGE(tp,e1_1,SOME(e3_1),e2_1),c);
    case (TUPLE(expl),source,target)
      equation 
        expl_1 = Util.list_map_2_2(expl, replace_exp, source, target);
        (expl_1,cnt) = Util.split_tuple2_list(expl_1);
        cnt_1 = Util.list_reduce(cnt, int_add); then (TUPLE(expl_1),cnt_1);
    case (CAST(tp,e1),source,target)
      equation 
        (e1_1,c) = replace_exp(e1, source, target); then (CAST(tp,e1_1),c);
    case (ASUB(e1,i),source,target)
      equation 
        (e1_1,c) = replace_exp(e1, source, target); then (ASUB(e1_1,i),c);
    case (SIZE(e1,NONE),source,target)
      equation 
        (e1_1,c) = replace_exp(e1, source, target); then (SIZE(e1_1,NONE),c);
    case (SIZE(e1,SOME(e2)),source,target)
      equation 
        (e1_1,c1) = replace_exp(e1, source, target);
        (e2_1,c2) = replace_exp(e2, source, target);
        c = c1 + c2; then (SIZE(e1_1,SOME(e2_1)),c);
    case (CODE(a,b),source,target)
      local
        Absyn.Code a;
        Type b;
      equation 
        print("replace_exp on CODE not impl.\n"); then (CODE(a,b),0);
    case (REDUCTION(p,e,id,r),source,target)
      equation 
        (e_1,c1) = replace_exp(e, source, target);
        (r_1,c2) = replace_exp(r, source, target);
        c = c1 + c2; then (REDUCTION(p,e_1,id,r_1),c);
    case (e,s,_) then (e,0); 
  end matchcontinue;
end replace_exp;

public function stringify_component_ref
  input ComponentRef cr;
  output ComponentRef out_componentref;
protected 
  Ident crs;
algorithm 
  crs := print_component_ref_str(cr);
  out_componentref := CREF_IDENT(crs,{});
end stringify_component_ref;

public function stringify_crefs " relation: stringify_crefs
  This function takes an expression and transforms all component reference names
  contained in the expression to a simpler form.
  For instance CREF_QUAL(\"a\",{}, CREF_IDENT(\"b\",{})) becomes
  CREF_IDENT(\"a.b\",{})
 "
  input Exp in_exp;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp)
    local
      Exp e,e1_1,e2_1,e1,e2,e_1,e3_1,e3;
      ComponentRef cr_1,cr;
      Type t;
      Operator op;
      ExpList expl_1,expl;
      Absyn.Path p;
      Boolean b;
      Integer i;
      Ident id;
    case ((e = ICONST(_))) then e; 
    case ((e = RCONST(_))) then e; 
    case ((e = SCONST(_))) then e; 
    case ((e = BCONST(_))) then e; 
    case (CREF(cr,t))
      equation 
        cr_1 = stringify_component_ref(cr); then CREF(cr_1,t);
    case (BINARY(e1,op,e2))
      equation 
        e1_1 = stringify_crefs(e1);
        e2_1 = stringify_crefs(e2); then BINARY(e1_1,op,e2_1);
    case (UNARY(op,e))
      equation 
        e_1 = stringify_crefs(e); then UNARY(op,e_1);
    case (LBINARY(e1,op,e2))
      equation 
        e1_1 = stringify_crefs(e1);
        e2_1 = stringify_crefs(e2); then LBINARY(e1_1,op,e2_1);
    case (LUNARY(op,e))
      equation 
        e_1 = stringify_crefs(e); then LUNARY(op,e_1);
    case (RELATION(e1,op,e2))
      equation 
        e1_1 = stringify_crefs(e1);
        e2_1 = stringify_crefs(e2); then RELATION(e1_1,op,e2_1);
    case (IFEXP(e1,e2,e3))
      equation 
        e1_1 = stringify_crefs(e1);
        e2_1 = stringify_crefs(e2);
        e3_1 = stringify_crefs(e3); then IFEXP(e1_1,e2_1,e3_1);
    case (CALL(p,expl,t,b))
      local Boolean t;
      equation 
        expl_1 = Util.list_map(expl, stringify_crefs); then CALL(p,expl_1,t,b);
    case (ARRAY(t,b,expl))
      equation 
        expl_1 = Util.list_map(expl, stringify_crefs); then ARRAY(t,b,expl_1);
    case ((e = MATRIX(t,b,expl)))
      local
        Integer b;
        TupleExpBooleanListList expl;
      equation 
        print("stringify_crefs on MATRIX not impl. yet\n"); then e;
    case (RANGE(t,e1,SOME(e2),e3))
      equation 
        e1_1 = stringify_crefs(e1);
        e2_1 = stringify_crefs(e2);
        e3_1 = stringify_crefs(e3); then RANGE(t,e1_1,SOME(e2_1),e3_1);
    case (RANGE(t,e1,NONE,e3))
      equation 
        e1_1 = stringify_crefs(e1);
        e3_1 = stringify_crefs(e3); then RANGE(t,e1_1,NONE,e3_1);
    case (TUPLE(expl))
      equation 
        expl_1 = Util.list_map(expl, stringify_crefs); then TUPLE(expl_1);
    case (CAST(t,e1))
      equation 
        e1_1 = stringify_crefs(e1); then CAST(t,e1_1);
    case (ASUB(e1,i))
      equation 
        e1_1 = stringify_crefs(e1); then ASUB(e1_1,i);
    case (SIZE(e1,SOME(e2)))
      equation 
        e1_1 = stringify_crefs(e1);
        e2_1 = stringify_crefs(e2); then SIZE(e1_1,SOME(e2_1));
    case (SIZE(e1,NONE))
      equation 
        e1_1 = stringify_crefs(e1); then SIZE(e1_1,NONE);
    case ((e = CODE(_,_))) then e; 
    case (REDUCTION(p,e1,id,e2))
      equation 
        e1_1 = stringify_crefs(e1);
        e2_1 = stringify_crefs(e2); then REDUCTION(p,e1_1,id,e2_1);
    case END() then END(); 
    case (e) then e; 
  end matchcontinue;
end stringify_crefs;

public function dump_exp_graphviz
  input Exp in_exp;
  output Graphviz.Node out_node;
algorithm 
  out_node:=
  matchcontinue (in_exp)
    local
      type Graphviz_NodeList = list<Graphviz.Node>;
      Ident s,s_1,s_2,sym,fs,tystr,istr,id;
      Integer x,i;
      ComponentRef c;
      Graphviz.Node lt,rt,ct,tt,ft,t1,t2,t3,crt,dimt,expt,itert;
      Exp e1,e2,e,t,f,start,stop,step,cr,dim,exp,iterexp;
      Operator op;
      Graphviz_NodeList argnodes,nodes;
      Absyn.Path fcn;
      ExpList args,es;
      Type ty;
    case (END()) then Graphviz.NODE("END",{},{}); 
    case (ICONST(x))
      equation 
        s = int_string(x); then Graphviz.LNODE("ICONST",{s},{},{});
    case (RCONST(x))
      local Real x;
      equation 
        s = real_string(x); then Graphviz.LNODE("RCONST",{s},{},{});
    case (SCONST(s))
      equation 
        s_1 = string_append("\"", s);
        s_2 = string_append(s_1, "\""); then Graphviz.LNODE("SCONST",{s_2},{},{});
    case (BCONST(false)) then Graphviz.LNODE("BCONST",{"false"},{},{}); 
    case (BCONST(true)) then Graphviz.LNODE("BCONST",{"true"},{},{}); 
    case (CREF(c,_))
      equation 
        s = print_component_ref_str(c); then Graphviz.LNODE("CREF",{s},{},{});
    case (BINARY(e1,op,e2))
      equation 
        sym = binop_symbol(op);
        lt = dump_exp_graphviz(e1);
        rt = dump_exp_graphviz(e2); then Graphviz.LNODE("BINARY",{sym},{},{lt,rt});
    case (UNARY(op,e))
      equation 
        sym = unaryop_symbol(op);
        ct = dump_exp_graphviz(e); then Graphviz.LNODE("UNARY",{sym},{},{ct});
    case (LBINARY(e1,op,e2))
      equation 
        sym = lbinop_symbol(op);
        lt = dump_exp_graphviz(e1);
        rt = dump_exp_graphviz(e2); then Graphviz.LNODE("LBINARY",{sym},{},{lt,rt});
    case (LUNARY(op,e))
      equation 
        sym = lunaryop_symbol(op);
        ct = dump_exp_graphviz(e); then Graphviz.LNODE("LUNARY",{sym},{},{ct});
    case (RELATION(e1,op,e2))
      equation 
        sym = relop_symbol(op);
        lt = dump_exp_graphviz(e1);
        rt = dump_exp_graphviz(e2); then Graphviz.LNODE("RELATION",{sym},{},{lt,rt});
    case (IFEXP(c,t,f))
      local Exp c;
      equation 
        ct = dump_exp_graphviz(c);
        tt = dump_exp_graphviz(t);
        ft = dump_exp_graphviz(f); then Graphviz.NODE("IFEXP",{},{ct,tt,ft});
    case (CALL(fcn,args,_,_))
      equation 
        fs = Absyn.path_string(fcn);
        argnodes = Util.list_map(args, dump_exp_graphviz); then Graphviz.LNODE("CALL",{fs},{},argnodes);
    case (ARRAY(_,_,es))
      equation 
        nodes = Util.list_map(es, dump_exp_graphviz); then Graphviz.NODE("ARRAY",{},nodes);
    case (TUPLE(es))
      equation 
        nodes = Util.list_map(es, dump_exp_graphviz); then Graphviz.NODE("TUPLE",{},nodes);
    case (MATRIX(_,_,es))
      local TupleExpBooleanListList es;
      equation 
        s = print_list_str(es, print_row_str, "},{");
        s_1 = string_append("{{", s);
        s_2 = string_append(s_1, "}}"); then Graphviz.LNODE("MATRIX",{s_2},{},{});
    case (RANGE(_,start,NONE,stop))
      equation 
        t1 = dump_exp_graphviz(start);
        t2 = Graphviz.NODE(":",{},{});
        t3 = dump_exp_graphviz(stop); then Graphviz.NODE("RANGE",{},{t1,t2,t3});
    case (RANGE(_,start,SOME(step),stop))
      equation 
        t1 = dump_exp_graphviz(start);
        t2 = dump_exp_graphviz(step);
        t3 = dump_exp_graphviz(stop); then Graphviz.NODE("RANGE",{},{t1,t2,t3});
    case (CAST(ty,e))
      equation 
        tystr = type_string(ty);
        ct = dump_exp_graphviz(e); then Graphviz.LNODE("CAST",{tystr},{},{ct});
    case (ASUB(e,i))
      equation 
        ct = dump_exp_graphviz(e);
        istr = int_string(i);
        s = Util.string_append_list({"[",istr,"]"}); then Graphviz.LNODE("ASUB",{s},{},{ct});
    case (SIZE(cr,SOME(dim)))
      equation 
        crt = dump_exp_graphviz(cr);
        dimt = dump_exp_graphviz(dim); then Graphviz.NODE("SIZE",{},{crt,dimt});
    case (SIZE(cr,NONE))
      equation 
        crt = dump_exp_graphviz(cr); then Graphviz.NODE("SIZE",{},{crt});
    case (REDUCTION(fcn,exp,id,iterexp))
      equation 
        fs = Absyn.path_string(fcn);
        expt = dump_exp_graphviz(exp);
        itert = dump_exp_graphviz(iterexp); then Graphviz.LNODE("REDUCTION",{fs},{},{expt,itert});
    case (_) then Graphviz.NODE("#UNKNOWN EXPRESSION# ----eeestr ",{},{}); 
  end matchcontinue;
end dump_exp_graphviz;

protected function gen_string_n_time
  input String in_string;
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_string,in_integer)
    local
      Ident str,new_str,res_str;
      Integer new_level,level;
    case (str,0) then ""; 
    case (str,level)
      equation 
        new_level = level + -1;
        new_str = gen_string_n_time(str, new_level);
        res_str = string_append(str, new_str); then res_str;
  end matchcontinue;
end gen_string_n_time;

public function dump_exp_str
  input Exp in_exp;
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_exp,in_integer)
    local
      type IdentList = list<Ident>;
      Ident gen_str,res_str,s,s_1,s_2,sym,lt,rt,ct,tt,ft,fs,argnodes_1,nodes_1,t1,t2,t3,tystr,istr,crt,dimt,expt,itert,id;
      Integer level,x,new_level1,new_level2,new_level3,i;
      ComponentRef c;
      Exp e1,e2,e,t,f,start,stop,step,cr,dim,exp,iterexp;
      Operator op;
      IdentList argnodes,nodes;
      Absyn.Path fcn;
      ExpList args,es;
      Type ty;
    case (END(),level)
      equation 
        gen_str = gen_string_n_time("   |", level);
        res_str = Util.string_append_list({gen_str,"END","\n"}); then res_str;
    case (ICONST(x),level) "Graphviz.LNODE(\"ICONST\",{s},{},{})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        s = int_string(x);
        res_str = Util.string_append_list({gen_str,"ICONST ",s,"\n"}); then res_str;
    case (RCONST(x),level) "Graphviz.LNODE(\"RCONST\",{s},{},{})"
      local Real x;
      equation 
        gen_str = gen_string_n_time("   |", level);
        s = real_string(x);
        res_str = Util.string_append_list({gen_str,"RCONST ",s,"\n"}); then res_str;
    case (SCONST(s),level) "Graphviz.LNODE(\"SCONST\",{s''},{},{})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        s_1 = string_append("\"", s);
        s_2 = string_append(s_1, "\"");
        res_str = Util.string_append_list({gen_str,"SCONST ",s_2,"\n"}); then res_str;
    case (BCONST(false),level) "Graphviz.LNODE(\"BCONST\",{\"false\"},{},{})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        res_str = Util.string_append_list({gen_str,"BCONST ","false","\n"}); then res_str;
    case (BCONST(true),level) "Graphviz.LNODE(\"BCONST\",{\"true\"},{},{})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        res_str = Util.string_append_list({gen_str,"BCONST ","true","\n"}); then res_str;
    case (CREF(c,_),level) "Graphviz.LNODE(\"CREF\",{s},{},{})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        s = print_component_ref_str(c);
        res_str = Util.string_append_list({gen_str,"CREF ",s,"\n"}); then res_str;
    case (BINARY(e1,op,e2),level) "Graphviz.LNODE(\"BINARY\",{sym},{},{lt,rt})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        new_level2 = level + 1;
        sym = binop_symbol(op);
        lt = dump_exp_str(e1, new_level1);
        rt = dump_exp_str(e2, new_level2);
        res_str = Util.string_append_list({gen_str,"BINARY ",sym,"\n",lt,rt,""}); then res_str;
    case (UNARY(op,e),level) "Graphviz.LNODE(\"UNARY\",{sym},{},{ct})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        sym = unaryop_symbol(op);
        ct = dump_exp_str(e, new_level1);
        res_str = Util.string_append_list({gen_str,"UNARY ",sym,"\n",ct,""}); then res_str;
    case (LBINARY(e1,op,e2),level) "Graphviz.LNODE(\"LBINARY\",{sym},{},{lt,rt})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        new_level2 = level + 1;
        sym = lbinop_symbol(op);
        lt = dump_exp_str(e1, new_level1);
        rt = dump_exp_str(e2, new_level2);
        res_str = Util.string_append_list({gen_str,"LBINARY ",sym,"\n",lt,rt,""}); then res_str;
    case (LUNARY(op,e),level) "Graphviz.LNODE(\"LUNARY\",{sym},{},{ct})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        sym = lunaryop_symbol(op);
        ct = dump_exp_str(e, new_level1);
        res_str = Util.string_append_list({gen_str,"LUNARY ",sym,"\n",ct,""}); then res_str;
    case (RELATION(e1,op,e2),level) "Graphviz.LNODE(\"RELATION\",{sym},{},{lt,rt})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        new_level2 = level + 1;
        sym = relop_symbol(op);
        lt = dump_exp_str(e1, new_level1);
        rt = dump_exp_str(e2, new_level2);
        res_str = Util.string_append_list({gen_str,"RELATION ",sym,"\n",lt,rt,""}); then res_str;
    case (IFEXP(c,t,f),level) "Graphviz.NODE(\"IFEXP\",{},{ct,tt,ft})"
      local Exp c;
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        new_level2 = level + 1;
        new_level3 = level + 1;
        ct = dump_exp_str(c, new_level1);
        tt = dump_exp_str(t, new_level2);
        ft = dump_exp_str(f, new_level3);
        res_str = Util.string_append_list({gen_str,"IFEXP ","\n",ct,tt,ft,""}); then res_str;
    case (CALL(fcn,args,_,_),level) "Graphviz.LNODE(\"CALL\",{fs},{},argnodes)"
      equation 
        gen_str = gen_string_n_time("   |", level);
        fs = Absyn.path_string(fcn);
        new_level1 = level + 1;
        argnodes = Util.list_map_1(args, dump_exp_str, new_level1);
        argnodes_1 = Util.string_append_list(argnodes);
        res_str = Util.string_append_list({gen_str,"CALL ",fs,"\n",argnodes_1,""}); then res_str;
    case (ARRAY(_,_,es),level) "Graphviz.NODE(\"ARRAY\",{},nodes)"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        nodes = Util.list_map_1(es, dump_exp_str, new_level1);
        nodes_1 = Util.string_append_list(nodes);
        res_str = Util.string_append_list({gen_str,"ARRAY ",nodes_1,"\n"}); then res_str;
    case (TUPLE(es),level) "Graphviz.NODE(\"TUPLE\",{},nodes)"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        nodes = Util.list_map_1(es, dump_exp_str, new_level1);
        nodes_1 = Util.string_append_list(nodes);
        res_str = Util.string_append_list({gen_str,"TUPLE ",nodes_1,"\n"}); then res_str;
    case (MATRIX(_,_,es),level) "Graphviz.LNODE(\"MATRIX\",{s''},{},{})"
      local TupleExpBooleanListList es;
      equation 
        gen_str = gen_string_n_time("   |", level);
        s = print_list_str(es, print_row_str, "},{");
        s_1 = string_append("{{", s);
        s_2 = string_append(s_1, "}}");
        res_str = Util.string_append_list({gen_str,"MATRIX ","\n",s_2,"","\n"}); then res_str;
    case (RANGE(_,start,NONE,stop),level) "Graphviz.NODE(\"RANGE\",{},{t1,t2,t3})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        new_level2 = level + 1;
        t1 = dump_exp_str(start, new_level1);
        t2 = ":";
        t3 = dump_exp_str(stop, new_level2);
        res_str = Util.string_append_list({gen_str,"RANGE ","\n",t1,t2,t3,""}); then res_str;
    case (RANGE(_,start,SOME(step),stop),level) "Graphviz.NODE(\"RANGE\",{},{t1,t2,t3})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        new_level2 = level + 1;
        new_level3 = level + 1;
        t1 = dump_exp_str(start, new_level1);
        t2 = dump_exp_str(step, new_level2);
        t3 = dump_exp_str(stop, new_level3);
        res_str = Util.string_append_list({gen_str,"RANGE ","\n",t1,t2,t3,""}); then res_str;
    case (CAST(ty,e),level) "Graphviz.LNODE(\"CAST\",{tystr},{},{ct})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        tystr = type_string(ty);
        ct = dump_exp_str(e, new_level1);
        res_str = Util.string_append_list({gen_str,"CAST ","\n",ct,""}); then res_str;
    case (ASUB(e,i),level) "Graphviz.LNODE(\"ASUB\",{s},{},{ct})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        ct = dump_exp_str(e, new_level1);
        istr = int_string(i);
        s = Util.string_append_list({"[",istr,"]"});
        res_str = Util.string_append_list({gen_str,"ASUB ","\n",s,ct,""}); then res_str;
    case (SIZE(cr,SOME(dim)),level) "Graphviz.NODE(\"SIZE\",{},{crt,dimt})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        new_level2 = level + 1;
        crt = dump_exp_str(cr, new_level1);
        dimt = dump_exp_str(dim, new_level2);
        res_str = Util.string_append_list({gen_str,"SIZE ","\n",crt,dimt,""}); then res_str;
    case (SIZE(cr,NONE),level) "Graphviz.NODE(\"SIZE\",{},{crt})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        crt = dump_exp_str(cr, new_level1);
        res_str = Util.string_append_list({gen_str,"SIZE ","\n",crt,""}); then res_str;
    case (REDUCTION(fcn,exp,id,iterexp),level) "Graphviz.LNODE(\"REDUCTION\",{fs},{},{expt,itert})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        new_level1 = level + 1;
        new_level2 = level + 1;
        fs = Absyn.path_string(fcn);
        expt = dump_exp_str(exp, new_level1);
        itert = dump_exp_str(iterexp, new_level2);
        res_str = Util.string_append_list({gen_str,"REDUCTION ","\n",expt,itert,""}); then res_str;
    case (_,level) "Graphviz.NODE(\"#UNKNOWN EXPRESSION# ----eeestr \",{},{})"
      equation 
        gen_str = gen_string_n_time("   |", level);
        res_str = Util.string_append_list({gen_str," UNKNOWN EXPRESSION ","\n"}); then res_str;
  end matchcontinue;
end dump_exp_str;

public function solve "rhs 
  solve for "
  input Exp in_exp1;
  input Exp in_exp2;
  input Exp in_exp3;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_exp2,in_exp3)
    local
      Exp crexp,rhs,lhs,res,res_1,cr,e1,e2,e3;
      ComponentRef cr1,cr2;
    case ((crexp = CREF(cr1,_)),rhs,CREF(cr2,_)) "lhs  rhs  solve for "
      equation 
        true = cref_equal(cr1, cr2) "Special case when already solved, otherwise division by zero 
	 when dividing with derivative ";
        false = exp_contains(rhs, crexp); then rhs;
    case (lhs,(crexp = CREF(cr1,_)),CREF(cr2,_))
      equation 
        true = cref_equal(cr1, cr2) "Special case when already solved, otherwise division by zero 
	 when dividing with derivative ";
        false = exp_contains(lhs, crexp); then lhs;
    case (lhs,rhs,(cr = CREF(_,_)))
      equation 
        res = solve2(lhs, rhs, cr);
        res_1 = simplify(res); then res_1;
    case (e1,e2,e3)
      equation 
        Print.print_buf("solve failed: Not an equation or trying to solve for\n");
        Print.print_buf("a non-component-reference, or a non-linear equation:\n");
        print_exp(e1);
        Print.print_buf("\n");
        print_exp(e2);
        Print.print_buf("\n"); then fail();
  end matchcontinue;
end solve;

protected function solve2 " relation: solve2
  This function solves an equation e1 = e2 with respect to the variable
  given as an expression e3 "
  input Exp in_exp1;
  input Exp in_exp2;
  input Exp in_exp3;
  output Exp out_exp;
algorithm 
  out_exp:=
  matchcontinue (in_exp1,in_exp2,in_exp3)
    local
      Exp lhs,lhsder,lhsder_1,lhszero,lhszero_1,rhs,rhs_1,e1,e2,crexp;
      ComponentRef cr;
    case (e1,e2,(crexp = CREF(cr,_))) "e1 e2 e3"
      equation 
        lhs = BINARY(e1,SUB(REAL()),e2);
        lhsder = Derive.differentiate_exp(lhs, cr);
        lhsder_1 = simplify(lhsder);
        false = exp_contains(lhsder_1, crexp);
        (lhszero,_) = replace_exp(lhs, crexp, RCONST(0.0));
        lhszero_1 = simplify(lhszero);
        rhs = UNARY(UMINUS(REAL()),BINARY(lhszero_1,DIV(REAL()),lhsder_1));
        rhs_1 = simplify(rhs) "
	& dump_exp_graphviz lhs => lhsnode 
	& Print.print_buf \"------------------ LHS -----------------\n\"
	& Graphviz.dump lhsnode
	& Print.print_buf \"------------------ /LHS -----------------\n\"
	 "; then rhs_1;
    case (e1,e2,(crexp = CREF(cr,_)))
      equation 
        lhs = BINARY(e1,SUB(REAL()),e2);
        lhsder = Derive.differentiate_exp(lhs, cr);
        lhsder_1 = simplify(lhsder);
        true = exp_contains(lhsder_1, crexp);
        Print.print_buf("solve2 failed: Not linear: ");
        print_exp(e1);
        Print.print_buf(" = ");
        print_exp(e2);
        Print.print_buf("\nsolving for: ");
        print_exp(crexp);
        Print.print_buf("\n");
        Print.print_buf("derivative: ");
        print_exp(lhsder);
        Print.print_buf("\n"); then fail();
    case (e1,e2,(crexp = CREF(cr,_)))
      equation 
        lhs = BINARY(e1,SUB(REAL()),e2);
        lhsder = Derive.differentiate_exp(lhs, cr);
        lhsder_1 = simplify(lhsder);
        Print.print_buf("solve2 failed: ");
        print_exp(e1);
        Print.print_buf(" = ");
        print_exp(e2);
        Print.print_buf("\nsolving for: ");
        print_exp(crexp);
        Print.print_buf("\nDerivative :");
        print_exp(lhsder_1);
        Print.print_buf("\n"); then fail();
  end matchcontinue;
end solve2;

protected function get_terms_containing_x " relation get_terms_containing_x
 
  Retrieves all terms of an expression containng a variable, given
  as second argument (in the form of an Exp)
 "
  input Exp in_exp1;
  input Exp in_exp2;
  output Exp out_exp1;
  output Exp out_exp2;
algorithm 
  (out_exp1,out_exp2):=
  matchcontinue (in_exp1,in_exp2)
    local
      Exp xt1,nonxt1,xt2,nonxt2,xt,nonxt,e1,e2,cr,e;
      Type ty;
      Boolean res;
    case (BINARY(e1,ADD(ty),e2),(cr = CREF(_,_)))
      equation 
        (xt1,nonxt1) = get_terms_containing_x(e1, cr);
        (xt2,nonxt2) = get_terms_containing_x(e2, cr);
        xt = BINARY(xt1,ADD(ty),xt2);
        nonxt = BINARY(nonxt1,ADD(ty),nonxt2); then (xt,nonxt);
    case (BINARY(e1,SUB(ty),e2),(cr = CREF(_,_)))
      equation 
        (xt1,nonxt1) = get_terms_containing_x(e1, cr);
        (xt2,nonxt2) = get_terms_containing_x(e2, cr);
        xt = BINARY(xt1,SUB(ty),xt2);
        nonxt = BINARY(nonxt1,SUB(ty),nonxt2); then (xt,nonxt);
    case (UNARY(UPLUS(ty),e),(cr = CREF(_,_)))
      equation 
        (xt1,nonxt1) = get_terms_containing_x(e, cr);
        xt = UNARY(UPLUS(ty),xt1);
        nonxt = UNARY(UPLUS(ty),nonxt1); then (xt,nonxt);
    case (UNARY(UMINUS(ty),e),(cr = CREF(_,_)))
      equation 
        (xt1,nonxt1) = get_terms_containing_x(e, cr);
        xt = UNARY(UMINUS(ty),xt1);
        nonxt = UNARY(UMINUS(ty),nonxt1); then (xt,nonxt);
    case (e,(cr = CREF(_,_)))
      equation 
        res = exp_contains(e, cr);
        xt = Util.if_(res, e, RCONST(0.0));
        nonxt = Util.if_(res, RCONST(0.0), e); then (xt,nonxt);
    case (e,cr)
      equation 
        Print.print_buf("get_terms_containing_x failed: ");
        print_exp(e);
        Print.print_buf("\nsolving for: ");
        print_exp(cr);
        Print.print_buf("\n"); then fail();
  end matchcontinue;
end get_terms_containing_x;

protected function print_bool
  input Boolean b;
protected 
  Ident str;
algorithm 
  str := Util.if_(b, "true", "false");
  Print.print_buf(str);
end print_bool;

public function exp_contains "Only component references can be checked so far,
   i.e. check whether an expression contains a given component reference "
  input Exp in_exp1;
  input Exp in_exp2;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp1,in_exp2)
    local
      type BooleanList = list<Boolean>;
      Integer i;
      Exp cr,c1,c2,e1,e2,e,c,t,f;
      Boolean res,res1,res2,res3;
      ComponentRef cr1,cr2;
      Operator op;
      BooleanList reslist;
      Absyn.Path fcn;
      ExpList args;
    case (ICONST(i),(cr = CREF(_,_))) then false; 
    case (RCONST(i),(cr = CREF(_,_)))
      local Real i; then false;
    case (SCONST(i),(cr = CREF(_,_)))
      local Ident i; then false;
    case (BCONST(i),(cr = CREF(_,_)))
      local Boolean i; then false;
    case ((c1 = CREF(cr1,_)),(c2 = CREF(cr2,_)))
      equation 
        res = cref_equal(cr1, cr2); then res;
    case (BINARY(e1,op,e2),(cr = CREF(_,_)))
      equation 
        res1 = exp_contains(e1, cr);
        res2 = exp_contains(e2, cr);
        res = bool_or(res1, res2); then res;
    case (UNARY(op,e),(cr = CREF(_,_)))
      equation 
        res = exp_contains(e, cr); then res;
    case (LBINARY(e1,op,e2),(cr = CREF(_,_)))
      equation 
        res1 = exp_contains(e1, cr);
        res2 = exp_contains(e2, cr);
        res = bool_or(res1, res2); then res;
    case (LUNARY(op,e),(cr = CREF(_,_)))
      equation 
        res = exp_contains(e, cr); then res;
    case (RELATION(e1,op,e2),(cr = CREF(_,_)))
      equation 
        res1 = exp_contains(e1, cr);
        res2 = exp_contains(e2, cr);
        res = bool_or(res1, res2); then res;
    case (IFEXP(c,t,f),(cr = CREF(_,_)))
      equation 
        res1 = exp_contains(c, cr);
        res2 = exp_contains(t, cr);
        res3 = exp_contains(f, cr);
        res = Util.bool_or_list({res1,res2,res3}); then res;
    case (CALL(fcn,args,_,_),(cr = CREF(_,_)))
      equation 
        reslist = Util.list_map_1(args, exp_contains, cr);
        res = Util.bool_or_list(reslist); then res;
    case (CAST(REAL(),ICONST(i)),(cr = CREF(_,_))) then false; 
    case (CAST(REAL(),e),(cr = CREF(_,_)))
      equation 
        res = exp_contains(e, cr); then res;
    case (ASUB(e,i),(cr = CREF(_,_)))
      equation 
        res = exp_contains(e, cr); then res;
    case (e,cr)
      equation 
        Print.print_buf("exp_contains failed: When looking for \"");
        print_exp(cr);
        Print.print_buf("\" in \"");
        print_exp(e);
        Print.print_buf("\n"); then fail();
  end matchcontinue;
end exp_contains;

public function get_cref_from_exp
  input Exp in_exp;
  output ComponentRefList out_componentreflist;
protected 
  type ComponentRefList = list<ComponentRef>;
algorithm 
  out_componentreflist:=
  matchcontinue (in_exp)
    local
      type ComponentRefListList = list<ComponentRefList>;
      ComponentRef cr;
      ComponentRefList l1,l2,res,res1,l3,res2;
      Exp e1,e2,e3,e;
      Operator op;
      ExpList farg,expl;
    case (ICONST(_)) then {}; 
    case (RCONST(_)) then {}; 
    case (SCONST(_)) then {}; 
    case (BCONST(_)) then {}; 
    case (CREF(cr,_)) then {cr}; 
    case (BINARY(e1,op,e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res = list_append(l1, l2); then res;
    case (UNARY(op,e1))
      equation 
        res = get_cref_from_exp(e1); then res;
    case (LBINARY(e1,op,e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res = list_append(l1, l2); then res;
    case (LUNARY(op,e1))
      equation 
        res = get_cref_from_exp(e1); then res;
    case (RELATION(e1,op,e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res = list_append(l1, l2); then res;
    case (IFEXP(e1,e2,e3))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res1 = list_append(l1, l2);
        l3 = get_cref_from_exp(e3);
        res = list_append(res1, l3); then res;
    case (CALL(_,farg,_,_))
      local ComponentRefListList res;
      equation 
        res = Util.list_map(farg, get_cref_from_exp);
        res2 = Util.list_flatten(res); then res2;
    case (ARRAY(_,_,expl))
      local ComponentRefListList res1;
      equation 
        res1 = Util.list_map(expl, get_cref_from_exp);
        res = Util.list_flatten(res1); then res;
    case (MATRIX(_,_,_))
      equation 
        print("get_cref_from_exp MATRIX not impl. yet\n"); then {};
    case (RANGE(_,e1,SOME(e3),e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res1 = list_append(l1, l2);
        l3 = get_cref_from_exp(e3);
        res = list_append(res1, l3); then res;
    case (RANGE(_,e1,NONE,e2))
      equation 
        l1 = get_cref_from_exp(e1);
        l2 = get_cref_from_exp(e2);
        res = list_append(l1, l2); then res;
    case (TUPLE(expl))
      equation 
        Print.print_buf("Not implemented yet\n"); then {};
    case (CAST(_,e))
      equation 
        res = get_cref_from_exp(e); then res;
    case (ASUB(e,_))
      equation 
        res = get_cref_from_exp(e); then res;
    case (_) then {}; 
  end matchcontinue;
end get_cref_from_exp;

public function get_function_calls_list " relation: get_function_calls_list
  calls get_function_calls for a list of exps
 "
  input ExpList exps;
  output ExpList res;
protected 
  type ExpListList = list<ExpList>;
  ExpListList explists;
algorithm 
  explists := Util.list_map(exps, get_function_calls);
  res := Util.list_flatten(explists);
end get_function_calls_list;

public function get_function_calls " relation: get_function_calls
  Return all exps that are function calls.
  Inner call exps are returned separately but not extracted from the exp they
  are in, e.g. CALL(foo, {CALL(bar)}) will return
  {CALL(foo, {CALL(bar)}), CALL(bar,{})}
 "
  input Exp in_exp;
  output ExpList out_explist;
algorithm 
  out_explist:=
  matchcontinue (in_exp)
    local
      ExpList argexps,exps,args,a,b,res,elts,elst,elist;
      Exp e,e1,e2,e3;
      Absyn.Path path;
      Boolean tuple_,builtin;
      TupleExpBooleanList flatexplst;
      TupleExpBooleanListList explst;
      ExpOption optexp;
    case ((e = CALL(path,args,tuple_,builtin)))
      equation 
        argexps = get_function_calls_list(args);
        exps = list_append({e}, argexps); then exps;
    case (BINARY(e1,_,e2)) "Binary "
      equation 
        a = get_function_calls(e1);
        b = get_function_calls(e2);
        res = list_append(a, b); then res;
    case (UNARY(_,e)) "Unary "
      equation 
        res = get_function_calls(e); then res;
    case (LBINARY(e1,_,e2)) "LBinary "
      equation 
        a = get_function_calls(e1);
        b = get_function_calls(e2);
        res = list_append(a, b); then res;
    case (LUNARY(_,e)) "LUnary "
      equation 
        res = get_function_calls(e); then res;
    case (RELATION(e1,_,e2)) "Relation "
      equation 
        a = get_function_calls(e1);
        b = get_function_calls(e2);
        res = list_append(a, b); then res;
    case (IFEXP(e1,e2,e3))
      equation 
        res = get_function_calls_list({e1,e2,e3}); then res;
    case (ARRAY(_,_,elts)) "Array "
      equation 
        res = get_function_calls_list(elts); then res;
    case (MATRIX(_,_,explst)) "Matrix "
      equation 
        flatexplst = Util.list_flatten(explst);
        elst = Util.list_map(flatexplst, Util.tuple2_1);
        res = get_function_calls_list(elst); then res;
    case (RANGE(_,e1,optexp,e2)) "Range "
      local ExpList e3;
      equation 
        e3 = Util.option_to_list(optexp);
        elist = list_append({e1,e2}, e3);
        res = get_function_calls_list(elist); then res;
    case (TUPLE(exps)) "Tuple "
      equation 
        res = get_function_calls_list(exps); then res;
    case (CAST(_,e))
      equation 
        res = get_function_calls(e); then res;
    case (SIZE(e1,e2)) "Size "
      local ExpOption e2;
      equation 
        a = Util.option_to_list(e2);
        elist = list_append(a, {e1});
        res = get_function_calls_list(elist); then res;
    case (_) then {}; 
  end matchcontinue;
end get_function_calls;
end Exp;

