package Graphviz "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Graphviz.rml
  module:      Graphviz
  description: Graphviz is a tool for drawing graphs from a textual
  representation. This module generates the textual input to graphviz from a
  tree defined using the data structures defined here, e.g. Node for tree
  nodes. See http://www.research.att.com/sw/tools/graphviz/ .
 
  Input: The tree constructed from data structures in Graphviz
  Output: Textual input to graphviz, written to stdout.
 
  RCS: $Id: Graphviz.rml 1733 2005-04-27 06:47:45Z haklu $
 
 "

public 
type Type = String;

public 
type Ident = String;

public 
type Label = String;

public 
type LabelList = list<Label>;

uniontype Node
  record NODE
    Type type_;
    Attributes attributes;
    Children children;
  end NODE;
  record LNODE
    Type type_;
    LabelList labellist;
    Attributes attributes;
    Children children;
  end LNODE;
end Node;

public 
type Children = list<Node>;

public 
type Attributes = list<Attribute>;

public 
uniontype Attribute
  record ATTR
    String name "name ";
    String value "value ";
  end ATTR;
end Attribute;

public constant Attribute box=ATTR("shape","box");

public function dump "Relations "
  input Node node;
protected 
  Label nm;
algorithm 
  print("graph AST {\n");
  nm := dump_node(node);
  print("}\n");
end dump;

protected function dump_node
  input Node in_node;
  output Ident out_ident;
algorithm 
  out_ident:=
  matchcontinue (in_node)
    local
      Label nm,typlbl,out,typ,lblstr;
      Attributes newattr,attr;
      Children children;
      LabelList lbl_1,lbl;
    case (NODE(typ,attr,children))
      equation 
        nm = nodename(typ);
        typlbl = make_label({typ});
        newattr = list_append({ATTR("label",typlbl)}, attr);
        out = make_node(nm, newattr);
        print(out);
        dump_children(nm, children); then nm;
    case (LNODE(typ,lbl,attr,children))
      equation 
        nm = nodename(typ);
        lbl_1 = list_append({typ}, lbl);
        lblstr = make_label(lbl_1);
        newattr = list_append({ATTR("label",lblstr)}, attr);
        out = make_node(nm, newattr);
        print(out);
        dump_children(nm, children); then nm;
  end matchcontinue;
end dump_node;

protected function make_label
  input StringList sl;
  output String s2;
protected 
  type StringList = list<String>;
  Label s0,s1;
algorithm 
  s0 := make_label_req(sl);
  s1 := string_append("\"", s0);
  s2 := string_append(s1, "\"");
end make_label;

protected function make_label_req
  input StringList in_stringlist;
  output String out_string;
protected 
  type StringList = list<String>;
algorithm 
  out_string:=
  matchcontinue (in_stringlist)
    local
      Label s,res,s1,s2,old;
      LabelList rest;
    case {s} then s; 
    case {s1,s2}
      equation 
        s = string_append(s1, "\\n");
        res = string_append(s, s2); then res;
    case s1 :: rest
      equation 
        old = make_label_req(rest);
        s = string_append(s1, "\\n");
        res = string_append(s, old); then res;
  end matchcontinue;
end make_label_req;

protected function dump_children
  input Ident in_ident;
  input Children in_children;
algorithm 
  _:=
  matchcontinue (in_ident,in_children)
    local
      Label nm,parent;
      Node node;
      Children rest;
    case (_,{}) then (); 
    case (parent,node :: rest)
      equation 
        nm = dump_node(node);
        print_edge(nm, parent);
        dump_children(parent, rest); then ();
  end matchcontinue;
end dump_children;

protected function nodename "changed use of str as part of nodename, since it may contain spaces
"
  input String str;
  output String s;
protected 
  Integer i;
  Label is;
algorithm 
  i := tick();
  is := int_string(i);
  s := string_append("GVNOD", is);
end nodename;

protected function print_edge
  input Ident n1;
  input Ident n2;
protected 
  Label str;
algorithm 
  str := make_edge(n1, n2);
  print(str);
  print(";\n");
end print_edge;

protected function make_edge
  input Ident n1;
  input Ident n2;
  output String str;
protected 
  Label s;
algorithm 
  s := string_append(n1, " -- ");
  str := string_append(s, n2);
end make_edge;

protected function make_node
  input Ident nm;
  input Attributes attr;
  output String str;
protected 
  Label s,s_1;
algorithm 
  s := make_attr(attr);
  s_1 := string_append(nm, s);
  str := string_append(s_1, ";");
end make_node;

protected function make_attr
  input AttributeList l;
  output String str;
protected 
  type AttributeList = list<Attribute>;
  Label res,s;
algorithm 
  res := make_attr_req(l);
  s := string_append("[", res);
  str := string_append(s, "]");
end make_attr;

protected function make_attr_req
  input AttributeList in_attributelist;
  output String out_string;
protected 
  type AttributeList = list<Attribute>;
algorithm 
  out_string:=
  matchcontinue (in_attributelist)
    local
      Label s,str,name,v,old,s_1,s_2;
      AttributeList rest;
    case {ATTR(name,v)}
      equation 
        s = string_append(name, "=");
        str = string_append(s, v); then str;
    case (ATTR(name,v) :: rest)
      equation 
        old = make_attr_req(rest);
        s = string_append(name, "=");
        s_1 = string_append(s, v);
        s_2 = string_append(s_1, ",");
        str = string_append(s_2, old); then str;
  end matchcontinue;
end make_attr_req;
end Graphviz;

