package Lookup "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Lookup.rml
  module:      Lookup
  description: Scoping rules
 
  RCS: $Id: Lookup.rml 1835 2005-06-27 13:42:20Z petar $
 
 
  This module is responsible for the lookup mechanism in Modelica.
  It is responsible for looking up classes, variables, etc. in the
  environment 'Env' by following the lookup rules.
  The most important functions are:
  lookup_class - to find a class 
  lookup_type - to find types (e.g. functions, types, etc.)
  lookup_var - to find a variable in the instance hierarchy.
 "

public import OpenModelica.Compiler.ClassInf;

public import OpenModelica.Compiler.Types;

public import OpenModelica.Compiler.Absyn;

public import OpenModelica.Compiler.Exp;

public import OpenModelica.Compiler.Env;

public import OpenModelica.Compiler.SCode;

protected import OpenModelica.Compiler.Parser;

protected import OpenModelica.Compiler.Debug;

protected import OpenModelica.Compiler.Dump;

protected import OpenModelica.Compiler.Inst;

protected import OpenModelica.Compiler.Mod;

protected import OpenModelica.Compiler.Prefix;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Builtin;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.ModUtil;

protected import OpenModelica.Compiler.Static;

public function lookup_type
  input Env.Env in_env;
  input Absyn.Path in_path;
  input Boolean in_boolean;
  output Types.Type out_type;
  output Env.Env out_env;
algorithm 
  (out_type,out_env):=
  matchcontinue (in_env,in_path,in_boolean)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      TupleTypes_TTypeAbsyn_PathOption t,c_1;
      Env_FrameList env_1,env,env_2,env3,env2,env_3;
      Absyn.Path path,p;
      Boolean msg,encflag;
      SCode.Class c;
      String id,pack,s1,s2;
      SCode.Restriction restr;
      ClassInf.State ci_state,cistate1;
    case (env,(path = Absyn.IDENT(_)),msg) "msg flag  Lookup of simple names "
      equation 
        (t,env_1) = lookup_type_in_env(env, path) "Debug.fprint (\"lotype\",\"lookup_type(\") &
	Debug.fcall (\"lotype\",Env.print_env, env) &
	Debug.fprint (\"lotype\",\", \") & 
	Debug.fcall (\"lotype\",Dump.print_path,path) &
	Debug.fprint (\"lottype\",\")\n\") &"; then (t,env_1);
    case (env,(path = Absyn.IDENT(_)),msg)
      equation 
        (c = SCode.CLASS(id,_,encflag,SCode.R_FUNCTION(),_),env_1) = lookup_class(env, path, false) "If we didn't find the type, but found a class definition 
	  * that is a function with the same name then we implicitly instantiate that
	  function, and try again. ";
        env_2 = Inst.implicit_function_type_instantiation(env_1, c);
        (t,env3) = lookup_type_in_env(env_2, path); then (t,env3);
    case (env,(path = Absyn.IDENT(_)),msg)
      equation 
        (c = SCode.CLASS(id,_,encflag,SCode.R_EXT_FUNCTION(),_),env_1) = lookup_class(env, path, msg) "If we didn't find the type, but found a class definition 
	  * that is an external function with the same name then we implicitly 
	  * instantiate that function, and try again. ";
        env_2 = Inst.implicit_function_type_instantiation(env_1, c);
        (t,env3) = lookup_type_in_env(env_2, path); then (t,env3);
    case (env,Absyn.QUALIFIED(pack,path),msg) "Lookup of qualified name "
      equation 
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class(env, Absyn.IDENT(pack), false);
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (_,env_2,_,cistate1,_) = Inst.inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {}, true, false) "Instantiate implicit (last argument = true)  true";
        failure(ClassInf.valid(cistate1, SCode.R_PACKAGE()));
        (t,env_3) = lookup_type_in_class(env_2, c, path, true) "Has to do additional check for encapsulated classes, see rule below "; then (t,env_3);
    case (env,(p = Absyn.QUALIFIED(pack,path)),msg)
      equation 
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class(env, Absyn.IDENT(pack), msg);
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (_,env_2,_,cistate1,_) = Inst.inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {}, true, false) "true";
        ClassInf.valid(cistate1, SCode.R_PACKAGE());
        (c_1,env_3) = lookup_type_in_class(env_2, c, path, false) "Has NOT to do additional check for encapsulated classes, see rule above "; then (c_1,env_3);
    case (_,path,true)
      equation 
        s1 = Absyn.path_string(path);
        s2 = Util.string_append_list({"# Error, type ",s1," not found.\n"});
        Print.print_error_buf(s2); then fail();
  end matchcontinue;
end lookup_type;

public function lookup_class " relation: lookup_class
  
  Tries to find a specified class in an environment
  
  Arg1: The enviroment where to look
  Arg2: The path for the class
  Arg3: A Bool to control the output of error-messages. If it is true
        then it outputs a error message if the class is not found.
 "
  input Env.Env in_env;
  input Absyn.Path in_path;
  input Boolean in_boolean;
  output SCode.Class out_class;
  output Env.Env out_env;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_env,in_path,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      SCode.Class c,c_1;
      String id,name,pack;
      Boolean encflag,msgflag;
      SCode.Restriction restr;
      Env_FrameList env_1,env,env2,env_2,env_3,env1,env4,env5;
      Absyn.Path path,ep,packp,p;
      ClassInf.State ci_state,cistate1;
    case (env,(path = Absyn.IDENT(name)),msgflag)
      equation 
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class_in_env(env, path, msgflag) "print \"searching for \" & print name & print \"\n\" &  Env.print_env env & Print.get_string => str & print str & "; then (c,env_1);
    case (env,(p = Absyn.QUALIFIED(_,_)),msgflag)
      equation 
        SOME(ep) = Env.get_env_path(env) "If we search for A1.A2....An.x while in scope A1.A2...An
	 , just search for x. Must do like this to ensure finite recursion ";
        packp = Absyn.strip_last(p);
        true = ModUtil.path_equal(ep, packp);
        id = Absyn.path_last_ident(p);
        (c,env_1) = lookup_class(env, Absyn.IDENT(id), msgflag); then (c,env_1);
    case (env,(p = Absyn.QUALIFIED(pack,path)),msgflag)
      equation 
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class(env, Absyn.IDENT(pack), msgflag);
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (env_2,cistate1) = Inst.partial_inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {});
        failure(ClassInf.valid(cistate1, SCode.R_PACKAGE()));
        (c_1,env_3) = lookup_class(env_2, path, msgflag) "Has to do additional check for encapsulated classes, see rule below "; then (c_1,env_3);
    case (env,(p = Absyn.QUALIFIED(pack,path)),msgflag)
      equation 
        (c = SCode.CLASS(id,_,encflag,restr,_),env1) = lookup_class(env, Absyn.IDENT(pack), msgflag);
        env2 = Env.open_scope(env1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (env4,cistate1) = Inst.partial_inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {});
        ClassInf.valid(cistate1, SCode.R_PACKAGE());
        (c_1,env5) = lookup_class(env4, path, msgflag) "Has NOT to do additional check for encapsulated classes, see rule above "; then (c_1,env5);
    case (env,path,true) then fail(); 
  end matchcontinue;
end lookup_class;

protected function lookup_qualified_imported_class_in_env
  input Env.Env in_env1;
  input Env.Env in_env2;
  input Absyn.Path in_path3;
  output SCode.Class out_class;
  output Env.Env out_env;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_env1,in_env2,in_path3)
    local
      type Env_FrameList = list<Env.Frame>;
      type StringOption = Option<String>;
      type Env_ItemList = list<Env.Item>;
      SCode.Class c,c_1;
      Env_FrameList env_1,env,fs,totenv,env2,env4,env_2;
      StringOption sid;
      Env_ItemList items,imps;
      String name,id,pack;
      Boolean encflag;
      SCode.Restriction restr;
      ClassInf.State ci_state;
      Absyn.Path path;
      Env.Frame f;
    case ((env = Env.FRAME(sid,_,_,items,_) :: fs),totenv,Absyn.IDENT(name))
      equation 
        (c,env_1) = lookup_qualified_imported_class_in_frame(items, totenv, name); then (c,env_1);
    case ((env = Env.FRAME(sid,_,_,imps,_) :: fs),totenv,Absyn.QUALIFIED(pack,path))
      equation 
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_qualified_imported_class_in_frame(imps, totenv, pack);
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (env4,_) = Inst.partial_inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {});
        (c_1,env_2) = lookup_class(env4, path, false); then (c_1,env_2);
    case (f :: fs,env,id)
      local Absyn.Path id;
      equation 
        (c,env_1) = lookup_qualified_imported_class_in_env(fs, env, id); then (c,env_1);
  end matchcontinue;
end lookup_qualified_imported_class_in_env;

protected function lookup_qualified_imported_var_in_frame
  input Env_ItemList in_env_itemlist;
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output Types.Attributes out_attributes;
  output Types.Type out_type;
  output Types.Binding out_binding;
protected 
  type Env_ItemList = list<Env.Item>;
algorithm 
  (out_attributes,out_type,out_binding):=
  matchcontinue (in_env_itemlist,in_env,in_ident)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      Env.Frame fr;
      Types.Attributes attr;
      TupleTypes_TTypeAbsyn_PathOption ty;
      Types.Binding bind;
      String id,ident,str;
      Env_ItemList fs;
      Env_FrameList env;
      Exp.ComponentRef cref;
      Absyn.Path strippath,path;
      SCode.Class c2;
    case (Env.IMPORT(Absyn.QUAL_IMPORT(Absyn.IDENT(id))) :: fs,env,ident)
      equation 
        equality(id = ident) "For imported paths A, not possible to assert sub-path package";
        fr = Env.top_frame(env);
        (attr,ty,bind) = lookup_var({fr}, Exp.CREF_IDENT(ident,{})); then (attr,ty,bind);
    case (Env.IMPORT(Absyn.QUAL_IMPORT(path)) :: fs,env,ident)
      equation 
        id = Absyn.path_last_ident(path) "For imported path A.B.C, assert A.B is package";
        equality(id = ident);
        fr = Env.top_frame(env);
        cref = Exp.path_to_cref(path);
        (attr,ty,bind) = lookup_var_in_packages({fr}, cref);
        strippath = Absyn.strip_last(path);
        (c2,_) = lookup_class({fr}, strippath, true);
        assert_package(c2); then (attr,ty,bind);
    case (Env.IMPORT(Absyn.QUAL_IMPORT(path)) :: fs,env,ident)
      equation 
        id = Absyn.path_last_ident(path) "If not package, error ";
        equality(id = ident);
        fr = Env.top_frame(env);
        cref = Exp.path_to_cref(path);
        (attr,ty,bind) = lookup_var_in_packages({fr}, cref);
        strippath = Absyn.strip_last(path);
        (c2,_) = lookup_class({fr}, strippath, true);
        failure(assert_package(c2));
        Print.print_error_buf("# Error, ");
        str = Absyn.path_string(strippath);
        Print.print_error_buf(str);
        Print.print_error_buf(" is not a package\n"); then fail();
    case (Env.IMPORT(Absyn.NAMED_IMPORT(id,path)) :: fs,env,ident)
      equation 
        equality(id = ident) "Named imports";
        fr = Env.top_frame(env);
        cref = Exp.path_to_cref(path);
        (attr,ty,bind) = lookup_var_in_packages({fr}, cref);
        strippath = Absyn.strip_last(path);
        (c2,_) = lookup_class({fr}, strippath, true);
        assert_package(c2); then (attr,ty,bind);
    case (Env.IMPORT(Absyn.NAMED_IMPORT(id,path)) :: fs,env,ident)
      equation 
        equality(id = ident) "Assert package for Named imports ";
        fr = Env.top_frame(env);
        cref = Exp.path_to_cref(path);
        (attr,ty,bind) = lookup_var_in_packages({fr}, cref);
        strippath = Absyn.strip_last(path);
        (c2,_) = lookup_class({fr}, strippath, true);
        failure(assert_package(c2));
        Print.print_error_buf("# Error, ");
        str = Absyn.path_string(strippath);
        Print.print_error_buf(str);
        Print.print_error_buf(" is not a package\n"); then fail();
    case (_ :: fs,env,ident)
      equation 
        (attr,ty,bind) = lookup_qualified_imported_var_in_frame(fs, env, ident); then (attr,ty,bind);
  end matchcontinue;
end lookup_qualified_imported_var_in_frame;

protected function more_lookup_unqualified_imported_var_in_frame
  input Env_ItemList in_env_itemlist;
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output Boolean out_boolean;
protected 
  type Env_ItemList = list<Env.Item>;
algorithm 
  out_boolean:=
  matchcontinue (in_env_itemlist,in_env,in_ident)
    local
      type Env_FrameList = list<Env.Frame>;
      Env.Frame fr;
      SCode.Class c;
      String id,ident;
      Boolean encflag,res;
      SCode.Restriction restr;
      Env_FrameList env_1,env2,env_2,env;
      ClassInf.State ci_state;
      Absyn.Path path;
      Env_ItemList fs;
    case (Env.IMPORT(Absyn.UNQUAL_IMPORT(path)) :: fs,env,ident)
      equation 
        fr = Env.top_frame(env);
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class({fr}, path, false);
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (env_2,_) = Inst.partial_inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {});
        (_,_,_) = lookup_var_in_packages(env_2, Exp.CREF_IDENT(ident,{})); then true;
    case (_ :: fs,env,ident)
      equation 
        res = more_lookup_unqualified_imported_var_in_frame(fs, env, ident); then res;
    case ({},_,_) then false; 
  end matchcontinue;
end more_lookup_unqualified_imported_var_in_frame;

protected function lookup_unqualified_imported_var_in_frame "unique"
  input Env_ItemList in_env_itemlist;
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output Types.Attributes out_attributes;
  output Types.Type out_type;
  output Types.Binding out_binding;
  output Boolean out_boolean;
protected 
  type Env_ItemList = list<Env.Item>;
algorithm 
  (out_attributes,out_type,out_binding,out_boolean):=
  matchcontinue (in_env_itemlist,in_env,in_ident)
    local
      type Env_FrameList = list<Env.Frame>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Env.Frame fr;
      Exp.ComponentRef cref;
      SCode.Class c;
      String id,ident;
      Boolean encflag,more,unique;
      SCode.Restriction restr;
      Env_FrameList env_1,env2,env_2,env;
      ClassInf.State ci_state;
      Types.Attributes attr;
      TupleTypes_TTypeAbsyn_PathOption ty;
      Types.Binding bind;
      Absyn.Path path;
      Env_ItemList fs;
    case (Env.IMPORT(Absyn.UNQUAL_IMPORT(path)) :: fs,env,ident) "unique"
      equation 
        fr = Env.top_frame(env);
        cref = Exp.path_to_cref(path);
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class({fr}, path, false);
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (_,env_2,_,_,_) = Inst.inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {}, false, false);
        (attr,ty,bind) = lookup_var_in_packages(env_2, Exp.CREF_IDENT(ident,{}));
        more = more_lookup_unqualified_imported_var_in_frame(fs, env, ident);
        unique = bool_not(more); then (attr,ty,bind,unique);
    case (_ :: fs,env,ident)
      equation 
        (attr,ty,bind,unique) = lookup_unqualified_imported_var_in_frame(fs, env, ident); then (attr,ty,bind,unique);
  end matchcontinue;
end lookup_unqualified_imported_var_in_frame;

protected function lookup_qualified_imported_class_in_frame
  input Env_ItemList in_env_itemlist;
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output SCode.Class out_class;
  output Env.Env out_env;
protected 
  type Env_ItemList = list<Env.Item>;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_env_itemlist,in_env,in_ident)
    local
      type Env_FrameList = list<Env.Frame>;
      Env.Frame fr;
      SCode.Class c,c2;
      Env_FrameList env_1,env;
      String id,ident,str;
      Env_ItemList fs;
      Absyn.Path strippath,path;
    case (Env.IMPORT(Absyn.QUAL_IMPORT(Absyn.IDENT(id))) :: fs,env,ident)
      equation 
        equality(id = ident) "For imported paths A, not possible to assert sub-path package";
        fr = Env.top_frame(env);
        (c,env_1) = lookup_class({fr}, Absyn.IDENT(id), true); then (c,env_1);
    case (Env.IMPORT(Absyn.QUAL_IMPORT(path)) :: fs,env,ident)
      equation 
        id = Absyn.path_last_ident(path) "For imported path A.B.C, assert A.B is package";
        equality(id = ident);
        fr = Env.top_frame(env);
        (c,env_1) = lookup_class({fr}, path, true);
        strippath = Absyn.strip_last(path);
        (c2,_) = lookup_class({fr}, strippath, true);
        assert_package(c2); then (c,env_1);
    case (Env.IMPORT(Absyn.QUAL_IMPORT(path)) :: fs,env,ident)
      equation 
        id = Absyn.path_last_ident(path) "If not package, error ";
        equality(id = ident);
        fr = Env.top_frame(env);
        (c,env_1) = lookup_class({fr}, path, true);
        strippath = Absyn.strip_last(path);
        (c2,_) = lookup_class({fr}, strippath, true);
        failure(assert_package(c2));
        Print.print_error_buf("# Error, ");
        str = Absyn.path_string(strippath);
        Print.print_error_buf(str);
        Print.print_error_buf(" is not a package\n"); then fail();
    case (Env.IMPORT(Absyn.NAMED_IMPORT(id,path)) :: fs,env,ident)
      equation 
        equality(id = ident) "Named imports";
        fr = Env.top_frame(env);
        (c,env_1) = lookup_class({fr}, path, true) "	Print.print_buf \"NAMED IMPORT, top frame:\" & 
	Env.print_env {fr} &";
        strippath = Absyn.strip_last(path);
        (c2,_) = lookup_class({fr}, strippath, true);
        assert_package(c2); then (c,env_1);
    case (Env.IMPORT(Absyn.NAMED_IMPORT(id,path)) :: fs,env,ident)
      equation 
        equality(id = ident) "Assert package for Named imports ";
        fr = Env.top_frame(env);
        (c,env_1) = lookup_class({fr}, path, true);
        strippath = Absyn.strip_last(path);
        (c2,_) = lookup_class({fr}, strippath, true);
        failure(assert_package(c2));
        Print.print_error_buf("# Error, ");
        str = Absyn.path_string(strippath);
        Print.print_error_buf(str);
        Print.print_error_buf(" is not a package\n"); then fail();
    case (_ :: fs,env,ident)
      equation 
        (c,env_1) = lookup_qualified_imported_class_in_frame(fs, env, ident); then (c,env_1);
  end matchcontinue;
end lookup_qualified_imported_class_in_frame;

protected function more_lookup_unqualified_imported_class_in_frame
  input Env_ItemList in_env_itemlist;
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output Boolean out_boolean;
protected 
  type Env_ItemList = list<Env.Item>;
algorithm 
  out_boolean:=
  matchcontinue (in_env_itemlist,in_env,in_ident)
    local
      type Env_FrameList = list<Env.Frame>;
      Env.Frame fr;
      SCode.Class c;
      String id,ident;
      Boolean encflag,res;
      SCode.Restriction restr;
      Env_FrameList env_1,env2,env_2,env;
      ClassInf.State ci_state;
      Absyn.Path path;
      Env_ItemList fs;
    case (Env.IMPORT(Absyn.UNQUAL_IMPORT(path)) :: fs,env,ident)
      equation 
        fr = Env.top_frame(env);
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class({fr}, path, false);
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (env_2,_) = Inst.partial_inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {});
        (_,_) = lookup_class(env_2, Absyn.IDENT(ident), false); then true;
    case (_ :: fs,env,ident)
      equation 
        res = more_lookup_unqualified_imported_class_in_frame(fs, env, ident); then res;
    case ({},_,_) then false; 
  end matchcontinue;
end more_lookup_unqualified_imported_class_in_frame;

protected function lookup_unqualified_imported_class_in_frame "unique"
  input Env_ItemList in_env_itemlist;
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output SCode.Class out_class;
  output Env.Env out_env;
  output Boolean out_boolean;
protected 
  type Env_ItemList = list<Env.Item>;
algorithm 
  (out_class,out_env,out_boolean):=
  matchcontinue (in_env_itemlist,in_env,in_ident)
    local
      type Env_FrameList = list<Env.Frame>;
      Env.Frame fr;
      SCode.Class c,c_1;
      String id,ident;
      Boolean encflag,more,unique;
      SCode.Restriction restr;
      Env_FrameList env_1,env2,env_2,env_3,env;
      ClassInf.State ci_state,cistate1;
      Absyn.Path path;
      Env_ItemList fs;
    case (Env.IMPORT(Absyn.UNQUAL_IMPORT(path)) :: fs,env,ident) "unique"
      equation 
        fr = Env.top_frame(env);
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class({fr}, path, false);
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (env_2,cistate1) = Inst.partial_inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {});
        (c_1,env_3) = lookup_class(env_2, Absyn.IDENT(ident), false);
        more = more_lookup_unqualified_imported_class_in_frame(fs, env, ident);
        unique = bool_not(more); then (c_1,env_3,unique);
    case (_ :: fs,env,ident)
      equation 
        (c,env_1,unique) = lookup_unqualified_imported_class_in_frame(fs, env, ident); then (c,env_1,unique);
  end matchcontinue;
end lookup_unqualified_imported_class_in_frame;

public function lookup_record_constructor_class " relation: lookup_record_constructor_class
  
  
  
 "
  input Env.Env in_env;
  input Absyn.Path in_path;
  output SCode.Class out_class;
  output Env.Env out_env;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_env,in_path)
    local
      type Env_FrameList = list<Env.Frame>;
      SCode.Class c;
      Env_FrameList env_1,env;
      Absyn.Path path;
    case (env,path)
      equation 
        (c,env_1) = lookup_recconst_in_env(env, path); then (c,env_1);
  end matchcontinue;
end lookup_record_constructor_class;

public function complete_path " relation: complete_path
 
  This function takes a type name and an env and looks up the class.
  Then it determines the full path for the type, such that it can be looked up
  from any environment.
 "
  input Env.Env in_env;
  input Absyn.Path in_path;
  output Absyn.Path out_path;
algorithm 
  out_path:=
  matchcontinue (in_env,in_path)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList env,env_1;
      Absyn.Path path,path1,path_1;
      String id,pack;
    case (env,path)
      equation 
        (_,Env.FRAME(NONE,_,_,_,_) :: _) = lookup_class(env, path, true) "Class found on top level. Nothing to complete. "; then path;
    case (env,(path = Absyn.IDENT(_)))
      equation 
        (SCode.CLASS(id,_,_,_,_),env_1) = lookup_class(env, path, true);
        SOME(path1) = Env.get_env_path(env_1);
        path_1 = Absyn.join_paths(path1, path) "	& Print.print_buf \"completed to \" &
	Absyn.path_string path' => s1 & Print.print_buf s1 &
	Print.print_buf \" class = \" & Print.print_buf id & 
	Print.print_buf \"\nenv =\" &
	Env.print_env env'  "; then path_1;
    case (env,(path = Absyn.QUALIFIED(pack,_)))
      equation 
        (_,Env.FRAME(NONE,_,_,_,_) :: _) = lookup_class(env, Absyn.IDENT(pack), true); then path;
    case (env,(path = Absyn.QUALIFIED(pack,_)))
      equation 
        (SCode.CLASS(id,_,_,_,_),env_1) = lookup_class(env, path, true) "Absyn.IDENT(pack)";
        SOME(path1) = Env.get_env_path(env_1);
        path_1 = Absyn.join_paths(path1, Absyn.IDENT(id)) "	& Print.print_buf \"Qualified completed to \" 
	& Absyn.path_string path' => s1 & Print.print_buf s1 &
	Print.print_buf \" class = \" & Print.print_buf id & 
	Print.print_buf \"\n\" &
	Print.print_buf \"\nenv =\" &
	Env.print_env env'  "; then path_1;
    case (env,path) then fail(); 
  end matchcontinue;
end complete_path;

public function lookup_var
  input Env.Env in_env;
  input Exp.ComponentRef in_componentref;
  output Types.Attributes out_attributes;
  output Types.Type out_type;
  output Types.Binding out_binding;
algorithm 
  (out_attributes,out_type,out_binding):=
  matchcontinue (in_env,in_componentref)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      Types.Attributes attr;
      TupleTypes_TTypeAbsyn_PathOption ty;
      Types.Binding binding;
      Env_FrameList env;
      Exp.ComponentRef cref;
    case (env,cref) "try the old lookup_var "
      equation 
        (attr,ty,binding) = lookup_var_internal(env, cref); then (attr,ty,binding);
    case (env,cref) "then look in classes (implicitly instantiated packages)
	 "
      equation 
        (attr,ty,binding) = lookup_var_in_packages(env, cref); then (attr,ty,binding);
    case (_,_) then fail(); 
  end matchcontinue;
end lookup_var;

protected function lookup_var_internal
  input Env.Env in_env;
  input Exp.ComponentRef in_componentref;
  output Types.Attributes out_attributes;
  output Types.Type out_type;
  output Types.Binding out_binding;
algorithm 
  (out_attributes,out_type,out_binding):=
  matchcontinue (in_env,in_componentref)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type StringOption = Option<String>;
      type Env_ItemList = list<Env.Item>;
      type Env_FrameList = list<Env.Frame>;
      Types.Attributes attr;
      TupleTypes_TTypeAbsyn_PathOption ty;
      Types.Binding binding;
      StringOption sid;
      Env.BinTree ht;
      Env_ItemList imps;
      Env_FrameList fs;
      Exp.ComponentRef ref;
    case (Env.FRAME(sid,ht,_,imps,_) :: fs,ref)
      equation 
        (attr,ty,binding) = lookup_var_f(ht, ref); then (attr,ty,binding);
    case (_ :: fs,ref)
      equation 
        (attr,ty,binding) = lookup_var_internal(fs, ref); then (attr,ty,binding);
  end matchcontinue;
end lookup_var_internal;

protected function lookup_var_in_packages
  input Env.Env in_env;
  input Exp.ComponentRef in_componentref;
  output Types.Attributes out_attributes;
  output Types.Type out_type;
  output Types.Binding out_binding;
algorithm 
  (out_attributes,out_type,out_binding):=
  matchcontinue (in_env,in_componentref)
    local
      type Env_FrameList = list<Env.Frame>;
      type Types_VarList = list<Types.Var>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Exp_SubscriptList = list<Exp.Subscript>;
      type StringOption = Option<String>;
      type Env_ItemList = list<Env.Item>;
      SCode.Class c;
      String n,id1,id;
      Boolean encflag;
      SCode.Restriction r;
      Env_FrameList env2,env3,env5,env,fs;
      ClassInf.State ci_state;
      Types_VarList types;
      Types.Attributes attr;
      TupleTypes_TTypeAbsyn_PathOption ty;
      Types.Binding bind;
      Exp.ComponentRef id2,cref,cr;
      Exp_SubscriptList sb;
      StringOption sid;
      Env_ItemList items;
      Env.Frame f;
    case (env,Exp.CREF_QUAL(id1,{},(id2 = Exp.CREF_IDENT(_,_))))
      equation 
        (c = SCode.CLASS(n,_,encflag,r = SCode.R_ENUMERATION(),_),env2) = lookup_class(env, Absyn.IDENT(id1), false) "Special case for looking up enumerations ";
        env3 = Env.open_scope(env2, encflag, SOME(n));
        ci_state = ClassInf.start(r, n);
        (_,env5,_,_,types) = Inst.inst_class_in(env3, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {}, false, false);
        (attr,ty,bind) = lookup_var_in_packages(env5, id2); then (attr,ty,bind);
    case (env,Exp.CREF_QUAL(id,{},cref))
      equation 
        (c = SCode.CLASS(n,_,encflag,r,_),env2) = lookup_class(env, Absyn.IDENT(id), false);
        env3 = Env.open_scope(env2, encflag, SOME(n));
        ci_state = ClassInf.start(r, n);
        (_,env5,_,_,types) = Inst.inst_class_in(env3, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {}, true, false) "true";
        (attr,ty,bind) = lookup_var_in_packages(env5, cref); then (attr,ty,bind);
    case (env,(cr = Exp.CREF_IDENT(id,sb)))
      equation 
        (attr,ty,bind) = lookup_var_local(env, cr); then (attr,ty,bind);
    case ((env = Env.FRAME(sid,_,_,items,_) :: fs),(cr = Exp.CREF_IDENT(id,sb)))
      equation 
        (attr,ty,bind) = lookup_qualified_imported_var_in_frame(items, env, id); then (attr,ty,bind);
    case ((env = Env.FRAME(sid,_,_,items,_) :: fs),(cr = Exp.CREF_IDENT(id,sb)))
      equation 
        (attr,ty,bind,true) = lookup_unqualified_imported_var_in_frame(items, env, id); then (attr,ty,bind);
    case ((env = Env.FRAME(sid,_,_,items,_) :: fs),(cr = Exp.CREF_IDENT(id,sb)))
      equation 
        (attr,ty,bind,false) = lookup_unqualified_imported_var_in_frame(items, env, id);
        Print.print_error_buf(
          "#Error, constant found in several unqualified import statements\n"); then fail();
    case (f :: fs,(cr = Exp.CREF_IDENT(id,sb))) "to find imports in previous frames, search next frame."
      equation 
        (attr,ty,bind) = lookup_var_in_packages(fs, cr); then (attr,ty,bind);
    case (env,(cr = Exp.CREF_IDENT(id,sb)))
      equation 
        failure((attr,ty,bind) = lookup_var_local(env, cr)); then fail();
    case (env,cr) then fail(); 
  end matchcontinue;
end lookup_var_in_packages;

public function lookup_var_local
  input Env.Env in_env;
  input Exp.ComponentRef in_componentref;
  output Types.Attributes out_attributes;
  output Types.Type out_type;
  output Types.Binding out_binding;
algorithm 
  (out_attributes,out_type,out_binding):=
  matchcontinue (in_env,in_componentref)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type StringOption = Option<String>;
      type Env_FrameList = list<Env.Frame>;
      Types.Attributes attr;
      TupleTypes_TTypeAbsyn_PathOption ty;
      Types.Binding binding;
      StringOption sid;
      Env.BinTree ht;
      Env_FrameList fs,env;
      Exp.ComponentRef ref;
    case (Env.FRAME(sid,ht,_,_,_) :: fs,ref)
      equation 
        (attr,ty,binding) = lookup_var_f(ht, ref); then (attr,ty,binding);
    case (Env.FRAME(SOME("$for loop scope$"),_,_,_,_) :: env,ref)
      equation 
        (attr,ty,binding) = lookup_var_local(env, ref) "Exception, when in for loop scope allow search of next scope "; then (attr,ty,binding);
  end matchcontinue;
end lookup_var_local;

public function lookup_ident_local " relation: lookup_ident_local
 
 
 
 "
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output Types.Var out_var;
  output TupleSCode_ElementTypes_ModOption out_tuplescode_elementtypes_modoption;
  output Boolean out_boolean;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModOption = Option<TupleSCode_ElementTypes_Mod>;
algorithm 
  (out_var,out_tuplescode_elementtypes_modoption,out_boolean):=
  matchcontinue (in_env,in_ident)
    local
      type StringOption = Option<String>;
      Types.Var fv;
      TupleSCode_ElementTypes_ModOption c;
      Boolean i;
      StringOption sid;
      Env.BinTree ht;
      String id;
    case (Env.FRAME(sid,ht,_,_,_) :: _,id)
      equation 
        (fv,c,i) = lookup_var2(ht, id); then (fv,c,i);
  end matchcontinue;
end lookup_ident_local;

public function lookup_ident
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output Types.Var out_var;
  output TupleSCode_ElementTypes_ModOption out_tuplescode_elementtypes_modoption;
  output Boolean out_boolean;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModOption = Option<TupleSCode_ElementTypes_Mod>;
algorithm 
  (out_var,out_tuplescode_elementtypes_modoption,out_boolean):=
  matchcontinue (in_env,in_ident)
    local
      type StringOption = Option<String>;
      type Env_FrameList = list<Env.Frame>;
      Types.Var fv;
      TupleSCode_ElementTypes_ModOption c;
      Boolean i;
      StringOption sid;
      Env.BinTree ht;
      String id;
      Env_FrameList rest;
    case (Env.FRAME(sid,ht,_,_,_) :: _,id)
      equation 
        (fv,c,i) = lookup_var2(ht, id); then (fv,c,i);
    case (_ :: rest,id)
      equation 
        (fv,c,i) = lookup_ident(rest, id); then (fv,c,i);
  end matchcontinue;
end lookup_ident;

public function lookup_functions_in_env "  Function lookup 
   relation: lookup_functions_in_env
 
  Returns a list of types that the function has. 
 
 "
  input Env.Env in_env;
  input Absyn.Path in_path;
  output Types_TypeList out_types_typelist;
protected 
  type Types_TypeList = list<Types.Type>;
algorithm 
  out_types_typelist:=
  matchcontinue (in_env,in_path)
    local
      type StringOption = Option<String>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type TupleTypes_TTypeAbsyn_PathOptionList = list<TupleTypes_TTypeAbsyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      Absyn.Path id,iid,path;
      StringOption sid;
      Env.BinTree ht,httypes;
      TupleTypes_TTypeAbsyn_PathOptionList reslist,c1,c2,res;
      Env_FrameList env,fs,env_1,env2,env_2;
      String pack;
      SCode.Class c;
      Boolean encflag;
      SCode.Restriction restr;
      ClassInf.State ci_state,cistate1;
      Env.Frame f;
    case ({},id) then {}; 
    case (env,(iid = Absyn.IDENT(id)))
      local String id;
      equation 
        _ = Static.elab_builtin_handler(id) "Check for builtin operators";
        Env.FRAME(sid,ht,httypes,_,_) = Env.top_frame(env);
        reslist = lookup_functions_in_frame(ht, httypes, env, id); then reslist;
    case ((env = Env.FRAME(sid,ht,httypes,_,_) :: fs),(iid = Absyn.IDENT(id)))
      local String id;
      equation 
        c1 = lookup_functions_in_frame(ht, httypes, env, id);
        c2 = lookup_functions_in_env(fs, iid);
        reslist = list_append(c1, c2); then reslist;
    case ((env = Env.FRAME(sid,ht,httypes,_,_) :: fs),(iid = Absyn.QUALIFIED(pack,path)))
      local String id;
      equation 
        (c = SCode.CLASS(id,_,encflag,restr,_),env_1) = lookup_class(env, Absyn.IDENT(pack), false) "For qualified function names, e.g. Modelica.Math.sin ";
        env2 = Env.open_scope(env_1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (_,env_2,_,cistate1,_) = Inst.inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {}, true, false) "Instantiate implicit (last argument = true) ";
        reslist = lookup_functions_in_env(env_2, path); then reslist;
    case (f :: _,(path = Absyn.IDENT(id)))
      local String id;
      equation 
        (c = SCode.CLASS(_,_,_,SCode.R_FUNCTION(),_),env_1) = lookup_class({f}, path, false) "If we find class that is function. ";
        (env = Env.FRAME(sid,ht,httypes,_,_) :: _) = Inst.implicit_function_type_instantiation(env_1, c);
        res = lookup_functions_in_frame(ht, httypes, env, id); then res;
    case (f :: fs,id) "Did not match. Continue"
      local TupleTypes_TTypeAbsyn_PathOptionList c;
      equation 
        c = lookup_functions_in_env(fs, id); then c;
    case (_,_)
      equation 
        Debug.fprintln("failtrace", "lookup_functions_in_env failed"); then fail();
  end matchcontinue;
end lookup_functions_in_env;

protected function lookup_type_in_env " - Internal functions 
   Type lookup 
   relation: lookup_type_in_env
  
 
 "
  input Env.Env in_env;
  input Absyn.Path in_path;
  output Types.Type out_type;
  output Env.Env out_env;
algorithm 
  (out_type,out_env):=
  matchcontinue (in_env,in_path)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      type StringOption = Option<String>;
      TupleTypes_TTypeAbsyn_PathOption c;
      Env_FrameList env_1,env,fs;
      StringOption sid;
      Env.BinTree ht,httypes;
      String id;
      Env.Frame f;
    case ((env = Env.FRAME(sid,ht,httypes,_,_) :: fs),Absyn.IDENT(id))
      equation 
        (c,env_1) = lookup_type_in_frame(ht, httypes, env, id); then (c,env_1);
    case (f :: fs,id)
      local Absyn.Path id;
      equation 
        (c,env_1) = lookup_type_in_env(fs, id); then (c,f :: env_1);
  end matchcontinue;
end lookup_type_in_env;

protected function lookup_type_in_frame "types "
  input Env.BinTree in_bintree1;
  input Env.BinTree in_bintree2;
  input Env.Env in_env3;
  input SCode.Ident in_ident4;
  output Types.Type out_type;
  output Env.Env out_env;
algorithm 
  (out_type,out_env):=
  matchcontinue (in_bintree1,in_bintree2,in_env3,in_ident4)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      type Types_VarList = list<Types.Var>;
      TupleTypes_TTypeAbsyn_PathOption t,ftype,ty;
      Env.BinTree ht,httypes;
      Env_FrameList env,cenv,env_1,env_2;
      String id,n;
      SCode.Class cdef;
      Absyn.Path fpath;
      Types_VarList varlst;
    case (ht,httypes,env,id) "Classes and vars types "
      equation 
        Env.TYPE(t :: _) = Env.tree_get(httypes, id, Env.myhash); then (t,env);
    case (ht,httypes,env,id)
      equation 
        Env.VAR(_,_,_) = Env.tree_get(ht, id, Env.myhash);
        Print.print_error_buf("# Error while looking up type ");
        Print.print_error_buf(id);
        Print.print_error_buf(": found a component\n"); then fail();
    case (ht,httypes,env,id)
      equation 
        Env.CLASS(cdef = SCode.CLASS(n,_,_,SCode.R_RECORD(),_),_) = Env.tree_get(ht, id, Env.myhash) "Each time a record constructor function is looked up, this rule will create the function. An improvement (perhaps needing lot of code) is to add the function to the environment, which is  returned from this function.";
        fpath = Inst.make_fully_qualified(env, Absyn.IDENT(n));
        varlst = build_record_constructor_varlst(cdef, env);
        ftype = Types.make_function_type(fpath, varlst); then (ftype,env);
    case (ht,httypes,env,id)
      equation 
        Env.CLASS(cdef = SCode.CLASS(_,_,_,SCode.R_FUNCTION(),_),cenv) = Env.tree_get(ht, id, Env.myhash) "If we found class that is function ";
        (env_1,_) = Inst.implicit_function_instantiation(cenv, Types.NOMOD(), Prefix.NOPRE(), {}, cdef, {}, false);
        (ty,env_2) = lookup_type_in_env(env_1, Absyn.IDENT(id)); then (ty,env_2);
    case (ht,httypes,env,id)
      equation 
        Env.CLASS(cdef = SCode.CLASS(_,_,_,SCode.R_EXT_FUNCTION(),_),cenv) = Env.tree_get(ht, id, Env.myhash) "If we found class that is external function ";
        env_1 = Inst.implicit_function_type_instantiation(cenv, cdef);
        (ty,env_2) = lookup_type_in_env(env_1, Absyn.IDENT(id)); then (ty,env_2);
  end matchcontinue;
end lookup_type_in_frame;

protected function lookup_functions_in_frame "Types "
  input Env.BinTree in_bintree1;
  input Env.BinTree in_bintree2;
  input Env.Env in_env3;
  input SCode.Ident in_ident4;
  output Types_TypeList out_types_typelist;
protected 
  type Types_TypeList = list<Types.Type>;
algorithm 
  out_types_typelist:=
  matchcontinue (in_bintree1,in_bintree2,in_env3,in_ident4)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type TupleTypes_TTypeAbsyn_PathOptionList = list<TupleTypes_TTypeAbsyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      type Types_VarList = list<Types.Var>;
      TupleTypes_TTypeAbsyn_PathOptionList tps;
      Env.BinTree ht,httypes;
      Env_FrameList env,cenv,env_1;
      String id,n;
      SCode.Class cdef;
      Types_VarList varlst;
      Absyn.Path fpath;
      TupleTypes_TTypeAbsyn_PathOption ftype;
    case (ht,httypes,env,id) "Classes and vars Types "
      equation 
        Env.TYPE(tps) = Env.tree_get(httypes, id, Env.myhash); then tps;
    case (ht,httypes,env,id)
      equation 
        Env.VAR(_,_,_) = Env.tree_get(ht, id, Env.myhash);
        Print.print_error_buf("# Error while looking up type ");
        Print.print_error_buf(id);
        Print.print_error_buf(": found a component\n"); then fail();
    case (ht,httypes,env,id)
      equation 
        Env.CLASS(cdef = SCode.CLASS(n,_,_,SCode.R_RECORD(),_),cenv) = Env.tree_get(ht, id, Env.myhash);
        varlst = build_record_constructor_varlst(cdef, env);
        fpath = Inst.make_fully_qualified(cenv, Absyn.IDENT(n));
        ftype = Types.make_function_type(fpath, varlst); then {ftype};
    case (ht,httypes,env,id)
      equation 
        Env.CLASS(cdef = SCode.CLASS(_,_,_,SCode.R_FUNCTION(),_),cenv) = Env.tree_get(ht, id, Env.myhash) "If found class that is function.";
        env_1 = Inst.implicit_function_type_instantiation(cenv, cdef);
        tps = lookup_functions_in_env(env_1, Absyn.IDENT(id)); then tps;
    case (ht,httypes,env,id)
      equation 
        Env.CLASS(cdef = SCode.CLASS(_,_,_,SCode.R_EXT_FUNCTION(),_),cenv) = Env.tree_get(ht, id, Env.myhash) "If found class that is external function.";
        env_1 = Inst.implicit_function_type_instantiation(cenv, cdef);
        tps = lookup_functions_in_env(env_1, Absyn.IDENT(id)); then tps;
  end matchcontinue;
end lookup_functions_in_frame;

protected function lookup_recconst_in_env " relation: lookup_recconst_in_env
  
  
  
 "
  input Env.Env in_env;
  input Absyn.Path in_path;
  output SCode.Class out_class;
  output Env.Env out_env;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_env,in_path)
    local
      type Env_FrameList = list<Env.Frame>;
      type StringOption = Option<String>;
      type Env_ItemList = list<Env.Item>;
      SCode.Class c;
      Env_FrameList env,fs;
      StringOption sid;
      Env.BinTree ht;
      Env_ItemList imps;
      String id;
      Env.Frame f;
    case ((env = Env.FRAME(sid,ht,_,imps,_) :: fs),Absyn.IDENT(id))
      equation 
        (c,_) = lookup_recconst_in_frame(ht, env, id); then (c,env);
    case (f :: fs,id)
      local Absyn.Path id;
      equation 
        (c,_) = lookup_recconst_in_env(fs, id); then (c,f :: fs);
  end matchcontinue;
end lookup_recconst_in_env;

protected function lookup_recconst_in_frame
  input Env.BinTree in_bintree;
  input Env.Env in_env;
  input SCode.Ident in_ident;
  output SCode.Class out_class;
  output Env.Env out_env;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_bintree,in_env,in_ident)
    local
      type Env_FrameList = list<Env.Frame>;
      Env.BinTree ht;
      Env_FrameList env;
      String id;
      SCode.Class cdef;
    case (ht,env,id)
      equation 
        Env.VAR(_,_,_) = Env.tree_get(ht, id, Env.myhash);
        Print.print_error_buf("# Error while looking up record constructor class ");
        Print.print_error_buf(id);
        Print.print_error_buf(": found a component\n"); then fail();
    case (ht,env,id)
      equation 
        Env.CLASS(cdef = SCode.CLASS(_,_,_,SCode.R_RECORD(),_),_) = Env.tree_get(ht, id, Env.myhash);
        cdef = build_record_constructor_class(cdef, env); then (cdef,env);
  end matchcontinue;
end lookup_recconst_in_frame;

protected function build_record_constructor_class " relation: build_record_constructor_class
  
  
  
 "
  input SCode.Class in_class;
  input Env.Env in_env;
  output SCode.Class out_class;
algorithm 
  out_class:=
  matchcontinue (in_class,in_env)
    local
      type SCode_ElementList = list<SCode.Element>;
      type Env_FrameList = list<Env.Frame>;
      SCode_ElementList funcelts,elts;
      SCode.Element reselt;
      SCode.Class cl;
      String id;
      SCode.Restriction restr;
      Env_FrameList env;
    case ((cl = SCode.CLASS(id,_,_,restr,SCode.PARTS(elts,_,_,_,_,_))),env)
      equation 
        funcelts = build_record_constructor_elts(elts, env);
        reselt = build_record_constructor_result_elt(elts, id, env); then SCode.CLASS(id,false,false,SCode.R_FUNCTION(),
          SCode.PARTS(reselt :: funcelts,{},{},{},{},NONE));
  end matchcontinue;
end build_record_constructor_class;

protected function build_record_constructor_elts " relation: build_record_constructor_elts
  
  
  
 "
  input SCode_ElementList in_scode_elementlist;
  input Env.Env in_env;
  output SCode_ElementList out_scode_elementlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_scode_elementlist:=
  matchcontinue (in_scode_elementlist,in_env)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type Env_FrameList = list<Env.Frame>;
      SCode_ElementList res,rest;
      SCode.Element comp;
      String id;
      Boolean fl,repl,prot,f;
      Absyn_SubscriptList d;
      SCode.Accessibility ac;
      SCode.Variability var;
      Absyn.Direction dir;
      Absyn.Path tp;
      SCode.Mod mod;
      Absyn_PathOption bc;
      Absyn_CommentOption comment;
      Env_FrameList env;
    case ((comp = SCode.COMPONENT(id,fl,repl,prot,SCode.ATTR(d,f,ac,var,dir),tp,mod,bc,comment)) :: rest,env)
      equation 
        res = build_record_constructor_elts(rest, env); then SCode.COMPONENT(id,fl,repl,prot,SCode.ATTR(d,f,ac,var,Absyn.INPUT()),tp,
          mod,bc,comment) :: res;
    case ({},_) then {}; 
  end matchcontinue;
end build_record_constructor_elts;

protected function build_record_constructor_result_elt " relation: build_record_constructor_result_elt
  
  This function builds the result element of a record constructor function, 
  i.e. the returned variable
  
 "
  input SCode_ElementList elts;
  input SCode.Ident id;
  input Env.Env env;
  output SCode.Element out_element;
protected 
  type SCode_ElementList = list<SCode.Element>;
  type SCode_SubModList = list<SCode.SubMod>;
  SCode_SubModList submodlst;
algorithm 
  submodlst := build_record_constructor_result_mod(elts);
  out_element := SCode.COMPONENT("result",false,false,false,
          SCode.ATTR({},false,SCode.RW(),SCode.VAR(),Absyn.OUTPUT()),Absyn.IDENT(id),SCode.MOD(false,submodlst,NONE),NONE,NONE);
end build_record_constructor_result_elt;

protected function build_record_constructor_result_mod " relation: build_record_constructor_result_mod
 
  This function builds up the modification list for the output element of a record constructor.
  Example: 
    record foo
       Real x;
       String y;
       end foo;
   => modifier list become 'result.x=x, result.y=y'
 "
  input SCode_ElementList in_scode_elementlist;
  output SCode_SubModList out_scode_submodlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
  type SCode_SubModList = list<SCode.SubMod>;
algorithm 
  out_scode_submodlist:=
  matchcontinue (in_scode_elementlist)
    local
      SCode_SubModList restmod;
      String id;
      SCode_ElementList rest;
    case (SCode.COMPONENT(id,_,_,_,_,_,_,_,_) :: rest)
      equation 
        restmod = build_record_constructor_result_mod(rest); then SCode.NAMEMOD("result",
          SCode.MOD(false,
          {
          SCode.NAMEMOD(id,
          SCode.MOD(false,{},SOME(Absyn.CREF(Absyn.CREF_IDENT(id,{})))))},NONE)) :: restmod;
    case ({}) then {}; 
  end matchcontinue;
end build_record_constructor_result_mod;

protected function build_record_constructor_varlst " relation: build_record_constructor_varlst
 
  This function takes a class  (`SCode.Class') which holds a definition of a record
  and builds a list of variables of the record used for constructing a record constructor function.
 "
  input SCode.Class in_class;
  input Env.Env in_env;
  output Types_VarList out_types_varlist;
protected 
  type Types_VarList = list<Types.Var>;
algorithm 
  out_types_varlist:=
  matchcontinue (in_class,in_env)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type SCode_ElementList = list<SCode.Element>;
      type Env_FrameList = list<Env.Frame>;
      Types_VarList inputvarlst;
      TupleTypes_TTypeAbsyn_PathOption ty;
      SCode.Class cl;
      SCode_ElementList elts;
      Env_FrameList env;
    case ((cl = SCode.CLASS(_,_,_,_,SCode.PARTS(elts,_,_,_,_,_))),env)
      equation 
        inputvarlst = build_varlst_from_elts(elts, env);
        (_,_,_,ty,_) = Inst.inst_class(env, Types.NOMOD(), Prefix.NOPRE(), {}, cl, {}, true, 
          Inst.TOP_CALL()) "FIXME: impl"; then Types.VAR("result",
          Types.ATTR(false,SCode.RW(),SCode.VAR(),Absyn.OUTPUT()),false,ty,Types.UNBOUND()) :: inputvarlst;
    case (_,_)
      equation 
        Debug.fprint("failtrace", "build_record_constructor_varlst failed\n"); then fail();
  end matchcontinue;
end build_record_constructor_varlst;

protected function build_varlst_from_elts " relation: build_varlst_from_elts
  
  
  
 "
  input SCode_ElementList in_scode_elementlist;
  input Env.Env in_env;
  output Types_VarList out_types_varlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
  type Types_VarList = list<Types.Var>;
algorithm 
  out_types_varlist:=
  matchcontinue (in_scode_elementlist,in_env)
    local
      type Env_FrameList = list<Env.Frame>;
      Types_VarList vars;
      Types.Var var;
      SCode.Element comp;
      SCode_ElementList rest;
      Env_FrameList env;
    case ((comp = SCode.COMPONENT(_,_,_,_,_,_,_,_,_)) :: rest,env)
      equation 
        vars = build_varlst_from_elts(rest, env);
        var = Inst.inst_record_constructor_elt(env, comp, true) "P.A Here we need to do a lookup of the type. Therefore we need the env passed along from lookup_xxxx function.  FIXME: impl"; then var :: vars;
    case ({},_) then {}; 
    case (_,_) then fail(); 
  end matchcontinue;
end build_varlst_from_elts;

public function is_in_builtin_env "  Class lookup "
  input Absyn.Path in_path;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_path)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList i_env;
      Absyn.Path path;
    case (path)
      equation 
        i_env = Builtin.initial_env();
        {} = lookup_functions_in_env(i_env, path); then false;
    case (path)
      equation 
        i_env = Builtin.initial_env();
        _ = lookup_functions_in_env(i_env, path); then true;
    case (path)
      equation 
        Debug.fprintln("failtrace", "is_in_builtin_env failed"); then fail();
  end matchcontinue;
end is_in_builtin_env;

protected function lookup_class_in_env " relation: lookup_class_in_env
  
  
  
 "
  input Env.Env in_env;
  input Absyn.Path in_path;
  input Boolean in_boolean;
  output SCode.Class out_class;
  output Env.Env out_env;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_env,in_path,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type StringOption = Option<String>;
      SCode.Class c;
      Env_FrameList env_1,env,fs,i_env;
      Env.Frame frame,f;
      String id,sid;
      Boolean msg,msgflag;
      Absyn.Path aid;
    case ((env = frame :: fs),Absyn.IDENT(id),msg)
      equation 
        (c,env_1) = lookup_class_in_frame(frame, frame :: fs, id, msg) "print \"looking in env for \" & print id &"; then (c,env_1);
    case ((env = Env.FRAME(SOME(sid),_,_,_,true) :: fs),(aid = Absyn.IDENT(id)),_)
      equation 
        equality(id = sid) "Special case if looking up the class that -is- encapsulated. That must be allowed. ";
        (c,env) = lookup_class_in_env(fs, aid, true); then (c,env);
    case ((env = Env.FRAME(sid,_,_,_,true) :: fs),(aid = Absyn.IDENT(id)),true)
      local StringOption sid;
      equation 
        i_env = Builtin.initial_env();
        failure((_,_) = lookup_class_in_env(i_env, aid, false));
        Print.print_error_buf("# Class ");
        Print.print_error_buf(id);
        Print.print_error_buf(" not found and the lookup reached an encapsulated class\n"); then fail();
    case (Env.FRAME(sid,_,_,_,true) :: fs,(aid = Absyn.IDENT(id)),false)
      local StringOption sid;
      equation 
        i_env = Builtin.initial_env();
        failure((_,_) = lookup_class_in_env(i_env, aid, false)); then fail();
    case (Env.FRAME(sid,_,_,_,true) :: fs,(aid = Absyn.IDENT(id)),msgflag)
      local StringOption sid;
      equation 
        i_env = Builtin.initial_env();
        (c,env_1) = lookup_class_in_env(i_env, aid, msgflag); then (c,env_1);
    case ((f = Env.FRAME(sid,_,_,_,false)) :: fs,id,msgflag)
      local
        StringOption sid;
        Absyn.Path id;
      equation 
        (c,env_1) = lookup_class_in_env(fs, id, msgflag); then (c,env_1);
  end matchcontinue;
end lookup_class_in_env;

protected function lookup_type_in_class " relation: lookup_type_in_class
  This function looks up an type inside a class (The outer class can be a package)
  Environment is passed along in case it needs to be modified...
  bool determines whether we restrict lookup for encapsulated class (true).
   
 "
  input Env.Env in_env;
  input SCode.Class in_class;
  input Absyn.Path in_path;
  input Boolean in_boolean;
  output Types.Type out_type;
  output Env.Env out_env;
algorithm 
  (out_type,out_env):=
  matchcontinue (in_env,in_class,in_path,in_boolean)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      TupleTypes_TTypeAbsyn_PathOption tp,t;
      Env_FrameList env_1,env,env_2,env3,env2,env4,env5,env1;
      SCode.Class cdef,c;
      Absyn.Path classname,p1,path;
      String id,c1,cname;
      Boolean encflag;
      SCode.Restriction restr;
      ClassInf.State ci_state;
    case (env,cdef,(classname = Absyn.IDENT(_)),_)
      equation 
        (tp,env_1) = lookup_type_in_env(env, classname) ", true"; then (tp,env_1);
    case (env,cdef,(classname = Absyn.IDENT(_)),_)
      equation 
        (c = SCode.CLASS(_,_,_,SCode.R_FUNCTION(),_),env_1) = lookup_class_in_env(env, classname, false) "If not found, look for classdef that is function and instantiate. ";
        env_2 = Inst.implicit_function_type_instantiation(env_1, c);
        (t,env3) = lookup_type_in_env(env_2, classname); then (t,env3);
    case (env,cdef,(classname = Absyn.IDENT(_)),_)
      equation 
        (c = SCode.CLASS(_,_,_,SCode.R_EXT_FUNCTION(),_),env_1) = lookup_class_in_env(env, classname, false) "If not found, look for classdef that is external function and instantiate. ";
        env_2 = Inst.implicit_function_type_instantiation(env_1, c);
        (t,env3) = lookup_type_in_env(env_2, classname); then (t,env3);
    case (env,cdef,Absyn.QUALIFIED(c1,p1),true)
      equation 
        (c = SCode.CLASS(id,_,encflag = true,restr,_),env) = lookup_class_in_env(env, Absyn.IDENT(c1), false) "Restrict lookup to encapsulated elements only ";
        env2 = Env.open_scope(env, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (_,env4,_,_,_) = Inst.inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {}, true, false);
        (t,env5) = lookup_type_in_class(env4, c, p1, false); then (t,env5);
    case (env,cdef,Absyn.QUALIFIED(c1,p1),true)
      equation 
        (c = SCode.CLASS(id,_,encflag = false,restr,_),env) = lookup_class_in_env(env, Absyn.IDENT(c1), false) "Restrict lookup to encapsulated elements only ";
        Print.print_error_buf(
          "# Error: attempt to search qualified class name went into non-encapsulated class (when it is still qualified)\n") "Restrict lookup to encapsulated elements only "; then fail();
    case (env,cdef,Absyn.QUALIFIED(c1,p1),false)
      equation 
        (c = SCode.CLASS(id,_,encflag,restr,_),env1) = lookup_class_in_env(env, Absyn.IDENT(c1), false) "Lookup not restricted to encapsulated elts. only ";
        env2 = Env.open_scope(env1, encflag, SOME(id));
        ci_state = ClassInf.start(restr, id);
        (_,env4,_,_,_) = Inst.inst_class_in(env2, Types.NOMOD(), Prefix.NOPRE(), {}, ci_state, c, 
          false, {}, true, false) "	Print.print_buf \"instanitating class \" &
	Print.print_buf id &
	Print.print_buf \" in envpath:\n\" &
	Env.print_env_path(env2') &
	Print.print_buf \"\n\" &";
        (t,env5) = lookup_type_in_class(env4, c, p1, false); then (t,env5);
    case (_,SCode.CLASS(cname,_,_,_,_),path,_) then fail(); 
  end matchcontinue;
end lookup_type_in_class;

protected function lookup_class_in_parts " relation: lookup_class_in_parts
  This function takes an `Env' and an `Ident' and an `Element' list, and returns the `Class' 
  having the name specified by the `Ident'. If not found, the function fails.
 "
  input Env.Env in_env;
  input SCode.Ident in_ident;
  input SCode_ElementList in_scode_elementlist;
  output SCode.Class out_class;
  output Env.Env out_env;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_env,in_ident,in_scode_elementlist)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList env;
      String name,name2;
      SCode.Class c;
      SCode_ElementList xs;
    case (env,name,SCode.CLASSDEF(name2,_,_,c,_) :: xs)
      equation 
        equality(name = name2); then (c,env);
    case (env,name,_ :: xs)
      equation 
        (c,env) = lookup_class_in_parts(env, name, xs); then (c,env);
    case (env,name,{}) then fail(); 
  end matchcontinue;
end lookup_class_in_parts;

protected function lookup_class_in_frame
  input Env.Frame in_frame;
  input Env.Env in_env;
  input SCode.Ident in_ident;
  input Boolean in_boolean;
  output SCode.Class out_class;
  output Env.Env out_env;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_frame,in_env,in_ident,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type StringOption = Option<String>;
      type Env_ItemList = list<Env.Item>;
      SCode.Class c;
      Env_FrameList env,totenv,env_1;
      StringOption sid;
      Env.BinTree ht;
      String id,name;
      Env_ItemList items;
    case (Env.FRAME(sid,ht,_,_,_),totenv,id,_)
      equation 
        Env.CLASS(c,env) = Env.tree_get(ht, id, Env.myhash) "print \"looking for class \" & print id & print \" in frame\n\" & & print \"found \" & print id & print \"\n\" "; then (c,totenv);
    case (Env.FRAME(sid,ht,_,_,_),_,id,true)
      equation 
        Env.VAR(_,_,_) = Env.tree_get(ht, id, Env.myhash);
        Print.print_error_buf("# Error while looking up class ");
        Print.print_error_buf(id);
        Print.print_error_buf(": found a component\n"); then fail();
    case (Env.FRAME(sid,_,_,items,_),totenv,name,_)
      equation 
        (c,env_1) = lookup_qualified_imported_class_in_frame(items, totenv, name); then (c,env_1);
    case (Env.FRAME(sid,_,_,items,_),totenv,name,_)
      equation 
        (c,env_1,true) = lookup_unqualified_imported_class_in_frame(items, totenv, name) "unique"; then (c,env_1);
    case (Env.FRAME(sid,_,_,items,_),totenv,name,_)
      equation 
        (c,env_1,false) = lookup_unqualified_imported_class_in_frame(items, totenv, name) "unique";
        Print.print_error_buf(
          "# Error, class found in several unqualified import statements\n") "unique"; then fail();
  end matchcontinue;
end lookup_class_in_frame;

protected function lookup_var2
  input Env.BinTree in_bintree;
  input SCode.Ident in_ident;
  output Types.Var out_var;
  output TupleSCode_ElementTypes_ModOption out_tuplescode_elementtypes_modoption;
  output Boolean out_boolean;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModOption = Option<TupleSCode_ElementTypes_Mod>;
algorithm 
  (out_var,out_tuplescode_elementtypes_modoption,out_boolean):=
  matchcontinue (in_bintree,in_ident)
    local
      Types.Var fv;
      TupleSCode_ElementTypes_ModOption c;
      Boolean i;
      Env.BinTree ht;
      String id;
    case (ht,id)
      equation 
        Env.VAR(fv,c,i) = Env.tree_get(ht, id, Env.myhash); then (fv,c,i);
  end matchcontinue;
end lookup_var2;

protected function check_subscripts
  input Types.Type in_type;
  input Exp_SubscriptList in_exp_subscriptlist;
  output Types.Type out_type;
protected 
  type Exp_SubscriptList = list<Exp.Subscript>;
algorithm 
  out_type:=
  matchcontinue (in_type,in_exp_subscriptlist)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Exp_ExpList = list<Exp.Exp>;
      TupleTypes_TTypeAbsyn_PathOption t,t_1;
      Types.ArrayDim dim;
      Absyn_PathOption p;
      Exp_SubscriptList ys,s;
      Integer sz,ind;
      Exp_ExpList se;
    case (t,{}) then t; 
    case ((Types.T_ARRAY(dim,t),p),Exp.WHOLEDIM() :: ys)
      equation 
        t_1 = check_subscripts(t, ys); then ((Types.T_ARRAY(dim,t_1),p));
    case ((Types.T_ARRAY(Types.DIM(SOME(sz)),t),p),Exp.SLICE(Exp.ARRAY(_,_,se)) :: ys)
      local Integer dim;
      equation 
        t_1 = check_subscripts(t, ys);
        dim = list_length(se) "FIXME: Check range "; then ((Types.T_ARRAY(Types.DIM(SOME(dim)),t_1),p));
    case ((Types.T_ARRAY(Types.DIM(SOME(sz)),t),_),Exp.INDEX(Exp.ICONST(ind)) :: ys)
      equation 
        (ind > 0) = true;
        (ind <= sz) = true;
        t_1 = check_subscripts(t, ys); then t_1;
    case ((Types.T_ARRAY(Types.DIM(SOME(sz)),t),_),Exp.INDEX(_) :: ys) "HJ: Subscrits needn't be constant. No range-checking can
	       be done "
      equation 
        t_1 = check_subscripts(t, ys); then t_1;
    case ((Types.T_ARRAY(Types.DIM(NONE),t),_),Exp.INDEX(_) :: ys)
      equation 
        t_1 = check_subscripts(t, ys); then t_1;
    case ((Types.T_ARRAY(Types.DIM(SOME(sz)),t),_),Exp.WHOLEDIM() :: ys)
      equation 
        t_1 = check_subscripts(t, ys); then t_1;
    case ((Types.T_ARRAY(Types.DIM(NONE),t),_),Exp.WHOLEDIM() :: ys)
      equation 
        t_1 = check_subscripts(t, ys); then t_1;
    case ((Types.T_ARRAY(Types.DIM(SOME(sz)),t),_),Exp.SLICE(_) :: ys)
      equation 
        t_1 = check_subscripts(t, ys); then t_1;
    case ((Types.T_ARRAY(Types.DIM(NONE),t),_),Exp.SLICE(_) :: ys)
      equation 
        t_1 = check_subscripts(t, ys); then t_1;
    case (t,s) then fail(); 
  end matchcontinue;
end check_subscripts;

protected function lookup_in_var
  input Types.Type in_type;
  input Exp.ComponentRef in_componentref;
  output Types.Attributes out_attributes;
  output Types.Type out_type;
  output Types.Binding out_binding;
algorithm 
  (out_attributes,out_type,out_binding):=
  matchcontinue (in_type,in_componentref)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Exp_SubscriptList = list<Exp.Subscript>;
      Boolean fl;
      SCode.Accessibility acc;
      SCode.Variability vt;
      Absyn.Direction di;
      TupleTypes_TTypeAbsyn_PathOption ty_1,ty_2,ty,ty_3;
      Types.Binding binding;
      String id;
      Exp_SubscriptList ss;
      Types.Attributes attr;
      Exp.ComponentRef vs;
    case (ty,Exp.CREF_IDENT(id,ss)) "Public components "
      equation 
        (Types.VAR(_,Types.ATTR(fl,acc,vt,di),false,ty_1,binding)) = Types.lookup_component(ty, id);
        ty_2 = check_subscripts(ty_1, ss); then (Types.ATTR(fl,acc,vt,di),ty_2,binding);
    case (ty,Exp.CREF_IDENT(id,ss)) "Protected components "
      equation 
        (Types.VAR(_,_,true,_,_)) = Types.lookup_component(ty, id);
        Print.print_error_buf("# Error, referencing protected elements not allowed.\n"); then fail();
    case (ty,Exp.CREF_QUAL(id,ss,vs))
      equation 
        (Types.VAR(_,Types.ATTR(fl,acc,vt,di),_,ty_1,_)) = Types.lookup_component(ty, id);
        ty_2 = check_subscripts(ty_1, ss);
        (attr,ty_3,binding) = lookup_in_var(ty_2, vs); then (attr,ty_3,binding);
    case (_,_) then fail(); 
  end matchcontinue;
end lookup_in_var;

protected function lookup_var_f
  input Env.BinTree in_bintree;
  input Exp.ComponentRef in_componentref;
  output Types.Attributes out_attributes;
  output Types.Type out_type;
  output Types.Binding out_binding;
algorithm 
  (out_attributes,out_type,out_binding):=
  matchcontinue (in_bintree,in_componentref)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Exp_SubscriptList = list<Exp.Subscript>;
      String n,id;
      Boolean f;
      SCode.Accessibility acc;
      SCode.Variability vt;
      Absyn.Direction di;
      TupleTypes_TTypeAbsyn_PathOption ty,ty_1,ty_2;
      Types.Binding bind,binding;
      Env.BinTree ht;
      Exp_SubscriptList ss;
      Types.Attributes attr;
      Exp.ComponentRef ids;
    case (ht,Exp.CREF_IDENT(id,ss))
      equation 
        (Types.VAR(n,Types.ATTR(f,acc,vt,di),_,ty,bind),_,_) = lookup_var2(ht, id);
        ty_1 = check_subscripts(ty, ss); then (Types.ATTR(f,acc,vt,di),ty_1,bind);
    case (ht,Exp.CREF_QUAL(id,ss,ids))
      equation 
        (Types.VAR(n,Types.ATTR(f,acc,vt,di),_,ty,_),_,_) = lookup_var2(ht, id);
        ty_1 = check_subscripts(ty, ss);
        (attr,ty_2,binding) = lookup_in_var(ty_1, ids) "FIXME: Check for non-array "; then (attr,ty_2,binding);
  end matchcontinue;
end lookup_var_f;

protected function assert_package " relation: assert_package
  
  This function checks that a class definition is a package.  This
  breaks the everything-can-be-generalized-to-class principle, since
  it requires that the keyword `package' is used in the package file.
 "
  input SCode.Class in_class;
algorithm 
  _:=
  matchcontinue (in_class)
    case SCode.CLASS(_,_,_,SCode.R_PACKAGE(),_) then (); 
  end matchcontinue;
end assert_package;

protected function lookup_class_in_package
  input SCode.Class in_class;
  input Absyn.Path in_path;
  output SCode.Class out_class;
  output Env.Env out_env;
algorithm 
  (out_class,out_env):=
  matchcontinue (in_class,in_path)
    local
      type SCode_ElementList = list<SCode.Element>;
      SCode.Class c;
      SCode_ElementList els;
      Absyn.Path path;
    case (SCode.CLASS(_,_,_,_,SCode.PARTS(els,_,_,_,_,_)),path)
      equation 
        c = lookup_class_in_elements(els, path); then (c,Env.empty_env);
  end matchcontinue;
end lookup_class_in_package;

protected function lookup_class_in_elements
  input SCode_ElementList in_scode_elementlist;
  input Absyn.Path in_path;
  output SCode.Class out_class;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_class:=
  matchcontinue (in_scode_elementlist,in_path)
    local
      type Env_FrameList = list<Env.Frame>;
      String n1,n2;
      SCode.Class c,c_1;
      Env_FrameList env;
      Absyn.Path path;
      SCode_ElementList es;
    case (SCode.CLASSDEF(n1,_,_,c,_) :: _,Absyn.IDENT(n2))
      equation 
        equality(n1 = n2); then c;
    case (SCode.CLASSDEF(n1,_,_,c,_) :: _,Absyn.QUALIFIED(n2,path))
      equation 
        equality(n1 = n2);
        assert_package(c);
        (c_1,env) = lookup_class_in_package(c, path); then c_1;
    case (_ :: es,path)
      equation 
        c = lookup_class_in_elements(es, path); then c;
  end matchcontinue;
end lookup_class_in_elements;
end Lookup;

