package Main "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:        Main.rml
  module:      Main
  description: Modelica main program
 
  RCS: $Id: Main.rml 1821 2005-06-21 12:12:32Z petar $
 
  This is the main program in the Modelica specification. 
  It either tranlates a file given as a command line argument
  or starts a server loop communicating through CORBA or sockets
  (The Win32 implementation only implements CORBA)
 "

protected import OpenModelica.Compiler.Absyn;

protected import OpenModelica.Compiler.ModUtil;

protected import OpenModelica.Compiler.Parser;

protected import OpenModelica.Compiler.Dump;

protected import OpenModelica.Compiler.DumpGraphviz;

protected import OpenModelica.Compiler.SCode;

protected import OpenModelica.Compiler.DAE;

protected import OpenModelica.Compiler.DAELow;

protected import OpenModelica.Compiler.Inst;

protected import OpenModelica.Compiler.Interactive;

protected import OpenModelica.Compiler.RTOpts;

protected import OpenModelica.Compiler.Debug;

protected import OpenModelica.Compiler.Codegen;

protected import OpenModelica.Compiler.Socket;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Corba;

protected import OpenModelica.Compiler.System;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.TaskGraph;

protected import OpenModelica.Compiler.TaskGraphExt;

protected import OpenModelica.Compiler.SimCodegen;

protected function server_loop " relation: server_loop
 
  This function is the main loop of the server listening to a port
    which recieves modelica expressions,  
"
  input Integer in_integer;
  input Interactive.InteractiveSymbolTable in_interactivesymboltable;
  output Interactive.InteractiveSymbolTable out_interactivesymboltable;
algorithm 
  out_interactivesymboltable:=
  matchcontinue (in_integer,in_interactivesymboltable)
    local
      String str,replystr;
      Interactive.InteractiveSymbolTable newsymb,ressymb,isymb;
      Integer shandle;
    case (shandle,isymb)
      equation 
        str = Socket.handlerequest(shandle);
        Debug.fprint("interactivedump", 
          "------- Recieved Data from client -----\n");
        Debug.fprint("interactivedump", str);
        Debug.fprint("interactivedump", "------- End recieved Data-----\n");
        Print.clear_buf();
        Print.clear_error_buf();
        (true,replystr,newsymb) = handle_command(str, isymb);
        Socket.sendreply(shandle, replystr);
        ressymb = server_loop(shandle, newsymb); then ressymb;
    case (shandle,isymb)
      equation 
        str = Socket.handlerequest(shandle) "2004-11-27 - adrpo added this part to make the loop deterministic ";
        Debug.fprint("interactivedump", 
          "------- Recieved Data from client -----\n") "2004-11-27 - adrpo added this part to make the loop deterministic ";
        Debug.fprint("interactivedump", str);
        Debug.fprint("interactivedump", "------- End recieved Data-----\n");
        Print.clear_error_buf();
        Print.clear_buf();
        (false,replystr,newsymb) = handle_command(str, isymb);
        Print.print_buf("Exiting\n");
        Socket.sendreply(shandle, "quit requested, shutting server down\n");
        Socket.close(shandle);
        Socket.cleanup(); then isymb;
  end matchcontinue;
end server_loop;

protected function charlistcompare
  input charList in_charlist1;
  input charList in_charlist2;
  input Integer in_integer3;
  output Boolean out_boolean;
protected 
  type charList = list<char>;
algorithm 
  out_boolean:=
  matchcontinue (in_charlist1,in_charlist2,in_integer3)
    local
      char a,b;
      Integer n1,n;
      charList l1,l2;
    case (a :: _,b :: _,1)
      equation 
        equality(a = b); then true;
    case (a :: _,b :: _,1)
      equation 
        failure(equality(a = b)); then false;
    case (a :: l1,b :: l2,n)
      equation 
        n1 = n - 1;
        equality(a = b);
        true = charlistcompare(l1, l2, n1); then true;
    case (_,_,_) then false; 
  end matchcontinue;
end charlistcompare;

protected function strncmp
  input String in_string1;
  input String in_string2;
  input Integer in_integer3;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_string1,in_string2,in_integer3)
    local
      type charList = list<char>;
      charList clst1,clst2;
      Integer s1len,s2len,n;
      String s1,s2;
    case (s1,s2,n)
      equation 
        clst1 = string_list(s1);
        clst2 = string_list(s2);
        s1len = string_length(s1);
        s2len = string_length(s2);
        (s1len >= n) = true;
        (s2len >= n) = true;
        true = charlistcompare(clst1, clst2, n); then true;
    case (_,_,_) then false; 
  end matchcontinue;
end strncmp;

protected function check_classdef
  input String in_string;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_string)
    local
      type charList = list<char>;
      charList clst,clst_1;
      String str_1,str;
      Boolean res;
    case (str) "Need to check for a whitespace after as well to get the keyword,
	e.g typeOf function would be taken as a type definition otherwise "
      equation 
        true = strncmp(" ", str, 1);
        clst = string_list(str);
        clst_1 = list_delete(clst, 0);
        str_1 = list_string(clst_1);
        res = check_classdef(str_1); then res;
    case str "Need to check for a whitespace after as well to get the keyword,
	e.g typeOf function would be taken as a type definition otherwise "
      equation 
        false = strncmp("end ", str, 4);
        false = strncmp("type ", str, 5);
        false = strncmp("class ", str, 6);
        false = strncmp("model ", str, 6);
        false = strncmp("block ", str, 6);
        false = strncmp("within ", str, 7);
        false = strncmp("record ", str, 7);
        false = strncmp("package ", str, 8);
        false = strncmp("partial ", str, 8);
        false = strncmp("function ", str, 9);
        false = strncmp("connector ", str, 10);
        false = strncmp("encapsulated ", str, 12); then false;
    case _ then true; 
  end matchcontinue;
end check_classdef;

protected function make_debug_result
  input String flagstr;
  input String res;
  output String res_1;
protected 
  String debugstr,res_with_debug,res_1;
  Boolean dumpflag;
algorithm 
  debugstr := Print.get_string();
  res_with_debug := Util.string_append_list(
          {res,"\n---DEBUG(",flagstr,")---\n",debugstr,"\n---/DEBUG(",
          flagstr,")---\n"});
  dumpflag := RTOpts.debug_flag(flagstr);
  res_1 := Util.if_(dumpflag, res_with_debug, res);
end make_debug_result;

protected function handle_command
  input String in_string;
  input Interactive.InteractiveSymbolTable in_interactivesymboltable;
  output Boolean out_boolean;
  output String out_string;
  output Interactive.InteractiveSymbolTable out_interactivesymboltable;
algorithm 
  (out_boolean,out_string,out_interactivesymboltable):=
  matchcontinue (in_string,in_interactivesymboltable)
    local
      type Interactive_InteractiveVariableList = list<Interactive.InteractiveVariable>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type TupleAbsyn_PathTupleTypes_TTypeAbsyn_PathOption = tuple<Absyn.Path,TupleTypes_TTypeAbsyn_PathOption>;
      type TupleAbsyn_PathTupleTypes_TTypeAbsyn_PathOptionList = list<TupleAbsyn_PathTupleTypes_TTypeAbsyn_PathOption>;
      type SCode_ClassList = list<SCode.Class>;
      type Interactive_InstantiatedClassList = list<Interactive.InstantiatedClass>;
      String str,msg,res_1,res,resultstr,expmsg,debugstr;
      Interactive.InteractiveSymbolTable isymb,newisymb;
      Absyn.Program p,p_1,newprog,iprog;
      Interactive_InteractiveVariableList vars_1,vars;
      TupleAbsyn_PathTupleTypes_TTypeAbsyn_PathOptionList cf_1,cf;
      SCode_ClassList a;
      Interactive_InstantiatedClassList b;
      Interactive.InteractiveStmts exp;
    case (str,isymb)
      equation 
        true = strncmp("quit()", str, 6); then (false,"Ok\n",isymb);
    case (str,Interactive.SYMBOLTABLE(iprog,a,b,vars,cf))
      equation 
        Debug.fcall0("dump", Print.clear_buf);
        Debug.fcall0("dumpgraphviz", Print.clear_buf);
        Debug.fprint("dump", "\nTrying to parse class definition...\n");
        (p,msg) = Parser.parsestring(str);
        equality(msg = "Ok") "Always succeeds, check msg for errors ";
        p_1 = Interactive.add_scope(p, vars);
        vars_1 = Interactive.update_scope(p, vars);
        newprog = Interactive.update_program(p_1, iprog);
        cf_1 = Interactive.remove_compiled_functions(p, cf);
        Debug.fprint("dump", 
          "\n--------------- Parsed program ---------------\n");
        Debug.fcall("dumpgraphviz", DumpGraphviz.dump, newprog);
        Debug.fcall("dump", Dump.dump, newprog);
        res_1 = make_debug_result("dump", msg);
        res = make_debug_result("dumpgraphviz", res_1); then (true,res,Interactive.SYMBOLTABLE(newprog,a,b,vars_1,cf_1));
    case (str,isymb)
      equation 
        Debug.fcall0("dump", Print.clear_buf);
        Debug.fcall0("dumpgraphviz", Print.clear_buf);
        Debug.fprint("dump", 
          "\nNot a class definition, trying expresion parser\n");
        (exp,msg) = Parser.parsestringexp(str);
        equality(msg = "Ok") "always succeeds, check msg for errors ";
        (resultstr,newisymb) = Interactive.evaluate(exp, isymb, false);
        Debug.fprint("dump", 
          "\n--------------- Parsed expression ---------------\n");
        Debug.fcall("dump", Dump.dump_istmt, exp);
        res_1 = make_debug_result("dump", resultstr);
        res = make_debug_result("dumpgraphviz", res_1); then (true,res,newisymb);
    case (str,isymb)
      local Interactive.InteractiveStmts p;
      equation 
        Debug.fcall0("failtrace", Print.clear_buf);
        (p,msg) = Parser.parsestring(str);
        (p,expmsg) = Parser.parsestringexp(str);
        failure(equality(msg = "Ok"));
        failure(equality(expmsg = "Ok"));
        Debug.fprint("failtrace", 
          "\nBoth parser and expression parser failed: \n");
        Debug.fprintl("failtrace", {"parser: \n",msg,"\n"});
        Debug.fprintl("failtrace", {"expparser: \n",expmsg,"\n"});
        res = make_debug_result("failtrace", msg); then (true,res,isymb);
    case (_,isymb)
      equation 
        Print.print_buf("Error occured building AST\n");
        debugstr = Print.get_string();
        str = string_append(debugstr, "Syntax Error\n"); then (true,str,isymb);
  end matchcontinue;
end handle_command;

protected function is_modelica_file
  input String in_string;
algorithm 
  _:=
  matchcontinue (in_string)
    local
      type StringList = list<String>;
      StringList lst;
      String last,filename;
    case (filename)
      equation 
        lst = System.strtok(filename, ".");
        last :: _ = list_reverse(lst);
        equality(last = "mo"); then ();
    case (filename)
      equation 
        lst = System.strtok(filename, ".");
        last :: _ = list_reverse(lst);
        equality(last = "mof"); then ();
  end matchcontinue;
end is_modelica_file;

protected function is_flat_modelica_file
  input String filename;
protected 
  type StringList = list<String>;
  StringList lst;
  String last;
algorithm 
  lst := System.strtok(filename, ".");
  last :: _ := list_reverse(lst);
  equality(last := "mof");
end is_flat_modelica_file;

protected function is_modelica_script_file
  input String filename;
protected 
  type StringList = list<String>;
  StringList lst;
  String last;
algorithm 
  lst := System.strtok(filename, ".");
  last :: _ := list_reverse(lst);
  equality(last := "mos");
end is_modelica_script_file;

protected function transform_if_flat " relation: transform_if_flat
  Checks is a modelica file is a flat modelica file
  and if so, translates all variables on the form a.b.c to a_b_c
 "
  input String in_string;
  input DAE.DAElist in_daelist;
  output DAE.DAElist out_daelist;
algorithm 
  out_daelist:=
  matchcontinue (in_string,in_daelist)
    local
      DAE.DAElist d,dae;
      String f;
    case (f,dae)
      equation 
        is_flat_modelica_file(f);
        d = DAE.to_modelica_form(dae); then d;
    case (_,d) then d; 
  end matchcontinue;
end transform_if_flat;

protected function is_not_function
  input SCode.Class cls;
  output Boolean res;
algorithm 
  res := SCode.is_function(cls);
  res := bool_not(res);
end is_not_function;

protected function remove_functions
  input SCode.Program in_program;
  output SCode.Program out_program;
algorithm 
  out_program:=
  matchcontinue (in_program)
    local
      type SCode_ClassList = list<SCode.Class>;
      SCode_ClassList classes_1,classes;
    case classes
      equation 
        classes_1 = Util.list_select(classes, is_not_function); then classes_1;
    case _
      equation 
        print("#-- remove_functions failed"); then fail();
  end matchcontinue;
end remove_functions;

protected function keep_only_functions
  input SCode.Program in_program;
  output SCode.Program out_program;
algorithm 
  out_program:=
  matchcontinue (in_program)
    local
      type SCode_ClassList = list<SCode.Class>;
      SCode_ClassList classes_1,classes;
    case classes
      equation 
        classes_1 = Util.list_select(classes, SCode.is_function); then classes_1;
    case _
      equation 
        print("#-- keep_only_functions failed"); then fail();
  end matchcontinue;
end keep_only_functions;

protected function translate_file " relation: translate_file
 
  This function invokes the translator on a source file.  The
  argument should be a list with a single file name.
 "
  input StringList in_stringlist;
protected 
  type StringList = list<String>;
algorithm 
  _:=
  matchcontinue (in_stringlist)
    local
      type SCode_ClassList = list<SCode.Class>;
      Absyn.Program p;
      SCode_ClassList p_1;
      DAE.DAElist d_1,d;
      String s,str,f,res;
      Absyn.Path cname;
      Boolean silent,notsilent;
      Interactive.InteractiveStmts stmts;
      Interactive.InteractiveSymbolTable newst;
    case {f}
      equation 
        is_modelica_file(f) "Print.print_buf \"Parsing\n\" &	 ";
        p = Parser.parse(f);
        Debug.fprint("dump", 
          "\n--------------- Parsed program ---------------\n");
        Debug.fcall("dumpgraphviz", DumpGraphviz.dump, p);
        Debug.fcall("dump", Dump.dump, p);
        Debug.fprint("info", 
          "\n------------------------------------------------------------ \n");
        Debug.fprint("info", "---elaborating\n");
        p_1 = SCode.elaborate(p) "	 & Dump.dump p'  	& Debug.fprint (\"info\", \"---Removing functions\n\")
	& remove_functions p' => pnofunc'
	& Debug.fprint (\"info\", \"---Removed functions\n---Filtering functions\n\")
	& keep_only_functions p' => pfunc'
	& Debug.fprint (\"info\", \"---Filtered functions\n\")
	& Debug.fprint (\"info\", \"---instantiating\n\")
";
        Debug.fprint("info", 
          "\n------------------------------------------------------------ \n") "	 & Dump.dump p'  	& Debug.fprint (\"info\", \"---Removing functions\n\")
	& remove_functions p' => pnofunc'
	& Debug.fprint (\"info\", \"---Removed functions\n---Filtering functions\n\")
	& keep_only_functions p' => pfunc'
	& Debug.fprint (\"info\", \"---Filtered functions\n\")
	& Debug.fprint (\"info\", \"---instantiating\n\")
";
        Debug.fprint("info", "---instantiating\n") "	 & Dump.dump p'  	& Debug.fprint (\"info\", \"---Removing functions\n\")
	& remove_functions p' => pnofunc'
	& Debug.fprint (\"info\", \"---Removed functions\n---Filtering functions\n\")
	& keep_only_functions p' => pfunc'
	& Debug.fprint (\"info\", \"---Filtered functions\n\")
	& Debug.fprint (\"info\", \"---instantiating\n\")
";
        d_1 = Inst.instantiate(p_1);
        Debug.fprint("beforefixmodout", "Explicit part:\n");
        Debug.fcall("beforefixmodout", DAE.dump_debug, d_1) "	& Inst.instantiate_implicit(pfunc') => dimpl'
	& Debug.fprint (\"beforefixmodout\", \"Implicit part:\n\")
	& Debug.fcall (\"beforefixmodout\", DAE.dump_debug, dimpl')
";
        d = fix_modelica_output(d_1) "	& fix_modelica_output (dimpl') => dimpl
 & Debug.	print \"\nJust instantiated\" ";
        Print.clear_buf() "	& fix_modelica_output (dimpl') => dimpl
 & Debug.	print \"\nJust instantiated\" ";
        Debug.fprint("info", "---dumping\n") "	& fix_modelica_output (dimpl') => dimpl
 & Debug.	print \"\nJust instantiated\" ";
        s = Debug.fcallret("flatmodelica", DAE.dump_str, d, "");
        Debug.fcall("flatmodelica", Print.print_buf, s) "
	& Debug.fprint (\"flatmodelica\", \"Implicit:\n\")
	& Debug.fcall (\"flatmodelica\", DAE.dump, dimpl)
";
        s = Debug.fcallret("none", DAE.dump_str, d, "");
        Debug.fcall("none", Print.print_buf, s);
        Debug.fcall("daedump", DAE.dump, d);
        Debug.fcall("daedump2", DAE.dump2, d);
        Debug.fcall("daedumpdebug", DAE.dump_debug, d);
        Debug.fcall("daedumpgraphv", DAE.dump_graphviz, d) "& ModUtil.string_prefix_params d => d' & Debug.fcall (\"codegen\", Codegen.generate_functions, d)  ";
        cname = Absyn.last_classname(p) "	& optimize_dae(d, dimpl, cname)
";
        str = Print.get_string();
        silent = RTOpts.silent();
        notsilent = bool_not(silent);
        Debug.bcall(notsilent, print, str);
        optimize_dae(d, d, cname); then ();
    case {f}
      equation 
        is_modelica_script_file(f) "Print.print_buf \"Parsing\n\" &
";
        stmts = Parser.parseexp(f);
        (res,newst) = Interactive.evaluate(stmts, 
          Interactive.SYMBOLTABLE(Absyn.PROGRAM({},Absyn.TOP()),{},{},{},{}), true);
        print(res); then ();
    case _ :: _ :: _
      equation 
        Print.print_error_buf("# Too many arguments\n"); then fail();
    case {}
      equation 
        Print.print_buf("Usage: omc <options> filename \n");
        Print.print_buf("omc accepts .mo (Modelica files) \n");
        Print.print_buf("            .mof (Flat Modelica files) \n");
        Print.print_buf("            .mos (Modelica Script files) \n");
        Print.print_buf("Options:\n========");
        Print.print_buf("+s    Generate simulation code\n");
        Print.print_buf("+q    Run in quiet mode, ouput nothing\n");
        Print.print_buf("+d=flags, set flags: \n");
        Print.print_buf("    blt               apply blt transformation\n");
        Print.print_buf("    interactive       run in interactive mode\n");
        Print.print_buf(
          "    interactiveCorba  run in interactive mode using Corba\n");
        Print.print_buf("    ..., see DEBUG.TXT for further flags\n"); then fail();
  end matchcontinue;
end translate_file;

protected function run_backend_q "Determine if backend, i.e. BLT etc. should be run.
It should be run if either \"blt\" flag is set or if 
parallelization is enabled by giving flag -n=<no proc.>
 "
  output Boolean res_1;
protected 
  Boolean bltflag,sim_cg,par,res,res_1;
  Integer n;
algorithm 
  bltflag := RTOpts.debug_flag("blt");
  sim_cg := RTOpts.simulation_cg();
  n := RTOpts.no_proc();
  par := (n > 0);
  res := bool_or(bltflag, par);
  res_1 := bool_or(res, sim_cg);
end run_backend_q;

protected function optimize_dae "Used for both parallization and for normal execution."
  input DAE.DAElist in_daelist1;
  input DAE.DAElist in_daelist2;
  input Absyn.Path in_path3;
algorithm 
  _:=
  matchcontinue (in_daelist1,in_daelist2,in_path3)
    local
      type IntegerList = list<Integer>;
      type IntegerListArray = IntegerList[:];
      type IntegerVector = Integer[:];
      type IntegerListList = list<IntegerList>;
      DAELow.DAELow dlow,dlow_1;
      IntegerListArray m,mT;
      IntegerVector v1,v2;
      IntegerListList comps;
      DAE.DAElist dae,daeimpl;
      Absyn.Path classname;
    case (dae,daeimpl,classname)
      equation 
        true = run_backend_q();
        dlow = DAELow.lower(dae);
        Debug.fcall("dumpdaelow", DAELow.dump, dlow);
        m = DAELow.incidence_matrix(dlow);
        mT = DAELow.transpose_matrix(m);
        Debug.fcall("bltdump", DAELow.dump_incidence_matrix, m);
        Debug.fcall("bltdump", DAELow.dump_incidence_matrix_t, mT);
        (v1,v2,dlow_1,m,mT) = DAELow.matching_algorithm(dlow, m, mT);
        Debug.fcall("bltdump", DAELow.dump_incidence_matrix, m);
        Debug.fcall("bltdump", DAELow.dump_incidence_matrix_t, mT);
        Debug.fcall("bltdump", DAELow.dump, dlow_1);
        Debug.fcall("bltdump", DAELow.dump_matching, v1);
        (comps) = DAELow.strong_components(m, mT, v1, v2);
        modpar(dlow_1, v1, v2, comps);
        simcodegen(classname, daeimpl, dlow_1, v1, v2, m, mT, comps) "	& DAELow.dump_components(comps) "; then ();
    case (_,_,_)
      equation 
        true = run_backend_q() "so main can print error messages "; then fail();
    case (_,_,_) "If not running backend. "
      equation 
        false = run_backend_q(); then ();
  end matchcontinue;
end optimize_dae;

protected function modpar " The automatic paralellzation module."
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input IntegerListList in_integerlistlist4;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
algorithm 
  _:=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integerlistlist4)
    local
      Integer n,nx,ny,np;
      DAELow.DAELow indexed_dae,indexed_dae_1,dae;
      Real l,b,t1,t2,time;
      String timestr,nps;
      IntegerVector ass1,ass2;
      IntegerListList comps;
    case (_,_,_,_)
      equation 
        n = RTOpts.no_proc() "If modpar not enabled, nproc = 0, return ";
        (n == 0) = true; then ();
    case (dae,ass1,ass2,comps)
      equation 
        print("translating dae.\n") "Otherwise, build task graph  print \"old dae:\" & DAELow.dump dae &";
        indexed_dae = DAELow.translate_dae(dae);
        indexed_dae_1 = DAELow.calculate_values(indexed_dae);
        print("building task graph\n") "	print \"new dae:\" & DAELow.dump indexed_dae' &";
        TaskGraph.build_taskgraph(indexed_dae_1, ass1, ass2, comps);
        TaskGraphExt.dump_graph("model.viz");
        l = RTOpts.latency();
        b = RTOpts.bandwidth();
        t1 = clock();
        TaskGraphExt.merge_tasks(l, b);
        t2 = clock();
        time = t2 -. t1;
        timestr = real_string(time);
        print("task merging took ");
        print(timestr);
        print(" seconds\n");
        TaskGraphExt.dump_merged_graph("merged_model.viz");
        n = RTOpts.no_proc();
        TaskGraphExt.schedule(n);
        (nx,ny,np) = DAELow.calculate_sizes(indexed_dae_1);
        nps = int_string(np);
        print("=======\nnp =");
        print(nps);
        print("=======\n");
        TaskGraphExt.generate_code(nx, ny, np);
        print("done\n"); then ();
    case (_,_,_,_)
      equation 
        Debug.fprint("failtrace", "-modpar failed\n"); then fail();
  end matchcontinue;
end modpar;

protected function simcodegen "ass1 
  ass2 
  blocks "
  input Absyn.Path in_path1;
  input DAE.DAElist in_daelist2;
  input DAELow.DAELow in_daelow3;
  input IntegerVector in_integervector4;
  input IntegerVector in_integervector5;
  input DAELow.IncidenceMatrix in_incidencematrix6;
  input DAELow.IncidenceMatrixT in_incidencematrixt7;
  input IntegerListList in_integerlistlist8;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
algorithm 
  _:=
  matchcontinue (in_path1,in_daelist2,in_daelow3,in_integervector4,in_integervector5,in_incidencematrix6,in_incidencematrixt7,in_integerlistlist8)
    local
      type IntegerListArray = IntegerList[:];
      DAELow.DAELow indexed_dlow,indexed_dlow_1,dlow;
      String cname_str,filename,funcfilename,init_filename,makefilename;
      Absyn.Path classname;
      DAE.DAElist dae;
      IntegerVector ass1,ass2;
      IntegerListArray m,mt;
      IntegerListList comps;
    case (classname,dae,dlow,ass1,ass2,m,mt,comps) "classname  ass1  ass2  blocks "
      equation 
        true = RTOpts.simulation_cg();
        Print.clear_error_buf();
        Print.clear_buf();
        indexed_dlow = DAELow.translate_dae(dlow);
        indexed_dlow_1 = DAELow.calculate_values(indexed_dlow);
        cname_str = Absyn.path_string(classname);
        filename = Util.string_append_list({cname_str,".cpp"});
        funcfilename = Util.string_append_list({cname_str,"_functions.cpp"});
        init_filename = Util.string_append_list({cname_str,"_init.txt"});
        makefilename = Util.string_append_list({cname_str,".makefile"});
        SimCodegen.generate_functions(dae, indexed_dlow_1, classname, funcfilename);
        SimCodegen.generate_simulation_code(dae, indexed_dlow_1, ass1, ass2, m, mt, comps, classname, 
          filename);
        SimCodegen.generate_init_data(indexed_dlow_1, classname, init_filename, 0.0, 1.0, 0.01);
        SimCodegen.generate_makefile(makefilename, classname); then ();
    case (_,_,_,_,_,_,_,_) "If something above failed. fail so Main can print errors "
      equation 
        true = RTOpts.simulation_cg(); then fail();
    case (_,_,_,_,_,_,_,_) "If not generating simulation code"
      equation 
        false = RTOpts.simulation_cg(); then ();
  end matchcontinue;
end simcodegen;

protected function run_modpar_q
  output Boolean res;
protected 
  Integer n;
algorithm 
  n := RTOpts.no_proc();
  res := (n > 0);
end run_modpar_q;

protected function fix_modelica_output
  input DAE.DAElist in_daelist;
  output DAE.DAElist out_daelist;
algorithm 
  out_daelist:=
  matchcontinue (in_daelist)
    local
      type DAE_ElementList = list<DAE.Element>;
      DAE_ElementList dae_1,dae;
      DAE.DAElist d;
    case DAE.DAE(dae)
      equation 
        true = RTOpts.modelica_output();
        dae_1 = Inst.init_vars_modelica_output(dae); then DAE.DAE(dae_1);
    case ((d = DAE.DAE(dae)))
      equation 
        false = RTOpts.modelica_output(); then d;
  end matchcontinue;
end fix_modelica_output;

protected function interactivemode
  input StringList in_stringlist;
protected 
  type StringList = list<String>;
algorithm 
  _:=
  matchcontinue (in_stringlist)
    local Integer shandle;
    case _
      equation 
        shandle = Socket.waitforconnect(29500);
        _ = server_loop(shandle, 
          Interactive.SYMBOLTABLE(Absyn.PROGRAM({},Absyn.TOP()),{},{},{},{})); then ();
  end matchcontinue;
end interactivemode;

protected function interactivemode_corba
  input StringList in_stringlist;
protected 
  type StringList = list<String>;
algorithm 
  _:=
  matchcontinue (in_stringlist)
    case _
      equation 
        Corba.initialize();
        _ = server_loop_corba(
          Interactive.SYMBOLTABLE(Absyn.PROGRAM({},Absyn.TOP()),{},{},{},{})); then ();
  end matchcontinue;
end interactivemode_corba;

protected function server_loop_corba " relation: server_loop_corba
 
  This function is the main loop of the server for a CORBA impl.
"
  input Interactive.InteractiveSymbolTable in_interactivesymboltable;
  output Interactive.InteractiveSymbolTable out_interactivesymboltable;
algorithm 
  out_interactivesymboltable:=
  matchcontinue (in_interactivesymboltable)
    local
      String str,replystr;
      Interactive.InteractiveSymbolTable newsymb,ressymb,isymb;
    case (isymb)
      equation 
        str = Corba.wait_for_command();
        Print.clear_buf();
        (true,replystr,newsymb) = handle_command(str, isymb);
        Corba.sendreply(replystr);
        ressymb = server_loop_corba(newsymb); then ressymb;
    case (isymb)
      equation 
        str = Corba.wait_for_command() "start - 2005-06-12 - adrpo added this part to make the loop deterministic ";
        Print.clear_buf() "start - 2005-06-12 - adrpo added this part to make the loop deterministic ";
        (false,replystr,newsymb) = handle_command(str, isymb);
        Print.print_buf("Exiting\n");
        Corba.sendreply("quit requested, shutting server down\n");
        Corba.close(); then isymb;
  end matchcontinue;
end server_loop_corba;

public function main " relation: main
 
  This is the main function that the RML runtime system calls to
  start the translation.
 "
  input StringList in_stringlist;
protected 
  type StringList = list<String>;
algorithm 
  _:=
  matchcontinue (in_stringlist)
    local
      StringList args_1,args;
      Boolean ismode,icmode,imode,imode_1;
      String str;
    case args
      equation 
        args_1 = RTOpts.args(args);
        ismode = RTOpts.debug_flag("interactive");
        icmode = RTOpts.debug_flag("interactiveCorba");
        imode = bool_or(ismode, icmode);
        imode_1 = bool_not(imode);
        Debug.bcall(ismode, interactivemode, args_1);
        Debug.bcall(icmode, interactivemode_corba, args_1);
        Debug.bcall(imode_1, translate_file, args_1); then ();
    case _
      equation 
        str = Print.get_error_string() "If main fails, retrieve error messages and print to std out. ";
        print(str); then fail();
  end matchcontinue;
end main;
end Main;

