package ModUtil "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 ModUtil.rml
  module:      ModUtil
  description: Miscellanous modelica related utilities (The horror, THE HORROR)
 
  RCS: $Id: ModUtil.rml 1733 2005-04-27 06:47:45Z haklu $
  
  This module contains various utilities. For example
  converting a path to a string and comparing two paths.
  It is used pretty much everywhere. The difference between this 
  module and the Util module is that ModUtil contains modelica 
  related utilities. The Util module only contains \"low-level\" 
  rml utilities, for example finding elements in lists.
  
 "

public import OpenModelica.Compiler.Absyn;

public import OpenModelica.Compiler.DAE;

public import OpenModelica.Compiler.Exp;

protected import OpenModelica.Compiler.RTOpts;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Print;

public function string_prefix_component_refs
  input String in_string;
  input FuncType1 in_functype1;
  input Type_b in_type_b;
  input Exp_ExpList in_exp_explist;
  output Exp_ExpList out_exp_explist;
protected 
  function FuncType1
    input Exp.ComponentRef in_componentref;
    input Type_b in_type_b;
  protected 
    replaceable type Type_b;
  end FuncType1;
  replaceable type Type_b;
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  out_exp_explist:=
  matchcontinue (in_string,in_functype1,in_type_b,in_exp_explist)
    local
      Exp_ExpList res,rest;
      Exp.Exp e_1,e;
      String str;
      FuncType1 r;
      Type_b rarg;
    case (_,_,_,{}) then {}; 
    case (str,r,rarg,e :: rest)
      equation 
        res = string_prefix_component_refs(str, r, rarg, rest);
        e_1 = string_prefix_component_ref(str, r, rarg, e); then e_1 :: res;
  end matchcontinue;
end string_prefix_component_refs;

public function string_prefix_component_ref
  input String in_string;
  input FuncType1 in_functype1;
  input Type_b in_type_b;
  input Exp.Exp in_exp;
  output Exp.Exp out_exp;
protected 
  function FuncType1
    input Exp.ComponentRef in_componentref;
    input Type_b in_type_b;
  protected 
    replaceable type Type_b;
  end FuncType1;
  replaceable type Type_b;
algorithm 
  out_exp:=
  matchcontinue (in_string,in_functype1,in_type_b,in_exp)
    local
      type Exp_ExpList = list<Exp.Exp>;
      type Exp_ExpListList = list<Exp_ExpList>;
      type BooleanList = list<Boolean>;
      type BooleanListList = list<BooleanList>;
      type TupleExp_ExpBoolean = tuple<Exp.Exp,Boolean>;
      type TupleExp_ExpBooleanList = list<TupleExp_ExpBoolean>;
      type TupleExp_ExpBooleanListList = list<TupleExp_ExpBooleanList>;
      Exp.ComponentRef cr_1,cr;
      String str;
      FuncType1 r;
      Type_b rarg;
      Exp.Type t,ty;
      Exp.Exp e1_1,e2_1,e1,e2,e3_1,e3,e;
      Exp.Operator op;
      Exp_ExpList el_1,el;
      Absyn.Path p;
      Boolean b,bi,a;
      BooleanListList bl;
      TupleExp_ExpBooleanListList ell_1,ell;
      Integer i;
    case (str,r,rarg,Exp.CREF(cr,t))
      equation 
        r(cr, rarg);
        cr_1 = string_prefix_cref(str, cr); then Exp.CREF(cr_1,t);
    case (_,r,rarg,Exp.CREF(cr,t))
      equation 
        failure(r(cr, rarg)); then Exp.CREF(cr,t);
    case (str,r,rarg,Exp.BINARY(e1,op,e2))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1);
        e2_1 = string_prefix_component_ref(str, r, rarg, e2); then Exp.BINARY(e1_1,op,e2_1);
    case (str,r,rarg,Exp.UNARY(op,e1))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1); then Exp.UNARY(op,e1_1);
    case (str,r,rarg,Exp.LBINARY(e1,op,e2))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1);
        e2_1 = string_prefix_component_ref(str, r, rarg, e2); then Exp.LBINARY(e1_1,op,e2_1);
    case (str,r,rarg,Exp.LUNARY(op,e1))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1); then Exp.LUNARY(op,e1_1);
    case (str,r,rarg,Exp.RELATION(e1,op,e2))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1);
        e2_1 = string_prefix_component_ref(str, r, rarg, e2); then Exp.RELATION(e1_1,op,e2_1);
    case (str,r,rarg,Exp.IFEXP(e1,e2,e3))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1);
        e2_1 = string_prefix_component_ref(str, r, rarg, e2);
        e3_1 = string_prefix_component_ref(str, r, rarg, e3); then Exp.IFEXP(e1_1,e2_1,e3_1);
    case (str,r,rarg,Exp.CALL(p,el,b,bi))
      equation 
        el_1 = string_prefix_component_refs(str, r, rarg, el); then Exp.CALL(p,el_1,b,bi);
    case (str,r,rarg,Exp.ARRAY(t,a,el))
      equation 
        el_1 = string_prefix_component_refs(str, r, rarg, el); then Exp.ARRAY(t,a,el_1);
    case (str,r,rarg,Exp.MATRIX(t,a,ell))
      local
        Exp_ExpListList el,el_1;
        Integer a;
      equation 
        el = Util.list_list_map(ell, Util.tuple2_1);
        bl = Util.list_list_map(ell, Util.tuple2_2);
        el_1 = string_prefix_component_refs_list(str, r, rarg, el);
        ell_1 = Util.list_list_thread_tuple(el_1, bl); then Exp.MATRIX(t,a,ell_1);
    case (str,r,rarg,Exp.RANGE(t,e1,NONE,e2))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1);
        e2_1 = string_prefix_component_ref(str, r, rarg, e2); then Exp.RANGE(t,e1_1,NONE,e2_1);
    case (str,r,rarg,Exp.RANGE(t,e1,SOME(e2),e3))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1);
        e2_1 = string_prefix_component_ref(str, r, rarg, e2);
        e3_1 = string_prefix_component_ref(str, r, rarg, e3); then Exp.RANGE(t,e1_1,SOME(e2_1),e3_1);
    case (str,r,rarg,Exp.TUPLE(el))
      equation 
        el_1 = string_prefix_component_refs(str, r, rarg, el); then Exp.TUPLE(el_1);
    case (str,r,rarg,Exp.CAST(ty,e1))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1); then Exp.CAST(ty,e1_1);
    case (str,r,rarg,Exp.ASUB(e1,i))
      equation 
        e1_1 = string_prefix_component_ref(str, r, rarg, e1); then Exp.ASUB(e1_1,i);
    case (str,r,rarg,e) then e; 
  end matchcontinue;
end string_prefix_component_ref;

protected function string_prefix_component_refs_list
  input String in_string;
  input FuncType1 in_functype1;
  input Type_b in_type_b;
  input Exp_ExpListList in_exp_explistlist;
  output Exp_ExpListList out_exp_explistlist;
protected 
  function FuncType1
    input Exp.ComponentRef in_componentref;
    input Type_b in_type_b;
  protected 
    replaceable type Type_b;
  end FuncType1;
  replaceable type Type_b;
  type Exp_ExpList = list<Exp.Exp>;
  type Exp_ExpListList = list<Exp_ExpList>;
algorithm 
  out_exp_explistlist:=
  matchcontinue (in_string,in_functype1,in_type_b,in_exp_explistlist)
    local
      Exp_ExpList el_1,el;
      Exp_ExpListList res,rest;
      String str;
      FuncType1 r;
      Type_b rarg;
    case (_,_,_,{}) then {}; 
    case (str,r,rarg,el :: rest)
      equation 
        el_1 = string_prefix_component_refs(str, r, rarg, el);
        res = string_prefix_component_refs_list(str, r, rarg, rest); then el_1 :: res;
  end matchcontinue;
end string_prefix_component_refs_list;

protected function string_prefix_cref
  input String in_string;
  input Exp.ComponentRef in_componentref;
  output Exp.ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_string,in_componentref)
    local
      type Exp_SubscriptList = list<Exp.Subscript>;
      String s_1,str,s;
      Exp_SubscriptList si;
      Exp.ComponentRef cr;
    case (str,Exp.CREF_IDENT(s,si))
      equation 
        s_1 = string_append(str, s); then Exp.CREF_IDENT(s_1,si);
    case (str,Exp.CREF_QUAL(s,si,cr))
      equation 
        s_1 = string_append(str, s); then Exp.CREF_QUAL(s_1,si,cr);
  end matchcontinue;
end string_prefix_cref;

public function string_prefix_elements
  input String in_string1;
  input DAE_ElementList in_dae_elementlist2;
  input DAE_ElementList in_dae_elementlist3;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_string1,in_dae_elementlist2,in_dae_elementlist3)
    local
      DAE.Element el_1,el;
      DAE_ElementList res,dae,rest;
      String str;
    case (_,_,{}) then {}; 
    case (str,dae,el :: rest)
      equation 
        el_1 = string_prefix_element(str, dae, el);
        res = string_prefix_elements(str, dae, rest); then el_1 :: res;
  end matchcontinue;
end string_prefix_elements;

protected function string_prefix_element
  input String in_string;
  input DAE_ElementList in_dae_elementlist;
  input DAE.Element in_element;
  output DAE.Element out_element;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_element:=
  matchcontinue (in_string,in_dae_elementlist,in_element)
    local
      type Exp_SubscriptList = list<Exp.Subscript>;
      type Exp_ExpOption = Option<Exp.Exp>;
      type Absyn_PathList = list<Absyn.Path>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.Exp exp_1,exp,exp1_1,exp2_1,exp1,exp2;
      String str,n;
      DAE_ElementList dae,dae_1,dae1;
      Exp.ComponentRef cr;
      DAE.VarKind vk;
      DAE.VarDirection vd;
      DAE.Type ty;
      Exp_SubscriptList inst_dims;
      Exp_ExpOption start;
      DAE.Flow flow_;
      Absyn_PathList cl;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      Algorithm.Algorithm alg;
      DAE.ExternalDecl decl;
      DAE.Element e;
    case (str,dae,DAE.VAR(cr,vk,vd,ty,SOME(exp),inst_dims,start,flow_,cl,dae_var_attr,comment))
      equation 
        exp_1 = string_prefix_component_ref(str, is_parameter_daelist, dae, exp); then DAE.VAR(cr,vk,vd,ty,SOME(exp_1),inst_dims,start,flow_,cl,
          dae_var_attr,comment);
    case (str,dae,DAE.DEFINE(cr,exp))
      equation 
        exp_1 = string_prefix_component_ref(str, is_parameter_daelist, dae, exp); then DAE.DEFINE(cr,exp_1);
    case (str,dae,DAE.EQUATION(exp1,exp2))
      equation 
        exp1_1 = string_prefix_component_ref(str, is_parameter_daelist, dae, exp1);
        exp2_1 = string_prefix_component_ref(str, is_parameter_daelist, dae, exp2); then DAE.EQUATION(exp1_1,exp2_1);
    case (str,dae,DAE.ALGORITHM(alg)) then DAE.ALGORITHM(alg); 
    case (str,dae1,DAE.COMP(n,DAE.DAE(dae))) " What happens if a variable is not found among dae, should we check dae1,
    i.e. where the COMP and FUNCTION was found? *"
      equation 
        dae_1 = string_prefix_elements(str, dae, dae); then DAE.COMP(n,DAE.DAE(dae_1));
    case (str,dae1,DAE.FUNCTION(n,DAE.DAE(dae),ty))
      local
        Absyn.Path n;
        TupleTypes_TTypeAbsyn_PathOption ty;
      equation 
        dae_1 = string_prefix_elements(str, dae, dae); then DAE.FUNCTION(n,DAE.DAE(dae_1),ty);
    case (str,dae1,DAE.EXTFUNCTION(n,DAE.DAE(dae),ty,decl))
      local
        Absyn.Path n;
        TupleTypes_TTypeAbsyn_PathOption ty;
      equation 
        dae_1 = string_prefix_elements(str, dae, dae); then DAE.EXTFUNCTION(n,DAE.DAE(dae_1),ty,decl);
    case (str,dae,e) then e; 
  end matchcontinue;
end string_prefix_element;

protected function is_parameter_daelist
  input Exp.ComponentRef in_componentref;
  input DAE_ElementList in_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  _:=
  matchcontinue (in_componentref,in_dae_elementlist)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      Exp.ComponentRef cr,crv;
      DAE.VarDirection vd;
      DAE.Type ty;
      Exp_ExpOption e;
      DAE_ElementList rest;
      DAE.VarKind vk;
    case (cr,DAE.VAR(crv,DAE.PARAM(),vd,ty,e,_,_,_,_,_,_) :: rest)
      equation 
        true = Exp.cref_equal(cr, crv); then ();
    case (cr,DAE.VAR(crv,vk,vd,ty,e,_,_,_,_,_,_) :: rest)
      equation 
        true = Exp.cref_equal(cr, crv); then fail();
    case (cr,e :: rest)
      local DAE.Element e;
      equation 
        is_parameter_daelist(cr, rest); then ();
  end matchcontinue;
end is_parameter_daelist;

public function string_prefix_params
  input DAE.DAElist in_daelist;
  output DAE.DAElist out_daelist;
algorithm 
  out_daelist:=
  matchcontinue (in_daelist)
    local
      type DAE_ElementList = list<DAE.Element>;
      DAE_ElementList dae_1,dae;
    case DAE.DAE(dae)
      equation 
        dae_1 = string_prefix_elements("params->", dae, dae); then DAE.DAE(dae_1);
  end matchcontinue;
end string_prefix_params;

protected function get_component_ref_next
  input Exp.ComponentRef in_componentref;
  output Exp.ComponentRef out_componentref;
algorithm 
  out_componentref:=
  matchcontinue (in_componentref)
    local
      type Exp_SubscriptList = list<Exp.Subscript>;
      String n;
      Exp_SubscriptList sl;
      Exp.ComponentRef cr;
    case Exp.CREF_QUAL(n,sl,cr) then cr; 
    case ((cr = Exp.CREF_IDENT(n,sl)))
      equation 
        Print.print_buf(
          "#-- Cannot get subname of an unqualified component reference\n");
        Exp.print_component_ref(cr); then fail();
  end matchcontinue;
end get_component_ref_next;

protected function get_var_value_expr
  input DAE_ElementList in_dae_elementlist;
  input Exp.ComponentRef in_componentref;
  output Exp.Exp out_exp;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_exp:=
  matchcontinue (in_dae_elementlist,in_componentref)
    local
      Exp.ComponentRef cr2,cr;
      DAE.VarKind vk;
      DAE.VarDirection vd;
      DAE.Type ty;
      DAE_ElementList dae;
      Exp.Exp value;
    case (dae,cr)
      equation 
        SOME(DAE.VAR(cr2,vk,vd,ty,NONE,_,_,_,_,_,_)) = find_var(dae, cr);
        Print.print_buf(
          "#-- get_var_value_expr: Warning: Variable is not initialized\n");
        Exp.print_component_ref(cr);
        Print.print_buf("\n"); then Exp.SCONST("UNDEF");
    case (dae,cr)
      equation 
        SOME(DAE.VAR(cr2,vk,vd,ty,SOME(value),_,_,_,_,_,_)) = find_var(dae, cr); then value;
    case (dae,cr)
      equation 
        Print.print_buf("#-- get_var_value_expr: Error: Variable not found\n");
        Exp.print_component_ref(cr);
        Print.print_buf("\n"); then fail();
  end matchcontinue;
end get_var_value_expr;

protected function print_exp_no_funcall
  input Exp.Exp in_exp;
algorithm 
  _:=
  matchcontinue (in_exp)
    local
      type Exp_ExpList = list<Exp.Exp>;
      String fs;
      Absyn.Path fcn;
      Exp_ExpList args;
      Exp.Exp e;
    case Exp.CALL(fcn,args,_,_)
      equation 
        fs = path_string(fcn);
        Print.print_buf(fs); then ();
    case e
      equation 
        Exp.print_exp(e); then ();
  end matchcontinue;
end print_exp_no_funcall;

public function opt_path_string
  input Absyn_PathOption in_absyn_pathoption;
  output String out_string;
protected 
  type Absyn_PathOption = Option<Absyn.Path>;
algorithm 
  out_string:=
  matchcontinue (in_absyn_pathoption)
    local
      String str;
      Absyn.Path p;
    case (NONE) then ""; 
    case (SOME(p))
      equation 
        str = path_string(p); then str;
  end matchcontinue;
end opt_path_string;

public function path_string
  input Absyn.Path in_path;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_path)
    local
      String s;
      Absyn.Path path;
    case path
      equation 
        true = RTOpts.modelica_output();
        s = path_string2(path, "__"); then s;
    case path
      equation 
        false = RTOpts.modelica_output();
        s = path_string2(path, "."); then s;
  end matchcontinue;
end path_string;

public function path_string2
  input Absyn.Path in_path;
  input String in_string;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_path,in_string)
    local
      String s,ns,s1,ss,str;
      Absyn.Path n;
    case (Absyn.IDENT(s),_) then s; 
    case (Absyn.QUALIFIED(s,n),str)
      equation 
        ns = path_string2(n, str);
        s1 = string_append(s, str);
        ss = string_append(s1, ns); then ss;
  end matchcontinue;
end path_string2;

protected function find_var
  input DAE_ElementList in_dae_elementlist;
  input Exp.ComponentRef in_componentref;
  output DAE_ElementOption out_dae_elementoption;
protected 
  type DAE_ElementList = list<DAE.Element>;
  type DAE_ElementOption = Option<DAE.Element>;
algorithm 
  out_dae_elementoption:=
  matchcontinue (in_dae_elementlist,in_componentref)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      DAE.Element e;
      Exp.ComponentRef cr1,cr2,cr;
      DAE.VarKind vk;
      DAE.VarDirection vd;
      DAE.Type ty;
      Exp_ExpOption value;
      DAE_ElementList rest;
      DAE_ElementOption e_1;
    case ({},_) then NONE; 
    case ((e = DAE.VAR(cr1,vk,vd,ty,value,_,_,_,_,_,_)) :: rest,cr2)
      equation 
        true = Exp.cref_equal(cr1, cr2); then SOME(e);
    case (_ :: rest,cr)
      equation 
        e_1 = find_var(rest, cr); then e_1;
  end matchcontinue;
end find_var;

public function path_equal
  input Absyn.Path in_path1;
  input Absyn.Path in_path2;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_path1,in_path2)
    local
      String id1,id2;
      Boolean res;
      Absyn.Path path1,path2;
    case (Absyn.IDENT(id1),Absyn.IDENT(id2))
      equation 
        equality(id1 = id2); then true;
    case (Absyn.IDENT(id1),Absyn.IDENT(id2))
      equation 
        failure(equality(id1 = id2)); then false;
    case (Absyn.QUALIFIED(id1,path1),Absyn.QUALIFIED(id2,path2))
      equation 
        equality(id1 = id2);
        res = path_equal(path1, path2); then res;
    case (Absyn.QUALIFIED(id1,path1),Absyn.QUALIFIED(id2,path2))
      equation 
        failure(equality(id1 = id2)); then false;
    case (_,_) then false; 
  end matchcontinue;
end path_equal;

public function path_subsuffix_equal "This function returns true if the first argument is a sub suffix of the second argument.
  Examples:
 path_subsuffix_equal(A.B, X.Y.A.B) => true
 path_subsuffix_equal(A.B.C, X.Y.A.B) => false
"
  input Absyn.Path in_path1;
  input Absyn.Path in_path2;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_path1,in_path2)
    local
      String p1str,p2str,id1;
      Absyn.Path p1_1,p2_1,p1,p2;
      Boolean res;
    case ((p1 = Absyn.QUALIFIED(_,_)),(p2 = Absyn.QUALIFIED(_,_)))
      equation 
        p1str = Absyn.path_last_ident(p1);
        p2str = Absyn.path_last_ident(p2);
        equality(p1str = p2str);
        p1_1 = Absyn.strip_last(p1);
        p2_1 = Absyn.strip_last(p2);
        res = path_subsuffix_equal(p1_1, p2_1); then res;
    case ((p1 = Absyn.IDENT(id1)),p2)
      equation 
        p2str = Absyn.path_last_ident(p2);
        equality(id1 = p2str); then true;
    case (_,_) then false; 
  end matchcontinue;
end path_subsuffix_equal;
end ModUtil;

