package SCode "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 scode.rml
  module:      SCode
  description: SCode intermediate form
 
  RCS: $Id: SCode.rml 1837 2005-06-28 10:01:44Z petar $
 
  This module contains data structures to describe a Modelica model
  in a more convenient way than the `Absyn' module does.  The most
  important function in this module is the `elaborate' function
  which turns an abstract syntax tree into an SCode representation.
 
  The SCode representaion is used as input to the Inst module
 "

public import OpenModelica.Compiler.Absyn;

public 
type Ident = Absyn.Ident " Some definitions are borrowed from `Absyn' ";

public 
type Path = Absyn.Path;

public 
type Subscript = Absyn.Subscript;

public 
uniontype Restriction
  record R_CLASS
  end R_CLASS;
  record R_MODEL
  end R_MODEL;
  record R_RECORD
  end R_RECORD;
  record R_BLOCK
  end R_BLOCK;
  record R_CONNECTOR
  end R_CONNECTOR;
  record R_TYPE
  end R_TYPE;
  record R_PACKAGE
  end R_PACKAGE;
  record R_FUNCTION
  end R_FUNCTION;
  record R_EXT_FUNCTION "Added c.t. Absyn "
  end R_EXT_FUNCTION;
  record R_ENUMERATION
  end R_ENUMERATION;
  record R_PREDEFINED_INT
  end R_PREDEFINED_INT;
  record R_PREDEFINED_REAL
  end R_PREDEFINED_REAL;
  record R_PREDEFINED_STRING
  end R_PREDEFINED_STRING;
  record R_PREDEFINED_BOOL
  end R_PREDEFINED_BOOL;
  record R_PREDEFINED_ENUM
  end R_PREDEFINED_ENUM;
end Restriction;

public 
type SubModList = list<SubMod>;

type Absyn_ExpOption = Option<Absyn.Exp>;

type ElementList = list<Element>;

uniontype Mod " - Modifications "
  record MOD
    Boolean final_ "final ";
    SubModList submodlist;
    Absyn_ExpOption absyn_expoption;
  end MOD;
  record REDECL
    Boolean final_ "final ";
    ElementList elementlist;
  end REDECL;
  record NOMOD
  end NOMOD;
end Mod;

public 
type SubscriptList = list<Subscript>;

uniontype SubMod
  record NAMEMOD
    Ident ident;
    Mod mod "A named component ";
  end NAMEMOD;
  record IDXMOD
    SubscriptList subscriptlist;
    Mod mod "An array element   Modifications are represented in an more structured way than in 
    the `Absyn' module.  Modifications using qualified names
    (such as in `x.y =  z') are normalized (to `x(y = z)').  And a
    special case when arrays are subscripted in a modification.
     - Programs ";
  end IDXMOD;
end SubMod;

public 
type Program = list<Class>;

public 
uniontype Class " As in the AST, a program is simply a list of class definitions. 
   - Classes "
  record CLASS
    Ident name "Name ";
    Boolean partial_ "Partial ";
    Boolean encapsulated_ "Encapsulated ";
    Restriction restricion "Restricion ";
    ClassDef parts "Parts ";
  end CLASS;
end Class;

public 
type EquationList = list<Equation>;

type AlgorithmList = list<Algorithm>;

type Absyn_ExternalDeclOption = Option<Absyn.ExternalDecl>;

type Absyn_ArrayDimOption = Option<Absyn.ArrayDim>;

type IdentList = list<Ident>;

type Absyn_PathList = list<Absyn.Path>;

uniontype ClassDef
  record PARTS
    ElementList elementlist1;
    EquationList equationlist2;
    EquationList equationlist3 "InitialEquation ";
    AlgorithmList algorithmlist4;
    AlgorithmList algorithmlist5 "InitialAlgorithm ";
    Absyn_ExternalDeclOption absyn_externaldecloption6 "Used by external functions";
  end PARTS;
  record DERIVED
    Path path "short class definition ";
    Absyn_ArrayDimOption absyn_arraydimoption;
    Mod mod;
  end DERIVED;
  record ENUMERATION
    IdentList identlist;
  end ENUMERATION;
  record OVERLOAD
    Absyn_PathList absyn_pathlist " The major difference between these types and their `Absyn' ";
  end OVERLOAD;
end ClassDef;

public 
type Absyn_PathOption = Option<Absyn.Path>;

uniontype Equation " counterparts is that the `PARTS' constructor contains separate 
   lists for elements, equations and algorithms. 
  - Equations "
  record EQUATION
    EEquation eequation;
    Absyn_PathOption absyn_pathoption "baseclassname if in bclass";
  end EQUATION;
end Equation;

public 
type EEquationList = list<EEquation>;

type TupleAbsyn_ExpEEquationList = tuple<Absyn.Exp,EEquationList>;

type TupleAbsyn_ExpEEquationListList = list<TupleAbsyn_ExpEEquationList>;

uniontype EEquation
  record EQ_IF
    Absyn.Exp conditional1 "conditional ";
    EEquationList eequationlist2 "true branch ";
    EEquationList eequationlist3 "false branch ";
  end EQ_IF;
  record EQ_EQUALS
    Absyn.Exp exp1;
    Absyn.Exp exp2;
  end EQ_EQUALS;
  record EQ_CONNECT
    Absyn.ComponentRef componentref1;
    Absyn.ComponentRef componentref2;
  end EQ_CONNECT;
  record EQ_FOR
    Ident ident;
    Absyn.Exp exp;
    EEquationList eequationlist;
  end EQ_FOR;
  record EQ_WHEN
    Absyn.Exp exp;
    EEquationList eequationlist;
    TupleAbsyn_ExpEEquationListList tupleabsyn_expeequationlistlist;
  end EQ_WHEN;
  record EQ_ASSERT
    Absyn.Exp exp1;
    Absyn.Exp exp2 "condition message string ";
  end EQ_ASSERT;
  record EQ_REINIT
    Absyn.ComponentRef componentref;
    Absyn.Exp exp "state variable  the new value ";
  end EQ_REINIT;
end EEquation;

public 
type Absyn_AlgorithmList = list<Absyn.Algorithm>;

uniontype Algorithm " These are almost identical to the `Absyn' versions.  In `EQ_IF', 
   the `elseif' branches are represented as normal `else' branches 
   with a single `if' statement in them. 
   - Algorithms "
  record ALGORITHM
    Absyn_AlgorithmList absyn_algorithmlist;
    Absyn_PathOption absyn_pathoption "baseclass name if in baseclass  The `Absyn' module uses the terminology from the grammar, where ";
  end ALGORITHM;
end Algorithm;

public 
type PathOption = Option<Path>;

type Absyn_CommentOption = Option<Absyn.Comment>;

uniontype Element " `algorithm' means an algorithmic statement.  But here, 
   `Algorithm' means a whole algorithm section. 
   - Elements "
  record EXTENDS
    Path path;
    Mod mod;
  end EXTENDS;
  record CLASSDEF
    Ident name "name ";
    Boolean final_ "final ";
    Boolean replaceable_ "replaceable ";
    Class class_;
    PathOption pathoption "baseclass name if in baseclass";
  end CLASSDEF;
  record IMPORT
    Absyn.Import import_;
  end IMPORT;
  record COMPONENT
    Ident ident "component name ";
    Boolean final_ "final ";
    Boolean replaceable_ "replaceable ";
    Boolean protected_ "protected ";
    Attributes attributes;
    Path path "type name ";
    Mod mod;
    PathOption pathoption "baseclass name if in baseclass";
    Absyn_CommentOption absyn_commentoption "this if for extraction comments and annotations from Absyn ";
  end COMPONENT;
end Element;

public 
uniontype Attributes " There are four types of elements in a declaration, represented 
   by the constructors `EXTENDS' (for `extends' clauses), 
   `CLASSDEF' (for local class definitions)  `COMPONENT' (for 
   local variables). and `IMPORT' (for `import' clauses) 
   The baseclass name is initially NONE in the translation, and
    if an element is inherited from a base class it is filled in during the
    instantiation process. 
   - Attributes "
  record ATTR
    Absyn.ArrayDim arraydim;
    Boolean flow_ "flow ";
    Accessibility accessibility "RW, RO, WO ";
    Variability parameter_ "parameter ";
    Absyn.Direction direction "input, output or bidirectional ";
  end ATTR;
end Attributes;

public 
uniontype Variability
  record VAR
  end VAR;
  record DISCRETE
  end DISCRETE;
  record PARAM
  end PARAM;
  record STRUCTPARAM
  end STRUCTPARAM;
  record CONST
  end CONST;
end Variability;

public 
uniontype Accessibility
  record RW "read/write "
  end RW;
  record RO "read-only "
  end RO;
  record WO "write-only (not used)   These are attributes that apply to a declared component. "
  end WO;
end Accessibility;

protected import OpenModelica.Compiler.Dump;

protected import OpenModelica.Compiler.Debug;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.Util;

public function elaborate " relation: elaborate
 
  This function takes an `Absyn.Program' and constructs a `Program'
  from it.
 "
  input Absyn.Program in_program;
  output Program out_program;
algorithm 
  out_program:=
  matchcontinue (in_program)
    local
      type Absyn_ClassList = list<Absyn.Class>;
      Class c_1;
      Program cs_1;
      Absyn.Class c;
      Absyn_ClassList cs;
      Absyn.Within w;
    case (Absyn.PROGRAM({},_)) then {}; 
    case (Absyn.PROGRAM(c :: cs,w))
      equation 
        c_1 = elab_class(c);
        cs_1 = elaborate(Absyn.PROGRAM(cs,w)); then c_1 :: cs_1;
  end matchcontinue;
end elaborate;

public function elab_class " relation: elab_class
 
  This functions converts an `Absyn.Class' to a `Class'.
 "
  input Absyn.Class in_class;
  output Class out_class;
algorithm 
  out_class:=
  matchcontinue (in_class)
    local
      ClassDef d_1;
      Restriction r_1;
      Absyn.Class c;
      String n,file;
      Boolean p,f,e;
      Absyn.Restriction r;
      Absyn.ClassDef d;
    case ((c = Absyn.CLASS(n,p,f,e,r,d,file)))
      equation 
        d_1 = elab_classdef(d);
        r_1 = elab_restriction(c, r); then CLASS(n,p,e,r_1,d_1);
  end matchcontinue;
end elab_class;

protected function elab_restriction
  input Absyn.Class in_class;
  input Absyn.Restriction in_restriction;
  output Restriction out_restriction;
algorithm 
  out_restriction:=
  matchcontinue (in_class,in_restriction)
    local Absyn.Class d;
    case (d,Absyn.R_FUNCTION())
      equation 
        true = contain_external_func_decl(d); then R_EXT_FUNCTION();
    case (d,Absyn.R_FUNCTION())
      equation 
        false = contain_external_func_decl(d); then R_FUNCTION();
    case (_,Absyn.R_CLASS()) then R_CLASS(); 
    case (_,Absyn.R_MODEL()) then R_MODEL(); 
    case (_,Absyn.R_RECORD()) then R_RECORD(); 
    case (_,Absyn.R_BLOCK()) then R_BLOCK(); 
    case (_,Absyn.R_CONNECTOR()) then R_CONNECTOR(); 
    case (_,Absyn.R_TYPE()) then R_TYPE(); 
    case (_,Absyn.R_PACKAGE()) then R_PACKAGE(); 
    case (_,Absyn.R_ENUMERATION()) then R_ENUMERATION(); 
    case (_,Absyn.R_PREDEFINED_INT()) then R_PREDEFINED_INT(); 
    case (_,Absyn.R_PREDEFINED_REAL()) then R_PREDEFINED_REAL(); 
    case (_,Absyn.R_PREDEFINED_STRING()) then R_PREDEFINED_STRING(); 
    case (_,Absyn.R_PREDEFINED_BOOL()) then R_PREDEFINED_BOOL(); 
    case (_,Absyn.R_PREDEFINED_ENUM()) then R_PREDEFINED_ENUM(); 
  end matchcontinue;
end elab_restriction;

protected function contain_external_func_decl
  input Absyn.Class in_class;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_class)
    local
      type Absyn_ClassPartList = list<Absyn.ClassPart>;
      type StringOption = Option<String>;
      Boolean res,b,c,d;
      String a,file;
      Absyn.Restriction e;
      Absyn_ClassPartList rest;
      StringOption cmt;
    case (Absyn.CLASS(_,_,_,_,_,Absyn.PARTS(Absyn.EXTERNAL(_,_) :: _,_),_)) then true; 
    case (Absyn.CLASS(a,b,c,d,e,Absyn.PARTS(_ :: rest,cmt),file))
      equation 
        res = contain_external_func_decl(Absyn.CLASS(a,b,c,d,e,Absyn.PARTS(rest,cmt),file)); then res;
    case (_) then false; 
  end matchcontinue;
end contain_external_func_decl;

protected function elab_classdef " relation: elab_classdef
 
  This function converts an `Absyn.ClassDef' to a `ClassDef'.  For
  the `DERIVED' case, the conversion is fairly trivial, but for the
  `PARTS' case more work is needed.  The result contains separate
  lists for elements, equations and algorithms, which are mixed in
  the input.
 
   LS: Divided the elab_classdef into separate functions for
  collecting the different parts
 "
  input Absyn.ClassDef in_classdef;
  output ClassDef out_classdef;
algorithm 
  out_classdef:=
  matchcontinue (in_classdef)
    local
      type SubscriptListOption = Option<SubscriptList>;
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      type Absyn_ClassPartList = list<Absyn.ClassPart>;
      type StringOption = Option<String>;
      type StringList = list<String>;
      type Absyn_EnumLiteralList = list<Absyn.EnumLiteral>;
      type PathList = list<Absyn.Path>;
      Mod mod;
      Absyn.Path p;
      SubscriptListOption d;
      Absyn.ElementAttributes attr;
      Absyn_ElementArgList a;
      Absyn_CommentOption cmt;
      ElementList els;
      EquationList eqs,initeqs;
      AlgorithmList als,initals;
      Absyn_ExternalDeclOption decl;
      Absyn_ClassPartList parts;
      StringList lst_1;
      Absyn_EnumLiteralList lst;
    case (Absyn.DERIVED(p,d,attr,a,cmt))
      equation 
        mod = build_mod(SOME(Absyn.CLASSMOD(a,NONE)), false) "TODO: attributes of derived classes "; then DERIVED(p,d,mod);
    case (Absyn.PARTS(parts,cmt))
      local StringOption cmt;
      equation 
        els = elab_classdef_elements(parts);
        eqs = elab_classdef_equations(parts);
        initeqs = elab_classdef_initialequations(parts);
        als = elab_classdef_algorithms(parts);
        initals = elab_classdef_initialalgorithms(parts);
        decl = elab_classdef_externaldecls(parts); then PARTS(els,eqs,initeqs,als,initals,decl);
    case (Absyn.ENUMERATION(lst,_))
      equation 
        lst_1 = elab_enumlist(lst); then ENUMERATION(lst_1);
    case (Absyn.OVERLOAD(lst,_))
      local PathList lst; then OVERLOAD(lst);
  end matchcontinue;
end elab_classdef;

protected function elab_enumlist
  input Absyn_EnumLiteralList in_absyn_enumliterallist;
  output IdentList out_identlist;
protected 
  type Absyn_EnumLiteralList = list<Absyn.EnumLiteral>;
algorithm 
  out_identlist:=
  matchcontinue (in_absyn_enumliterallist)
    local
      type StringList = list<String>;
      StringList res;
      String id;
      Absyn_EnumLiteralList rest;
    case ({}) then {}; 
    case (Absyn.ENUMLITERAL(id,_) :: rest)
      equation 
        res = elab_enumlist(rest); then id :: res;
  end matchcontinue;
end elab_enumlist;

protected function elab_classdef_elements
  input Absyn_ClassPartList in_absyn_classpartlist;
  output ElementList out_elementlist;
protected 
  type Absyn_ClassPartList = list<Absyn.ClassPart>;
algorithm 
  out_elementlist:=
  matchcontinue (in_absyn_classpartlist)
    local
      type Absyn_ElementItemList = list<Absyn.ElementItem>;
      ElementList els,es_1,els_1;
      Absyn_ElementItemList es;
      Absyn_ClassPartList rest;
    case {} then {}; 
    case (Absyn.PUBLIC(es) :: rest)
      equation 
        els = elab_classdef_elements(rest);
        es_1 = elab_eitemlist(es, false);
        els_1 = list_append(es_1, els); then els_1;
    case (Absyn.PROTECTED(es) :: rest)
      equation 
        els = elab_classdef_elements(rest);
        es_1 = elab_eitemlist(es, true);
        els_1 = list_append(es_1, els); then els_1;
    case _ :: rest " ignore all other than PUBLIC and PROTECTED, i.e. elements *"
      equation 
        els = elab_classdef_elements(rest); then els;
  end matchcontinue;
end elab_classdef_elements;

protected function elab_classdef_equations
  input Absyn_ClassPartList in_absyn_classpartlist;
  output EquationList out_equationlist;
protected 
  type Absyn_ClassPartList = list<Absyn.ClassPart>;
algorithm 
  out_equationlist:=
  matchcontinue (in_absyn_classpartlist)
    local
      type Absyn_EquationItemList = list<Absyn.EquationItem>;
      EquationList eqs,eql_1,eqs_1;
      Absyn_EquationItemList eql;
      Absyn_ClassPartList rest;
    case {} then {}; 
    case (Absyn.EQUATIONS(eql) :: rest)
      equation 
        eqs = elab_classdef_equations(rest);
        eql_1 = elab_equations(eql);
        eqs_1 = list_append(eqs, eql_1); then eqs_1;
    case _ :: rest " ignore everthing other than equations *"
      equation 
        eqs = elab_classdef_equations(rest); then eqs;
  end matchcontinue;
end elab_classdef_equations;

protected function elab_classdef_initialequations
  input Absyn_ClassPartList in_absyn_classpartlist;
  output EquationList out_equationlist;
protected 
  type Absyn_ClassPartList = list<Absyn.ClassPart>;
algorithm 
  out_equationlist:=
  matchcontinue (in_absyn_classpartlist)
    local
      type Absyn_EquationItemList = list<Absyn.EquationItem>;
      EquationList eqs,eql_1,eqs_1;
      Absyn_EquationItemList eql;
      Absyn_ClassPartList rest;
    case {} then {}; 
    case (Absyn.INITIALEQUATIONS(eql) :: rest)
      equation 
        eqs = elab_classdef_initialequations(rest);
        eql_1 = elab_equations(eql);
        eqs_1 = list_append(eqs, eql_1); then eqs_1;
    case _ :: rest " ignore everthing other than equations *"
      equation 
        eqs = elab_classdef_initialequations(rest); then eqs;
  end matchcontinue;
end elab_classdef_initialequations;

protected function elab_classdef_algorithms
  input Absyn_ClassPartList in_absyn_classpartlist;
  output AlgorithmList out_algorithmlist;
protected 
  type Absyn_ClassPartList = list<Absyn.ClassPart>;
algorithm 
  out_algorithmlist:=
  matchcontinue (in_absyn_classpartlist)
    local
      type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
      AlgorithmList als,als_1;
      Absyn_AlgorithmList al_1;
      Absyn_AlgorithmItemList al;
      Absyn_ClassPartList rest;
    case {} then {}; 
    case (Absyn.ALGORITHMS(al) :: rest)
      equation 
        als = elab_classdef_algorithms(rest);
        al_1 = elab_classdef_algorithmitems(al);
        als_1 = ALGORITHM(al_1,NONE) :: als; then als_1;
    case _ :: rest " ignore everthing other than algorithms *"
      equation 
        als = elab_classdef_algorithms(rest); then als;
  end matchcontinue;
end elab_classdef_algorithms;

protected function elab_classdef_initialalgorithms
  input Absyn_ClassPartList in_absyn_classpartlist;
  output AlgorithmList out_algorithmlist;
protected 
  type Absyn_ClassPartList = list<Absyn.ClassPart>;
algorithm 
  out_algorithmlist:=
  matchcontinue (in_absyn_classpartlist)
    local
      type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
      AlgorithmList als,als_1;
      Absyn_AlgorithmList al_1;
      Absyn_AlgorithmItemList al;
      Absyn_ClassPartList rest;
    case {} then {}; 
    case (Absyn.INITIALALGORITHMS(al) :: rest)
      equation 
        als = elab_classdef_initialalgorithms(rest);
        al_1 = elab_classdef_algorithmitems(al);
        als_1 = ALGORITHM(al_1,NONE) :: als; then als_1;
    case _ :: rest " ignore everthing other than algorithms *"
      equation 
        als = elab_classdef_initialalgorithms(rest); then als;
  end matchcontinue;
end elab_classdef_initialalgorithms;

protected function elab_classdef_algorithmitems
  input Absyn_AlgorithmItemList in_absyn_algorithmitemlist;
  output Absyn_AlgorithmList out_absyn_algorithmlist;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
algorithm 
  out_absyn_algorithmlist:=
  matchcontinue (in_absyn_algorithmitemlist)
    local
      Absyn_AlgorithmList res;
      Absyn.Algorithm alg;
      Absyn_AlgorithmItemList rest;
    case {} then {}; 
    case (Absyn.ALGORITHMITEM(alg,_) :: rest)
      equation 
        res = elab_classdef_algorithmitems(rest); then alg :: res;
    case (_ :: rest)
      equation 
        res = elab_classdef_algorithmitems(rest); then res;
  end matchcontinue;
end elab_classdef_algorithmitems;

protected function elab_classdef_externaldecls
  input Absyn_ClassPartList in_absyn_classpartlist;
  output Absyn_ExternalDeclOption out_absyn_externaldecloption;
protected 
  type Absyn_ClassPartList = list<Absyn.ClassPart>;
algorithm 
  out_absyn_externaldecloption:=
  matchcontinue (in_absyn_classpartlist)
    local
      Absyn.ExternalDecl decl;
      Absyn_ExternalDeclOption res;
      Absyn_ClassPartList rest;
    case (Absyn.EXTERNAL(decl,_) :: _) then SOME(decl); 
    case (_ :: rest)
      equation 
        res = elab_classdef_externaldecls(rest); then res;
    case ({}) then NONE; 
  end matchcontinue;
end elab_classdef_externaldecls;

protected function elab_eitemlist " relation: elab_eitemlist
 
  This function converts a list of `Absyn.ElementItem' to a list of
  `Element'.  The boolean argument flags whether the elements are
  pretected. Annotations are not elaborated, i.e. they are removed when converting to SCode.
 "
  input Absyn_ElementItemList in_absyn_elementitemlist;
  input Boolean in_boolean;
  output ElementList out_elementlist;
protected 
  type Absyn_ElementItemList = list<Absyn.ElementItem>;
algorithm 
  out_elementlist:=
  matchcontinue (in_absyn_elementitemlist,in_boolean)
    local
      ElementList l,e_1,es_1;
      Absyn_ElementItemList es;
      Boolean prot;
      Absyn.Element e;
    case ({},_) then {}; 
    case (Absyn.ANNOTATIONITEM(_) :: es,prot)
      equation 
        l = elab_eitemlist(es, prot); then l;
    case (Absyn.ELEMENTITEM(e) :: es,prot)
      equation 
        e_1 = elab_element(e, prot);
        es_1 = elab_eitemlist(es, prot);
        l = list_append(e_1, es_1); then l;
  end matchcontinue;
end elab_eitemlist;

protected function elab_element " relation: elab_element
 
  This function converts an `Absyn.Element' to a list of
  `Element's.  The original element may declare several components
  at once, and those are separated to several declarations in the
  result.
 "
  input Absyn.Element in_element;
  input Boolean in_boolean;
  output ElementList out_elementlist;
algorithm 
  out_elementlist:=
  matchcontinue (in_element,in_boolean)
    local
      ElementList es;
      Boolean f,repl,prot;
      Absyn.ElementSpec s;
      Integer line;
    case (Absyn.ELEMENT(f,repl,_,_,s,_,line,_),prot)
      equation 
        es = elab_elementspec(f, repl, prot, s); then es;
  end matchcontinue;
end elab_element;

protected function elab_elementspec " relation: elab_elementspec
 
  This function turns an `Absyn.ElementSpec' to a list of
  `Element's.  The boolean arguments say if the element is final and
  protected, respectively.
 "
  input Boolean in_boolean1;
  input Boolean in_boolean2;
  input Boolean in_boolean3;
  input Absyn.ElementSpec in_elementspec4;
  output ElementList out_elementlist;
algorithm 
  out_elementlist:=
  matchcontinue (in_boolean1,in_boolean2,in_boolean3,in_elementspec4)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      type Absyn_ModificationOption = Option<Absyn.Modification>;
      type Absyn_ComponentItemList = list<Absyn.ComponentItem>;
      ClassDef de_1;
      Restriction re_1;
      Boolean final_,repl,prot,rp,pa,fi,e,fl;
      Absyn.Class cl;
      String n,file,ns;
      Absyn.Restriction re;
      Absyn.ClassDef de;
      Mod mod;
      Absyn.Path t;
      Absyn_ElementArgList args;
      ElementList xs_1;
      Variability pa_1;
      SubscriptList tot_dim,ad,d;
      Absyn.ElementAttributes attr;
      Absyn.Direction di;
      Absyn_ModificationOption m;
      Absyn_CommentOption comment;
      Absyn_ComponentItemList xs;
      Absyn.Import imp;
    case (final_,repl,prot,Absyn.CLASSDEF(rp,(cl = Absyn.CLASS(n,pa,fi,e,re,de,file))))
      equation 
        de_1 = elab_classdef(de);
        re_1 = elab_restriction(cl, re); then {CLASSDEF(n,final_,rp,CLASS(n,pa,e,re_1,de_1),NONE)};
    case (final_,repl,prot,Absyn.EXTENDS(n,args))
      local Absyn.Path n;
      equation 
        mod = build_mod(SOME(Absyn.CLASSMOD(args,NONE)), false);
        ns = Absyn.path_string(n); then {EXTENDS(n,mod)};
    case (_,_,_,Absyn.COMPONENTS(_,_,{})) then {}; 
    case (final_,repl,prot,Absyn.COMPONENTS((attr = Absyn.ATTR(fl,pa,di,ad)),t,Absyn.COMPONENTITEM(Absyn.COMPONENT(n,d,m),comment) :: xs))
      local Absyn.Variability pa;
      equation 
        xs_1 = elab_elementspec(final_, repl, prot, Absyn.COMPONENTS(attr,t,xs));
        mod = build_mod(m, false);
        pa_1 = elab_variability(pa) "PR. This adds the arraydimension that may be specified together with
	 the type of the component.  ";
        tot_dim = list_append(d, ad); then COMPONENT(n,final_,repl,prot,ATTR(tot_dim,fl,RW(),pa_1,di),t,mod,
          NONE,comment) :: xs_1;
    case (final_,repl,prot,Absyn.IMPORT(imp,_)) then {IMPORT(imp)}; 
  end matchcontinue;
end elab_elementspec;

protected function elab_variability
  input Absyn.Variability in_variability;
  output Variability out_variability;
algorithm 
  out_variability:=
  matchcontinue (in_variability)
    case (Absyn.VAR()) then VAR(); 
    case (Absyn.DISCRETE()) then DISCRETE(); 
    case (Absyn.PARAM()) then PARAM(); 
    case (Absyn.CONST()) then CONST(); 
  end matchcontinue;
end elab_variability;

protected function elab_equations " relation: elab_equations
 
  This function transforms a list of `Absyn.Equation's to a list of
  `Equations's, by applying the `elab_equation' function to each
  equation.
 "
  input Absyn_EquationItemList in_absyn_equationitemlist;
  output EquationList out_equationlist;
protected 
  type Absyn_EquationItemList = list<Absyn.EquationItem>;
algorithm 
  out_equationlist:=
  matchcontinue (in_absyn_equationitemlist)
    local
      EEquation e_1;
      EquationList es_1;
      Absyn.Equation e;
      Absyn_EquationItemList es;
    case {} then {}; 
    case Absyn.EQUATIONITEM(e,_) :: es
      equation 
        e_1 = elab_equation(e);
        es_1 = elab_equations(es); then EQUATION(e_1,NONE) :: es_1;
    case Absyn.EQUATIONITEMANN(_) :: es
      equation 
        es_1 = elab_equations(es); then es_1;
  end matchcontinue;
end elab_equations;

protected function elab_e_equations
  input Absyn_EquationItemList in_absyn_equationitemlist;
  output EEquationList out_eequationlist;
protected 
  type Absyn_EquationItemList = list<Absyn.EquationItem>;
algorithm 
  out_eequationlist:=
  matchcontinue (in_absyn_equationitemlist)
    local
      EEquation e_1;
      EEquationList es_1;
      Absyn.Equation e;
      Absyn_EquationItemList es;
    case {} then {}; 
    case Absyn.EQUATIONITEM(e,_) :: es
      equation 
        e_1 = elab_equation(e);
        es_1 = elab_e_equations(es); then e_1 :: es_1;
    case Absyn.EQUATIONITEMANN(_) :: es
      equation 
        es_1 = elab_e_equations(es); then es_1;
  end matchcontinue;
end elab_e_equations;

public function equation_str " relation: equation_str
  author: PA
 
  Return the equation as a string.
 "
  input EEquation in_eequation;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_eequation)
    local
      type StringList = list<String>;
      String s1,s2,s3,res,id;
      StringList tb_strs,fb_strs,str_lst;
      Absyn.Exp exp,e1,e2;
      EEquationList tb,fb,eqn_lst;
      Absyn.ComponentRef cr1,cr2,cr;
    case (EQ_IF(exp,tb,fb))
      equation 
        s1 = Dump.print_exp_str(exp);
        tb_strs = Util.list_map(tb, equation_str);
        fb_strs = Util.list_map(fb, equation_str);
        s2 = Util.string_delimit_list(tb_strs, "\n");
        s3 = Util.string_delimit_list(fb_strs, "\n");
        res = Util.string_append_list({"if ",s1," then ",s2,"else ",s3,"end if;"}); then res;
    case (EQ_EQUALS(e1,e2))
      equation 
        s1 = Dump.print_exp_str(e1);
        s2 = Dump.print_exp_str(e2);
        res = Util.string_append_list({s1," = ",s2,";"}); then res;
    case (EQ_CONNECT(cr1,cr2))
      equation 
        s1 = Dump.print_component_ref_str(cr1);
        s2 = Dump.print_component_ref_str(cr2);
        res = Util.string_append_list({"connect(",s1,", ",s2,");"}); then res;
    case (EQ_FOR(id,exp,eqn_lst))
      equation 
        s1 = Dump.print_exp_str(exp);
        str_lst = Util.list_map(eqn_lst, equation_str);
        s2 = Util.string_delimit_list(str_lst, "\n");
        res = Util.string_append_list({"for ",id," in ",s1," loop\n",s2,"\nend for;"}); then res;
    case (EQ_WHEN(_,_,_)) then "EQ_WHEN(... not impl ...)"; 
    case (EQ_ASSERT(e1,e2))
      equation 
        s1 = Dump.print_exp_str(e1);
        s2 = Dump.print_exp_str(e2);
        res = Util.string_append_list({"assert(",s1,", ",s2,");"}); then res;
    case (EQ_REINIT(cr,e1))
      equation 
        s1 = Dump.print_component_ref_str(cr);
        s2 = Dump.print_exp_str(e1);
        res = Util.string_append_list({"reinit(",s1,", ",s2,");"}); then res;
  end matchcontinue;
end equation_str;

protected function elab_equation " relation: elab_equation
 
  The translation of equations are straightforward, with one
  exception.  `If' clauses are translated so that the SCode only
  contains simple `if'-`else' constructs, and no `elseif'.
 
  PR Arrays seem to keep their Absyn.rml structure. "
  input Absyn.Equation in_equation;
  output EEquation out_eequation;
algorithm 
  out_eequation:=
  matchcontinue (in_equation)
    local
      type Absyn_EquationItemList = list<Absyn.EquationItem>;
      type TupleAbsyn_ExpAbsyn_EquationItemList = tuple<Absyn.Exp,Absyn_EquationItemList>;
      type TupleAbsyn_ExpAbsyn_EquationItemListList = list<TupleAbsyn_ExpAbsyn_EquationItemList>;
      EEquationList tb_1,fb_1,eb_1,l_1;
      Absyn.Exp e,ee,econd_1,cond,econd,e1,e2;
      Absyn_EquationItemList tb,fb,ei,eb,l;
      EEquation eq;
      TupleAbsyn_ExpAbsyn_EquationItemListList eis,elsewhen_;
      TupleAbsyn_ExpEEquationListList elsewhen_1;
      Absyn.ComponentRef c1,c2,cr;
      String i;
    case Absyn.EQ_IF(e,tb,{},fb) "REMOVE - If noone complains, I will remove this 2004-11-01
 axiom	elab_equation Absyn.EQ_EXPR(e) => EQ_EXPR(e) "
      equation 
        tb_1 = elab_e_equations(tb);
        fb_1 = elab_e_equations(fb); then EQ_IF(e,tb_1,fb_1);
    case Absyn.EQ_IF(e,tb,(ee,ei) :: eis,fb)
      equation 
        eq = elab_equation(
          Absyn.EQ_IF(e,tb,{},
          {Absyn.EQUATIONITEM(Absyn.EQ_IF(ee,ei,eis,fb),NONE)})); then eq;
    case Absyn.EQ_WHEN_E(cond,tb,(econd,eb) :: elsewhen_)
      equation 
        tb_1 = elab_e_equations(tb);
        EQ_WHEN(econd_1,eb_1,elsewhen_1) = elab_equation(Absyn.EQ_WHEN_E(econd,eb,elsewhen_)); then EQ_WHEN(cond,tb_1,(econd_1,eb_1) :: elsewhen_1);
    case Absyn.EQ_WHEN_E(cond,tb,{})
      equation 
        tb_1 = elab_e_equations(tb); then EQ_WHEN(cond,tb_1,{});
    case Absyn.EQ_EQUALS(e1,e2) then EQ_EQUALS(e1,e2); 
    case Absyn.EQ_CONNECT(c1,c2) then EQ_CONNECT(c1,c2); 
    case Absyn.EQ_FOR(i,e,l)
      equation 
        l_1 = elab_e_equations(l); then EQ_FOR(i,e,l_1);
    case Absyn.EQ_NORETCALL("assert",Absyn.FUNCTIONARGS({e1,e2},{})) then EQ_ASSERT(e1,e2); 
    case Absyn.EQ_NORETCALL("reinit",Absyn.FUNCTIONARGS({Absyn.CREF(cr),e2},{})) then EQ_REINIT(cr,e2); 
  end matchcontinue;
end elab_equation;

public function build_mod " - Modification management 
   relation: build_mod
 
  Builds an `SCode.Mod' from an `Absyn.Modification'.  The boolean
  argument flags whether the modification is `final'.
 "
  input Absyn_ModificationOption in_absyn_modificationoption;
  input Boolean in_boolean;
  output Mod out_mod;
protected 
  type Absyn_ModificationOption = Option<Absyn.Modification>;
algorithm 
  out_mod:=
  matchcontinue (in_absyn_modificationoption,in_boolean)
    local
      type Absyn_ElementArgList = list<Absyn.ElementArg>;
      Absyn_ExpOption e;
      Boolean final_;
      SubModList subs;
      Absyn_ElementArgList l;
    case (NONE,_) then NOMOD(); 
    case (SOME(Absyn.CLASSMOD({},(e = SOME(_)))),final_) then MOD(final_,{},e); 
    case (SOME(Absyn.CLASSMOD(l,e)),final_)
      equation 
        subs = build_args(l); then MOD(final_,subs,e);
  end matchcontinue;
end build_mod;

public function strip_submod " relation: strip_submod
  author: PA
 
  Removes all submodifiers from the Mod.
 "
  input Mod in_mod;
  output Mod out_mod;
algorithm 
  out_mod:=
  matchcontinue (in_mod)
    local
      Boolean f;
      SubModList subs;
      Absyn_ExpOption e;
      Mod m;
    case (MOD(f,subs,e)) then MOD(f,{},e); 
    case (m) then m; 
  end matchcontinue;
end strip_submod;

protected function build_args " LS: Adding elaborate for the elementspec in the redeclaration
 "
  input Absyn_ElementArgList in_absyn_elementarglist;
  output SubModList out_submodlist;
protected 
  type Absyn_ElementArgList = list<Absyn.ElementArg>;
algorithm 
  out_submodlist:=
  matchcontinue (in_absyn_elementarglist)
    local
      type Absyn_ModificationOption = Option<Absyn.Modification>;
      type StringOption = Option<String>;
      type Absyn_ElementSpecOption = Option<Absyn.ElementSpec>;
      SubModList subs;
      Mod mod_1;
      SubMod sub;
      Boolean final_;
      Absyn.Each each_;
      Absyn.ComponentRef cref;
      Absyn_ModificationOption mod;
      StringOption cmt;
      Absyn_ElementArgList xs;
      String n;
      ElementList elist;
      Absyn.ElementSpec spec;
      Absyn_ElementSpecOption constropt;
    case {} then {}; 
    case (Absyn.MODIFICATION(final_,each_,cref,mod,cmt) :: xs)
      equation 
        subs = build_args(xs);
        mod_1 = build_mod(mod, final_);
        sub = build_sub(cref, mod_1); then sub :: subs;
    case (Absyn.REDECLARATION(final_,each_,spec,constropt) :: xs)
      equation 
        subs = build_args(xs);
        n = Absyn.element_spec_name(spec);
        elist = elab_elementspec(final_, false, false, spec) " LS:: don't know what to use for \"protected\", so using false *  LS:: don't know what to use for \"replaceable\", so using false *"; then NAMEMOD(n,REDECL(final_,elist)) :: subs;
  end matchcontinue;
end build_args;

protected function build_sub " relation: build_sub
 
  This function converts a `ComponentRef' into a number of nested
  `SUBMOD's.
 "
  input Absyn.ComponentRef in_componentref;
  input Mod in_mod;
  output SubMod out_submod;
algorithm 
  out_submod:=
  matchcontinue (in_componentref,in_mod)
    local
      String str,i;
      Absyn.ComponentRef c,path;
      Mod mod_1,mod;
      SubscriptList ss;
      SubMod sub;
    case ((c = Absyn.CREF_IDENT(_,_ :: _)),MOD(_,_ :: _,_)) "First some rules to prevent bad modifications "
      equation 
        Print.print_error_buf("# Illegal modification of ") "First some rules to prevent bad modifications ";
        str = Dump.print_component_ref_str(c);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
    case ((c = Absyn.CREF_QUAL(_,_ :: _,_)),MOD(_,_ :: _,_))
      equation 
        Print.print_error_buf("# Illegal modification of ");
        str = Dump.print_component_ref_str(c);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
    case (Absyn.CREF_IDENT(i,ss),mod) "Then the normal rules "
      equation 
        mod_1 = build_sub_sub(ss, mod); then NAMEMOD(i,mod_1);
    case (Absyn.CREF_QUAL(i,ss,path),mod)
      equation 
        sub = build_sub(path, mod);
        mod = MOD(false,{sub},NONE);
        mod_1 = build_sub_sub(ss, mod); then NAMEMOD(i,mod_1);
  end matchcontinue;
end build_sub;

protected function build_sub_sub " relation: build_sub_sub
 
  This function is used to handle the case when a array component is
  indexed in the modification, so that only one or a limitied number
  of array elements should be modified.
 "
  input SubscriptList in_subscriptlist;
  input Mod in_mod;
  output Mod out_mod;
algorithm 
  out_mod:=
  matchcontinue (in_subscriptlist,in_mod)
    local
      Mod m;
      SubscriptList l;
    case ({},m) then m; 
    case (l,m) then MOD(false,{IDXMOD(l,m)},NONE); 
  end matchcontinue;
end build_sub_sub;

public function get_element_named
  input Ident in_ident;
  input Class in_class;
  output Element out_element;
algorithm 
  out_element:=
  matchcontinue (in_ident,in_class)
    local
      Element elt;
      String id;
      ElementList elts;
    case (id,CLASS(_,_,_,_,PARTS(elts,_,_,_,_,_)))
      equation 
        elt = get_element_named_from_elts(id, elts); then elt;
  end matchcontinue;
end get_element_named;

protected function get_element_named_from_elts
  input Ident in_ident;
  input ElementList in_elementlist;
  output Element out_element;
algorithm 
  out_element:=
  matchcontinue (in_ident,in_elementlist)
    local
      Element elt,comp,cdef;
      String id2,id1;
      ElementList xs;
    case (id2,COMPONENT(id1,_,_,_,_,_,_,_,_) :: xs)
      equation 
        failure(equality(id1 = id2));
        elt = get_element_named_from_elts(id2, xs); then elt;
    case (id2,CLASSDEF(id1,_,_,_,_) :: xs)
      equation 
        failure(equality(id1 = id2));
        elt = get_element_named_from_elts(id2, xs); then elt;
    case (id2,EXTENDS(_,_) :: xs)
      equation 
        elt = get_element_named_from_elts(id2, xs); then elt;
    case (id2,(comp = COMPONENT(id1,_,_,_,_,_,_,_,_)) :: _)
      equation 
        equality(id1 = id2); then comp;
    case (id2,(cdef = CLASSDEF(id1,_,_,_,_)) :: _)
      equation 
        equality(id1 = id2); then cdef;
  end matchcontinue;
end get_element_named_from_elts;

public function print_mod " relation: print_mod
 
  This function prints a modification.  The code is excluded from
  the report for brevity.
 
  !ignorecode"
  input Mod m;
protected 
  String s;
algorithm 
  s := print_mod_str(m);
  Print.print_buf(s);
end print_mod;

public function print_mod_str
  input Mod in_mod;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_mod)
    local
      type StringList = list<String>;
      String final_str,str,res,subs_str,ass_str;
      StringList strs;
      Boolean b,final_;
      ElementList elist;
      SubModList subs;
      Absyn_ExpOption ass;
    case (NOMOD()) then "NOMOD"; 
    case REDECL(b,elist)
      equation 
        Print.print_buf("redeclare(");
        final_str = Util.if_(b, "final", "");
        strs = Util.list_map(elist, print_element_str);
        str = Util.string_delimit_list(strs, ",");
        res = Util.string_append_list({"redeclare(",final_str,str,")"}); then res;
    case MOD(final_,subs,ass)
      equation 
        final_str = Util.if_(final_, "final", "");
        subs_str = print_subs1_str(subs);
        ass_str = print_eqmod_str(ass);
        res = Util.string_append_list({final_str,subs_str,ass_str}); then res;
    case _
      equation 
        Print.print_buf("#-- print_mod_str failed\n"); then fail();
  end matchcontinue;
end print_mod_str;

public function restr_string ""
  input Restriction in_restriction;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_restriction)
    case R_CLASS() then "CLASS"; 
    case R_MODEL() then "MODEL"; 
    case R_RECORD() then "RECORD"; 
    case R_BLOCK() then "BLOCK"; 
    case R_CONNECTOR() then "CONNECTOR"; 
    case R_TYPE() then "TYPE"; 
    case R_PACKAGE() then "PACKAGE"; 
    case R_FUNCTION() then "FUNCTION"; 
    case R_EXT_FUNCTION() then "EXTFUNCTION"; 
    case R_PREDEFINED_INT() then "PREDEFINED_INT"; 
    case R_PREDEFINED_REAL() then "PREDEFINED_REAL"; 
    case R_PREDEFINED_STRING() then "PREDEFINED_STRING"; 
    case R_PREDEFINED_BOOL() then "PREDEFINED_BOOL"; 
  end matchcontinue;
end restr_string;

public function print_restr
  input Restriction restr;
protected 
  String str;
algorithm 
  str := restr_string(restr);
  Print.print_buf(str);
end print_restr;

protected function print_final
  input Boolean in_boolean;
algorithm 
  _:=
  matchcontinue (in_boolean)
    case false then (); 
    case true
      equation 
        Print.print_buf(" final "); then ();
  end matchcontinue;
end print_final;

protected function print_subs_str
  input SubModList in_submodlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_submodlist)
    local
      String s,res,n,mod_str,str,sub_str;
      Mod mod;
      SubModList subs;
      SubscriptList ss;
    case {} then ""; 
    case {NAMEMOD(n,mod)}
      equation 
        Print.print_buf(n);
        s = print_mod_str(mod);
        res = string_append(n, s); then res;
    case NAMEMOD(n,mod) :: subs
      equation 
        mod_str = print_mod_str(mod);
        str = print_subs_str(subs);
        res = Util.string_append_list({"n",mod_str,", ",str}); then res;
    case {IDXMOD(ss,mod)}
      equation 
        str = Dump.print_subscripts_str(ss);
        mod_str = print_mod_str(mod);
        res = string_append(str, mod_str); then res;
    case IDXMOD(ss,mod) :: subs
      equation 
        str = Dump.print_subscripts_str(ss);
        mod_str = print_mod_str(mod);
        sub_str = print_subs_str(subs);
        res = Util.string_append_list({str,mod_str,", ",sub_str}); then res;
  end matchcontinue;
end print_subs_str;

protected function print_subs1_str
  input SubModList in_submodlist;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_submodlist)
    local
      String s,res;
      SubModList l;
    case {} then ""; 
    case l
      equation 
        Print.print_buf("(");
        s = print_subs_str(l);
        Print.print_buf(")");
        res = Util.string_append_list({"(",s,")"}); then res;
  end matchcontinue;
end print_subs1_str;

protected function print_eqmod
  input Absyn_ExpOption m;
protected 
  String s;
algorithm 
  s := print_eqmod_str(m);
  Print.print_buf(s);
end print_eqmod;

protected function print_eqmod_str
  input Absyn_ExpOption in_absyn_expoption;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_absyn_expoption)
    local
      String str,res;
      Absyn.Exp e;
    case NONE then ""; 
    case SOME(e)
      equation 
        str = Dump.print_exp_str(e);
        res = string_append(" = ", str); then res;
  end matchcontinue;
end print_eqmod_str;

public function print_element_list "!includecode"
  input ElementList in_elementlist;
algorithm 
  _:=
  matchcontinue (in_elementlist)
    local
      Element x;
      ElementList xs;
    case ({}) then (); 
    case (x :: xs)
      equation 
        print_element(x);
        print_element_list(xs); then ();
  end matchcontinue;
end print_element_list;

public function print_element
  input Element elt;
protected 
  String str;
algorithm 
  str := print_element_str(elt);
  Print.print_buf(str);
end print_element;

public function print_element_str
  input Element in_element;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_element)
    local
      String str,res,n,mod_str,s,vs;
      Absyn.Path path,typath;
      Mod mod;
      Boolean final_,repl,prot;
      Class cl;
      Variability var;
      Absyn_CommentOption comment;
      Attributes attr;
    case EXTENDS(path,mod)
      equation 
        str = Absyn.path_string(path);
        res = Util.string_append_list({"EXTENDS(",str,")"}); then res;
    case CLASSDEF(n,final_,repl,cl,SOME(path))
      equation 
        str = Absyn.path_string(path);
        res = Util.string_append_list({"CLASSDEF(",n,", from basclass: ",str,")"}); then res;
    case COMPONENT(n,final_,repl,prot,ATTR(_,_,_,var,_),typath,mod,SOME(path),comment)
      equation 
        mod_str = print_mod_str(mod);
        s = Absyn.path_string(typath);
        vs = variability_string(var);
        str = Absyn.path_string(path);
        res = Util.string_append_list(
          {"COMPONENT(",n,mod_str,s,"var :",vs,", from baseclass: ",
          str,")"}); then res;
    case CLASSDEF(n,final_,repl,cl,NONE)
      equation 
        res = Util.string_append_list({"CLASSDEF(",n,")"}); then res;
    case COMPONENT(n,final_,repl,prot,attr,typath,mod,NONE,comment)
      equation 
        mod_str = print_mod_str(mod);
        s = Absyn.path_string(typath);
        res = Util.string_append_list({"COMPONENT(",n," mod:",mod_str,", tp: ",s,")"}); then res;
    case (IMPORT(_)) then "IMPORT(_)"; 
  end matchcontinue;
end print_element_str;

public function variability_string
  input Variability in_variability;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_variability)
    case (VAR()) then "VAR"; 
    case (DISCRETE()) then "DISCRETE"; 
    case (PARAM()) then "PARAM"; 
    case (STRUCTPARAM()) then "STRUCTPARAM"; 
    case (CONST()) then "CONST"; 
  end matchcontinue;
end variability_string;

public function is_parameter_or_const
  input Variability in_variability;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_variability)
    local Variability STRUCPARAM;
    case (VAR()) then false; 
    case (DISCRETE()) then false; 
    case (PARAM()) then true; 
    case (STRUCPARAM) then true; 
    case (CONST()) then true; 
  end matchcontinue;
end is_parameter_or_const;

public function count_parts
  input Class in_class;
  output Integer out_integer;
algorithm 
  out_integer:=
  matchcontinue (in_class)
    local
      Integer res;
      ElementList elts;
    case CLASS(_,_,_,_,PARTS(elts,_,_,_,_,_))
      equation 
        res = list_length(elts); then res;
    case _ then 0; 
  end matchcontinue;
end count_parts;

public function component_names
  input Class in_class;
  output StringList out_stringlist;
protected 
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_class)
    local
      StringList res;
      ElementList elts;
    case (CLASS(_,_,_,_,PARTS(elts,_,_,_,_,_)))
      equation 
        res = component_names_from_elts(elts); then res;
    case (_) then {}; 
  end matchcontinue;
end component_names;

protected function component_names_from_elts
  input ElementList in_elementlist;
  output StringList out_stringlist;
protected 
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_elementlist)
    local
      StringList res;
      String id;
      ElementList rest;
    case ({}) then {}; 
    case (COMPONENT(id,_,_,_,_,_,_,_,_) :: rest)
      equation 
        res = component_names_from_elts(rest); then id :: res;
  end matchcontinue;
end component_names_from_elts;

public function is_function
  input Class in_class;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_class)
    local
      String n;
      ClassDef def;
    case CLASS(n,_,_,R_FUNCTION(),def) then true; 
    case CLASS(n,_,_,R_EXT_FUNCTION(),def) then true; 
    case _ then false; 
  end matchcontinue;
end is_function;

public function class_name
  input Class in_class;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_class)
    local String n;
    case CLASS(n,_,_,_,_) then n; 
    case _ then "Not a class"; 
  end matchcontinue;
end class_name;
end SCode;

