package SimCodegen "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 SimCodegen.rml
  module:      SimCodegen
  description: Generate Simulation code for connecting to solver.
  This can be done in two different ways.
  1. Generation of simulation code on residual form. This will generate 
  code on the form g(\dot{x},x,y,t)=0.
  2. Generation of simulation code on solved form. This will generate
  code on ode form. \dot{x} = f(x,y,t). This means that \"function\" f will
  contain code for solving states from equations, some of them might be 
  system of equations, linear or non-linear.
 
  Outputs: the simulation code in C/C++ to a given filename.
  Input: DAELow
  Uses: DAELow, Absyn, Exp
 
 
  RCS: $Id: SimCodegen.rml 1831 2005-06-23 13:37:38Z petar $
 
 "

public import OpenModelica.Compiler.DAE;

public import OpenModelica.Compiler.DAELow;

public import OpenModelica.Compiler.Absyn;

public import OpenModelica.Compiler.Exp;

protected constant String TAB="    ";

protected import OpenModelica.Compiler.Util "TAB is four whitespaces ";

protected import OpenModelica.Compiler.RTOpts;

protected import OpenModelica.Compiler.Debug;

protected import OpenModelica.Compiler.System;

protected import OpenModelica.Compiler.Values;

protected import OpenModelica.Compiler.Codegen;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.ModUtil;

protected import OpenModelica.Compiler.VarTransform;

public function generate_makefile " relation generate_makefile
  This function generates a makefile for the simulation code.
  It uses OPENMODELICAHOME as a reference to find c_runtime with libraries.
 "
  input String filename;
  input Absyn.Path class_;
protected 
  String cname,cpp_file,str;
algorithm 
  cname := Absyn.path_string(class_);
  cpp_file := Util.string_append_list({cname,".cpp"});
  str := Util.string_append_list(
          {"#Makefile generated by OpenModelica\n\n","CXX=g++\n",
          cname,": ",cpp_file,"\n","\t $(CXX) -o ",cname," ",cpp_file,
          " -L$(OPENMODELICAHOME)/c_runtime/"," -I$(OPENMODELICAHOME)/c_runtime/ "," -lsim -lg2c -lc_runtime\n"});
  System.write_file(filename, str);
end generate_makefile;

public function generate_simulation_code "ass1 
  ass2 
  blocks
  classname "
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input DAELow.IncidenceMatrix in_incidencematrix5;
  input DAELow.IncidenceMatrixT in_incidencematrixt6;
  input IntegerListList in_integerlistlist7;
  input Absyn.Path in_path8;
  input String in_string9;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
algorithm 
  _:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_incidencematrix5,in_incidencematrixt6,in_integerlistlist7,in_path8,in_string9)
    local
      type IntegerListArray = IntegerList[:];
      String cname,out_str,in_str,cglobal,coutput,cstate,c_ode,s_code,cwhen,funcfilename,res,filename;
      IntegerListList blt_states,blt_no_states,comps;
      Integer n_o,n_i;
      DAE.DAElist dae;
      DAELow.DAELow dlow;
      IntegerVector ass1,ass2;
      IntegerListArray m,mt;
      Absyn.Path class_;
    case (dae,dlow,ass1,ass2,m,mt,comps,class_,filename) "ass1  ass2  blocks classname "
      equation 
        cname = Absyn.path_string(class_);
        (blt_states,blt_no_states) = DAELow.generate_state_partition(comps, dlow, ass1, ass2, m, mt);
        (out_str,n_o) = generate_output_function_code(dlow);
        (in_str,n_i) = generate_input_function_code(dlow);
        cglobal = generate_global_data(class_, dlow, n_o, n_i);
        coutput = generate_compute_output(cname, dae, dlow, ass1, ass2, blt_no_states);
        cstate = generate_compute_residual_state(cname, dae, dlow, ass1, ass2, blt_states);
        c_ode = generate_ode_code(dlow, blt_states, ass1, ass2, m, mt, class_);
        s_code = generate_start_attr_code(dlow);
        cwhen = generate_when_clauses(cname, dae, dlow, ass1, ass2, comps);
        funcfilename = string_append(cname, "_functions.cpp");
        res = Util.string_append_list(
          {"//Simulation code for ",cname,
          "\n//Generated by OpenModelica.\n","\n#include \"modelica.h\"\n","\n#include \"",funcfilename,"\"\n",
          cglobal,coutput,in_str,out_str,cstate,cwhen,c_ode,s_code});
        System.write_file(filename, res); then ();
    case (_,_,_,_,_,_,_,_,_)
      equation 
        print("-generate_simulation_code failed\n"); then fail();
  end matchcontinue;
end generate_simulation_code;

protected function generate_global_data " relation: generate_global_data
  This function generates the C-code for the global data: arrays for states, derivatives
  and algebraic variables, etc.
  arg1
  arg2
  arg3 an int which shows the number of output variables on top level
  arg4 an int which shows the number of input variables on top level
 "
  input Absyn.Path class_;
  input DAELow.DAELow dlow;
  input Integer no;
  input Integer ni;
  output String str;
protected 
  Integer nx,ny,np;
  String class_str,nx_str,ny_str,np_str,no_str,ni_str,varnames,paramnames;
algorithm 
  (nx,ny,np) := DAELow.calculate_sizes(dlow);
  class_str := Absyn.path_string(class_);
  nx_str := int_string(nx);
  ny_str := int_string(ny);
  np_str := int_string(np);
  no_str := int_string(no);
  ni_str := int_string(ni);
  varnames := generate_var_names(dlow, nx, ny);
  paramnames := generate_parameter_names(dlow, np);
  str := Util.string_append_list(
          {"\n","#define NX ",nx_str,"\n","#define NY ",ny_str,"\n",
          "#define NP ",np_str,"\n","#define NO ",no_str,
          " // number of outputvar on topmodel\n","#define NI ",ni_str," // number of inputvar on topmodel\n",
          "#define MAXORD 5\n","#define time (*t)\n","\n",
          "#if NX > 0 /* some c-compilers does not like a static array of |a|==0 */\n","double x[NX];\n","#else\n","double *x;\n","#endif\n","#if NX > 0\n",
          "double xd[NX];\n","#else\n","double *xd;\n","#endif\n","#if NX > 0\n",
          "double dummy_delta[NX];\n","#else\n","double *dummy_delta;\n","#endif\n","#if NY > 0\n",
          "double y[NY];\n","#else\n","double *y;\n","#endif\n","#if NP > 0\n","double p[NP];\n",
          "#else\n","double *p;\n","#endif\n","#if NO > 0\n","double out_y[NO];\n",
          "#else\n","double *out_y;\n","#endif\n","#if NI > 0\n","double in_y[NI];\n",
          "#else\n","double *in_y;\n","#endif\n",
          "double rwork[50+(MAXORD+4)*NX+NX*NX];\n","long iwork[20+NX];\n","long liw = 20+NX;\n",
          "long lrw = 50+(MAXORD+4)*NX+NX*NX;\n","long nx = NX;\n","long ny = NY;\n","long np = NP;\n",
          "long no = NO;\n","long ni = NI;\n","char *model_name=\"",class_str,"\";\n",varnames,
          "\n",paramnames,"\n"});
end generate_global_data;

protected function generate_var_names " relation: generate_var_names
  Generates an array of the original variable names.
 "
  input DAELow.DAELow in_daelow1;
  input Integer in_integer2;
  input Integer in_integer3;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_daelow1,in_integer2,in_integer3)
    local
      type DAELow_VarList = list<DAELow.Var>;
      type StringList = list<String>;
      DAELow_VarList var_lst,knvar_lst;
      StringList state_lst,der_lst,algvar_lst,inputvar_lst,varnames_lst;
      String str,tot_elts_str,res;
      Integer nxtwo,tot_elts,nx,ny;
      DAELow.Variables vars,knvars;
    case (DAELow.DAELOW(vars,knvars,_,_,_,_,_),nx,ny)
      equation 
        var_lst = DAELow.var_list(vars);
        knvar_lst = DAELow.var_list(knvars);
        state_lst = generate_var_names_states(var_lst);
        der_lst = generate_var_names_derivatives(var_lst);
        algvar_lst = generate_var_names_algvars(var_lst);
        inputvar_lst = generate_var_names_inputvars(knvar_lst);
        varnames_lst = Util.list_flatten({state_lst,der_lst,algvar_lst,inputvar_lst});
        str = Util.string_delimit_list(varnames_lst, ", ");
        nxtwo = nx + nx;
        tot_elts = ny + nxtwo;
        tot_elts_str = int_string(tot_elts);
        res = Util.string_append_list({"char* varnames[",tot_elts_str,"] = {",str,"};\n"}); then res;
  end matchcontinue;
end generate_var_names;

protected function generate_parameter_names "generate_parameter_names
  this function generates a char* parameternames{}=... that contain
  the name of the parameters. 
 "
  input DAELow.DAELow in_daelow;
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_daelow,in_integer)
    local
      type DAELow_VarList = list<DAELow.Var>;
      type StringList = list<String>;
      DAELow.Variables vars,knvars;
      Integer np;
      DAELow_VarList knvar_lst;
      StringList param_str_lst;
      String str,np_str,res;
    case (DAELow.DAELOW(vars,knvars,_,_,_,_,_),np)
      equation 
        {} = DAELow.var_list(knvars) "special case if no variables "; then "";
    case (DAELow.DAELOW(vars,knvars,_,_,_,_,_),np)
      equation 
        knvar_lst = DAELow.var_list(knvars);
        param_str_lst = generate_parameter_names2(knvar_lst);
        str = Util.string_delimit_list(param_str_lst, ", ");
        np_str = int_string(np);
        res = Util.string_append_list({"char* parameternames[",np_str,"] = {",str,"};\n"}); then res;
  end matchcontinue;
end generate_parameter_names;

protected function generate_parameter_names2 " relation: generate_parameter_names2
  Helper function to generate_parameter_names
 "
  input DAELow_VarList in_daelow_varlist;
  output StringList out_stringlist;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelow_varlist)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      StringList res;
      String origname_1,origname;
      Exp.ComponentRef cr;
      Exp_ExpOption value;
      Integer indx;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList vs;
    case ({}) then {}; 
    case (DAELow.VAR(cr,DAELow.PARAM(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_parameter_names2(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (_ :: vs)
      equation 
        res = generate_parameter_names2(vs); then res;
  end matchcontinue;
end generate_parameter_names2;

protected function generate_var_names_states " relation: generate_var_names_states
  Helper function to generate_var_names
 "
  input DAELow_VarList in_daelow_varlist;
  output StringList out_stringlist;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelow_varlist)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      StringList res;
      String origname_1,origname;
      Exp.ComponentRef cr;
      Exp_ExpOption value;
      Integer indx;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList vs;
    case ({}) then {}; 
    case (DAELow.VAR(cr,DAELow.STATE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_states(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (_ :: vs)
      equation 
        res = generate_var_names_states(vs); then res;
  end matchcontinue;
end generate_var_names_states;

protected function generate_var_names_derivatives " relation: generate_var_names_derivatives
  Helper function to generate_var_names
 "
  input DAELow_VarList in_daelow_varlist;
  output StringList out_stringlist;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelow_varlist)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      StringList res;
      String origname_1,origname;
      Exp.ComponentRef cr;
      Exp_ExpOption value;
      Integer indx;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList vs;
    case ({}) then {}; 
    case (DAELow.VAR(cr,DAELow.STATE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_derivatives(vs);
        origname_1 = Util.string_append_list({"\"der(",origname,")\""}); then origname_1 :: res;
    case (_ :: vs)
      equation 
        res = generate_var_names_derivatives(vs); then res;
  end matchcontinue;
end generate_var_names_derivatives;

protected function generate_var_names_algvars " relation: generate_var_names_algvars
  Helper function to generate_var_names
 "
  input DAELow_VarList in_daelow_varlist;
  output StringList out_stringlist;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelow_varlist)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      StringList res;
      String origname_1,origname;
      Exp.ComponentRef cr;
      Exp_ExpOption value;
      Integer indx;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList vs;
    case ({}) then {}; 
    case (DAELow.VAR(cr,DAELow.VARIABLE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_algvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (DAELow.VAR(cr,DAELow.DISCRETE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_algvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (DAELow.VAR(cr,DAELow.DUMMY_DER(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_algvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (DAELow.VAR(cr,DAELow.DUMMY_STATE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_algvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (DAELow.VAR(cr,DAELow.DISCRETE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_algvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (_ :: vs)
      equation 
        res = generate_var_names_algvars(vs); then res;
  end matchcontinue;
end generate_var_names_algvars;

protected function generate_input_function_code "
   this generates the input_function for all the variables
   that are INPUT and on top model
  
"
  input DAELow.DAELow in_daelow;
  output String out_string;
  output Integer out_integer;
algorithm 
  (out_string,out_integer):=
  matchcontinue (in_daelow)
    local
      type DAELow_VarList = list<DAELow.Var>;
      type StringList = list<String>;
      type Algorithm_AlgorithmArray = Algorithm.Algorithm[:];
      DAELow_VarList knvars_lst;
      StringList res1;
      Integer lst_lenght;
      String res1_1,res;
      DAELow.Variables vars,knvars;
      DAELow.EquationArray eqns,se,ie;
      Algorithm_AlgorithmArray al;
      DAELow.EventInfo ev;
    case (DAELow.DAELOW(vars,knvars,eqns,se,ie,al,ev))
      equation 
        knvars_lst = DAELow.var_list(knvars);
        res1 = generate_input_function_code2(knvars_lst, 0);
        lst_lenght = list_length(res1);
        res1_1 = Util.string_delimit_list_no_empty(res1, "\n  ");
        res = Util.string_append_list(
          {
          "\n/*\n*/\nvoid input_function(double *in_x,double *x, double *xd, double *y, double *p,\n","                   int nx, int ny, int np)\n","{\n  ",res1_1,"\n}\n"}); then (res,lst_lenght);
  end matchcontinue;
end generate_input_function_code;

protected function generate_input_function_code2
  input DAELow_VarList in_daelow_varlist;
  input Integer in_integer;
  output StringList out_stringlist;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelow_varlist,in_integer)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type Values_ValueOption = Option<Values.Value>;
      type Exp_SubscriptList = list<Exp.Subscript>;
      type Absyn_PathList = list<Absyn.Path>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Integer int,i_1,index,i;
      String i_str,cr_str,assign_str,name;
      StringList res;
      DAELow.Var var;
      Exp.ComponentRef cr;
      DAE.VarDirection dir;
      DAE.Type tp;
      Exp_ExpOption exp,st;
      Values_ValueOption v;
      Exp_SubscriptList dim;
      Absyn_PathList classes;
      DAE_VariableAttributesOption attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList rest;
    case ({},int) then {}; 
    case ((var = DAELow.VAR(cr,_,dir,tp,exp,v,dim,st,index,name,classes,attr,comment,flow_)) :: rest,i)
      equation 
        true = DAELow.is_var_on_top_level_and_input(var);
        i_str = int_string(i);
        i_1 = i + 1;
        cr_str = Exp.print_component_ref_str(cr);
        assign_str = Util.string_append_list({cr_str,"= in_x[",i_str,"];"});
        res = generate_input_function_code2(rest, i_1); then assign_str :: res;
    case (var :: rest,index)
      equation 
        res = generate_input_function_code2(rest, index); then res;
  end matchcontinue;
end generate_input_function_code2;

protected function generate_output_function_code "
   this generates the output_function for all the variables
   that are OUTPUT and on top model
  
"
  input DAELow.DAELow in_daelow;
  output String out_string;
  output Integer out_integer;
algorithm 
  (out_string,out_integer):=
  matchcontinue (in_daelow)
    local
      type DAELow_VarList = list<DAELow.Var>;
      type StringList = list<String>;
      type Algorithm_AlgorithmArray = Algorithm.Algorithm[:];
      DAELow_VarList knvars_lst,vars_lst,vars_lst_1;
      StringList res1;
      Integer lst_lenght;
      String res1_1,res;
      DAELow.Variables vars,knvars;
      DAELow.EquationArray eqns,se,ie;
      Algorithm_AlgorithmArray al;
      DAELow.EventInfo ev;
    case (DAELow.DAELOW(vars,knvars,eqns,se,ie,al,ev))
      equation 
        knvars_lst = DAELow.var_list(knvars);
        vars_lst = DAELow.var_list(vars);
        vars_lst_1 = list_append(knvars_lst, vars_lst);
        res1 = generate_output_function_code2(vars_lst_1, 0);
        lst_lenght = list_length(res1);
        res1_1 = Util.string_delimit_list_no_empty(res1, "\n  ");
        res = Util.string_append_list(
          {
          "\n/*\n*/\nvoid output_function(double *out_x,double *x, double *xd, double *y, double *p,\n","                   int nx, int ny, int np)\n","{\n  ",res1_1,"\n}\n"}); then (res,lst_lenght);
  end matchcontinue;
end generate_output_function_code;

protected function generate_output_function_code2
  input DAELow_VarList in_daelow_varlist;
  input Integer in_integer;
  output StringList out_stringlist;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelow_varlist,in_integer)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type Values_ValueOption = Option<Values.Value>;
      type Exp_SubscriptList = list<Exp.Subscript>;
      type Absyn_PathList = list<Absyn.Path>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Integer int,i_1,index,i;
      String i_str,cr_str,assign_str,name;
      StringList res;
      DAELow.Var var;
      Exp.ComponentRef cr;
      DAE.VarDirection dir;
      DAE.Type tp;
      Exp_ExpOption exp,st;
      Values_ValueOption v;
      Exp_SubscriptList dim;
      Absyn_PathList classes;
      DAE_VariableAttributesOption attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList rest;
    case ({},int) then {}; 
    case ((var = DAELow.VAR(cr,_,dir,tp,exp,v,dim,st,index,name,classes,attr,comment,flow_)) :: rest,i)
      equation 
        true = DAELow.is_var_on_top_level_and_output(var);
        i_str = int_string(i);
        i_1 = i + 1;
        cr_str = Exp.print_component_ref_str(cr);
        assign_str = Util.string_append_list({"out_y[",i_str,"] =",cr_str,";"});
        res = generate_output_function_code2(rest, i_1); then assign_str :: res;
    case (var :: rest,index)
      equation 
        res = generate_output_function_code2(rest, index); then res;
  end matchcontinue;
end generate_output_function_code2;

protected function generate_var_names_inputvars " relation: generate_var_names_inputvars
  Helper function to generate_var_names
 "
  input DAELow_VarList in_daelow_varlist;
  output StringList out_stringlist;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelow_varlist)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      StringList res;
      String origname_1,origname;
      Exp.ComponentRef cr;
      Exp_ExpOption value;
      Integer indx;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList vs;
    case ({}) then {}; 
    case (DAELow.VAR(cr,DAELow.VARIABLE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_inputvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (DAELow.VAR(cr,DAELow.DISCRETE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_inputvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (DAELow.VAR(cr,DAELow.DUMMY_DER(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_inputvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (DAELow.VAR(cr,DAELow.DUMMY_STATE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_inputvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (DAELow.VAR(cr,DAELow.DISCRETE(),_,_,_,_,_,value,indx,origname,_,dae_var_attr,comment,flow_) :: vs)
      equation 
        res = generate_var_names_inputvars(vs);
        origname_1 = Util.string_append_list({"\"",origname,"\""}); then origname_1 :: res;
    case (_ :: vs)
      equation 
        res = generate_var_names_inputvars(vs); then res;
  end matchcontinue;
end generate_var_names_inputvars;

protected function generate_start_attr_code "generate_start_attr_code
  Because of that some of the veriables are of the form: Real x(start=p*4) 
  this function generatese a start_function in c to calculate the start-function
  of the variables
 "
  input DAELow.DAELow in_daelow;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_daelow)
    local
      type DAELow_VarList = list<DAELow.Var>;
      type StringList = list<String>;
      type Algorithm_AlgorithmArray = Algorithm.Algorithm[:];
      DAELow_VarList vars_lst;
      StringList res1;
      String res1_1,res;
      DAELow.Variables vars,knvars;
      DAELow.EquationArray eqns,se,ie;
      Algorithm_AlgorithmArray al;
      DAELow.EventInfo ev;
    case (DAELow.DAELOW(vars,knvars,eqns,se,ie,al,ev))
      equation 
        vars_lst = DAELow.var_list(vars);
        res1 = Util.list_map(vars_lst, generate_start_attr_code2);
        res1_1 = Util.string_delimit_list_no_empty(res1, ";\n  ");
        res = Util.string_append_list(
          {
          "\n/*\n*/\nvoid start_function(double *x, double *xd, double *y, double *p,\n","                   int nx, int ny, int np)\n","{\n  ",res1_1,
          ";\n}\n"}); then res;
    case (_) then ""; 
  end matchcontinue;
end generate_start_attr_code;

protected function generate_start_attr_code2
  input DAELow.Var in_var;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_var)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type Values_ValueOption = Option<Values.Value>;
      type Exp_SubscriptList = list<Exp.Subscript>;
      type Absyn_PathList = list<Absyn.Path>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      String cr_str,st_str,res,name;
      Exp.ComponentRef cr;
      DAE.VarDirection dir;
      DAE.Type tp;
      Exp_ExpOption exp;
      Values_ValueOption v;
      Exp_SubscriptList dim;
      Exp.Exp st;
      Integer index;
      Absyn_PathList classes;
      DAE_VariableAttributesOption attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
    case (DAELow.VAR(cr,_,dir,tp,exp,v,dim,SOME(st),index,name,classes,attr,comment,flow_))
      equation 
        cr_str = Exp.print_component_ref_str(cr);
        st_str = print_exp_cpp_str(st);
        res = Util.string_append_list({cr_str,"=",st_str}); then res;
    case (_) then ""; 
  end matchcontinue;
end generate_start_attr_code2;

protected function generate_ode_code "components
  ass1 
  ass2 "
  input DAELow.DAELow dlow;
  input IntegerListList comps;
  input IntegerVector ass1;
  input IntegerVector ass2;
  input DAELow.IncidenceMatrix m;
  input DAELow.IncidenceMatrixT mt;
  input Absyn.Path class_;
  output String res;
protected 
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
  type IntegerVector = Integer[:];
  String cname,block_code,extra_funcs,cheader,cfooter;
  IntegerListList blt_states,blt_no_states;
algorithm 
  cname := Absyn.path_string(class_);
  (blt_states,blt_no_states) := DAELow.generate_state_partition(comps, dlow, ass1, ass2, m, mt);
  (block_code,extra_funcs) := generate_ode_blocks(dlow, ass1, ass2, blt_states);
  cheader := generate_ode_header();
  cfooter := generate_ode_footer();
  res := Util.string_append_list({extra_funcs,cheader,block_code,cfooter});
end generate_ode_code;

protected function generate_ode_header " relation: generate_ode_header
  author: PA
 
  Geneates the header of the function for calcuating the explicit ode
 "
  output String res;
algorithm 
  res := Util.string_append_list(
          {
          "void functionODE(double *x, double *xd, double *y, double *p, \n","                  int nx, int ny, int np, double *t)\n","{\n"});
end generate_ode_header;

protected function generate_ode_footer " relation: generate_ode_footer
  author: PA
 
  Geneates the footer of the function for calcuating the explicit ode
 "
  output String res;
algorithm 
  res := Util.string_append_list({"}\n"});
end generate_ode_footer;

protected function generate_ode_blocks "block code
  extra functions code"
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input IntegerListList in_integerlistlist4;
  output String out_string1;
  output String out_string2;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
algorithm 
  (out_string1,out_string2):=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integerlistlist4)
    local
      String s1,f1,s2,f2,res,res2;
      DAELow.DAELow dae;
      IntegerVector ass1,ass2;
      IntegerList block_;
      IntegerListList blocks;
      Integer eqn;
    case (_,_,_,{}) then ("",""); 
    case (dae,ass1,ass2,(block_ = _ :: _ :: _) :: blocks)
      equation 
        (s1,f1) = generate_ode_system(dae, ass1, ass2, block_) "For system of equations ";
        (s2,f2) = generate_ode_blocks(dae, ass1, ass2, blocks);
        res = string_append(s1, s2);
        res2 = string_append(f1, f2); then (res,res2);
    case (dae,ass1,ass2,(block_ = {eqn}) :: blocks)
      equation 
        (s1,f1) = generate_ode_equation(dae, ass1, ass2, eqn) "for single equations ";
        (s2,f2) = generate_ode_blocks(dae, ass1, ass2, blocks);
        res = string_append(s1, s2);
        res2 = string_append(f1, f2); then (res,res2);
    case (_,_,_,_)
      equation 
        print("-generate_ode_blocks failed\n"); then fail();
  end matchcontinue;
end generate_ode_blocks;

protected function generate_ode_system "system code "
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input IntegerList in_integerlist4;
  output String out_string1;
  output String out_string2;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
algorithm 
  (out_string1,out_string2):=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integerlist4)
    local
      type DAELow_EquationList = list<DAELow.Equation>;
      type DAELow_VarList = list<DAELow.Var>;
      type IntegerListArray = IntegerList[:];
      type TupleIntegerIntegerDAELow_Equation = tuple<Integer,Integer,DAELow.Equation>;
      type TupleIntegerIntegerDAELow_EquationList = list<TupleIntegerIntegerDAELow_Equation>;
      type Algorithm_AlgorithmArray = Algorithm.Algorithm[:];
      DAELow_EquationList eqn_lst;
      DAELow_VarList var_lst,var_lst_1;
      DAELow.Variables vars_1,vars,knvars;
      DAELow.EquationArray eqns_1,eqns,se,ie;
      DAELow.DAELow subsystem_dae,daelow;
      IntegerListArray m,m_1,mt_1;
      TupleIntegerIntegerDAELow_EquationList jac;
      DAELow.JacobianType jac_tp;
      String str,f1,jac_tp_str,res;
      Algorithm_AlgorithmArray al;
      DAELow.EventInfo ev;
      IntegerVector ass1,ass2;
      IntegerList block_;
    case ((daelow = DAELow.DAELOW(vars,knvars,eqns,se,ie,al,ev)),ass1,ass2,block_) "system code  extra function code"
      equation 
        (eqn_lst,var_lst) = Util.list_map_3_2(block_, get_equation_and_solved_var, eqns, vars, ass2) " extract the variables and equations of the block.";
        var_lst_1 = Util.list_map(var_lst, transform_x_to_xd);
        vars_1 = DAELow.list_var(var_lst_1);
        eqns_1 = DAELow.list_equation(eqn_lst);
        subsystem_dae = DAELow.DAELOW(vars_1,knvars,eqns_1,se,ie,al,ev) "not used";
        m = DAELow.incidence_matrix(subsystem_dae);
        m_1 = DAELow.abs_incidence_matrix(m);
        mt_1 = DAELow.transpose_matrix(m_1);
        jac = DAELow.calculate_jacobian(vars_1, eqns_1, m_1, mt_1) "calculate jacobian. If constant, linear system of equations. Otherwise nonlinear ";
        jac_tp = DAELow.analyze_jacobian(subsystem_dae, jac);
        (str,f1) = generate_ode_system2(subsystem_dae, jac, jac_tp);
        jac_tp_str = DAELow.jacobian_type_str(jac_tp);
        res = Util.string_append_list(
          {"/* system of equations of type ",jac_tp_str,"*/\n",str,
          "/* End system of equations */\n"}); then (res,f1);
  end matchcontinue;
end generate_ode_system;

protected function generate_ode_system2
  input DAELow.DAELow in_daelow;
  input TupleIntegerIntegerDAELow_EquationList in_tupleintegerintegerdaelow_equationlist;
  input DAELow.JacobianType in_jacobiantype;
  output String out_string1;
  output String out_string2;
protected 
  type TupleIntegerIntegerDAELow_Equation = tuple<Integer,Integer,DAELow.Equation>;
  type TupleIntegerIntegerDAELow_EquationList = list<TupleIntegerIntegerDAELow_Equation>;
algorithm 
  (out_string1,out_string2):=
  matchcontinue (in_daelow,in_tupleintegerintegerdaelow_equationlist,in_jacobiantype)
    local
      type DAELow_EquationList = list<DAELow.Equation>;
      type DAELow_VarList = list<DAELow.Var>;
      type Exp_ComponentRefList = list<Exp.ComponentRef>;
      Integer eqn_size;
      String s1,f1,s2,s3,s4,str;
      DAELow.DAELow d;
      DAELow.Variables v,kv;
      DAELow.EquationArray eqn;
      TupleIntegerIntegerDAELow_EquationList jac;
      DAELow_EquationList eqn_lst;
      DAELow_VarList var_lst;
      Exp_ComponentRefList crefs;
    case ((d = DAELow.DAELOW(v,kv,eqn,_,_,_,_)),jac,DAELow.JAC_CONSTANT()) "system code  extra functions code  constant jacobians. Linear system of equations (A x = b) where
	   A and b are constants. "
      equation 
        eqn_size = DAELow.equation_size(eqn);
        (s1,f1) = generate_ode_system2(d, jac, DAELow.JAC_TIME_VARYING()); then (s1,f1);
    case ((d = DAELow.DAELOW(v,kv,eqn,_,_,_,_)),jac,DAELow.JAC_TIME_VARYING()) "Time varying jacobian. Linear system of equations that needs to 
	 * be solved during runtime. "
      equation 
        eqn_size = DAELow.equation_size(eqn);
        s1 = generate_ode_system2_declaration(eqn_size);
        s2 = generate_ode_system2_populate_Ab(jac, v, eqn);
        s3 = generate_ode_system2_solve_call(eqn_size);
        s4 = generate_ode_system2_collect_results(v);
        str = Util.string_append_list({s1,s2,s3,s4}); then (str,"");
    case (DAELow.DAELOW(v,kv,eqn,_,_,_,_),jac,DAELow.JAC_NONLINEAR()) "Time varying nonlinear jacobian. Non-linear system of equations"
      equation 
        eqn_lst = DAELow.equation_list(eqn);
        var_lst = DAELow.var_list(v);
        crefs = Util.list_map(var_lst, DAELow.var_cref);
        (str,f1) = generate_ode_system2_nonlinear_residuals(crefs, eqn_lst); then (str,f1);
    case (_,_,_)
      equation 
        print("-generate_ode_system2 failed\n"); then fail();
  end matchcontinue;
end generate_ode_system2;

protected function generate_ode_system2_nonlinear_residuals "solve code "
  input Exp_ComponentRefList in_exp_componentreflist;
  input DAELow_EquationList in_daelow_equationlist;
  output String out_string1;
  output String out_string2;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
  type DAELow_EquationList = list<DAELow.Equation>;
algorithm 
  (out_string1,out_string2):=
  matchcontinue (in_exp_componentreflist,in_daelow_equationlist)
    local
      VarTransform.VariableReplacements repl;
      String s1,str_id,size_str,func,s2,s3,s4,res;
      Integer id,eqn_size;
      Exp_ComponentRefList crs;
      DAELow_EquationList eqns;
    case (crs,eqns) "solve code  extra functions: residual func"
      equation 
        repl = make_residual_replacements(crs);
        s1 = generate_ode_system2_nonlinear_residuals2(eqns, 0, repl);
        id = tick();
        str_id = int_string(id);
        eqn_size = list_length(eqns);
        size_str = int_string(eqn_size);
        func = Util.string_append_list(
          {"void residualFunc",str_id,
          "(int* n, double* xloc, double*res, int* iflag)\n","{\n",s1,"}\n"});
        s2 = generate_ode_system2_nonlinear_setvector(crs, 0);
        s3 = generate_ode_system2_nonlinear_call(str_id);
        s4 = generate_ode_system2_nonlinear_store_results(crs, 0);
        res = Util.string_append_list(
          {TAB,"start_nonlinear_system(",size_str,");\n",s2,s3,s4,TAB,
          "end_nonlinear_system();\n"}); then (res,func);
    case (crs,eqns)
      equation 
        print("generate_ode_system2_nonlinear_residuals failed\n"); then fail();
  end matchcontinue;
end generate_ode_system2_nonlinear_residuals;

protected function make_residual_replacements
  input Exp_ComponentRefList crefs;
  output VarTransform.VariableReplacements repl_1;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
  VarTransform.VariableReplacements repl,repl_1;
algorithm 
  repl := VarTransform.empty_replacements();
  repl_1 := make_residual_replacements2(repl, crefs, 0);
end make_residual_replacements;

protected function make_residual_replacements2
  input VarTransform.VariableReplacements in_variablereplacements;
  input Exp_ComponentRefList in_exp_componentreflist;
  input Integer in_integer;
  output VarTransform.VariableReplacements out_variablereplacements;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_variablereplacements:=
  matchcontinue (in_variablereplacements,in_exp_componentreflist,in_integer)
    local
      VarTransform.VariableReplacements repl,repl_1,repl_2;
      String pstr,str;
      Integer pos_1,pos;
      Exp.ComponentRef cr;
      Exp_ComponentRefList crs;
    case (repl,{},_) then repl; 
    case (repl,cr :: crs,pos)
      equation 
        pstr = int_string(pos);
        str = Util.string_append_list({"xloc[",pstr,"]"});
        repl_1 = VarTransform.add_replacement(repl, cr, Exp.CREF_IDENT(str,{}));
        pos_1 = pos + 1;
        repl_2 = make_residual_replacements2(repl_1, crs, pos_1); then repl_2;
  end matchcontinue;
end make_residual_replacements2;

protected function generate_ode_system2_nonlinear_setvector "index iterator "
  input Exp_ComponentRefList in_exp_componentreflist;
  input Integer in_integer;
  output String out_string;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_string:=
  matchcontinue (in_exp_componentreflist,in_integer)
    local
      String cr_str,indx_str,s1,res;
      Integer indx_1,indx;
      Exp.ComponentRef cr;
      Exp_ComponentRefList crs;
    case ({},_) then ""; 
    case (cr :: crs,indx)
      equation 
        cr_str = Exp.print_component_ref_str(cr);
        indx_str = int_string(indx);
        indx_1 = indx + 1;
        s1 = generate_ode_system2_nonlinear_setvector(crs, indx_1);
        res = Util.string_append_list({TAB,"nls_x[",indx_str,"] = ",cr_str,";\n",s1}); then res;
  end matchcontinue;
end generate_ode_system2_nonlinear_setvector;

protected function generate_ode_system2_nonlinear_call
  input String func_id;
  output String res;
algorithm 
  res := Util.string_append_list(
          {TAB,"hybrd_(residualFunc",func_id,
          ",&n, nls_x,nls_fvec,&xtol,&maxfev,&ml,&mu,","&epsfcn,\n",TAB,TAB,"nls_diag,&mode,&factor,&nprint,",
          "&info,&nfev,nls_fjac,&ldfjac,\n",TAB,TAB,"nls_r,","&lr,nls_qtf,nls_wa1,nls_wa2,nls_wa3,",
          "nls_wa4);\n",TAB,"if (info == 0) {\n",TAB,TAB,"printf(\"improper ",
          "input parameters to nonlinear eq. syst.\\n\");\n",TAB,"}\n",TAB,"if (info >= 2 && info <= 5) {\n",TAB,TAB,
          "printf(\"error solving nonlinear"," system.\\n\");\n",TAB,"}\n"});
end generate_ode_system2_nonlinear_call;

protected function generate_ode_system2_nonlinear_store_results
  input Exp_ComponentRefList in_exp_componentreflist;
  input Integer in_integer;
  output String out_string;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  out_string:=
  matchcontinue (in_exp_componentreflist,in_integer)
    local
      String cr_str,indx_str,s1,res;
      Integer indx_1,indx;
      Exp.ComponentRef cr;
      Exp_ComponentRefList crs;
    case ({},_) then ""; 
    case (cr :: crs,indx)
      equation 
        cr_str = Exp.print_component_ref_str(cr);
        indx_str = int_string(indx);
        indx_1 = indx + 1;
        s1 = generate_ode_system2_nonlinear_store_results(crs, indx_1);
        res = Util.string_append_list({TAB,cr_str," = nls_x[",indx_str,"];\n",s1}); then res;
  end matchcontinue;
end generate_ode_system2_nonlinear_store_results;

protected function generate_ode_system2_nonlinear_residuals2 "index iterator"
  input DAELow_EquationList in_daelow_equationlist;
  input Integer in_integer;
  input VarTransform.VariableReplacements in_variablereplacements;
  output String out_string;
protected 
  type DAELow_EquationList = list<DAELow.Equation>;
algorithm 
  out_string:=
  matchcontinue (in_daelow_equationlist,in_integer,in_variablereplacements)
    local
      Exp.Type tp;
      Exp.Exp res_exp,res_exp_1,res_exp_2,e1,e2,e;
      String exp_str,indx_str,s1,res;
      Integer indx_1,indx;
      DAELow_EquationList rest;
      VarTransform.VariableReplacements repl;
    case ({},_,_) then ""; 
    case (DAELow.EQUATION(e1,e2) :: rest,indx,repl)
      equation 
        tp = Exp.typeof(e1);
        res_exp = Exp.BINARY(e1,Exp.SUB(tp),e2);
        res_exp_1 = Exp.simplify(res_exp);
        res_exp_2 = VarTransform.replace_exp(res_exp_1, repl);
        exp_str = print_exp_cpp_str(res_exp_2);
        indx_str = int_string(indx);
        indx_1 = indx + 1;
        s1 = generate_ode_system2_nonlinear_residuals2(rest, indx_1, repl);
        res = Util.string_append_list({TAB,"res[",indx_str,"] = ",exp_str,";\n",s1}); then res;
    case (DAELow.RESIDUAL_EQUATION(e) :: rest,indx,repl)
      equation 
        res_exp_1 = Exp.simplify(e);
        res_exp_2 = VarTransform.replace_exp(res_exp_1, repl);
        exp_str = print_exp_cpp_str(res_exp_2);
        indx_str = int_string(indx);
        indx_1 = indx + 1;
        s1 = generate_ode_system2_nonlinear_residuals2(rest, indx_1, repl);
        res = Util.string_append_list({TAB,"res[",indx_str,"] = ",exp_str,";\n",s1}); then res;
  end matchcontinue;
end generate_ode_system2_nonlinear_residuals2;

protected function generate_ode_system2_declaration " relation: generate_ode_system2_declaration
  author: PA
  
  Generates code for the declaration of A and b when
  solving linear systems of equations.
 "
  input Integer size;
  output String res;
protected 
  String size_str;
algorithm 
  size_str := int_string(size);
  res := Util.string_append_list(
          {TAB,"declare_matrix(A,",size_str,",",size_str,");\n",TAB,
          "declare_vector(b,",size_str,");\n"});
end generate_ode_system2_declaration;

protected function generate_ode_system2_populate_Ab
  input TupleIntegerIntegerDAELow_EquationList jac;
  input DAELow.Variables vars;
  input DAELow.EquationArray eqns;
  output String res;
protected 
  type TupleIntegerIntegerDAELow_Equation = tuple<Integer,Integer,DAELow.Equation>;
  type TupleIntegerIntegerDAELow_EquationList = list<TupleIntegerIntegerDAELow_Equation>;
  String s1,s2;
algorithm 
  s1 := generate_ode_system2_populate_A(jac, vars, eqns);
  s2 := generate_ode_system2_populate_b(jac, vars, eqns);
  res := string_append(s1, s2);
end generate_ode_system2_populate_Ab;

protected function generate_ode_system2_populate_A
  input TupleIntegerIntegerDAELow_EquationList jac;
  input DAELow.Variables vars;
  input DAELow.EquationArray eqns;
  output String res;
protected 
  type TupleIntegerIntegerDAELow_Equation = tuple<Integer,Integer,DAELow.Equation>;
  type TupleIntegerIntegerDAELow_EquationList = list<TupleIntegerIntegerDAELow_Equation>;
  Integer n_rows;
algorithm 
  n_rows := DAELow.equation_size(eqns);
  res := generate_ode_system2_populate_A2(jac, vars, eqns, n_rows);
end generate_ode_system2_populate_A;

protected function generate_ode_system2_populate_A2 "n rows "
  input TupleIntegerIntegerDAELow_EquationList in_tupleintegerintegerdaelow_equationlist;
  input DAELow.Variables in_variables;
  input DAELow.EquationArray in_equationarray;
  input Integer in_integer;
  output String out_string;
protected 
  type TupleIntegerIntegerDAELow_Equation = tuple<Integer,Integer,DAELow.Equation>;
  type TupleIntegerIntegerDAELow_EquationList = list<TupleIntegerIntegerDAELow_Equation>;
algorithm 
  out_string:=
  matchcontinue (in_tupleintegerintegerdaelow_equationlist,in_variables,in_equationarray,in_integer)
    local
      Integer r_1,c_1,r,c,n_rows;
      String rs,rc,n_rows_str,es,s1,res;
      Exp.Exp exp;
      TupleIntegerIntegerDAELow_EquationList jac;
      DAELow.Variables vars;
      DAELow.EquationArray eqn;
    case ({},_,_,_) then ""; 
    case ((r,c,DAELow.RESIDUAL_EQUATION(exp)) :: jac,vars,eqn,n_rows)
      equation 
        r_1 = r - 1;
        c_1 = c - 1;
        rs = int_string(r_1);
        rc = int_string(c_1);
        n_rows_str = int_string(n_rows);
        es = print_exp_cpp_str(exp);
        s1 = generate_ode_system2_populate_A(jac, vars, eqn);
        res = Util.string_append_list(
          {TAB,"set_matrix_elt(A,",rs,", ",rc,", ",n_rows_str,", ",es,
          ");\n",s1}); then res;
  end matchcontinue;
end generate_ode_system2_populate_A2;

protected function generate_ode_system2_populate_b
  input TupleIntegerIntegerDAELow_EquationList jac;
  input DAELow.Variables vars;
  input DAELow.EquationArray eqns;
  output String res;
protected 
  type TupleIntegerIntegerDAELow_Equation = tuple<Integer,Integer,DAELow.Equation>;
  type TupleIntegerIntegerDAELow_EquationList = list<TupleIntegerIntegerDAELow_Equation>;
  type DAELow_EquationList = list<DAELow.Equation>;
  DAELow_EquationList eqn_lst;
algorithm 
  eqn_lst := DAELow.equation_list(eqns);
  res := generate_ode_system2_populate_b2(eqn_lst, vars, 0);
end generate_ode_system2_populate_b;

protected function generate_ode_system2_populate_b2 "index iterator"
  input DAELow_EquationList in_daelow_equationlist;
  input DAELow.Variables in_variables;
  input Integer in_integer;
  output String out_string;
protected 
  type DAELow_EquationList = list<DAELow.Equation>;
algorithm 
  out_string:=
  matchcontinue (in_daelow_equationlist,in_variables,in_integer)
    local
      Exp.Type tp;
      Exp.Exp new_exp,rhs_exp,rhs_exp_1,rhs_exp_2,e1,e2;
      String rhs_str,index_str,s1,res;
      Integer index_1,index;
      DAELow_EquationList rest;
      DAELow.Variables v;
    case ({},_,_) then ""; 
    case (DAELow.EQUATION(e1,e2) :: rest,v,index)
      equation 
        tp = Exp.typeof(e1);
        new_exp = Exp.BINARY(e1,Exp.SUB(tp),e2);
        rhs_exp = DAELow.get_eqnsys_rhs_exp(new_exp, v);
        rhs_exp_1 = Exp.UNARY(Exp.UMINUS(tp),rhs_exp);
        rhs_exp_2 = Exp.simplify(rhs_exp_1);
        rhs_str = print_exp_cpp_str(rhs_exp_2);
        index_str = int_string(index);
        index_1 = index + 1;
        s1 = generate_ode_system2_populate_b2(rest, v, index_1);
        res = Util.string_append_list({TAB,"set_vector_elt(b,",index_str,", ",rhs_str,");\n",s1}); then res;
  end matchcontinue;
end generate_ode_system2_populate_b2;

protected function generate_ode_system2_solve_call
  input Integer eqn_size;
  output String res;
protected 
  String size_str;
algorithm 
  size_str := int_string(eqn_size);
  res := Util.string_append_list({TAB,"solve_linear_equation_system(A,b,",size_str,");\n"});
end generate_ode_system2_solve_call;

protected function generate_ode_system2_collect_results
  input DAELow.Variables vars;
  output String res;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
  type StringList = list<String>;
  DAELow_VarList var_lst,var_lst_1;
  Exp_ComponentRefList crefs;
  StringList strs;
algorithm 
  var_lst := DAELow.var_list(vars);
  var_lst_1 := list_reverse(var_lst);
  crefs := Util.list_map(var_lst, DAELow.var_cref);
  strs := Util.list_map(crefs, Exp.print_component_ref_str);
  res := generate_ode_system2_collect_results2(strs, 0);
end generate_ode_system2_collect_results;

protected function generate_ode_system2_collect_results2
  input StringList in_stringlist;
  input Integer in_integer;
  output String out_string;
protected 
  type StringList = list<String>;
algorithm 
  out_string:=
  matchcontinue (in_stringlist,in_integer)
    local
      Integer index_1,index;
      String s1,index_str,res,str;
      StringList strs;
    case ({},_) then ""; 
    case (str :: strs,index)
      equation 
        index_1 = index + 1;
        s1 = generate_ode_system2_collect_results2(strs, index_1);
        index_str = int_string(index);
        res = Util.string_append_list({TAB,str," = get_vector_elt(b,",index_str,");\n",s1}); then res;
  end matchcontinue;
end generate_ode_system2_collect_results2;

protected function transform_x_to_xd " relation transform_x_to_xd
  author: PA
  this function transforms x variables (in the state vector)
  to corresponding xd variable (in the derivatives vector)
 "
  input DAELow.Var in_var;
  output DAELow.Var out_var;
algorithm 
  out_var:=
  matchcontinue (in_var)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type Values_ValueOption = Option<Values.Value>;
      type Exp_SubscriptList = list<Exp.Subscript>;
      type Absyn_PathList = list<Absyn.Path>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      String index_str,res,name;
      Exp.ComponentRef cr;
      DAE.VarDirection dir;
      DAE.Type tp;
      Exp_ExpOption exp,st;
      Values_ValueOption v;
      Exp_SubscriptList dim;
      Integer index;
      Absyn_PathList classes;
      DAE_VariableAttributesOption attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
    case (DAELow.VAR(cr,DAELow.STATE(),dir,tp,exp,v,dim,st,index,name,classes,attr,comment,flow_))
      equation 
        index_str = int_string(index);
        res = Util.string_append_list({"xd[",index_str,"]"}); then DAELow.VAR(Exp.CREF_IDENT(res,{}),DAELow.STATE(),dir,tp,exp,v,dim,st,
          index,name,classes,attr,comment,flow_);
    case (v)
      local DAELow.Var v; then v;
  end matchcontinue;
end transform_x_to_xd;

protected function get_equation_and_solved_var "assignments2"
  input Integer in_integer;
  input DAELow.EquationArray in_equationarray;
  input DAELow.Variables in_variables;
  input IntegerVector in_integervector;
  output DAELow.Equation out_equation;
  output DAELow.Var out_var;
protected 
  type IntegerVector = Integer[:];
algorithm 
  (out_equation,out_var):=
  matchcontinue (in_integer,in_equationarray,in_variables,in_integervector)
    local
      Integer e_1,v,e;
      DAELow.Equation eqn;
      DAELow.Var var;
      DAELow.EquationArray eqns;
      DAELow.Variables vars;
      IntegerVector ass2;
    case (e,eqns,vars,ass2) "equation no.  assignments2"
      equation 
        e_1 = e - 1;
        eqn = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        var = DAELow.get_var_at(vars, v); then (eqn,var);
  end matchcontinue;
end get_equation_and_solved_var;

protected function is_non_state
  input DAELow.VarKind in_varkind;
algorithm 
  _:=
  matchcontinue (in_varkind)
    case (DAELow.VARIABLE()) then (); 
    case (DAELow.DUMMY_DER()) then (); 
    case (DAELow.DUMMY_STATE()) then (); 
  end matchcontinue;
end is_non_state;

protected function generate_ode_equation "assignements 1 
  assignements 2 
  equation number
  equation code "
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input Integer in_integer4;
  output String out_string1;
  output String out_string2;
protected 
  type IntegerVector = Integer[:];
algorithm 
  (out_string1,out_string2):=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integer4)
    local
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Exp.Exp e1,e2,varexp,expr;
      DAELow.Var v;
      Exp.ComponentRef cr,cr_1;
      DAELow.VarKind kind;
      String origname,expr_str,cr_str,res,indxs,id,f1;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow.Variables vars;
      DAELow.EquationArray eqns;
      IntegerVector ass1,ass2;
      Integer e,indx;
      DAELow.Equation eqn;
    case (DAELow.DAELOW(vars,_,eqns,_,_,_,_),ass1,ass2,e) "assignements 1  assignements 2  equation number equation code  extra functions "
      equation 
        (DAELow.EQUATION(e1,e2),v = DAELow.VAR(cr,kind,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_)) = get_equation_and_solved_var(e, eqns, vars, ass2) "Solving for non-states ";
        is_non_state(kind);
        varexp = Exp.CREF(cr,Exp.REAL());
        expr = Exp.solve(e1, e2, varexp);
        expr_str = print_exp_cpp_str(expr);
        cr_str = Exp.print_component_ref_str(cr);
        res = Util.string_append_list({"  ",cr_str,"=",expr_str,";\n"}); then (res,"");
    case (DAELow.DAELOW(vars,_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        (DAELow.EQUATION(e1,e2),DAELow.VAR(cr,DAELow.STATE(),_,_,_,_,_,_,indx,origname,_,dae_var_attr,comment,flow_)) = get_equation_and_solved_var(e, eqns, vars, ass2) "Solving the state s means solving for der(s) ";
        indxs = int_string(indx);
        id = Util.string_append_list({"xd[",indxs,"]"});
        cr_1 = Exp.CREF_IDENT(id,{});
        varexp = Exp.CREF(cr_1,Exp.REAL());
        expr = Exp.solve(e1, e2, varexp);
        expr_str = print_exp_cpp_str(expr);
        cr_str = Exp.print_component_ref_str(cr_1);
        res = Util.string_append_list({"  ",cr_str,"=",expr_str,";\n"}); then (res,"");
    case (DAELow.DAELOW(vars,_,eqns,_,_,_,_),ass1,ass2,e) "state nonlinear "
      equation 
        (eqn = DAELow.EQUATION(e1,e2),DAELow.VAR(cr,DAELow.STATE(),_,_,_,_,_,_,indx,origname,_,dae_var_attr,comment,flow_)) = get_equation_and_solved_var(e, eqns, vars, ass2);
        indxs = int_string(indx);
        id = Util.string_append_list({"xd[",indxs,"]"});
        cr_1 = Exp.CREF_IDENT(id,{});
        varexp = Exp.CREF(cr_1,Exp.REAL());
        failure(_ = Exp.solve(e1, e2, varexp));
        (res,f1) = generate_ode_system2_nonlinear_residuals({cr_1}, {eqn}); then (res,f1);
  end matchcontinue;
end generate_ode_equation;

protected function generate_nonlinear_equations "residuals "
  input Exp_ComponentRefList in_exp_componentreflist;
  input Exp_ExpList in_exp_explist;
  output String out_string;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  out_string:=
  matchcontinue (in_exp_componentreflist,in_exp_explist)
    case (_,_) then "/* Generate non-linear equations not impl. yet.*/\n"; 
  end matchcontinue;
end generate_nonlinear_equations;

public function generate_functions " relation generate_functions
 
  Finds the called functions in daelow and generates code for them
  from the given DAE. Hence, the functions must exist in the DAE.Element list.
"
  input DAE.DAElist in_daelist;
  input DAELow.DAELow in_daelow;
  input Absyn.Path in_path;
  input String in_string;
algorithm 
  _:=
  matchcontinue (in_daelist,in_daelow,in_path,in_string)
    local
      type Absyn_PathList = list<Absyn.Path>;
      type StringList = list<String>;
      type DAE_ElementList = list<DAE.Element>;
      Absyn_PathList funcpaths;
      StringList debugpathstrs;
      String debugpathstr,debugstr,filename;
      DAE_ElementList funcelems,elements;
      DAE.DAElist dae;
      DAELow.DAELow dlow;
      Absyn.Path path;
    case ((dae = DAE.DAE(elements)),dlow,path,filename)
      equation 
        funcpaths = get_called_functions(dae, dlow);
        Debug.fprint("info", "Found called functions: ");
        debugpathstrs = Util.list_map(funcpaths, Absyn.path_string) "debug";
        debugpathstr = Util.string_delimit_list(debugpathstrs, ", ") "debug";
        Debug.fprintln("info", debugpathstr) "debug";
        funcelems = Util.list_map_1(funcpaths, DAE.get_named_function, elements);
        funcelems = Util.list_flatten(funcelems);
        debugstr = Print.get_string();
        Print.clear_buf();
        Codegen.generate_functions(DAE.DAE(funcelems));
        Print.write_buf(filename); then ();
    case (_,_,_,_)
      equation 
        print("-generate_functions failed\n"); then fail();
  end matchcontinue;
end generate_functions;

public function generate_init_data "filename 
  start time
  stop time 
  step size "
  input DAELow.DAELow in_daelow1;
  input Absyn.Path in_path2;
  input String in_string3;
  input Real in_real4;
  input Real in_real5;
  input Real in_real6;
algorithm 
  _:=
  matchcontinue (in_daelow1,in_path2,in_string3,in_real4,in_real5,in_real6)
    local
      String start_str,stop_str,step_str,nx_str,ny_str,np_str,init_str,str,filename;
      Integer nx,ny,np;
      DAELow.DAELow dlow;
      Absyn.Path class_;
      Real start,stop,step;
    case (dlow,class_,filename,start,stop,step) "classname  filename  start time stop time  step size "
      equation 
        start_str = real_string(start);
        stop_str = real_string(stop);
        step_str = real_string(step);
        (nx,ny,np) = DAELow.calculate_sizes(dlow);
        nx_str = int_string(nx);
        ny_str = int_string(ny);
        np_str = int_string(np);
        init_str = generate_init_data2(dlow, nx, ny, np);
        str = Util.string_append_list(
          {start_str," // start value\n",stop_str," // stop value\n",
          step_str," // step value\n",nx_str," // n states\n",ny_str," // n alg vars\n",
          np_str," //n parameters\n",init_str});
        System.write_file(filename, str); then ();
    case (_,_,_,_,_,_)
      equation 
        print("-generate_init_data failed\n"); then fail();
  end matchcontinue;
end generate_init_data;

protected function generate_init_data2 "nx
  ny
  np"
  input DAELow.DAELow in_daelow1;
  input Integer in_integer2;
  input Integer in_integer3;
  input Integer in_integer4;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_daelow1,in_integer2,in_integer3,in_integer4)
    local
      type DAELow_VarList = list<DAELow.Var>;
      type StringArray = String[:];
      type StringList = list<String>;
      type Algorithm_AlgorithmArray = Algorithm.Algorithm[:];
      DAELow_VarList var_lst,knvar_lst;
      StringArray nxarr,nxdarr,nyarr,nparr,nxarr1,nxdarr1,nyarr1,nparr1,nxarr2,nxdarr2,nyarr2,nparr2,nxarr3,nxdarr3,nyarr3,nparr3;
      StringList nx_lst,nxd_lst,ny_lst,np_lst,whole_lst;
      String res;
      DAELow.Variables vars,knvars;
      DAELow.EquationArray initeqn;
      Algorithm_AlgorithmArray alg;
      Integer nx,ny,np;
    case (DAELow.DAELOW(vars,knvars,_,_,initeqn,alg,_),nx,ny,np) "nx ny np"
      equation 
        var_lst = DAELow.var_list(vars);
        knvar_lst = DAELow.var_list(knvars);
        nxarr = fill("", nx);
        nxdarr = fill("0.0", nx);
        nyarr = fill("", ny);
        nparr = fill("", np);
        (nxarr1,nxdarr1,nyarr1,nparr1) = generate_init_data3(var_lst, nxarr, nxdarr, nyarr, nparr);
        (nxarr2,nxdarr2,nyarr2,nparr2) = generate_init_data3(knvar_lst, nxarr1, nxdarr1, nyarr1, nparr1);
        (nxarr3,nxdarr3,nyarr3,nparr3) = generate_init_data4(knvar_lst, nxarr2, nxdarr2, nyarr2, nparr2);
        nx_lst = array_list(nxarr3);
        nxd_lst = array_list(nxdarr3);
        ny_lst = array_list(nyarr3);
        np_lst = array_list(nparr3);
        whole_lst = Util.list_flatten({nx_lst,nxd_lst,ny_lst,np_lst});
        res = Util.string_delimit_list(whole_lst, "\n"); then res;
  end matchcontinue;
end generate_init_data2;

protected function print_exp_str_opt " relation: print_exp_str_opt
  Helper function to generate_init_data2
  Prints expression value that is opional for initial values.
  If NONE is passed. The default value 0.0 is returned.
 "
  input Exp_ExpOption in_exp_expoption;
  output String out_string;
protected 
  type Exp_ExpOption = Option<Exp.Exp>;
algorithm 
  out_string:=
  matchcontinue (in_exp_expoption)
    local
      String str;
      Exp.Exp e;
    case NONE then "0.0"; 
    case SOME(e)
      equation 
        str = Exp.print_exp_str(e); then str;
  end matchcontinue;
end print_exp_str_opt;

protected function generate_init_data3 "state strings 
  derivative strings
  alg. var strings 
  param. strings 
  updated state strings 
  updated derivative strings
  updated alg. var strings 
  updated param. strings "
  input DAELow_VarList in_daelow_varlist1;
  input StringArray in_stringarray2;
  input StringArray in_stringarray3;
  input StringArray in_stringarray4;
  input StringArray in_stringarray5;
  output StringArray out_stringarray1;
  output StringArray out_stringarray2;
  output StringArray out_stringarray3;
  output StringArray out_stringarray4;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringArray = String[:];
algorithm 
  (out_stringarray1,out_stringarray2,out_stringarray3,out_stringarray4):=
  matchcontinue (in_daelow_varlist1,in_stringarray2,in_stringarray3,in_stringarray4,in_stringarray5)
    local
      type Exp_ExpOption = Option<Exp.Exp>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      StringArray nxarr,nxdarr,nyarr,nparr,nxarr_1,nxdarr_1,nyarr_1,nparr_1;
      String v,str,origname;
      Exp.ComponentRef cr;
      Exp_ExpOption start;
      Integer indx;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList rest;
    case ({},nxarr,nxdarr,nyarr,nparr) then (nxarr,nxdarr,nyarr,nparr); 
    case (DAELow.VAR(cr,DAELow.VARIABLE(),_,_,_,_,_,start,indx,origname,_,dae_var_attr,comment,flow_) :: rest,nxarr,nxdarr,nyarr,nparr)
      equation 
        v = print_exp_opt_str_if_const(start) "algebraic variables ";
        str = Util.string_append_list({v," // ",origname});
        nyarr = array_update(nyarr, indx + 1, str);
        (nxarr,nxdarr,nyarr,nparr) = generate_init_data3(rest, nxarr, nxdarr, nyarr, nparr); then (nxarr,nxdarr,nyarr,nparr);
    case (DAELow.VAR(cr,DAELow.DISCRETE(),_,_,_,_,_,start,indx,origname,_,dae_var_attr,comment,flow_) :: rest,nxarr,nxdarr,nyarr,nparr)
      equation 
        v = print_exp_opt_str_if_const(start) "algebraic variables ";
        str = Util.string_append_list({v," // ",origname});
        nyarr = array_update(nyarr, indx + 1, str);
        (nxarr,nxdarr,nyarr,nparr) = generate_init_data3(rest, nxarr, nxdarr, nyarr, nparr); then (nxarr,nxdarr,nyarr,nparr);
    case (DAELow.VAR(_,DAELow.STATE(),_,_,_,_,_,start,indx,origname,_,dae_var_attr,comment,flow_) :: rest,nxarr,nxdarr,nyarr,nparr)
      equation 
        v = print_exp_opt_str_if_const(start) "State variables ";
        str = Util.string_append_list({v," // ",origname});
        nxarr = array_update(nxarr, indx + 1, str);
        (nxarr,nxdarr,nyarr,nparr) = generate_init_data3(rest, nxarr, nxdarr, nyarr, nparr); then (nxarr,nxdarr,nyarr,nparr);
    case (DAELow.VAR(_,DAELow.DUMMY_DER(),_,_,_,_,_,start,indx,origname,_,dae_var_attr,comment,flow_) :: rest,nxarr,nxdarr,nyarr,nparr)
      equation 
        v = print_exp_opt_str_if_const(start) "dummy derivatives => algebraic variables ";
        str = Util.string_append_list({v," // ",origname});
        nyarr = array_update(nyarr, indx + 1, str);
        (nxarr,nxdarr,nyarr,nparr) = generate_init_data3(rest, nxarr, nxdarr, nyarr, nparr); then (nxarr,nxdarr,nyarr,nparr);
    case (DAELow.VAR(_,DAELow.DUMMY_STATE(),_,_,_,_,_,start,indx,origname,_,dae_var_attr,comment,flow_) :: rest,nxarr,nxdarr,nyarr,nparr)
      equation 
        v = print_exp_opt_str_if_const(start) "Dummy states => algebraic variables ";
        str = Util.string_append_list({v," // ",origname});
        nyarr = array_update(nyarr, indx + 1, str);
        (nxarr,nxdarr,nyarr,nparr) = generate_init_data3(rest, nxarr, nxdarr, nyarr, nparr); then (nxarr,nxdarr,nyarr,nparr);
    case (_ :: rest,nxarr,nxdarr,nyarr,nparr)
      equation 
        (nxarr_1,nxdarr_1,nyarr_1,nparr_1) = generate_init_data3(rest, nxarr, nxdarr, nyarr, nparr); then (nxarr_1,nxdarr_1,nyarr_1,nparr_1);
  end matchcontinue;
end generate_init_data3;

public function print_exp_opt_str_if_const
  input Exp_ExpOption in_exp_expoption;
  output String out_string;
protected 
  type Exp_ExpOption = Option<Exp.Exp>;
algorithm 
  out_string:=
  matchcontinue (in_exp_expoption)
    local
      String res;
      Exp.Exp e;
    case (SOME(e))
      equation 
        true = Exp.is_const(e);
        res = print_exp_str_opt(SOME(e)); then res;
    case (_)
      equation 
        res = print_exp_str_opt(NONE); then "0.0";
  end matchcontinue;
end print_exp_opt_str_if_const;

protected function generate_init_data4
  input DAELow_VarList in_daelow_varlist1;
  input StringArray in_stringarray2;
  input StringArray in_stringarray3;
  input StringArray in_stringarray4;
  input StringArray in_stringarray5;
  output StringArray out_stringarray1;
  output StringArray out_stringarray2;
  output StringArray out_stringarray3;
  output StringArray out_stringarray4;
protected 
  type DAELow_VarList = list<DAELow.Var>;
  type StringArray = String[:];
algorithm 
  (out_stringarray1,out_stringarray2,out_stringarray3,out_stringarray4):=
  matchcontinue (in_daelow_varlist1,in_stringarray2,in_stringarray3,in_stringarray4,in_stringarray5)
    local
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      StringArray nxarr,nxdarr,nyarr,nparr,nxarr_1,nxdarr_1,nyarr_1,nparr_1;
      String v,str,origname;
      Values.Value value;
      Integer indx;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList rest,vs;
    case ({},nxarr,nxdarr,nyarr,nparr) then (nxarr,nxdarr,nyarr,nparr); 
    case (DAELow.VAR(_,DAELow.PARAM(),_,_,_,SOME(value),_,_,indx,origname,_,dae_var_attr,comment,flow_) :: rest,nxarr,nxdarr,nyarr,nparr)
      equation 
        v = Values.val_string(value) "Parameters ";
        str = Util.string_append_list({v," // ",origname});
        nparr = array_update(nparr, indx + 1, str);
        (nxarr,nxdarr,nyarr,nparr) = generate_init_data4(rest, nxarr, nxdarr, nyarr, nparr); then (nxarr,nxdarr,nyarr,nparr);
    case (DAELow.VAR(_,DAELow.PARAM(),_,_,_,NONE,_,_,indx,origname,_,dae_var_attr,comment,flow_) :: rest,nxarr,nxdarr,nyarr,nparr)
      equation 
        str = Util.string_append_list({"0.0  // ",origname}) "Parameters without value binding. Should normally not happen ";
        nparr = array_update(nparr, indx + 1, str);
        (nxarr,nxdarr,nyarr,nparr) = generate_init_data4(rest, nxarr, nxdarr, nyarr, nparr); then (nxarr,nxdarr,nyarr,nparr);
    case (_ :: vs,nxarr,nxdarr,nyarr,nparr)
      equation 
        (nxarr_1,nxdarr_1,nyarr_1,nparr_1) = generate_init_data4(vs, nxarr, nxdarr, nyarr, nparr) "Skip alg. vars that are removed 
	 In future we should compare eliminated variables 
	 intial values to their aliases to detect inconsistent
	 initial values.
	 "; then (nxarr_1,nxdarr_1,nyarr_1,nparr_1);
  end matchcontinue;
end generate_init_data4;

protected function dump_when_clauses_str
  input DAELow_WhenClauseList in_daelow_whenclauselist;
  output String out_string;
protected 
  type DAELow_WhenClauseList = list<DAELow.WhenClause>;
algorithm 
  out_string:=
  matchcontinue (in_daelow_whenclauselist)
    local
      String str,str2,res;
      DAELow.WhenClause c;
      DAELow_WhenClauseList xs;
    case {} then ""; 
    case c :: xs
      equation 
        str = dump_when_clause_str(c);
        str2 = dump_when_clauses_str(xs);
        res = string_append(str, str2); then res;
  end matchcontinue;
end dump_when_clauses_str;

protected function dump_when_clause_str
  input DAELow.WhenClause in_whenclause;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_whenclause)
    local
      String str1,res;
      Exp.Exp exp;
    case DAELow.WHEN_CLAUSE(exp)
      equation 
        str1 = Exp.print_exp_str(exp);
        res = Util.string_append_list({"when ",str1,"\n"}); then res;
  end matchcontinue;
end dump_when_clause_str;

protected function generate_when_clauses "assignments1 
  assignments2 
  blocks "
  input String in_string1;
  input DAE.DAElist in_daelist2;
  input DAELow.DAELow in_daelow3;
  input IntegerVector in_integervector4;
  input IntegerVector in_integervector5;
  input IntegerListList in_integerlistlist6;
  output String out_string;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
algorithm 
  out_string:=
  matchcontinue (in_string1,in_daelist2,in_daelow3,in_integervector4,in_integervector5,in_integerlistlist6)
    local
      type DAELow_WhenClauseList = list<DAELow.WhenClause>;
      String when_str,wcs,res,cname;
      DAE.DAElist dae;
      DAELow.DAELow dlow;
      DAELow_WhenClauseList wc;
      IntegerVector ass1,ass2;
      IntegerListList blocks;
    case (cname,dae,(dlow = DAELow.DAELOW(_,_,_,_,_,_,DAELow.EVENT_INFO(wc,_))),ass1,ass2,blocks) "assignments1  assignments2  blocks "
      equation 
        when_str = build_when_blocks(dae, dlow, ass1, ass2, blocks);
        wcs = dump_when_clauses_str(wc);
        res = Util.string_append_list({"/*\n",wcs,when_str,"*/\n"}); then res;
    case (_,_,_,_,_,_)
      equation 
        print("-generate_when_clauses failed\n"); then fail();
  end matchcontinue;
end generate_when_clauses;

protected function build_when_blocks
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input IntegerListList in_integerlistlist5;
  output String out_string;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
algorithm 
  out_string:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_integerlistlist5)
    local
      String eqn_str,eqn_str2,res;
      DAE.DAElist dae;
      DAELow.DAELow dlow;
      IntegerVector ass1,ass2;
      IntegerList block_;
      Integer eqn;
      IntegerListList blocks;
    case (_,_,_,_,{}) then ""; 
    case (dae,dlow,ass1,ass2,(block_ = {eqn}) :: blocks)
      equation 
        eqn_str = build_when_equation(dae, dlow, ass1, ass2, eqn);
        eqn_str2 = build_when_blocks(dae, dlow, ass1, ass2, blocks);
        res = string_append(eqn_str, eqn_str2); then res;
    case (_,_,_,_,_) then ""; 
    case (_,_,_,_,_)
      equation 
        print("-generate_when_blocks failed\n"); then fail();
  end matchcontinue;
end build_when_blocks;

protected function build_when_equation "assignments1
  assignments2
  equation no. "
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input Integer in_integer5;
  output String out_string;
protected 
  type IntegerVector = Integer[:];
algorithm 
  out_string:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_integer5)
    local
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Integer e_1,v,v_1;
      DAELow.WhenEquation e;
      Exp.ComponentRef cr;
      Exp.Exp expr;
      DAELow.Var va;
      DAELow.VarKind kind;
      String origname,assignment_string;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAE.DAElist dae;
      DAELow.VariableArray vararr;
      DAELow.EquationArray eqns;
      IntegerVector ass1,ass2;
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e) "assignments1 assignments2 equation no. "
      equation 
        e_1 = e - 1;
        DAELow.WHEN_EQUATION(e = DAELow.WHEN_EQ(_,cr,expr)) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1;
        (va = DAELow.VAR(cr,kind,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_)) = DAELow.vararray_nth(vararr, v_1);
        assignment_string = build_assignment(dae, cr, expr, origname); then assignment_string;
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1;
        DAELow.WHEN_EQUATION(e = DAELow.WHEN_EQ(_,cr,expr)) = DAELow.equation_nth(eqns, e_1);
        print("-build_when_equation failed\n"); then fail();
    case (_,_,_,_,_) then ""; 
  end matchcontinue;
end build_when_equation;

protected function generate_compute_residual_state "assignments1 
  assignments2 
  blocks "
  input String cname;
  input DAE.DAElist dae;
  input DAELow.DAELow dlow;
  input IntegerVector ass1;
  input IntegerVector ass2;
  input IntegerListList blocks;
  output String res;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
  type StringList = list<String>;
  StringList eqn_str_list;
  String cstate;
algorithm 
  eqn_str_list := build_residual_blocks(dae, dlow, ass1, ass2, blocks);
  cstate := Util.string_append_list(eqn_str_list);
  res := Util.string_append_list(
          {
          "int\nfunctionDAE_res(double *t, double *x, double *xd, double *delta, long int *ires, double *rpar, long int* ipar)\n{\n",cstate,TAB,"return 0;\n\n}\n"});
end generate_compute_residual_state;

protected function generate_compute_output
  input String cname;
  input DAE.DAElist dae;
  input DAELow.DAELow dlow;
  input IntegerVector ass1;
  input IntegerVector ass2;
  input IntegerListList blocks;
  output String coutput;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
  type StringList = list<String>;
  String func_header,func_footer;
  StringList eqn_str_lst,res;
algorithm 
  func_header := "int\nfunctionDAE_output(double *t, double* x, double* xd, double* y, double* p)\n{\n";
  func_footer := "return 0;\n\n}\n";
  eqn_str_lst := build_solved_blocks(dae, dlow, ass1, ass2, blocks);
  res := Util.list_flatten({{func_header},eqn_str_lst,{func_footer}});
  coutput := Util.string_append_list(res);
end generate_compute_output;

protected function build_solved_blocks "assignments1
  assignments2
  list of blocks "
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input IntegerListList in_integerlistlist5;
  output StringList out_stringlist;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_integerlistlist5)
    local
      String eqn_str,block_str;
      StringList res;
      DAE.DAElist dae;
      DAELow.DAELow dlow;
      IntegerVector ass1,ass2;
      IntegerList block_;
      Integer eqn;
      IntegerListList blocks;
    case (_,_,_,_,{}) then {}; 
    case (dae,dlow,ass1,ass2,(block_ = {eqn}) :: blocks)
      equation 
        eqn_str = build_equation(dae, dlow, ass1, ass2, eqn) "for single equations ";
        res = build_solved_blocks(dae, dlow, ass1, ass2, blocks); then eqn_str :: res;
    case (dae,dlow,ass1,ass2,block_ :: blocks)
      equation 
        block_str = build_block(dae, dlow, ass1, ass1, block_) "for blocks ";
        res = build_solved_blocks(dae, dlow, ass1, ass2, blocks); then block_str :: res;
    case (_,_,_,_,_)
      equation 
        print("-build_solved_blocks failed\n"); then fail();
  end matchcontinue;
end build_solved_blocks;

protected function build_block "assignments1 
  assignments2 
  block of equations"
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input IntegerList in_integerlist5;
  output String out_string;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
algorithm 
  out_string:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_integerlist5)
    local
      type Exp_ExpList = list<Exp.Exp>;
      type StringList = list<String>;
      Integer e_1,indx,e;
      Exp_ExpList inputs,outputs;
      Algorithm.Algorithm alg;
      StringList stmt_strs;
      String res;
      DAE.DAElist dae;
      DAELow.DAELow dlow;
      DAELow.EquationArray eqns;
      IntegerVector ass1,as2,ass2;
      IntegerList block_;
    case (dae,(dlow = DAELow.DAELOW(_,_,eqns,_,_,alg,_)),ass1,as2,(block_ = e :: _)) "assignments1  assignments2  block of equations"
      equation 
        true = all_same_algorithm(dlow, block_);
        e_1 = e - 1;
        DAELow.ALGORITHM(indx,inputs,outputs) = DAELow.equation_nth(eqns, e_1);
        alg = alg[indx + 1];
        (Codegen.CFUNCTION(_,_,_,_,_,_,stmt_strs,_),_) = Codegen.generate_algorithm(DAE.ALGORITHM(alg), 1);
        res = Util.string_delimit_list(stmt_strs, "\n"); then res;
    case (dae,dlow,ass1,ass2,block_)
      equation 
        print("#Solving of equation systems not implemented yet.\n"); then fail();
  end matchcontinue;
end build_block;

protected function all_same_algorithm "blocks"
  input DAELow.DAELow in_daelow;
  input IntegerList in_integerlist;
  output Boolean out_boolean;
protected 
  type IntegerList = list<Integer>;
algorithm 
  out_boolean:=
  matchcontinue (in_daelow,in_integerlist)
    local
      type Algorithm_AlgorithmArray = Algorithm.Algorithm[:];
      Integer e_1,indx,e;
      Boolean res;
      DAELow.DAELow dlow;
      DAELow.EquationArray eqns;
      Algorithm_AlgorithmArray alg;
      IntegerList block_;
    case ((dlow = DAELow.DAELOW(_,_,eqns,_,_,alg,_)),(block_ = e :: _)) "blocks"
      equation 
        e_1 = e - 1 "extract index of first algorithm and check that entire block
	 * has that index. ";
        DAELow.ALGORITHM(indx,_,_) = DAELow.equation_nth(eqns, e_1);
        res = all_same_algorithm2(dlow, block_, indx); then res;
    case (_,_) then false; 
  end matchcontinue;
end all_same_algorithm;

protected function all_same_algorithm2 "block 
  alg. index"
  input DAELow.DAELow in_daelow;
  input IntegerList in_integerlist;
  input Integer in_integer;
  output Boolean out_boolean;
protected 
  type IntegerList = list<Integer>;
algorithm 
  out_boolean:=
  matchcontinue (in_daelow,in_integerlist,in_integer)
    local
      type Algorithm_AlgorithmArray = Algorithm.Algorithm[:];
      Integer e_1,indx2,e,indx;
      Boolean b1;
      DAELow.DAELow dlow;
      DAELow.EquationArray eqns;
      Algorithm_AlgorithmArray alg;
      IntegerList es;
    case (_,{},_) then true; 
    case ((dlow = DAELow.DAELOW(_,_,eqns,_,_,alg,_)),e :: es,indx)
      equation 
        e_1 = e - 1;
        DAELow.ALGORITHM(indx2,_,_) = DAELow.equation_nth(eqns, e_1);
        (indx == indx2) = true;
        b1 = all_same_algorithm2(dlow, es, indx); then b1;
    case ((dlow = DAELow.DAELOW(_,_,eqns,_,_,alg,_)),e :: es,indx)
      equation 
        e_1 = e - 1;
        DAELow.ALGORITHM(indx2,_,_) = DAELow.equation_nth(eqns, e_1);
        (indx == indx2) = false; then false;
    case (_,_,_) then false; 
  end matchcontinue;
end all_same_algorithm2;

protected function build_equation "assignments1
  assignments2
  equation no"
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input Integer in_integer5;
  output String out_string;
protected 
  type IntegerVector = Integer[:];
algorithm 
  out_string:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_integer5)
    local
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type Exp_ExpList = list<Exp.Exp>;
      type StringList = list<String>;
      type Algorithm_AlgorithmArray = Algorithm.Algorithm[:];
      Integer e_1,v,v_1,e,indx;
      DAELow.VarKind kind;
      DAE.Flow flow_;
      DAELow.VariableArray vararr;
      IntegerVector ass2,ass1;
      Exp.Exp e1,e2,varexp,expr,simplify_exp,new_varexp;
      DAELow.Var va;
      Exp.ComponentRef cr;
      String origname,assigment_string,res;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.DAElist dae;
      DAELow.EquationArray eqns;
      Exp_ExpList inputs,outputs;
      Algorithm.Algorithm alg;
      StringList stmt_strs;
      Algorithm_AlgorithmArray algs;
    case (_,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,_,_,_,_,_),_,ass2,e) "assignments1 assignments2 equation no"
      equation 
        e_1 = e - 1 "filter discrete variables ";
        v = ass2[e_1 + 1];
        v_1 = v - 1;
        (DAELow.VAR(_,kind,_,_,_,_,_,_,_,_,_,_,_,flow_)) = DAELow.vararray_nth(vararr, v_1);
        DAELow.is_discrete(kind); then "";
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1 "Solving for non-states ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1;
        (va = DAELow.VAR(cr,kind,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_)) = DAELow.vararray_nth(vararr, v_1);
        true = DAELow.is_non_state(kind);
        varexp = Exp.CREF(cr,Exp.REAL());
        expr = Exp.solve(e1, e2, varexp);
        simplify_exp = Exp.simplify(expr);
        assigment_string = build_assignment(dae, cr, simplify_exp, origname); then assigment_string;
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1 "Solving the state s means solving for der(s) ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1 "v == variable no solved in this equation ";
        DAELow.VAR(cr,kind,_,_,_,_,_,_,indx,origname,_,dae_var_attr,comment,flow_) = DAELow.vararray_nth(vararr, v_1);
        new_varexp = Exp.CREF(cr,Exp.REAL());
        expr = Exp.solve(e1, e2, new_varexp);
        simplify_exp = Exp.simplify(expr);
        assigment_string = build_assignment(dae, cr, simplify_exp, origname); then assigment_string;
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1 "probably, solved failed in rule above. This means that we have 
	 a non-linear equation. ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1 "v==variable no solved in this equation ";
        DAELow.VAR(cr,_,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_) = DAELow.vararray_nth(vararr, v_1);
        varexp = Exp.CREF(cr,Exp.REAL());
        failure(_ = Exp.solve(e1, e2, varexp));
        print("nonlinear equation not implemented yet\n"); then fail();
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,algs,_),ass1,ass2,e)
      equation 
        e_1 = e - 1 "Algorithms  Each algorithm should only be genated once.";
        DAELow.ALGORITHM(indx,inputs,outputs) = DAELow.equation_nth(eqns, e_1);
        alg = algs[indx + 1];
        (Codegen.CFUNCTION(_,_,_,_,_,_,stmt_strs,_),_) = Codegen.generate_algorithm(DAE.ALGORITHM(alg), 1);
        res = Util.string_delimit_list(stmt_strs, "\n"); then res;
    case (_,_,_,_,_)
      equation 
        print("-build_equation failed\n"); then "//build_equation failed\n";
  end matchcontinue;
end build_equation;

protected function build_residual_blocks " relation build_residual_blocks
  This function generates code for blocks on residual form, i.e.
  g(\dot{x},x,y,t) = 0
 "
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input IntegerListList in_integerlistlist5;
  output StringList out_stringlist;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_integerlistlist5)
    local
      String eqn_str;
      StringList res;
      DAE.DAElist dae;
      DAELow.DAELow dlow;
      IntegerVector ass1,ass2;
      IntegerList block_;
      IntegerListList blocks;
      Integer eqn;
    case (_,_,_,_,{}) then {}; 
    case (dae,dlow,ass1,ass2,(block_ = _ :: _ :: _) :: blocks)
      equation 
        eqn_str = build_residual_system(dae, dlow, ass1, ass2, block_) "For system of equations ";
        res = build_residual_blocks(dae, dlow, ass1, ass2, blocks); then eqn_str :: res;
    case (dae,dlow,ass1,ass2,(block_ = {eqn}) :: blocks)
      equation 
        eqn_str = build_residual_equation(dae, dlow, ass1, ass2, eqn) "for single equations ";
        res = build_residual_blocks(dae, dlow, ass1, ass2, blocks); then eqn_str :: res;
    case (_,_,_,_,_)
      equation 
        print("-build_residual_blocks failed\n"); then fail();
  end matchcontinue;
end build_residual_blocks;

protected function build_residual_system "assignments1
  assignments2
  equation no. "
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input IntegerList in_integerlist5;
  output String out_string;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
algorithm 
  out_string:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_integerlist5)
    local
      String s1,s2,str;
      DAE.DAElist dae;
      DAELow.DAELow dlow;
      IntegerVector ass1,ass2;
      Integer eqn;
      IntegerList eqns;
    case (_,_,_,_,{}) then ""; 
    case (dae,dlow,ass1,ass2,eqn :: eqns)
      equation 
        s1 = build_residual_equation(dae, dlow, ass1, ass2, eqn);
        s2 = build_residual_system(dae, dlow, ass1, ass2, eqns);
        str = string_append(s1, s2); then str;
  end matchcontinue;
end build_residual_system;

protected function build_residual_equation "assignments1
  assignments2
  equation no. "
  input DAE.DAElist in_daelist1;
  input DAELow.DAELow in_daelow2;
  input IntegerVector in_integervector3;
  input IntegerVector in_integervector4;
  input Integer in_integer5;
  output String out_string;
protected 
  type IntegerVector = Integer[:];
algorithm 
  out_string:=
  matchcontinue (in_daelist1,in_daelow2,in_integervector3,in_integervector4,in_integer5)
    local
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Integer e_1,v,v_1,e,indx;
      DAELow.VarKind kind;
      DAE.Flow flow_;
      DAELow.VariableArray vararr;
      IntegerVector ass2,ass1;
      Exp.Exp e1,e2,varexp,expr,simplify_exp,exp;
      DAELow.Var va;
      Exp.ComponentRef cr,new_cr;
      String origname,assigment_string,indx_str,cr_str;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.DAElist dae;
      DAELow.EquationArray eqns;
    case (_,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,_,_,_,_,_),_,ass2,e) "assignments1 assignments2 equation no. "
      equation 
        e_1 = e - 1 "Do not output equations for discrete variables here ";
        v = ass2[e_1 + 1];
        v_1 = v - 1;
        (DAELow.VAR(_,kind,_,_,_,_,_,_,_,_,_,_,_,flow_)) = DAELow.vararray_nth(vararr, v_1);
        DAELow.is_discrete(kind); then "";
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1 "Solving for non-states ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1;
        (va = DAELow.VAR(cr,kind,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_)) = DAELow.vararray_nth(vararr, v_1);
        true = DAELow.is_non_state(kind);
        varexp = Exp.CREF(cr,Exp.REAL()) "print \"Solving for non-states\n\" &";
        expr = Exp.solve(e1, e2, varexp);
        simplify_exp = Exp.simplify(expr);
        assigment_string = build_assignment(dae, cr, simplify_exp, origname); then assigment_string;
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1 "Solving the state s, caluate residual form. ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1;
        DAELow.VAR(cr,kind,_,_,_,_,_,_,indx,origname,_,dae_var_attr,comment,flow_) = DAELow.vararray_nth(vararr, v_1);
        indx_str = int_string(indx);
        exp = Exp.BINARY(e1,Exp.SUB(Exp.REAL()),e2);
        simplify_exp = Exp.simplify(exp);
        cr_str = Util.string_append_list({"delta[",indx_str,"]"}) "Use array named 'delta' for residuals ";
        new_cr = Exp.CREF_IDENT(cr_str,{});
        assigment_string = build_assignment(dae, new_cr, simplify_exp, origname); then assigment_string;
    case (_,DAELow.DAELOW(_,_,eqns,_,_,_,_),_,_,e)
      local DAELow.WhenEquation e;
      equation 
        e_1 = e - 1 "when-equations are not part of the residual equations ";
        DAELow.WHEN_EQUATION(e) = DAELow.equation_nth(eqns, e_1); then "";
    case (dae,DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1;
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1 "v==variable no solved in this equation ";
        DAELow.VAR(cr,_,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_) = DAELow.vararray_nth(vararr, v_1);
        varexp = Exp.CREF(cr,Exp.REAL());
        failure(_ = Exp.solve(e1, e2, varexp));
        print("nonlinear equation not implemented yet\n"); then fail();
    case (_,_,_,_,_)
      equation 
        print("-build_residual_equation failed\n"); then "//build_residual_equation failed\n";
  end matchcontinue;
end build_residual_equation;

protected function build_assignment "varname 
  expression
  orig. name"
  input DAE.DAElist in_daelist;
  input Exp.ComponentRef in_componentref;
  input Exp.Exp in_exp;
  input String in_string;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_daelist,in_componentref,in_exp,in_string)
    local
      type DAE_ElementList = list<DAE.Element>;
      type Exp_ExpList = list<Exp.Exp>;
      String cr_str,exp_str,eqn_str,origname;
      DAE.DAElist dae;
      DAE_ElementList elements;
      Exp.ComponentRef cr;
      Exp.Exp exp;
      Absyn.Path path;
      Exp_ExpList args;
      Boolean tuple_,builtin;
    case ((dae = DAE.DAE(elements)),cr,(exp = Exp.CALL(path,args,(tuple_ = false),builtin)),origname) "varname  expression orig. name 
  rule  Exp.print_component_ref_str(cr) => cr_str &	
        print_exp_cpp_str(exp) => exp_str & 
	DAE.get_named_function(path, elements) => funcdae &
	DAE.get_output_vars funcdae => {outvar} &
        Util.string_append_list({\"  \",cr_str, \" = \", exp_str, \";\n\"}) => eqn_str 
	-------------------------------------------------------------------------
	build_assignment(dae as DAE.DAElist(elements),cr,exp as Exp.CALL(path,args,tuple as false,builtin),origname) => eqn_str
"
      equation 
        cr_str = Exp.print_component_ref_str(cr);
        exp_str = print_exp_cpp_str(exp);
        eqn_str = Util.string_append_list({"  ",cr_str," = ",exp_str,";\n"}); then eqn_str;
    case (dae,cr,(exp = Exp.CALL(path,args,(tuple_ = true),builtin)),origname)
      equation 
        print(
          "-simcodegen: build_assignment: Tuple return values from functions not implemented\n"); then fail();
    case (dae,cr,exp,origname)
      equation 
        cr_str = Exp.print_component_ref_str(cr);
        exp_str = print_exp_cpp_str(exp);
        eqn_str = Util.string_append_list({"  ",cr_str," = ",exp_str,";\n"}); then eqn_str;
    case (dae,cr,exp,origname)
      equation 
        print("-build_assignment failed\n"); then fail();
  end matchcontinue;
end build_assignment;

public function print_exp_cpp_str " relation: print_exp_cpp_str
 
  This function prints a complete expression on a C/C++ format.
 "
  input Exp.Exp e;
  output String s;
algorithm 
  s := print_exp2_str(e, 0);
end print_exp_cpp_str;

protected function lbinop_symbol
  input Exp.Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (Exp.AND()) then " && "; 
    case (Exp.OR()) then " || "; 
  end matchcontinue;
end lbinop_symbol;

protected function lunaryop_symbol
  input Exp.Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (Exp.NOT()) then " !"; 
  end matchcontinue;
end lunaryop_symbol;

protected function relop_symbol
  input Exp.Operator in_operator;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_operator)
    case (Exp.LESS(_)) then " < "; 
    case (Exp.LESSEQ(_)) then " <= "; 
    case (Exp.GREATER(_)) then " > "; 
    case (Exp.GREATEREQ(_)) then " >= "; 
    case (Exp.EQUAL(_)) then " = "; 
    case (Exp.NEQUAL(_)) then " <> "; 
  end matchcontinue;
end relop_symbol;

protected function print_exp2_str " relation: print_exp2_str
  Helper function to print_exp_str
 "
  input Exp.Exp in_exp;
  input Integer in_integer;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_exp,in_integer)
    local
      type Exp_ExpList = list<Exp.Exp>;
      type TupleExp_ExpBoolean = tuple<Exp.Exp,Boolean>;
      type TupleExp_ExpBooleanList = list<TupleExp_ExpBoolean>;
      type TupleExp_ExpBooleanListList = list<TupleExp_ExpBooleanList>;
      type Exp_ComponentRefList = list<Exp.ComponentRef>;
      String s,s_1,s_2,res,sym,s1,s2,s3,s4,s_3,res_1,ifstr,thenstr,elsestr,s_4,slast,fs,argstr,s5,s_5,res2,crstr,dimstr,str,expstr,iterstr,id;
      Integer x,pri2_1,pri2,pri3,pri1,ival,i;
      Real two_1,two,rval;
      Exp.ComponentRef c;
      Exp.Exp e1,e2,e21,e22,e,t,f,start,stop,step,cr,dim,exp,iterexp;
      Exp.Operator op;
      Exp.Type ty,ty2,REAL;
      Absyn.Path fcn;
      Exp_ExpList args,es;
      Boolean builtin;
      Exp_ComponentRefList cref_list;
    case (Exp.END(),_)
      equation 
        print("# equation contain undefined symbols"); then fail();
    case (Exp.ICONST(x),_)
      equation 
        s = int_string(x); then s;
    case (Exp.RCONST(x),_)
      local Real x;
      equation 
        s = real_string(x); then s;
    case (Exp.SCONST(s),_)
      equation 
        s_1 = string_append("\"", s);
        s_2 = string_append(s_1, "\""); then s_2;
    case (Exp.BCONST(false),_) then "false"; 
    case (Exp.BCONST(true),_) then "true"; 
    case (Exp.CREF(c,_),_)
      equation 
        res = Exp.print_component_ref_str(c); then res;
    case (Exp.BINARY(e1,(op = Exp.SUB(ty)),(e2 = Exp.BINARY(e21,Exp.SUB(ty2),e22))),pri1)
      equation 
        sym = Exp.binop_symbol(op);
        pri2_1 = Exp.binop_priority(op);
        pri2 = pri2_1 + 1;
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2) "binary minus have higher priority than itself ";
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, sym);
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (Exp.BINARY(e1,(op = Exp.POW(_)),Exp.ICONST(2)),pri1)
      equation 
        pri2 = Exp.binop_priority(op) "x^2 => x*x ";
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s4 = Exp.print_rightpar_str(pri1, pri2);
        res = Util.string_append_list({s1,s2,"*",s2,s4}); then res;
    case (Exp.BINARY(e1,(op = Exp.POW(_)),Exp.RCONST(two)),pri1)
      equation 
        two_1 = int_real(2) "x^2 => x*x ";
        (two ==. two) = true;
        pri2 = Exp.binop_priority(op);
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s4 = Exp.print_rightpar_str(pri1, pri2);
        res = Util.string_append_list({s1,s2,"*",s2,s4}); then res;
    case (Exp.BINARY(e1,(op = Exp.POW(_)),e2),pri1)
      equation 
        pri2 = Exp.binop_priority(op);
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append("pow(", s);
        s_2 = string_append(s_1, ",");
        s_3 = string_append(s_2, s3);
        res = string_append(s_3, ")");
        res_1 = string_append(res, s4); then res_1;
    case (Exp.BINARY(e1,op,e2),pri1)
      equation 
        sym = Exp.binop_symbol(op);
        pri2 = Exp.binop_priority(op);
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, sym);
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (Exp.UNARY(op,e),pri1)
      equation 
        sym = Exp.unaryop_symbol(op);
        pri2 = Exp.unaryop_priority(op);
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e, pri3);
        s3 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, sym);
        s_1 = string_append(s, s2);
        s_2 = string_append(s_1, s3); then s_2;
    case (Exp.LBINARY(e1,op,e2),pri1)
      equation 
        sym = lbinop_symbol(op);
        pri2 = Exp.lbinop_priority(op);
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, sym);
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (Exp.LUNARY(op,e),pri1)
      equation 
        sym = lunaryop_symbol(op);
        pri2 = Exp.lunaryop_priority(op);
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e, pri3);
        s3 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, sym);
        s_1 = string_append(s, s2);
        s_2 = string_append(s_1, s3); then s_2;
    case (Exp.RELATION(e1,op,e2),pri1)
      equation 
        sym = relop_symbol(op);
        pri2 = Exp.relop_priority(op);
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e1, pri3);
        s3 = print_exp2_str(e2, pri2);
        s4 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, sym);
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (Exp.IFEXP(c,t,f),_)
      local Exp.Exp c;
      equation 
        ifstr = print_exp2_str(c, 0);
        thenstr = print_exp2_str(t, 0);
        elsestr = print_exp2_str(f, 0);
        s = string_append("(( ", ifstr);
        s_1 = string_append(s, " ) ? ( ");
        s_2 = string_append(s_1, thenstr);
        s_3 = string_append(s_2, " ) : ( ");
        s_4 = string_append(s_3, elsestr);
        slast = string_append(s_4, " )) "); then slast;
    case (Exp.CALL(fcn,args,_,(builtin = true)),_)
      equation 
        fs = Absyn.path_string2(fcn, "_");
        argstr = Exp.print_list_str(args, print_exp_cpp_str, ",");
        s = Util.string_append_list({fs,"(",argstr,")"}); then s;
    case (Exp.CALL(fcn,args,_,(builtin = false)),_)
      equation 
        fs = Absyn.path_string2(fcn, "_");
        argstr = Exp.print_list_str(args, print_exp_cpp_str, ",");
        s = Util.string_append_list({fs,"(",argstr,").",fs,"_rettype_1"}); then s;
    case (Exp.ARRAY(_,_,es),_)
      equation 
        s = Exp.print_list_str(es, print_exp_cpp_str, ",");
        s_1 = string_append("{", s);
        s_2 = string_append(s_1, "}"); then s_2;
    case (Exp.TUPLE(es),_)
      equation 
        s = Exp.print_list_str(es, print_exp_cpp_str, ",");
        s_1 = string_append("(", s);
        s_2 = string_append(s_1, ")"); then s_2;
    case (Exp.MATRIX(_,_,es),_)
      local TupleExp_ExpBooleanListList es;
      equation 
        s = Exp.print_list_str(es, Exp.print_row_str, "},{");
        s_1 = string_append("{{", s);
        s_2 = string_append(s_1, "}}"); then s_2;
    case (Exp.RANGE(_,start,NONE,stop),pri1)
      equation 
        pri2 = 41;
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(start, pri3);
        s3 = print_exp2_str(stop, pri3);
        s4 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, ":");
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, s4); then s_3;
    case (Exp.RANGE(_,start,SOME(step),stop),pri1)
      equation 
        pri2 = 41;
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(start, pri3);
        s3 = print_exp2_str(step, pri3);
        s4 = print_exp2_str(stop, pri3);
        s5 = Exp.print_rightpar_str(pri1, pri2);
        s = string_append(s1, s2);
        s_1 = string_append(s, ":");
        s_2 = string_append(s_1, s3);
        s_3 = string_append(s_2, ":");
        s_4 = string_append(s_3, s4);
        s_5 = string_append(s_4, s5); then s_5;
    case (Exp.CAST(REAL,Exp.ICONST(ival)),_)
      equation 
        false = RTOpts.modelica_output();
        rval = int_real(ival);
        res = real_string(rval); then res;
    case (Exp.CAST(REAL,Exp.UNARY(Exp.UMINUS(_),Exp.ICONST(ival))),_)
      equation 
        false = RTOpts.modelica_output();
        rval = int_real(ival);
        res = real_string(rval);
        res2 = string_append("-", res); then res2;
    case (Exp.CAST(Exp.REAL(),e),_)
      equation 
        false = RTOpts.modelica_output();
        s = print_exp_cpp_str(e);
        s_1 = string_append("(float)(", s);
        s_2 = string_append(s_1, ")"); then s_2;
    case (Exp.CAST(Exp.REAL(),e),_)
      equation 
        true = RTOpts.modelica_output();
        s = print_exp_cpp_str(e); then s;
    case (Exp.ASUB(e,i),pri1)
      equation 
        pri2 = 51;
        cref_list = Exp.get_cref_from_exp(e);
        (s1,pri3) = Exp.print_leftpar_str(pri1, pri2);
        s2 = print_exp2_str(e, pri3);
        s3 = Exp.print_rightpar_str(pri1, pri2);
        s4 = int_string(i);
        s = string_append(s1, s2);
        s_1 = string_append(s, s3);
        s_2 = string_append(s_1, "[");
        s_3 = string_append(s_2, s4);
        s_4 = string_append(s_3, "]"); then s_4;
    case (Exp.SIZE(cr,SOME(dim)),_)
      equation 
        crstr = print_exp_cpp_str(cr);
        dimstr = print_exp_cpp_str(dim);
        str = Util.string_append_list({"size(",crstr,",",dimstr,")"}); then str;
    case (Exp.SIZE(cr,NONE),_)
      equation 
        crstr = print_exp_cpp_str(cr);
        str = Util.string_append_list({"size(",crstr,")"}); then str;
    case (Exp.REDUCTION(fcn,exp,id,iterexp),_)
      equation 
        fs = Absyn.path_string(fcn);
        expstr = print_exp_cpp_str(exp);
        iterstr = print_exp_cpp_str(iterexp);
        str = Util.string_append_list({"<reduction>",fs,"(",expstr," for ",id," in ",iterstr,")"}); then str;
    case (_,_) then "#UNKNOWN EXPRESSION# ----eee "; 
  end matchcontinue;
end print_exp2_str;

public function cref_modelica_str " relation: cref_modelica_str
  Converts Exp.ComponentRef, i.e. variables, to Modelica friendly variables.
  This means that dots are converted to underscore, etc.
 "
  input Exp.ComponentRef in_componentref;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_componentref)
    local
      String res_1,res_2,res_3,s,ns,ss;
      Exp.ComponentRef n;
    case (Exp.CREF_IDENT(s,_))
      equation 
        res_1 = Util.string_replace_char(s, #".", #"_");
        res_2 = Util.string_replace_char(res_1, #"[", #"_");
        res_3 = Util.string_replace_char(res_2, #"]", #"_") "& Util.string_append_list({\"_\",res,\"_\"}) => res'"; then res_3;
    case (Exp.CREF_QUAL(s,_,n))
      equation 
        ns = cref_modelica_str(n);
        ss = string_append(s, ns) " 	string_append(s,\"_\") => s1 &   s1"; then ss;
  end matchcontinue;
end cref_modelica_str;

protected function get_called_functions " relation: get_called_functions
 
  Goes through the DAELow structure, finds all function calls and returns them
  in a list. Removes duplicates.
 "
  input DAE.DAElist dae;
  input DAELow.DAELow dlow;
  output Absyn_PathList res;
protected 
  type Absyn_PathList = list<Absyn.Path>;
  type Exp_ExpList = list<Exp.Exp>;
  Exp_ExpList explist,fcallexps,fcallexps_1;
  Absyn_PathList calledfuncs,res1,res2,restemp;
algorithm 
  explist := DAELow.get_all_exps(dlow);
  fcallexps := Exp.get_function_calls_list(explist);
  fcallexps_1 := Util.list_select(fcallexps, is_not_builtin_call);
  calledfuncs := Util.list_map(fcallexps_1, get_call_path);
  res1 := remove_duplicate_paths(calledfuncs);
  res2 := get_called_functions_in_functions(res1, dae);
  restemp := list_append(res1, res2);
  res := remove_duplicate_paths(restemp);
end get_called_functions;

public function get_called_functions_in_functions " relation: get_called_functions_in_functions
 
  Goes through the given DAE, finds the given functions and collects the names
  of the functions called from within those functions
 "
  input Absyn_PathList paths;
  input DAE.DAElist dae;
  output Absyn_PathList res;
protected 
  type Absyn_PathList = list<Absyn.Path>;
  type Absyn_PathListList = list<Absyn_PathList>;
  Absyn_PathListList pathslist;
algorithm 
  pathslist := Util.list_map_1(paths, get_called_functions_in_function, dae);
  res := Util.list_flatten(pathslist);
end get_called_functions_in_functions;

public function get_called_functions_in_function " relation: get_called_functions_in_function 
 
  Goes through the given DAE, finds the given function and collects the names
  of the functions called from within those functions
 "
  input Absyn.Path in_path;
  input DAE.DAElist in_daelist;
  output Absyn_PathList out_absyn_pathlist;
protected 
  type Absyn_PathList = list<Absyn.Path>;
algorithm 
  out_absyn_pathlist:=
  matchcontinue (in_path,in_daelist)
    local
      type DAE_ElementList = list<DAE.Element>;
      type Exp_ExpList = list<Exp.Exp>;
      type StringList = list<String>;
      String pathstr,debugpathstr;
      Absyn.Path path;
      DAE_ElementList elements,funcelems;
      Exp_ExpList explist,fcallexps,fcallexps_1;
      Absyn_PathList calledfuncs,res1,res2,res;
      StringList debugpathstrs;
      DAE.DAElist dae;
    case (path,DAE.DAE(elements)) "Don't fail here, ceval will generate the function later "
      equation 
        {} = DAE.get_named_function(path, elements);
        Print.print_error_buf("Function not found: ");
        pathstr = Absyn.path_string(path);
        Print.print_error_buf(pathstr);
        Print.print_error_buf("\n");
        Debug.fprint("info", "Function not found: ");
        Debug.fprintln("info", pathstr); then {};
    case (path,(dae = DAE.DAE(elements)))
      equation 
        funcelems = DAE.get_named_function(path, elements);
        explist = DAE.get_all_exps(funcelems);
        fcallexps = Exp.get_function_calls_list(explist);
        fcallexps_1 = Util.list_select(fcallexps, is_not_builtin_call);
        calledfuncs = Util.list_map(fcallexps_1, get_call_path);
        res1 = remove_duplicate_paths(calledfuncs);
        Debug.fprint("info", "Found called functions: ");
        debugpathstrs = Util.list_map(res1, Absyn.path_string) "debug";
        debugpathstr = Util.string_delimit_list(debugpathstrs, ", ") "debug";
        Debug.fprintln("info", debugpathstr) "debug";
        res2 = get_called_functions_in_functions(res1, dae);
        res = list_append(res1, res2); then res;
  end matchcontinue;
end get_called_functions_in_function;

protected function is_not_builtin_call " relation: is_not_builtin_call
 
  return true if the given Exp.CALL is a call but not to a builtin function.
  checks the builtin flag in Exp.CALL
 "
  input Exp.Exp in_exp;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_exp)
    local
      Boolean res,builtin;
      Exp.Exp e;
    case Exp.CALL(_,_,_,builtin)
      equation 
        res = bool_not(builtin); then res;
    case e then false; 
  end matchcontinue;
end is_not_builtin_call;

protected function get_call_path
  input Exp.Exp in_exp;
  output Absyn.Path out_path;
algorithm 
  out_path:=
  matchcontinue (in_exp)
    local Absyn.Path path;
    case Exp.CALL(path,_,_,_) then path; 
  end matchcontinue;
end get_call_path;

protected function remove_duplicate_paths
  input Absyn_PathList in_absyn_pathlist;
  output Absyn_PathList out_absyn_pathlist;
protected 
  type Absyn_PathList = list<Absyn.Path>;
algorithm 
  out_absyn_pathlist:=
  matchcontinue (in_absyn_pathlist)
    local
      Absyn_PathList restwithoutfirst,recresult,rest;
      Absyn.Path first;
    case {} then {}; 
    case first :: rest
      equation 
        restwithoutfirst = remove_path_from_list(rest, first);
        recresult = remove_duplicate_paths(restwithoutfirst); then first :: recresult;
  end matchcontinue;
end remove_duplicate_paths;

protected function remove_path_from_list
  input Absyn_PathList in_absyn_pathlist;
  input Absyn.Path in_path;
  output Absyn_PathList out_absyn_pathlist;
protected 
  type Absyn_PathList = list<Absyn.Path>;
algorithm 
  out_absyn_pathlist:=
  matchcontinue (in_absyn_pathlist,in_path)
    local
      Absyn_PathList res,rest;
      Absyn.Path first,path;
    case ({},_) then {}; 
    case (first :: rest,path)
      equation 
        true = ModUtil.path_equal(first, path);
        res = remove_path_from_list(rest, path); then res;
    case (first :: rest,path)
      equation 
        false = ModUtil.path_equal(first, path);
        res = remove_path_from_list(rest, path); then first :: res;
  end matchcontinue;
end remove_path_from_list;
end SimCodegen;

