package TaskGraph "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 TaskGraph.rml
  module:      TaskGraph
  description: Building of task graphs from expressions, and equation systems.
 
  RCS: %W% %E%
  
 
  
  This module is used in the modpar part of OpenModelica for bulding task graphs
  from the BLT decomposition for automatic parallelization.
  The exported function 'build_taskgraph' takes the lowered form of the DAE defined in
  'DAELow' and two assignments vectors (which variable is solved in which equation) and
  the list of blocks given by the BLT decomposition.

  The module uses 'TaskGraphExt' for the task graph datastructure itself, which
  is implemented using Boost Graph Library in C++
 "

public import OpenModelica.Compiler.Exp;

public import OpenModelica.Compiler.DAELow;

protected import OpenModelica.Compiler.TaskGraphExt;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Absyn;

protected import OpenModelica.Compiler.DAE;

protected import OpenModelica.Compiler.Ceval;

protected import OpenModelica.Compiler.Values;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.VarTransform;

protected import OpenModelica.Compiler.SimCodegen;

public function build_taskgraph
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input IntegerListList in_integerlistlist4;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
algorithm 
  _:=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integerlistlist4)
    local
      type DAELow_VarList = list<DAELow.Var>;
      Integer starttask,endtask;
      DAELow_VarList vars,knvars;
      DAELow.DAELow dae;
      DAELow.VariableArray vararr,knvararr;
      IntegerVector ass1,ass2;
      IntegerListList blocks;
    case ((dae = DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),DAELow.VARIABLES(_,knvararr,_,_),_,_,_,_,_)),ass1,ass2,blocks)
      equation 
        print("starting buildtaskgraph\n");
        starttask = TaskGraphExt.new_task("start");
        endtask = TaskGraphExt.new_task("end");
        TaskGraphExt.set_execcost(starttask, 1.0);
        TaskGraphExt.set_execcost(starttask, 1.0);
        TaskGraphExt.register_startstop(starttask, endtask);
        vars = DAELow.vararray_list(vararr);
        knvars = DAELow.vararray_list(knvararr);
        add_variables(vars, starttask);
        add_variables(knvars, starttask);
        add_variables(
          {
          DAELow.VAR(Exp.CREF_IDENT("sim_time",{}),DAELow.VARIABLE(),
          DAE.INPUT(),DAE.REAL(),NONE,NONE,{},NONE,0,"time",{},NONE,NONE,
          DAE.NON_CONNECTOR())}, starttask);
        build_blocks(dae, ass1, ass2, blocks);
        print("done building taskgraph, about to build inits.\n");
        build_inits(dae);
        print("leaving build_taskgraph\n"); then ();
    case (_,_,_,_)
      equation 
        print("-build_taskgraph failed\n"); then fail();
  end matchcontinue;
end build_taskgraph;

protected function build_inits " relation: build_inits
  This function traverses the DAE and calls external functions to build 
  the initialization values for the DAE
  This is implemented in C++ as a set of vectors
"
  input DAELow.DAELow in_daelow;
algorithm 
  _:=
  matchcontinue (in_daelow)
    local
      type DAELow_VarList = list<DAELow.Var>;
      DAELow_VarList vars,kvars;
      DAELow.VariableArray vararr,kvararr;
    case (DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),DAELow.VARIABLES(_,kvararr,_,_),_,_,_,_,_))
      equation 
        vars = DAELow.vararray_list(vararr);
        kvars = DAELow.vararray_list(kvararr);
        build_inits2(vars);
        build_inits2(kvars); then ();
  end matchcontinue;
end build_inits;

protected function build_inits2
  input DAELow_VarList in_daelow_varlist;
protected 
  type DAELow_VarList = list<DAELow.Var>;
algorithm 
  _:=
  matchcontinue (in_daelow_varlist)
    local
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      String v,origname;
      Exp.Exp value;
      Integer indx;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow_VarList rest;
    case ({}) then (); 
    case (DAELow.VAR(_,DAELow.VARIABLE(),_,_,_,_,_,SOME(value),indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        v = Exp.print_exp_str(value);
        TaskGraphExt.add_initvar(indx, v, origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.VARIABLE(),_,_,_,_,_,NONE,indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        TaskGraphExt.add_initvar(indx, "0.0", origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.STATE(),_,_,_,_,_,SOME(value),indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        v = Exp.print_exp_str(value);
        TaskGraphExt.add_initstate(indx, v, origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.STATE(),_,_,_,_,_,NONE,indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        TaskGraphExt.add_initstate(indx, "0.0", origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.DUMMY_DER(),_,_,_,_,_,SOME(value),indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        v = Exp.print_exp_str(value);
        TaskGraphExt.add_initvar(indx, v, origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.DUMMY_DER(),_,_,_,_,_,NONE,indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        TaskGraphExt.add_initvar(indx, "0.0", origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.DUMMY_STATE(),_,_,_,_,_,SOME(value),indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        v = Exp.print_exp_str(value);
        TaskGraphExt.add_initvar(indx, v, origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.DUMMY_STATE(),_,_,_,_,_,NONE,indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        TaskGraphExt.add_initvar(indx, "0.0", origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.PARAM(),_,_,_,SOME(value),_,_,indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      local Values.Value value;
      equation 
        v = Values.val_string(value);
        TaskGraphExt.add_initparam(indx, v, origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.PARAM(),_,_,_,NONE,_,_,indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        TaskGraphExt.add_initparam(indx, "0.0", origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.CONST(),_,_,_,SOME(value),_,_,indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      local Values.Value value;
      equation 
        v = Values.val_string(value);
        TaskGraphExt.add_initparam(indx, v, origname);
        build_inits2(rest); then ();
    case (DAELow.VAR(_,DAELow.CONST(),_,_,_,NONE,_,_,indx,origname,_,dae_var_attr,comment,flow_) :: rest)
      equation 
        TaskGraphExt.add_initparam(indx, "0.0", origname);
        build_inits2(rest); then ();
  end matchcontinue;
end build_inits2;

protected function add_variables
  input DAELow_VarList in_daelow_varlist;
  input Integer in_integer;
protected 
  type DAELow_VarList = list<DAELow.Var>;
algorithm 
  _:=
  matchcontinue (in_daelow_varlist,in_integer)
    local
      Integer start;
      DAELow.Var v;
      DAELow_VarList vs;
    case ({},start) then (); 
    case (v :: vs,start)
      equation 
        add_variable(v, start);
        add_variables(vs, start); then ();
  end matchcontinue;
end add_variables;

protected function build_blocks
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input IntegerListList in_integerlistlist4;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type IntegerListList = list<IntegerList>;
algorithm 
  _:=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integerlistlist4)
    local
      DAELow.DAELow dae;
      IntegerVector ass1,ass2;
      IntegerList block_;
      IntegerListList blocks;
      Integer eqn;
    case (_,_,_,{}) then (); 
    case (dae,ass1,ass2,(block_ = _ :: _ :: _) :: blocks)
      equation 
        build_system(dae, ass1, ass2, block_) "For system of equations ";
        build_blocks(dae, ass1, ass2, blocks); then ();
    case (dae,ass1,ass2,(block_ = {eqn}) :: blocks)
      equation 
        build_equation(dae, ass1, ass2, eqn) "for single equations ";
        build_blocks(dae, ass1, ass2, blocks); then ();
    case (_,_,_,_)
      equation 
        print("-build_blocks failed\n"); then fail();
  end matchcontinue;
end build_blocks;

protected function is_non_state
  input DAELow.VarKind in_varkind;
algorithm 
  _:=
  matchcontinue (in_varkind)
    case (DAELow.VARIABLE()) then (); 
    case (DAELow.DUMMY_DER()) then (); 
    case (DAELow.DUMMY_STATE()) then (); 
  end matchcontinue;
end is_non_state;

protected function build_equation "Build task graph for a single equation."
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input Integer in_integer4;
protected 
  type IntegerVector = Integer[:];
algorithm 
  _:=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integer4)
    local
      type DAELow_VarList = list<DAELow.Var>;
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Integer e_1,v_1,e,indx;
      Exp.Exp e1,e2,varexp,expr;
      DAELow.Var v;
      DAELow_VarList varlst;
      Exp.ComponentRef cr,cr_1;
      DAELow.VarKind kind;
      String origname,indxs,id;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      DAELow.Variables vars;
      DAELow.EquationArray eqns;
      IntegerVector ass1,ass2;
    case (DAELow.DAELOW(vars,_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1 "Solving for non-states ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1 "v == variable no solved in this equation ";
        varlst = DAELow.var_list(vars);
        (v = DAELow.VAR(cr,kind,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_)) = list_nth(varlst, v_1);
        is_non_state(kind);
        varexp = Exp.CREF(cr,Exp.REAL()) "print \"Solving for non-states\n\" &";
        expr = Exp.solve(e1, e2, varexp);
        build_assignment(cr, expr, origname) "	Exp.print_exp_str e1 => e1s &
	Exp.print_exp_str e2 => e2s &
	print \"Equation \" & print e1s & print \" = \" & print e2s &
	print \" solved for \" & Exp.print_exp_str varexp => s &
	print s & print \" giving \" &
	Exp.print_exp_str expr => s2 & print s2 & print \"\n\" &"; then ();
    case (DAELow.DAELOW(vars,_,eqns,_,_,_,_),ass1,ass2,e)
      local Integer v;
      equation 
        e_1 = e - 1 "Solving the state s means solving for der(s) ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1 "v == variable no solved in this equation ";
        varlst = DAELow.var_list(vars);
        DAELow.VAR(cr,DAELow.STATE(),_,_,_,_,_,_,indx,origname,_,dae_var_attr,comment,flow_) = list_nth(varlst, v_1);
        indxs = int_string(indx) "	print \"solving for state\n\" &";
        id = Util.string_append_list({"xd[",indxs,"]"});
        cr_1 = Exp.CREF_IDENT(id,{});
        varexp = Exp.CREF(cr_1,Exp.REAL());
        expr = Exp.solve(e1, e2, varexp);
        build_assignment(cr_1, expr, origname) "	Exp.print_exp_str e1 => e1s &
	Exp.print_exp_str e2 => e2s &
	print \"Equation \" & print e1s & print \" = \" & print e2s &
	print \"solved for \" & Exp.print_exp_str varexp => s &
	print s & print \"giving \" &
	Exp.print_exp_str expr => s2 & print s2 & print \"\n\" &"; then ();
    case (DAELow.DAELOW(vars,_,eqns,_,_,_,_),ass1,ass2,e) " rule	int_sub(e,1) => e' &
	DAELow.equation_nth(eqns,e') => DAELow.EQUATION(e1,e2) &
	vector_nth(ass2,e') => v & (* v==variable no solved in this equation *)
	int_sub(v,1) => v' &
	DAELow.vararray_nth(vararr,v') => DAELow.VAR(cr,_,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow) &  
	let varexp = Exp.CREF(cr,Exp.REAL) &
	not Exp.solve(e1,e2,varexp) => _ &
	print \"nonlinear equation not implemented yet\n\" 
	--------------------------------
	build_equation(DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_),ass1,ass2,e) => fail
"
      local Integer v;
      equation 
        e_1 = e - 1 "state nonlinear ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1 "v == variable no solved in this equation ";
        varlst = DAELow.var_list(vars);
        DAELow.VAR(cr,DAELow.STATE(),_,_,_,_,_,_,indx,origname,_,dae_var_attr,_,flow_) = list_nth(varlst, v_1);
        indxs = int_string(indx);
        id = Util.string_append_list({"xd[",indxs,"]"});
        cr_1 = Exp.CREF_IDENT(id,{});
        varexp = Exp.CREF(cr_1,Exp.REAL());
        failure(_ = Exp.solve(e1, e2, varexp));
        build_nonlinear_equations({varexp}, {Exp.BINARY(e1,Exp.SUB(Exp.REAL()),e2)}); then ();
    case (DAELow.DAELOW(vars,_,eqns,_,_,_,_),ass1,ass2,e)
      equation 
        e_1 = e - 1 "Solving nonlinear for non-states ";
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1 "v == variable no solved in this equation ";
        varlst = DAELow.var_list(vars);
        (v = DAELow.VAR(cr,kind,_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_)) = list_nth(varlst, v_1);
        is_non_state(kind);
        varexp = Exp.CREF(cr,Exp.REAL()) "print \"Solving for non-states\n\" &";
        failure(expr = Exp.solve(e1, e2, varexp));
        build_nonlinear_equations({varexp}, {Exp.BINARY(e1,Exp.SUB(Exp.REAL()),e2)}); then ();
    case (_,_,_,_)
      equation 
        print("-build_equation failed\n"); then fail();
  end matchcontinue;
end build_equation;

protected function build_nonlinear_equations "residuals "
  input Exp_ExpList in_exp_explist1;
  input Exp_ExpList in_exp_explist2;
protected 
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  _:=
  matchcontinue (in_exp_explist1,in_exp_explist2)
    local
      type StringList = list<String>;
      Integer size,tid;
      String size_str,taskname;
      StringList varnames;
      Exp_ExpList vars,residuals;
    case (vars,residuals) "variables  residuals "
      equation 
        size = list_length(vars);
        size_str = int_string(size);
        taskname = build_residual_code(vars, residuals);
        tid = TaskGraphExt.new_task(taskname);
        TaskGraphExt.set_tasktype(tid, 3);
        build_nonlinear_equations2(tid, vars, residuals) "See TaskType in TaskGraph.hpp ";
        varnames = Util.list_map(vars, Exp.print_exp_str);
        store_multiple_results(varnames, tid); then ();
    case (vars,residuals)
      equation 
        print("build_nonlinear_equatins failed\n"); then fail();
  end matchcontinue;
end build_nonlinear_equations;

protected function build_residual_code "residuals "
  input Exp_ExpList in_exp_explist1;
  input Exp_ExpList in_exp_explist2;
  output String out_string;
protected 
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  out_string:=
  matchcontinue (in_exp_explist1,in_exp_explist2)
    local
      VarTransform.VariableReplacements repl;
      String res;
      Exp_ExpList vars,es;
    case (vars,es) "vars  residuals "
      equation 
        repl = make_residual_replacements(vars);
        res = build_residual_code2(es, 0, repl); then res;
    case (_,_)
      equation 
        print("build_residual_code failed\n"); then fail();
  end matchcontinue;
end build_residual_code;

protected function make_residual_replacements
  input Exp_ExpList expl;
  output VarTransform.VariableReplacements repl_1;
protected 
  type Exp_ExpList = list<Exp.Exp>;
  VarTransform.VariableReplacements repl,repl_1;
algorithm 
  repl := VarTransform.empty_replacements();
  repl_1 := make_residual_replacements2(repl, expl, 0);
end make_residual_replacements;

protected function make_residual_replacements2
  input VarTransform.VariableReplacements in_variablereplacements;
  input Exp_ExpList in_exp_explist;
  input Integer in_integer;
  output VarTransform.VariableReplacements out_variablereplacements;
protected 
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  out_variablereplacements:=
  matchcontinue (in_variablereplacements,in_exp_explist,in_integer)
    local
      VarTransform.VariableReplacements repl,repl_1,repl_2;
      String pstr,str;
      Integer pos_1,pos;
      Exp.ComponentRef cr;
      Exp_ExpList es;
    case (repl,{},_) then repl; 
    case (repl,Exp.CREF(cr,_) :: es,pos)
      equation 
        pstr = int_string(pos);
        str = Util.string_append_list({"xloc[",pstr,"]"});
        repl_1 = VarTransform.add_replacement(repl, cr, Exp.CREF_IDENT(str,{}));
        pos_1 = pos + 1;
        repl_2 = make_residual_replacements2(repl_1, es, pos_1); then repl_2;
  end matchcontinue;
end make_residual_replacements2;

protected function build_residual_code2
  input Exp_ExpList in_exp_explist;
  input Integer in_integer;
  input VarTransform.VariableReplacements in_variablereplacements;
  output String out_string;
protected 
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  out_string:=
  matchcontinue (in_exp_explist,in_integer,in_variablereplacements)
    local
      Exp.Exp e_1,e;
      String s1,s2,pstr,res;
      Integer pos_1,pos;
      Exp_ExpList es;
      VarTransform.VariableReplacements repl;
    case ({},_,_) then ""; 
    case (e :: es,pos,repl)
      equation 
        e_1 = VarTransform.replace_exp(e, repl);
        s1 = SimCodegen.print_exp_cpp_str(e_1);
        pos_1 = pos + 1;
        s2 = build_residual_code2(es, pos_1, repl);
        pstr = int_string(pos);
        res = Util.string_append_list({"res[",pstr,"]=",s1,";\n",s2}); then res;
    case (_,_,_)
      equation 
        print("build_residual_code2 failed\n"); then fail();
  end matchcontinue;
end build_residual_code2;

protected function store_multiple_results "task id"
  input StringList in_stringlist;
  input Integer in_integer;
protected 
  type StringList = list<String>;
algorithm 
  _:=
  matchcontinue (in_stringlist,in_integer)
    local
      String result_str;
      StringList varnames;
      Integer tid;
    case (varnames,tid) "var names task id"
      equation 
        result_str = Util.string_delimit_list(varnames, ";");
        TaskGraphExt.store_result(result_str, tid, true, result_str); then ();
    case (_,_)
      equation 
        print("store_multiple_results failed\n"); then fail();
  end matchcontinue;
end store_multiple_results;

protected function build_nonlinear_equations2 "vars 
  residuals "
  input Integer in_integer1;
  input Exp_ExpList in_exp_explist2;
  input Exp_ExpList in_exp_explist3;
protected 
  type Exp_ExpList = list<Exp.Exp>;
algorithm 
  _:=
  matchcontinue (in_integer1,in_exp_explist2,in_exp_explist3)
    local
      type Exp_ComponentRefList = list<Exp.ComponentRef>;
      type Exp_ComponentRefListList = list<Exp_ComponentRefList>;
      Integer tid;
      Exp_ComponentRefList vars1,vars2,vars1_1,vars;
      Exp_ComponentRefListList vars_1;
      Exp.Exp res,e;
      Exp_ExpList residuals;
      String es;
    case (tid,_,{}) then (); 
    case (tid,vars,res :: residuals)
      equation 
        vars1 = Exp.get_cref_from_exp(res) "Collect all variables and construct
	 a string for the residual, that can be directly used in codegen.";
        vars_1 = Util.list_map(vars, Exp.get_cref_from_exp);
        vars2 = Util.list_flatten(vars_1);
        vars1_1 = Util.list_union_p(vars1, vars2, Exp.cref_equal) "No duplicate elements ";
        vars = Util.list_setdifference_p(vars1_1, vars2, Exp.cref_equal);
        add_edges_from_vars(vars, tid, 0); then ();
    case (_,_,e :: _)
      equation 
        print("build_nonlinear_equations2 failed\n");
        es = Exp.print_exp_str(e);
        print("first residual :");
        print(es);
        print("\n"); then fail();
  end matchcontinue;
end build_nonlinear_equations2;

protected function add_edges_from_vars "task 
  priority"
  input Exp_ComponentRefList in_exp_componentreflist1;
  input Integer in_integer2;
  input Integer in_integer3;
protected 
  type Exp_ComponentRefList = list<Exp.ComponentRef>;
algorithm 
  _:=
  matchcontinue (in_exp_componentreflist1,in_integer2,in_integer3)
    local
      String v_str;
      Integer predt,prio_1,tid,prio;
      Exp.ComponentRef v;
      Exp_ComponentRefList vs;
    case ({},_,_) then (); 
    case (v :: vs,tid,prio)
      equation 
        v_str = Exp.cref_str(v);
        predt = TaskGraphExt.get_task(v_str);
        TaskGraphExt.add_edge(predt, tid, v_str, prio);
        prio_1 = prio + 1;
        add_edges_from_vars(vs, tid, prio_1); then ();
    case (v :: vs,_,_)
      equation 
        v_str = Exp.cref_str(v);
        failure(_ = TaskGraphExt.get_task(v_str));
        print("task ");
        print(v_str);
        print(" not found\n"); then fail();
    case (_,_,_)
      equation 
        print("add_edges_from_vars failed\n"); then fail();
  end matchcontinue;
end add_edges_from_vars;

protected function build_system "Build task graph for a system of equations "
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input IntegerList in_integerlist4;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
algorithm 
  _:=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integerlist4)
    local
      type StringList = list<String>;
      Integer tid;
      StringList predtasks;
      IntegerList predtaskids,system;
      DAELow.DAELow dae;
      IntegerVector ass1,ass2;
    case (dae,ass1,ass2,system)
      equation 
        print("build system\n");
        tid = TaskGraphExt.new_task("equation system");
        predtasks = build_system2(dae, ass1, ass2, system, tid);
        predtaskids = Util.list_map(predtasks, TaskGraphExt.get_task);
        add_predecessors(tid, predtaskids, predtasks, 0); then ();
    case (_,_,_,_)
      equation 
        print("build_system failed\n"); then fail();
  end matchcontinue;
end build_system;

protected function build_system2
  input DAELow.DAELow in_daelow1;
  input IntegerVector in_integervector2;
  input IntegerVector in_integervector3;
  input IntegerList in_integerlist4;
  input Integer in_integer5;
  output StringList out_stringlist;
protected 
  type IntegerVector = Integer[:];
  type IntegerList = list<Integer>;
  type StringList = list<String>;
algorithm 
  out_stringlist:=
  matchcontinue (in_daelow1,in_integervector2,in_integervector3,in_integerlist4,in_integer5)
    local
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type Exp_ComponentRefList = list<Exp.ComponentRef>;
      DAELow.DAELow dae;
      IntegerVector ass1,ass2;
      Integer tid,e_1,v_1,e;
      Exp.Exp e1,e2;
      DAELow.Var v;
      Exp.ComponentRef cr;
      String origname,crstr;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      Exp_ComponentRefList cr1,cr2,crs,crs_1;
      StringList crs_2,crs2,res;
      DAELow.VariableArray vararr;
      DAELow.EquationArray eqns;
      IntegerList rest;
    case (dae,ass1,ass2,{},tid) then {}; 
    case ((dae = DAELow.DAELOW(DAELow.VARIABLES(_,vararr,_,_),_,eqns,_,_,_,_)),ass1,ass2,e :: rest,tid)
      equation 
        e_1 = e - 1;
        DAELow.EQUATION(e1,e2) = DAELow.equation_nth(eqns, e_1);
        v = ass2[e_1 + 1];
        v_1 = v - 1 "v == variable no solved in this equation ";
        (v = DAELow.VAR(cr,DAELow.VARIABLE(),_,_,_,_,_,_,_,origname,_,dae_var_attr,comment,flow_)) = DAELow.vararray_nth(vararr, v_1);
        cr1 = Exp.get_cref_from_exp(e1);
        cr2 = Exp.get_cref_from_exp(e2);
        crs = list_append(cr1, cr2);
        crs_1 = Util.list_deletemember(crs, cr);
        crs_2 = Util.list_map(crs_1, Exp.cref_str);
        crstr = Exp.cref_str(cr);
        TaskGraphExt.store_result(crstr, tid, true, origname);
        crs2 = build_system2(dae, ass1, ass2, rest, tid);
        res = Util.list_union(crs_2, crs2); then res;
    case (_,_,_,_,_)
      equation 
        print("build_system2 failed\n"); then fail();
  end matchcontinue;
end build_system2;

protected function add_variable
  input DAELow.Var in_var;
  input Integer in_integer;
algorithm 
  _:=
  matchcontinue (in_var,in_integer)
    local
      type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      String cfs,name;
      Exp.ComponentRef cf;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      DAE.Flow flow_;
      Integer start;
    case (DAELow.VAR(cf,_,_,_,_,_,_,_,_,name,_,dae_var_attr,comment,flow_),start)
      equation 
        cfs = Exp.cref_str(cf);
        TaskGraphExt.store_result(cfs, start, false, name) "print \"adding variable \" & print cfs & print \"\n\" &"; then ();
  end matchcontinue;
end add_variable;

protected function build_assignment "expression
  orig. name"
  input Exp.ComponentRef in_componentref;
  input Exp.Exp in_exp;
  input String in_string;
algorithm 
  _:=
  matchcontinue (in_componentref,in_exp,in_string)
    local
      Integer task,tid;
      String str,cr2s,crs,origname;
      Exp.ComponentRef cr,cr2;
      Exp.Exp exp;
      Exp.Type tp;
    case (cr,(exp = Exp.CREF(cr2,tp)),origname) "varname  expression orig. name"
      equation 
        (task,str) = build_expression(exp) "special rule for equation a:=b ";
        tid = TaskGraphExt.new_task("copy");
        cr2s = Exp.cref_str(cr2);
        TaskGraphExt.add_edge(task, tid, cr2s, 0);
        crs = Exp.cref_str(cr);
        TaskGraphExt.store_result(crs, tid, true, origname);
        TaskGraphExt.set_tasktype(tid, 6) "See TaskType in TaskGraph.hpp "; then ();
    case (cr,exp,origname)
      equation 
        (task,str) = build_expression(exp);
        crs = Exp.cref_str(cr);
        TaskGraphExt.store_result(crs, task, true, origname); then ();
    case (cr,exp,origname)
      equation 
        print("-build_assignment failed\n"); then fail();
  end matchcontinue;
end build_assignment;

protected function build_expression "Builds the task graph for the expression and
 returns the task no that calculates the result of the expr "
  input Exp.Exp in_exp;
  output Integer out_integer;
  output String out_string;
algorithm 
  (out_integer,out_string):=
  matchcontinue (in_exp)
    local
      type IntegerList = list<Integer>;
      type StringList = list<String>;
      type Exp_ExpList = list<Exp.Exp>;
      String is,rs,crs,s1,istr,ts,s2,ops,s3,funcstr,s,es;
      Integer tid,i,t1,ival,t,t2,t3,numargs;
      Real r,rval;
      Exp.ComponentRef cr;
      Exp.Exp e1,e2,e3,e;
      Exp.Operator op,relop;
      IntegerList tasks;
      StringList strs;
      Absyn.Path func;
      Exp_ExpList expl;
    case (Exp.ICONST(i))
      equation 
        is = int_string(i);
        tid = TaskGraphExt.new_task(is) "&
	TaskGraphExt.get_starttask() => st &
	TaskGraphExt.add_edge(st,tid,\"\") &
	TaskGraphExt.set_commcost(st,tid,0)"; then (tid,"");
    case (Exp.RCONST(r))
      equation 
        rs = real_string(r);
        tid = TaskGraphExt.new_task(rs) "&
	TaskGraphExt.get_starttask() => st &
	TaskGraphExt.add_edge(st,tid,\"\") &
	TaskGraphExt.set_commcost(st,tid,0) "; then (tid,"");
    case (Exp.CREF(cr,_))
      equation 
        crs = Exp.cref_str(cr) "for state variables and alg. variables";
        tid = TaskGraphExt.get_task(crs); then (tid,crs);
    case (Exp.CREF(Exp.CREF_IDENT("time",_),_))
      equation 
        tid = TaskGraphExt.get_task("sim_time") "for state variables and alg. variables"; then (tid,"sim_time");
    case (Exp.CREF(cr,_))
      equation 
        crs = Exp.cref_str(cr) "for constants and parameters, no data to send from proc0 ";
        tid = TaskGraphExt.new_task(crs); then (tid,crs);
    case (Exp.BINARY(e1,Exp.POW(_),Exp.RCONST(rval)))
      equation 
        (t1,s1) = build_expression(e1) "special case for pow ";
        ival = real_int(rval);
        istr = int_string(ival);
        ts = Util.string_append_list({"pow(%s,",istr,")"});
        t = TaskGraphExt.new_task(ts);
        TaskGraphExt.add_edge(t1, t, s1, 0); then (t,"");
    case (Exp.BINARY(e1,op,e2))
      equation 
        (t1,s1) = build_expression(e1);
        (t2,s2) = build_expression(e2);
        ops = Exp.binop_symbol_1(op);
        ts = Util.string_append_list({"%s",ops,"%s"});
        t = TaskGraphExt.new_task(ts);
        TaskGraphExt.add_edge(t1, t, s1, 0);
        TaskGraphExt.add_edge(t2, t, s2, 1); then (t,"");
    case (Exp.LBINARY(e1,op,e2))
      equation 
        (t1,s1) = build_expression(e1);
        (t2,s2) = build_expression(e2);
        ops = Exp.binop_symbol_1(op);
        ts = Util.string_append_list({"%s",ops,"%s"});
        t = TaskGraphExt.new_task(ts);
        TaskGraphExt.add_edge(t1, t, s1, 0);
        TaskGraphExt.add_edge(t2, t, s2, 1); then (t,"");
    case (Exp.UNARY(op,e1))
      equation 
        (t1,s1) = build_expression(e1);
        ops = Exp.unaryop_symbol(op);
        ts = Util.string_append_list({ops,"%s"});
        t = TaskGraphExt.new_task(ts);
        TaskGraphExt.add_edge(t1, t, s1, 0); then (t,"");
    case (Exp.LUNARY(op,e1))
      equation 
        (t1,s1) = build_expression(e1);
        ops = Exp.lunaryop_symbol(op);
        ts = Util.string_append_list({ops,"%s"});
        t = TaskGraphExt.new_task(ts);
        TaskGraphExt.add_edge(t1, t, s1, 0); then (t,"");
    case (Exp.RELATION(e1,relop,e2))
      equation 
        (t1,s1) = build_expression(e1);
        (t2,s2) = build_expression(e2);
        ops = Exp.relop_symbol(relop);
        ts = Util.string_append_list({"%s",ops,"%s"});
        t = TaskGraphExt.new_task(ts);
        TaskGraphExt.add_edge(t1, t, s1, 0);
        TaskGraphExt.add_edge(t2, t, s2, 1); then (t,"");
    case (Exp.IFEXP(e1,e2,e3))
      equation 
        (t1,s1) = build_expression(e1);
        (t2,s2) = build_expression(e2);
        (t3,s3) = build_expression(e3);
        ts = Util.string_append_list({"%s ? %s : %s"});
        t = TaskGraphExt.new_task(ts);
        TaskGraphExt.add_edge(t1, t, s1, 0);
        TaskGraphExt.add_edge(t2, t, s2, 1);
        TaskGraphExt.add_edge(t3, t, s3, 2); then (t,"");
    case (Exp.CALL(func,expl,_,_))
      equation 
        funcstr = Absyn.path_string(func);
        numargs = list_length(expl);
        ts = build_call_str(funcstr, numargs);
        (tasks,strs) = Util.list_map__2(expl, build_expression);
        t = TaskGraphExt.new_task(ts);
        add_predecessors(t, tasks, strs, 0); then (t,"");
    case (Exp.ARRAY(_,_,_))
      equation 
        print("build_expression(ARRAY) not impl. yet\n"); then fail();
    case (Exp.ARRAY(_,_,_))
      equation 
        print("build_expression(MATRIX) not impl. yet\n"); then fail();
    case (Exp.RANGE(_,_,_,_))
      equation 
        print("build_expression(RANGE) not impl. yet\n"); then fail();
    case (Exp.TUPLE(_))
      equation 
        print("build_expression(TUPLE) not impl. yet\n"); then fail();
    case (Exp.CAST(t,e))
      equation 
        (t,s) = build_expression(e); then (t,s);
    case (Exp.ASUB(_,_))
      equation 
        print("build_expression(ASUB) not impl. yet\n"); then fail();
    case (Exp.SIZE(_,_))
      equation 
        print("build_expression(SIZE) not impl. yet\n"); then fail();
    case (Exp.CODE(_,_))
      equation 
        print("build_expression(CODE) not impl. yet\n"); then fail();
    case (Exp.REDUCTION(_,_,_,_))
      equation 
        print("build_expression(REDUCTION) not impl. yet\n"); then fail();
    case (Exp.END())
      equation 
        print("build_expression(END) not impl. yet\n"); then fail();
    case (e)
      equation 
        print("-build_expression failed\n Exp = ");
        es = Exp.print_exp_str(e);
        print(es);
        print("\n"); then fail();
  end matchcontinue;
end build_expression;

protected function build_call_str
  input String str;
  input Integer n;
  output String res;
protected 
  type StringList = list<String>;
  StringList ns;
  String ns_1;
algorithm 
  ns := Util.list_fill("%s", n);
  ns_1 := Util.string_delimit_list(ns, ", ");
  res := Util.string_append_list({str,"(",ns_1,")"});
end build_call_str;

protected function add_predecessors
  input Integer in_integer1;
  input IntegerList in_integerlist2;
  input StringList in_stringlist3;
  input Integer in_integer4;
protected 
  type IntegerList = list<Integer>;
  type StringList = list<String>;
algorithm 
  _:=
  matchcontinue (in_integer1,in_integerlist2,in_stringlist3,in_integer4)
    local
      Integer prio_1,t,t1,prio;
      IntegerList ts;
      String s;
      StringList strs;
    case (_,{},{},_) then (); 
    case (t,t1 :: ts,s :: strs,prio)
      equation 
        TaskGraphExt.add_edge(t1, t, s, prio);
        prio_1 = prio + 1;
        add_predecessors(t, ts, strs, prio_1); then ();
  end matchcontinue;
end add_predecessors;
end TaskGraph;

