package Inst "
    Copyright (c) PELAB, Dept. Computer and Information Science, Linkoping University,
    Sweden, in the following denoted PELAB/LIU.

   
    This file is part of OpenModelica, in the following denoted OSM.


    The OSM License, Version 1.1, March 4, 2005.


    1 Preface

    The aim of this license is to lay down the conditions enabling you to use, 
    modify and circulate OSM. However, PELAB/LIU remain the authors of 
    OSM and so retain property rights and the use of all ancillary rights.

    2 Definitions

    OSM is defined as all successive versions of the OSM software and their 
    documentation that have been developed by PELAB/LIU and including
    accepted contributions from other contributors according to this license.

    OSM DERIVED SOFTWARE is defined as all or part of OSM that you have 
    modified and/or translated and/or adapted.


    3 Dual License

    OSM is made available under the OSM licensing scheme, which is a dual
    licensing scheme with two options, a) and b):

    a) OSM OPEN SOURCE LICENSE:

    If you wish to write Open Source software you can use the Open Source version 
    of OSM, released under the OSM license which include GPL as its open source 
    licensing option. If you use the OSM Open Source version you must release your 
    Application using OSM including this Application's source code under the GPL as well.

    This OSM license text, and Copyright (c) PELAB/Linkoping University, must
    be present in your copy of OSM and in OSM DERIVED SOFTWARE.

    You should have received a copy of the GPL - GNU General Public License
    along with OpenModelica; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


    b) OSM COMMERCIAL LICENSE

    If you are using OSM commercially - that is, for commercial usage or
    for creating proprietary software for sale or use in a commercial setting
    - you must purchase a commercial license of OSM from PELAB/L
  
  file:	 Inst.rml
  module:      Inst
  description: Model instantiation
 
  RCS: $Id: Inst.rml 1837 2005-06-28 10:01:44Z petar $
 
  This module is responsible for instantiation of Modelica
   models. The instantation is the process of instantiating model
  components, flattening inheritance and generating equations from
  connect statements.
  The instantiation process takes Modelica AST as defined in SCode
  and produces variables and equations and algorithms, etc. as
  defined in DAE.
  
  This module uses 'Lookup' to lookup classes and variables from the
  environment defined in 'Env'. It uses 'Connect' for generating equations from
  connect statements. The type system defined in 'Types' is used for
  variable instantiation and type . 'Mod' is used for modifiers and
  merging of modifiers. 
  
 
   debug flags: inst "

public import OpenModelica.Compiler.ClassInf;

public import OpenModelica.Compiler.Connect;

public import OpenModelica.Compiler.DAE;

public import OpenModelica.Compiler.Env;

public import OpenModelica.Compiler.Exp;

public import OpenModelica.Compiler.SCode;

public import OpenModelica.Compiler.Mod;

public import OpenModelica.Compiler.Prefix;

public import OpenModelica.Compiler.Types;

public import OpenModelica.Compiler.Absyn;

public 
uniontype CallingScope "Calling scope is used to determine when unconnected flow variables 
    should be set to zero. "
  record TOP_CALL
  end TOP_CALL;
  record INNER_CALL
  end INNER_CALL;
end CallingScope;

public 
type InstDims = list<Exp.Subscript>;

public 
uniontype Initial
  record INITIAL
  end INITIAL;
  record NON_INITIAL
  end NON_INITIAL;
end Initial;

public 
type Exp_ExpOption = Option<Exp.Exp>;

uniontype DimExp
  record DIMINT
    Integer integer;
  end DIMINT;
  record DIMEXP
    Exp.Subscript subscript;
    Exp_ExpOption exp_expoption;
  end DIMEXP;
end DimExp;

protected import OpenModelica.Compiler.Debug;

protected import OpenModelica.Compiler.Interactive;

protected import OpenModelica.Compiler.Util;

protected import OpenModelica.Compiler.Algorithm;

protected import OpenModelica.Compiler.Builtin;

protected import OpenModelica.Compiler.Dump;

protected import OpenModelica.Compiler.Lookup;

protected import OpenModelica.Compiler.ModUtil;

protected import OpenModelica.Compiler.RTOpts;

protected import OpenModelica.Compiler.Static;

protected import OpenModelica.Compiler.Values;

protected import OpenModelica.Compiler.Print;

protected import OpenModelica.Compiler.System;

protected import OpenModelica.Compiler.Ceval;

protected constant String for_scope_name="$for loop scope$";

protected 
type Prefix = Prefix.Prefix "
  These type aliases are introduced to make the code a little more
  readable.
 ";

protected 
type Mod = Types.Mod;

protected 
type Ident = Exp.Ident;

protected 
type Env = Env.Env;

protected function print_dims
  input DimExpList in_dimexplist;
protected 
  type DimExpList = list<DimExp>;
algorithm 
  _:=
  matchcontinue (in_dimexplist)
    case (x :: xs)
      equation 
        print_dim({SOME(x)});
        print_dims(xs); then ();
    case ({}) then (); 
  end matchcontinue;
end print_dims;

public function new_ident " relation: new_ident
 
  This function creates a new, unique identifer.  The same name is
  never returned twice.
 "
  output Exp.ComponentRef out_componentref;
protected 
  Integer i;
  String is,s;
algorithm 
  i := tick();
  is := int_string(i);
  s := string_append("__TMP__", is);
  out_componentref := Exp.CREF_IDENT(s,{});
end new_ident;

protected function select " relation: select
 
  This utility function selects one of two objects depending on a
  boolean variable.
 "
  input Boolean in_boolean1;
  input Type_a in_type_a2;
  input Type_a in_type_a3;
  output Type_a out_type_a;
protected 
  replaceable type Type_a;
algorithm 
  out_type_a:=
  matchcontinue (in_boolean1,in_type_a2,in_type_a3)
    local Type_a x;
    case (true,x,_) then x; 
    case (false,_,x) then x; 
  end matchcontinue;
end select;

protected function is_not_function
  input SCode.Class cls;
  output Boolean res;
algorithm 
  res := SCode.is_function(cls);
  res := bool_not(res);
end is_not_function;

public function instantiate " relation: instantiate
 
  To instantiate a Modelica program, an initial environment is
  built, containing the predefined types. Then the program is
  instantiated by the function `inst_program'
 "
  input SCode.Program in_program;
  output DAE.DAElist out_daelist;
algorithm 
  out_daelist:=
  matchcontinue (in_program)
    case (p)
      equation 
        Debug.fprintln("insttr", "instantiate");
        pnofunc = Util.list_select(p, is_not_function);
        pfunc = Util.list_select(p, SCode.is_function);
        env = Builtin.initial_env();
        Debug.fprintln("insttr", "Instantiating functions");
        pfuncnames = Util.list_map(pfunc, SCode.class_name);
        str1 = Util.string_delimit_list(pfuncnames, ", ");
        Debug.fprint("insttr", "Instantiating functions: ");
        Debug.fprintln("insttr", str1);
        envimpl = Env.extend_frame_classes(env, pfunc);
        (lfunc,envimpl_1) = inst_program_implicit(envimpl, pfunc);
        Debug.fprint("insttr", "Instantiating other classes: ");
        pnofuncnames = Util.list_map(pnofunc, SCode.class_name);
        str2 = Util.string_delimit_list(pnofuncnames, ", ");
        Debug.fprintln("insttr", str2);
        lnofunc = inst_program(envimpl_1, pnofunc);
        l = list_append(lfunc, lnofunc); then DAE.DAE(l);
    case _
      equation 
        Debug.fprintln("failtrace", "instantiate failed"); then fail();
  end matchcontinue;
end instantiate;

public function instantiate_implicit " Implicit instantiation of a program can be used for e.g. code generation of functions, since a function must be implicitly instantiated in order to generate code from it. "
  input SCode.Program in_program;
  output DAE.DAElist out_daelist;
algorithm 
  out_daelist:=
  matchcontinue (in_program)
    case (p)
      equation 
        Debug.fprintln("insttr", "instantiate_implicit");
        env = Builtin.initial_env();
        env_1 = Env.extend_frame_classes(env, p);
        (l,_) = inst_program_implicit(env_1, p); then DAE.DAE(l);
    case _
      equation 
        Debug.fprintln("failtrace", "instantiate_implicit failed"); then fail();
  end matchcontinue;
end instantiate_implicit;

public function instantiate_class " relation: instantiate_class
 
 
  To enable interactive instantiation, an arbitrary class in the program needs to be
  possible to instantiate. This function performs the same action as `inst_program', 
   but given a specific class to instantiate.
   First all the class
  definitions is added to the environment without modifications, and
  then the specified class is instantiated in the function `inst_class'
 "
  input SCode.Program in_program;
  input SCode.Path in_path;
  output DAE.DAElist out_daelist;
  output Env.Env out_env;
algorithm 
  (out_daelist,out_env):=
  matchcontinue (in_program,in_path)
    case ({},cr)
      equation 
        Print.print_error_buf("# Empty program\n"); then fail();
    case ((cdecls = _ :: _),(path = Absyn.IDENT(name2)))
      equation 
        env = Builtin.initial_env();
        (env_1,dae1) = inst_class_decls(env, cdecls, path);
        (dae,env_2) = inst_class_in_program(env_1, cdecls, path); then (DAE.DAE({DAE.COMP(name2,DAE.DAE(dae))}),env_2);
    case ((cdecls = _ :: _),(path = Absyn.QUALIFIED(name,_)))
      equation 
        env = Builtin.initial_env();
        (env_1,_) = inst_class_decls(env, cdecls, path);
        (cdef = SCode.CLASS(n,_,_,_,_),env_2) = Lookup.lookup_class(env_1, path, true);
        (dae,env_2,_,_,_) = inst_class(env_2, Types.NOMOD, Prefix.NOPRE, {}, cdef, {}, false, 
          TOP_CALL) "impl";
        pathstr = Absyn.path_string(path); then (DAE.DAE({DAE.COMP(pathstr,DAE.DAE(dae))}),env_2);
    case (_,_) then fail(); 
  end matchcontinue;
end instantiate_class;

protected function inst_class_in_program
  input Env in_env;
  input SCode.Program in_program;
  input SCode.Path in_path;
  output DAE_ElementList out_dae_elementlist;
  output Env.Env out_env;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env):=
  matchcontinue (in_env,in_program,in_path)
    case (env,(c = SCode.CLASS(name,_,_,_,_)) :: cs,Absyn.IDENT(name2))
      equation 
        equality(name = name2);
        (dae,env_1,_,_,_) = inst_class(env, Types.NOMOD, Prefix.NOPRE, {}, c, {}, false, TOP_CALL) "impl"; then (dae,env_1);
    case (env,(c = SCode.CLASS(name,_,_,_,_)) :: cs,(path = Absyn.IDENT(name2)))
      equation 
        failure(equality(name = name2));
        (dae,env) = inst_class_in_program(env, cs, path); then (dae,env);
    case (env,{},_) then ({},env); 
    case (env,_,_) then fail(); 
  end matchcontinue;
end inst_class_in_program;

protected function inst_class_decls " relation: inst_class_decls
 
  This function instantiated class definitions, i.e. adding the class definitions to the
  environment. See also partial_inst_class_in.
 "
  input Env in_env;
  input SCode.Program in_program;
  input SCode.Path in_path;
  output Env.Env out_env;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_env,out_dae_elementlist):=
  matchcontinue (in_env,in_program,in_path)
    case (env,(c = SCode.CLASS(name,_,_,_,_)) :: cs,(ref = Absyn.IDENT(name2)))
      equation 
        failure(equality(name = name2));
        (env_1,dae1) = inst_class_decl(env, Types.NOMOD, Prefix.NOPRE, {}, c, {}, false) "packimpl";
        (env_2,dae2) = inst_class_decls(env_1, cs, ref);
        dae = list_append(dae1, dae2); then (env_2,dae);
    case (env,(c = SCode.CLASS(name,_,_,_,_)) :: cs,(ref = Absyn.IDENT(name2)))
      equation 
        equality(name = name2);
        (env_1,dae2) = inst_class_decls(env, cs, ref); then (env_1,dae2);
    case (env,(c = SCode.CLASS(name,_,_,_,_)) :: cs,(ref = Absyn.QUALIFIED(name2,_)))
      equation 
        equality(name = name2);
        (env_1,dae1) = inst_class_decl(env, Types.NOMOD, Prefix.NOPRE, {}, c, {}, false) "How should be do here ???  packimpl";
        (env_2,dae2) = inst_class_decls(env_1, cs, ref);
        dae = list_append(dae1, dae2); then (env_2,dae);
    case (env,(c = SCode.CLASS(name,_,_,_,_)) :: cs,(ref = Absyn.QUALIFIED(name2,_)))
      equation 
        failure(equality(name = name2));
        (env_1,dae1) = inst_class_decl(env, Types.NOMOD, Prefix.NOPRE, {}, c, {}, false) "packimpl";
        (env_2,dae2) = inst_class_decls(env_1, cs, ref);
        dae = list_append(dae1, dae2); then (env_2,dae);
    case (env,{},_) then (env,{}); 
    case (_,_,ref)
      equation 
        print("inst_class_decls failed\n ref =");
        str = Absyn.path_string(ref);
        print(str);
        print("\n"); then fail();
  end matchcontinue;
end inst_class_decls;

public function make_env_from_program " relation: make_env_from_program
 
  This function takes a `SCode.Program' and builds an environment, excluding the class in A1.
"
  input SCode.Program prog;
  input SCode.Path c;
  output Env.Env env_1;
algorithm 
  env := Builtin.initial_env();
  env_1 := add_program_to_env(env, prog, c);
end make_env_from_program;

public function make_simple_env_from_program
  input SCode.Program prog;
  input SCode.Path c;
  output Env.Env env_1;
algorithm 
  env := Builtin.simple_initial_env();
  env_1 := add_program_to_env(env, prog, c);
end make_simple_env_from_program;

protected function add_program_to_env
  input Env.Env env;
  input SCode.Program p;
  input SCode.Path path;
  output Env.Env env_1;
algorithm 
  (env_1,_) := inst_class_decls(env, p, path);
end add_program_to_env;

protected function inst_program " relation: inst_program
 
  Instantiating a Modelica program is the same as instantiating the
  last class definition in the source file. First all the class
  definitions is added to the environment without modifications, and
  then the last class is instantiated in the function `inst_class'.
  This is used when calling the compiler with a Modelica source code file.
  It is not used in the interactive environment when instantiating a class.
 "
  input Env in_env;
  input SCode.Program in_program;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_env,in_program)
    case (env,{})
      equation 
        Print.print_error_buf("# Empty program\n"); then fail();
    case (env,{(c = SCode.CLASS(n,_,_,_,_))})
      equation 
        Debug.fprint("insttr", "inst_program1: ");
        Debug.fprint("insttr", n);
        Debug.fprintln("insttr", "");
        (dae,env_1,csets,_,_) = inst_class(env, Types.NOMOD, Prefix.NOPRE, {}, c, {}, false, TOP_CALL) "Env.extend_frame_c(env,c) => env' &  packimp"; then {DAE.COMP(n,DAE.DAE(dae))};
    case (env,c :: (cs = _ :: _))
      equation 
        Debug.fprintln("insttr", "inst_program2");
        (env_1,dae1) = inst_class_decl(env, Types.NOMOD, Prefix.NOPRE, {}, c, {}, false) "packimpl";
        dae2 = inst_program(env_1, cs) "Env.extend_frame_c(env,c) => env' & ";
        dae = list_append(dae1, dae2); then dae;
    case (_,_)
      equation 
        Debug.fprintln("failtrace", "- inst_program failed"); then fail();
  end matchcontinue;
end inst_program;

protected function inst_program_implicit
  input Env in_env;
  input SCode.Program in_program;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env):=
  matchcontinue (in_env,in_program)
    case (env,(c = SCode.CLASS(n,_,_,restr,_)) :: cs)
      equation 
        Debug.fprint("insttr", "inst_program_implicit: ");
        Debug.fprint("insttr", n);
        Debug.fprintln("insttr", "");
        (env_1,dae1) = implicit_instantiation(env, Types.NOMOD, Prefix.NOPRE, {}, c, {}, false) "packimpl";
        (dae2,env_2) = inst_program_implicit(env_1, cs);
        dae = list_append(dae1, dae2); then (dae,env_2);
    case (env,{})
      equation 
        Debug.fprintln("insttr", "inst_program_implicit (end)"); then ({},env);
  end matchcontinue;
end inst_program_implicit;

public function inst_class "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input SCode.Class in_class;
  input InstDims in_instdims;
  input Boolean in_boolean;
  input CallingScope in_callingscope;
  output DAE_ElementList out_dae_elementlist;
  output Env.Env out_env;
  output Connect.Sets out_sets;
  output Types.Type out_type;
  output ClassInf.State out_state;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_type,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_class,in_instdims,in_boolean,in_callingscope)
    case (env,mod,pre,csets,SCode.CLASS(n,true,_,_,_),_,(impl = false),_) "impl"
      equation 
        Print.print_error_buf("# Can't instantiate partial class ") "impl";
        Print.print_error_buf(n);
        Print.print_error_buf("\n"); then fail();
    case (env,mod,pre,csets,(c = SCode.CLASS(n,_,encflag,r,_)),inst_dims,impl,callscope)
      equation 
        env_1 = Env.open_scope(env, encflag, SOME(n));
        ci_state = ClassInf.start(r, n);
        (dae1,env_3,csets_1,ci_state_1,tys) = inst_class_in(env_1, mod, pre, csets, ci_state, c, false, inst_dims, 
          impl, false) "accesibility packimpl";
        fq_class = make_fully_qualified(env, Absyn.IDENT(n));
        dae1_1 = DAE.set_component_type(dae1, fq_class);
        callscope_1 = is_top_call(callscope);
        dae2 = Connect.equations(csets_1);
        dae3 = Connect.unconnected_flow_equations(csets_1, dae1, env_3, callscope_1);
        dae = Util.list_flatten({dae1_1,dae2,dae3});
        typename = make_fully_qualified(env, Absyn.IDENT(n));
        ty = mktype(typename, ci_state_1, tys); then (dae,env_3,{},ty,ci_state_1);
    case (_,_,_,_,SCode.CLASS(n,_,_,_,_),_,impl,_)
      equation 
        Debug.fprint("failtrace", "- inst_class ");
        Debug.fprint("failtrace", n);
        Debug.fprint("failtrace", " failed\n"); then fail();
  end matchcontinue;
end inst_class;

public function inst_class_in "Accesibility
  implicit instantiation
  implicit function (ls:package?) instantiation "
  input Env in_env1;
  input Mod in_mod2;
  input Prefix in_prefix3;
  input Connect.Sets in_sets4;
  input ClassInf.State in_state5;
  input SCode.Class in_class6;
  input Boolean in_boolean7;
  input InstDims in_instdims8;
  input Boolean in_boolean9;
  input Boolean in_boolean10;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
  output Types_VarList out_types_varlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
  type Types_VarList = list<Types.Var>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state,out_types_varlist):=
  matchcontinue (in_env1,in_mod2,in_prefix3,in_sets4,in_state5,in_class6,in_boolean7,in_instdims8,in_boolean9,in_boolean10)
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS("Real",_,_,_,_)),_,_,impl,packimpl) then ({},env,{},ci_state,{});  "No DAE "
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS("Integer",_,_,_,_)),_,_,impl,packimpl) then ({},env,{},ci_state,{});  "No DAE "
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS("String",_,_,_,_)),_,_,impl,packimpl) then ({},env,{},ci_state,{});  "No DAE "
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS("Boolean",_,_,_,_)),_,_,impl,packimpl) then ({},env,{},ci_state,{});  "No DAE "
    case (env,mods,pre,cset,ci_state,cls,_,_,(impl = false),packimpl) "No DAE  Ignore functions if not implicit instantiation  No DAE "
      equation 
        true = SCode.is_function(cls);
        Debug.fprint("insttr", "Ignoring function in explicit instantiation: ");
        clsname = SCode.class_name(cls);
        Debug.fprint("insttr", clsname);
        Debug.fprint("insttr", "\n"); then ({},env,{},ci_state,{});
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS(n,_,_,r,d)),prot,inst_dims,impl,packimpl)
      equation 
        clsname = SCode.class_name(c) "print \"inst_class_in\" & print n & print \"\n\" &";
        Debug.fprint("insttr", "Instantiating class: ") "print \"inst_class_in\" & print n & print \"\n\" &";
        implstr = Util.if_(impl, " (implicit) ", " (explicit) ");
        Debug.fprint("insttr", implstr);
        Debug.fprint("insttr", clsname);
        Debug.fprint("insttr", "\n");
        (l,env_1,csets_1,ci_state_1,tys) = inst_classdef(env, mods, pre, csets, ci_state, d, r, prot, inst_dims, 
          impl, packimpl); then (l,env_1,csets_1,ci_state_1,tys);
    case (_,_,_,csets,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_class_in failed\n"); then fail();
  end matchcontinue;
end inst_class_in;

public function partial_inst_class_in
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.Class in_class;
  input Boolean in_boolean;
  input InstDims in_instdims;
  output Env out_env;
  output ClassInf.State out_state;
algorithm 
  (out_env,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_class,in_boolean,in_instdims)
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS("Real",_,_,_,_)),_,_) then (env,ci_state); 
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS("Integer",_,_,_,_)),_,_) then (env,ci_state); 
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS("String",_,_,_,_)),_,_) then (env,ci_state); 
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS("Boolean",_,_,_,_)),_,_) then (env,ci_state); 
    case (env,mods,pre,csets,ci_state,(c = SCode.CLASS(n,_,_,r,d)),prot,inst_dims)
      equation 
        (env_1,ci_state_1) = partial_inst_classdef(env, mods, pre, csets, ci_state, d, r, prot, inst_dims); then (env_1,ci_state_1);
  end matchcontinue;
end partial_inst_class_in;

protected function inst_classdef "implicit inst
  implicit package inst"
  input Env in_env1;
  input Mod in_mod2;
  input Prefix in_prefix3;
  input Connect.Sets in_sets4;
  input ClassInf.State in_state5;
  input SCode.ClassDef in_classdef6;
  input SCode.Restriction in_restriction7;
  input Boolean in_boolean8;
  input InstDims in_instdims9;
  input Boolean in_boolean10;
  input Boolean in_boolean11;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
  output Types_VarList out_types_varlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
  type Types_VarList = list<Types.Var>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state,out_types_varlist):=
  matchcontinue (in_env1,in_mod2,in_prefix3,in_sets4,in_state5,in_classdef6,in_restriction7,in_boolean8,in_instdims9,in_boolean10,in_boolean11)
    case (env,mods,pre,csets,ci_state,SCode.PARTS(els,eqs,initeqs,alg,initalg,_),re,prot,inst_dims,impl,(packimpl = false)) "implicit inst implicit package inst  This rule describes how to instantiate an explicit
	 ** class definition "
      equation 
        ci_state1 = ClassInf.trans(ci_state, ClassInf.NEWDEF) "Prefix.print_prefix_str pre => prestr & Debug.fprintl (\"insttr\", {\"inst_class parts: pre=\",prestr,\"\n\"}) &";
        cdefelts = classdef_elts(els);
        compelts = component_elts(els);
        extendselts = extends_elts(els);
        env1 = add_classdefs_to_env(env, cdefelts, impl) "1. CLASSDEF & IMPORT nodes and COMPONENT nodes(add to env)";
        (env2,emods,extcomps,eqs2,initeqs2,alg2,initalg2) = inst_extends_list(env1, mods, extendselts, ci_state, impl) "2. EXTENDS Nodes inst_extends_list only flatten inhteritance structure.  It does not perform component instantiations. ";
        compelts_1 = add_nomod(compelts) "Problem. Modifiers on inherited components are unelabed, loosing their 
	   type information. This will not work, since the modifier type can not always be found.
	   for instance. 
	   model B extends B2; end B; model B2 Integer ni=1; end B2;
	   model test
	     Integer n=2;
	      B b(ni=n);
	   end test;
	   The modifier (n=n) will be untypes when B is instantiated and the variable n can not be 
	   found, since the component b is instantiated in env of B.
	   Solution:
	   Redesign inst_extends_list to return (SCode.Element, Types.Mod) list and
	   convert other component elements to the same format, such that inst_element can 
	   handle the new format uniformely.
	   ";
        cdefelts_1 = add_nomod(cdefelts);
        compelts_1 = Util.list_flatten({extcomps,compelts_1,cdefelts_1});
        env3 = add_components_to_env(env2, emods, pre, csets, ci_state, compelts_1, compelts_1, 
          inst_dims, impl) "Add variables to env, wihtout type and binding,  which will be added later in inst_element_list   (where update_variable is called) ";
        (compelts_2,env4) = update_compelts_mods(env3, pre, compelts_1, ci_state, impl) "Update the modifiers of elements to typed ones, needed for modifiers
	   on components that are inherited. ";
        eqs_1 = list_append(eqs, eqs2) "Add components from base classes to be instantiated in 3 as well. ";
        initeqs_1 = list_append(initeqs, initeqs2);
        alg_1 = list_append(alg, alg2);
        initalg_1 = list_append(initalg, initalg2);
        (dae1,env5,csets1,ci_state2,tys) = inst_element_list(env4, mods, pre, csets, ci_state1, compelts_2, inst_dims, 
          impl, packimpl) "3. Instantiate components ";
        (dae2,_,csets2,ci_state3) = inst_list(env5, mods, pre, csets1, ci_state2, inst_equation, eqs_1, 
          impl);
        (dae3,_,csets3,ci_state4) = inst_list(env5, mods, pre, csets2, ci_state3, inst_initialequation, 
          initeqs_1, impl);
        (dae4,_,csets4,ci_state5) = inst_list(env5, mods, pre, csets3, ci_state4, inst_algorithm, alg_1, 
          impl);
        (dae5,_,csets5,ci_state6) = inst_list(env5, mods, pre, csets4, ci_state5, inst_initialalgorithm, 
          initalg_1, impl);
        dae6 = list_append(dae1, dae2);
        dae7 = list_append(dae6, dae3);
        dae8 = list_append(dae7, dae4);
        dae = list_append(dae8, dae5); then (dae,env5,csets5,ci_state6,tys);
    case (env,mods,pre,csets,ci_state,SCode.PARTS(els,eqs,initeqs,alg,initalg,_),re,prot,inst_dims,impl,true) " This rule describes how to instantiate an explicit
	 ** class definition  packimpl"
      equation 
        ci_state1 = ClassInf.trans(ci_state, ClassInf.NEWDEF) "Debug.fprintl (\"insttr\", {\"inst_class parts: pre=\",prestr,\"\n\"}) & ";
        cdefelts = classdef_elts(els);
        compelts = component_elts(els);
        extendselts = extends_elts(els);
        env1 = add_classdefs_to_env(env, cdefelts, impl) "1. CLASSDEF & IMPORT nodes and COMPONENT nodes(add to env)";
        (env2,emods,extcomps,eqs2,initeqs2,alg2,initalg2) = inst_extends_list(env1, mods, extendselts, ci_state, impl) "2. EXTENDS Nodes inst_extends_list only flatten inhteritance structure.  It does not perform component instantiations. ";
        compelts_1 = add_nomod(compelts);
        cdefelts_1 = add_nomod(cdefelts);
        compelts_1 = Util.list_flatten({extcomps,compelts_1,cdefelts_1});
        env3 = add_components_to_env(env2, emods, pre, csets, ci_state, compelts_1, compelts_1, 
          inst_dims, impl) "Add variables to env, wihtout type and binding,  which will be added later in inst_element_list   (where update_variable is called) ";
        eqs_1 = list_append(eqs, eqs2) "Add components from base classes to be instantiated in 3 as well. ";
        initeqs_1 = list_append(initeqs, initeqs2);
        alg_1 = list_append(alg, alg2);
        initalg_1 = list_append(initalg, initalg2);
        (dae1,env4,csets1,ci_state2,tys) = inst_element_list(env3, mods, pre, csets, ci_state1, compelts_1, inst_dims, 
          impl, true) "3. Instantiate components  packimpl"; then (dae1,env4,csets1,ci_state2,tys);
    case (env,mods,pre,csets,ci_state,SCode.DERIVED(cn,ad,mod),re,prot,inst_dims,impl,packimpl) "packimpl  This rule describes how to instantiate a derived class   definition "
      equation 
        (c = SCode.CLASS(cn2,_,enc2,r,_),cenv) = Lookup.lookup_class(env, cn, true) "Prefix.print_prefix_str pre => prestr &
	Absyn.path_string cn => cnstr & Debug.fprintl (\"insttr\", {\"inst_class derived: \", cnstr, \", pre=\",prestr,\"\n\"}) &";
        cenv_2 = Env.open_scope(cenv, enc2, SOME(cn2));
        m = Mod.lookup_modification_p(mods, cn);
        mod_1 = Mod.elab_mod(env, pre, mod, impl);
        new_ci_state = ClassInf.start(r, cn2);
        mods_1 = Mod.merge(mods, m, cenv_2, pre);
        mods_2 = Mod.merge(mods_1, mod_1, cenv_2, pre);
        (dae,env_2,csets_1,ci_state_1,tys) = inst_class_in(cenv_2, mods_2, pre, csets, new_ci_state, c, prot, 
          inst_dims, impl, false) "packimpl";
        ClassInf.assert_valid(ci_state_1, re) "Check for restriction violations "; then (dae,env_2,csets_1,ci_state_1,tys);
    case (env,mods,pre,csets,ci_state,SCode.DERIVED(cn,ad,mod),re,prot,inst_dims,impl,packimpl) " If the class is derived from a class that can not be   found in the environment, this rule prints an error   message. "
      equation 
        failure((_,_) = Lookup.lookup_class(env, cn, false));
        cns = Absyn.path_string(cn);
        Print.print_error_buf("# unknown class: ");
        Print.print_error_buf(cns);
        Print.print_error_buf(" in scope: ");
        str = Env.print_env_path_str(env);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
    case (_,_,_,_,_,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_classdef failed\n"); then fail();
  end matchcontinue;
end inst_classdef;

protected function partial_inst_classdef
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.ClassDef in_classdef;
  input SCode.Restriction in_restriction;
  input Boolean in_boolean;
  input InstDims in_instdims;
  output Env out_env;
  output ClassInf.State out_state;
algorithm 
  (out_env,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_classdef,in_restriction,in_boolean,in_instdims)
    case (env,mods,pre,csets,ci_state,SCode.PARTS(els,eqs,initeqs,alg,initalg,_),re,prot,inst_dims)
      equation 
        ci_state1 = ClassInf.trans(ci_state, ClassInf.NEWDEF);
        cdefelts = classdef_elts(els);
        extendselts = extends_elts(els);
        env1 = add_classdefs_to_env(env, cdefelts, true) "1. CLASSDEF & IMPORT nodes and COMPONENT nodes(add to env)";
        (env2,emods,extcomps,eqs2,initeqs2,alg2,initalg2) = inst_extends_list(env1, mods, extendselts, ci_state, true) "2. EXTENDS Nodes inst_extends_list only flatten inhteritance structure.  It does not perform component instantiations. "; then (env2,ci_state1);
    case (env,mods,pre,csets,ci_state,SCode.DERIVED(cn,ad,mod),re,prot,inst_dims) " This rule describes how to instantiate a derived class   definition "
      equation 
        (c = SCode.CLASS(cn2,_,enc2,r,_),cenv) = Lookup.lookup_class(env, cn, true);
        cenv_2 = Env.open_scope(cenv, enc2, SOME(cn2));
        m = Mod.lookup_modification_p(mods, cn);
        mod_1 = Mod.elab_mod(env, pre, mod, false) "FIXME: impl";
        new_ci_state = ClassInf.start(r, cn2);
        mods_1 = Mod.merge(mods, m, cenv_2, pre);
        mods_2 = Mod.merge(mods_1, mod_1, cenv_2, pre);
        (env_2,new_ci_state_1) = partial_inst_class_in(cenv_2, mods_2, pre, csets, new_ci_state, c, prot, 
          inst_dims); then (env_2,new_ci_state_1);
    case (env,mods,pre,csets,ci_state,SCode.DERIVED(cn,ad,mod),re,prot,inst_dims) " If the class is derived from a class that can not be   found in the environment, this rule prints an error   message. "
      equation 
        failure((_,_) = Lookup.lookup_class(env, cn, false));
        cns = Absyn.path_string(cn);
        Print.print_error_buf("# unknown class: ");
        Print.print_error_buf(cns);
        Print.print_error_buf(" in scope: ");
        str = Env.print_env_path_str(env);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
  end matchcontinue;
end partial_inst_classdef;

protected function update_compelts_mods
  input Env.Env in_env;
  input Prefix.Prefix in_prefix;
  input TupleSCode_ElementTypes_ModList in_tuplescode_elementtypes_modlist;
  input ClassInf.State in_state;
  input Boolean in_boolean;
  output TupleSCode_ElementTypes_ModList out_tuplescode_elementtypes_modlist;
  output Env.Env out_env;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
algorithm 
  (out_tuplescode_elementtypes_modlist,out_env):=
  matchcontinue (in_env,in_prefix,in_tuplescode_elementtypes_modlist,in_state,in_boolean)
    case (env,pre,{},_,_) then ({},env); 
    case (env,pre,(comp,cmod) :: xs,ci_state,impl)
      equation 
        umod = Mod.unelab_mod(cmod);
        crefs = get_cref_from_mod(umod);
        env2 = update_components_in_env(cmod, crefs, env, ci_state, impl);
        cmod_1 = Mod.update_mod(env2, pre, cmod, impl);
        (res,env3) = update_compelts_mods(env2, pre, xs, ci_state, impl); then ((comp,cmod_1) :: res,env3);
  end matchcontinue;
end update_compelts_mods;

protected function get_option_arraydim " relation: get_option_arraydim
 
  Return the Arraydim of an optional arradim. Empty list returned if no arraydim present.
 "
  input Absyn_ArrayDimOption in_absyn_arraydimoption;
  output Absyn.ArrayDim out_arraydim;
protected 
  type Absyn_ArrayDimOption = Option<Absyn.ArrayDim>;
algorithm 
  out_arraydim:=
  matchcontinue (in_absyn_arraydimoption)
    case (SOME(dim)) then dim; 
    case (NONE) then {}; 
  end matchcontinue;
end get_option_arraydim;

protected function inst_extends_list "inherited initial equations 
  inherited algorithms 
  inherited initial algorithms "
  input Env.Env in_env;
  input Types.Mod in_mod;
  input SCode_ElementList in_scode_elementlist;
  input ClassInf.State in_state;
  input Boolean in_boolean;
  output Env.Env out_env1;
  output Types.Mod out_mod2;
  output TupleSCode_ElementTypes_ModList out_tuplescode_elementtypes_modlist3;
  output SCode_EquationList out_scode_equationlist4;
  output SCode_EquationList out_scode_equationlist5;
  output SCode_AlgorithmList out_scode_algorithmlist6;
  output SCode_AlgorithmList out_scode_algorithmlist7;
protected 
  type SCode_ElementList = list<SCode.Element>;
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
  type SCode_EquationList = list<SCode.Equation>;
  type SCode_AlgorithmList = list<SCode.Algorithm>;
algorithm 
  (out_env1,out_mod2,out_tuplescode_elementtypes_modlist3,out_scode_equationlist4,out_scode_equationlist5,out_scode_algorithmlist6,out_scode_algorithmlist7):=
  matchcontinue (in_env,in_mod,in_scode_elementlist,in_state,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type TupleSCode_ElementMod = tuple<SCode.Element,Mod>;
      type TupleSCode_ElementModList = list<TupleSCode_ElementMod>;
      SCode.Class c;
      String cn;
      Boolean encf,impl;
      SCode.Restriction r;
      Env_FrameList cenv,cenv1,cenv3,env2,env,env_1;
      Types.Mod outermod,mod_1,mod_2,mods,mods_1,emod_1,mod;
      SCode_ElementList els,els_1,rest;
      SCode_EquationList eq1,ieq1,eq2,ieq2,eq3,ieq3,eq,ieq,initeq2;
      SCode_AlgorithmList alg1,ialg1,alg2,ialg2,alg3,ialg3,alg,ialg;
      Absyn.Path tp_1,tp;
      ClassInf.State new_ci_state,ci_state;
      TupleSCode_ElementModList compelts1,compelts2,compelts,compelts3;
      SCode.Mod emod;
      SCode.Element elt;
    case (env,mod,SCode.EXTENDS(tp,emod) :: rest,ci_state,impl) "inherited initial equations  inherited algorithms  inherited initial algorithms "
      equation 
        (c = SCode.CLASS(cn,_,encf,r,_),cenv) = Lookup.lookup_class(env, tp, true);
        outermod = Mod.lookup_modification_p(mod, Absyn.IDENT(cn));
        (cenv1,els,eq1,ieq1,alg1,ialg1) = inst_derived_classes(cenv, outermod, c, impl);
        tp_1 = make_fully_qualified(cenv1, tp);
        els_1 = add_inherit_scope(els, tp_1);
        cenv3 = Env.open_scope(cenv1, encf, SOME(cn));
        new_ci_state = ClassInf.start(r, cn);
        mod_1 = Mod.elab_untyped_mod(emod, cenv3, Prefix.NOPRE);
        mod_2 = Mod.merge(outermod, mod_1, cenv3, Prefix.NOPRE);
        (_,mods,compelts1,eq2,ieq2,alg2,ialg2) = inst_extends_list(cenv1, outermod, els_1, ci_state, impl) "env";
        (env2,mods_1,compelts2,eq3,ieq3,alg3,ialg3) = inst_extends_list(env, mod, rest, ci_state, impl);
        emod_1 = Mod.elab_untyped_mod(emod, env2, Prefix.NOPRE) "corresponding elements. But emod is Absyn.Mod and can not   Must merge(mod,emod) here and then apply the bindings to the  be elaborated, because for instance extends A(x=y) can reference  a variable y defined in A and will thus not be found.  On the other hand: A(n=4), n might be a structural parameter that must be set to  instantiate A. How could this be solved?  Solution: made new function elab_untyped_mod which transforms  to a Mod, but set the type information to unknown.  We can then perform the merge, and update untyped modifications  later (using update_mod), when we are instantiating the components. ";
        mod_1 = Mod.merge(mod, mods_1, env2, Prefix.NOPRE);
        mods_1 = Mod.merge(mod_1, emod_1, env2, Prefix.NOPRE);
        compelts = list_append(compelts1, compelts2);
        compelts3 = update_components(compelts, mods_1, env2) "	  Print.print_buf \"\n compelts list: \" &
	  SCode.print_element_list(compelts) &
	Print.print_buf \"mod =\" & 
	Mod.print_mod mods' &	
	Print.get_string() => str1 & print str1 &";
        eq = Util.list_flatten({eq1,eq2,eq3});
        ieq = Util.list_flatten({ieq1,ieq2,ieq3});
        alg = Util.list_flatten({alg1,alg2,alg3});
        ialg = Util.list_flatten({ialg1,ialg2,ialg3}); then (env2,mods_1,compelts3,eq,ieq,alg,ialg);
    case (env,mod,SCode.EXTENDS(tp,emod) :: rest,ci_state,impl)
      equation 
        failure((c = SCode.CLASS(cn,_,encf,r,_),cenv) = Lookup.lookup_class(env, tp, true));
        Debug.fprint("failtrace", "Failed lookup of inherited class ");
        Debug.fcall("failtrace", Dump.print_path, tp);
        Debug.fprint("failtrace", ", env: ");
        Debug.fcall("failtrace", Env.print_env, env);
        Debug.fprint("failtrace", "\n"); then fail();
    case (env,mod,SCode.EXTENDS(tp,emod) :: rest,ci_state,impl)
      equation 
        Debug.fprint("failtrace", "Failed inst_extends_list on EXTENDS\n env:");
        Env.print_env(env); then fail();
    case (env,mod,elt :: rest,ci_state,impl)
      equation 
        (env_1,mods,compelts2,eq2,initeq2,alg2,ialg2) = inst_extends_list(env, mod, rest, ci_state, impl); then (env_1,mods,(elt,Types.NOMOD) :: compelts2,eq2,initeq2,alg2,ialg2);
    case (env,mod,{},ci_state,impl) then (env,mod,{},{},{},{},{}); 
    case (_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_extends_list failed\n"); then fail();
  end matchcontinue;
end inst_extends_list;

protected function add_inherit_scope
  input SCode_ElementList in_scode_elementlist;
  input Absyn.Path in_path;
  output SCode_ElementList out_scode_elementlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_scode_elementlist:=
  matchcontinue (in_scode_elementlist,in_path)
    local
      type Absyn_CommentOption = Option<Absyn.Comment>;
      SCode_ElementList res,xs;
      String a;
      Boolean b,c,d;
      SCode.Attributes e;
      Absyn.Path f,tp;
      SCode.Mod g;
      Absyn_CommentOption comment;
      SCode.Element x;
    case ({},_) then {}; 
    case (SCode.COMPONENT(a,b,c,d,e,f,g,_,comment) :: xs,tp)
      equation 
        res = add_inherit_scope(xs, tp); then SCode.COMPONENT(a,b,c,d,e,f,g,SOME(tp),comment) :: res;
    case (SCode.CLASSDEF(a,b,c,d,_) :: xs,tp)
      local SCode.Class d;
      equation 
        res = add_inherit_scope(xs, tp); then SCode.CLASSDEF(a,b,c,d,SOME(tp)) :: res;
    case (x :: xs,tp)
      equation 
        res = add_inherit_scope(xs, tp); then x :: res;
    case (_,_)
      equation 
        print("add_inherit_scope failed\n"); then fail();
  end matchcontinue;
end add_inherit_scope;

public function add_nomod " relation: add_nomod 
 
  This function takes an SCode.Element list and tranforms it into a (SCode.Element* Types.Mod) list
  by inserting Types.NOMOD for each element. Used to transform elements into a uniform list 
  combined from inherited elements and ordinary elements.
 "
  input SCode_ElementList in_scode_elementlist;
  output TupleSCode_ElementTypes_ModList out_tuplescode_elementtypes_modlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
algorithm 
  out_tuplescode_elementtypes_modlist:=
  matchcontinue (in_scode_elementlist)
    local
      type TupleSCode_ElementMod = tuple<SCode.Element,Mod>;
      type TupleSCode_ElementModList = list<TupleSCode_ElementMod>;
      TupleSCode_ElementModList res;
      SCode.Element x;
      SCode_ElementList xs;
    case {} then {}; 
    case (x :: xs)
      equation 
        res = add_nomod(xs); then (x,Types.NOMOD) :: res;
  end matchcontinue;
end add_nomod;

protected function update_components
  input TupleSCode_ElementTypes_ModList in_tuplescode_elementtypes_modlist;
  input Types.Mod in_mod;
  input Env.Env in_env;
  output TupleSCode_ElementTypes_ModList out_tuplescode_elementtypes_modlist;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
algorithm 
  out_tuplescode_elementtypes_modlist:=
  matchcontinue (in_tuplescode_elementtypes_modlist,in_mod,in_env)
    local
      type TupleSCode_ElementMod = tuple<SCode.Element,Mod>;
      type TupleSCode_ElementModList = list<TupleSCode_ElementMod>;
      type Env_FrameList = list<Env.Frame>;
      Types.Mod cmod2,mod_1,cmod,mod,emod;
      TupleSCode_ElementModList res,xs;
      SCode.Element comp,c;
      String id;
      Env_FrameList env;
    case ({},_,_) then {}; 
    case (((comp = SCode.COMPONENT(id,_,_,_,_,_,_,_,_)),cmod) :: xs,mod,env)
      equation 
        cmod2 = Mod.lookup_comp_modification(mod, id);
        mod_1 = Mod.merge(cmod, cmod2, env, Prefix.NOPRE);
        res = update_components(xs, mod, env); then (comp,mod_1) :: res;
    case (((c = SCode.EXTENDS(_,_)),emod) :: xs,mod,env)
      equation 
        res = update_components(xs, mod, env); then (c,emod) :: res;
    case (((c = SCode.CLASSDEF(_,_,_,_,_)),cmod) :: xs,mod,env)
      equation 
        res = update_components(xs, mod, env); then (c,cmod) :: res;
    case (((c = SCode.IMPORT(_)),_) :: xs,mod,env)
      equation 
        res = update_components(xs, mod, env); then (c,Types.NOMOD) :: res;
    case (_,_,_)
      equation 
        Debug.fprint("failtrace", "-update_components failed\n"); then fail();
  end matchcontinue;
end update_components;

protected function no_import_elements " relation: no_import_elements
 
  Returns all elements except imports, i.e. filter out import elements.
 "
  input SCode_ElementList in_scode_elementlist;
  output SCode_ElementList out_scode_elementlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_scode_elementlist:=
  matchcontinue (in_scode_elementlist)
    local
      SCode_ElementList elt,rest;
      SCode.Element e;
    case {} then {}; 
    case SCode.IMPORT(_) :: rest
      equation 
        elt = no_import_elements(rest); then elt;
    case e :: rest
      equation 
        elt = no_import_elements(rest); then e :: elt;
  end matchcontinue;
end no_import_elements;

protected function inst_derived_classes
  input Env.Env in_env;
  input Types.Mod in_mod;
  input SCode.Class in_class;
  input Boolean in_boolean;
  output Env.Env out_env1;
  output SCode_ElementList out_scode_elementlist2;
  output SCode_EquationList out_scode_equationlist3;
  output SCode_EquationList out_scode_equationlist4;
  output SCode_AlgorithmList out_scode_algorithmlist5;
  output SCode_AlgorithmList out_scode_algorithmlist6;
protected 
  type SCode_ElementList = list<SCode.Element>;
  type SCode_EquationList = list<SCode.Equation>;
  type SCode_AlgorithmList = list<SCode.Algorithm>;
algorithm 
  (out_env1,out_scode_elementlist2,out_scode_equationlist3,out_scode_equationlist4,out_scode_algorithmlist5,out_scode_algorithmlist6):=
  matchcontinue (in_env,in_mod,in_class,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      SCode_ElementList elt_1,elt;
      Env_FrameList env,cenv;
      Types.Mod mod;
      SCode_EquationList eq,ieq;
      SCode_AlgorithmList alg,ialg;
      SCode.Class c;
      Absyn.Path tp;
      SCode.Mod dmod;
      Boolean impl;
    case (env,mod,SCode.CLASS(_,_,_,_,SCode.PARTS(elt,eq,ieq,alg,ialg,_)),_)
      equation 
        elt_1 = no_import_elements(elt); then (env,elt_1,eq,ieq,alg,ialg);
    case (env,mod,SCode.CLASS(_,_,_,_,SCode.DERIVED(tp,_,dmod)),impl)
      equation 
        (c,cenv) = Lookup.lookup_class(env, tp, true);
        (env,elt,eq,ieq,alg,ialg) = inst_derived_classes(cenv, mod, c, impl) "Mod.lookup_modification_p(mod, c) => innermod &  We have to merge and apply modifications as well! "; then (env,elt,eq,ieq,alg,ialg);
    case (_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_derived_classes failed\n"); then fail();
  end matchcontinue;
end inst_derived_classes;

protected function inst_element_list "impl
  packimpl"
  input Env in_env1;
  input Mod in_mod2;
  input Prefix in_prefix3;
  input Connect.Sets in_sets4;
  input ClassInf.State in_state5;
  input TupleSCode_ElementTypes_ModList in_tuplescode_elementtypes_modlist6;
  input InstDims in_instdims7;
  input Boolean in_boolean8;
  input Boolean in_boolean9;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
  output Types_VarList out_types_varlist;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
  type DAE_ElementList = list<DAE.Element>;
  type Types_VarList = list<Types.Var>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state,out_types_varlist):=
  matchcontinue (in_env1,in_mod2,in_prefix3,in_sets4,in_state5,in_tuplescode_elementtypes_modlist6,in_instdims7,in_boolean8,in_boolean9)
    case (env,_,_,csets,ci_state,{},_,_,_) then ({},env,csets,ci_state,{}); 
    case (env,mod,pre,csets,ci_state,el :: els,inst_dims,impl,packimpl)
      equation 
        (dae1,env_1,csets_1,ci_state_1,tys1) = inst_element(env, mod, pre, csets, ci_state, el, inst_dims, impl, 
          packimpl);
        (dae2,env_2,csets_2,ci_state_2,tys2) = inst_element_list(env_1, mod, pre, csets_1, ci_state_1, els, inst_dims, 
          impl, packimpl);
        tys = list_append(tys1, tys2);
        dae = list_append(dae1, dae2); then (dae,env_2,csets_2,ci_state_2,tys);
    case (_,_,_,_,_,els,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_element_list failed\n"); then fail();
  end matchcontinue;
end inst_element_list;

protected function classdef_elts " relation: classdef_elts
  author: PA
 
  This function filters out the class definitions of an Element list.
 "
  input SCode_ElementList in_scode_elementlist;
  output SCode_ElementList out_scode_elementlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_scode_elementlist:=
  matchcontinue (in_scode_elementlist)
    local
      SCode_ElementList res,xs;
      SCode.Element cdef,imp;
    case ({}) then {}; 
    case ((cdef = SCode.CLASSDEF(_,_,_,_,_)) :: xs)
      equation 
        res = classdef_elts(xs); then cdef :: res;
    case ((imp = SCode.IMPORT(_)) :: xs)
      equation 
        res = classdef_elts(xs); then imp :: res;
    case (_ :: xs)
      equation 
        res = classdef_elts(xs); then res;
  end matchcontinue;
end classdef_elts;

protected function extends_elts " relation: extends_elts
  author: PA
 
  This function filters out the extends Element's in an Element list
 "
  input SCode_ElementList in_scode_elementlist;
  output SCode_ElementList out_scode_elementlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_scode_elementlist:=
  matchcontinue (in_scode_elementlist)
    local
      SCode_ElementList res,xs;
      SCode.Element cdef;
    case ({}) then {}; 
    case ((cdef = SCode.EXTENDS(_,_)) :: xs)
      equation 
        res = extends_elts(xs); then cdef :: res;
    case (_ :: xs)
      equation 
        res = extends_elts(xs); then res;
  end matchcontinue;
end extends_elts;

protected function component_elts " relation: component_elts
  author: PA
 
  This function filters out the component Element's in an Element list
 "
  input SCode_ElementList in_scode_elementlist;
  output SCode_ElementList out_scode_elementlist;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_scode_elementlist:=
  matchcontinue (in_scode_elementlist)
    local
      SCode_ElementList res,xs;
      SCode.Element cdef;
    case ({}) then {}; 
    case ((cdef = SCode.COMPONENT(_,_,_,_,_,_,_,_,_)) :: xs)
      equation 
        res = component_elts(xs); then cdef :: res;
    case (_ :: xs)
      equation 
        res = component_elts(xs); then res;
  end matchcontinue;
end component_elts;

protected function add_classdefs_to_env " relation: add_classdefs_to_env
  author: PA
 
  This function adds classdefinitions and import statements to the environment.
 "
  input Env in_env;
  input SCode_ElementList in_scode_elementlist;
  input Boolean in_boolean;
  output Env out_env;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_env:=
  matchcontinue (in_env,in_scode_elementlist,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList env,env_1,env_2;
      SCode.Class cl;
      SCode_ElementList xs;
      Boolean impl;
      Absyn.Import imp;
    case (env,{},_) then env; 
    case (env,SCode.CLASSDEF(_,_,_,cl,_) :: xs,impl)
      equation 
        env_1 = Env.extend_frame_c(env, cl);
        env_2 = add_classdefs_to_env(env_1, xs, impl); then env_2;
    case (env,SCode.IMPORT(imp) :: xs,impl)
      equation 
        env_1 = Env.extend_frame_i(env, imp);
        env_2 = add_classdefs_to_env(env_1, xs, impl); then env_2;
  end matchcontinue;
end add_classdefs_to_env;

protected function is_structural_parameter "components variability
  component name 
  all elements "
  input SCode.Variability in_variability;
  input Absyn.ComponentRef in_componentref;
  input TupleSCode_ElementTypes_ModList in_tuplescode_elementtypes_modlist;
  output Boolean out_boolean;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
algorithm 
  out_boolean:=
  matchcontinue (in_variability,in_componentref,in_tuplescode_elementtypes_modlist)
    local
      type Absyn_ComponentRefList = list<Absyn.ComponentRef>;
      type TupleSCode_ElementMod = tuple<SCode.Element,Mod>;
      type TupleSCode_ElementModList = list<TupleSCode_ElementMod>;
      Absyn_ComponentRefList crefs;
      Boolean res;
      SCode.Variability param;
      Absyn.ComponentRef compname;
      TupleSCode_ElementModList allcomps;
    case (param,compname,allcomps)
      equation 
        true = SCode.is_parameter_or_const(param);
        crefs = get_crefs_from_compdims(allcomps);
        res = member_crefs(compname, crefs); then res;
    case (_,_,_) then false; 
  end matchcontinue;
end is_structural_parameter;

protected function add_components_to_env
  input Env in_env1;
  input Mod in_mod2;
  input Prefix in_prefix3;
  input Connect.Sets in_sets4;
  input ClassInf.State in_state5;
  input TupleSCode_ElementTypes_ModList in_tuplescode_elementtypes_modlist6;
  input TupleSCode_ElementTypes_ModList in_tuplescode_elementtypes_modlist7;
  input InstDims in_instdims8;
  input Boolean in_boolean9;
  output Env out_env;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
algorithm 
  out_env:=
  matchcontinue (in_env1,in_mod2,in_prefix3,in_sets4,in_state5,in_tuplescode_elementtypes_modlist6,in_tuplescode_elementtypes_modlist7,in_instdims8,in_boolean9)
    local
      type Env_FrameList = list<Env.Frame>;
      type Connect_SetList = list<Connect.Set>;
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type TupleSCode_ElementMod = tuple<SCode.Element,Mod>;
      type TupleSCode_ElementModList = list<TupleSCode_ElementMod>;
      Env_FrameList env,env_1,env_2;
      Types.Mod mod,cmod;
      Prefix.Prefix pre;
      Connect_SetList csets;
      ClassInf.State cistate;
      SCode.Element comp;
      String n;
      Boolean final_,repl,prot,flow_,impl;
      SCode.Attributes attr;
      Absyn_SubscriptList ad;
      SCode.Accessibility acc;
      SCode.Variability param;
      Absyn.Direction dir;
      Absyn.Path t;
      SCode.Mod m;
      Absyn_PathOption bc;
      Absyn_CommentOption comment;
      TupleSCode_ElementModList xs,allcomps,comps;
      InstDims instdims;
    case (env,_,_,_,_,{},_,_,_) then env; 
    case (env,mod,pre,csets,cistate,((comp = SCode.COMPONENT(n,final_,repl,prot,(attr = SCode.ATTR(ad,flow_,acc,param,dir)),t,m,bc,comment)),cmod) :: xs,allcomps,instdims,impl)
      equation 
        failure(ClassInf.is_function(cistate)) "Check if the component is a structural parameter, change it's
	 attribute to STRUCTPARAM. ";
        true = is_structural_parameter(param, Absyn.CREF_IDENT(n,{}), allcomps) "Functions should not be considered ";
        env_1 = add_components_to_env2(env, mod, pre, csets, cistate, 
          {
          (
          SCode.COMPONENT(n,final_,repl,prot,
          SCode.ATTR(ad,flow_,acc,SCode.STRUCTPARAM,dir),t,m,bc,comment),cmod)}, instdims, impl);
        env_2 = add_components_to_env(env_1, mod, pre, csets, cistate, xs, allcomps, instdims, 
          impl); then env_2;
    case (env,mod,pre,csets,cistate,((comp = SCode.COMPONENT(n,final_,repl,prot,(attr = SCode.ATTR(ad,flow_,acc,param,dir)),t,m,bc,comment)),cmod) :: xs,allcomps,instdims,impl)
      equation 
        env_1 = add_components_to_env2(env, mod, pre, csets, cistate, 
          {
          (
          SCode.COMPONENT(n,final_,repl,prot,SCode.ATTR(ad,flow_,acc,param,dir),t,m,
          bc,comment),cmod)}, instdims, impl) "Not structural parameter. No Change.";
        env_2 = add_components_to_env(env_1, mod, pre, csets, cistate, xs, allcomps, instdims, 
          impl); then env_2;
    case (env,mod,pre,csets,cistate,(SCode.IMPORT(_),_) :: xs,allcomps,instdims,impl)
      equation 
        env_2 = add_components_to_env(env, mod, pre, csets, cistate, xs, allcomps, instdims, impl); then env_2;
    case (env,mod,pre,csets,cistate,(SCode.EXTENDS(_,_),_) :: xs,allcomps,instdims,impl)
      equation 
        env_2 = add_components_to_env(env, mod, pre, csets, cistate, xs, allcomps, instdims, impl); then env_2;
    case (env,mod,pre,csets,cistate,(SCode.CLASSDEF(_,_,_,_,_),_) :: xs,allcomps,instdims,impl)
      equation 
        env_2 = add_components_to_env(env, mod, pre, csets, cistate, xs, allcomps, instdims, impl); then env_2;
    case (_,_,_,_,_,comps,_,_,_)
      equation 
        Debug.fprint("failtrace", "- add_components_to_env failed\n"); then fail();
  end matchcontinue;
end add_components_to_env;

protected function add_components_to_env2
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input TupleSCode_ElementTypes_ModList in_tuplescode_elementtypes_modlist;
  input InstDims in_instdims;
  input Boolean in_boolean;
  output Env out_env;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
algorithm 
  out_env:=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_tuplescode_elementtypes_modlist,in_instdims,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type Connect_SetList = list<Connect.Set>;
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type TupleSCode_ElementMod = tuple<SCode.Element,Mod>;
      type TupleSCode_ElementModList = list<TupleSCode_ElementMod>;
      Types.Mod compmod,cmod_1,mods,cmod;
      Env_FrameList env_1,env_2,env;
      Prefix.Prefix pre;
      Connect_SetList csets;
      ClassInf.State ci_state;
      SCode.Element comp;
      String n;
      Boolean final_,repl,prot,flow_,impl;
      SCode.Attributes attr;
      Absyn_SubscriptList ad;
      SCode.Accessibility acc;
      SCode.Variability param;
      Absyn.Direction dir;
      Absyn.Path t;
      SCode.Mod m;
      Absyn_PathOption bc;
      Absyn_CommentOption comment;
      TupleSCode_ElementModList xs,comps;
      InstDims inst_dims;
    case (env,mods,pre,csets,ci_state,((comp = SCode.COMPONENT(n,final_,repl,prot,(attr = SCode.ATTR(ad,flow_,acc,param,dir)),t,m,bc,comment)),cmod) :: xs,inst_dims,impl)
      equation 
        compmod = Mod.lookup_comp_modification(mods, n) " PA: PROBLEM, Modifiers should be merged in this phase, but
	   since undeclared components can not be found (is done in this phase)
	   the modifiers can not be elaborated to get a variable binding.
	   Thus, we need to store the merged modifier for elaboration in the 
	   next stage. 
	   Solution: Save all modifiers in environment...
	   * Use type T_NOTYPE instead of as earier trying to instantiate, since instanitation might fail without having correct
	 modifications (e.g. when instanitating a partial class that must be redeclared through a modification)";
        cmod_1 = Mod.merge(compmod, cmod, env, pre);
        env_1 = Env.extend_frame_v(env, 
          Types.VAR(n,Types.ATTR(flow_,acc,param,dir),prot,
          (Types.T_NOTYPE,NONE),Types.UNBOUND), SOME((comp,cmod_1)), false);
        env_2 = add_components_to_env2(env_1, mods, pre, csets, ci_state, xs, inst_dims, impl); then env_2;
    case (env,_,_,_,_,{},_,_) then env; 
    case (env,_,_,_,_,comps,_,_)
      equation 
        Debug.fprint("failtrace", "- add_components_to_env2 failed\n");
        Debug.fprint("failtrace", "\n\n"); then fail();
  end matchcontinue;
end add_components_to_env2;

protected function get_crefs_from_compdims " relation: get_crefs_from_compdims
  author: PA
 
  This function collects all variables from the dimensionalities of component elements.
  These variables are candidates for structural parameters.
 "
  input TupleSCode_ElementTypes_ModList in_tuplescode_elementtypes_modlist;
  output Absyn_ComponentRefList out_absyn_componentreflist;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type TupleSCode_ElementTypes_ModList = list<TupleSCode_ElementTypes_Mod>;
  type Absyn_ComponentRefList = list<Absyn.ComponentRef>;
algorithm 
  out_absyn_componentreflist:=
  matchcontinue (in_tuplescode_elementtypes_modlist)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type TupleSCode_ElementMod = tuple<SCode.Element,Mod>;
      type TupleSCode_ElementModList = list<TupleSCode_ElementMod>;
      Absyn_ComponentRefList crefs1,crefs2,crefs;
      Absyn_SubscriptList arraydim;
      TupleSCode_ElementModList xs;
    case ({}) then {}; 
    case ((SCode.COMPONENT(_,_,_,_,SCode.ATTR(arraydim,_,_,_,_),_,_,_,_),_) :: xs)
      equation 
        crefs1 = get_cref_from_dim(arraydim);
        crefs2 = get_crefs_from_compdims(xs) "Util.list_map(crefs1,Absyn.cref_get_first) => crefs1' &";
        crefs = list_append(crefs1, crefs2) "Util.list_map(crefs2,Absyn.cref_get_first) => crefs2' &"; then crefs;
    case (_ :: xs)
      equation 
        crefs = get_crefs_from_compdims(xs); then crefs;
  end matchcontinue;
end get_crefs_from_compdims;

protected function member_crefs " relation member_cref
  author: PA
 
  This function checks if a componentreferece is a member of a list of component references, 
  disregarding subscripts.
 "
  input Absyn.ComponentRef in_componentref;
  input Absyn_ComponentRefList in_absyn_componentreflist;
  output Boolean out_boolean;
protected 
  type Absyn_ComponentRefList = list<Absyn.ComponentRef>;
algorithm 
  out_boolean:=
  matchcontinue (in_componentref,in_absyn_componentreflist)
    local
      Absyn.ComponentRef cr,cr1;
      Absyn_ComponentRefList xs;
      Boolean res;
    case (cr,cr1 :: xs)
      equation 
        true = Absyn.cref_equal(cr, cr1); then true;
    case (cr,cr1 :: xs)
      equation 
        false = Absyn.cref_equal(cr, cr1);
        res = member_crefs(cr, xs); then res;
    case (_,_) then false; 
  end matchcontinue;
end member_crefs;

protected function inst_element
  input Env in_env1;
  input Mod in_mod2;
  input Prefix in_prefix3;
  input Connect.Sets in_sets4;
  input ClassInf.State in_state5;
  input TupleSCode_ElementTypes_Mod in_tuplescode_elementtypes_mod6;
  input InstDims in_instdims7;
  input Boolean in_boolean8;
  input Boolean in_boolean9;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
  output Types_VarList out_types_varlist;
protected 
  type TupleSCode_ElementTypes_Mod = tuple<SCode.Element,Types.Mod>;
  type DAE_ElementList = list<DAE.Element>;
  type Types_VarList = list<Types.Var>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state,out_types_varlist):=
  matchcontinue (in_env1,in_mod2,in_prefix3,in_sets4,in_state5,in_tuplescode_elementtypes_mod6,in_instdims7,in_boolean8,in_boolean9)
    case (env,mod,pre,csets,ci_state,(SCode.IMPORT(imp),_),instdims,_,_) " imports
	   ** imports are simply added to the current frame, so that the lookup rule can find them.
	   *"
      equation 
        env_1 = Env.extend_frame_i(env, imp); then ({},env_1,csets,ci_state,{});
    case (env,mods,pre,csets,ci_state,(SCode.COMPONENT(n,final_,repl,prot,_,_,_,_,_),_),_,_,_) "If a variable is declared multiple times, the first is used "
      equation 
        (_,NONE,true) = Lookup.lookup_ident_local(env, n); then ({},env,csets,ci_state,{});
    case (env,mods,pre,csets,ci_state,(SCode.CLASSDEF(n,_,_,_,_),_),_,_,_) "Illegal redeclarations "
      equation 
        (_,_,_) = Lookup.lookup_ident_local(env, n);
        Print.print_error_buf("# Trying to redeclare the class ");
        Print.print_error_buf(n);
        Print.print_error_buf(" as a variable\n"); then fail();
    case (env,mods,pre,csets,ci_state,(SCode.CLASSDEF(n,_,true,c,_),_),inst_dims,impl,packimpl) "  A new class definition
	   **
	   ** Put it in the current frame in the environment
	   *"
      equation 
        (classmod = Types.REDECL(final_,{(SCode.CLASSDEF(n2,f2,repl2,cls2,_),_)})) = Mod.lookup_modification_p(mods, Absyn.IDENT(n)) "Redeclare of class definition, replaceable is true ";
        (env_1,dae) = inst_class_decl(env, classmod, pre, csets, cls2, inst_dims, packimpl) "Debug.fprintln (\"insttr\", \"*--*Classdef mods\") &
	Debug.fcall (\"insttr\", Mod.print_mod, classmod) &
	Debug.fprintln (\"insttr\", \"*--*All mods\") &
	Debug.fcall (\"insttr\", Mod.print_mod, mods) &"; then (dae,env_1,csets,ci_state,{});
    case (env,mods,pre,csets,ci_state,(SCode.CLASSDEF(n,_,false,c,_),_),inst_dims,impl,packimpl)
      equation 
        (classmod = Types.REDECL(final_,{(SCode.CLASSDEF(n2,f2,repl2,cls2,_),_)})) = Mod.lookup_modification_p(mods, Absyn.IDENT(n)) "Redeclare of class definition, replaceable is false ";
        Print.print_error_buf("# Trying to redeclare class ") "Redeclare of class definition, replaceable is false ";
        Print.print_error_buf(n);
        Print.print_error_buf(" but class not declared as replacable.\n"); then fail();
    case (env,mods,pre,csets,ci_state,(SCode.CLASSDEF(n,_,_,c,_),_),inst_dims,impl,packimpl)
      equation 
        classmod = Mod.lookup_modification_p(mods, Absyn.IDENT(n)) "Classdefinition without redeclaration ";
        (env_1,dae) = inst_class_decl(env, classmod, pre, csets, c, inst_dims, packimpl) "Debug.fprintln (\"insttr\", \"**Classdef mods\") &
	Debug.fcall (\"insttr\", Mod.print_mod, classmod) &
	Debug.fprintln (\"insttr\", \"**All mods\") &
	Debug.fcall (\"insttr\", Mod.print_mod, mods) &"; then (dae,env_1,csets,ci_state,{});
    case (env,mods,pre,csets,ci_state,((comp = SCode.COMPONENT(n,final_,repl,prot,(attr = SCode.ATTR(ad,flow_,acc,param,dir)),t,m,bc,comment)),cmod),inst_dims,impl,packimpl) "  A component
	   **
	   ** This is the rule for instantiating a model component.  A
	   ** component can be a structured subcomponent or a variable,
	   ** parameter or constant.  All of these are treated in a
	   ** similar way.
	   **
	   ** Lookup the class name, apply modifications and add the
	   ** variable to the current frame in the environment. Then
	   ** instantiate the class with an extended prefix.
	   *"
      equation 
        vn = Prefix.prefix_cref(pre, Exp.CREF_IDENT(n,{})) "Debug.fprint(\"insttr\", \"Instantiating component \") &
	Debug.fprint(\"insttr\", n) & Debug.fprint(\"insttr\", \"\n\") &";
        classmod = Mod.lookup_modification_p(mods, t) " The class definition is fetched from the environment.   Then the set of modifications is calculated.  The   modificions is the result of merging the modifications   from several sources.  The modification stored with the   class definition is put in the variable `classmod', the   modification passed to the function_ is extracted and put   in the variable `mm', and the modification that is   included in the variable declaration is in the variable   `m'.  All of these are merged so that the correct   precedence rules are followed.  Debug.fprint (\"insttr\", \"Looking up class modifications\") &";
        mm = Mod.lookup_comp_modification(mods, n) "Debug.fcall (\"insttr\", Mod.print_mod, classmod) & Debug.fprint (\"insttr\", \"\nLooking up component modifications\") &";
        owncref = Absyn.CREF_IDENT(n,{}) "Debug.fcall (\"insttr\", Mod.print_mod, mm) & Debug.fprint (\"insttr\", \"\n\") & The types in the environment does not have correct Binding.
	   We must update those variables that is found in m into a new environment.";
        crefs = get_cref_from_mod(m);
        crefs2 = get_cref_from_dim(ad);
        crefs_1 = Util.list_flatten({crefs,crefs2});
        crefs_2 = remove_cref_from_crefs(crefs_1, owncref);
        env2 = update_components_in_env(mods, crefs_2, env, ci_state, impl);
        mods_1 = Mod.update_mod(env2, pre, mods, impl) "Debug.fprint(\"insttr\", \"updated variables in env\n\") & Update the untyped modifiers to typed ones, and extract class  and component modifiers again.";
        classmod_1 = Mod.lookup_modification_p(mods_1, t) "Debug.fprint(\"insttr\", \"updated modifiers to typed ones: \") &
	Debug.fcall(\"insttr\", Mod.print_mod, mods') &
	Debug.fprint(\"insttr\", \"\n\") & ";
        mm_1 = Mod.lookup_comp_modification(mods_1, n);
        m_1 = Mod.elab_mod(env2, pre, m, impl) "Debug.fprintln (\"insttr\", \"Elaborating current modifications\n env\") &
	Debug.fcall (\"insttr\", Env.print_env, env2)  &
	Debug.fprintln (\"insttr\", \"\n\n\") &";
        mod = Mod.merge(classmod_1, mm_1, env2, pre) "Debug.fcall (\"insttr\", Mod.print_mod, m') &";
        mod1 = Mod.merge(mod, m_1, env2, pre);
        mod1_1 = Mod.merge(cmod, mod1, env2, pre);
        (SCode.COMPONENT(n,final_,repl,prot,attr = SCode.ATTR(ad,flow_,acc,param,dir),t,m,bc,comment),mod_1,env2_1) = redeclare_type(mod1_1, comp, env2, pre, ci_state, impl);
        env_1 = get_derived_env(env, bc) "	Debug.fprint(\"insttr\", \"looking for type \") &
	Debug.fcall(\"insttr\",Dump.print_path,t) &
	Debug.fprint(\"insttr\",\" \n\") &";
        (cl,cenv) = Lookup.lookup_class(env_1, t, true);
        check_prot(prot, mm_1, vn) "	Debug.fprintl (\"insttr\", {\"found type for: \", n, \"\n\"}) &  If the element is `protected', and an external   modification is applied, it is an error. ";
        Debug.fcall("myeqmod", Mod.print_mod, mod_1) "	Debug.fprintln (\"insttr\", \"Protection checked\") &";
        Debug.fprintln("myeqmod", "\n") "	Debug.fprintln (\"insttr\", \"Protection checked\") &";
        eq = Mod.mod_equation(mod_1);
        dims = elab_arraydim(env2_1, owncref, ad, eq, impl, NONE) "	Debug.fprintln (\"insttr\", \"mod equation done, from mod:\") &
	Debug.fcall(\"insttr\",Mod.print_mod, mod') & Debug.fprintln (\"insttr\", \"\n\") &  The variable declaration and the (optional) equation   modification are inspected for array dimensions. ";
        (dae,csets_1,ty) = inst_var(cenv, ci_state, mod_1, pre, csets, n, cl, attr, dims, {}, 
          inst_dims, impl, comment) "	Debug.fprintln (\"insttr\", \"arraydim elabed\") &
	Debug.fprintln (\"arraydim\", \"arraydim elabed :\") &
	Debug.fcall(\"arraydim\",print_dim2, dims) &  Instantiate the component  Dump.select_string(impl,\"IMPLICIT!!!\n\",\"\") => s &
	print s &";
        binding = make_binding(env2_1, attr, eq) "	Debug.fprintl (\"insttr\", {\"instantiated: \", n, \"\n\"}) &  The environment is extended (updated) with the new variable   binding. ";
        env_1 = Env.update_frame_v(env2_1, 
          Types.VAR(n,Types.ATTR(flow_,acc,param,dir),prot,ty,binding), true) "	Debug.fcall (\"insttr\", Types.print_binding, binding) &
	Debug.fcall (\"insttr\", print, \"\n\") & true in update_frame means the variable is now instantiated.  type info present Now we can also put the binding into the dae   If the type is one of the simple, predifined types a   simple variable declaration is added to the DAE.  & Debug.fprint(\"insttr\",\"inst_element Component succeeded\n\") "; then (dae,env_1,csets_1,ci_state,{
          Types.VAR(n,Types.ATTR(flow_,acc,param,dir),prot,ty,binding)});
    case (env,_,pre,csets,ci_state,(SCode.COMPONENT(n,final_,repl,prot,_,t,_,_,_),_),_,_,_) " If the class lookup in the previous rule fails, this
	 ** rule catches the error and prints an error message about
	 ** the unknown class. 
	 *   Failure          => ({},env,csets,ci_state,{}) "
      equation 
        failure((cl,cenv) = Lookup.lookup_class(env, t, false));
        s = Absyn.path_string(t);
        Print.print_error_buf("# unknown class '");
        Print.print_error_buf(s);
        Print.print_error_buf("' while instantiating ");
        pre_1 = Prefix.prefix_add(n, {}, pre);
        ns = Prefix.print_prefix_str(pre_1);
        Print.print_error_buf(ns);
        Print.print_error_buf("\n"); then fail();
    case (_,_,_,_,_,(el,_),_,_,_) "        => ({},env,csets,ci_state,{}) "
      equation 
        Debug.fprint("failtrace", "- inst_element failed\n");
        Debug.fcall("failtrace", SCode.print_element, el);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_element;

protected function get_derived_env " relation: get_derived_env
 
  This function returns the environment of a baseclass.
  It is used when instantiating a component defined in a baseclass.
"
  input Env.Env in_env;
  input Absyn_PathOption in_absyn_pathoption;
  output Env.Env out_env;
protected 
  type Absyn_PathOption = Option<Absyn.Path>;
algorithm 
  out_env:=
  matchcontinue (in_env,in_absyn_pathoption)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList env,env_1;
      Absyn.Path tp;
    case (env,NONE) then env; 
    case (env,SOME(tp))
      equation 
        (_,env_1) = Lookup.lookup_class(env, tp, true); then env_1;
  end matchcontinue;
end get_derived_env;

protected function remove_cref_from_crefs
  input Absyn_ComponentRefList in_absyn_componentreflist;
  input Absyn.ComponentRef in_componentref;
  output Absyn_ComponentRefList out_absyn_componentreflist;
protected 
  type Absyn_ComponentRefList = list<Absyn.ComponentRef>;
algorithm 
  out_absyn_componentreflist:=
  matchcontinue (in_absyn_componentreflist,in_componentref)
    local
      String n1,n2;
      Absyn_ComponentRefList rest_1,rest;
      Absyn.ComponentRef cr1,cr2;
    case ({},_) then {}; 
    case (cr1 :: rest,cr2)
      equation 
        Absyn.CREF_IDENT(n1,{}) = cr1;
        Absyn.CREF_IDENT(n2,{}) = cr2;
        equality(n1 = n2);
        rest_1 = remove_cref_from_crefs(rest, cr2); then rest_1;
    case (cr1 :: rest,cr2)
      equation 
        rest_1 = remove_cref_from_crefs(rest, cr2); then cr1 :: rest_1;
  end matchcontinue;
end remove_cref_from_crefs;

protected function redeclare_type "Implicit instantation "
  input Types.Mod in_mod;
  input SCode.Element in_element;
  input Env.Env in_env;
  input Prefix.Prefix in_prefix;
  input ClassInf.State in_state;
  input Boolean in_boolean;
  output SCode.Element out_element;
  output Types.Mod out_mod;
  output Env.Env out_env;
algorithm 
  (out_element,out_mod,out_env):=
  matchcontinue (in_mod,in_element,in_env,in_prefix,in_state,in_boolean)
    case ((m = Types.REDECL(_,((redecl = SCode.COMPONENT(n1,final_,repl,prot,_,t,mod,bc,comment)),rmod) :: rest)),SCode.COMPONENT(n2,false,repl2,prot2,_,t2,_,_,_),env,pre,ci_state,impl) "Implicit instantation "
      equation 
        equality(n1 = n2);
        crefs = get_cref_from_mod(mod);
        env_1 = update_components_in_env(Types.NOMOD, crefs, env, ci_state, impl) "m ";
        m_1 = Mod.elab_mod(env_1, pre, mod, impl);
        m_2 = Mod.merge(rmod, m_1, env_1, pre); then (redecl,m_2,env_1);
    case ((mod = Types.REDECL(redfin,((redecl = SCode.COMPONENT(n1,final_,repl,prot,_,t,_,bc,comment)),rmod) :: rest)),(comp = SCode.COMPONENT(n2,false,repl2,prot2,_,t2,_,_,comment2)),env,pre,ci_state,impl)
      equation 
        failure(equality(n1 = n2));
        (newcomp,mod,env_1) = redeclare_type(Types.REDECL(redfin,rest), comp, env, pre, ci_state, impl); then (newcomp,mod,env_1);
    case (Types.REDECL(redfin,_ :: rest),comp,env,pre,ci_state,impl)
      equation 
        (newcomp,mod,env_1) = redeclare_type(Types.REDECL(redfin,rest), comp, env, pre, ci_state, impl); then (newcomp,mod,env_1);
    case (Types.REDECL(redfin,{}),comp,env,pre,ci_state,impl) then (comp,Types.NOMOD,env); 
    case (mod,comp,env,pre,ci_state,impl) then (comp,mod,env); 
    case (_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- redeclare_type failed\n"); then fail();
  end matchcontinue;
end redeclare_type;

protected function inst_var "impl"
  input Env.Env in_env;
  input ClassInf.State in_state;
  input Types.Mod in_mod;
  input Prefix.Prefix in_prefix;
  input Connect.Sets in_sets;
  input Ident in_ident;
  input SCode.Class in_class;
  input SCode.Attributes in_attributes;
  input DimExpList in_dimexplist;
  input IntegerList in_integerlist;
  input InstDims in_instdims;
  input Boolean in_boolean;
  input Absyn_CommentOption in_absyn_commentoption;
  output DAE_ElementList out_dae_elementlist;
  output Connect.Sets out_sets;
  output Types.Type out_type;
protected 
  type DimExpList = list<DimExp>;
  type IntegerList = list<Integer>;
  type Absyn_CommentOption = Option<Absyn.Comment>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_sets,out_type):=
  matchcontinue (in_env,in_state,in_mod,in_prefix,in_sets,in_ident,in_class,in_attributes,in_dimexplist,in_integerlist,in_instdims,in_boolean,in_absyn_commentoption)
    case (env,ci_state,mod,pre,csets,n,(cl = SCode.CLASS(id,_,_,SCode.R_TYPE,_)),attr,dims,idxs,inst_dims,impl,comment) "impl"
      equation 
        (dims_1 = _ :: _) = get_usertype_dimensions(env, mod, pre, cl, inst_dims, impl) "Variables of userdefined type, e.g. Point p => Real p{3};  These must be handled separately since even if they do not appear to 
	 be an array, they can. Therefore we need to collect the full dimensionality
	 and call inst_var2  Collect dimensions ";
        (dae,csets_1,ty_1) = inst_var2(env, ci_state, mod, pre, csets, n, cl, attr, dims_1, idxs, 
          inst_dims, impl, comment);
        ty = make_array_type(dims_1, ty_1); then (dae,csets_1,ty);
    case (env,ci_state,mod,pre,csets,n,(cl = SCode.CLASS(id,_,_,_,_)),attr,dims,idxs,inst_dims,impl,comment)
      equation 
        (dae,csets_1,ty_1) = inst_var2(env, ci_state, mod, pre, csets, n, cl, attr, dims, idxs, 
          inst_dims, impl, comment) "Generic case: fall trough "; then (dae,csets_1,ty_1);
  end matchcontinue;
end inst_var;

protected function inst_var2 "True = Implicit instantiation, should only be 
			    used when instantiating functions. Then we can 
			    not always find dimensional sizes. e.g. 
			    input Real x{:}; "
  input Env.Env in_env;
  input ClassInf.State in_state;
  input Types.Mod in_mod;
  input Prefix.Prefix in_prefix;
  input Connect.Sets in_sets;
  input Ident in_ident;
  input SCode.Class in_class;
  input SCode.Attributes in_attributes;
  input DimExpList in_dimexplist;
  input IntegerList in_integerlist;
  input InstDims in_instdims;
  input Boolean in_boolean;
  input Absyn_CommentOption in_absyn_commentoption;
  output DAE_ElementList out_dae_elementlist;
  output Connect.Sets out_sets;
  output Types.Type out_type;
protected 
  type DimExpList = list<DimExp>;
  type IntegerList = list<Integer>;
  type Absyn_CommentOption = Option<Absyn.Comment>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_sets,out_type):=
  matchcontinue (in_env,in_state,in_mod,in_prefix,in_sets,in_ident,in_class,in_attributes,in_dimexplist,in_integerlist,in_instdims,in_boolean,in_absyn_commentoption)
    case (env,ci_state,(mod = Types.MOD(_,_,SOME(Types.TYPED(e,_)))),pre,csets,n,cl,SCode.ATTR(_,flow_,acc,(vt = SCode.CONST),dir),{},idxs,inst_dims,impl,comment) "True = Implicit instantiation, should only be 
			    used when instantiating functions. Then we can 
			    not always find dimensional sizes. e.g. 
			    input Real x{:};   The class is instantiated with the calculated   modification, and an extended prefix.   LS: Removed the part which checks if modelica_output is true
	 ** and generates variables with initialization expression from the
	 ** modifications, because it cannot handle right hand side which is a
	 ** component (T_COMPLEX) anyway. This case is handled by the rule below
	 ** which generates correct equations according to the modification.
	 ** Separate code can parse the DAE and put the rhs of the latest
	 ** equation inside the variable declaration, and discard all the
	 ** equations.
	 *  Rules for normal instantiation, will resolv dimensional sizes, etc.  as false"
      equation 
        idxs_1 = list_reverse(idxs) "Constants ";
        pre_1 = Prefix.prefix_add(n, idxs_1, pre);
        (dae1,env_1,csets_1,ty,st) = inst_class(env, mod, pre_1, csets, cl, inst_dims, impl, INNER_CALL);
        dae1_1 = fix_direction(dae1, dir);
        subs = Exp.int_subscripts(idxs_1);
        cr = Prefix.prefix_cref(pre, Exp.CREF_IDENT(n,subs));
        dae3 = dae_declare(cr, ci_state, ty, SCode.ATTR({},flow_,acc,vt,dir), 
          SOME(e), inst_dims, NONE, NONE, comment) "inst_mod_equation(cr,ty,mod) => dae2 &";
        dae = list_append(dae1_1, dae3) "	& Debug.fcall (\"instvardae\",DAE.dump2,DAE.DAE(dae))"; then (dae,csets_1,ty);
    case (env,ci_state,(mod = Types.MOD(_,_,SOME(Types.TYPED(e,_)))),pre,csets,n,cl,SCode.ATTR(_,flow_,acc,(vt = SCode.PARAM),dir),{},idxs,inst_dims,impl,comment) "as false"
      equation 
        idxs_1 = list_reverse(idxs) "Parameters ";
        pre_1 = Prefix.prefix_add(n, idxs_1, pre);
        (dae1,env_1,csets_1,ty,st) = inst_class(env, mod, pre_1, csets, cl, inst_dims, impl, INNER_CALL);
        dae1_1 = fix_direction(dae1, dir);
        subs = Exp.int_subscripts(idxs_1);
        cr = Prefix.prefix_cref(pre, Exp.CREF_IDENT(n,subs));
        dae3 = dae_declare(cr, ci_state, ty, SCode.ATTR({},flow_,acc,vt,dir), 
          SOME(e), inst_dims, NONE, NONE, comment) "	inst_mod_equation(cr,ty,mod) => dae2 &";
        dae = list_append(dae1_1, dae3) "	& Debug.fcall (\"instvardae\",DAE.dump2,DAE.DAE(dae))"; then (dae,csets_1,ty);
    case (env,ci_state,mod,pre,csets,n,cl,SCode.ATTR(_,flow_,acc,vt,dir),{},idxs,inst_dims,impl,comment)
      equation 
        idxs_1 = list_reverse(idxs) "Scalar Variables ";
        pre_1 = Prefix.prefix_add(n, idxs_1, pre);
        prefix_str = Prefix.print_prefix_str(pre_1) "This is where redeclare should be handled ";
        Debug.fprintl("insttr", 
          {"instantiating var class: ",n," prefix ",prefix_str,"\n"});
        (dae1,_,csets_1,ty,st) = inst_class(env, mod, pre_1, csets, cl, inst_dims, impl, INNER_CALL) "Debug.fcall (\"envprint\", Env.print_env, env) & ";
        dae1_1 = fix_direction(dae1, dir);
        subs = Exp.int_subscripts(idxs_1) "& Debug.fcall (\"insttr\",DAE.dump2,DAE.DAE(dae'))";
        cr = Prefix.prefix_cref(pre, Exp.CREF_IDENT(n,subs));
        dae2 = inst_mod_equation(cr, ty, mod, impl);
        index_string = Util.list_map(idxs_1, int_string);
        Debug.fprint("insttrind", "\n ******************\n ");
        Debug.fprint("insttrind", "\n index_string ");
        Debug.fprintl("insttr", index_string);
        Debug.fprint("insttrind", "\n component ref ");
        Debug.fcall("insttr", Exp.print_component_ref, cr);
        Debug.fprint("insttrind", "\n ******************\n ");
        Debug.fprint("insttrind", "\n ");
        start = inst_start_binding_exp(mod, ty, idxs_1);
        dae_var_attr = inst_dae_variable_attributes(mod, ty, idxs_1);
        str_tmp = DAE.dump_variable_attributes_str(dae_var_attr);
        Debug.fprintln("failtrace", str_tmp);
        dae3 = dae_declare(cr, ci_state, ty, SCode.ATTR({},flow_,acc,vt,dir), NONE, 
          inst_dims, start, dae_var_attr, comment);
        daex = list_append(dae1_1, dae2) "         dae_declare(cr,ty,SCode.ATTR({}, flow, acc, vt, dir), NONE,    		            inst_dims, start, NONE) => dae3 &  ";
        dae = list_append(daex, dae3) "	& Debug.fcall (\"instvardae\",DAE.dump2,DAE.DAE(dae))
	 & Debug.fprint(\"instvartr\", \"\nEnd of inst_var variables\n\")"; then (dae,csets_1,ty);
    case (env,(ci_state = ClassInf.FUNCTION(_)),mod,pre,csets,n,cl,attr,(dims = _ :: _),idxs,inst_dims,impl,comment)
      equation 
        dims_1 = inst_dim_exp_lst(dims, impl) "Array variables with binding in functions , e.g. input Real x{:}=Y  Do not flatten because it is a function ";
        SOME(Types.TYPED(e,p)) = Mod.mod_equation(mod) "get the equation modification ";
        (_,_,_,ty,st) = inst_class(env, mod, pre, csets, cl, inst_dims, impl, INNER_CALL) "Instantiate type of the component ";
        ty_1 = make_array_type(dims, ty) "Make it an array type since we are not flattening ";
        cr = Prefix.prefix_cref(pre, Exp.CREF_IDENT(n,{}));
        ty_2 = Types.elab_type(ty_1);
        e_1 = Types.match_prop(e, Types.PROP(ty_1,false), p);
        daeeq = make_dae_equation(Exp.CREF(cr,ty_2), e_1, NON_INITIAL) "Put the mod equation in the dae so that code will be generated ";
        dae1 = dae_declare(cr, ci_state, ty, attr, NONE, dims_1, NONE, NONE, comment);
        dae = list_append(dae1, {daeeq}); then (dae,csets,ty_1);
    case (env,(ci_state = ClassInf.FUNCTION(_)),mod,pre,csets,n,cl,attr,(dims = _ :: _),idxs,inst_dims,impl,comment)
      equation 
        (_,_,_,ty,st) = inst_class(env, mod, pre, csets, cl, inst_dims, impl, INNER_CALL) "Array variables without binding in functions , e.g. input Real x{:}  Do not flatten because it is a function ";
        cr = Prefix.prefix_cref(pre, Exp.CREF_IDENT(n,{}));
        dims_1 = inst_dim_exp_lst(dims, impl) "Do all dimensions... ";
        dae = dae_declare(cr, ci_state, ty, attr, NONE, dims_1, NONE, NONE, comment);
        arrty = make_array_type(dims, ty); then (dae,csets,arrty);
    case (env,ci_state,mod,pre,csets,n,cl,attr,dim :: dims,idxs,inst_dims,impl,comment) "FIXME: make a similar rule: if implicit=true and we fail to flatten, we should leave it unflattened "
      equation 
        dime = inst_dim_exp(dim, impl) "Array variables , e.g. Real x{3} flatten ";
        inst_dims_1 = list_append(inst_dims, {dime});
        (dae,csets_1,ty) = inst_array(env, ci_state, mod, pre, csets, n, (cl,attr), 1, dim, 
          dims, idxs, inst_dims_1, impl, comment);
        dimt = inst_dim_type(dim);
        ty_1 = Types.lift_array(ty, dimt); then (dae,csets_1,ty_1);
    case (_,_,_,_,_,n,_,_,_,_,_,_,_) "Rules for instantation of function variables (e.g. input and output 
  parameters and protected variables)  
rule	Print.print_buf \"!!! Error, implicit instantiation of variables\n\"
	-------------------------------
	inst_var2(_,_,_,_,_,_,_,_,_,_,_,impl as true) => fail
"
      equation 
        Debug.fprint("failtrace", "- inst_var2 failed: ");
        Debug.fprint("failtrace", n);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_var2;

protected function make_array_type
  input DimExpList in_dimexplist;
  input Types.Type in_type;
  output Types.Type out_type;
protected 
  type DimExpList = list<DimExp>;
algorithm 
  out_type:=
  matchcontinue (in_dimexplist,in_type)
    case ({},ty) then ty; 
    case (DIMINT(i) :: xs,(ty,p))
      equation 
        ty_1 = make_array_type(xs, (ty,p)); then ((Types.T_ARRAY(Types.DIM(SOME(i)),ty_1),p));
    case (DIMEXP(_,_) :: xs,(ty,p))
      equation 
        ty_1 = make_array_type(xs, (ty,p)); then ((Types.T_ARRAY(Types.DIM(NONE),ty_1),p));
    case (_,_)
      equation 
        Debug.fprint("failtrace", "- make_array_type failed\n"); then fail();
  end matchcontinue;
end make_array_type;

protected function get_usertype_dimensions "impl"
  input Env.Env in_env;
  input Types.Mod in_mod;
  input Prefix.Prefix in_prefix;
  input SCode.Class in_class;
  input InstDims in_instdims;
  input Boolean in_boolean;
  output DimExpList out_dimexplist;
protected 
  type DimExpList = list<DimExp>;
algorithm 
  out_dimexplist:=
  matchcontinue (in_env,in_mod,in_prefix,in_class,in_instdims,in_boolean)
    case (_,_,_,SCode.CLASS("Real",_,_,_,_),_,_) then {}; 
    case (_,_,_,SCode.CLASS("Integer",_,_,_,_),_,_) then {}; 
    case (_,_,_,SCode.CLASS("String",_,_,_,_),_,_) then {}; 
    case (_,_,_,SCode.CLASS("Boolean",_,_,_,_),_,_) then {}; 
    case (env,mods,pre,SCode.CLASS(id,_,_,SCode.R_TYPE,SCode.DERIVED(cn,ad,mod)),dims,impl)
      equation 
        (cl,cenv) = Lookup.lookup_class(env, cn, true);
        owncref = Absyn.CREF_IDENT(id,{});
        ad_1 = get_option_arraydim(ad);
        mod_1 = Mod.elab_mod(env, pre, mod, impl);
        mods_2 = Mod.merge(mods, mod_1, env, pre);
        eq = Mod.mod_equation(mods_2);
        mods_3 = Mod.lookup_comp_modification(mods_2, id);
        dim1 = get_usertype_dimensions(cenv, mods_3, pre, cl, dims, impl);
        dim2 = elab_arraydim(env, owncref, ad_1, eq, impl, NONE);
        res = list_append(dim2, dim1); then res;
    case (_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- get_usertype_dimensions failed\n"); then fail();
  end matchcontinue;
end get_usertype_dimensions;

protected function get_cref_from_mod " relation: get_cref_from_mod
  author: PA
 
  Return all variables in a modifier, SCode.Mod.
  This is needed to prepare the second pass of instantiation, because a component can not be 
  instantiated unless the types of the modifiers are known. Therefore the variables in all 
  modifiers must be instantiated before the component itself is instantiated. This is done by 
  backpatching in the instantiation process. NOTE: This means that a recursive modification 
  structure (which is not allowed in Modelica) will currently run the compiler into
  infinite recursion.
 "
  input SCode.Mod in_mod;
  output Absyn_ComponentRefList out_absyn_componentreflist;
protected 
  type Absyn_ComponentRefList = list<Absyn.ComponentRef>;
algorithm 
  out_absyn_componentreflist:=
  matchcontinue (in_mod)
    local
      type SCode_ElementList = list<SCode.Element>;
      type SCode_SubModList = list<SCode.SubMod>;
      Absyn_ComponentRefList res1,res2,res,l1,l2;
      Boolean b;
      String n;
      SCode.Mod m,mod;
      SCode_ElementList xs;
      SCode_SubModList submods;
      Absyn.Exp e;
    case (SCode.REDECL(b,SCode.COMPONENT(n,_,_,_,_,_,m,_,_) :: xs))
      equation 
        res1 = get_cref_from_mod(SCode.REDECL(b,xs)) "For redeclarations e.g \"redeclare B2 b(cref=<expr>)\", find cref ";
        res2 = get_cref_from_mod(m);
        res = list_append(res1, res2) "&Util.list_map(res,Absyn.cref_get_first) => res'"; then res;
    case (SCode.REDECL(b,_ :: xs))
      equation 
        res = get_cref_from_mod(SCode.REDECL(b,xs)) "For redeclarations e.g \"redeclare B2 b(cref=<expr>)\", find cref "; then res;
    case (SCode.REDECL(b,{})) then {}; 
    case ((mod = SCode.MOD(_,submods,SOME(e))))
      equation 
        l1 = get_cref_from_submods(submods) "Find in sub modifications e.g A(B=3) find B ";
        l2 = Absyn.get_cref_from_exp(e);
        res = list_append(l1, l2); then res;
    case (SCode.MOD(_,submods,NONE))
      equation 
        res = get_cref_from_submods(submods); then res;
    case (_) then {}; 
    case (_)
      equation 
        Debug.fprint("failtrace", "- get_cref_from_mod failed\n"); then fail();
  end matchcontinue;
end get_cref_from_mod;

protected function get_cref_from_dim " relation: get_cref_from_dim
  author: PA
 
  Similar to get_cref_from_mod, but investigates array dimensionalitites instead.
 "
  input Absyn.ArrayDim in_arraydim;
  output Absyn_ComponentRefList out_absyn_componentreflist;
protected 
  type Absyn_ComponentRefList = list<Absyn.ComponentRef>;
algorithm 
  out_absyn_componentreflist:=
  matchcontinue (in_arraydim)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      Absyn_ComponentRefList l1,l2,res;
      Absyn.Exp exp;
      Absyn_SubscriptList rest;
    case (Absyn.SUBSCRIPT(exp) :: rest)
      equation 
        l1 = get_cref_from_dim(rest);
        l2 = Absyn.get_cref_from_exp(exp);
        res = list_append(l1, l2); then res;
    case (Absyn.NOSUB :: rest)
      equation 
        res = get_cref_from_dim(rest); then res;
    case ({}) then {}; 
    case (_)
      equation 
        Debug.fprint("failtrace", "- get_cref_from_dim failed\n"); then fail();
  end matchcontinue;
end get_cref_from_dim;

protected function get_cref_from_submods " relation: get_cref_from_submods
 
  Helper function to get_cref_from_mod, investigates sub modifiers.
 "
  input SCode_SubModList in_scode_submodlist;
  output Absyn_ComponentRefList out_absyn_componentreflist;
protected 
  type SCode_SubModList = list<SCode.SubMod>;
  type Absyn_ComponentRefList = list<Absyn.ComponentRef>;
algorithm 
  out_absyn_componentreflist:=
  matchcontinue (in_scode_submodlist)
    local
      Absyn_ComponentRefList res1,res2,res;
      SCode.Mod mod;
      SCode_SubModList rest;
    case (SCode.NAMEMOD(_,mod) :: rest)
      equation 
        res1 = get_cref_from_mod(mod);
        res2 = get_cref_from_submods(rest);
        res = list_append(res1, res2); then res;
    case ({}) then {}; 
  end matchcontinue;
end get_cref_from_submods;

protected function update_components_in_env "Implicit instantiation"
  input Types.Mod in_mod;
  input Absyn_ComponentRefList in_absyn_componentreflist;
  input Env.Env in_env;
  input ClassInf.State in_state;
  input Boolean in_boolean;
  output Env.Env out_env;
protected 
  type Absyn_ComponentRefList = list<Absyn.ComponentRef>;
algorithm 
  out_env:=
  matchcontinue (in_mod,in_absyn_componentreflist,in_env,in_state,in_boolean)
    case (mods,cr :: rest,env,ci_state,impl) "Implicit instantiation"
      equation 
        env_1 = update_component_in_env(mods, cr, env, ci_state, impl);
        env_2 = update_components_in_env(mods, rest, env_1, ci_state, impl); then env_2;
    case (_,{},env,ci_state,impl) then env; 
  end matchcontinue;
end update_components_in_env;

protected function update_component_in_env
  input Types.Mod in_mod;
  input Absyn.ComponentRef in_componentref;
  input Env.Env in_env;
  input ClassInf.State in_state;
  input Boolean in_boolean;
  output Env.Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_mod,in_componentref,in_env,in_state,in_boolean)
    case (mods,(cref = Absyn.CREF_IDENT(id,subscr)),env,ci_state,impl)
      equation 
        (ty,SOME((SCode.COMPONENT(n,final_,repl,prot,attr = SCode.ATTR(ad,flow_,acc,param,dir),t,m,bc,comment),cmod)),_) = Lookup.lookup_ident(env, id) "Variables that have Element in Environment, i.e. no type information are 
	 instnatiated here to get the type.";
        (cl,cenv) = Lookup.lookup_class(env, t, false);
        crefs = get_cref_from_mod(m);
        crefs2 = get_cref_from_dim(ad);
        crefs_1 = list_append(crefs, crefs2);
        crefs_2 = remove_cref_from_crefs(crefs_1, cref);
        env2 = update_components_in_env(mods, crefs_2, env, ci_state, impl);
        m_1 = Mod.elab_mod(env2, Prefix.NOPRE, m, impl) "Prefix does not matter, since we only update types in env, and does
	   not make any dae elements, etc.. ";
        classmod = Mod.lookup_modification_p(mods, t);
        mm = Mod.lookup_comp_modification(mods, n);
        mod = Mod.merge(classmod, mm, env2, Prefix.NOPRE);
        mod_1 = Mod.merge(mod, m_1, env2, Prefix.NOPRE);
        mod_2 = Mod.merge(cmod, mod_1, env2, Prefix.NOPRE);
        mod_3 = Mod.update_mod(env2, Prefix.NOPRE, mod_2, impl);
        eq = Mod.mod_equation(mod_3);
        dims = elab_arraydim(env2, cref, ad, eq, impl, NONE) " The variable declaration and the (optional) equation   modification are inspected for array dimensions. ";
        (dae1,csets_1,ty) = inst_var(cenv, ci_state, mod_3, Prefix.NOPRE, {}, n, cl, attr, 
          dims, {}, {}, impl, NONE) " Instantiate the component ";
        binding = make_binding(env2, attr, eq) " The environment is extended with the new variable   binding. ";
        (env2_1,binding_1) = check_structural_param_binding(param, binding, env2) "check if binding makes other variables into structural parameters";
        env_1 = Env.update_frame_v(env2_1, 
          Types.VAR(n,Types.ATTR(flow_,acc,param,dir),prot,ty,binding_1), false) "type info present 	& Debug.fprint(\"insttr\",\"Updated variable \") &
	Debug.fprint(\"insttr\", id) &
	Debug.fprint(\"insttr\",\" in env.\n \") "; then env_1;
    case (mods,(cref = Absyn.CREF_IDENT(id,subscr)),env,ci_state,impl)
      equation 
        (ty,NONE,_) = Lookup.lookup_ident(env, id) "Variables that already has type has no element. "; then env;
    case (mods,Absyn.CREF_QUAL(id,_,_),env,ci_state,impl)
      equation 
        (cl,cenv) = Lookup.lookup_class(env, Absyn.IDENT(id), false) "If first part of ident is a class, e.g StateSelect.None, nothing to update "; then env;
    case (mods,Absyn.CREF_QUAL(id,_,_),env,ci_state,impl)
      equation 
        (ty,NONE,_) = Lookup.lookup_ident(env, id) "Nothing to update. "; then env;
    case (mods,Absyn.CREF_QUAL(id,_,_),env,ci_state,impl)
      equation 
        (ty,SOME((SCode.COMPONENT(n,final_,repl,prot,attr = SCode.ATTR(ad,flow_,acc,param,dir),t,m,_,comment),cmod)),_) = Lookup.lookup_ident(env, id) "For qualified names, e.g. a.b.c, instanitate component a ";
        (cl,cenv) = Lookup.lookup_class(env, t, false);
        crefs = get_cref_from_mod(m);
        env2_1 = update_components_in_env(mods, crefs, env, ci_state, impl);
        crefs2 = get_cref_from_dim(ad);
        env2 = update_components_in_env(mods, crefs2, env2_1, ci_state, impl);
        m_1 = Mod.elab_mod(env2, Prefix.NOPRE, m, impl) "Prefix does not matter, since we only update types in env, and does
	   not make any dae elements, etc.. ";
        classmod = Mod.lookup_modification_p(mods, t);
        mm = Mod.lookup_comp_modification(mods, n);
        mod = Mod.merge(classmod, mm, env2, Prefix.NOPRE);
        mod_1 = Mod.merge(mod, m_1, env2, Prefix.NOPRE);
        mod_2 = Mod.merge(cmod, mod_1, env2, Prefix.NOPRE);
        mod_3 = Mod.update_mod(env2, Prefix.NOPRE, mod_2, impl);
        eq = Mod.mod_equation(mod_3);
        owncref = Absyn.CREF_IDENT(n,{}) " The variable declaration and the (optional) equation   modification are inspected for array dimensions. ";
        dims = elab_arraydim(env2, owncref, ad, eq, impl, NONE);
        (dae1,csets_1,ty) = inst_var(cenv, ci_state, mod_3, Prefix.NOPRE, {}, n, cl, attr, 
          dims, {}, {}, false, NONE) " Instantiate the component ";
        binding = make_binding(env2, attr, eq) " The environment is extended with the new variable   binding. ";
        env_1 = Env.update_frame_v(env2, 
          Types.VAR(n,Types.ATTR(flow_,acc,param,dir),prot,ty,binding), false) "type info present"; then env_1;
    case (mod,cref,env,ci_state,impl)
      equation 
        Debug.fprint("failtrace", "-update_component_in_env failed, ident = ");
        str = Debug.fcallret("failtrace", Dump.print_component_ref_str, cref, "");
        Debug.fprint("failtrace", str);
        Debug.fprint("failtrace", "\n mods:");
        str2 = Debug.fcallret("failtrace", Mod.print_mod_str, mod, "");
        Debug.fprint("failtrace", str2);
        Debug.fprint("failtrace", "\n   env:   ");
        str3 = Debug.fcallret("failtrace", Env.print_env_str, env, "");
        Debug.fprint("failtrace", str3);
        Debug.fprint("failtrace", "\n"); then env;
  end matchcontinue;
end update_component_in_env;

protected function check_structural_param_binding
  input SCode.Variability in_variability;
  input Types.Binding in_binding;
  input Env.Env in_env;
  output Env.Env out_env;
  output Types.Binding out_binding;
algorithm 
  (out_env,out_binding):=
  matchcontinue (in_variability,in_binding,in_env)
    local
      type Exp_ComponentRefList = list<Exp.ComponentRef>;
      type Env_FrameList = list<Env.Frame>;
      Exp_ComponentRefList crefs;
      String str;
      Env_FrameList env_1,env;
      Exp.Exp exp;
      Boolean const_;
      Types.Binding bind;
    case (SCode.STRUCTPARAM,Types.EQBOUND(exp,const_),env) "collect varnames from binding and make them structural. Also make sure
	   * that binding has constant = true.
	   "
      equation 
        crefs = Exp.get_cref_from_exp(exp);
        str = Exp.print_exp_str(exp);
        env_1 = Util.list_fold(crefs, make_structural_in_env, env); then (env_1,Types.EQBOUND(exp,true));
    case (_,bind,env) then (env,bind); 
  end matchcontinue;
end check_structural_param_binding;

protected function make_structural_in_env " relation: make_structural_in_env
  author: PA
 
  This function is used to update a parameter in the environment to a 
  structural parameter.
 "
  input Exp.ComponentRef in_componentref;
  input Env.Env in_env;
  output Env.Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_componentref,in_env)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      type Env_FrameList = list<Env.Frame>;
      String n,a,id,s;
      Boolean flow_,prot,b,c,d,f,typed;
      SCode.Accessibility acc,g;
      Absyn.Direction dir,h;
      TupleTypes_TTypeAbsyn_PathOption tp;
      Types.Binding bind;
      Absyn_SubscriptList e;
      Absyn.Path i;
      SCode.Mod j;
      Absyn_PathOption k;
      Absyn_CommentOption l;
      Types.Mod m;
      Env_FrameList env_1,env;
      Exp.ComponentRef cr;
    case (Exp.CREF_IDENT(id,_),env)
      equation 
        (Types.VAR(n,Types.ATTR(flow_,acc,SCode.PARAM,dir),prot,tp,bind),SOME((SCode.COMPONENT(a,b,c,d,SCode.ATTR(e,f,g,SCode.PARAM,h),i,j,k,l),m)),typed) = Lookup.lookup_ident(env, id);
        env_1 = Env.extend_frame_v(env, 
          Types.VAR(n,Types.ATTR(flow_,acc,SCode.STRUCTPARAM,dir),prot,tp,bind), 
          SOME(
          (
          SCode.COMPONENT(a,b,c,d,SCode.ATTR(e,f,g,SCode.STRUCTPARAM,h),i,j,k,l),m)), false) "replace variable, relys on hash_add to replace node. "; then env_1;
    case (cr,env)
      equation 
        print("make_structural_in_env failed for component ");
        s = Exp.print_component_ref_str(cr);
        print(s);
        print("\n"); then env;
  end matchcontinue;
end make_structural_in_env;

protected function inst_dim_exp_lst " relation: inst_dim_exp_lst
 
  Instantiates dimension expressions, DimExp, which are transformed to Exp.Subscript's
 "
  input DimExpList in_dimexplist;
  input Boolean in_boolean;
  output Exp_SubscriptList out_exp_subscriptlist;
protected 
  type DimExpList = list<DimExp>;
  type Exp_SubscriptList = list<Exp.Subscript>;
algorithm 
  out_exp_subscriptlist:=
  matchcontinue (in_dimexplist,in_boolean)
    case ({},_) then {}; 
    case (x :: xs,b)
      equation 
        res = inst_dim_exp_lst(xs, b);
        r = inst_dim_exp(x, b); then r :: res;
  end matchcontinue;
end inst_dim_exp_lst;

protected function inst_dim_exp " relation: inst_dim_exp_lst
 
  instantiates one dimension expression, See also isnt_dim_exp_lst.
 "
  input DimExp in_dimexp;
  input Boolean in_boolean;
  output Exp.Subscript out_subscript;
algorithm 
  out_subscript:=
  matchcontinue (in_dimexp,in_boolean)
    case (DIMEXP(Exp.WHOLEDIM,_),(impl = false)) "impl FIXME: Fix slicing, e.g. Exp.SLICE, for impl=true "
      equation 
        Print.print_error_buf("# Error, can not determine size of array.\n") "impl FIXME: Fix slicing, e.g. Exp.SLICE, for impl=true "; then fail();
    case (DIMEXP(Exp.SLICE(_),_),(impl = false))
      equation 
        Print.print_error_buf("# Error, can not determine size of array slice\n"); then fail();
    case (DIMEXP((e = Exp.WHOLEDIM),_),(impl = true)) then e; 
    case (DIMINT(i),_) then Exp.INDEX(Exp.ICONST(i)); 
    case (DIMEXP((e = Exp.INDEX(_)),_),_) then e; 
  end matchcontinue;
end inst_dim_exp;

protected function inst_dim_type " relation inst_dim_type
  Retrieves the dimension expression as an integer option. 
  Non constant dimensions give NONE.
 "
  input DimExp in_dimexp;
  output IntegerOption out_integeroption;
protected 
  type IntegerOption = Option<Integer>;
algorithm 
  out_integeroption:=
  matchcontinue (in_dimexp)
    local Integer i;
    case DIMINT(i) then SOME(i); 
    case DIMEXP(_,_) then NONE; 
  end matchcontinue;
end inst_dim_type;

protected function fix_direction
  input DAE_ElementList in_dae_elementlist;
  input Absyn.Direction in_direction;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_dae_elementlist,in_direction)
    case (lst,Absyn.BIDIR) then lst; 
    case ({},_) then {}; 
    case (DAE.VAR(cr,vk,DAE.BIDIR,t,e,id,start,flow_,class_,dae_var_attr,comment) :: r,dir)
      equation 
        dir_1 = absyn_dir_to_dae_dir(dir) "bidirectional variables are changed to input or output if component has such
	 * prefix.
	 ";
        r_1 = fix_direction(r, dir); then DAE.VAR(cr,vk,dir_1,t,e,id,start,flow_,class_,dae_var_attr,comment) :: r_1;
    case (DAE.VAR(cr,vk,DAE.INPUT,t,e,id,start,flow_,class_,dae_var_attr,comment) :: r,dir)
      equation 
        Print.print_error_buf("#Error, component declared as ");
        str = Dump.direction_symbol(dir);
        Print.print_error_buf(str);
        Print.print_error_buf(" when having the variable ");
        str = Exp.print_component_ref_str(cr);
        Print.print_error_buf(str);
        Print.print_error_buf(" declared as input\n"); then fail();
    case (DAE.VAR(cr,vk,DAE.OUTPUT,t,e,id,start,flow_,class_,dae_var_attr,comment) :: r,dir)
      equation 
        Print.print_error_buf("#Error, component declared as ");
        str = Dump.direction_symbol(dir);
        Print.print_error_buf(str);
        Print.print_error_buf(" when having the variable ");
        str = Exp.print_component_ref_str(cr);
        Print.print_error_buf(str);
        Print.print_error_buf(" declared as input\n"); then fail();
    case (DAE.COMP(id,DAE.DAE(lst)) :: r,dir)
      equation 
        lst_1 = fix_direction(lst, dir);
        r_1 = fix_direction(r, dir); then DAE.COMP(id,DAE.DAE(lst_1)) :: r_1;
    case (x :: r,dir)
      equation 
        r_1 = fix_direction(r, dir); then x :: r_1;
  end matchcontinue;
end fix_direction;

protected function absyn_dir_to_dae_dir " relation: absyn_dir_to_dae_dir
 
  Helper relatoin to fix_direction. 
  Translates Absyn.Direction to DAE.VarDirection. Needed so that input, output is transferred
  to DAE.
 "
  input Absyn.Direction in_direction;
  output DAE.VarDirection out_vardirection;
algorithm 
  out_vardirection:=
  matchcontinue (in_direction)
    case Absyn.INPUT then DAE.INPUT; 
    case Absyn.OUTPUT then DAE.OUTPUT; 
    case Absyn.BIDIR then DAE.BIDIR; 
  end matchcontinue;
end absyn_dir_to_dae_dir;

protected function inst_array
  input Env.Env in_env;
  input ClassInf.State in_state;
  input Types.Mod in_mod;
  input Prefix.Prefix in_prefix;
  input Connect.Sets in_sets;
  input Ident in_ident;
  input TupleSCode_ClassSCode_Attributes in_tuplescode_classscode_attributes;
  input Integer in_integer;
  input DimExp in_dimexp;
  input DimExpList in_dimexplist;
  input IntegerList in_integerlist;
  input InstDims in_instdims;
  input Boolean in_boolean;
  input Absyn_CommentOption in_absyn_commentoption;
  output DAE_ElementList out_dae_elementlist;
  output Connect.Sets out_sets;
  output Types.Type out_type;
protected 
  type TupleSCode_ClassSCode_Attributes = tuple<SCode.Class,SCode.Attributes>;
  type DimExpList = list<DimExp>;
  type IntegerList = list<Integer>;
  type Absyn_CommentOption = Option<Absyn.Comment>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_sets,out_type):=
  matchcontinue (in_env,in_state,in_mod,in_prefix,in_sets,in_ident,in_tuplescode_classscode_attributes,in_integer,in_dimexp,in_dimexplist,in_integerlist,in_instdims,in_boolean,in_absyn_commentoption)
    case (env,(ci_state = ClassInf.FUNCTION(_)),mod,pre,csets,n,(cl,attr),i,DIMEXP(_,_),dims,idxs,inst_dims,impl,comment)
      equation 
        SOME(Types.TYPED(e,p)) = Mod.mod_equation(mod) "If is a function var.";
        (_,_,_,ty,st) = inst_class(env, mod, pre, csets, cl, inst_dims, true, INNER_CALL) "Which has an 
							  expression binding ";
        cr = Prefix.prefix_cref(pre, Exp.CREF_IDENT(n,{})) "Check their types... ";
        ty_1 = Types.elab_type(ty);
        e_1 = Types.match_prop(e, Types.PROP(ty,false), p);
        dae = make_dae_equation(Exp.CREF(cr,ty_1), e_1, NON_INITIAL); then ({dae},{},ty);
    case (env,ci_state,mod,pre,csets,n,(cl,attr),i,DIMEXP(_,_),dims,idxs,inst_dims,impl,comment)
      equation 
        (dae,csets,ty) = inst_var2(env, ci_state, mod, pre, csets, n, cl, attr, dims, 
          i :: idxs, inst_dims, impl, comment); then (dae,csets,ty);
    case (env,ci_state,mod,pre,csets,n,(cl,attr),i,DIMINT(stop),dims,idxs,inst_dims,impl,comment)
      equation 
        (i > stop) = true; then ({},csets,(Types.T_NOTYPE,NONE));
    case (env,ci_state,mod,pre,csets,n,(cl,attr),i,DIMINT(stop),dims,idxs,inst_dims,impl,comment)
      equation 
        SOME(Types.TYPED(e,p)) = Mod.mod_equation(mod) "Modifiers of arrays that are functioncall, eg. Real x{:}=foo(...)  Should only generate -one- functioncall ";
        true = Exp.contain_functioncall(e);
        (dae1,csets_1,ty) = inst_var2(env, ci_state, Types.NOMOD, pre, csets, n, cl, attr, dims, 
          i :: idxs, inst_dims, impl, comment);
        i_1 = i + 1;
        (dae2,csets_2,arrty) = inst_array(env, ci_state, Types.NOMOD, pre, csets_1, n, (cl,attr), 
          i_1, DIMINT(stop), dims, idxs, inst_dims, impl, comment);
        cr = Prefix.prefix_cref(pre, Exp.CREF_IDENT(n,{})) "Make the equation containing the functioncall ";
        arrty_1 = Types.elab_type(arrty);
        dae3 = make_dae_equation(Exp.CREF(cr,arrty_1), e, NON_INITIAL);
        dae = Util.list_flatten({dae1,dae2,{dae3}}); then (dae,csets_1,ty);
    case (env,ci_state,mod,pre,csets,n,(cl,attr),i,DIMINT(stop),dims,idxs,inst_dims,impl,comment)
      equation 
        mod_1 = Mod.lookup_idx_modification(mod, i);
        (dae1,csets_1,ty) = inst_var2(env, ci_state, mod_1, pre, csets, n, cl, attr, dims, 
          i :: idxs, inst_dims, impl, comment);
        i_1 = i + 1;
        (dae2,csets_2,_) = inst_array(env, ci_state, mod, pre, csets_1, n, (cl,attr), i_1, 
          DIMINT(stop), dims, idxs, inst_dims, impl, comment);
        dae = list_append(dae1, dae2); then (dae,csets_2,ty);
    case (_,_,_,_,_,n,(_,_),_,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_array failed: ");
        Debug.fcall("failtrace", Print.print_buf, n);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_array;

public function elab_component_arraydim_from_env
  input Env.Env in_env;
  input Exp.ComponentRef in_componentref;
  output DimExpList out_dimexplist;
protected 
  type DimExpList = list<DimExp>;
algorithm 
  out_dimexplist:=
  matchcontinue (in_env,in_componentref)
    case (env,(cref = Exp.CREF_IDENT(id,_)))
      equation 
        (ty,SOME((SCode.COMPONENT(n,final_,repl,prot,attr = SCode.ATTR(ad,flow_,acc,param,dir),t,m,bc,comment),cmod)),_) = Lookup.lookup_ident(env, id);
        cmod_1 = Types.strip_submod(cmod);
        m_1 = SCode.strip_submod(m);
        m_2 = Mod.elab_mod(env, Prefix.NOPRE, m_1, false);
        mod_2 = Mod.merge(cmod_1, m_2, env, Prefix.NOPRE);
        SOME(eq) = Mod.mod_equation(mod_2);
        dims = elab_component_arraydim_from_env_2(eq, env); then dims;
  end matchcontinue;
end elab_component_arraydim_from_env;

protected function elab_component_arraydim_from_env_2
  input Types.EqMod in_eqmod;
  input Env.Env in_env;
  output DimExpList out_dimexplist;
protected 
  type DimExpList = list<DimExp>;
algorithm 
  out_dimexplist:=
  matchcontinue (in_eqmod,in_env)
    case (Types.TYPED(e,Types.PROP(t,c)),env)
      equation 
        lst = Types.get_dimension_sizes(t);
        lst_1 = Util.list_map(lst, make_dimexp_from_int); then lst_1;
  end matchcontinue;
end elab_component_arraydim_from_env_2;

protected function make_dimexp_from_int
  input Integer in_integer;
  output DimExp out_dimexp;
algorithm 
  out_dimexp:=
  matchcontinue (in_integer)
    case (i) then DIMINT(i); 
  end matchcontinue;
end make_dimexp_from_int;

protected function elab_arraydim
  input Env.Env in_env;
  input Absyn.ComponentRef in_componentref;
  input Absyn.ArrayDim in_arraydim;
  input Types_EqModOption in_types_eqmodoption;
  input Boolean in_boolean;
  input Interactive_InteractiveSymbolTableOption in_interactive_interactivesymboltableoption;
  output DimExpList out_dimexplist;
protected 
  type Types_EqModOption = Option<Types.EqMod>;
  type Interactive_InteractiveSymbolTableOption = Option<Interactive.InteractiveSymbolTable>;
  type DimExpList = list<DimExp>;
algorithm 
  out_dimexplist:=
  matchcontinue (in_env,in_componentref,in_arraydim,in_types_eqmodoption,in_boolean,in_interactive_interactivesymboltableoption)
    local
      type DimExpOption = Option<DimExp>;
      type DimExpOptionList = list<DimExpOption>;
      type Env_FrameList = list<Env.Frame>;
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      DimExpOptionList dim,dim1,dim2;
      DimExpList dim_1,dim3;
      Env_FrameList env;
      Absyn.ComponentRef cref;
      Absyn_SubscriptList ad;
      Boolean impl,c;
      Interactive_InteractiveSymbolTableOption st;
      Exp.Exp e,e_1;
      TupleTypes_TTypeAbsyn_PathOption t;
      String str;
    case (env,cref,ad,NONE,impl,st)
      equation 
        dim = elab_arraydim_decl(env, cref, ad, impl, st);
        dim_1 = complete_arraydim(dim); then dim_1;
    case (env,cref,ad,SOME(Types.TYPED(e,Types.PROP(t,c))),impl,st)
      equation 
        dim1 = elab_arraydim_decl(env, cref, ad, impl, st);
        dim2 = elab_arraydim_type(t, ad);
        dim3 = compatible_arraydim(dim1, dim2); then dim3;
    case (env,cref,ad,SOME(Types.UNTYPED(e)),impl,st)
      local Absyn.Exp e;
      equation 
        (e_1,Types.PROP(t,c),_) = Static.elab_exp(env, e, impl, st);
        dim1 = elab_arraydim_decl(env, cref, ad, impl, st);
        dim2 = elab_arraydim_type(t, ad);
        dim3 = compatible_arraydim(dim1, dim2); then dim3;
    case (env,cref,ad,SOME(Types.TYPED(e,Types.PROP(t,c))),impl,st)
      equation 
        dim1 = elab_arraydim_decl(env, cref, ad, impl, st);
        dim2 = elab_arraydim_type(t, ad);
        failure(dim3 = compatible_arraydim(dim1, dim2));
        Print.print_error_buf("# Array dimension mismatch\n");
        Print.print_error_buf("    expression ");
        str = Exp.print_exp_str(e);
        Print.print_error_buf(str);
        Print.print_error_buf(" has type ");
        str = Types.unparse_type(t);
        Print.print_error_buf(str);
        Print.print_error_buf("\n");
        Print.print_error_buf("    expected array dimensions [");
        str = print_dim_str(dim1);
        Print.print_error_buf(str);
        Print.print_error_buf("]\n"); then fail();
    case (_,cref,ad,_,_,_)
      equation 
        Debug.fprint("failtrace", "- elab_arraydim failed\n cref:");
        Debug.fcall("failtrace", Dump.print_component_ref, cref);
        Debug.fprint("failtrace", " dim: ");
        Debug.fcall("failtrace", Dump.print_arraydim, ad);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end elab_arraydim;

protected function print_dim_str " relation: print_dim
 
  This function prints array dimensions.  The code is not included
  in the report.
 
  !ignorecode"
  input DimExpOptionList in_dimexpoptionlist;
  output String out_string;
protected 
  type DimExpOption = Option<DimExp>;
  type DimExpOptionList = list<DimExpOption>;
algorithm 
  out_string:=
  matchcontinue (in_dimexpoptionlist)
    local
      String s,str,res,s2,s1;
      Integer x;
      DimExpOptionList xs;
    case {NONE} then ":"; 
    case {SOME(DIMINT(x))}
      equation 
        s = int_string(x); then s;
    case {SOME(DIMEXP(x,_))}
      local Exp.Subscript x;
      equation 
        s = Exp.print_subscript_str(x); then s;
    case NONE :: xs
      equation 
        str = print_dim_str(xs);
        res = string_append(":,", str); then res;
    case SOME(DIMINT(x)) :: xs
      equation 
        s = int_string(x);
        s2 = print_dim_str(xs);
        res = Util.string_append_list({s,",",s2}); then res;
    case SOME(DIMEXP(x,_)) :: xs
      local Exp.Subscript x;
      equation 
        s1 = Exp.print_subscript_str(x);
        s2 = print_dim_str(xs);
        res = Util.string_append_list({s1,",",s2}); then res;
  end matchcontinue;
end print_dim_str;

protected function print_dim " relation: print_dim
 
  Prints a dimension expression option list to the print buffer.
 "
  input DimExpOptionList dims;
protected 
  type DimExpOption = Option<DimExp>;
  type DimExpOptionList = list<DimExpOption>;
algorithm 
  str := print_dim_str(dims);
  Print.print_buf(str);
end print_dim;

protected function print_dim2 " relation print_dim2
 
  Helper function to print_dim
 "
  input DimExpList in_dimexplist;
protected 
  type DimExpList = list<DimExp>;
algorithm 
  _:=
  matchcontinue (in_dimexplist)
    case {DIMINT(x)}
      equation 
        s = int_string(x);
        Print.print_buf(s); then ();
    case {DIMEXP(x,_)}
      equation 
        Exp.print_subscript(x); then ();
    case DIMINT(x) :: xs
      equation 
        s = int_string(x);
        Print.print_buf(s);
        Print.print_buf(",");
        print_dim2(xs); then ();
    case DIMEXP(x,_) :: xs
      equation 
        Exp.print_subscript(x);
        Print.print_buf(",");
        print_dim2(xs); then ();
    case {} then (); 
  end matchcontinue;
end print_dim2;

protected function elab_arraydim_decl
  input Env.Env in_env;
  input Absyn.ComponentRef in_componentref;
  input Absyn.ArrayDim in_arraydim;
  input Boolean in_boolean;
  input Interactive_InteractiveSymbolTableOption in_interactive_interactivesymboltableoption;
  output DimExpOptionList out_dimexpoptionlist;
protected 
  type Interactive_InteractiveSymbolTableOption = Option<Interactive.InteractiveSymbolTable>;
  type DimExpOption = Option<DimExp>;
  type DimExpOptionList = list<DimExpOption>;
algorithm 
  out_dimexpoptionlist:=
  matchcontinue (in_env,in_componentref,in_arraydim,in_boolean,in_interactive_interactivesymboltableoption)
    local
      type Env_FrameList = list<Env.Frame>;
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      DimExpOptionList l;
      Env_FrameList env;
      Absyn.ComponentRef cref,cr;
      Absyn_SubscriptList ds;
      Boolean impl,cnst;
      Interactive_InteractiveSymbolTableOption st;
      Exp.Exp e;
      Integer i;
      Absyn.Exp d;
      TupleTypes_TTypeAbsyn_PathOption t;
      String str;
    case (_,_,{},_,_) then {}; 
    case (env,cref,Absyn.NOSUB :: ds,impl,st)
      equation 
        l = elab_arraydim_decl(env, cref, ds, impl, st) "Print.print_buf \"NOSUB in elab_arraydim_decl\n\" & "; then NONE :: l;
    case (env,cref,Absyn.SUBSCRIPT(Absyn.CALL(Absyn.CREF_IDENT("size",_),Absyn.FUNCTIONARGS({Absyn.CREF(cr),_},_))) :: ds,impl,st)
      equation 
        true = Absyn.cref_equal(cref, cr) "For functions, this can occur: Real x{:,size(x,1)} ,i.e. referencing 
	 the variable itself but a different dimension.";
        l = elab_arraydim_decl(env, cref, ds, impl, st); then NONE :: l;
    case (env,cref,Absyn.SUBSCRIPT(d) :: ds,impl,st) "as false"
      equation 
        Debug.fprintln("insttr", "elab_arraydim_decl5");
        (e,Types.PROP((Types.T_INTEGER(_),_),cnst = true),_) = Static.elab_exp(env, d, impl, st);
        (Values.INTEGER(i),_) = Ceval.ceval(env, e, impl, st, NONE) "_ Was true, but structural parameters must be given 
	       the numerical value in order to instantiate successfully.
	       PA: This is true for normal models, e.g. model test Real x{n}; end test;
	       It is NOT true for functions, or other implicit instantiated models. 
	     ";
        l = elab_arraydim_decl(env, cref, ds, impl, st); then SOME(DIMINT(i)) :: l;
    case (env,cref,Absyn.SUBSCRIPT(d) :: ds,impl,st)
      equation 
        Debug.fprintln("insttr", "elab_arraydim_decl6") "as false";
        (e,Types.PROP((Types.T_INTEGER(_),_),cnst),_) = Static.elab_exp(env, d, impl, st);
        l = elab_arraydim_decl(env, cref, ds, impl, st); then SOME(DIMEXP(Exp.INDEX(e),NONE)) :: l;
    case (env,cref,Absyn.SUBSCRIPT(d) :: ds,impl,st)
      equation 
        (e = Exp.SIZE(_,_),Types.PROP(t,_),_) = Static.elab_exp(env, d, impl, st) "Size(x,1) in e.g. functions => Unknown dimension ";
        l = elab_arraydim_decl(env, cref, ds, impl, st); then SOME(DIMEXP(Exp.INDEX(e),NONE)) :: l;
    case (env,cref,Absyn.SUBSCRIPT(d) :: ds,impl,st)
      equation 
        (e,Types.PROP(t,_),_) = Static.elab_exp(env, d, impl, st);
        Print.print_error_buf("# Array dimensions must be integer\n");
        Print.print_error_buf("    expression: ");
        str = Exp.print_exp_str(e);
        Print.print_error_buf(str);
        Print.print_error_buf(" (");
        str = Types.unparse_type(t);
        Print.print_error_buf(str);
        Print.print_error_buf(")\n"); then fail();
    case (_,_,_,_,_)
      equation 
        Print.print_error_buf("elab_arraydim_decl failed\n");
        Debug.fprint("failtrace", "- elab_arraydim_decl failed\n"); then fail();
  end matchcontinue;
end elab_arraydim_decl;

protected function complete_arraydim " relation: complete_arraydim
 
  This function converts a list of optional integers to a list of
  integers.  If one element of the list is `NONE', this function
  will fail.
 
  This is used to check that an array specification contain fully
  specified array dimension sizes.
 "
  input DimExpOptionList in_dimexpoptionlist;
  output DimExpList out_dimexplist;
protected 
  type DimExpOption = Option<DimExp>;
  type DimExpOptionList = list<DimExpOption>;
  type DimExpList = list<DimExp>;
algorithm 
  out_dimexplist:=
  matchcontinue (in_dimexpoptionlist)
    local
      DimExpList xs_1;
      DimExp x;
      DimExpOptionList xs;
    case {} then {}; 
    case SOME(x) :: xs
      equation 
        xs_1 = complete_arraydim(xs); then x :: xs_1;
    case NONE :: xs
      equation 
        xs_1 = complete_arraydim(xs); then DIMEXP(Exp.WHOLEDIM,NONE) :: xs_1;
  end matchcontinue;
end complete_arraydim;

protected function compatible_arraydim
  input DimExpOptionList in_dimexpoptionlist1;
  input DimExpOptionList in_dimexpoptionlist2;
  output DimExpList out_dimexplist;
protected 
  type DimExpOption = Option<DimExp>;
  type DimExpOptionList = list<DimExpOption>;
  type DimExpList = list<DimExp>;
algorithm 
  out_dimexplist:=
  matchcontinue (in_dimexpoptionlist1,in_dimexpoptionlist2)
    local
      DimExpList l;
      DimExp x,y,de;
      DimExpOptionList xs,ys;
      Exp_ExpOption e,e1,e2;
    case ({},{}) then {}; 
    case (SOME(x) :: xs,NONE :: ys)
      equation 
        l = compatible_arraydim(xs, ys); then x :: l;
    case (NONE :: xs,SOME(y) :: ys)
      equation 
        l = compatible_arraydim(xs, ys); then y :: l;
    case (SOME(DIMINT(x)) :: xs,SOME(DIMINT(y)) :: ys)
      local Integer x,y;
      equation 
        equality(x = y);
        l = compatible_arraydim(xs, ys); then DIMINT(x) :: l;
    case (SOME(DIMINT(x)) :: xs,SOME(DIMEXP(y,e)) :: ys)
      local
        Integer x;
        Exp.Subscript y;
      equation 
        de = arraydim_condition(DIMEXP(Exp.INDEX(Exp.ICONST(x)),NONE), DIMEXP(y,e));
        l = compatible_arraydim(xs, ys); then de :: l;
    case (SOME(DIMEXP(x,e)) :: xs,SOME(DIMINT(y)) :: ys)
      local
        Exp.Subscript x;
        Integer y;
      equation 
        de = arraydim_condition(DIMEXP(Exp.INDEX(Exp.ICONST(y)),NONE), DIMEXP(x,e));
        l = compatible_arraydim(xs, ys); then de :: l;
    case (SOME(DIMEXP(x,e1)) :: xs,SOME(DIMEXP(y,e2)) :: ys)
      local Exp.Subscript x,y;
      equation 
        de = arraydim_condition(DIMEXP(x,e1), DIMEXP(y,e2));
        l = compatible_arraydim(xs, ys); then de :: l;
    case (NONE :: xs,NONE :: ys)
      equation 
        l = compatible_arraydim(xs, ys); then DIMEXP(Exp.WHOLEDIM,NONE) :: l;
    case (_,_)
      equation 
        Print.print_buf("-compatible_arraydim failed\n");
        Debug.fprint("failtrace", "- compatible_arraydim failed\n"); then fail();
  end matchcontinue;
end compatible_arraydim;

protected function arraydim_condition " relation arraydim_condition
  
  This function checks that the two arraydim expressions have the same dimension.
  FIXME: no check performed yet, just return first DimExp.
 "
  input DimExp in_dimexp1;
  input DimExp in_dimexp2;
  output DimExp out_dimexp;
algorithm 
  out_dimexp:=
  matchcontinue (in_dimexp1,in_dimexp2)
    local DimExp de;
    case (de,_) then de; 
  end matchcontinue;
end arraydim_condition;

protected function elab_arraydim_type
  input Types.Type in_type;
  input Absyn.ArrayDim in_arraydim;
  output DimExpOptionList out_dimexpoptionlist;
protected 
  type DimExpOption = Option<DimExp>;
  type DimExpOptionList = list<DimExpOption>;
algorithm 
  out_dimexpoptionlist:=
  matchcontinue (in_type,in_arraydim)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      DimExpOptionList l;
      TupleTypes_TTypeAbsyn_PathOption t;
      Absyn_SubscriptList ad;
      Integer i;
    case ((Types.T_ARRAY(Types.DIM(NONE),t),_),_ :: ad)
      equation 
        l = elab_arraydim_type(t, ad); then NONE :: l;
    case ((Types.T_ARRAY(Types.DIM(SOME(i)),t),_),_ :: ad)
      equation 
        l = elab_arraydim_type(t, ad); then SOME(DIMINT(i)) :: l;
    case (_,{}) then {}; 
    case (t,_ :: ad) "PR, for debugging "
      equation 
        Debug.fprint("failtrace", "Undefined!");
        Debug.fprint("failtrace", " The type detected: ");
        Debug.fcall("failtrace", Types.print_type, t); then fail();
  end matchcontinue;
end elab_arraydim_type;

public function inst_class_decl "packimpl"
  input Env.Env in_env;
  input Types.Mod in_mod;
  input Prefix.Prefix in_prefix;
  input Connect.Sets in_sets;
  input SCode.Class in_class;
  input InstDims in_instdims;
  input Boolean in_boolean;
  output Env.Env out_env;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_env,out_dae_elementlist):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_class,in_instdims,in_boolean)
    case (env,mod,pre,csets,(c = SCode.CLASS(n,_,_,restr,_)),inst_dims,packimpl) "packimpl"
      equation 
        env_1 = Env.extend_frame_c(env, c) "Debug.fprint(\"insttr\",\"+\tinst_class_decl called\n\") &";
        (env_2,dae) = implicit_instantiation(env_1, Types.NOMOD, pre, csets, c, inst_dims, packimpl) "	Debug.fprint(\"insttr\",\"+\textended frame\n\") &"; then (env_2,dae);
    case (env,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_class_decl failed\n"); then fail();
  end matchcontinue;
end inst_class_decl;

public function implicit_instantiation "packimpl"
  input Env.Env in_env;
  input Types.Mod in_mod;
  input Prefix.Prefix in_prefix;
  input Connect.Sets in_sets;
  input SCode.Class in_class;
  input InstDims in_instdims;
  input Boolean in_boolean;
  output Env.Env out_env;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_env,out_dae_elementlist):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_class,in_instdims,in_boolean)
    case (env,mod,pre,csets,(c = SCode.CLASS(n,_,_,SCode.R_FUNCTION,_)),inst_dims,(packimpl = false)) "packimpl"
      equation 
        (dae,_,csets_1,ty,st) = inst_class(env, mod, pre, csets, c, inst_dims, true, INNER_CALL) "Implicit instantiation of functions  impl";
        env_1 = Env.extend_frame_t(env, n, ty);
        fpath = make_fully_qualified(env, Absyn.IDENT(n)); then (env_1,{DAE.FUNCTION(fpath,DAE.DAE(dae),ty)});
    case (env,mod,pre,csets,(c = SCode.CLASS(n,_,_,(restr = SCode.R_EXT_FUNCTION),(parts = SCode.PARTS(els,_,_,_,_,_)))),inst_dims,(packimpl = false))
      equation 
        (dae,_,csets_1,ty,st) = inst_class(env, mod, pre, csets, c, inst_dims, true, INNER_CALL) "Implicit instantiation of external functions ";
        env_1 = Env.extend_frame_t(env, n, ty);
        fpath = make_fully_qualified(env, Absyn.IDENT(n));
        prot = false;
        (_,tempenv,_,_,_) = inst_classdef(env, mod, pre, csets, ClassInf.UNKNOWN(n), parts, restr, 
          prot, inst_dims, true, packimpl) "impl";
        extdecl = inst_ext_decl(tempenv, n, parts, true) "impl"; then (env_1,{DAE.EXTFUNCTION(fpath,DAE.DAE(dae),ty,extdecl)});
    case (env,mod,pre,csets,(c = SCode.CLASS(n,_,_,SCode.R_TYPE,SCode.ENUMERATION(l))),inst_dims,packimpl)
      equation 
        enumclass = inst_enumeration(n, l) "enumerations ";
        env_2 = Env.extend_frame_c(env, enumclass); then (env_2,{});
    case (env,mod,pre,csets,c,_,_) then (env,{}); 
  end matchcontinue;
end implicit_instantiation;

public function make_fully_qualified " relation: make_fully_qualified
  author: PA
 
  Transforms a class name to its fully qualified name by investigating the environment.
  For instance, the model Resistor in Modelica.Electrical.Analog.Basic will given the 
  correct environment have the fully qualified name: Modelica.Electrical.Analog.Basic.Resistor
 "
  input Env.Env in_env;
  input Absyn.Path in_path;
  output Absyn.Path out_path;
algorithm 
  out_path:=
  matchcontinue (in_env,in_path)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList env,env_1;
      Absyn.Path path,path_1,path_2;
      String class_name;
    case (env,path)
      equation 
        NONE = Env.get_env_path(env); then path;
    case (env,path)
      equation 
        (_,env_1) = Lookup.lookup_class(env, path, false) " To make a class fully qualified, the class path
	 ** is looked up in the environment.
	 ** The FQ path consist of the simple class name
	 ** appended to the environment path of the looked up class.
	 *";
        SOME(path_1) = Env.get_env_path(env_1);
        class_name = Absyn.path_last_ident(path);
        path_2 = Absyn.join_paths(path_1, Absyn.IDENT(class_name)); then path_2;
    case (env,path) then path; 
  end matchcontinue;
end make_fully_qualified;

public function implicit_function_instantiation
  input Env.Env in_env;
  input Types.Mod in_mod;
  input Prefix.Prefix in_prefix;
  input Connect.Sets in_sets;
  input SCode.Class in_class;
  input InstDims in_instdims;
  input Boolean in_boolean;
  output Env.Env out_env;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_env,out_dae_elementlist):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_class,in_instdims,in_boolean)
    case (env,mod,pre,csets,(c = SCode.CLASS(n,_,_,SCode.R_FUNCTION,_)),inst_dims,(packimpl = false))
      equation 
        (dae,_,csets_1,ty,st) = inst_class(env, mod, pre, csets, c, inst_dims, true, INNER_CALL) "Prefix.prefix_add(n,{},pre) => pre' &";
        env_1 = Env.extend_frame_t(env, n, ty);
        fpath = make_fully_qualified(env, Absyn.IDENT(n)); then (env_1,{DAE.FUNCTION(fpath,DAE.DAE(dae),ty)});
    case (env,mod,pre,csets,(c = SCode.CLASS(n,_,_,(restr = SCode.R_EXT_FUNCTION),(parts = SCode.PARTS(els,_,_,_,_,_)))),inst_dims,(packimpl = false))
      equation 
        (dae,_,csets_1,ty,st) = inst_class(env, mod, pre, csets, c, inst_dims, true, INNER_CALL) "External functions should also have their type in env, but no dae! ";
        env_1 = Env.extend_frame_t(env, n, ty);
        fpath = make_fully_qualified(env, Absyn.IDENT(n));
        prot = false;
        (_,tempenv,_,_,_) = inst_classdef(env_1, mod, pre, csets, ClassInf.FUNCTION(n), parts, 
          restr, prot, inst_dims, true, packimpl) "how to get this?  impl";
        extdecl = inst_ext_decl(tempenv, n, parts, true) "impl"; then (env_1,{DAE.EXTFUNCTION(fpath,DAE.DAE(dae),ty,extdecl)});
    case (env,mod,pre,csets,(c = SCode.CLASS(n,_,_,(restr = SCode.R_FUNCTION),SCode.OVERLOAD(funcnames))),inst_dims,(packimpl = false))
      equation 
        (env_1,daefuncs) = inst_overloaded_functions(env, n, funcnames) "Overloaded functions "; then (env_1,daefuncs);
    case (_,_,_,_,_,_,_) then fail(); 
  end matchcontinue;
end implicit_function_instantiation;

public function implicit_function_type_instantiation
  input Env.Env in_env;
  input SCode.Class in_class;
  output Env.Env out_env;
algorithm 
  out_env:=
  matchcontinue (in_env,in_class)
    case (env,SCode.CLASS(id,p,e,r,SCode.PARTS(elts,_,_,_,_,_)))
      equation 
        stripped_class = SCode.CLASS(id,p,e,r,SCode.PARTS(elts,{},{},{},{},NONE));
        (env_1,_) = implicit_function_instantiation(env, Types.NOMOD, Prefix.NOPRE, {}, stripped_class, {}, 
          false); then env_1;
  end matchcontinue;
end implicit_function_type_instantiation;

protected function inst_overloaded_functions
  input Env.Env in_env;
  input Absyn.Ident in_ident;
  input Absyn_PathList in_absyn_pathlist;
  output Env.Env out_env;
  output DAE_ElementList out_dae_elementlist;
protected 
  type Absyn_PathList = list<Absyn.Path>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_env,out_dae_elementlist):=
  matchcontinue (in_env,in_ident,in_absyn_pathlist)
    case (env,_,{}) then (env,{}); 
    case (env,overloadname,fn :: fns)
      equation 
        (c = SCode.CLASS(id,_,encflag,SCode.R_FUNCTION,_),cenv) = Lookup.lookup_class(env, fn, true) "Instantiate each function, add its FQ name to the type, needed when deoverloading ";
        (dae,_,_,(Types.T_FUNCTION(args,tp),_),st) = inst_class(cenv, Types.NOMOD, Prefix.NOPRE, {}, c, {}, true, 
          INNER_CALL);
        fpath = make_fully_qualified(env, Absyn.IDENT(overloadname));
        ovlfpath = make_fully_qualified(cenv, Absyn.IDENT(id));
        ty = (Types.T_FUNCTION(args,tp),SOME(ovlfpath));
        env_1 = Env.extend_frame_t(env, overloadname, ty);
        (env_2,dae1) = inst_overloaded_functions(env_1, overloadname, fns); then (env_2,DAE.FUNCTION(fpath,DAE.DAE(dae),ty) :: dae1);
    case (env,_,_)
      equation 
        Debug.fprint("failtrace", "inst_overloaded_functions failed\n"); then fail();
  end matchcontinue;
end inst_overloaded_functions;

protected function inst_ext_decl " relation: inst_ext_decl
  author: LS
 
  This function handles the external declaration. If there is an explicit call
  of the external function, the component references are looked up and
  inserted in the argument list, otherwise the input and output parameters are
  inserted in the argument list with their order. The return type is
  determined according to the specification; if there is a explicit call and a
  lhs, which must be an output parameter, the type of the function is that
  type. If no explicit call and only one output parameter exists, then this
  will be the return type of the function, otherwise the return type will be
  void. 
 "
  input Env.Env in_env;
  input Ident in_ident;
  input SCode.ClassDef in_classdef;
  input Boolean in_boolean;
  output DAE.ExternalDecl out_externaldecl;
algorithm 
  out_externaldecl:=
  matchcontinue (in_env,in_ident,in_classdef,in_boolean)
    local
      type DAE_ExtArgList = list<DAE.ExtArg>;
      type Env_FrameList = list<Env.Frame>;
      type SCode_ElementList = list<SCode.Element>;
      String fname,lang,n;
      DAE_ExtArgList fargs;
      DAE.ExtArg rettype;
      DAE.ExternalDecl daeextdecl;
      Env_FrameList env;
      Absyn.ExternalDecl extdecl,orgextdecl;
      Boolean impl;
      SCode_ElementList els;
    case (env,n,SCode.PARTS(_,_,_,_,_,SOME(extdecl)),impl) "impl"
      equation 
        is_ext_explicit_call(extdecl);
        fname = inst_ext_get_fname(extdecl, n);
        fargs = inst_ext_get_fargs(env, extdecl, impl);
        rettype = inst_ext_get_rettype(env, extdecl, impl);
        lang = inst_ext_get_lang(extdecl);
        daeextdecl = DAE.EXTERNALDECL(fname,fargs,rettype,lang); then daeextdecl;
    case (env,n,SCode.PARTS(els,_,_,_,_,SOME(orgextdecl)),impl)
      equation 
        failure(is_ext_explicit_call(orgextdecl));
        extdecl = inst_ext_make_externaldecl(n, els, orgextdecl);
        fname = inst_ext_get_fname(extdecl, n);
        fargs = inst_ext_get_fargs(env, extdecl, impl);
        rettype = inst_ext_get_rettype(env, extdecl, impl);
        lang = inst_ext_get_lang(extdecl);
        daeextdecl = DAE.EXTERNALDECL(fname,fargs,rettype,lang); then daeextdecl;
    case (env,_,_,_)
      equation 
        Debug.fprint("failtrace", "#-- inst_ext_decl failed"); then fail();
  end matchcontinue;
end inst_ext_decl;

protected function is_ext_explicit_call " relation: is_ext_explicit_call
  
  If the external function id is present, then a function call must
  exist, i.e. explicit call was written in the external clause.
 "
  input Absyn.ExternalDecl in_externaldecl;
algorithm 
  _:=
  matchcontinue (in_externaldecl)
    local String id;
    case Absyn.EXTERNALDECL(SOME(id),_,_,_) then (); 
  end matchcontinue;
end is_ext_explicit_call;

protected function inst_ext_make_externaldecl
  input Ident in_ident;
  input SCode_ElementList in_scode_elementlist;
  input Absyn.ExternalDecl in_externaldecl;
  output Absyn.ExternalDecl out_externaldecl;
protected 
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_externaldecl:=
  matchcontinue (in_ident,in_scode_elementlist,in_externaldecl)
    local
      type Absyn_ExpList = list<Absyn.Exp>;
      type Absyn_ExpListList = list<Absyn_ExpList>;
      type StringOption = Option<String>;
      SCode.Element outvar;
      SCode_ElementList invars,els,inoutvars;
      Absyn_ExpListList explists;
      Absyn_ExpList exps;
      Absyn.ComponentRef retcref;
      Absyn.ExternalDecl extdecl;
      String id;
      StringOption lang;
    case (id,els,Absyn.EXTERNALDECL(_,lang,_,_)) "the case with only one output var, and that cannot be array, otherwise
      inst_ext_make_crefs outvar will fail "
      equation 
        outvar :: {} = Util.list_matching(els, is_output_var);
        invars = Util.list_matching(els, is_input_var);
        explists = Util.list_map(invars, inst_ext_make_crefs);
        exps = Util.list_flatten(explists);
        {Absyn.CREF(retcref)} = inst_ext_make_crefs(outvar);
        extdecl = Absyn.EXTERNALDECL(SOME(id),lang,SOME(retcref),exps); then extdecl;
    case (id,els,Absyn.EXTERNALDECL(_,lang,_,_))
      equation 
        inoutvars = Util.list_matching(els, is_inout_var);
        explists = Util.list_map(inoutvars, inst_ext_make_crefs);
        exps = Util.list_flatten(explists);
        extdecl = Absyn.EXTERNALDECL(SOME(id),lang,NONE,exps); then extdecl;
    case (_,_,_)
      equation 
        Debug.fprint("failtrace", "#-- inst_ext_make_externaldecl failed\n"); then fail();
  end matchcontinue;
end inst_ext_make_externaldecl;

protected function is_inout_var " relation: is_inout_var 
 
  Succeds for Elements that are input or output components
 "
  input SCode.Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    local SCode.Element e;
    case e
      equation 
        is_output_var(e); then ();
    case e
      equation 
        is_input_var(e); then ();
  end matchcontinue;
end is_inout_var;

protected function is_output_var " relation: is_output_var
 
  Succeds for element that is output component
"
  input SCode.Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    case SCode.COMPONENT(_,_,_,_,SCode.ATTR(_,_,_,_,Absyn.OUTPUT),_,_,_,_) then (); 
  end matchcontinue;
end is_output_var;

protected function is_input_var " relation: is_input_var
 
  Succeds for element that is input component
 "
  input SCode.Element in_element;
algorithm 
  _:=
  matchcontinue (in_element)
    case SCode.COMPONENT(_,_,_,_,SCode.ATTR(_,_,_,_,Absyn.INPUT),_,_,_,_) then (); 
  end matchcontinue;
end is_input_var;

protected function inst_ext_make_crefs " relation: inst_ext_make_crefs
  author: LS
 
  This function is used in external function declarations. It collects the component
  identifier and the dimension sizes and returns as a Absyn.Exp list
"
  input SCode.Element in_element;
  output Absyn_ExpList out_absyn_explist;
protected 
  type Absyn_ExpList = list<Absyn.Exp>;
algorithm 
  out_absyn_explist:=
  matchcontinue (in_element)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      type Absyn_CommentOption = Option<Absyn.Comment>;
      Absyn_ExpList sizelist,crlist;
      String id;
      Boolean fi,re,pr;
      Absyn_SubscriptList dims;
      Absyn.Path path;
      SCode.Mod mod;
      Absyn_CommentOption comment;
    case SCode.COMPONENT(id,fi,re,pr,SCode.ATTR(dims,_,_,_,_),path,mod,_,comment)
      equation 
        sizelist = inst_ext_make_crefs_2(id, dims, 1);
        crlist = Absyn.CREF(Absyn.CREF_IDENT(id,{})) :: sizelist; then crlist;
  end matchcontinue;
end inst_ext_make_crefs;

protected function inst_ext_make_crefs_2 " relation: inst_ext_make_crefs_2
 
  Helper function to inst_ext_make_crefs, collects array dimension sizes.
 "
  input SCode.Ident in_ident;
  input Absyn.ArrayDim in_arraydim;
  input Integer in_integer;
  output Absyn_ExpList out_absyn_explist;
protected 
  type Absyn_ExpList = list<Absyn.Exp>;
algorithm 
  out_absyn_explist:=
  matchcontinue (in_ident,in_arraydim,in_integer)
    local
      type Absyn_SubscriptList = list<Absyn.Subscript>;
      String id;
      Integer nextdimno,dimno;
      Absyn_ExpList restlist,exps;
      Absyn.Subscript dim;
      Absyn_SubscriptList restdim;
    case (id,{},_) then {}; 
    case (id,dim :: restdim,dimno)
      equation 
        nextdimno = dimno + 1;
        restlist = inst_ext_make_crefs_2(id, restdim, nextdimno);
        exps = Absyn.CALL(Absyn.CREF_IDENT("size",{}),
          Absyn.FUNCTIONARGS({Absyn.CREF(Absyn.CREF_IDENT(id,{})),Absyn.INTEGER(dimno)},
          {})) :: restlist; then exps;
  end matchcontinue;
end inst_ext_make_crefs_2;

protected function inst_ext_get_fname " relation: inst_ex_get_fname
  
  Returns the function name of the externally defined function.
 "
  input Absyn.ExternalDecl in_externaldecl;
  input Ident in_ident;
  output Ident out_ident;
algorithm 
  out_ident:=
  matchcontinue (in_externaldecl,in_ident)
    local String id,fid;
    case (Absyn.EXTERNALDECL(SOME(id),_,_,_),fid) then id; 
    case (Absyn.EXTERNALDECL(NONE,_,_,_),fid) then fid; 
  end matchcontinue;
end inst_ext_get_fname;

protected function inst_ext_get_lang " relation: inst_ext_get_lang
 
  Return the implementation language of the external function declaration
  Defaults to \"C\" if no language specified.
 "
  input Absyn.ExternalDecl in_externaldecl;
  output String out_string;
algorithm 
  out_string:=
  matchcontinue (in_externaldecl)
    local String lang;
    case Absyn.EXTERNALDECL(_,SOME(lang),_,_) then lang; 
    case Absyn.EXTERNALDECL(_,NONE,_,_) then "C"; 
  end matchcontinue;
end inst_ext_get_lang;

protected function elab_exp_list_ext
  input Env.Env in_env;
  input Absyn_ExpList in_absyn_explist;
  input Boolean in_boolean;
  input Interactive_InteractiveSymbolTableOption in_interactive_interactivesymboltableoption;
  output Exp_ExpList out_exp_explist;
  output Types_PropertiesList out_types_propertieslist;
  output Interactive_InteractiveSymbolTableOption out_interactive_interactivesymboltableoption;
protected 
  type Absyn_ExpList = list<Absyn.Exp>;
  type Interactive_InteractiveSymbolTableOption = Option<Interactive.InteractiveSymbolTable>;
  type Exp_ExpList = list<Exp.Exp>;
  type Types_PropertiesList = list<Types.Properties>;
algorithm 
  (out_exp_explist,out_types_propertieslist,out_interactive_interactivesymboltableoption):=
  matchcontinue (in_env,in_absyn_explist,in_boolean,in_interactive_interactivesymboltableoption)
    local
      type Env_FrameList = list<Env.Frame>;
      Boolean impl;
      Interactive_InteractiveSymbolTableOption st,st_1,st_2;
      Exp.Exp exp;
      Types.Properties p;
      Exp_ExpList exps;
      Types_PropertiesList props;
      Env_FrameList env;
      Absyn.Exp e;
      Absyn_ExpList rest;
    case (_,{},impl,st) then ({},{},st); 
    case (env,e :: rest,impl,st)
      equation 
        (exp,p,st_1) = elab_exp_ext(env, e, impl, st);
        (exps,props,st_2) = elab_exp_list_ext(env, rest, impl, st_1); then (exp :: exps,p :: props,st_2);
  end matchcontinue;
end elab_exp_list_ext;

protected function elab_exp_ext
  input Env.Env in_env;
  input Absyn.Exp in_exp;
  input Boolean in_boolean;
  input Interactive_InteractiveSymbolTableOption in_interactive_interactivesymboltableoption;
  output Exp.Exp out_exp;
  output Types.Properties out_properties;
  output Interactive_InteractiveSymbolTableOption out_interactive_interactivesymboltableoption;
protected 
  type Interactive_InteractiveSymbolTableOption = Option<Interactive.InteractiveSymbolTable>;
algorithm 
  (out_exp,out_properties,out_interactive_interactivesymboltableoption):=
  matchcontinue (in_env,in_exp,in_boolean,in_interactive_interactivesymboltableoption)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      type Absyn_ExpList = list<Absyn.Exp>;
      type Absyn_NamedArgList = list<Absyn.NamedArg>;
      Exp.Exp dimp,arraycrefe,exp,e;
      TupleTypes_TTypeAbsyn_PathOption dimty;
      Boolean dim_const,impl;
      Types.Properties arraycrprop,prop;
      Env_FrameList env;
      Absyn.Exp call,arraycr,dim;
      Absyn_ExpList args;
      Absyn_NamedArgList nargs;
      Interactive_InteractiveSymbolTableOption st;
    case (env,(call = Absyn.CALL(Absyn.CREF_IDENT("size",_),Absyn.FUNCTIONARGS((args = {arraycr,dim}),nargs))),impl,st)
      equation 
        (dimp,Types.PROP(dimty,dim_const),_) = Static.elab_exp(env, dim, impl, NONE) "special case for  size";
        (arraycrefe,arraycrprop,_) = Static.elab_exp(env, arraycr, impl, NONE);
        exp = Exp.SIZE(arraycrefe,SOME(dimp)); then (exp,Types.PROP((Types.T_INTEGER({}),NONE),false),st);
    case (env,exp,impl,st)
      local Absyn.Exp exp;
      equation 
        (e,prop,st) = Static.elab_exp(env, exp, impl, st) "For all other expressions, use normal elaboration"; then (e,prop,st);
    case (env,exp,impl,st)
      local Absyn.Exp exp;
      equation 
        Debug.fprint("failtrace", "-elab_exp_ext failed\n"); then fail();
  end matchcontinue;
end elab_exp_ext;

protected function inst_ext_get_fargs " relation: inst_ext_get_fargs
  author: LS
 
  instantiates function arguments, i.e. actual parameters, in external declaration.
 "
  input Env.Env in_env;
  input Absyn.ExternalDecl in_externaldecl;
  input Boolean in_boolean;
  output DAE_ExtArgList out_dae_extarglist;
protected 
  type DAE_ExtArgList = list<DAE.ExtArg>;
algorithm 
  out_dae_extarglist:=
  matchcontinue (in_env,in_externaldecl,in_boolean)
    local
      type Exp_ExpList = list<Exp.Exp>;
      type Types_PropertiesList = list<Types.Properties>;
      type Env_FrameList = list<Env.Frame>;
      type StringOption = Option<String>;
      type Absyn_ComponentRefOption = Option<Absyn.ComponentRef>;
      type Absyn_ExpList = list<Absyn.Exp>;
      Exp_ExpList exps;
      Types_PropertiesList props;
      DAE_ExtArgList extargs;
      Env_FrameList env;
      StringOption id,lang;
      Absyn_ComponentRefOption retcr;
      Absyn_ExpList absexps;
      Boolean impl;
    case (env,Absyn.EXTERNALDECL(id,lang,retcr,absexps),impl) "impl"
      equation 
        (exps,props,_) = elab_exp_list_ext(env, absexps, impl, NONE);
        extargs = inst_ext_get_fargs2(env, exps, props); then extargs;
    case (_,_,impl)
      equation 
        Debug.fprint("failtrace", "- inst_ext_get_fargs failed\n"); then fail();
  end matchcontinue;
end inst_ext_get_fargs;

protected function inst_ext_get_fargs2 " relation: inst_ext_get_fargs2
  author: LS
 
  Helper function to inst_ext_get_fargs
 "
  input Env.Env in_env;
  input Exp_ExpList in_exp_explist;
  input Types_PropertiesList in_types_propertieslist;
  output DAE_ExtArgList out_dae_extarglist;
protected 
  type Exp_ExpList = list<Exp.Exp>;
  type Types_PropertiesList = list<Types.Properties>;
  type DAE_ExtArgList = list<DAE.ExtArg>;
algorithm 
  out_dae_extarglist:=
  matchcontinue (in_env,in_exp_explist,in_types_propertieslist)
    local
      type Env_FrameList = list<Env.Frame>;
      DAE_ExtArgList extargs;
      DAE.ExtArg extarg;
      Env_FrameList env;
      Exp.Exp e;
      Exp_ExpList exps;
      Types.Properties p;
      Types_PropertiesList props;
    case (_,{},_) then {}; 
    case (env,e :: exps,p :: props)
      equation 
        extargs = inst_ext_get_fargs2(env, exps, props);
        extarg = inst_ext_get_fargs_single(env, e, p); then extarg :: extargs;
  end matchcontinue;
end inst_ext_get_fargs2;

protected function inst_ext_get_fargs_single " relation: inst_ext_get_fargs_single
  author: LS
 
  Helper function to inst_ext_get_fargs2, does the work for one argument.
 "
  input Env.Env in_env;
  input Exp.Exp in_exp;
  input Types.Properties in_properties;
  output DAE.ExtArg out_extarg;
algorithm 
  out_extarg:=
  matchcontinue (in_env,in_exp,in_properties)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Env_FrameList = list<Env.Frame>;
      Types.Attributes attr;
      TupleTypes_TTypeAbsyn_PathOption ty,varty;
      Types.Binding bnd;
      Env_FrameList env;
      Exp.ComponentRef cref;
      Exp.Type crty;
      Boolean cnst;
      String crefstr,str;
      Exp.Exp dim,exp;
      Types.Properties prop;
    case (env,Exp.CREF(cref,crty),Types.PROP(ty,cnst))
      equation 
        (attr,ty,bnd) = Lookup.lookup_var_local(env, cref); then DAE.EXTARG(cref,attr,ty);
    case (env,Exp.CREF(cref,crty),Types.PROP(ty,cnst))
      equation 
        failure((attr,ty,bnd) = Lookup.lookup_var_local(env, cref));
        Print.print_buf("inst_ext_get_fargs_single lookup_var_local failed\n");
        crefstr = Exp.print_component_ref_str(cref);
        str = Util.string_append_list({"#-- Unknown component ",crefstr," in external clause\n"});
        Print.print_error_buf(str); then fail();
    case (env,Exp.SIZE(Exp.CREF(cref,crty),SOME(dim)),Types.PROP(ty,cnst))
      equation 
        (attr,varty,bnd) = Lookup.lookup_var_local(env, cref); then DAE.EXTARGSIZE(cref,attr,varty,dim);
    case (env,exp,Types.PROP(ty,cnst)) then DAE.EXTARGEXP(exp,ty); 
    case (_,exp,prop)
      equation 
        Debug.fprint("failtrace", "#-- inst_ext_get_fargs_single failed\n");
        Debug.fcall("failtrace", Exp.print_exp, exp);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_ext_get_fargs_single;

protected function inst_ext_get_rettype " relation: inst_ext_get_rettype
  author: LS
 
  Instantiates the return type of an external declaration.
 "
  input Env.Env in_env;
  input Absyn.ExternalDecl in_externaldecl;
  input Boolean in_boolean;
  output DAE.ExtArg out_extarg;
algorithm 
  out_extarg:=
  matchcontinue (in_env,in_externaldecl,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type StringOption = Option<String>;
      type Absyn_ExpList = list<Absyn.Exp>;
      Exp.Exp exp;
      Types.Properties prop;
      SCode.Accessibility acc;
      DAE.ExtArg extarg;
      Env_FrameList env;
      StringOption n,lang;
      Absyn.ComponentRef cref;
      Absyn_ExpList args;
      Boolean impl;
    case (_,Absyn.EXTERNALDECL(_,_,NONE,_),_) then DAE.NOEXTARG; 
    case (env,Absyn.EXTERNALDECL(n,lang,SOME(cref),args),impl)
      equation 
        (exp,prop,acc) = Static.elab_cref(env, cref, impl);
        extarg = inst_ext_get_fargs_single(env, exp, prop); then extarg;
    case (_,_,_)
      equation 
        Debug.fprint("failtrace", "#-- inst_ext_rettype failed\n"); then fail();
  end matchcontinue;
end inst_ext_get_rettype;

protected function inst_enumeration " relation: inst_enumeration
  author: PA
 
  This function takes an 'Ident' and list of strings, and returns an enumeration class.
"
  input SCode.Ident n;
  input StringList l;
  output SCode.Class out_class;
protected 
  type StringList = list<String>;
  type SCode_ElementList = list<SCode.Element>;
  SCode_ElementList comp;
algorithm 
  comp := make_enum_components(l);
  out_class := SCode.CLASS(n,false,false,SCode.R_ENUMERATION,
          SCode.PARTS(comp,{},{},{},{},NONE));
end inst_enumeration;

protected function make_enum_components " relation: make_enum_components
  author: PA
 
  This function takes a list of strings and returns the elements of 
  type 'EnumType' each corresponding to one of the enumeration values.
 "
  input StringList in_stringlist;
  output SCode_ElementList out_scode_elementlist;
protected 
  type StringList = list<String>;
  type SCode_ElementList = list<SCode.Element>;
algorithm 
  out_scode_elementlist:=
  matchcontinue (in_stringlist)
    local
      String str;
      SCode_ElementList els;
      StringList x;
    case ({str}) then {
          SCode.COMPONENT(str,true,false,false,
          SCode.ATTR({},false,SCode.RO,SCode.CONST,Absyn.BIDIR),Absyn.IDENT("EnumType"),SCode.NOMOD,NONE,NONE)}; 
    case (str :: (x = _ :: _))
      equation 
        els = make_enum_components(x); then SCode.COMPONENT(str,true,false,false,
          SCode.ATTR({},false,SCode.RO,SCode.CONST,Absyn.BIDIR),Absyn.IDENT("EnumType"),SCode.NOMOD,NONE,NONE) :: els;
  end matchcontinue;
end make_enum_components;

protected function dae_declare
  input Exp.ComponentRef in_componentref;
  input ClassInf.State in_state;
  input Types.Type in_type;
  input SCode.Attributes in_attributes;
  input Exp_ExpOption in_exp_expoption;
  input InstDims in_instdims;
  input DAE.StartValue in_startvalue;
  input DAE_VariableAttributesOption in_dae_variableattributesoption;
  input Absyn_CommentOption in_absyn_commentoption;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
  type Absyn_CommentOption = Option<Absyn.Comment>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_componentref,in_state,in_type,in_attributes,in_exp_expoption,in_instdims,in_startvalue,in_dae_variableattributesoption,in_absyn_commentoption)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      DAE.Flow flow_1;
      DAE_ElementList dae;
      Exp.ComponentRef vn;
      ClassInf.State ci_state;
      TupleTypes_TTypeAbsyn_PathOption ty;
      Boolean flow_;
      SCode.Variability par;
      Absyn.Direction dir;
      Exp_ExpOption e,start;
      InstDims inst_dims;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
    case (vn,ci_state,ty,SCode.ATTR(_,flow_,_,par,dir),e,inst_dims,start,dae_var_attr,comment)
      equation 
        flow_1 = DAE.to_flow(flow_, ci_state) "Debug.fprint( \"insttr\", \"dae_declare(\") &
	Debug.fcall ( \"insttr\", Exp.print_component_ref,vn) &
	Debug.fprint ( \"insttr\", \",\") &
	Debug.fcall ( \"insttr\", Types.print_type,ty)  &
	Debug.fprint (\"insttr\", \"...)\n\") &";
        dae = dae_declare2(vn, ty, flow_1, par, dir, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (_,_,_,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- dae_declare failed\n"); then fail();
  end matchcontinue;
end dae_declare;

protected function dae_declare2
  input Exp.ComponentRef in_componentref;
  input Types.Type in_type;
  input DAE.Flow in_flow;
  input SCode.Variability in_variability;
  input Absyn.Direction in_direction;
  input Exp_ExpOption in_exp_expoption;
  input InstDims in_instdims;
  input DAE.StartValue in_startvalue;
  input DAE_VariableAttributesOption in_dae_variableattributesoption;
  input Absyn_CommentOption in_absyn_commentoption;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
  type Absyn_CommentOption = Option<Absyn.Comment>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_componentref,in_type,in_flow,in_variability,in_direction,in_exp_expoption,in_instdims,in_startvalue,in_dae_variableattributesoption,in_absyn_commentoption)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      DAE_ElementList dae;
      Exp.ComponentRef vn;
      TupleTypes_TTypeAbsyn_PathOption ty;
      DAE.Flow flow_;
      Absyn.Direction dir;
      Exp_ExpOption e,start;
      InstDims inst_dims;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
    case (vn,ty,flow_,SCode.VAR,dir,e,inst_dims,start,dae_var_attr,comment)
      equation 
        dae = dae_declare3(vn, ty, flow_, DAE.VARIABLE, dir, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (vn,ty,flow_,SCode.DISCRETE,dir,e,inst_dims,start,dae_var_attr,comment)
      equation 
        dae = dae_declare3(vn, ty, flow_, DAE.DISCRETE, dir, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (vn,ty,flow_,SCode.PARAM,dir,e,inst_dims,start,dae_var_attr,comment)
      equation 
        dae = dae_declare3(vn, ty, flow_, DAE.PARAM, dir, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (vn,ty,flow_,SCode.CONST,dir,e,inst_dims,start,dae_var_attr,comment)
      equation 
        dae = dae_declare3(vn, ty, flow_, DAE.CONST, dir, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (vn,ty,flow_,SCode.STRUCTPARAM,dir,e,inst_dims,start,dae_var_attr,comment)
      equation 
        dae = dae_declare3(vn, ty, flow_, DAE.CONST, dir, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (_,_,_,_,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- dae_declare2 failed\n"); then fail();
  end matchcontinue;
end dae_declare2;

protected function dae_declare3
  input Exp.ComponentRef in_componentref;
  input Types.Type in_type;
  input DAE.Flow in_flow;
  input DAE.VarKind in_varkind;
  input Absyn.Direction in_direction;
  input Exp_ExpOption in_exp_expoption;
  input InstDims in_instdims;
  input DAE.StartValue in_startvalue;
  input DAE_VariableAttributesOption in_dae_variableattributesoption;
  input Absyn_CommentOption in_absyn_commentoption;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
  type Absyn_CommentOption = Option<Absyn.Comment>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_componentref,in_type,in_flow,in_varkind,in_direction,in_exp_expoption,in_instdims,in_startvalue,in_dae_variableattributesoption,in_absyn_commentoption)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      DAE_ElementList dae;
      Exp.ComponentRef vn;
      TupleTypes_TTypeAbsyn_PathOption ty;
      DAE.Flow fl;
      DAE.VarKind vk;
      Exp_ExpOption e,start;
      InstDims inst_dims;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
    case (vn,ty,fl,vk,Absyn.INPUT,e,inst_dims,start,dae_var_attr,comment)
      equation 
        dae = dae_declare4(vn, ty, fl, vk, DAE.INPUT, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (vn,ty,fl,vk,Absyn.OUTPUT,e,inst_dims,start,dae_var_attr,comment)
      equation 
        dae = dae_declare4(vn, ty, fl, vk, DAE.OUTPUT, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (vn,ty,fl,vk,Absyn.BIDIR,e,inst_dims,start,dae_var_attr,comment)
      equation 
        dae = dae_declare4(vn, ty, fl, vk, DAE.BIDIR, e, inst_dims, start, 
          dae_var_attr, comment); then dae;
    case (_,_,_,_,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "#- dae_declare3 failed\n"); then fail();
  end matchcontinue;
end dae_declare3;

protected function dae_declare4
  input Exp.ComponentRef in_componentref;
  input Types.Type in_type;
  input DAE.Flow in_flow;
  input DAE.VarKind in_varkind;
  input DAE.VarDirection in_vardirection;
  input Exp_ExpOption in_exp_expoption;
  input InstDims in_instdims;
  input DAE.StartValue in_startvalue;
  input DAE_VariableAttributesOption in_dae_variableattributesoption;
  input Absyn_CommentOption in_absyn_commentoption;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
  type Absyn_CommentOption = Option<Absyn.Comment>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_componentref,in_type,in_flow,in_varkind,in_vardirection,in_exp_expoption,in_instdims,in_startvalue,in_dae_variableattributesoption,in_absyn_commentoption)
    local
      type StringList = list<String>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.ComponentRef vn,c;
      DAE.Flow fl;
      DAE.VarKind kind;
      DAE.VarDirection dir;
      Exp_ExpOption e,start;
      InstDims inst_dims;
      DAE_VariableAttributesOption dae_var_attr;
      Absyn_CommentOption comment;
      StringList l;
      TupleTypes_TTypeAbsyn_PathOption ty;
    case (vn,(Types.T_INTEGER(_),_),fl,kind,dir,e,inst_dims,start,dae_var_attr,comment) then {
          DAE.VAR(vn,kind,dir,DAE.INT,e,inst_dims,start,fl,{},dae_var_attr,
          comment)}; 
    case (vn,(Types.T_REAL(_),_),fl,kind,dir,e,inst_dims,start,dae_var_attr,comment) then {
          DAE.VAR(vn,kind,dir,DAE.REAL,e,inst_dims,start,fl,{},dae_var_attr,
          comment)}; 
    case (vn,(Types.T_BOOL(_),_),fl,kind,dir,e,inst_dims,start,dae_var_attr,comment) then {
          DAE.VAR(vn,kind,dir,DAE.BOOL,e,inst_dims,start,fl,{},dae_var_attr,
          comment)}; 
    case (vn,(Types.T_STRING(_),_),fl,kind,dir,e,inst_dims,start,dae_var_attr,comment) then {
          DAE.VAR(vn,kind,dir,DAE.STRING,e,inst_dims,start,fl,{},
          dae_var_attr,comment)}; 
    case (vn,(Types.T_ENUM,_),fl,kind,dir,e,inst_dims,start,dae_var_attr,comment) then {}; 
    case (vn,(Types.T_ENUMERATION(l,_),_),fl,kind,dir,e,inst_dims,start,dae_var_attr,comment) then {
          DAE.VAR(vn,kind,dir,DAE.ENUMERATION(l),e,inst_dims,start,fl,{},
          dae_var_attr,comment)}; 
    case (c,ty,_,_,_,_,_,_,_,_) then {}; 
  end matchcontinue;
end dae_declare4;

protected function inst_equation "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.Equation in_equation;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_equation,in_boolean)
    case (env,mods,pre,csets,ci_state,SCode.EQUATION(eq,bc),impl) "impl"
      equation 
        env_1 = get_derived_env(env, bc) "Equation inherited from base class ";
        (dae,_,csets_1,ci_state_1) = inst_equation_common(env_1, mods, pre, csets, ci_state, eq, NON_INITIAL, impl); then (dae,env,csets_1,ci_state_1);
    case (_,_,_,_,_,_,impl)
      equation 
        Debug.fprint("failtrace", "- inst_equation failed\n"); then fail();
  end matchcontinue;
end inst_equation;

protected function inst_e_equation "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.EEquation in_eequation;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_eequation,in_boolean)
    case (env,mods,pre,csets,ci_state,eq,impl) "impl"
      equation 
        (dae,_,csets_1,ci_state_1) = inst_equation_common(env, mods, pre, csets, ci_state, eq, NON_INITIAL, impl); then (dae,env,csets_1,ci_state_1);
  end matchcontinue;
end inst_e_equation;

protected function inst_initialequation "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.Equation in_equation;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_equation,in_boolean)
    case (env,mods,pre,csets,ci_state,SCode.EQUATION(eq,bc),impl) "impl"
      equation 
        env_1 = get_derived_env(env, bc) "Equation inherited from base class ";
        (dae,_,csets_1,ci_state_1) = inst_equation_common(env_1, mods, pre, csets, ci_state, eq, INITIAL, impl); then (dae,env,csets_1,ci_state_1);
    case (_,_,_,_,_,_,impl)
      equation 
        Debug.fprint("failtrace", "- inst_initialequation failed\n"); then fail();
  end matchcontinue;
end inst_initialequation;

protected function inst_e_initialequation "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.EEquation in_eequation;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_eequation,in_boolean)
    case (env,mods,pre,csets,ci_state,eq,impl) "impl"
      equation 
        (dae,_,csets_1,ci_state_1) = inst_equation_common(env, mods, pre, csets, ci_state, eq, INITIAL, impl); then (dae,env,csets_1,ci_state_1);
  end matchcontinue;
end inst_e_initialequation;

protected function inst_equation_common "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.EEquation in_eequation;
  input Initial in_initial;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_eequation,in_initial,in_boolean)
    local
      type Connect_SetList = list<Connect.Set>;
      type Env_FrameList = list<Env.Frame>;
      type Interactive_InteractiveSymbolTableOption = Option<Interactive.InteractiveSymbolTable>;
      type SCode_EEquationList = list<SCode.EEquation>;
      type TupleAbsyn_ExpSCode_EEquationList = tuple<Absyn.Exp,SCode_EEquationList>;
      type TupleAbsyn_ExpSCode_EEquationListList = list<TupleAbsyn_ExpSCode_EEquationList>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Connect_SetList csets_1,csets;
      DAE_ElementList dae,dae1,dae2;
      ClassInf.State ci_state_1,ci_state,ci_state_2;
      Env_FrameList env,env_1,env_2;
      Types.Mod mods,mod;
      Prefix.Prefix pre;
      Absyn.ComponentRef c1,c2,cr;
      Initial initial_;
      Boolean impl,cond;
      String n,i;
      Absyn.Exp e2,e1,e,ee;
      Exp.Exp e1_1,e2_1,e1_2,e2_2,e_1;
      Types.Properties prop1,prop2;
      SCode_EEquationList b,tb,fb,el,eel;
      TupleAbsyn_ExpSCode_EEquationListList eex;
      TupleTypes_TTypeAbsyn_PathOption id_t;
      Values.Value v;
      Exp.ComponentRef cr_1;
    case (env,mods,pre,csets,ci_state,SCode.EQ_CONNECT(c1,c2),initial_,impl) "impl   
	 ** Handle connect statements
	 *"
      equation 
        (csets_1,dae) = inst_connect(csets, env, pre, c1, c2, impl);
        ci_state_1 = inst_equation_common_ci_trans(ci_state, initial_); then (dae,env,csets_1,ci_state_1);
    case (env,mods,pre,csets,ci_state,SCode.EQ_EQUALS(Absyn.CREF(Absyn.CREF_IDENT(n,{})),e2),initial_,impl) " The following rule handles shadowed (replaced) equations.   If an equation has a simple name on the left-hand side,   and that component has an equation modifier, this equation   is discarded. "
      equation 
        (Types.VAR(_,_,_,_,Types.EQBOUND(_,_)),_,_) = Lookup.lookup_ident_local(env, n);
        ci_state_1 = inst_equation_common_ci_trans(ci_state, initial_); then ({},env,csets,ci_state_1);
    case (env,mods,pre,csets,ci_state,SCode.EQ_EQUALS(e1,e2),initial_,impl)
      local Interactive_InteractiveSymbolTableOption c1,c2;
      equation 
        (e1_1,prop1,c1) = Static.elab_exp(env, e1, impl, NONE) "
	 Do static analysis and constant evaluation of expressions. 
	 Gives expression and properties 
	 (Type * bool | (Type * Const as (bool | Const list))).
	 For a function, it checks the funtion name. 
	 Also the function call's in parameters are type checked with
	 the functions definition's inparameters. This is done with
	 regard to the position of the input arguments.

	 Returns the output parameters from the funktion.
	";
        (e2_1,prop2,c2) = Static.elab_exp(env, e2, impl, NONE);
        e1_2 = Prefix.prefix_exp(env, e1_1, pre);
        e2_2 = Prefix.prefix_exp(env, e2_1, pre);
        dae = inst_eq_equation(e1_2, prop1, e2_2, prop2, initial_, impl) "Check that the lefthandside and the righthandside get along. ";
        ci_state_1 = inst_equation_common_ci_trans(ci_state, initial_); then (dae,env,csets,ci_state_1);
    case (env,mod,pre,csets,ci_state,SCode.EQ_IF(e,tb,fb),NON_INITIAL,impl) "  `if' statements
	 **
	 ** If statements are instantiated by evaluating the
	 ** conditional expression, and selecting the branch that
	 ** should be used.
	 * When the condition is constant evaluate it and select the correct branch "
      equation 
        (e_1,Types.PROP((Types.T_BOOL(_),_),true),_) = Static.elab_exp(env, e, impl, NONE);
        (Values.BOOL(cond),_) = Ceval.ceval(env, e_1, impl, NONE, NONE);
        b = select(cond, tb, fb);
        (dae,env_1,csets_1,ci_state_1) = inst_list(env, mod, pre, csets, ci_state, inst_e_equation, b, impl); then (dae,env_1,csets_1,ci_state_1);
    case (env,mod,pre,csets,ci_state,SCode.EQ_IF(e,tb,fb),INITIAL,impl)
      equation 
        (e_1,Types.PROP((Types.T_BOOL(_),_),true),_) = Static.elab_exp(env, e, impl, NONE);
        (Values.BOOL(cond),_) = Ceval.ceval(env, e_1, impl, NONE, NONE);
        b = select(cond, tb, fb);
        (dae,env_1,csets_1,ci_state_1) = inst_list(env, mod, pre, csets, ci_state, inst_e_initialequation, b, 
          impl); then (dae,env_1,csets_1,ci_state_1);
    case (env,mod,pre,csets,ci_state,SCode.EQ_IF(e,tb,fb),NON_INITIAL,impl)
      equation 
        (e_1,Types.PROP((Types.T_BOOL(_),_),false),_) = Static.elab_exp(env, e, impl, NONE);
        (dae1,env_1,_,ci_state_1) = inst_list(env, mod, pre, csets, ci_state, inst_e_equation, tb, impl);
        (dae2,env_2,_,ci_state_2) = inst_list(env_1, mod, pre, csets, ci_state, inst_e_equation, fb, impl) "There are no connections inside if-clauses. "; then ({DAE.IF_EQUATION(e_1,dae1,dae2)},env_1,csets,ci_state_1);
    case (env,mod,pre,csets,ci_state,SCode.EQ_WHEN(e,el,(ee,eel) :: eex),(initial_ = NON_INITIAL),impl) "  `when equation' statement, modelica 1.1 
	 **
	 ** When statements are instantiated by evaluating the
	 ** conditional expression.
	 *"
      local DAE.Element dae2;
      equation 
        (e_1,_,_) = Static.elab_exp(env, e, impl, NONE);
        (dae1,env_1,_,_) = inst_list(env, mod, pre, csets, ci_state, inst_e_equation, el, impl);
        (dae2 :: _,env_2,_,ci_state_1) = inst_equation_common(env_1, mod, pre, csets, ci_state, 
          SCode.EQ_WHEN(ee,eel,eex), initial_, impl);
        ci_state_2 = inst_equation_common_ci_trans(ci_state_1, initial_); then ({DAE.WHEN_EQUATION(e_1,dae1,SOME(dae2))},env_2,csets,ci_state_2);
    case (env,mod,pre,csets,ci_state,SCode.EQ_WHEN(e,el,{}),(initial_ = NON_INITIAL),impl)
      equation 
        (e_1,_,_) = Static.elab_exp(env, e, impl, NONE);
        (dae1,env_1,_,_) = inst_list(env, mod, pre, csets, ci_state, inst_e_equation, el, impl);
        ci_state_1 = inst_equation_common_ci_trans(ci_state, initial_); then ({DAE.WHEN_EQUATION(e_1,dae1,NONE)},env_1,csets,ci_state_1);
    case (env,mod,pre,csets,ci_state,SCode.EQ_FOR(i,e,el),initial_,impl) "seems unnecessary to handle when equations that are initial    `for' loops
	 **
	 ** The loop expression is evaluated to a constant array of
	 ** integers, and then the loop is unrolled.
	 **
	 *  FIXME: Why lookup after add_for_loop_scope ? *"
      equation 
        (e_1,Types.PROP((Types.T_ARRAY(Types.DIM(_),id_t),_),true),_) = Static.elab_exp(env, e, impl, NONE) "Debug.fprintln (\"insttr\", \"inst_equation_common_eqfor_1\") &";
        env_1 = add_for_loop_scope(env, i, id_t) "Debug.fprintln (\"insti\", \"for expression elaborated\") &";
        (Types.ATTR(false,SCode.RW,SCode.VAR,_),(Types.T_INTEGER(_),_),Types.UNBOUND) = Lookup.lookup_var(env_1, Exp.CREF_IDENT(i,{})) "	Debug.fprintln (\"insti\", \"loop-variable added to scope\") &";
        (v,_) = Ceval.ceval(env, e_1, impl, NONE, NONE) "	Debug.fprintln (\"insti\", \"loop variable looked up\") & FIXME: Check bounds ";
        (dae,csets_1) = unroll(env_1, mod, pre, csets, ci_state, i, v, el, initial_, impl) "	Debug.fprintln (\"insti\", \"for expression evaluated\") &";
        ci_state_1 = inst_equation_common_ci_trans(ci_state, initial_) "	Debug.fprintln (\"insti\", \"for expression unrolled\") & 	& Debug.fprintln (\"insttr\", \"inst_equation_common_eqfor_1 succeeded\")"; then (dae,env,csets_1,ci_state_1);
    case (env,mod,pre,csets,ci_state,SCode.EQ_FOR(i,e,el),initial_,impl)
      equation 
        (Types.ATTR(false,SCode.RW,SCode.VAR,_),(Types.T_INTEGER(_),_),Types.UNBOUND) = Lookup.lookup_var(env, Exp.CREF_IDENT(i,{})) "for loops with non-constant iteration bounds ";
        (e_1,Types.PROP((Types.T_ARRAY(Types.DIM(_),(Types.T_INTEGER(_),_)),_),false),_) = Static.elab_exp(env, e, impl, NONE);
        Print.print_error_buf("# Error: Non-constant iteration bounds not supported\n"); then fail();
    case (env,mod,pre,csets,ci_state,SCode.EQ_ASSERT(e1,e2),initial_,impl)
      equation 
        (e1_1,Types.PROP((Types.T_BOOL(_),_),_),_) = Static.elab_exp(env, e1, impl, NONE) "assert statement ";
        (e2_1,Types.PROP((Types.T_STRING(_),_),_),_) = Static.elab_exp(env, e2, impl, NONE); then ({
          DAE.ASSERT(Exp.CALL(Absyn.IDENT("assert"),{e1_1,e2_1},false,false))},env,csets,ci_state);
    case (env,mod,pre,csets,ci_state,SCode.EQ_REINIT(cr,e2),initial_,impl)
      equation 
        (Exp.CREF(cr_1,_),_,_) = Static.elab_cref(env, cr, impl) "reinit statement ";
        (e2_1,_,_) = Static.elab_exp(env, e2, impl, NONE); then ({DAE.REINIT(cr_1,e2_1)},env,csets,ci_state);
    case (_,_,_,_,_,eqn,_,impl)
      equation 
        Debug.fprint("failtrace", "- inst_equation_common failed for eqn: ");
        s = SCode.equation_str(eqn);
        Debug.fprint("failtrace", s);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_equation_common;

protected function inst_equation_common_ci_trans
  input ClassInf.State in_state;
  input Initial in_initial;
  output ClassInf.State out_state;
algorithm 
  out_state:=
  matchcontinue (in_state,in_initial)
    local ClassInf.State ci_state_1,ci_state;
    case (ci_state,NON_INITIAL)
      equation 
        ci_state_1 = ClassInf.trans(ci_state, ClassInf.FOUND_EQUATION); then ci_state_1;
    case (ci_state,INITIAL) then ci_state; 
  end matchcontinue;
end inst_equation_common_ci_trans;

protected function add_for_loop_scope " relation: add_for_loop_scope
  author: HJ
 
  Adds a scope on the environment used in for loops.
  The name of the scope is for_scope_name, defined as a value.
 "
  input Env.Env env;
  input Ident i;
  input Types.Type typ;
  output Env.Env env_2;
protected 
  type Env_FrameList = list<Env.Frame>;
  Env_FrameList env_1,env_2;
algorithm 
  env_1 := Env.open_scope(env, false, SOME(for_scope_name));
  env_2 := Env.extend_frame_v(env_1, 
          Types.VAR(i,Types.ATTR(false,SCode.RW,SCode.VAR,Absyn.BIDIR),false,
          typ,Types.UNBOUND), NONE, false);
end add_for_loop_scope;

protected function is_parameter " relation: is_parameter
  author: LS
  
  Succeeds if a variable is a parameter.
 "
  input Exp.ComponentRef cr;
  input Env.Env env;
algorithm 
  (Types.ATTR(fl,acc,SCode.PARAM,dir),ty,bnd) := Lookup.lookup_var(env, cr) "Env.print_env env &";
end is_parameter;

protected function inst_eq_equation "impl"
  input Exp.Exp in_exp1;
  input Types.Properties in_properties2;
  input Exp.Exp in_exp3;
  input Types.Properties in_properties4;
  input Initial in_initial5;
  input Boolean in_boolean6;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_exp1,in_properties2,in_exp3,in_properties4,in_initial5,in_boolean6)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.Exp e1_1,e1,e2,e2_1;
      DAE_ElementList dae;
      Types.Properties p1,p2;
      TupleTypes_TTypeAbsyn_PathOption t1,t2,t;
      Initial initial_;
      Boolean impl;
      String str;
    case (e1,(p1 = Types.PROP(t1,_)),e2,(p2 = Types.PROP(t2,_)),initial_,impl) "impl PR. e1= lefthandside, e2=righthandside
	* This seem to be a strange function. 
	* wich rule is matched? or is both rules matched?
	 LS: Static.type_convert in Static.match_prop can probably fail,
	 * then the first rule will not match. Question if whether the second
	 * rule can match in that case.
	  This rule is matched first, if it fail the next rule is matched.
	  If it fails then this rule is matched. "
      equation 
        e1_1 = Types.match_prop(e1, p1, p2) "Debug.print(\"\ninst_eq_equation (match e1) PROP, PROP\") &";
        dae = inst_eq_equation_2(e1_1, e2, t2, initial_); then dae;
    case (e1,(p1 = Types.PROP(t1,_)),e2,(p2 = Types.PROP(t2,_)),initial_,impl) "If it fails then this rule is matched. "
      equation 
        e2_1 = Types.match_prop(e2, p2, p1) "Debug.print(\"\ninst_eq_equation (match e2) PROP, PROP\") &";
        dae = inst_eq_equation_2(e1, e2_1, t1, initial_) "	Debug.print(\"\n Second rule of function_ inst_eq_equation \") & 	& Debug.print(\"\n Second rule complete. \") "; then dae;
    case (e1,(p1 = Types.PROP_TUPLE(t1,_)),e2,(p2 = Types.PROP_TUPLE(t2,_)),initial_,impl) "PR. "
      equation 
        e1_1 = Types.match_prop(e1, p1, p2) "Debug.print(\"\ninst_eq_equation(e1) PROP_TUPLE, PROP_TUPLE\") & Exp.print_exp (e1) & ";
        dae = inst_eq_equation_2(e1_1, e2, t2, initial_) "Exp.print_exp (e1') & "; then dae;
    case (e1,(p1 = Types.PROP_TUPLE(t1,_)),e2,(p2 = Types.PROP_TUPLE(t2,_)),initial_,impl) "PR.  
	   ** An assignment to a varaible of T_ENUMERATION type is an explicit 
	   ** assignment to the value componnent of the enumeration, i.e. having 
	   ** a type T_ENUM
	   *"
      equation 
        e2_1 = Types.match_prop(e2, p2, p1) "Debug.print(\"\ninst_eq_equation(e2) PROP_TUPLE, PROP_TUPLE\") &
	Debug.print \"\n About to do a static match e2. \" &";
        dae = inst_eq_equation_2(e1, e2_1, t1, initial_) "	Debug.print(\"\n Second rule of function_ inst_eq_equation \") & 	& Debug.print(\"\n Second rule complete. \") "; then dae;
    case ((e1 = Exp.CREF(_,_)),Types.PROP((Types.T_ENUMERATION(_,_),_),_),e2,Types.PROP((t = (Types.T_ENUM,_)),_),initial_,impl) "
	   ** An assignment to a varaible of T_ENUMERATION type is an explicit 
	   ** assignment to the value componnent of the enumeration, i.e. having 
	   ** a type T_ENUM
	   *"
      equation 
        dae = inst_eq_equation_2(e1, e2, t, initial_) "Debug.fprint (\"insttr\", \"Found assignment to T_ENUMERATION type. Rhs type must be T_ENUM or T_ENUMERATION.\n\") &"; then dae;
    case ((e1 = Exp.CREF(_,_)),Types.PROP((Types.T_ENUMERATION(_,_),_),_),e2,Types.PROP((t = (Types.T_ENUMERATION(_,_),_)),_),initial_,impl)
      equation 
        dae = inst_eq_equation_2(e1, e2, t, initial_) "Debug.fprint (\"insttr\", \"Found assignment to T_ENUMERATION type. Rhs type must be T_ENUM or T_ENUMERATION.\n\") &"; then dae;
    case (e1,Types.PROP(t1,_),e2,Types.PROP(t2,_),initial_,impl)
      equation 
        Print.print_error_buf("# Illegal types in equation\n");
        Print.print_error_buf("    lhs: ");
        str = Exp.print_exp_str(e1);
        Print.print_error_buf(str);
        Print.print_error_buf(" :: ");
        str = Types.unparse_type(t1);
        Print.print_error_buf(str);
        Print.print_error_buf("\n");
        Print.print_error_buf("    rhs: ");
        str = Exp.print_exp_str(e2);
        Print.print_error_buf(str);
        Print.print_error_buf(" :: ");
        str = Types.unparse_type(t2);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
    case (e1,p1,e2,p2,initial_,impl) "No match. "
      equation 
        Print.print_error_buf("# No match in inst_eq_equation\n");
        Print.print_error_buf("    lhs: ");
        str = Exp.print_exp_str(e1);
        Print.print_error_buf(str);
        Print.print_error_buf(" :: ");
        str = Types.print_prop_str(p1);
        Print.print_error_buf(str);
        Print.print_error_buf("\n");
        Print.print_error_buf("    rhs: ");
        str = Exp.print_exp_str(e2);
        Print.print_error_buf(str);
        Print.print_error_buf(" :: ");
        str = Types.print_prop_str(p2);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
  end matchcontinue;
end inst_eq_equation;

protected function inst_eq_equation_2
  input Exp.Exp in_exp1;
  input Exp.Exp in_exp2;
  input Types.Type in_type3;
  input Initial in_initial4;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_exp1,in_exp2,in_type3,in_initial4)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Types_VarList = list<Types.Var>;
      DAE.Element dae;
      Exp.Exp e1,e2;
      Initial initial_;
      Exp.ComponentRef cr,c1_1,c2_1,c1,c2;
      Exp.Type t,t1,t2;
      DAE_ElementList dae1,dae2,decl;
      Types.ArrayDim ad;
      ClassInf.State cs;
      String n;
      Types_VarList vs;
      Absyn_PathOption p;
    case (e1,e2,(Types.T_INTEGER(_),_),initial_)
      equation 
        dae = make_dae_equation(e1, e2, initial_); then {dae};
    case (e1,e2,(Types.T_REAL(_),_),initial_)
      equation 
        dae = make_dae_equation(e1, e2, initial_); then {dae};
    case (e1,e2,(Types.T_STRING(_),_),initial_)
      equation 
        dae = make_dae_equation(e1, e2, initial_); then {dae};
    case (e1,e2,(Types.T_BOOL(_),_),initial_)
      equation 
        dae = make_dae_equation(e1, e2, initial_); then {dae};
    case (Exp.CREF(cr,t),e2,(Types.T_ENUM,_),initial_)
      equation 
        dae = make_dae_define(cr, e2, initial_); then {dae};
    case (Exp.CREF(cr,t),e2,(Types.T_ENUMERATION(_,_),_),initial_)
      equation 
        dae = make_dae_define(cr, e2, initial_); then {dae};
    case (e1,e2,(Types.T_ARRAY(ad,t),_),initial_)
      local
        DAE_ElementList dae;
        TupleTypes_TTypeAbsyn_PathOption t;
      equation 
        true = RTOpts.split_arrays();
        false = Exp.contain_functioncall(e1);
        false = Exp.contain_functioncall(e2);
        dae = inst_array_equation(e1, e2, ad, t, initial_) "	 For debugging. 	Print.print_buf \"  Array equation: \" &
	Exp.print_exp e1 & Print.print_buf \" = \" &
	Exp.print_exp e2 & Print.print_buf \"\n\" & "; then dae;
    case (e1,e2,(Types.T_TUPLE(_),_),initial_) "PR. The muliple outputs from functions is outputed as is. "
      equation 
        dae = make_dae_equation(e1, e2, initial_); then {dae};
    case (e1,e2,(Types.T_ARRAY(ad,t),_),initial_)
      local TupleTypes_TTypeAbsyn_PathOption t;
      equation 
        dae = make_dae_equation(e1, e2, initial_) "RTOpts.split_arrays => false &"; then {dae};
    case (e1,e2,(Types.T_COMPLEX(_,{}),_),initial_) then {}; 
    case (Exp.CREF(c1,t1),Exp.CREF(c2,t2),(Types.T_COMPLEX(cs,Types.VAR(n,_,_,t,_) :: vs),p),initial_)
      local
        DAE_ElementList dae;
        TupleTypes_TTypeAbsyn_PathOption t;
      equation 
        c1_1 = Exp.extend_cref(c1, n, {});
        c2_1 = Exp.extend_cref(c2, n, {});
        dae1 = inst_eq_equation_2(Exp.CREF(c1_1,t1), Exp.CREF(c2_1,t2), t, initial_);
        dae2 = inst_eq_equation_2(Exp.CREF(c1,t1), Exp.CREF(c2,t2), 
          (Types.T_COMPLEX(cs,vs),p), initial_);
        dae = list_append(dae1, dae2); then dae;
    case (e1,(e2 = Exp.CREF(_,_)),(t = (Types.T_COMPLEX(_,_),_)),initial_) " When the type of the expressions is a complex type, and   the left-hand side of the equation is not a component   reference, a new variable is introduced to be able to   dereference components of the expression.     This is rather ugly, since it doesn't really solve the   problem of describing the semantics.  Now the semantics of   composite equations are defined in terms of other   composite equations.  To make this a little cleaner, the   equation that equates the new name to the expression is   stored using DAE.DEFINE rather than DAE.EQUATION.  This   makes it a little clearer. "
      local
        Exp.ComponentRef n;
        DAE_ElementList dae;
        TupleTypes_TTypeAbsyn_PathOption t;
      equation 
        n = new_ident();
        decl = dae_declare(n, ClassInf.UNKNOWN(""), t, 
          SCode.ATTR({},false,SCode.RW,SCode.VAR,Absyn.BIDIR), NONE, {}, NONE, NONE, NONE);
        dae1 = inst_eq_equation_2(Exp.CREF(n,Exp.OTHER), e2, t, initial_);
        dae = list_append(decl, DAE.DEFINE(n,e1) :: dae1); then dae;
    case (e1,e2,(t = (Types.T_COMPLEX(_,_),_)),initial_) " When the right-hand side is not a component reference a   similar trick is applied.  This also catched the case   where none of the sides is a component reference "
      local
        Exp.ComponentRef n;
        DAE_ElementList dae;
        TupleTypes_TTypeAbsyn_PathOption t;
      equation 
        n = new_ident();
        decl = dae_declare(n, ClassInf.UNKNOWN(""), t, 
          SCode.ATTR({},false,SCode.RW,SCode.VAR,Absyn.BIDIR), NONE, {}, NONE, NONE, NONE);
        dae1 = inst_eq_equation_2(e1, Exp.CREF(n,Exp.OTHER), t, initial_);
        dae = list_append(decl, DAE.DEFINE(n,e2) :: dae1); then dae;
    case (e1,e2,t,initial_)
      local TupleTypes_TTypeAbsyn_PathOption t;
      equation 
        Debug.fprint("failtrace", "- inst_eq_equation_2 failed\n exp1=");
        Debug.fcall("failtrace", Exp.print_exp, e1);
        Debug.fprint("failtrace", " exp2=");
        Debug.fcall("failtrace", Exp.print_exp, e2);
        Debug.fprint("failtrace", " type =");
        Debug.fcall("failtrace", Types.print_type, t);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_eq_equation_2;

protected function make_dae_equation " relation: make_dae_equation
  author: LS, ELN 
 
  Constructs an equation in the DAE, they can be either an initial equation or an
  ordinary equation.
"
  input Exp.Exp in_exp1;
  input Exp.Exp in_exp2;
  input Initial in_initial3;
  output DAE.Element out_element;
algorithm 
  out_element:=
  matchcontinue (in_exp1,in_exp2,in_initial3)
    local Exp.Exp e1,e2;
    case (e1,e2,NON_INITIAL) then DAE.EQUATION(e1,e2); 
    case (e1,e2,INITIAL) then DAE.INITIALEQUATION(e1,e2); 
  end matchcontinue;
end make_dae_equation;

protected function make_dae_define " relation: make_dae_define
  author: LS, ELN 
 
 "
  input Exp.ComponentRef in_componentref;
  input Exp.Exp in_exp;
  input Initial in_initial;
  output DAE.Element out_element;
algorithm 
  out_element:=
  matchcontinue (in_componentref,in_exp,in_initial)
    local
      Exp.ComponentRef cr;
      Exp.Exp e2;
    case (cr,e2,NON_INITIAL) then DAE.DEFINE(cr,e2); 
    case (cr,e2,INITIAL) then DAE.INITIALDEFINE(cr,e2); 
  end matchcontinue;
end make_dae_define;

protected function inst_array_equation
  input Exp.Exp in_exp1;
  input Exp.Exp in_exp2;
  input Types.ArrayDim in_arraydim3;
  input Types.Type in_type4;
  input Initial in_initial5;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_exp1,in_exp2,in_arraydim3,in_type4,in_initial5)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      String str;
      Exp.Exp e1,e2;
      TupleTypes_TTypeAbsyn_PathOption t;
      Initial initial_;
      DAE_ElementList dae;
      Integer sz;
    case (e1,e2,Types.DIM(NONE),t,initial_)
      equation 
        Print.print_error_buf(
          "# Cannot instantiate equations with arrays of unknown size\n");
        Print.print_error_buf("    ");
        str = Exp.print_exp_str(e1);
        Print.print_error_buf(str);
        Print.print_error_buf(" = ");
        str = Exp.print_exp_str(e2);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
    case (e1,e2,Types.DIM(SOME(sz)),t,initial_)
      equation 
        dae = inst_array_el_eq(e1, e2, t, 1, sz, initial_); then dae;
    case (_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_array_equation failed\n"); then fail();
  end matchcontinue;
end inst_array_equation;

protected function inst_array_el_eq
  input Exp.Exp in_exp1;
  input Exp.Exp in_exp2;
  input Types.Type in_type3;
  input Integer in_integer4;
  input Integer in_integer5;
  input Initial in_initial6;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_exp1,in_exp2,in_type3,in_integer4,in_integer5,in_initial6)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.Exp e1_1,e2_1,e1,e2;
      DAE_ElementList dae1,dae2,dae;
      Integer i_1,i,sz;
      TupleTypes_TTypeAbsyn_PathOption t;
      Initial initial_;
    case (e1,e2,t,i,sz,initial_)
      equation 
        (i <= sz) = true;
        e1_1 = Exp.simplify(Exp.ASUB(e1,i));
        e2_1 = Exp.simplify(Exp.ASUB(e2,i));
        dae1 = inst_eq_equation_2(e1_1, e2_1, t, initial_);
        i_1 = i + 1;
        dae2 = inst_array_el_eq(e1, e2, t, i_1, sz, initial_);
        dae = list_append(dae1, dae2); then dae;
    case (e1,e2,t,i,sz,initial_)
      equation 
        (i <= sz) = false; then {};
    case (_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_array_el_eq failed\n"); then fail();
  end matchcontinue;
end inst_array_el_eq;

protected function unroll "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input Ident in_ident;
  input Values.Value in_value;
  input SCode_EEquationList in_scode_eequationlist;
  input Initial in_initial;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
  output Connect.Sets out_sets;
protected 
  type SCode_EEquationList = list<SCode.EEquation>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_sets):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_ident,in_value,in_scode_eequationlist,in_initial,in_boolean)
    case (_,_,_,csets,_,_,Values.ARRAY({}),_,_,_) then ({},csets); 
    case (env,mods,pre,csets,ci_state,i,Values.ARRAY(fst :: rest),eqs,(initial_ = NON_INITIAL),impl)
      equation 
        env_1 = Env.open_scope(env, false, SOME(for_scope_name));
        env_2 = Env.extend_frame_v(env_1, 
          Types.VAR(i,Types.ATTR(false,SCode.RO,SCode.CONST,Absyn.BIDIR),true,
          (Types.T_INTEGER({}),NONE),Types.VALBOUND(fst)), NONE, false);
        (dae1,env_3,csets_1,ci_state_1) = inst_list(env_2, mods, pre, csets, ci_state, inst_e_equation, eqs, 
          impl);
        (dae2,csets_2) = unroll(env, mods, pre, csets_1, ci_state_1, i, 
          Values.ARRAY(rest), eqs, initial_, impl);
        dae = list_append(dae1, dae2); then (dae,csets_2);
    case (env,mods,pre,csets,ci_state,i,Values.ARRAY(fst :: rest),eqs,(initial_ = INITIAL),impl)
      equation 
        env_1 = Env.open_scope(env, false, SOME(for_scope_name));
        env_2 = Env.extend_frame_v(env_1, 
          Types.VAR(i,Types.ATTR(false,SCode.RO,SCode.CONST,Absyn.BIDIR),true,
          (Types.T_INTEGER({}),NONE),Types.VALBOUND(fst)), NONE, false);
        (dae1,env_3,csets_1,ci_state_1) = inst_list(env_2, mods, pre, csets, ci_state, inst_e_initialequation, 
          eqs, impl);
        (dae2,csets_2) = unroll(env, mods, pre, csets_1, ci_state_1, i, 
          Values.ARRAY(rest), eqs, initial_, impl);
        dae = list_append(dae1, dae2); then (dae,csets_2);
    case (_,_,_,_,_,_,v,_,_,_)
      equation 
        Debug.fprint("failtrace", "- unroll ");
        Debug.fcall("failtrace", Values.print_val, v);
        Debug.fprint("failtrace", " failed\n"); then fail();
  end matchcontinue;
end unroll;

protected function inst_algorithm "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.Algorithm in_algorithm;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_algorithm,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type Algorithm_StatementList = list<Algorithm.Statement>;
      type Connect_SetList = list<Connect.Set>;
      type Absyn_AlgorithmList = list<Absyn.Algorithm>;
      type Absyn_PathOption = Option<Absyn.Path>;
      Env_FrameList env_1,env;
      Algorithm_StatementList statements_1;
      Connect_SetList csets;
      ClassInf.State ci_state;
      Absyn_AlgorithmList statements;
      Absyn_PathOption bc;
      Boolean impl;
    case (env,_,_,csets,ci_state,SCode.ALGORITHM(statements,bc),impl) "impl"
      equation 
        env_1 = get_derived_env(env, bc) "If algorithm is inherited, find base class environment";
        statements_1 = inst_statements(env_1, statements, impl); then ({DAE.ALGORITHM(Algorithm.ALGORITHM(statements_1))},env,csets,ci_state);
    case (_,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_algorithm failed\n"); then fail();
  end matchcontinue;
end inst_algorithm;

protected function inst_initialalgorithm "impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input SCode.Algorithm in_algorithm;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_dae_elementlist,out_env,out_sets,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_algorithm,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type Algorithm_StatementList = list<Algorithm.Statement>;
      type Connect_SetList = list<Connect.Set>;
      type Absyn_AlgorithmList = list<Absyn.Algorithm>;
      type Absyn_PathOption = Option<Absyn.Path>;
      Env_FrameList env_1,env;
      Algorithm_StatementList statements_1;
      Connect_SetList csets;
      ClassInf.State ci_state;
      Absyn_AlgorithmList statements;
      Absyn_PathOption bc;
      Boolean impl;
    case (env,_,_,csets,ci_state,SCode.ALGORITHM(statements,bc),impl) "impl"
      equation 
        env_1 = get_derived_env(env, bc);
        statements_1 = inst_statements(env, statements, impl); then ({DAE.INITIALALGORITHM(Algorithm.ALGORITHM(statements_1))},env,csets,ci_state);
    case (_,_,_,_,_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_algorithm failed\n"); then fail();
  end matchcontinue;
end inst_initialalgorithm;

protected function inst_statements
  input Env in_env;
  input Absyn_AlgorithmList in_absyn_algorithmlist;
  input Boolean in_boolean;
  output Algorithm_StatementList out_algorithm_statementlist;
protected 
  type Absyn_AlgorithmList = list<Absyn.Algorithm>;
  type Algorithm_StatementList = list<Algorithm.Statement>;
algorithm 
  out_algorithm_statementlist:=
  matchcontinue (in_env,in_absyn_algorithmlist,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList env;
      Boolean impl;
      Algorithm.Statement x_1;
      Algorithm_StatementList xs_1;
      Absyn.Algorithm x;
      Absyn_AlgorithmList xs;
    case (env,{},impl) then {}; 
    case (env,x :: xs,impl)
      equation 
        x_1 = inst_statement(env, x, impl);
        xs_1 = inst_statements(env, xs, impl); then x_1 :: xs_1;
  end matchcontinue;
end inst_statements;

protected function inst_algorithmitems
  input Env in_env;
  input Absyn_AlgorithmItemList in_absyn_algorithmitemlist;
  input Boolean in_boolean;
  output Algorithm_StatementList out_algorithm_statementlist;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
  type Algorithm_StatementList = list<Algorithm.Statement>;
algorithm 
  out_algorithm_statementlist:=
  matchcontinue (in_env,in_absyn_algorithmitemlist,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList env;
      Boolean impl;
      Algorithm.Statement x_1;
      Algorithm_StatementList xs_1;
      Absyn.Algorithm x;
      Absyn_AlgorithmItemList xs;
    case (env,{},impl) then {}; 
    case (env,Absyn.ALGORITHMITEM(x,_) :: xs,impl)
      equation 
        x_1 = inst_statement(env, x, impl);
        xs_1 = inst_algorithmitems(env, xs, impl); then x_1 :: xs_1;
    case (env,Absyn.ALGORITHMITEMANN(_) :: xs,impl)
      equation 
        xs_1 = inst_algorithmitems(env, xs, impl); then xs_1;
  end matchcontinue;
end inst_algorithmitems;

protected function inst_statement " relation: inst_statement
 
  This function Looks at an algorithm statement and uses functions
  in the `Algorithm' module to build a representation of it that can
  be used in the DAE output.
 "
  input Env in_env;
  input Absyn.Algorithm in_algorithm;
  input Boolean in_boolean;
  output Algorithm.Statement out_statement;
algorithm 
  out_statement:=
  matchcontinue (in_env,in_algorithm,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type Exp_ExpList = list<Exp.Exp>;
      type Types_PropertiesList = list<Types.Properties>;
      type Absyn_ExpList = list<Absyn.Exp>;
      type Algorithm_StatementList = list<Algorithm.Statement>;
      type TupleExp_ExpTypes_PropertiesAlgorithm_StatementList = tuple<Exp.Exp,Types.Properties,Algorithm_StatementList>;
      type TupleExp_ExpTypes_PropertiesAlgorithm_StatementListList = list<TupleExp_ExpTypes_PropertiesAlgorithm_StatementList>;
      type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
      type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
      type TupleAbsyn_ExpAbsyn_AlgorithmItemListList = list<TupleAbsyn_ExpAbsyn_AlgorithmItemList>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Exp.ComponentRef ce,ce_1;
      Exp.Type t;
      Types.Properties cprop,eprop,prop,msgprop;
      SCode.Accessibility acc;
      Exp.Exp e_1,cond_1,msg_1;
      Algorithm.Statement stmt;
      Env_FrameList env,env_1;
      Absyn.ComponentRef cr;
      Absyn.Exp e,cond,msg;
      Boolean impl;
      Exp_ExpList expl_1;
      Types_PropertiesList cprops;
      Absyn_ExpList expl;
      Algorithm_StatementList tb_1,fb_1,sl_1;
      TupleExp_ExpTypes_PropertiesAlgorithm_StatementListList eib_1;
      Absyn_AlgorithmItemList tb,fb,sl;
      TupleAbsyn_ExpAbsyn_AlgorithmItemListList eib,el;
      String i;
      Absyn.Algorithm alg;
    case (env,Absyn.ALG_ASSIGN(cr,e),impl) "impl"
      equation 
        (Exp.CREF(ce,t),cprop,acc) = Static.elab_cref(env, cr, impl);
        ce_1 = Static.canon_cref(env, ce, impl);
        (e_1,eprop,_) = Static.elab_exp(env, e, impl, NONE);
        stmt = Algorithm.make_assignment(Exp.CREF(ce_1,t), cprop, e_1, eprop, acc); then stmt;
    case (env,Absyn.ALG_ASSIGN(cr,e),impl)
      local Exp.Exp ce;
      equation 
        (ce,cprop,acc) = Static.elab_cref(env, cr, impl);
        (e_1,eprop,_) = Static.elab_exp(env, e, impl, NONE);
        stmt = Algorithm.make_assignment(ce, cprop, e_1, eprop, acc); then stmt;
    case (env,Absyn.ALG_TUPLE_ASSIGN(Absyn.TUPLE(expl),e),impl)
      equation 
        (e_1 = Exp.CALL(_,_,_,_),eprop,_) = Static.elab_exp(env, e, impl, NONE);
        (expl_1,cprops,_) = Static.elab_exp_list(env, expl, impl, NONE);
        stmt = Algorithm.make_tuple_assignment(expl_1, cprops, e_1, eprop); then stmt;
    case (env,Absyn.ALG_TUPLE_ASSIGN(Absyn.TUPLE(expl),e),impl)
      equation 
        Print.print_error_buf(
          "#Error, tuple assignment only allowed when rhs is function call.\n"); then fail();
    case (env,Absyn.ALG_IF(e,tb,eib,fb),impl)
      equation 
        (e_1,prop,_) = Static.elab_exp(env, e, impl, NONE);
        tb_1 = inst_algorithmitems(env, tb, impl);
        eib_1 = inst_elseifs(env, eib, impl);
        fb_1 = inst_algorithmitems(env, fb, impl);
        stmt = Algorithm.make_if(e_1, prop, tb_1, eib_1, fb_1); then stmt;
    case (env,Absyn.ALG_FOR(i,e,sl),impl)
      local TupleTypes_TTypeAbsyn_PathOption t;
      equation 
        (e_1,prop = Types.PROP((Types.T_ARRAY(_,t),_),_),_) = Static.elab_exp(env, e, impl, NONE);
        env_1 = add_for_loop_scope(env, i, t);
        sl_1 = inst_algorithmitems(env_1, sl, impl);
        stmt = Algorithm.make_for(i, e_1, prop, sl_1); then stmt;
    case (env,Absyn.ALG_WHILE(e,sl),impl)
      equation 
        (e_1,prop,_) = Static.elab_exp(env, e, impl, NONE);
        sl_1 = inst_algorithmitems(env, sl, impl);
        stmt = Algorithm.make_while(e_1, prop, sl_1); then stmt;
    case (env,Absyn.ALG_WHEN_A(e,sl,el),impl)
      equation 
        (e_1,prop,_) = Static.elab_exp(env, e, impl, NONE);
        sl_1 = inst_algorithmitems(env, sl, impl);
        stmt = Algorithm.make_when_a(e_1, prop, sl_1) "TODO elsewhen "; then stmt;
    case (env,Absyn.ALG_NORETCALL(Absyn.CREF_IDENT("assert",_),Absyn.FUNCTIONARGS({cond,msg},{})),impl) "1.0
 rule	Static.elab_exp (env,e,impl,NONE) => (e',prop,_) &
	inst_algorithmitems(env,sl,impl) => sl' &
	Algorithm.make_when(e',prop,sl') => stmt
	---------------------------------------
	inst_statement(env,Absyn.ALG_WHEN(e,sl),impl) => stmt
"
      equation 
        (cond_1,cprop,_) = Static.elab_exp(env, cond, impl, NONE);
        (msg_1,msgprop,_) = Static.elab_exp(env, msg, impl, NONE);
        stmt = Algorithm.make_assert(cond_1, msg_1, cprop, msgprop); then stmt;
    case (env,alg,impl)
      equation 
        Debug.fprint("failtrace", "- inst_statement failed\n alg:");
        Debug.fcall("failtrace", Dump.print_algorithm, alg);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_statement;

protected function inst_elseifs
  input Env.Env in_env;
  input TupleAbsyn_ExpAbsyn_AlgorithmItemListList in_tupleabsyn_expabsyn_algorithmitemlistlist;
  input Boolean in_boolean;
  output TupleExp_ExpTypes_PropertiesAlgorithm_StatementListList out_tupleexp_exptypes_propertiesalgorithm_statementlistlist;
protected 
  type Absyn_AlgorithmItemList = list<Absyn.AlgorithmItem>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemList = tuple<Absyn.Exp,Absyn_AlgorithmItemList>;
  type TupleAbsyn_ExpAbsyn_AlgorithmItemListList = list<TupleAbsyn_ExpAbsyn_AlgorithmItemList>;
  type Algorithm_StatementList = list<Algorithm.Statement>;
  type TupleExp_ExpTypes_PropertiesAlgorithm_StatementList = tuple<Exp.Exp,Types.Properties,Algorithm_StatementList>;
  type TupleExp_ExpTypes_PropertiesAlgorithm_StatementListList = list<TupleExp_ExpTypes_PropertiesAlgorithm_StatementList>;
algorithm 
  out_tupleexp_exptypes_propertiesalgorithm_statementlistlist:=
  matchcontinue (in_env,in_tupleabsyn_expabsyn_algorithmitemlistlist,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      Env_FrameList env;
      Boolean impl;
      Exp.Exp e_1;
      Types.Properties prop;
      Algorithm_StatementList stmts;
      TupleExp_ExpTypes_PropertiesAlgorithm_StatementListList tail_1;
      Absyn.Exp e;
      Absyn_AlgorithmItemList l;
      TupleAbsyn_ExpAbsyn_AlgorithmItemListList tail;
    case (env,{},impl) then {}; 
    case (env,(e,l) :: tail,impl)
      equation 
        (e_1,prop,_) = Static.elab_exp(env, e, impl, NONE);
        stmts = inst_algorithmitems(env, l, impl);
        tail_1 = inst_elseifs(env, tail, impl); then (e_1,prop,stmts) :: tail_1;
    case (_,_,_)
      equation 
        Debug.fprint("failtrace", "- inst_elseifs failed\n"); then fail();
  end matchcontinue;
end inst_elseifs;

protected function inst_connect "impl"
  input Connect.Sets in_sets1;
  input Env in_env2;
  input Prefix in_prefix3;
  input Absyn.ComponentRef in_componentref4;
  input Absyn.ComponentRef in_componentref5;
  input Boolean in_boolean6;
  output Connect.Sets out_sets;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_sets,out_dae_elementlist):=
  matchcontinue (in_sets1,in_env2,in_prefix3,in_componentref4,in_componentref5,in_boolean6)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Connect_SetList = list<Connect.Set>;
      type Env_FrameList = list<Env.Frame>;
      Exp.ComponentRef c1_1,c2_1,c1_2,c2_2;
      Exp.Type t1,t2;
      Types.Properties prop1,prop2;
      SCode.Accessibility acc;
      Types.Attributes attr1,attr2;
      Boolean flow1,impl;
      TupleTypes_TTypeAbsyn_PathOption ty1,ty2;
      Connect.Face f1,f2;
      Connect_SetList sets_1,sets;
      DAE_ElementList dae;
      Env_FrameList env;
      Prefix.Prefix pre;
      Absyn.ComponentRef c1,c2;
    case (sets,env,pre,c1,c2,impl) "impl"
      equation 
        (Exp.CREF(c1_1,t1),prop1,acc) = Static.elab_cref(env, c1, impl);
        (Exp.CREF(c2_1,t2),prop2,acc) = Static.elab_cref(env, c2, impl);
        c1_2 = Static.canon_cref(env, c1_1, impl);
        c2_2 = Static.canon_cref(env, c2_1, impl);
        (attr1 = Types.ATTR(flow1,_,_,_),ty1,_) = Lookup.lookup_var_local(env, c1_2);
        (attr2,ty2,_) = Lookup.lookup_var(env, c2_2);
        valid_connector(ty1) " Check that the types of the connectors are good. ";
        valid_connector(ty2);
        check_connect_types(c1_2, ty1, attr1, c2_2, ty2, attr2);
        f1 = component_face(c1_2);
        f2 = component_face(c2_2);
        (sets_1,dae) = connect_components(sets, env, pre, c1_2, f1, ty1, c2_2, f2, ty2, flow1); then (sets_1,dae);
    case (sets,_,_,_,_,impl)
      equation 
        Print.print_error_buf("# could not connect\n"); then fail();
  end matchcontinue;
end inst_connect;

protected function valid_connector " relation: valid_connector
 
  This function tests whether a type is a eligible to be used in
  connections.
 
 "
  input Types.Type in_type;
algorithm 
  _:=
  matchcontinue (in_type)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      ClassInf.State state;
      String str;
      TupleTypes_TTypeAbsyn_PathOption t;
    case ((Types.T_REAL(_),_)) then (); 
    case ((Types.T_COMPLEX(state,_),_))
      equation 
        ClassInf.valid(state, SCode.R_CONNECTOR); then ();
    case t
      equation 
        Print.print_error_buf("# Can't connect objects of type ");
        str = Types.unparse_type(t);
        Print.print_error_buf(str);
        Print.print_buf("\n"); then fail();
  end matchcontinue;
end valid_connector;

protected function check_connect_types
  input Exp.ComponentRef in_componentref1;
  input Types.Type in_type2;
  input Types.Attributes in_attributes3;
  input Exp.ComponentRef in_componentref4;
  input Types.Type in_type5;
  input Types.Attributes in_attributes6;
algorithm 
  _:=
  matchcontinue (in_componentref1,in_type2,in_attributes3,in_componentref4,in_type5,in_attributes6)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      String str;
      Exp.ComponentRef c1,c2;
      TupleTypes_TTypeAbsyn_PathOption t1,t2;
      Boolean flow1,flow2;
    case (c1,_,Types.ATTR(_,_,_,Absyn.INPUT),c2,_,Types.ATTR(_,_,_,Absyn.INPUT))
      equation 
        assert_different_faces(c1, c2);
        Print.print_error_buf("# Can't connect two input variables\n");
        Print.print_error_buf("    while connecting ");
        str = Exp.print_component_ref_str(c1);
        Print.print_error_buf(str);
        Print.print_error_buf(" to ");
        str = Exp.print_component_ref_str(c2);
        Print.print_error_buf(str);
        Print.print_error_buf(
          "\n unless one of them is inside connector and the other outside connector."); then fail();
    case (c1,_,Types.ATTR(_,_,_,Absyn.OUTPUT),c2,_,Types.ATTR(_,_,_,Absyn.OUTPUT))
      equation 
        assert_different_faces(c1, c2);
        Print.print_error_buf("# Can't connect two output variables\n");
        Print.print_error_buf("    while connecting ");
        str = Exp.print_component_ref_str(c1);
        Print.print_error_buf(str);
        Print.print_error_buf(" to ");
        str = Exp.print_component_ref_str(c2);
        Print.print_error_buf(str);
        Print.print_error_buf(
          "\n unless one of them is inside connector and the other outside connector."); then fail();
    case (_,t1,Types.ATTR(flow1,_,_,_),_,t2,Types.ATTR(flow2,_,_,_))
      equation 
        equality(flow1 = flow2);
        true = Types.equivtypes(t1, t2); then ();
    case (c1,_,Types.ATTR(true,_,_,_),c2,_,Types.ATTR(false,_,_,_))
      equation 
        Print.print_error_buf("# Can't connect flow component ");
        str = Exp.print_component_ref_str(c1);
        Print.print_error_buf(str);
        Print.print_error_buf(" to non-flow component ");
        str = Exp.print_component_ref_str(c2);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
    case (c1,_,Types.ATTR(false,_,_,_),c2,_,Types.ATTR(true,_,_,_))
      equation 
        Print.print_error_buf("# Can't connect non-flow component ");
        str = Exp.print_component_ref_str(c1);
        Print.print_error_buf(str);
        Print.print_error_buf(" to flow component ");
        str = Exp.print_component_ref_str(c2);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
    case (c1,_,_,c2,_,_)
      equation 
        Debug.fprint("failtrace", "- check_connect_types(");
        Debug.fcall("failtrace", Exp.print_component_ref, c1);
        Debug.fprint("failtrace", " <-> ");
        Debug.fcall("failtrace", Exp.print_component_ref, c2);
        Debug.fprint("failtrace", ") failed\n"); then fail();
  end matchcontinue;
end check_connect_types;

protected function assert_different_faces " relation assert_different_faces
 
  This function fails if two connectors have same faces, 
  e.g both inside or both outside connectors 
 "
  input Exp.ComponentRef in_componentref1;
  input Exp.ComponentRef in_componentref2;
algorithm 
  _:=
  matchcontinue (in_componentref1,in_componentref2)
    local Exp.ComponentRef c1,c2;
    case (c1,c2)
      equation 
        Connect.INNER = component_face(c1);
        Connect.OUTER = component_face(c1); then ();
    case (c1,c2)
      equation 
        Connect.OUTER = component_face(c1);
        Connect.INNER = component_face(c1); then ();
  end matchcontinue;
end assert_different_faces;

protected function connect_components
  input Connect.Sets in_sets1;
  input Env.Env in_env2;
  input Prefix in_prefix3;
  input Exp.ComponentRef in_componentref4;
  input Connect.Face in_face5;
  input Types.Type in_type6;
  input Exp.ComponentRef in_componentref7;
  input Connect.Face in_face8;
  input Types.Type in_type9;
  input Boolean in_boolean10;
  output Connect.Sets out_sets;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_sets,out_dae_elementlist):=
  matchcontinue (in_sets1,in_env2,in_prefix3,in_componentref4,in_face5,in_type6,in_componentref7,in_face8,in_type9,in_boolean10)
    local
      type Connect_SetList = list<Connect.Set>;
      type Env_FrameList = list<Env.Frame>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      type Types_VarList = list<Types.Var>;
      Exp.ComponentRef c1_1,c2_1,c1,c2;
      Connect_SetList sets_1,sets;
      Env_FrameList env;
      Prefix.Prefix pre;
      Connect.Face f1,f2;
      TupleTypes_TTypeAbsyn_PathOption t1,t2;
      SCode.Variability vr;
      Integer dim1,dim2;
      DAE_ElementList dae;
      Types_VarList l1,l2;
      String str;
    case (sets,env,pre,c1,f1,(Types.T_REAL(_),_),c2,f2,(Types.T_REAL(_),_),true) "  Flow type, must be a subtype of Real "
      equation 
        c1_1 = Prefix.prefix_cref(pre, c1);
        c2_1 = Prefix.prefix_cref(pre, c2);
        sets_1 = Connect.add_flow(sets, c1_1, f1, c2_1, f2); then (sets_1,{});
    case (sets,env,pre,c1,f1,(Types.T_ARRAY(_,t1),_),c2,f2,(Types.T_ARRAY(_,t2),_),true)
      equation 
        ((Types.T_REAL(_),_)) = Types.array_element_type(t1);
        ((Types.T_REAL(_),_)) = Types.array_element_type(t2);
        c1_1 = Prefix.prefix_cref(pre, c1);
        c2_1 = Prefix.prefix_cref(pre, c2);
        sets_1 = Connect.add_equ(sets, c1_1, c2_1); then (sets_1,{});
    case (sets,env,pre,c1,f1,(_,_),c2,f2,(_,_),false) "  Non-flow type   Parameters and constants generate assert statements "
      equation 
        c1_1 = Prefix.prefix_cref(pre, c1);
        c2_1 = Prefix.prefix_cref(pre, c2);
        (Types.ATTR(_,_,vr,_),t1,_) = Lookup.lookup_var_local(env, c1_1);
        true = SCode.is_parameter_or_const(vr);
        true = Types.basic_type(t1);
        (Types.ATTR(_,_,_,_),t2,_) = Lookup.lookup_var_local(env, c2_1);
        true = Types.basic_type(t2); then (sets,{
          DAE.ASSERT(
          Exp.CALL(Absyn.IDENT("assert"),
          {
          Exp.RELATION(Exp.CREF(c1_1,Exp.REAL),Exp.EQUAL(Exp.BOOL),
          Exp.CREF(c2_1,Exp.REAL)),Exp.SCONST("automatically generated from connect")},false,true))});
    case (sets,env,pre,c1,_,(Types.T_REAL(_),_),c2,_,(Types.T_REAL(_),_),false)
      equation 
        c1_1 = Prefix.prefix_cref(pre, c1);
        c2_1 = Prefix.prefix_cref(pre, c2);
        sets_1 = Connect.add_equ(sets, c1_1, c2_1); then (sets_1,{});
    case (sets,env,pre,c1,_,(Types.T_INTEGER(_),_),c2,_,(Types.T_INTEGER(_),_),false)
      equation 
        c1_1 = Prefix.prefix_cref(pre, c1);
        c2_1 = Prefix.prefix_cref(pre, c2);
        sets_1 = Connect.add_equ(sets, c1_1, c2_1); then (sets_1,{});
    case (sets,env,pre,c1,f1,(Types.T_ARRAY(Types.DIM(SOME(dim1)),t1),_),c2,f2,(Types.T_ARRAY(Types.DIM(SOME(dim2)),t2),_),false)
      equation 
        c1_1 = Prefix.prefix_cref(pre, c1);
        c2_1 = Prefix.prefix_cref(pre, c2);
        equality(dim1 = dim2);
        sets_1 = Connect.add_array_equ(sets, c1_1, c2_1, dim1); then (sets_1,{});
    case (sets,env,pre,c1,f1,(Types.T_COMPLEX(_,l1),_),c2,f2,(Types.T_COMPLEX(_,l2),_),_) "  Complex types "
      equation 
        c1_1 = Prefix.prefix_cref(pre, c1);
        c2_1 = Prefix.prefix_cref(pre, c2);
        (sets_1,dae) = connect_vars(sets, env, c1_1, f1, l1, c2_1, f2, l2); then (sets_1,dae);
    case (_,env,pre,c1,_,t1,c2,_,t2,_) "  Error "
      equation 
        c1_1 = Prefix.prefix_cref(pre, c1);
        c2_1 = Prefix.prefix_cref(pre, c2);
        Print.print_error_buf("# Wrong type in connection\n");
        Print.print_error_buf("  while making connectiion ");
        str = Exp.print_component_ref_str(c1);
        Print.print_error_buf(str);
        Print.print_error_buf(" <-> ");
        str = Exp.print_component_ref_str(c2);
        Print.print_error_buf(str);
        Print.print_error_buf("\n");
        Print.print_error_buf("  The type `");
        str = Types.unparse_type(t1);
        Print.print_error_buf(str);
        Print.print_error_buf("' is not allowed in connections\n"); then fail();
  end matchcontinue;
end connect_components;

protected function connect_vars
  input Connect.Sets in_sets1;
  input Env.Env in_env2;
  input Exp.ComponentRef in_componentref3;
  input Connect.Face in_face4;
  input Types_VarList in_types_varlist5;
  input Exp.ComponentRef in_componentref6;
  input Connect.Face in_face7;
  input Types_VarList in_types_varlist8;
  output Connect.Sets out_sets;
  output DAE_ElementList out_dae_elementlist;
protected 
  type Types_VarList = list<Types.Var>;
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_sets,out_dae_elementlist):=
  matchcontinue (in_sets1,in_env2,in_componentref3,in_face4,in_types_varlist5,in_componentref6,in_face7,in_types_varlist8)
    local
      type Connect_SetList = list<Connect.Set>;
      type Env_FrameList = list<Env.Frame>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Connect_SetList sets,sets_1,sets_2;
      Env_FrameList env;
      Exp.ComponentRef c1_1,c2_1,c1,c2;
      DAE_ElementList dae,dae2,dae_1;
      Connect.Face f1,f2;
      String n;
      Types.Attributes attr1,attr2;
      Boolean flow1,flow2;
      SCode.Variability vt1,vt2;
      TupleTypes_TTypeAbsyn_PathOption ty1,ty2;
      Types_VarList xs1,xs2;
    case (sets,env,_,_,{},_,_,{}) then (sets,{}); 
    case (sets,env,c1,f1,Types.VAR(n,(attr1 = Types.ATTR(flow1,_,vt1,_)),_,ty1,_) :: xs1,c2,f2,Types.VAR(_,(attr2 = Types.ATTR(flow2,_,vt2,_)),_,ty2,_) :: xs2)
      equation 
        c1_1 = Exp.extend_cref(c1, n, {});
        c2_1 = Exp.extend_cref(c2, n, {});
        check_connect_types(c1_1, ty1, attr1, c2_1, ty2, attr2);
        (sets_1,dae) = connect_components(sets, env, Prefix.NOPRE, c1_1, f1, ty1, c2_1, f2, ty2, 
          flow1);
        (sets_2,dae2) = connect_vars(sets_1, env, c1, f1, xs1, c2, f2, xs2);
        dae_1 = list_append(dae, dae2); then (sets_2,dae_1);
  end matchcontinue;
end connect_vars;

public function mktype " relation: mktype 
 
  From a class typename, its inference state, and a list of subcomponents, this
  relation returns `Types.Type'.  If the class inference state
  indicates that the type should be a built-in type, one of the
  built-in type constructors is used.  Otherwise, a `T_COMPLEX' is
  built.
 "
  input Absyn.Path in_path;
  input ClassInf.State in_state;
  input Types_VarList in_types_varlist;
  output Types.Type out_type;
protected 
  type Types_VarList = list<Types.Var>;
algorithm 
  out_type:=
  matchcontinue (in_path,in_state,in_types_varlist)
    local
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Absyn_PathOption somep;
      Absyn.Path p;
      Types_VarList v,vl,v1,l;
      TupleTypes_TTypeAbsyn_PathOption functype,enumtype;
      ClassInf.State st;
      String name;
    case (p,ClassInf.TYPE_INTEGER(_),v)
      equation 
        somep = get_opt_path(p); then ((Types.T_INTEGER(v),somep));
    case (p,ClassInf.TYPE_REAL(_),v)
      equation 
        somep = get_opt_path(p); then ((Types.T_REAL(v),somep));
    case (p,ClassInf.TYPE_STRING(_),v)
      equation 
        somep = get_opt_path(p); then ((Types.T_STRING(v),somep));
    case (p,ClassInf.TYPE_BOOL(_),v)
      equation 
        somep = get_opt_path(p); then ((Types.T_BOOL(v),somep));
    case (p,ClassInf.TYPE_ENUM(_),_)
      equation 
        somep = get_opt_path(p); then ((Types.T_ENUM,somep));
    case (p,(st = ClassInf.FUNCTION(name)),vl) "Insert function type construction here
	   after checking input/output arguments? 
	   see Types.rml T_FUNCTION "
      equation 
        functype = Types.make_function_type(p, vl); then functype;
    case (p,ClassInf.ENUMERATION(name),v1) "
  rule	Print.print_buf \" <function \" & Print.print_buf name & Print.print_buf \"> <\" &
	Dump.print_list (vl, Types.print_var, \", \") &
	Print.print_buf \" >\"
	--------------------------
	mktype(st as ClassInf.FUNCTION(name), vl) => Types.T_COMPLEX(st,vl)
"
      equation 
        enumtype = Types.make_enumeration_type(p, v1); then enumtype;
    case (p,st,l) "
  rule	Print.print_buf \" <function \" & Print.print_buf name & Print.print_buf \"> <\" &
	Dump.print_list (vl, Types.print_var, \", \") &
	Print.print_buf \" >\"
	--------------------------
	mktype(st as ClassInf.FUNCTION(name), vl) => Types.T_COMPLEX(st,vl)
"
      equation 
        somep = get_opt_path(p); then ((Types.T_COMPLEX(st,l),somep));
  end matchcontinue;
end mktype;

protected function get_opt_path " relation: get_opt_path
  
  Helper function to mktype
  Transforms a Path into a Path option.
 "
  input Absyn.Path in_path;
  output Absyn_PathOption out_absyn_pathoption;
protected 
  type Absyn_PathOption = Option<Absyn.Path>;
algorithm 
  out_absyn_pathoption:=
  matchcontinue (in_path)
    local Absyn.Path p;
    case Absyn.IDENT("") then NONE; 
    case p then SOME(p); 
  end matchcontinue;
end get_opt_path;

protected function inst_list "impl
  impl"
  input Env in_env;
  input Mod in_mod;
  input Prefix in_prefix;
  input Connect.Sets in_sets;
  input ClassInf.State in_state;
  input FuncType1 in_functype1;
  input VType_aList in_vtype_alist;
  input Boolean in_boolean;
  output VType_bList out_vtype_blist;
  output Env out_env;
  output Connect.Sets out_sets;
  output ClassInf.State out_state;
protected 
  function FuncType1
    input Env in_env;
    input Mod in_mod;
    input Prefix in_prefix;
    input Connect.Sets in_sets;
    input ClassInf.State in_state;
    input Type_a in_type_a;
    input Boolean in_boolean;
    output VType_bList out_vtype_blist;
    output Env out_env;
    output Connect.Sets out_sets;
    output ClassInf.State out_state;
  protected 
    replaceable type Type_a;
    replaceable type Type_b;
    type VType_bList = list<Type_b>;
  end FuncType1;
  replaceable type Type_a;
  type VType_aList = list<Type_a>;
  replaceable type Type_b;
  type VType_bList = list<Type_b>;
algorithm 
  (out_vtype_blist,out_env,out_sets,out_state):=
  matchcontinue (in_env,in_mod,in_prefix,in_sets,in_state,in_functype1,in_vtype_alist,in_boolean)
    local
      type Env_FrameList = list<Env.Frame>;
      type Connect_SetList = list<Connect.Set>;
      function FuncType1
        input Env_FrameList in_env_framelist;
        input Types.Mod in_mod;
        input Prefix.Prefix in_prefix;
        input Connect_SetList in_connect_setlist;
        input ClassInf.State in_state;
        input Type_a in_type_a;
        input Boolean in_boolean;
        output VType_bList out_vtype_blist;
        output Env_FrameList out_env_framelist;
        output Connect_SetList out_connect_setlist;
        output ClassInf.State out_state;
      end FuncType1;
      Env_FrameList env,env_1,env_2;
      Types.Mod mod;
      Prefix.Prefix pre;
      Connect_SetList csets,csets_1,csets_2;
      ClassInf.State ci_state,ci_state_1,ci_state_2;
      FuncType1 r;
      Boolean impl;
      VType_bList l,l_1,l_2;
      Type_a e;
      VType_aList es;
    case (env,mod,pre,csets,ci_state,r,{},impl) then ({},env,csets,ci_state); 
    case (env,mod,pre,csets,ci_state,r,e :: es,impl)
      equation 
        (l,env_1,csets_1,ci_state_1) = r(env, mod, pre, csets, ci_state, e, impl);
        (l_1,env_2,csets_2,ci_state_2) = inst_list(env_1, mod, pre, csets_1, ci_state_1, r, es, impl);
        l_2 = list_append(l, l_1); then (l_2,env_2,csets_2,ci_state_2);
  end matchcontinue;
end inst_list;

protected function component_face " relation: component_face
  
  This function determines whether a component reference refers to an
  inner or outer connector.
 "
  input Exp.ComponentRef in_componentref;
  output Connect.Face out_face;
algorithm 
  out_face:=
  matchcontinue (in_componentref)
    case Exp.CREF_QUAL(_,_,Exp.CREF_IDENT(_,_)) then Connect.INNER; 
    case Exp.CREF_QUAL(_,_,Exp.CREF_QUAL(_,_,_)) then Connect.INNER; 
    case Exp.CREF_IDENT(_,_) then Connect.OUTER; 
  end matchcontinue;
end component_face;

protected function inst_binding " relation: inst_binding
 
  This function investigates a modification and extracts the 
  <...> modification. E.g. Real x(<...>=1+3) => 1+3
  It also handles the case Integer T0{2}(final <...>={5,6})={9,10} becomes
  Integer T0{1}(<...>=5); Integer T0{2}(<...>=6);
 
  Arg 1 is the modification  
  Arg 2 is the expected type that the modification should have
  Arg 3 is the index list for the element: for T0{1,2} is {1,2} 
 
 "
  input Types.Mod in_mod;
  input Types.Type in_type;
  input IntegerList in_integerlist;
  input String in_string;
  output Exp_ExpOption out_exp_expoption;
protected 
  type IntegerList = list<Integer>;
algorithm 
  out_exp_expoption:=
  matchcontinue (in_mod,in_type,in_integerlist,in_string)
    case (mod,expected_type,{},bind_name)
      equation 
        mod2 = Mod.lookup_comp_modification(mod, bind_name);
        SOME(Types.TYPED(e,Types.PROP(ty2,c))) = Mod.mod_equation(mod2);
        (e_1,ty_1) = Types.match_type(e, ty2, expected_type); then SOME(e_1);
    case (mod,etype,index_list,bind_name) "maybe this should be index::{}"
      equation 
        mod2 = Mod.lookup_comp_modification(mod, bind_name);
        result = inst_binding2(mod2, etype, index_list, bind_name); then result;
    case (mod,expected_type,{},bind_name)
      equation 
        failure(mod2 = Mod.lookup_comp_modification(mod, bind_name)); then NONE;
    case (mod,etype,_,_) then NONE; 
  end matchcontinue;
end inst_binding;

protected function inst_binding2 " relation: inst_binding2
 
  This function investigates a modification and extracts the <...> modification if
  the modification is in array of components. 
  Help-relation to inst_binding
 "
  input Types.Mod in_mod;
  input Types.Type in_type;
  input IntegerList in_integerlist;
  input String in_string;
  output Exp_ExpOption out_exp_expoption;
protected 
  type IntegerList = list<Integer>;
algorithm 
  out_exp_expoption:=
  matchcontinue (in_mod,in_type,in_integerlist,in_string)
    case (mod,etype,index :: {},bind_name) "maybe this should be index::{}"
      equation 
        mod2 = Mod.lookup_idx_modification(mod, index) "this rule handles the case where there is only one element in the index-list";
        SOME(Types.TYPED(e,Types.PROP(ty2,c))) = Mod.mod_equation(mod2);
        (e_1,ty_1) = Types.match_type(e, ty2, etype); then SOME(e_1);
    case (mod,etype,index :: res,bind_name) "maybe this should be index::{}"
      equation 
        mod2 = Mod.lookup_idx_modification(mod, index) "this rule handles the case where there is several elements in the index-list";
        result = inst_binding2(mod2, etype, res, bind_name); then result;
    case (mod,etype,index :: res,bind_name) "maybe this should be index::{}"
      equation 
        failure(mod2 = Mod.lookup_idx_modification(mod, index)); then NONE;
    case (_,_,_,_) then fail(); 
  end matchcontinue;
end inst_binding2;

protected function inst_start_binding_exp " relation: inst_start_binding_exp
 
  This function investigates a modification and extracts the 
  start modification. E.g. Real x(start=1+3) => 1+3
  It also handles the case Integer T0{2}(final start={5,6})={9,10} becomes
  Integer T0{1}(start=5); Integer T0{2}(start=6);
 
  Arg 1 is the start modification  
  Arg 2 is the expected type that the modification should have
  Arg 3 is the index list for the element: for T0{1,2} is {1,2} 
 "
  input Types.Mod mod;
  input Types.Type etype;
  input IntegerList index_list;
  output DAE.StartValue result;
protected 
  type IntegerList = list<Integer>;
algorithm 
  result := inst_binding(mod, etype, index_list, "start");
end inst_start_binding_exp;

protected function inst_dae_variable_attributes " relation: inst_dae_variable_attributes 
 
  this function extracts the attributes from the modification
  It returns a DAE.VariableAttributes option because 
  somtimes a varible does not contain the variable-attr
 "
  input Types.Mod in_mod;
  input Types.Type in_type;
  input IntegerList in_integerlist;
  output DAE_VariableAttributesOption out_dae_variableattributesoption;
protected 
  type IntegerList = list<Integer>;
  type DAE_VariableAttributesOption = Option<DAE.VariableAttributes>;
algorithm 
  out_dae_variableattributesoption:=
  matchcontinue (in_mod,in_type,in_integerlist)
    case (mod,(Types.T_REAL(_),path),index_list)
      equation 
        quantity_str = inst_string_binding(mod, index_list, "quantity");
        unit_str = inst_string_binding(mod, index_list, "unit");
        displayunit_str = inst_string_binding(mod, index_list, "displayUnit");
        min_val = inst_real_binding(mod, index_list, "min");
        max_val = inst_real_binding(mod, index_list, "max");
        start_val = inst_real_binding(mod, index_list, "start");
        fixed_val = inst_bool_binding(mod, index_list, "fixed");
        nominal_val = inst_real_binding(mod, index_list, "nominal");
        exp_bind_select = inst_enumeration_binding(mod, index_list, "stateSelect");
        stateSelect_value = get_stateSelect_from_exp_option(exp_bind_select); then SOME(
          DAE.VAR_ATTR_REAL(quantity_str,unit_str,displayunit_str,(min_val,max_val),
          start_val,fixed_val,nominal_val,stateSelect_value));
    case (mod,(Types.T_INTEGER(_),_),index_list)
      equation 
        quantity_str = inst_string_binding(mod, index_list, "quantity");
        min_val = inst_int_binding(mod, index_list, "min");
        max_val = inst_int_binding(mod, index_list, "max");
        start_val = inst_int_binding(mod, index_list, "start");
        fixed_val = inst_bool_binding(mod, index_list, "fixed"); then SOME(
          DAE.VAR_ATTR_INT(quantity_str,(min_val,max_val),start_val,fixed_val));
    case (mod,(Types.T_BOOL(_),_),index_list)
      equation 
        quantity_str = inst_string_binding(mod, index_list, "quantity");
        start_val = inst_bool_binding(mod, index_list, "start");
        fixed_val = inst_bool_binding(mod, index_list, "fixed"); then SOME(DAE.VAR_ATTR_BOOL(quantity_str,start_val,fixed_val));
    case (mod,(Types.T_STRING(_),_),index_list)
      equation 
        quantity_str = inst_string_binding(mod, index_list, "quantity");
        start_val = inst_string_binding(mod, index_list, "start"); then SOME(DAE.VAR_ATTR_STRING(quantity_str,start_val));
    case (mod,(enumtype = (Types.T_ENUMERATION(_,_),_)),index_list)
      equation 
        quantity_str = inst_string_binding(mod, index_list, "quantity");
        exp_bind_min = inst_binding(mod, enumtype, index_list, "min");
        exp_bind_max = inst_binding(mod, enumtype, index_list, "max");
        exp_bind_start = inst_binding(mod, enumtype, index_list, "start");
        fixed_val = inst_bool_binding(mod, index_list, "fixed"); then SOME(
          DAE.VAR_ATTR_ENUMERATION(quantity_str,(exp_bind_min,exp_bind_max),exp_bind_start,
          fixed_val));
    case (mod,_,_) then NONE; 
  end matchcontinue;
end inst_dae_variable_attributes;

protected function inst_bool_binding " relation inst_bool_binding
  author: LP
 
  instantiates a bool binding and retrieves the value.
  FIXME: check the type of variable for the fixed because there is a difference between 
 parameters and variables
 "
  input Types.Mod in_mod;
  input IntegerList in_integerlist;
  input String in_string;
  output BooleanOption out_booleanoption;
protected 
  type IntegerList = list<Integer>;
  type BooleanOption = Option<Boolean>;
algorithm 
  out_booleanoption:=
  matchcontinue (in_mod,in_integerlist,in_string)
    case (mod,index_list,bind_name)
      equation 
        SOME(Exp.BCONST(result)) = inst_binding(mod, (Types.T_BOOL({}),NONE), index_list, bind_name); then SOME(result);
    case (mod,index_list,bind_name)
      equation 
        NONE = inst_binding(mod, (Types.T_BOOL({}),NONE), index_list, bind_name); then NONE;
    case (mod,index_list,_)
      equation 
        Print.print_error_buf("# Wrong type on predefined.\n"); then fail();
  end matchcontinue;
end inst_bool_binding;

protected function inst_real_binding " relation: inst_real_binding
  author: LP
 
  instantiates a real binding and retrieves the value.
 "
  input Types.Mod in_mod;
  input IntegerList in_integerlist;
  input String in_string;
  output RealOption out_realoption;
protected 
  type IntegerList = list<Integer>;
  type RealOption = Option<Real>;
algorithm 
  out_realoption:=
  matchcontinue (in_mod,in_integerlist,in_string)
    case (mod,index_list,bind_name)
      equation 
        SOME(Exp.RCONST(result)) = inst_binding(mod, (Types.T_REAL({}),NONE), index_list, bind_name); then SOME(result);
    case (mod,index_list,bind_name)
      equation 
        SOME(Exp.ICONST(result)) = inst_binding(mod, (Types.T_INTEGER({}),NONE), index_list, bind_name);
        r_result = int_real(result); then SOME(r_result);
    case (mod,index_list,bind_name)
      equation 
        NONE = inst_binding(mod, (Types.T_REAL({}),NONE), index_list, bind_name); then NONE;
    case (mod,index_list,_)
      equation 
        Print.print_error_buf("# Wrong type on predefined.\n"); then fail();
  end matchcontinue;
end inst_real_binding;

protected function inst_int_binding " relation: inst_int_binding
  author: LP
 
  instantiates an int binding and retrieves the value.
 "
  input Types.Mod in_mod;
  input IntegerList in_integerlist;
  input String in_string;
  output IntegerOption out_integeroption;
protected 
  type IntegerList = list<Integer>;
  type IntegerOption = Option<Integer>;
algorithm 
  out_integeroption:=
  matchcontinue (in_mod,in_integerlist,in_string)
    case (mod,index_list,bind_name)
      equation 
        SOME(Exp.ICONST(result)) = inst_binding(mod, (Types.T_INTEGER({}),NONE), index_list, bind_name); then SOME(result);
    case (mod,index_list,bind_name)
      equation 
        NONE = inst_binding(mod, (Types.T_INTEGER({}),NONE), index_list, bind_name); then NONE;
    case (mod,index_list,_)
      equation 
        Print.print_error_buf("# Wrong type on predefined.\n"); then fail();
  end matchcontinue;
end inst_int_binding;

protected function inst_string_binding " relation: inst_string_binding
  author: LP
 
  instantiates a string binding and retrieves the value.
 "
  input Types.Mod in_mod;
  input IntegerList in_integerlist;
  input String in_string;
  output StringOption out_stringoption;
protected 
  type IntegerList = list<Integer>;
  type StringOption = Option<String>;
algorithm 
  out_stringoption:=
  matchcontinue (in_mod,in_integerlist,in_string)
    case (mod,index_list,bind_name)
      equation 
        SOME(Exp.SCONST(result)) = inst_binding(mod, (Types.T_STRING({}),NONE), index_list, bind_name); then SOME(result);
    case (mod,index_list,bind_name)
      equation 
        NONE = inst_binding(mod, (Types.T_STRING({}),NONE), index_list, bind_name); then NONE;
    case (mod,index_list,_)
      equation 
        Print.print_error_buf("# Wrong type on predefined.\n"); then fail();
  end matchcontinue;
end inst_string_binding;

protected function inst_enumeration_binding " relation: inst_enumeration_binding
  author: LP
 
  instantiates a enumeration binding and retrieves the value.
 "
  input Types.Mod in_mod;
  input IntegerList in_integerlist;
  input String in_string;
  output Exp_ExpOption out_exp_expoption;
protected 
  type IntegerList = list<Integer>;
algorithm 
  out_exp_expoption:=
  matchcontinue (in_mod,in_integerlist,in_string)
    case (mod,index_list,bind_name)
      equation 
        result = inst_binding(mod, (Types.T_ENUMERATION({},{}),NONE), index_list, 
          bind_name); then result;
    case (mod,index_list,_)
      equation 
        Print.print_error_buf("# Wrong type on predefined.\n"); then fail();
  end matchcontinue;
end inst_enumeration_binding;

protected function get_stateSelect_from_exp_option " relation: get_stateSelect_from_exp_option
  author: LP
 
  Retrieves the stateSelect value, as defined in DAE,  from an Expression option.
 "
  input Exp_ExpOption in_exp_expoption;
  output DAE_StateSelectOption out_dae_stateselectoption;
protected 
  type DAE_StateSelectOption = Option<DAE.StateSelect>;
algorithm 
  out_dae_stateselectoption:=
  matchcontinue (in_exp_expoption)
    case (SOME(Exp.CREF(Exp.CREF_QUAL("StateSelect",{},Exp.CREF_IDENT("never",{})),Exp.ENUM))) then SOME(DAE.NEVER); 
    case (SOME(Exp.CREF(Exp.CREF_QUAL("StateSelect",{},Exp.CREF_IDENT("avoid",{})),Exp.ENUM))) then SOME(DAE.AVOID); 
    case (SOME(Exp.CREF(Exp.CREF_QUAL("StateSelect",{},Exp.CREF_IDENT("default",{})),Exp.ENUM))) then SOME(DAE.DEFAULT); 
    case (SOME(Exp.CREF(Exp.CREF_QUAL("StateSelect",{},Exp.CREF_IDENT("prefer",{})),Exp.ENUM))) then SOME(DAE.PREFER); 
    case (SOME(Exp.CREF(Exp.CREF_QUAL("StateSelect",{},Exp.CREF_IDENT("always",{})),Exp.ENUM))) then SOME(DAE.ALWAYS); 
    case (NONE) then NONE; 
    case (_) then NONE; 
  end matchcontinue;
end get_stateSelect_from_exp_option;

protected function inst_mod_equation
  input Exp.ComponentRef in_componentref;
  input Types.Type in_type;
  input Mod in_mod;
  input Boolean in_boolean;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_componentref,in_type,in_mod,in_boolean)
    case (cr,ty1,(mod = Types.MOD(_,_,SOME(Types.TYPED(e,prop2)))),impl) "impl"
      equation 
        t = Types.elab_type(ty1);
        dae = inst_eq_equation(Exp.CREF(cr,t), Types.PROP(ty1,false), e, prop2, 
          NON_INITIAL, impl) "
	Debug.print (\"Debug: inst_mod_equation calling inst_eq_equation\n\") &
	fix_mod_subscript (cr, e, prop2) => (e', prop2') &
	inst_eq_equation(Exp.CREF(cr,t),
			 Types.PROP(ty1,false(*FIXME*)), e',prop2',NON_INITIAL)
 => dae
 FIXME"; then dae;
    case (_,_,Types.MOD(_,_,NONE),impl) then {}; 
    case (_,_,Types.NOMOD,impl) then {}; 
    case (_,_,Types.REDECL(_,_),impl) then {}; 
    case (c,t,m,impl)
      equation 
        Debug.fprint("failtrace", "- inst_mod_equation failed\n type: ");
        Debug.fcall("failtrace", Types.print_type, t);
        Debug.fprint("failtrace", "\n  cref: ");
        Debug.fcall("failtrace", Exp.print_component_ref, c);
        Debug.fprint("failtrace", "\n mod:");
        Debug.fcall("failtrace", Mod.print_mod, m);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_mod_equation;

protected function check_prot "might be needed from inst_mod_equation

relation fix_mod_subscript : (Exp.ComponentRef, Exp.Exp, Types.Properties) 
	  => (Exp.Exp, Types.Properties) = 

  rule	get_last_subscripts cr1 => subs &
	put_last_subscripts (cr2, subs) => cr2' &
	list_length subs => sublength &
	unlift_array_times (typrop, sublength) => typrop' &
	get_exp_type typrop' => ty2' &
	let newexp = Exp.CREF(cr2', ty2') &
	let newprop = Types.PROP(typrop', flg) 
	--------------------------------------
	fix_mod_subscript (cr1, Exp.CREF(cr2,ty2), Types.PROP(typrop,flg)) => (newexp, newprop)
end

relation unlift_array_times : (Types.Type, int) => Types.Type =
	
  axiom	unlift_array_times (ty, 0) => ty

  rule	Types.unlift_array ty => ty' &
	int_sub (cnt,1) => cnt' &
	unlift_array_times (ty', cnt') => ty''
	-------------------
	unlift_array_times (ty, cnt) => ty''
			 
end

relation get_exp_type : Types.Type => Exp.Type =

  axiom	get_exp_type Types.T_INTEGER(_) => Exp.INT
  axiom	get_exp_type Types.T_REAL(_) => Exp.REAL
  axiom	get_exp_type Types.T_STRING(_) => Exp.STRING
  axiom	get_exp_type Types.T_BOOL(_) => Exp.BOOL
  axiom	get_exp_type Types.T_ENUM => Exp.ENUM
  axiom	get_exp_type _ => Exp.OTHER

end


relation get_last_subscripts : Exp.ComponentRef => Exp.Subscript list =

  axiom	get_last_subscripts Exp.CREF_IDENT (n, subs) => subs

  rule	get_last_subscripts rest => subs
	--------------------------------
	get_last_subscripts Exp.CREF_QUAL (n, subs, rest) => subs
end

relation put_last_subscripts : (Exp.ComponentRef, Exp.Subscript list) 
	  => Exp.ComponentRef =

  axiom	put_last_subscripts (Exp.CREF_IDENT(n, subs), subs') 
	  => Exp.CREF_IDENT(n,subs') 

  rule	put_last_subscripts (rest,subs) => rest'
	--------------------------------
	put_last_subscripts (Exp.CREF_QUAL(n, s, rest), subs) => Exp.CREF_QUAL(n,s,rest')
end

   relation: check_prot
 
  This function is used to check that a protected element is not
  modified.
 "
  input Boolean in_boolean;
  input Types.Mod in_mod;
  input Exp.ComponentRef in_componentref;
algorithm 
  _:=
  matchcontinue (in_boolean,in_mod,in_componentref)
    local
      Exp.ComponentRef cref;
      String str;
    case (false,_,cref) then (); 
    case (_,Types.NOMOD,_) then (); 
    case (true,_,cref)
      equation 
        Print.print_error_buf("# Trying to modify protected element ");
        str = Exp.print_component_ref_str(cref);
        Print.print_error_buf(str);
        Print.print_error_buf("\n"); then fail();
  end matchcontinue;
end check_prot;

public function make_binding
  input Env.Env in_env;
  input SCode.Attributes in_attributes;
  input Types_EqModOption in_types_eqmodoption;
  output Types.Binding out_binding;
protected 
  type Types_EqModOption = Option<Types.EqMod>;
algorithm 
  out_binding:=
  matchcontinue (in_env,in_attributes,in_types_eqmodoption)
    local
      type Env_FrameList = list<Env.Frame>;
      type Absyn_PathOption = Option<Absyn.Path>;
      type TupleTypes_TTypeAbsyn_PathOption = tuple<Types.TType,Absyn_PathOption>;
      Env_FrameList env;
      Integer i;
      Real r;
      String s;
      Boolean b,c;
      Exp.Exp e,e_1;
      Types.Properties prop;
      TupleTypes_TTypeAbsyn_PathOption t;
    case (_,_,NONE) then Types.UNBOUND; 
    case (env,_,SOME(Types.TYPED(Exp.ICONST(i),_))) then Types.VALBOUND(Values.INTEGER(i)); 
    case (env,_,SOME(Types.TYPED(Exp.RCONST(r),_))) then Types.VALBOUND(Values.REAL(r)); 
    case (env,_,SOME(Types.TYPED(Exp.SCONST(s),_))) then Types.VALBOUND(Values.STRING(s)); 
    case (env,_,SOME(Types.TYPED(Exp.BCONST(b),_))) then Types.VALBOUND(Values.BOOL(b)); 
    case (env,SCode.ATTR(_,_,_,SCode.CONST,_),SOME(Types.TYPED(e,prop))) then Types.EQBOUND(e,true); 
    case (_,_,SOME(Types.TYPED(e,Types.PROP(t,c)))) "default "
      equation 
        e_1 = Exp.simplify(e); then Types.EQBOUND(e_1,c);
    case (_,_,_)
      equation 
        Debug.fprint("failtrace", "- make_binding failed\n"); then fail();
  end matchcontinue;
end make_binding;

public function init_vars_modelica_output " relation init_vars_modelica_output
  author: LS
 
  This rule goes through the elements and for each variable, searches the rest
  of the list for \"equations\" which refer to that variable on the LHS, and puts
  their RHS in the variable as the initialization expression. This is needed
  for modelica output where parameters must be \"assigned\" (?) during
  declaration.
 "
  input DAE_ElementList l;
  output DAE_ElementList l_1;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  l_1 := init_vars_modelica_output_1({}, l);
end init_vars_modelica_output;

protected function init_vars_modelica_output_1
  input DAE_ElementList in_dae_elementlist1;
  input DAE_ElementList in_dae_elementlist2;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  out_dae_elementlist:=
  matchcontinue (in_dae_elementlist1,in_dae_elementlist2)
    case (done,{}) then done; 
    case (done,(v = DAE.VAR(cr,vk,vd,ty,exp,inst_dims,start,flow_,class_,dae_var_attr,comment)) :: todorest)
      equation 
        (exp_1,done_1) = init_vars_modelica_output_2(cr, exp, done);
        (exp_2,todorest_1) = init_vars_modelica_output_2(cr, exp_1, todorest);
        done_2 = list_append(done_1, 
          {
          DAE.VAR(cr,vk,vd,ty,exp_2,inst_dims,start,flow_,class_,
          dae_var_attr,comment)});
        done_3 = init_vars_modelica_output_1(done_2, todorest_1); then done_3;
    case (done,DAE.COMP(n,DAE.DAE(dae)) :: rest)
      equation 
        dae_1 = init_vars_modelica_output(dae);
        done_1 = list_append(done, {DAE.COMP(n,DAE.DAE(dae_1))});
        done_2 = init_vars_modelica_output_1(done_1, rest); then done_2;
    case (done,DAE.FUNCTION(fpath,DAE.DAE(dae),ty) :: rest)
      equation 
        dae_1 = init_vars_modelica_output(dae);
        done_1 = list_append(done, {DAE.FUNCTION(fpath,DAE.DAE(dae_1),ty)});
        done_2 = init_vars_modelica_output_1(done_1, rest); then done_2;
    case (done,e :: rest)
      equation 
        done_1 = list_append(done, {e});
        done_2 = init_vars_modelica_output_1(done_1, rest); then done_2;
  end matchcontinue;
end init_vars_modelica_output_1;

protected function init_vars_modelica_output_2
  input Exp.ComponentRef in_componentref;
  input Exp_ExpOption in_exp_expoption;
  input DAE_ElementList in_dae_elementlist;
  output Exp_ExpOption out_exp_expoption;
  output DAE_ElementList out_dae_elementlist;
protected 
  type DAE_ElementList = list<DAE.Element>;
algorithm 
  (out_exp_expoption,out_dae_elementlist):=
  matchcontinue (in_componentref,in_exp_expoption,in_dae_elementlist)
    case (cr,exp,{}) then (exp,{}); 
    case (cr,exp,DAE.EQUATION(Exp.CREF(e1cr,_),exp_1) :: rest) "Exp.OTHER"
      equation 
        true = Exp.cref_equal(cr, e1cr);
        (exp_2,rest_1) = init_vars_modelica_output_2(cr, SOME(exp_1), rest); then (exp_2,rest_1);
    case (cr,exp,(e1 = DAE.EQUATION(Exp.CREF(e1cr,_),exp_1)) :: rest) "Exp.OTHER"
      equation 
        false = Exp.cref_equal(cr, e1cr);
        (exp_2,rest_1) = init_vars_modelica_output_2(cr, exp, rest); then (exp_2,e1 :: rest_1);
    case (excr,exp,e1 :: rest)
      equation 
        (exp_1,rest_1) = init_vars_modelica_output_2(excr, exp, rest); then (exp_1,e1 :: rest_1);
  end matchcontinue;
end init_vars_modelica_output_2;

public function inst_record_constructor_elt
  input Env.Env in_env;
  input SCode.Element in_element;
  input Boolean in_boolean;
  output Types.Var out_var;
algorithm 
  out_var:=
  matchcontinue (in_env,in_element,in_boolean)
    case (env,SCode.COMPONENT(id,_,repl,prot,(attr = SCode.ATTR(dim,f,acc,var,dir)),t,mod,bc,comment),impl) "impl"
      equation 
        Debug.fprint("recconst", "inst_record_constructor_elt called\n") "impl";
        (cl,cenv) = Lookup.lookup_class(env, t, true);
        Debug.fprint("recconst", "looked up class\n");
        mod_1 = Mod.elab_mod(env, Prefix.NOPRE, mod, impl);
        owncref = Absyn.CREF_IDENT(id,{});
        dimexp = elab_arraydim(env, owncref, dim, NONE, false, NONE);
        Debug.fprint("recconst", "calling inst_var\n");
        (_,_,tp_1) = inst_var(cenv, ClassInf.FUNCTION(""), mod_1, Prefix.NOPRE, {}, id, 
          cl, attr, dimexp, {}, {}, impl, comment);
        Debug.fprint("recconst", "Type of argument:");
        Debug.fcall("recconst", Types.print_type, tp_1);
        Debug.fprint("recconst", "\nMod=");
        Debug.fcall("recconst", Mod.print_mod, mod_1);
        eq = Mod.mod_equation(mod_1);
        bind = make_binding(env, attr, eq); then Types.VAR(id,Types.ATTR(f,acc,var,Absyn.INPUT),prot,tp_1,bind);
    case (env,elt,impl)
      equation 
        Debug.fprint("failtrace", "- inst_record_constructor_elt failed.,elt:");
        str = SCode.print_element_str(elt);
        Debug.fprint("failtrace", str);
        Debug.fprint("failtrace", "\n"); then fail();
  end matchcontinue;
end inst_record_constructor_elt;

protected function is_top_call " relation: is_top_call
  author: PA
 
  The topmost instantiation call is treated specially with for instance unconnected connectors.
  This function returns true if the CallingScope indicates the top call.
 "
  input CallingScope in_callingscope;
  output Boolean out_boolean;
algorithm 
  out_boolean:=
  matchcontinue (in_callingscope)
    case TOP_CALL then true; 
    case INNER_CALL then false; 
  end matchcontinue;
end is_top_call;
end Inst;

